unit doom32_combined_unit;

interface

uses
	messages,
	classes,
	directx,
	windows,
	mmsystem,
	sysutils,
	winsock;


(* Unit m_rnd interface *)


{
    m_random.h, m_random.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// Random number LUT.
//     
//-----------------------------------------------------------------------------

// Returns a number from 0 to 255,
// from a lookup table.
function M_Random: integer;

// As M_Random, but used only by the play simulation.
function P_Random: integer;

// Fix randoms for demos.
procedure M_ClearRandom;

var
  rndindex: integer;
  prndindex: integer;


(* Unit f_wipe interface *)


{
    f_wipe.h, f_wipe.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Mission start screen wipe/melt, special effects. 
// Mission begin melt/wipe screen special effect. 
//   
//----------------------------------------------------------------------------- 

function wipe_StartScreen(x, y, width, height: integer): integer;

function wipe_EndScreen(x, y, width, height: integer): integer;

function wipe_ScreenWipe(wipeno, x, y, width, height, ticks: integer): boolean;

type
  wipe_t = (
    // simple gradual pixel change for 8-bit only
    wipe_ColorXForm,
    // weird screen melt
    wipe_Melt,
    wipe_NUMWIPES
  );


(* Unit d_textur interface *)


{
    d_textur.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Typedefs related to to textures etc., 
//   isolated here to make it easier separating modules. 
//     
//----------------------------------------------------------------------------- 

//
// Flats?
//
// a pic is an unmasked block of pixels
type
  pic_t = record
    width : byte;
    height : byte;
    data : byte;
  end;
  Ppic_t = ^pic_t;


(* Unit i_music interface *)


//
//  MUSIC I/O
//
procedure I_InitMusic;
procedure I_ShutdownMusic;

// Volume.
procedure I_SetMusicVolume(volume: integer);

// PAUSE game handling.
procedure I_PauseSong(handle: integer);
procedure I_ResumeSong(handle: integer);

// Registers a song handle to song data.
function I_RegisterSong(data: pointer; size: integer): integer;

// Called by anything that wishes to start music.
//  plays a song, and when the song is done,
//  starts playing it again in an endless loop.
// Horrible thing to do, considering.
procedure I_PlaySong(handle: integer; looping: boolean);

// See above (register), then think backwards
procedure I_UnRegisterSong(handle: integer);

procedure I_ProcessMusic;


(* Unit m_fixed interface *)


{
    m_fixed.h, m_fixed.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Fixed point arithemtics, implementation. 
// 
//----------------------------------------------------------------------------- 

// 
// Fixed point, 32bit as 16.16. 
//

const
  FRACBITS = 16;
  FRACUNIT = 1 shl FRACBITS;

type
  fixed_t = integer;
  Pfixed_t = ^fixed_t;
  fixed_tArray = packed array[0..$FFFF] of fixed_t;
  Pfixed_tArray = ^fixed_tArray;

function FixedMul(a: fixed_t; b: fixed_t): fixed_t;

function FixedDiv(a: fixed_t; b: fixed_t): fixed_t;

function FixedDiv2(a: fixed_t; b: fixed_t): fixed_t;


(* Unit m_bbox interface *)


              

{
    m_bbox.h, m_bbox.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//    Nil. 
//     
//----------------------------------------------------------------------------- 

// Bounding box coordinate storage.
const
  BOXTOP = 0;
  BOXBOTTOM = 1;
  BOXLEFT = 2;
  BOXRIGHT = 3;

// Bounding box functions.
procedure M_ClearBox(box: Pfixed_tArray);

procedure M_AddToBox(box: Pfixed_tArray; x: fixed_t; y: fixed_t);



(* Unit i_midi interface *)


procedure I_PlayMidi(index: integer);

procedure I_StopMidi;

procedure AddMidiFileToPlayList(MidiFile: string);

procedure DeleteMidiFileFromPlayList(MidiFile: string);

procedure ClearMidiFilePlayList;

function I_IsMidiPlaying: boolean;

procedure I_ResumeMidi;

procedure I_PauseMidi;

function _mciGetErrorString(const code: LongWord): string;


(* Unit tables interface *)


{
    tables.h, tables.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details.
// 
// DESCRIPTION: 
//  Lookup tables. 
//  Do not try to look them up :-). 
//  In the order of appearance:  
// 
//  int finetangent[4096]  - Tangens LUT. 
//   Should work with BAM fairly well (12 of 16bit, 
//      effectively, by shifting). 
// 
//  int finesine[10240]    - Sine lookup. 
//   Guess what, serves as cosine, too. 
//   Remarkable thing is, how to use BAMs with this?  
// 
//  int tantoangle[2049]  - ArcTan LUT, 
//    maps tan(angle) to angle fast. Gotta search.   
//     
//-----------------------------------------------------------------------------

              

const PI = 3.141592657;


const
  FINEANGLES = 8192;
  FINEMASK = FINEANGLES - 1;

// 0x100000000 to 0x2000
  ANGLETOFINESHIFT = 19;
  { Effective size is 10240. }

// Binary Angle Measument, BAM.

type
  angle_t = LongWord;

const
  ANG45 = $20000000;
  ANG90 = $40000000;
  ANG180 = $80000000;
  ANG270 = $c0000000;
  SLOPERANGE = 2048;
  SLOPEBITS = 11;
  DBITS = FRACBITS - SLOPEBITS;

var
// Effective size is 10240.
  finesine: array[0..((5 * FINEANGLES) div 4) - 1] of fixed_t;

// Re-use data, is just PI/2 pahse shift.
  finecosine: Pfixed_tArray;

// Effective size is 4096.
  finetangent : array[0..(FINEANGLES div 2) - 1] of fixed_t;

// Effective size is 2049;
// The +1 size is to handle the case when x==y
//  without additional checking.
  tantoangle: array[0..(SLOPERANGE + 1) - 1] of angle_t;

// Utility function,
// called by R_PointToAngle.

function SlopeDiv(const num: integer; const den: integer): integer;
// VJ was function SlopeDiv(num: LongWord; den: LongWord): integer;


(* Unit d_delphi interface *)


type
  PPointer = ^Pointer;
  
  PString = ^string;

  PBoolean = ^Boolean;

  PInteger = ^Integer;
  
  PLongWord = ^LongWord;
  
  TWordArray = packed array[0..$FFFF] of word;
  PWordArray = ^TWordArray;

  TIntegerArray = packed array[0..$FFFF] of integer;
  PIntegerArray = ^TIntegerArray;

  TLongWordArray = packed array[0..$FFFF] of LongWord;
  PLongWordArray = ^TLongWordArray;

  TSmallintArray = packed array[0..$FFFF] of Smallint;
  PSmallintArray = ^TSmallintArray;

  TByteArray = packed array[0..$FFFF] of Byte;
  PByteArray = ^TByteArray;

  TBooleanArray = packed array[0..$FFFF] of boolean;
  PBooleanArray = ^TBooleanArray;

  PProcedure = procedure;

  TStringArray = array[0..$FFFF] of string;
  PStringArray = ^TStringArray;

  TPointerArray = packed array[0..$FFFF] of pointer;
  PPointerArray = ^TPointerArray;

  PSmallInt = ^SmallInt;
  TSmallIntPArray = packed array[0..$FFFF] of PSmallIntArray;
  PSmallIntPArray = ^TSmallIntPArray;

  PWord = ^Word;
  TWordPArray = packed array[0..$FFFF] of PWordArray;
  PWordPArray = ^TWordPArray;

  PByte = ^Byte;
  TBytePArray = packed array[0..$FFFF] of PByteArray;
  PBytePArray = ^TBytePArray;

  TOutProc = procedure (const s: string);

var
  outproc: TOutProc = nil;

procedure sprintf(var s: string; const Fmt: string; const Args: array of const);

procedure printf(const str: string); overload;

procedure printf(const Fmt: string; const Args: array of const); overload;

function itoa(i: integer): string;

function atoi(const s: string): integer;

function atof(const s: string): single;

function memcpy(dst: pointer; const src: pointer; len: integer): pointer; overload;

function memcpy(dst: pointer; const src: string; len: integer): pointer; overload;

function memmove(dst: pointer; const src: pointer; len: integer): pointer;

function memset(buf: pointer; c: integer; len: integer): pointer;

function malloc(size: integer): Pointer;

procedure free(var p: pointer);

function IntToStrZfill(const z: integer; const x: integer): string;

function boolval(const x: integer): boolean; overload;

function boolval(const c: char): boolean; overload;

function boolval(const p: pointer): boolean; overload;

function intval(const b: boolean): integer;

function decide(const contition: boolean;
  const iftrue: integer; const iffalse: integer): integer; overload;

function decide(const contition: boolean;
  const iftrue: boolean; const iffalse: boolean): boolean; overload;

function decide(const contition: boolean;
  const iftrue: string; const iffalse: string): string; overload;

function decide(const contition: boolean;
  const iftrue: pointer; const iffalse: pointer): pointer; overload;

function decide(const contition: integer;
  const iftrue: integer; const iffalse: integer): integer; overload;

function decide(const contition: integer;
  const iftrue: boolean; const iffalse: boolean): boolean; overload;

function decide(const contition: integer;
  const iftrue: string; const iffalse: string): string; overload;

function decide(const contition: integer;
  const iftrue: pointer; const iffalse: pointer): pointer; overload;

function incp(var p: pointer; const size: integer = 1): pointer;

function pOperation(const p1, p2: pointer; const op: char; size: integer): integer;

function getenv(const env: string): string;

function fexists(const filename: string): boolean;

procedure fdelete(const filename: string);

function fext(const filename: string): string;

const
  fOpen = 0;
  fCreate = 1;

  sFromBeginning = 0;
  sFromCurrent = 1;
  sFromEnd = 2;

type
  TFile = class
  public
    f: file;
    constructor Create(const FileName: string; mode: word);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; virtual;
    function Write(const Buffer; Count: Longint): Longint; virtual;
    function Seek(Offset: Longint; Origin: Word): Longint; virtual;
    function Size: Longint;
  end;

  TCachedFile = class(TFile)
  private
    fBufSize: integer;
    fBuffer: pointer;
    fPosition: integer;
    fBufferStart: integer;
    fBufferEnd: integer;
    fSize: integer;
    fInitialized: boolean;
  protected
    procedure SetSize(NewSize: Longint); //override;
    procedure ResetBuffer; virtual;
  public
    constructor Create(const FileName: string; mode: word; ABufSize: integer = $FFFF); virtual;
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

procedure fprintf(var f: file; const str: string); overload;

procedure fprintf(var f: file; const Fmt: string; const Args: array of const); overload;

procedure fprintf(var f: TFile; const str: string); overload;

procedure fprintf(var f: TFile; const Fmt: string; const Args: array of const); overload;

function tan(const x: single): single;

function strupper(const S: string): string;

function toupper(ch: Char): Char;

function tolower(ch: Char): Char;

function _SHL(const x: integer; const bits: integer): integer;

function _SHLW(const x: LongWord; const bits: LongWord): LongWord;

function _SHR(const x: integer; const bits: integer): integer;

function _SHRW(const x: LongWord; const bits: LongWord): LongWord;

function StringVal(const Str: PChar): string;

procedure ZeroMemory(var X; Count: Integer);


(* Unit w_wad interface *)


               

{
    w_wad.h, w_wad.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	WAD I/O functions.
//	Handles WAD file header, directory, lump I/O.
//
//-----------------------------------------------------------------------------

//
// TYPES
//
type
  char8_t = array[0..7] of char;

  wadinfo_t = packed record
    // Should be "IWAD" or "PWAD".
    identification: integer;
    numlumps: integer;
    infotableofs: integer;
  end;
  Pwadinfo_t = ^wadinfo_t;

  filelump_t = packed record
    filepos: integer;
    size: integer;
    name: char8_t;
  end;
  Pfilelump_t = ^filelump_t;
  Tfilelump_tArray = packed array[0..$FFFF] of filelump_t;
  Pfilelump_tArray = ^Tfilelump_tArray;

//
// WADFILE I/O related stuff.
//
  lumpinfo_t = record
    handle: TCachedFile;
    position: integer;
    size: integer;
    case integer of
      0 : ( name: char8_t);
      1 : ( v1, v2: integer);
    end;

  Plumpinfo_t = ^lumpinfo_t;
  lumpinfo_tArray = array[0..$FFFF] of lumpinfo_t;
  Plumpinfo_tArray = ^lumpinfo_tArray;

function char8tostring(src: char8_t): string;

function stringtochar8(src: string): char8_t;

procedure W_InitMultipleFiles(filenames: PStringArray);

procedure W_Reload;

function W_CheckNumForName(const name: string): integer;

function W_GetNumForName(const name: string): integer;

function W_LumpLength(lump: integer): integer;
procedure W_ReadLump(lump: integer; dest: pointer);

function W_CacheLumpNum(lump: integer; tag: integer): pointer;
function W_CacheLumpName(const name: string; tag: integer): pointer;

var
  lumpinfo: Plumpinfo_tArray;


(* Unit i_io interface *)


                   

var
  debugfile: TFile;
  stderr: TFile;

procedure I_InitializeIO;

procedure I_ShutdownIO;

procedure I_IOMessageBox(const s: string);

procedure I_IOErrorMessageBox(const s: string);

procedure I_IOprintf(const s: string);


(* Unit d_think interface *)


{
    d_think.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  MapObj data. Map Objects or mobjs are actors, entities, 
//  thinker, take-your-pick... anything that moves, acts, or 
//  suffers state changes of more or less violent nature. 
// 
//----------------------------------------------------------------------------- 

type
  actionf_v = procedure;
  actionf_p1 = procedure(p1: pointer);
  actionf_p2 = procedure(p1, p2: pointer);

  actionf_t = record
    case longint of
      0 : ( acp1 : actionf_p1 );
      1 : ( acv : actionf_v );
      2 : ( acp2 : actionf_p2 );
    end;
  Pactionf_t = ^actionf_t;

  think_t = actionf_t;
  Pthink_t = ^think_t;

  Pthinker_t = ^thinker_t;
  thinker_t = record
    prev : Pthinker_t;
    next : Pthinker_t;
    _function : think_t;
  end;


(* Unit m_argv interface *)


{
    m_argv.h, m_argv.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Nil. 
//     
//----------------------------------------------------------------------------- 

// 
// MISC 
//

const
  MAXARGS = 256;

var
  myargc: integer;
  myargv: array[0..MAXARGS] of string;

{ Returns the position of the given parameter }
{ in the arg list (0 if not found). }
function M_CheckParm(const check: string): integer;

function M_CheckParmCDROM: boolean;

procedure M_InitArgv;


(* Unit p_saveg interface *)


               

{
    p_saveg.h, p_saveg.pas
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Savegame I/O, archiving, persistence.
//
//-----------------------------------------------------------------------------

// Persistent storage/archiving.
// These are the load / save game routines.
procedure P_ArchivePlayers;

procedure P_UnArchivePlayers;

procedure P_ArchiveWorld;

procedure P_UnArchiveWorld;

procedure P_ArchiveThinkers;

procedure P_UnArchiveThinkers;

procedure P_ArchiveSpecials;

procedure P_UnArchiveSpecials;

var
  save_p: PByteArray;


(* Unit m_cheat interface *)


{
    m_cheat.h, m_cheat.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Cheat code checking. 
// 
//----------------------------------------------------------------------------- 

// 
// CHEAT SEQUENCE PACKAGE 
// 

type
  cheatseq_t = record
    sequence: string;
    p: string;
  end;
  Pcheatseq_t = ^cheatseq_t;

function cht_CheckCheat(cht: Pcheatseq_t; key: char): boolean;

procedure cht_GetParam(cht: Pcheatseq_t; var buffer: string);

function get_cheatseq_string(const A: array of char): string; overload; // VJ

function get_cheatseq_string(const A: string): string; overload; // VJ

function get_cheatseq_string(const x: integer): string; overload; // VJ


(* Unit d_ticcmd interface *)


{
    d_ticcmd.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  System specific interface stuff. 
// 
//-----------------------------------------------------------------------------

// The data sampled per tick (single player)
// and transmitted to other peers (multiplayer).
// Mainly movements/button commands per game tick,
// plus a checksum for internal state consistency.
type
  ticcmd_t = packed record
    forwardmove: smallint; // *2048 for move // VJ changed from byte to smallint
    sidemove: smallint;    // *2048 for move // VJ changed from byte to smallint
    angleturn: smallint;   // <<16 for angle delta
    consistancy: smallint; // checks for net game
    chatchar: byte;
    buttons: byte;
    look: byte; // look/fly up/down/centering
  end;
  Pticcmd_t = ^ticcmd_t;


(* Unit z_zone interface *)


{
    z_zone.h, z_zone.c
}


// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//      Zone Memory Allocation, perhaps NeXT ObjectiveC inspired.
//	Remark: this was the only stuff that, according
//	 to John Carmack, might have been useful for
//	 Quake.
//
//---------------------------------------------------------------------

               

//
// ZONE MEMORY
// PU - purge tags.
// Tags < 100 are not overwritten until freed.

const
  PU_STATIC = 1;    // static entire execution time
  PU_SOUND = 2;     // static while playing
  PU_MUSIC = 3;     // static while playing
  PU_DAVE = 4;      // anything else Dave wants static
  PU_LEVEL = 50;    // static until level exited
  PU_LEVSPEC = 51;  // a special thinker in a level
  // Tags >= 100 are purgable whenever needed.
  PU_PURGELEVEL = 100;
  PU_CACHE = 101;

procedure Z_Init;

function Z_Malloc(size: integer; tag: integer; user: pointer): pointer;

procedure Z_Free(ptr: pointer);

procedure Z_FreeTags(lowtag: integer; hightag: integer);

procedure Z_DumpHeap(lowtag: integer; hightag: integer);

procedure Z_FileDumpHeap(var f: file); overload;

procedure Z_FileDumpHeap(const filename: string); overload;

procedure Z_CheckHeap;

procedure Z_ChangeTag(ptr: pointer; tag: integer);

procedure Z_ChangeTag2(ptr: pointer; tag: integer);

function Z_FreeMemory: integer;

  { including the header and possibly tiny fragments }
  { NULL if a free block }
  { purgelevel }
  { should be ZONEID }

type
  Pmemblock_t = ^memblock_t;

  memblock_t = record
    size: integer;  // including the header and possibly tiny fragments
    user: PPointer; // NULL if a free block
    tag: integer;   // purgelevel
    id: integer;    // should be ZONEID
    next: Pmemblock_t;
    prev: Pmemblock_t;
  end;


(* Unit i_input interface *)


procedure I_InitInput;

procedure I_ProcessInput;

procedure I_ShutdownInput;

procedure I_SynchronizeInput(active: boolean);

var
  usedirectinput: boolean = false;


(* Unit doomtype interface *)


{
    doomtype.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Simple basic typedefs, isolated here to make it easier 
//   separating modules. 
//     
//-----------------------------------------------------------------------------

type
  bool = boolean;


function MAXCHAR : char;

function MAXSHORT : smallint;

{ Max pos 32-bit int. }
function MAXINT : longint;

function MAXLONG : longint;

function MINCHAR : char;

function MINSHORT : smallint;

{ Max negative 32-bit integer. }
function MININT : longint;

function MINLONG : longint;


(* Unit info_h interface *)


              

{
    info.h
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//  Thing frame/state LUT,
//  generated by multigen utilitiy.
//  This one is the original DOOM version, preserved.
//
//-----------------------------------------------------------------------------

type
  spritenum_t = (
    SPR_TROO, SPR_SHTG, SPR_PUNG, SPR_PISG, SPR_PISF,
    SPR_SHTF, SPR_SHT2, SPR_CHGG, SPR_CHGF, SPR_MISG,
    SPR_MISF, SPR_SAWG, SPR_PLSG, SPR_PLSF, SPR_BFGG,
    SPR_BFGF, SPR_BLUD, SPR_PUFF, SPR_BAL1, SPR_BAL2,
    SPR_PLSS, SPR_PLSE, SPR_MISL, SPR_BFS1, SPR_BFE1,
    SPR_BFE2, SPR_TFOG, SPR_IFOG, SPR_PLAY, SPR_POSS,
    SPR_SPOS, SPR_VILE, SPR_FIRE, SPR_FATB, SPR_FBXP,
    SPR_SKEL, SPR_MANF, SPR_FATT, SPR_CPOS, SPR_SARG,
    SPR_HEAD, SPR_BAL7, SPR_BOSS, SPR_BOS2, SPR_SKUL,
    SPR_SPID, SPR_BSPI, SPR_APLS, SPR_APBX, SPR_CYBR,
    SPR_PAIN, SPR_SSWV, SPR_KEEN, SPR_BBRN, SPR_BOSF,
    SPR_ARM1, SPR_ARM2, SPR_BAR1, SPR_BEXP, SPR_FCAN,
    SPR_BON1, SPR_BON2, SPR_BKEY, SPR_RKEY, SPR_YKEY,
    SPR_BSKU, SPR_RSKU, SPR_YSKU, SPR_STIM, SPR_MEDI,
    SPR_SOUL, SPR_PINV, SPR_PSTR, SPR_PINS, SPR_MEGA,
    SPR_SUIT, SPR_PMAP, SPR_PVIS, SPR_CLIP, SPR_AMMO,
    SPR_ROCK, SPR_BROK, SPR_CELL, SPR_CELP, SPR_SHEL,
    SPR_SBOX, SPR_BPAK, SPR_BFUG, SPR_MGUN, SPR_CSAW,
    SPR_LAUN, SPR_PLAS, SPR_SHOT, SPR_SGN2, SPR_COLU,
    SPR_SMT2, SPR_GOR1, SPR_POL2, SPR_POL5, SPR_POL4,
    SPR_POL3, SPR_POL1, SPR_POL6, SPR_GOR2, SPR_GOR3,
    SPR_GOR4, SPR_GOR5, SPR_SMIT, SPR_COL1, SPR_COL2,
    SPR_COL3, SPR_COL4, SPR_CAND, SPR_CBRA, SPR_COL6,
    SPR_TRE1, SPR_TRE2, SPR_ELEC, SPR_CEYE, SPR_FSKU,
    SPR_COL5, SPR_TBLU, SPR_TGRN, SPR_TRED, SPR_SMBT,
    SPR_SMGT, SPR_SMRT, SPR_HDB1, SPR_HDB2, SPR_HDB3,
    SPR_HDB4, SPR_HDB5, SPR_HDB6, SPR_POB1, SPR_POB2,
    SPR_BRS1, SPR_TLMP, SPR_TLP2, NUMSPRITES
  );

  statenum_t = (
    S_NULL, S_LIGHTDONE, S_PUNCH, S_PUNCHDOWN,
    S_PUNCHUP, S_PUNCH1, S_PUNCH2, S_PUNCH3, S_PUNCH4,
    S_PUNCH5, S_PISTOL, S_PISTOLDOWN, S_PISTOLUP,
    S_PISTOL1, S_PISTOL2, S_PISTOL3, S_PISTOL4,
    S_PISTOLFLASH, S_SGUN, S_SGUNDOWN, S_SGUNUP,
    S_SGUN1, S_SGUN2, S_SGUN3, S_SGUN4, S_SGUN5,
    S_SGUN6, S_SGUN7, S_SGUN8, S_SGUN9, S_SGUNFLASH1,
    S_SGUNFLASH2, S_DSGUN, S_DSGUNDOWN, S_DSGUNUP,
    S_DSGUN1, S_DSGUN2, S_DSGUN3, S_DSGUN4, S_DSGUN5,
    S_DSGUN6, S_DSGUN7, S_DSGUN8, S_DSGUN9, S_DSGUN10,
    S_DSNR1, S_DSNR2, S_DSGUNFLASH1, S_DSGUNFLASH2,
    S_CHAIN, S_CHAINDOWN, S_CHAINUP, S_CHAIN1,
    S_CHAIN2, S_CHAIN3, S_CHAINFLASH1, S_CHAINFLASH2,
    S_MISSILE, S_MISSILEDOWN, S_MISSILEUP, S_MISSILE1,
    S_MISSILE2, S_MISSILE3, S_MISSILEFLASH1,
    S_MISSILEFLASH2, S_MISSILEFLASH3, S_MISSILEFLASH4,
    S_SAW, S_SAWB, S_SAWDOWN, S_SAWUP, S_SAW1, S_SAW2,
    S_SAW3, S_PLASMA, S_PLASMADOWN, S_PLASMAUP,
    S_PLASMA1, S_PLASMA2, S_PLASMAFLASH1, S_PLASMAFLASH2,
    S_BFG, S_BFGDOWN, S_BFGUP, S_BFG1, S_BFG2, S_BFG3,
    S_BFG4, S_BFGFLASH1, S_BFGFLASH2, S_BLOOD1,
    S_BLOOD2, S_BLOOD3, S_PUFF1, S_PUFF2, S_PUFF3,
    S_PUFF4, S_TBALL1, S_TBALL2, S_TBALLX1, S_TBALLX2,
    S_TBALLX3, S_RBALL1, S_RBALL2, S_RBALLX1,
    S_RBALLX2, S_RBALLX3, S_PLASBALL, S_PLASBALL2,
    S_PLASEXP, S_PLASEXP2, S_PLASEXP3, S_PLASEXP4,
    S_PLASEXP5, S_ROCKET, S_BFGSHOT, S_BFGSHOT2,
    S_BFGLAND, S_BFGLAND2, S_BFGLAND3, S_BFGLAND4,
    S_BFGLAND5, S_BFGLAND6, S_BFGEXP, S_BFGEXP2,
    S_BFGEXP3, S_BFGEXP4, S_EXPLODE1, S_EXPLODE2,
    S_EXPLODE3, S_TFOG, S_TFOG01, S_TFOG02, S_TFOG2,
    S_TFOG3, S_TFOG4, S_TFOG5, S_TFOG6, S_TFOG7,
    S_TFOG8, S_TFOG9, S_TFOG10, S_IFOG, S_IFOG01,
    S_IFOG02, S_IFOG2, S_IFOG3, S_IFOG4, S_IFOG5,
    S_PLAY, S_PLAY_RUN1, S_PLAY_RUN2, S_PLAY_RUN3,
    S_PLAY_RUN4, S_PLAY_ATK1, S_PLAY_ATK2, S_PLAY_PAIN,
    S_PLAY_PAIN2, S_PLAY_DIE1, S_PLAY_DIE2,
    S_PLAY_DIE3, S_PLAY_DIE4, S_PLAY_DIE5, S_PLAY_DIE6,
    S_PLAY_DIE7, S_PLAY_XDIE1, S_PLAY_XDIE2,
    S_PLAY_XDIE3, S_PLAY_XDIE4, S_PLAY_XDIE5,
    S_PLAY_XDIE6, S_PLAY_XDIE7, S_PLAY_XDIE8,
    S_PLAY_XDIE9, S_POSS_STND, S_POSS_STND2,
    S_POSS_RUN1, S_POSS_RUN2, S_POSS_RUN3, S_POSS_RUN4,
    S_POSS_RUN5, S_POSS_RUN6, S_POSS_RUN7, S_POSS_RUN8,
    S_POSS_ATK1, S_POSS_ATK2, S_POSS_ATK3, S_POSS_PAIN,
    S_POSS_PAIN2, S_POSS_DIE1, S_POSS_DIE2,
    S_POSS_DIE3, S_POSS_DIE4, S_POSS_DIE5, S_POSS_XDIE1,
    S_POSS_XDIE2, S_POSS_XDIE3, S_POSS_XDIE4,
    S_POSS_XDIE5, S_POSS_XDIE6, S_POSS_XDIE7,
    S_POSS_XDIE8, S_POSS_XDIE9, S_POSS_RAISE1,
    S_POSS_RAISE2, S_POSS_RAISE3, S_POSS_RAISE4,
    S_SPOS_STND, S_SPOS_STND2, S_SPOS_RUN1,
    S_SPOS_RUN2, S_SPOS_RUN3, S_SPOS_RUN4, S_SPOS_RUN5,
    S_SPOS_RUN6, S_SPOS_RUN7, S_SPOS_RUN8, S_SPOS_ATK1,
    S_SPOS_ATK2, S_SPOS_ATK3, S_SPOS_PAIN, S_SPOS_PAIN2,
    S_SPOS_DIE1, S_SPOS_DIE2, S_SPOS_DIE3, S_SPOS_DIE4,
    S_SPOS_DIE5, S_SPOS_XDIE1, S_SPOS_XDIE2,
    S_SPOS_XDIE3, S_SPOS_XDIE4, S_SPOS_XDIE5,
    S_SPOS_XDIE6, S_SPOS_XDIE7, S_SPOS_XDIE8,
    S_SPOS_XDIE9, S_SPOS_RAISE1, S_SPOS_RAISE2,
    S_SPOS_RAISE3, S_SPOS_RAISE4, S_SPOS_RAISE5,
    S_VILE_STND, S_VILE_STND2, S_VILE_RUN1,
    S_VILE_RUN2, S_VILE_RUN3, S_VILE_RUN4, S_VILE_RUN5,
    S_VILE_RUN6, S_VILE_RUN7, S_VILE_RUN8, S_VILE_RUN9,
    S_VILE_RUN10, S_VILE_RUN11, S_VILE_RUN12,
    S_VILE_ATK1, S_VILE_ATK2, S_VILE_ATK3, S_VILE_ATK4,
    S_VILE_ATK5, S_VILE_ATK6, S_VILE_ATK7, S_VILE_ATK8,
    S_VILE_ATK9, S_VILE_ATK10, S_VILE_ATK11,
    S_VILE_HEAL1, S_VILE_HEAL2, S_VILE_HEAL3,
    S_VILE_PAIN, S_VILE_PAIN2, S_VILE_DIE1,
    S_VILE_DIE2, S_VILE_DIE3, S_VILE_DIE4, S_VILE_DIE5,
    S_VILE_DIE6, S_VILE_DIE7, S_VILE_DIE8, S_VILE_DIE9,
    S_VILE_DIE10, S_FIRE1, S_FIRE2, S_FIRE3, S_FIRE4,
    S_FIRE5, S_FIRE6, S_FIRE7, S_FIRE8, S_FIRE9,
    S_FIRE10, S_FIRE11, S_FIRE12, S_FIRE13, S_FIRE14,
    S_FIRE15, S_FIRE16, S_FIRE17, S_FIRE18, S_FIRE19,
    S_FIRE20, S_FIRE21, S_FIRE22, S_FIRE23, S_FIRE24,
    S_FIRE25, S_FIRE26, S_FIRE27, S_FIRE28, S_FIRE29,
    S_FIRE30, S_SMOKE1, S_SMOKE2, S_SMOKE3, S_SMOKE4,
    S_SMOKE5, S_TRACER, S_TRACER2, S_TRACEEXP1,
    S_TRACEEXP2, S_TRACEEXP3, S_SKEL_STND, S_SKEL_STND2,
    S_SKEL_RUN1, S_SKEL_RUN2, S_SKEL_RUN3, S_SKEL_RUN4,
    S_SKEL_RUN5, S_SKEL_RUN6, S_SKEL_RUN7, S_SKEL_RUN8,
    S_SKEL_RUN9, S_SKEL_RUN10, S_SKEL_RUN11,
    S_SKEL_RUN12, S_SKEL_FIST1, S_SKEL_FIST2,
    S_SKEL_FIST3, S_SKEL_FIST4, S_SKEL_MISS1,
    S_SKEL_MISS2, S_SKEL_MISS3, S_SKEL_MISS4,
    S_SKEL_PAIN, S_SKEL_PAIN2, S_SKEL_DIE1,
    S_SKEL_DIE2, S_SKEL_DIE3, S_SKEL_DIE4, S_SKEL_DIE5,
    S_SKEL_DIE6, S_SKEL_RAISE1, S_SKEL_RAISE2,
    S_SKEL_RAISE3, S_SKEL_RAISE4, S_SKEL_RAISE5,
    S_SKEL_RAISE6, S_FATSHOT1, S_FATSHOT2, S_FATSHOTX1,
    S_FATSHOTX2, S_FATSHOTX3, S_FATT_STND, S_FATT_STND2,
    S_FATT_RUN1, S_FATT_RUN2, S_FATT_RUN3, S_FATT_RUN4,
    S_FATT_RUN5, S_FATT_RUN6, S_FATT_RUN7, S_FATT_RUN8,
    S_FATT_RUN9, S_FATT_RUN10, S_FATT_RUN11,
    S_FATT_RUN12, S_FATT_ATK1, S_FATT_ATK2,
    S_FATT_ATK3, S_FATT_ATK4, S_FATT_ATK5, S_FATT_ATK6,
    S_FATT_ATK7, S_FATT_ATK8, S_FATT_ATK9, S_FATT_ATK10,
    S_FATT_PAIN, S_FATT_PAIN2, S_FATT_DIE1,
    S_FATT_DIE2, S_FATT_DIE3, S_FATT_DIE4, S_FATT_DIE5,
    S_FATT_DIE6, S_FATT_DIE7, S_FATT_DIE8, S_FATT_DIE9,
    S_FATT_DIE10, S_FATT_RAISE1, S_FATT_RAISE2,
    S_FATT_RAISE3, S_FATT_RAISE4, S_FATT_RAISE5,
    S_FATT_RAISE6, S_FATT_RAISE7, S_FATT_RAISE8,
    S_CPOS_STND, S_CPOS_STND2, S_CPOS_RUN1,
    S_CPOS_RUN2, S_CPOS_RUN3, S_CPOS_RUN4, S_CPOS_RUN5,
    S_CPOS_RUN6, S_CPOS_RUN7, S_CPOS_RUN8, S_CPOS_ATK1,
    S_CPOS_ATK2, S_CPOS_ATK3, S_CPOS_ATK4, S_CPOS_PAIN,
    S_CPOS_PAIN2, S_CPOS_DIE1, S_CPOS_DIE2,
    S_CPOS_DIE3, S_CPOS_DIE4, S_CPOS_DIE5, S_CPOS_DIE6,
    S_CPOS_DIE7, S_CPOS_XDIE1, S_CPOS_XDIE2,
    S_CPOS_XDIE3, S_CPOS_XDIE4, S_CPOS_XDIE5,
    S_CPOS_XDIE6, S_CPOS_RAISE1, S_CPOS_RAISE2,
    S_CPOS_RAISE3, S_CPOS_RAISE4, S_CPOS_RAISE5,
    S_CPOS_RAISE6, S_CPOS_RAISE7, S_TROO_STND,
    S_TROO_STND2, S_TROO_RUN1, S_TROO_RUN2,
    S_TROO_RUN3, S_TROO_RUN4, S_TROO_RUN5, S_TROO_RUN6,
    S_TROO_RUN7, S_TROO_RUN8, S_TROO_ATK1, S_TROO_ATK2,
    S_TROO_ATK3, S_TROO_PAIN, S_TROO_PAIN2,
    S_TROO_DIE1, S_TROO_DIE2, S_TROO_DIE3, S_TROO_DIE4,
    S_TROO_DIE5, S_TROO_XDIE1, S_TROO_XDIE2,
    S_TROO_XDIE3, S_TROO_XDIE4, S_TROO_XDIE5,
    S_TROO_XDIE6, S_TROO_XDIE7, S_TROO_XDIE8,
    S_TROO_RAISE1, S_TROO_RAISE2, S_TROO_RAISE3,
    S_TROO_RAISE4, S_TROO_RAISE5, S_SARG_STND,
    S_SARG_STND2, S_SARG_RUN1, S_SARG_RUN2,
    S_SARG_RUN3, S_SARG_RUN4, S_SARG_RUN5, S_SARG_RUN6,
    S_SARG_RUN7, S_SARG_RUN8, S_SARG_ATK1, S_SARG_ATK2,
    S_SARG_ATK3, S_SARG_PAIN, S_SARG_PAIN2,
    S_SARG_DIE1, S_SARG_DIE2, S_SARG_DIE3, S_SARG_DIE4,
    S_SARG_DIE5, S_SARG_DIE6, S_SARG_RAISE1,
    S_SARG_RAISE2, S_SARG_RAISE3, S_SARG_RAISE4,
    S_SARG_RAISE5, S_SARG_RAISE6, S_HEAD_STND,
    S_HEAD_RUN1, S_HEAD_ATK1, S_HEAD_ATK2, S_HEAD_ATK3,
    S_HEAD_PAIN, S_HEAD_PAIN2, S_HEAD_PAIN3,
    S_HEAD_DIE1, S_HEAD_DIE2, S_HEAD_DIE3, S_HEAD_DIE4,
    S_HEAD_DIE5, S_HEAD_DIE6, S_HEAD_RAISE1,
    S_HEAD_RAISE2, S_HEAD_RAISE3, S_HEAD_RAISE4,
    S_HEAD_RAISE5, S_HEAD_RAISE6, S_BRBALL1,
    S_BRBALL2, S_BRBALLX1, S_BRBALLX2, S_BRBALLX3,
    S_BOSS_STND, S_BOSS_STND2, S_BOSS_RUN1,
    S_BOSS_RUN2, S_BOSS_RUN3, S_BOSS_RUN4, S_BOSS_RUN5,
    S_BOSS_RUN6, S_BOSS_RUN7, S_BOSS_RUN8, S_BOSS_ATK1,
    S_BOSS_ATK2, S_BOSS_ATK3, S_BOSS_PAIN, S_BOSS_PAIN2,
    S_BOSS_DIE1, S_BOSS_DIE2, S_BOSS_DIE3, S_BOSS_DIE4,
    S_BOSS_DIE5, S_BOSS_DIE6, S_BOSS_DIE7, S_BOSS_RAISE1,
    S_BOSS_RAISE2, S_BOSS_RAISE3, S_BOSS_RAISE4,
    S_BOSS_RAISE5, S_BOSS_RAISE6, S_BOSS_RAISE7,
    S_BOS2_STND, S_BOS2_STND2, S_BOS2_RUN1,
    S_BOS2_RUN2, S_BOS2_RUN3, S_BOS2_RUN4, S_BOS2_RUN5,
    S_BOS2_RUN6, S_BOS2_RUN7, S_BOS2_RUN8, S_BOS2_ATK1,
    S_BOS2_ATK2, S_BOS2_ATK3, S_BOS2_PAIN, S_BOS2_PAIN2,
    S_BOS2_DIE1, S_BOS2_DIE2, S_BOS2_DIE3, S_BOS2_DIE4,
    S_BOS2_DIE5, S_BOS2_DIE6, S_BOS2_DIE7, S_BOS2_RAISE1,
    S_BOS2_RAISE2, S_BOS2_RAISE3, S_BOS2_RAISE4,
    S_BOS2_RAISE5, S_BOS2_RAISE6, S_BOS2_RAISE7,
    S_SKULL_STND, S_SKULL_STND2, S_SKULL_RUN1,
    S_SKULL_RUN2, S_SKULL_ATK1, S_SKULL_ATK2,
    S_SKULL_ATK3, S_SKULL_ATK4, S_SKULL_PAIN,
    S_SKULL_PAIN2, S_SKULL_DIE1, S_SKULL_DIE2,
    S_SKULL_DIE3, S_SKULL_DIE4, S_SKULL_DIE5,
    S_SKULL_DIE6, S_SPID_STND, S_SPID_STND2,
    S_SPID_RUN1, S_SPID_RUN2, S_SPID_RUN3, S_SPID_RUN4,
    S_SPID_RUN5, S_SPID_RUN6, S_SPID_RUN7, S_SPID_RUN8,
    S_SPID_RUN9, S_SPID_RUN10, S_SPID_RUN11,
    S_SPID_RUN12, S_SPID_ATK1, S_SPID_ATK2,
    S_SPID_ATK3, S_SPID_ATK4, S_SPID_PAIN, S_SPID_PAIN2,
    S_SPID_DIE1, S_SPID_DIE2, S_SPID_DIE3, S_SPID_DIE4,
    S_SPID_DIE5, S_SPID_DIE6, S_SPID_DIE7, S_SPID_DIE8,
    S_SPID_DIE9, S_SPID_DIE10, S_SPID_DIE11,
    S_BSPI_STND, S_BSPI_STND2, S_BSPI_SIGHT,
    S_BSPI_RUN1, S_BSPI_RUN2, S_BSPI_RUN3, S_BSPI_RUN4,
    S_BSPI_RUN5, S_BSPI_RUN6, S_BSPI_RUN7, S_BSPI_RUN8,
    S_BSPI_RUN9, S_BSPI_RUN10, S_BSPI_RUN11,
    S_BSPI_RUN12, S_BSPI_ATK1, S_BSPI_ATK2,
    S_BSPI_ATK3, S_BSPI_ATK4, S_BSPI_PAIN, S_BSPI_PAIN2,
    S_BSPI_DIE1, S_BSPI_DIE2, S_BSPI_DIE3, S_BSPI_DIE4,
    S_BSPI_DIE5, S_BSPI_DIE6, S_BSPI_DIE7, S_BSPI_RAISE1,
    S_BSPI_RAISE2, S_BSPI_RAISE3, S_BSPI_RAISE4,
    S_BSPI_RAISE5, S_BSPI_RAISE6, S_BSPI_RAISE7,
    S_ARACH_PLAZ, S_ARACH_PLAZ2, S_ARACH_PLEX,
    S_ARACH_PLEX2, S_ARACH_PLEX3, S_ARACH_PLEX4,
    S_ARACH_PLEX5, S_CYBER_STND, S_CYBER_STND2,
    S_CYBER_RUN1, S_CYBER_RUN2, S_CYBER_RUN3,
    S_CYBER_RUN4, S_CYBER_RUN5, S_CYBER_RUN6,
    S_CYBER_RUN7, S_CYBER_RUN8, S_CYBER_ATK1,
    S_CYBER_ATK2, S_CYBER_ATK3, S_CYBER_ATK4,
    S_CYBER_ATK5, S_CYBER_ATK6, S_CYBER_PAIN,
    S_CYBER_DIE1, S_CYBER_DIE2, S_CYBER_DIE3,
    S_CYBER_DIE4, S_CYBER_DIE5, S_CYBER_DIE6,
    S_CYBER_DIE7, S_CYBER_DIE8, S_CYBER_DIE9,
    S_CYBER_DIE10, S_PAIN_STND, S_PAIN_RUN1,
    S_PAIN_RUN2, S_PAIN_RUN3, S_PAIN_RUN4, S_PAIN_RUN5,
    S_PAIN_RUN6, S_PAIN_ATK1, S_PAIN_ATK2, S_PAIN_ATK3,
    S_PAIN_ATK4, S_PAIN_PAIN, S_PAIN_PAIN2,
    S_PAIN_DIE1, S_PAIN_DIE2, S_PAIN_DIE3, S_PAIN_DIE4,
    S_PAIN_DIE5, S_PAIN_DIE6, S_PAIN_RAISE1,
    S_PAIN_RAISE2, S_PAIN_RAISE3, S_PAIN_RAISE4,
    S_PAIN_RAISE5, S_PAIN_RAISE6, S_SSWV_STND,
    S_SSWV_STND2, S_SSWV_RUN1, S_SSWV_RUN2,
    S_SSWV_RUN3, S_SSWV_RUN4, S_SSWV_RUN5, S_SSWV_RUN6,
    S_SSWV_RUN7, S_SSWV_RUN8, S_SSWV_ATK1, S_SSWV_ATK2,
    S_SSWV_ATK3, S_SSWV_ATK4, S_SSWV_ATK5, S_SSWV_ATK6,
    S_SSWV_PAIN, S_SSWV_PAIN2, S_SSWV_DIE1,
    S_SSWV_DIE2, S_SSWV_DIE3, S_SSWV_DIE4, S_SSWV_DIE5,
    S_SSWV_XDIE1, S_SSWV_XDIE2, S_SSWV_XDIE3,
    S_SSWV_XDIE4, S_SSWV_XDIE5, S_SSWV_XDIE6,
    S_SSWV_XDIE7, S_SSWV_XDIE8, S_SSWV_XDIE9,
    S_SSWV_RAISE1, S_SSWV_RAISE2, S_SSWV_RAISE3,
    S_SSWV_RAISE4, S_SSWV_RAISE5, S_KEENSTND,
    S_COMMKEEN, S_COMMKEEN2, S_COMMKEEN3, S_COMMKEEN4,
    S_COMMKEEN5, S_COMMKEEN6, S_COMMKEEN7, S_COMMKEEN8,
    S_COMMKEEN9, S_COMMKEEN10, S_COMMKEEN11,
    S_COMMKEEN12, S_KEENPAIN, S_KEENPAIN2, S_BRAIN,
    S_BRAIN_PAIN, S_BRAIN_DIE1, S_BRAIN_DIE2,
    S_BRAIN_DIE3, S_BRAIN_DIE4, S_BRAINEYE,
    S_BRAINEYESEE, S_BRAINEYE1, S_SPAWN1, S_SPAWN2,
    S_SPAWN3, S_SPAWN4, S_SPAWNFIRE1, S_SPAWNFIRE2,
    S_SPAWNFIRE3, S_SPAWNFIRE4, S_SPAWNFIRE5,
    S_SPAWNFIRE6, S_SPAWNFIRE7, S_SPAWNFIRE8,
    S_BRAINEXPLODE1, S_BRAINEXPLODE2, S_BRAINEXPLODE3,
    S_ARM1, S_ARM1A, S_ARM2, S_ARM2A, S_BAR1, S_BAR2,
    S_BEXP, S_BEXP2, S_BEXP3, S_BEXP4, S_BEXP5,
    S_BBAR1, S_BBAR2, S_BBAR3, S_BON1, S_BON1A,
    S_BON1B, S_BON1C, S_BON1D, S_BON1E, S_BON2,
    S_BON2A, S_BON2B, S_BON2C, S_BON2D, S_BON2E,
    S_BKEY, S_BKEY2, S_RKEY, S_RKEY2, S_YKEY, S_YKEY2,
    S_BSKULL, S_BSKULL2, S_RSKULL, S_RSKULL2,
    S_YSKULL, S_YSKULL2, S_STIM, S_MEDI, S_SOUL,
    S_SOUL2, S_SOUL3, S_SOUL4, S_SOUL5, S_SOUL6,
    S_PINV, S_PINV2, S_PINV3, S_PINV4, S_PSTR, S_PINS,
    S_PINS2, S_PINS3, S_PINS4, S_MEGA, S_MEGA2,
    S_MEGA3, S_MEGA4, S_SUIT, S_PMAP, S_PMAP2, S_PMAP3,
    S_PMAP4, S_PMAP5, S_PMAP6, S_PVIS, S_PVIS2,
    S_CLIP, S_AMMO, S_ROCK, S_BROK, S_CELL, S_CELP,
    S_SHEL, S_SBOX, S_BPAK, S_BFUG, S_MGUN, S_CSAW,
    S_LAUN, S_PLAS, S_SHOT, S_SHOT2, S_COLU, S_STALAG,
    S_BLOODYTWITCH, S_BLOODYTWITCH2, S_BLOODYTWITCH3,
    S_BLOODYTWITCH4, S_DEADTORSO, S_DEADBOTTOM,
    S_HEADSONSTICK, S_GIBS, S_HEADONASTICK,
    S_HEADCANDLES, S_HEADCANDLES2, S_DEADSTICK,
    S_LIVESTICK, S_LIVESTICK2, S_MEAT2, S_MEAT3,
    S_MEAT4, S_MEAT5, S_STALAGTITE, S_TALLGRNCOL,
    S_SHRTGRNCOL, S_TALLREDCOL, S_SHRTREDCOL,
    S_CANDLESTIK, S_CANDELABRA, S_SKULLCOL,
    S_TORCHTREE, S_BIGTREE, S_TECHPILLAR, S_EVILEYE,
    S_EVILEYE2, S_EVILEYE3, S_EVILEYE4, S_FLOATSKULL,
    S_FLOATSKULL2, S_FLOATSKULL3, S_HEARTCOL,
    S_HEARTCOL2, S_BLUETORCH, S_BLUETORCH2,
    S_BLUETORCH3, S_BLUETORCH4, S_GREENTORCH,
    S_GREENTORCH2, S_GREENTORCH3, S_GREENTORCH4,
    S_REDTORCH, S_REDTORCH2, S_REDTORCH3, S_REDTORCH4,
    S_BTORCHSHRT, S_BTORCHSHRT2, S_BTORCHSHRT3,
    S_BTORCHSHRT4, S_GTORCHSHRT, S_GTORCHSHRT2,
    S_GTORCHSHRT3, S_GTORCHSHRT4, S_RTORCHSHRT,
    S_RTORCHSHRT2, S_RTORCHSHRT3, S_RTORCHSHRT4,
    S_HANGNOGUTS, S_HANGBNOBRAIN, S_HANGTLOOKDN,
    S_HANGTSKULL, S_HANGTLOOKUP, S_HANGTNOBRAIN,
    S_COLONGIBS, S_SMALLPOOL, S_BRAINSTEM, S_TECHLAMP,
    S_TECHLAMP2, S_TECHLAMP3, S_TECHLAMP4, S_TECH2LAMP,
    S_TECH2LAMP2, S_TECH2LAMP3, S_TECH2LAMP4,
    NUMSTATES
  );

  state_t = record
    sprite: spritenum_t;
    frame: integer;
    tics: integer;
    action: actionf_t;
    nextstate: statenum_t;
    misc1: integer;
    misc2: integer;
  end;
  Pstate_t = ^state_t;

type
  mobjtype_t = (
    MT_PLAYER, MT_POSSESSED, MT_SHOTGUY, MT_VILE,
    MT_FIRE, MT_UNDEAD, MT_TRACER, MT_SMOKE, MT_FATSO,
    MT_FATSHOT, MT_CHAINGUY, MT_TROOP, MT_SERGEANT,
    MT_SHADOWS, MT_HEAD, MT_BRUISER, MT_BRUISERSHOT,
    MT_KNIGHT, MT_SKULL, MT_SPIDER, MT_BABY, MT_CYBORG,
    MT_PAIN, MT_WOLFSS, MT_KEEN, MT_BOSSBRAIN,
    MT_BOSSSPIT, MT_BOSSTARGET, MT_SPAWNSHOT,
    MT_SPAWNFIRE, MT_BARREL, MT_TROOPSHOT, MT_HEADSHOT,
    MT_ROCKET, MT_PLASMA, MT_BFG, MT_ARACHPLAZ,
    MT_PUFF, MT_BLOOD, MT_TFOG, MT_IFOG, MT_TELEPORTMAN,
    MT_EXTRABFG, MT_MISC0, MT_MISC1, MT_MISC2,
    MT_MISC3, MT_MISC4, MT_MISC5, MT_MISC6, MT_MISC7,
    MT_MISC8, MT_MISC9, MT_MISC10, MT_MISC11,
    MT_MISC12, MT_INV, MT_MISC13, MT_INS, MT_MISC14,
    MT_MISC15, MT_MISC16, MT_MEGA, MT_CLIP, MT_MISC17,
    MT_MISC18, MT_MISC19, MT_MISC20, MT_MISC21,
    MT_MISC22, MT_MISC23, MT_MISC24, MT_MISC25,
    MT_CHAINGUN, MT_MISC26, MT_MISC27, MT_MISC28,
    MT_SHOTGUN, MT_SUPERSHOTGUN, MT_MISC29,
    MT_MISC30, MT_MISC31, MT_MISC32, MT_MISC33,
    MT_MISC34, MT_MISC35, MT_MISC36, MT_MISC37,
    MT_MISC38, MT_MISC39, MT_MISC40, MT_MISC41,
    MT_MISC42, MT_MISC43, MT_MISC44, MT_MISC45,
    MT_MISC46, MT_MISC47, MT_MISC48, MT_MISC49,
    MT_MISC50, MT_MISC51, MT_MISC52, MT_MISC53,
    MT_MISC54, MT_MISC55, MT_MISC56, MT_MISC57,
    MT_MISC58, MT_MISC59, MT_MISC60, MT_MISC61,
    MT_MISC62, MT_MISC63, MT_MISC64, MT_MISC65,
    MT_MISC66, MT_MISC67, MT_MISC68, MT_MISC69,
    MT_MISC70, MT_MISC71, MT_MISC72, MT_MISC73,
    MT_MISC74, MT_MISC75, MT_MISC76, MT_MISC77,
    MT_MISC78, MT_MISC79, MT_MISC80, MT_MISC81,
    MT_MISC82, MT_MISC83, MT_MISC84, MT_MISC85,
    MT_MISC86,NUMMOBJTYPES
  );

{  mobjinfo_t = packed record
    doomednum: smallint;
    spawnstate: smallint;
    spawnhealth: smallint;
    seestate: smallint;
    seesound: smallint;
    reactiontime: integer;
    attacksound: smallint;
    painstate: smallint;
    painchance: integer;
    painsound: smallint;
    meleestate: smallint;
    missilestate: smallint;
    deathstate: smallint;
    xdeathstate: smallint;
    deathsound: smallint;
    speed: integer;
    radius: integer;
    height: integer;
    mass: integer;
    damage: integer;
    activesound: smallint;
    flags: integer;
    raisestate: smallint;
  end;
  Pmobjinfo_t = ^mobjinfo_t;}
  
  mobjinfo_t = record
    doomednum: integer;
    spawnstate: integer;
    spawnhealth: integer;
    seestate: integer;
    seesound: integer;
    reactiontime: integer;
    attacksound: integer;
    painstate: integer;
    painchance: integer;
    painsound: integer;
    meleestate: integer;
    missilestate: integer;
    deathstate: integer;
    xdeathstate: integer;
    deathsound: integer;
    speed: integer;
    radius: integer;
    height: integer;
    mass: integer;
    damage: integer;
    activesound: integer;
    flags: integer;
    raisestate: integer;
  end;
  Pmobjinfo_t = ^mobjinfo_t;



(* Unit info interface *)


                      

{
    info.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Thing frame/state LUT, 
//  generated by multigen utilitiy. 
//  This one is the original DOOM version, preserved. 
// 
//-----------------------------------------------------------------------------

type
  statesArray_t = packed array[0..ord(NUMSTATES) - 1] of state_t;
  PstatesArray_t = ^statesArray_t;

  sprnamesArray_t = packed array[0..ord(NUMSPRITES) - 1] of string[4];
  PsprnamesArray_t = ^sprnamesArray_t;

  mobjinfoArray_t = packed array[0..ord(NUMMOBJTYPES) - 1] of mobjinfo_t;
  PmobjinfoArray_t = ^mobjinfoArray_t;

var
  states: PstatesArray_t;
  sprnames: PsprnamesArray_t;
  mobjinfo: PmobjinfoArray_t;

procedure I_InitInfo;


(* Unit r_sky interface *)


{
    r_sky.h, r_sky.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Sky rendering.
//
//-----------------------------------------------------------------------------

// SKY, store the number for name.
const
  SKYFLATNAME = 'F_SKY1';

// The sky map is 256*128*4 maps.
  ANGLETOSKYSHIFT = 22;

var
  skyflatnum: integer;
  skytexture: integer;
  skytexturemid: integer;

procedure R_InitSkyMap;
  

(* Unit s_sound interface *)


{
    s_sound.h, s_sound.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	The not so system specific sound interface.
//
//-----------------------------------------------------------------------------

//
// Initializes sound stuff, including volume
// Sets channels, SFX and music volume,
//  allocates channel buffer, sets S_sfx lookup.
//
procedure S_Init(sfxVolume: integer; musicVolume: integer);


//
// Per level startup code.
// Kills playing sounds at start of level,
//  determines music if any, changes music.
//
procedure S_Start;


//
// Start sound for thing at <origin>
//  using <sound_id> from sounds.h
//
procedure S_StartSound(origin: pointer; sfx_id: integer);


// Will start a sound at a given volume.
procedure S_StartSoundAtVolume(origin_p: pointer; sfx_id: integer; volume: integer);


// Stop sound for thing at <origin>
procedure S_StopSound(origin: pointer);


// Start music using <music_id> from sounds.h
procedure S_StartMusic(music_id: integer);

// Start music using <music_id> from sounds.h,
//  and set whether looping
procedure S_ChangeMusic(musicnum: integer; looping: boolean);

// Stops the music fer sure.
procedure S_StopMusic;

// Stop and resume music, during game PAUSE.
procedure S_PauseSound;
procedure S_ResumeSound;


//
// Updates music & sounds
//
procedure S_UpdateSounds(listener_p: pointer);

procedure S_SetMusicVolume(volume: integer);
procedure S_SetSfxVolume(volume: integer);

var
// These are not used, but should be (menu).
// Maximum volume of a sound effect.
// Internal default is max out of 0-15.
  snd_SfxVolume: integer;

// Maximum volume of music. Useless so far.
  snd_MusicVolume: integer;

// following is set
//  by the defaults code in M_misc:
// number of channels available
  numChannels: integer;
  

(* Unit m_misc interface *)


{
    m_misc.h, m_misc.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// Main loop menu stuff. 
// Default Config File. 
// PCX Screenshots. 
// 
//     
//----------------------------------------------------------------------------- 

//
// MISC 
//

function M_WriteFile(const name: string; source: pointer; length: integer): boolean;

function M_ReadFile(const name: string; var buffer: Pointer): integer;

procedure M_ScreenShot;

procedure M_LoadDefaults;

procedure M_SaveDefaults;


(* Unit d_englsh interface *)


{
    d_englsh.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Printed strings for translation. 
//  English language support (default). 
// 
//----------------------------------------------------------------------------- 
// 
//  Printed strings for translation 
// 
// 
// D_Main.C 
//

const
  D_DEVSTR = 'Development mode ON.' + #13#10;
  D_CDROM = 'CD-ROM Version: default.cfg from c:\doomdata' + #13#10;

  { }
  {  M_Menu.C }
  { }
  PRESSKEY = 'press a key.';
  PRESSYN = 'press y or n.';
  QUITMSG = 'are you sure you want to' + #13#10 +
            'quit this great game?';
  LOADNET = 'you can''t do load while in a net game!' + #13#10 + #13#10 +
            PRESSKEY;
  QLOADNET = 'you can''t quickload during a netgame!' + #13#10 + #13#10 +
             PRESSKEY;
  QSAVESPOT = 'you haven''t picked a quicksave slot yet!' + #13#10 + #13#10 +
              PRESSKEY;
  SAVEDEAD = 'you can''t save if you aren''t playing!' + #13#10 + #13#10 +
             PRESSKEY;
  QSPROMPT = 'quicksave over your game named' + #13#10 + #13#10 +
             '''%s''?' + #13#10 + #13#10 +
             PRESSYN;
  QLPROMPT = 'do you want to quickload the game named' + #13#10 + #13#10 +
             '''%s''?' + #13#10 + #13#10 +
             PRESSYN;
  SNEWGAME = 'you can''t start a new game' + #13#10 +
             'while in a network game.' + #13#10 + #13#10 +
             PRESSKEY;
  SNIGHTMARE = 'are you sure? this skill level' + #13#10 +
               'isn''t even remotely fair.' + #13#10 +
               PRESSYN;
  SWSTRING = 'this is the shareware version of doom.' + #13#10 +
             'you need to order the entire trilogy.' + #13#10 +
             PRESSKEY;
  MSGOFF = 'Messages OFF';
  MSGON = 'Messages ON';
  NETEND =  'you can''t end a netgame!' + #13#10 + #13#10 +
            PRESSKEY;
  SENDGAME  = 'are you sure you want to end the game?' + #13#10 + #13#10 +
             PRESSYN;

  DOSY = '(press y to quit)';

  DETAILHI = 'High detail';
  DETAILLO = 'Low detail';
  GAMMALVL0 = 'Gamma correction OFF';
  GAMMALVL1 = 'Gamma correction level 1';
  GAMMALVL2 = 'Gamma correction level 2';
  GAMMALVL3 = 'Gamma correction level 3';
  GAMMALVL4 = 'Gamma correction level 4';
  EMPTYSTRING = 'empty slot';

  { }
  {  P_inter.C }
  { }
  GOTARMOR = 'Picked up the armor.';
  GOTMEGA = 'Picked up the MegaArmor!';
  GOTHTHBONUS = 'Picked up a health bonus.';
  GOTARMBONUS = 'Picked up an armor bonus.';
  GOTSTIM = 'Picked up a stimpack.';
  GOTMEDINEED = 'Picked up a medikit that you REALLY need!';
  GOTMEDIKIT = 'Picked up a medikit.';
  GOTSUPER = 'Supercharge!';

  GOTBLUECARD = 'Picked up a blue keycard.';
  GOTYELWCARD = 'Picked up a yellow keycard.';
  GOTREDCARD = 'Picked up a red keycard.';
  GOTBLUESKUL = 'Picked up a blue skull key.';
  GOTYELWSKUL = 'Picked up a yellow skull key.';
  GOTREDSKULL = 'Picked up a red skull key.';

  GOTINVUL = 'Invulnerability!';
  GOTBERSERK = 'Berserk!';
  GOTINVIS = 'Partial Invisibility';
  GOTSUIT = 'Radiation Shielding Suit';
  GOTMAP = 'Computer Area Map';
  GOTVISOR = 'Light Amplification Visor';
  GOTMSPHERE = 'MegaSphere!';

  GOTCLIP = 'Picked up a clip.';
  GOTCLIPBOX = 'Picked up a box of bullets.';
  GOTROCKET = 'Picked up a rocket.';
  GOTROCKBOX = 'Picked up a box of rockets.';
  GOTCELL = 'Picked up an energy cell.';
  GOTCELLBOX = 'Picked up an energy cell pack.';
  GOTSHELLS = 'Picked up 4 shotgun shells.';
  GOTSHELLBOX = 'Picked up a box of shotgun shells.';
  GOTBACKPACK = 'Picked up a backpack full of ammo!';

  GOTBFG9000 = 'You got the BFG9000!  Oh, yes.';
  GOTCHAINGUN = 'You got the chaingun!';
  GOTCHAINSAW = 'A chainsaw!  Find some meat!';
  GOTLAUNCHER = 'You got the rocket launcher!';
  GOTPLASMA = 'You got the plasma gun!';
  GOTSHOTGUN = 'You got the shotgun!';
  GOTSHOTGUN2 = 'You got the super shotgun!';

  { }
  { P_Doors.C }
  { }
  PD_BLUEO = 'You need a blue key to activate this object';
  PD_REDO = 'You need a red key to activate this object';
  PD_YELLOWO = 'You need a yellow key to activate this object';
  PD_BLUEK = 'You need a blue key to open this door';
  PD_REDK = 'You need a red key to open this door';
  PD_YELLOWK = 'You need a yellow key to open this door';

  { }
  { G_game.C }
  { }
  GGSAVED = 'game saved.';

  { }
  {  HU_stuff.C }
  { }
  HUSTR_MSGU = '[Message unsent]';

  HUSTR_E1M1 = 'E1M1: Hangar';
  HUSTR_E1M2 = 'E1M2: Nuclear Plant';
  HUSTR_E1M3 = 'E1M3: Toxin Refinery';
  HUSTR_E1M4 = 'E1M4: Command Control';
  HUSTR_E1M5 = 'E1M5: Phobos Lab';
  HUSTR_E1M6 = 'E1M6: Central Processing';
  HUSTR_E1M7 = 'E1M7: Computer Station';
  HUSTR_E1M8 = 'E1M8: Phobos Anomaly';
  HUSTR_E1M9 = 'E1M9: Military Base';

  HUSTR_E2M1 = 'E2M1: Deimos Anomaly';
  HUSTR_E2M2 = 'E2M2: Containment Area';
  HUSTR_E2M3 = 'E2M3: Refinery';
  HUSTR_E2M4 = 'E2M4: Deimos Lab';
  HUSTR_E2M5 = 'E2M5: Command Center';
  HUSTR_E2M6 = 'E2M6: Halls of the Damned';
  HUSTR_E2M7 = 'E2M7: Spawning Vats';
  HUSTR_E2M8 = 'E2M8: Tower of Babel';
  HUSTR_E2M9 = 'E2M9: Fortress of Mystery';

  HUSTR_E3M1 = 'E3M1: Hell Keep';
  HUSTR_E3M2 = 'E3M2: Slough of Despair';
  HUSTR_E3M3 = 'E3M3: Pandemonium';
  HUSTR_E3M4 = 'E3M4: House of Pain';
  HUSTR_E3M5 = 'E3M5: Unholy Cathedral';
  HUSTR_E3M6 = 'E3M6: Mt. Erebus';
  HUSTR_E3M7 = 'E3M7: Limbo';
  HUSTR_E3M8 = 'E3M8: Dis';
  HUSTR_E3M9 = 'E3M9: Warrens';

  HUSTR_E4M1 = 'E4M1: Hell Beneath';
  HUSTR_E4M2 = 'E4M2: Perfect Hatred';
  HUSTR_E4M3 = 'E4M3: Sever The Wicked';
  HUSTR_E4M4 = 'E4M4: Unruly Evil';
  HUSTR_E4M5 = 'E4M5: They Will Repent';
  HUSTR_E4M6 = 'E4M6: Against Thee Wickedly';
  HUSTR_E4M7 = 'E4M7: And Hell Followed';
  HUSTR_E4M8 = 'E4M8: Unto The Cruel';
  HUSTR_E4M9 = 'E4M9: Fear';

  HUSTR_1 = 'level 1: entryway';
  HUSTR_2 = 'level 2: underhalls';
  HUSTR_3 = 'level 3: the gantlet';
  HUSTR_4 = 'level 4: the focus';
  HUSTR_5 = 'level 5: the waste tunnels';
  HUSTR_6 = 'level 6: the crusher';
  HUSTR_7 = 'level 7: dead simple';
  HUSTR_8 = 'level 8: tricks and traps';
  HUSTR_9 = 'level 9: the pit';
  HUSTR_10 = 'level 10: refueling base';
  HUSTR_11 = 'level 11: ''o'' of destruction!';

  HUSTR_12 = 'level 12: the factory';
  HUSTR_13 = 'level 13: downtown';
  HUSTR_14 = 'level 14: the inmost dens';
  HUSTR_15 = 'level 15: industrial zone';
  HUSTR_16 = 'level 16: suburbs';
  HUSTR_17 = 'level 17: tenements';
  HUSTR_18 = 'level 18: the courtyard';
  HUSTR_19 = 'level 19: the citadel';
  HUSTR_20 = 'level 20: gotcha!';

  HUSTR_21 = 'level 21: nirvana';
  HUSTR_22 = 'level 22: the catacombs';
  HUSTR_23 = 'level 23: barrels o'' fun';
  HUSTR_24 = 'level 24: the chasm';
  HUSTR_25 = 'level 25: bloodfalls';
  HUSTR_26 = 'level 26: the abandoned mines';
  HUSTR_27 = 'level 27: monster condo';
  HUSTR_28 = 'level 28: the spirit world';
  HUSTR_29 = 'level 29: the living end';
  HUSTR_30 = 'level 30: icon of sin';

  HUSTR_31 = 'level 31: wolfenstein';
  HUSTR_32 = 'level 32: grosse';

  PHUSTR_1 = 'level 1: congo';
  PHUSTR_2 = 'level 2: well of souls';
  PHUSTR_3 = 'level 3: aztec';
  PHUSTR_4 = 'level 4: caged';
  PHUSTR_5 = 'level 5: ghost town';
  PHUSTR_6 = 'level 6: baron''s lair';
  PHUSTR_7 = 'level 7: caughtyard';
  PHUSTR_8 = 'level 8: realm';
  PHUSTR_9 = 'level 9: abattoire';
  PHUSTR_10 = 'level 10: onslaught';
  PHUSTR_11 = 'level 11: hunted';

  PHUSTR_12 = 'level 12: speed';
  PHUSTR_13 = 'level 13: the crypt';
  PHUSTR_14 = 'level 14: genesis';
  PHUSTR_15 = 'level 15: the twilight';
  PHUSTR_16 = 'level 16: the omen';
  PHUSTR_17 = 'level 17: compound';
  PHUSTR_18 = 'level 18: neurosphere';
  PHUSTR_19 = 'level 19: nme';
  PHUSTR_20 = 'level 20: the death domain';

  PHUSTR_21 = 'level 21: slayer';
  PHUSTR_22 = 'level 22: impossible mission';
  PHUSTR_23 = 'level 23: tombstone';
  PHUSTR_24 = 'level 24: the final frontier';
  PHUSTR_25 = 'level 25: the temple of darkness';
  PHUSTR_26 = 'level 26: bunker';
  PHUSTR_27 = 'level 27: anti-christ';
  PHUSTR_28 = 'level 28: the sewers';
  PHUSTR_29 = 'level 29: odyssey of noises';
  PHUSTR_30 = 'level 30: the gateway of hell';

  PHUSTR_31 = 'level 31: cyberden';
  PHUSTR_32 = 'level 32: go 2 it';

  THUSTR_1 = 'level 1: system control';
  THUSTR_2 = 'level 2: human bbq';
  THUSTR_3 = 'level 3: power control';
  THUSTR_4 = 'level 4: wormhole';
  THUSTR_5 = 'level 5: hanger';
  THUSTR_6 = 'level 6: open season';
  THUSTR_7 = 'level 7: prison';
  THUSTR_8 = 'level 8: metal';
  THUSTR_9 = 'level 9: stronghold';
  THUSTR_10 = 'level 10: redemption';
  THUSTR_11 = 'level 11: storage facility';

  THUSTR_12 = 'level 12: crater';
  THUSTR_13 = 'level 13: nukage processing';
  THUSTR_14 = 'level 14: steel works';
  THUSTR_15 = 'level 15: dead zone';
  THUSTR_16 = 'level 16: deepest reaches';
  THUSTR_17 = 'level 17: processing area';
  THUSTR_18 = 'level 18: mill';
  THUSTR_19 = 'level 19: shipping/respawning';
  THUSTR_20 = 'level 20: central processing';

  THUSTR_21 = 'level 21: administration center';
  THUSTR_22 = 'level 22: habitat';
  THUSTR_23 = 'level 23: lunar mining project';
  THUSTR_24 = 'level 24: quarry';
  THUSTR_25 = 'level 25: baron''s den';
  THUSTR_26 = 'level 26: ballistyx';
  THUSTR_27 = 'level 27: mount pain';
  THUSTR_28 = 'level 28: heck';
  THUSTR_29 = 'level 29: river styx';
  THUSTR_30 = 'level 30: last call';

  THUSTR_31 = 'level 31: pharaoh';
  THUSTR_32 = 'level 32: caribbean';

  HUSTR_CHATMACRO1 = 'I''m ready to kick butt!';
  HUSTR_CHATMACRO2 = 'I''m OK.';
  HUSTR_CHATMACRO3 = 'I''m not looking too good!';
  HUSTR_CHATMACRO4 = 'Help!';
  HUSTR_CHATMACRO5 = 'You suck!';
  HUSTR_CHATMACRO6 = 'Next time, scumbag...';
  HUSTR_CHATMACRO7 = 'Come here!';
  HUSTR_CHATMACRO8 = 'I''ll take care of it.';
  HUSTR_CHATMACRO9 = 'Yes';
  HUSTR_CHATMACRO0 = 'No';

  HUSTR_TALKTOSELF1 = 'You mumble to yourself';
  HUSTR_TALKTOSELF2 = 'Who''s there?';
  HUSTR_TALKTOSELF3 = 'You scare yourself';
  HUSTR_TALKTOSELF4 = 'You start to rave';
  HUSTR_TALKTOSELF5 = 'You''ve lost it...';

  HUSTR_MESSAGESENT = '[Message Sent]';

  { The following should NOT be changed unless it seems }
  { just AWFULLY necessary }
  HUSTR_PLRGREEN = 'Green: ';
  HUSTR_PLRINDIGO = 'Indigo: ';
  HUSTR_PLRBROWN = 'Brown: ';
  HUSTR_PLRRED = 'Red: ';

  HUSTR_KEYGREEN = 'g';
  HUSTR_KEYINDIGO = 'i';
  HUSTR_KEYBROWN = 'b';
  HUSTR_KEYRED = 'r';

  { }
  {  AM_map.C }
  { }
  AMSTR_FOLLOWON = 'Follow Mode ON';
  AMSTR_FOLLOWOFF = 'Follow Mode OFF';
  AMSTR_GRIDON = 'Grid ON';
  AMSTR_GRIDOFF = 'Grid OFF';
  AMSTR_MARKEDSPOT = 'Marked Spot';
  AMSTR_MARKSCLEARED = 'All Marks Cleared';

  { }
  {  ST_stuff.C }
  { }
  STSTR_MUS = 'Music Change';
  STSTR_NOMUS = 'IMPOSSIBLE SELECTION';
  STSTR_DQDON = 'Degreelessness Mode On';
  STSTR_DQDOFF = 'Degreelessness Mode Off';

  STSTR_KFAADDED = 'Very Happy Ammo Added';
  STSTR_FAADDED = 'Ammo (no keys) Added';

  STSTR_NCON = 'No Clipping Mode ON';
  STSTR_NCOFF = 'No Clipping Mode OFF';

  STSTR_BEHOLD = 'inVuln, Str, Inviso, Rad, Allmap, or Lite-amp';
  STSTR_BEHOLDX = 'Power-up Toggled';

  STSTR_CHOPPERS = '... doesn''t suck - GM';
  STSTR_CLEV = 'Changing Level...';

  { }
  { F_Finale.C }
  { }
  E1TEXT =
    'Once you beat the big badasses and' + #13#10 +
    'clean out the moon base you''re supposed' + #13#10 +
    'to win, aren''t you? Aren''t you? Where''s' + #13#10 +
    'your fat reward and ticket home? What' + #13#10 +
    'the hell is this? It''s not supposed to' + #13#10 +
    'end this way!' + #13#10 +
    ' ' + #13#10 +
    'It stinks like rotten meat, but looks' + #13#10 +
    'like the lost Deimos base.  Looks like' + #13#10 +
    'you''re stuck on The Shores of Hell.' + #13#10 +
    'The only way out is through.' + #13#10 +
    ' ' + #13#10 +
    'To continue the DOOM experience, play' + #13#10 +
    'The Shores of Hell and its amazing' + #13#10 +
    'sequel, Inferno!' + #13#10;


  E2TEXT =
    'You''ve done it! The hideous cyber-' + #13#10 +
    'demon lord that ruled the lost Deimos' + #13#10 +
    'moon base has been slain and you' + #13#10 +
    'are triumphant! But ... where are' + #13#10 +
    'you? You clamber to the edge of the' + #13#10 +
    'moon and look down to see the awful' + #13#10 +
    'truth.' + #13#10 +
    ' ' + #13#10 +
    'Deimos floats above Hell itself!' + #13#10 +
    'You''ve never heard of anyone escaping' + #13#10 +
    'from Hell, but you''ll make the bastards' + #13#10 +
    'sorry they ever heard of you! Quickly,' + #13#10 +
    'you rappel down to  the surface of' + #13#10 +
    'Hell.' + #13#10 +
    ' ' + #13#10 +
    'Now, it''s on to the final chapter of' + #13#10 +
    'DOOM! -- Inferno.';


  E3TEXT =
    'The loathsome spiderdemon that' + #13#10 +
    'masterminded the invasion of the moon' + #13#10 +
    'bases and caused so much death has had' + #13#10 +
    'its ass kicked for all time.' + #13#10 +
    '' + #13#10 +
    'A hidden doorway opens and you enter.' + #13#10 +
    'You''ve proven too tough for Hell to' + #13#10 +
    'contain, and now Hell at last plays' + #13#10 +
    'fair -- for you emerge from the door' + #13#10 +
    'to see the green fields of Earth!' + #13#10 +
    'Home at last.' + #13#10 +
    ' ' + #13#10 +
    'You wonder what''s been happening on' + #13#10 +
    'Earth while you were battling evil' + #13#10 +
    'unleashed. It''s good that no Hell-' + #13#10 +
    'spawn could have come through that' + #13#10 +
    'door with you ...';


  E4TEXT =
    'the spider mastermind must have sent forth' + #13#10 +
    'its legions of hellspawn before your' + #13#10 +
    'final confrontation with that terrible' + #13#10 +
    'beast from hell.  but you stepped forward' + #13#10 +
    'and brought forth eternal damnation and' + #13#10 +
    'suffering upon the horde as a true hero' + #13#10 +
    'would in the face of something so evil.' + #13#10 +
    ' ' + #13#10 +
    'besides, someone was gonna pay for what' + #13#10 +
    'happened to daisy, your pet rabbit.' + #13#10 +
    ' ' + #13#10 +
    'but now, you see spread before you more' + #13#10 +
    'potential pain and gibbitude as a nation' + #13#10 +
    'of demons run amok among our cities.' + #13#10 +
    ' ' + #13#10 +
    'next stop, hell on earth!';


// after level 6, put this:

  C1TEXT =
    'YOU HAVE ENTERED DEEPLY INTO THE INFESTED' + #13#10 +
    'STARPORT. BUT SOMETHING IS WRONG. THE' + #13#10 +
    'MONSTERS HAVE BROUGHT THEIR OWN REALITY' + #13#10 +
    'WITH THEM, AND THE STARPORT''S TECHNOLOGY' + #13#10 +
    'IS BEING SUBVERTED BY THEIR PRESENCE.' + #13#10 +
    ' ' + #13#10 +
    'AHEAD, YOU SEE AN OUTPOST OF HELL, A' + #13#10 +
    'FORTIFIED ZONE. IF YOU CAN GET PAST IT,' + #13#10 +
    'YOU CAN PENETRATE INTO THE HAUNTED HEART' + #13#10 +
    'OF THE STARBASE AND FIND THE CONTROLLING' + #13#10 +
    'SWITCH WHICH HOLDS EARTH''S POPULATION' + #13#10 +
    'HOSTAGE.';

// After level 11, put this:

  C2TEXT =
    'YOU HAVE WON! YOUR VICTORY HAS ENABLED' + #13#10 +
    'HUMANKIND TO EVACUATE EARTH AND ESCAPE' + #13#10 +
    'THE NIGHTMARE.  NOW YOU ARE THE ONLY' + #13#10 +
    'HUMAN LEFT ON THE FACE OF THE PLANET.' + #13#10 +
    'CANNIBAL MUTATIONS, CARNIVOROUS ALIENS,' + #13#10 +
    'AND EVIL SPIRITS ARE YOUR ONLY NEIGHBORS.' + #13#10 +
    'YOU SIT BACK AND WAIT FOR DEATH, CONTENT' + #13#10 +
    'THAT YOU HAVE SAVED YOUR SPECIES.' + #13#10 +
    ' ' + #13#10 +
    'BUT THEN, EARTH CONTROL BEAMS DOWN A' + #13#10 +
    'MESSAGE FROM SPACE: ''SENSORS HAVE LOCATED' + #13#10 +
    'THE SOURCE OF THE ALIEN INVASION. IF YOU' + #13#10 +
    'GO THERE, YOU MAY BE ABLE TO BLOCK THEIR' + #13#10 +
    'ENTRY.  THE ALIEN BASE IS IN THE HEART OF' + #13#10 +
    'YOUR OWN HOME CITY, NOT FAR FROM THE' + #13#10 +
    'STARPORT.'' SLOWLY AND PAINFULLY YOU GET' + #13#10 +
    'UP AND RETURN TO THE FRAY.';


// After level 20, put this:

  C3TEXT =
    'YOU ARE AT THE CORRUPT HEART OF THE CITY,' + #13#10 +
    'SURROUNDED BY THE CORPSES OF YOUR ENEMIES.' + #13#10 +
    'YOU SEE NO WAY TO DESTROY THE CREATURES' + #13#10 +
    'ENTRYWAY ON THIS SIDE, SO YOU CLENCH YOUR' + #13#10 +
    'TEETH AND PLUNGE THROUGH IT.' + #13#10 +
    ' ' + #13#10 +
    'THERE MUST BE A WAY TO CLOSE IT ON THE' + #13#10 +
    'OTHER SIDE. WHAT DO YOU CARE IF YOU''VE' + #13#10 +
    'GOT TO GO THROUGH HELL TO GET TO IT?';


// After level 29, put this:

  C4TEXT =
    'THE HORRENDOUS VISAGE OF THE BIGGEST' + #13#10 +
    'DEMON YOU''VE EVER SEEN CRUMBLES BEFORE' + #13#10 +
    'YOU, AFTER YOU PUMP YOUR ROCKETS INTO' + #13#10 +
    'HIS EXPOSED BRAIN. THE MONSTER SHRIVELS' + #13#10 +
    'UP AND DIES, ITS THRASHING LIMBS' + #13#10 +
    'DEVASTATING UNTOLD MILES OF HELL''S' + #13#10 +
    'SURFACE.' + #13#10 +
    ' ' + #13#10 +
    'YOU''VE DONE IT. THE INVASION IS OVER.' + #13#10 +
    'EARTH IS SAVED. HELL IS A WRECK. YOU' + #13#10 +
    'WONDER WHERE BAD FOLKS WILL GO WHEN THEY' + #13#10 +
    'DIE, NOW. WIPING THE SWEAT FROM YOUR' + #13#10 +
    'FOREHEAD YOU BEGIN THE LONG TREK BACK' + #13#10 +
    'HOME. REBUILDING EARTH OUGHT TO BE A' + #13#10 +
    'LOT MORE FUN THAN RUINING IT WAS.' + #13#10;



// Before level 31, put this:

  C5TEXT =
    'CONGRATULATIONS, YOU''VE FOUND THE SECRET' + #13#10 +
    'LEVEL! LOOKS LIKE IT''S BEEN BUILT BY' + #13#10 +
    'HUMANS, RATHER THAN DEMONS. YOU WONDER' + #13#10 +
    'WHO THE INMATES OF THIS CORNER OF HELL' + #13#10 +
    'WILL BE.';


// Before level 32, put this:

  C6TEXT =
    'CONGRATULATIONS, YOU''VE FOUND THE' + #13#10 +
    'SUPER SECRET LEVEL!  YOU''D BETTER' + #13#10 +
    'BLAZE THROUGH THIS ONE!' + #13#10;


// after map 06

  P1TEXT =
    'You gloat over the steaming carcass of the' + #13#10 +
    'Guardian.  With its death, you''ve wrested' + #13#10 +
    'the Accelerator from the stinking claws' + #13#10 +
    'of Hell.  You relax and glance around the' + #13#10 +
    'room.  Damn!  There was supposed to be at' + #13#10 +
    'least one working prototype, but you can''t' + #13#10 +
    'see it. The demons must have taken it.' + #13#10 +
    ' ' + #13#10 +
    'You must find the prototype, or all your' + #13#10 +
    'struggles will have been wasted. Keep' + #13#10 +
    'moving, keep fighting, keep killing.' + #13#10 +
    'Oh yes, keep living, too.';


// after map 11

  P2TEXT =
    'Even the deadly Arch-Vile labyrinth could' + #13#10 +
    'not stop you, and you''ve gotten to the' + #13#10 +
    'prototype Accelerator which is soon' + #13#10 +
    'efficiently and permanently deactivated.' + #13#10 +
    ' ' + #13#10 +
    'You''re good at that kind of thing.';


// after map 20

  P3TEXT =
    'You''ve bashed and battered your way into' + #13#10 +
    'the heart of the devil-hive.  Time for a' + #13#10 +
    'Search-and-Destroy mission, aimed at the' + #13#10 +
    'Gatekeeper, whose foul offspring is' + #13#10 +
    'cascading to Earth.  Yeah, he''s bad. But' + #13#10 +
    'you know who''s worse!' + #13#10 +
    ' ' + #13#10 +
    'Grinning evilly, you check your gear, and' + #13#10 +
    'get ready to give the bastard a little Hell' + #13#10 +
    'of your own making!';

// after map 30

  P4TEXT =
    'The Gatekeeper''s evil face is splattered' + #13#10 +
    'all over the place.  As its tattered corpse' + #13#10 +
    'collapses, an inverted Gate forms and' + #13#10 +
    'sucks down the shards of the last' + #13#10 +
    'prototype Accelerator, not to mention the' + #13#10 +
    'few remaining demons.  You''re done. Hell' + #13#10 +
    'has gone back to pounding bad dead folks ' + #13#10 +
    'instead of good live ones.  Remember to' + #13#10 +
    'tell your grandkids to put a rocket' + #13#10 +
    'launcher in your coffin. If you go to Hell' + #13#10 +
    'when you die, you''ll need it for some' + #13#10 +
    'final cleaning-up ...';

// before map 31

  P5TEXT =
    'You''ve found the second-hardest level we' + #13#10 +
    'got. Hope you have a saved game a level or' + #13#10 +
    'two previous.  If not, be prepared to die' + #13#10 +
    'aplenty. For master marines only.';

// before map 32

  P6TEXT =
    'Betcha wondered just what WAS the hardest' + #13#10 +
    'level we had ready for ya?  Now you know.' + #13#10 +
    'No one gets out alive.';


  T1TEXT =
    'You''ve fought your way out of the infested' + #13#10 +
    'experimental labs.   It seems that UAC has' + #13#10 +
    'once again gulped it down.  With their' + #13#10 +
    'high turnover, it must be hard for poor' + #13#10 +
    'old UAC to buy corporate health insurance' + #13#10 +
    'nowadays..' + #13#10 +
    ' ' + #13#10 +
    'Ahead lies the military complex, now' + #13#10 +
    'swarming with diseased horrors hot to get' + #13#10 +
    'their teeth into you. With luck, the' + #13#10 +
    'complex still has some warlike ordnance' + #13#10 +
    'laying around.';


  T2TEXT =
    'You hear the grinding of heavy machinery' + #13#10 +
    'ahead.  You sure hope they''re not stamping' + #13#10 +
    'out new hellspawn, but you''re ready to' + #13#10 +
    'ream out a whole herd if you have to.' + #13#10 +
    'They might be planning a blood feast, but' + #13#10 +
    'you feel about as mean as two thousand' + #13#10 +
    'maniacs packed into one mad killer.' + #13#10 +
    ' ' + #13#10 +
    'You don''t plan to go down easy.';


  T3TEXT =
    'The vista opening ahead looks real damn' + #13#10 +
    'familiar. Smells familiar, too -- like' + #13#10 +
    'fried excrement. You didn''t like this' + #13#10 +
    'place before, and you sure as hell ain''t' + #13#10 +
    'planning to like it now. The more you' + #13#10 +
    'brood on it, the madder you get.' + #13#10 +
    'Hefting your gun, an evil grin trickles' + #13#10 +
    'onto your face. Time to take some names.';

  T4TEXT =
    'Suddenly, all is silent, from one horizon' + #13#10 +
    'to the other. The agonizing echo of Hell' + #13#10 +
    'fades away, the nightmare sky turns to' + #13#10 +
    'blue, the heaps of monster corpses start ' + #13#10 +
    'to evaporate along with the evil stench ' + #13#10 +
    'that filled the air. Jeeze, maybe you''ve' + #13#10 +
    'done it. Have you really won?' + #13#10 +
    ' ' + #13#10 +
    'Something rumbles in the distance.' + #13#10 +
    'A blue light begins to glow inside the' + #13#10 +
    'ruined skull of the demon-spitter.';


  T5TEXT =
    'What now? Looks totally different. Kind' + #13#10 +
    'of like King Tut''s condo. Well,' + #13#10 +
    'whatever''s here can''t be any worse' + #13#10 +
    'than usual. Can it?  Or maybe it''s best' + #13#10 +
    'to let sleeping gods lie..';


  T6TEXT =
    'Time for a vacation. You''ve burst the' + #13#10 +
    'bowels of hell and by golly you''re ready' + #13#10 +
    'for a break. You mutter to yourself,' + #13#10 +
    'Maybe someone else can kick Hell''s ass' + #13#10 +
    'next time around. Ahead lies a quiet town,' + #13#10 +
    'with peaceful flowing water, quaint' + #13#10 +
    'buildings, and presumably no Hellspawn.' + #13#10 +
    ' ' + #13#10 +
    'As you step off the transport, you hear' + #13#10 +
    'the stomp of a cyberdemon''s iron shoe.';


//
// Character cast strings F_FINALE.C
//
  CC_ZOMBIE  = 'ZOMBIEMAN';
  CC_SHOTGUN = 'SHOTGUN GUY';
  CC_HEAVY = 'HEAVY WEAPON DUDE';
  CC_IMP = 'IMP';
  CC_DEMON = 'DEMON';
  CC_LOST = 'LOST SOUL';
  CC_CACO = 'CACODEMON';
  CC_HELL = 'HELL KNIGHT';
  CC_BARON = 'BARON OF HELL';
  CC_ARACH = 'ARACHNOTRON';
  CC_PAIN = 'PAIN ELEMENTAL';
  CC_REVEN = 'REVENANT';
  CC_MANCU = 'MANCUBUS';
  CC_ARCH = 'ARCH-VILE';
  CC_SPIDER = 'THE SPIDER MASTERMIND';
  CC_CYBER = 'THE CYBERDEMON';
  CC_HERO = 'OUR HERO';

  MSG_MODIFIEDGAME =
      '===========================================================================' + #13#10 +
      'ATTENTION:  This version of DOOM has been modified.  If you would like to' + #13#10 +
      'get a copy of the original game, call 1-800-IDGAMES or see the readme file.' + #13#10 +
      '        You will not receive technical support for modified games.' + #13#10 +
      '                      press enter to continue' + #13#10 +
      '===========================================================================' + #13#10;

  MSG_SHAREWARE =
        '===========================================================================' + #13#10 +
        '                                Shareware!' + #13#10 +
        '===========================================================================' + #13#10;

  MSG_COMMERCIAL =
        '===========================================================================' + #13#10 +
        '                 Commercial product - do not distribute!' + #13#10 +
        '         Please report software piracy to the SPA: 1-800-388-PIR8' + #13#10 +
        '===========================================================================' + #13#10;

  MSG_UNDETERMINED =
        '===========================================================================' + #13#10 +
        '                       Undetermined version! (Ouch)' + #13#10 +
        '===========================================================================' + #13#10;



(* Unit i_main interface *)


                             

{
    i_main.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//   Main program, simply calls D_DoomMain high level loop. 
//  
//----------------------------------------------------------------------------- 

var
  hMainWnd: HWND = 0;

const
  AppTitle = 'Delphi Doom';

procedure DoomMain;


(* Unit sounds interface *)


{
    sounds.h, sounds.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Created by the sound utility written by Dave Taylor.
//	Kept as a sample, DOOM2  sounds. Frozen.
//
//-----------------------------------------------------------------------------

//
// SoundFX struct.
//

type
  Psfxinfo_t = ^sfxinfo_t;

  sfxinfo_t = record
    // up to 6-character name
    name: string;

    // Sfx singularity (only one at a time)
    singularity: boolean;

    // Sfx priority
    priority: integer;

    // referenced sound if a link
    link: Psfxinfo_t;

    // pitch if a link
    pitch: integer;

    // volume if a link
    volume: integer;

    // sound data
    data: pointer;

    // this is checked every second to see if sound
    // can be thrown out (if 0, then decrement, if -1,
    // then throw out, if > 0, then it is in use)
    usefulness: integer;

    // lump number of sfx
    lumpnum: integer;
  end;

//
// MusicInfo struct.
//
  musicinfo_t = record
    // up to 6-character name
    name: string;

    // lump number of music
    lumpnum: integer;

    // music data
    data: pointer;

    // music handle once registered
    handle: integer;
  end;
  Pmusicinfo_t = ^musicinfo_t;

//
// Identifiers for all music in game.
//

  musicenum_t = (
    mus_None,
    mus_e1m1,
    mus_e1m2,
    mus_e1m3,
    mus_e1m4,
    mus_e1m5,
    mus_e1m6,
    mus_e1m7,
    mus_e1m8,
    mus_e1m9,
    mus_e2m1,
    mus_e2m2,
    mus_e2m3,
    mus_e2m4,
    mus_e2m5,
    mus_e2m6,
    mus_e2m7,
    mus_e2m8,
    mus_e2m9,
    mus_e3m1,
    mus_e3m2,
    mus_e3m3,
    mus_e3m4,
    mus_e3m5,
    mus_e3m6,
    mus_e3m7,
    mus_e3m8,
    mus_e3m9,
    mus_inter,
    mus_intro,
    mus_bunny,
    mus_victor,
    mus_introa,
    mus_runnin,
    mus_stalks,
    mus_countd,
    mus_betwee,
    mus_doom,
    mus_the_da,
    mus_shawn,
    mus_ddtblu,
    mus_in_cit,
    mus_dead,
    mus_stlks2,
    mus_theda2,
    mus_doom2,
    mus_ddtbl2,
    mus_runni2,
    mus_dead2,
    mus_stlks3,
    mus_romero,
    mus_shawn2,
    mus_messag,
    mus_count2,
    mus_ddtbl3,
    mus_ampie,
    mus_theda3,
    mus_adrian,
    mus_messg2,
    mus_romer2,
    mus_tense,
    mus_shawn3,
    mus_openin,
    mus_evil,
    mus_ultima,
    mus_read_m,
    mus_dm2ttl,
    mus_dm2int,
    NUMMUSIC
  );


//
// Identifiers for all sfx in game.
//

  sfxenum_t = (
    sfx_None,
    sfx_pistol,
    sfx_shotgn,
    sfx_sgcock,
    sfx_dshtgn,
    sfx_dbopn,
    sfx_dbcls,
    sfx_dbload,
    sfx_plasma,
    sfx_bfg,
    sfx_sawup,
    sfx_sawidl,
    sfx_sawful,
    sfx_sawhit,
    sfx_rlaunc,
    sfx_rxplod,
    sfx_firsht,
    sfx_firxpl,
    sfx_pstart,
    sfx_pstop,
    sfx_doropn,
    sfx_dorcls,
    sfx_stnmov,
    sfx_swtchn,
    sfx_swtchx,
    sfx_plpain,
    sfx_dmpain,
    sfx_popain,
    sfx_vipain,
    sfx_mnpain,
    sfx_pepain,
    sfx_slop,
    sfx_itemup,
    sfx_wpnup,
    sfx_oof,
    sfx_telept,
    sfx_posit1,
    sfx_posit2,
    sfx_posit3,
    sfx_bgsit1,
    sfx_bgsit2,
    sfx_sgtsit,
    sfx_cacsit,
    sfx_brssit,
    sfx_cybsit,
    sfx_spisit,
    sfx_bspsit,
    sfx_kntsit,
    sfx_vilsit,
    sfx_mansit,
    sfx_pesit,
    sfx_sklatk,
    sfx_sgtatk,
    sfx_skepch,
    sfx_vilatk,
    sfx_claw,
    sfx_skeswg,
    sfx_pldeth,
    sfx_pdiehi,
    sfx_podth1,
    sfx_podth2,
    sfx_podth3,
    sfx_bgdth1,
    sfx_bgdth2,
    sfx_sgtdth,
    sfx_cacdth,
    sfx_skldth,
    sfx_brsdth,
    sfx_cybdth,
    sfx_spidth,
    sfx_bspdth,
    sfx_vildth,
    sfx_kntdth,
    sfx_pedth,
    sfx_skedth,
    sfx_posact,
    sfx_bgact,
    sfx_dmact,
    sfx_bspact,
    sfx_bspwlk,
    sfx_vilact,
    sfx_noway,
    sfx_barexp,
    sfx_punch,
    sfx_hoof,
    sfx_metal,
    sfx_chgun,
    sfx_tink,
    sfx_bdopn,
    sfx_bdcls,
    sfx_itmbk,
    sfx_flame,
    sfx_flamst,
    sfx_getpow,
    sfx_bospit,
    sfx_boscub,
    sfx_bossit,
    sfx_bospn,
    sfx_bosdth,
    sfx_manatk,
    sfx_mandth,
    sfx_sssit,
    sfx_ssdth,
    sfx_keenpn,
    sfx_keendt,
    sfx_skeact,
    sfx_skesit,
    sfx_skeatk,
    sfx_radio,
    NUMSFX
  );

var
  S_music: array[0..67] of musicinfo_t = (
    (name: '';       lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m1';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m2';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m3';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m4';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m5';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m6';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m7';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m8';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e1m9';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m1';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m2';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m3';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m4';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m5';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m6';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m7';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m8';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e2m9';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m1';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m2';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m3';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m4';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m5';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m6';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m7';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m8';   lumpnum: 0; data: nil; handle: 0),
    (name: 'e3m9';   lumpnum: 0; data: nil; handle: 0),
    (name: 'inter';  lumpnum: 0; data: nil; handle: 0),
    (name: 'intro';  lumpnum: 0; data: nil; handle: 0),
    (name: 'bunny';  lumpnum: 0; data: nil; handle: 0),
    (name: 'victor'; lumpnum: 0; data: nil; handle: 0),
    (name: 'introa'; lumpnum: 0; data: nil; handle: 0),
    (name: 'runnin'; lumpnum: 0; data: nil; handle: 0),
    (name: 'stalks'; lumpnum: 0; data: nil; handle: 0),
    (name: 'countd'; lumpnum: 0; data: nil; handle: 0),
    (name: 'betwee'; lumpnum: 0; data: nil; handle: 0),
    (name: 'doom';   lumpnum: 0; data: nil; handle: 0),
    (name: 'the_da'; lumpnum: 0; data: nil; handle: 0),
    (name: 'shawn';  lumpnum: 0; data: nil; handle: 0),
    (name: 'ddtblu'; lumpnum: 0; data: nil; handle: 0),
    (name: 'in_cit'; lumpnum: 0; data: nil; handle: 0),
    (name: 'dead';   lumpnum: 0; data: nil; handle: 0),
    (name: 'stlks2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'theda2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'doom2';  lumpnum: 0; data: nil; handle: 0),
    (name: 'ddtbl2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'runni2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'dead2';  lumpnum: 0; data: nil; handle: 0),
    (name: 'stlks3'; lumpnum: 0; data: nil; handle: 0),
    (name: 'romero'; lumpnum: 0; data: nil; handle: 0),
    (name: 'shawn2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'messag'; lumpnum: 0; data: nil; handle: 0),
    (name: 'count2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'ddtbl3'; lumpnum: 0; data: nil; handle: 0),
    (name: 'ampie';  lumpnum: 0; data: nil; handle: 0),
    (name: 'theda3'; lumpnum: 0; data: nil; handle: 0),
    (name: 'adrian'; lumpnum: 0; data: nil; handle: 0),
    (name: 'messg2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'romer2'; lumpnum: 0; data: nil; handle: 0),
    (name: 'tense';  lumpnum: 0; data: nil; handle: 0),
    (name: 'shawn3'; lumpnum: 0; data: nil; handle: 0),
    (name: 'openin'; lumpnum: 0; data: nil; handle: 0),
    (name: 'evil';   lumpnum: 0; data: nil; handle: 0),
    (name: 'ultima'; lumpnum: 0; data: nil; handle: 0),
    (name: 'read_m'; lumpnum: 0; data: nil; handle: 0),
    (name: 'dm2ttl'; lumpnum: 0; data: nil; handle: 0),
    (name: 'dm2int'; lumpnum: 0; data: nil; handle: 0)
  );

  S_sfx: array[0..108] of sfxinfo_t = (
  // S_sfx[0] needs to be a dummy for odd reasons.
    (name: 'none';   singularity: false; priority:   0; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),

    (name: 'pistol'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'shotgn'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sgcock'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dshtgn'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dbopn';  singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dbcls';  singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dbload'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'plasma'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bfg';    singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sawup';  singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sawidl'; singularity: false; priority: 118; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sawful'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sawhit'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'rlaunc'; singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'rxplod'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'firsht'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'firxpl'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pstart'; singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pstop';  singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'doropn'; singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dorcls'; singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'stnmov'; singularity: false; priority: 119; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'swtchn'; singularity: false; priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'swtchx'; singularity: false; priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'plpain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dmpain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'popain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'vipain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'mnpain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pepain'; singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'slop';   singularity: false; priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'itemup'; singularity: true;  priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'wpnup';  singularity: true;  priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'oof';    singularity: false; priority:  96; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'telept'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'posit1'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'posit2'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'posit3'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bgsit1'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bgsit2'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sgtsit'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'cacsit'; singularity: true;  priority:  98; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'brssit'; singularity: true;  priority:  94; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'cybsit'; singularity: true;  priority:  92; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'spisit'; singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bspsit'; singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'kntsit'; singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'vilsit'; singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'mansit'; singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pesit';  singularity: true;  priority:  90; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sklatk'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sgtatk'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skepch'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'vilatk'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'claw';   singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skeswg'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pldeth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pdiehi'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'podth1'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'podth2'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'podth3'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bgdth1'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bgdth2'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sgtdth'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'cacdth'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skldth'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'brsdth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'cybdth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'spidth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bspdth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'vildth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'kntdth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'pedth';  singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skedth'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'posact'; singularity: true;  priority: 120; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bgact';  singularity: true;  priority: 120; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'dmact';  singularity: true;  priority: 120; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bspact'; singularity: true;  priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bspwlk'; singularity: true;  priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'vilact'; singularity: true;  priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'noway';  singularity: false; priority:  78; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'barexp'; singularity: false; priority:  60; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'punch';  singularity: false; priority:  64; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'hoof';   singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'metal';  singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'chgun';  singularity: false; priority:  64; link: @S_sfx[Ord(sfx_pistol)]; pitch: 150; volume: 0; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'tink';   singularity: false; priority:  60; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bdopn';  singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bdcls';  singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'itmbk';  singularity: false; priority: 100; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'flame';  singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'flamst'; singularity: false; priority:  32; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'getpow'; singularity: false; priority:  60; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bospit'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'boscub'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bossit'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bospn';  singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'bosdth'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'manatk'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'mandth'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'sssit';  singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'ssdth';  singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'keenpn'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'keendt'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skeact'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skesit'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'skeatk'; singularity: false; priority:  70; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0),
    (name: 'radio';  singularity: false; priority:  60; link: nil; pitch: -1; volume: -1; data: nil; usefulness: 0; lumpnum: 0)
  );



(* Unit i_sound interface *)


             

{
    i_sound.h, i_sound.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
//
// DESCRIPTION:
//	System interface, sound.
//
//-----------------------------------------------------------------------------

// Init at program start...
procedure I_InitSound;

// ... update sound buffer and audio device at runtime...
procedure I_UpdateSound;
procedure I_SubmitSound;

// ... shut down and relase at program termination.
procedure I_ShutdownSound;

//
//  SFX I/O
//

// Initialize channels?
procedure I_SetChannels;

// Get raw data lump index for sound descriptor.
function I_GetSfxLumpNum(sfxinfo: Psfxinfo_t): integer;


// Starts a sound in a particular sound channel.
function I_StartSound(id: integer; vol: integer; sep: integer;
  pitch: integer; priority: integer): integer;


// Stops a sound channel.
procedure I_StopSound(handle: integer);

// Called by S_*() functions
//  to see if a channel is still playing.
// Returns 0 if no longer playing, 1 if playing.
function I_SoundIsPlaying(handle: integer): boolean;

// Updates the volume, separation,
//  and pitch of a sound channel.
procedure I_UpdateSoundParams(handle: integer; vol: integer; sep: integer;
  pitch: integer);



(* Unit dstrings interface *)


{
    dstrings.h, dstrings.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details.
// 
// 
// $Log:$ 
// 
// DESCRIPTION: 
//	DOOM strings, by language. 
//
// DESCRIPTION:
// Globally defined strings.
//
//-----------------------------------------------------------------------------


const
// Misc. other strings.
  SAVEGAMENAME = 'doomsav';

//
// File locations,
// relative to current position.
// Path names are OS-sensitive.
//
  DEVMAPS = 'devmaps';
  DEVDATA = 'devdata';

// Not done in french?

// QuitDOOM messages
  NUM_QUITMESSAGES = 22;

var
  endmsg: array[0..NUM_QUITMESSAGES] of string;


(* Unit p_pspr_h interface *)


                      

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//
//-----------------------------------------------------------------------------

type
//
// Overlay psprites are scaled shapes
// drawn directly on the view screen,
// coordinates are given for a 320*200 view screen.
//
  psprnum_t = (
    ps_weapon,
    ps_flash,
    NUMPSPRITES
  );


  pspdef_t = record
    state: Pstate_t; // a NULL state means not active
    tics: integer;
    sx: fixed_t;
    sy: fixed_t;
  end;
  Ppspdef_t = ^pspdef_t;



(* Unit d_event interface *)


               

{
    d_event.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details. 
// 
// DESCRIPTION: 
// 
//     
//----------------------------------------------------------------------------- 

// 
// Event handling. 
//
// Input event types.

type
  evtype_t = (ev_keydown, ev_keyup, ev_mouse, ev_joystick);

// Event structure. 
// keys / mouse/joystick buttons 
// mouse/joystick x move 
// mouse/joystick y move 

  event_t = record
    _type : evtype_t;
    data1 : integer;
    data2 : integer;
    data3 : integer;
  end;
  Pevent_t = ^event_t;

  gameaction_t = (ga_nothing, ga_loadlevel, ga_newgame, ga_loadgame,
                  ga_savegame, ga_playdemo, ga_completed,
                  ga_victory, ga_worlddone, ga_screenshot);

// 
// Button/action code definitions. 
// 
// Press "Fire". 
// Use button, to open doors, activate switches. 
// Flag: game events, not really buttons. 
// Flag, weapon change pending. 
// If true, the next 3 bits hold weapon num. 
// The 3bit weapon mask and shift, convenience. 
// Pause the game. 
// Save the game at each console. 
// Savegame slot numbers 
//  occupy the second byte of buttons.

const
    // Press "Fire".
    BT_ATTACK = 1;
    // Use button, to open doors, activate switches.
    BT_USE = 2;

    // Flag: game events, not really buttons.
    BT_SPECIAL = 128;
    BT_SPECIALMASK = 3;

    // Flag, weapon change pending.
    // If true, the next 3 bits hold weapon num.
    BT_CHANGE = 4;
    // The 3bit weapon mask and shift, convenience.
    BT_WEAPONMASK = (8+16+32);
    BT_WEAPONSHIFT = 3;

    // Pause the game.
    BTS_PAUSE = 1;
    // Save the game at each console.
    BTS_SAVEGAME = 2;

    // Savegame slot numbers
    //  occupy the second byte of buttons.
    BTS_SAVEMASK = (4+8+16);
    BTS_SAVESHIFT = 2;

//
// GLOBAL VARIABLES 
//

const
  MAXEVENTS = 256;

var
//
// EVENT HANDLING
//
// Events are asynchronous inputs generally generated by the game user.
// Events can be discarded if no responder claims them
//
  events: array[0..(MAXEVENTS)-1] of event_t;
  eventhead: integer;
  eventtail: integer;


(* Unit i_system interface *)


                                     

{
    i_system.h, i_system.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  System specific interface stuff. 
// 
//----------------------------------------------------------------------------- 

procedure I_Init;

{ Called by startup code }
{ to get the ammount of memory to malloc }
{ for the zone management. }
function I_ZoneBase(var size: integer): pointer;

{ Called by D_DoomLoop, }
{ returns current time in tics. }
function I_GetTime: integer;

{ }
{ Called by D_DoomLoop, }
{ called before processing any tics in a frame }
{ (just after displaying a frame). }
{ Time consuming syncronous operations }
{ are performed here (joystick reading). }
{ Can call D_PostEvent. }
{ }
procedure I_StartFrame;

{ }
{ Called by D_DoomLoop, }
{ called before processing each tic in a frame. }
{ Quick syncronous operations are performed here. }
{ Can call D_PostEvent. }
procedure I_StartTic;

{ Asynchronous interrupt functions should maintain private queues }
{ that are read by the synchronous functions }
{ to be converted into events. }
{ Either returns a null ticcmd, }
{ or calls a loadable driver to build it. }
{ This ticcmd will then be modified by the gameloop }
{ for normal input. }
function I_BaseTiccmd: Pticcmd_t;

{ Called by M_Responder when quit is selected. }
{ Clean exit, displays sell blurb. }
procedure I_Quit;

procedure I_Destroy(const code: integer);

procedure I_Error(const error: string; const Args: array of const); overload;

procedure I_Error(const error: string); overload;

procedure I_ProcessWindows;

function I_GameFinished: boolean;

procedure I_WaitVBL(const count: integer);

var
  mb_used: integer;


(* Unit m_menu interface *)


              

{
    m_menu.h, m_menu.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//   Menu widget stuff, episode selection and such. 
//     
//----------------------------------------------------------------------------- 

// 
// MENUS 
// 

{ Called by main loop, }
{ saves config file and calls I_Quit when user exits. }
{ Even when the menu is not displayed, }
{ this can resize the view and change game parameters. }
{ Does all the real work of the menu interaction. }

function M_Responder(ev: Pevent_t): boolean;

{ Called by main loop, }
{ only used for menu (skull cursor) animation. }
procedure M_Ticker;

{ Called by main loop, }
{ draws the menus directly into the screen buffer. }
procedure M_Drawer;

{ Called by D_DoomMain, }
{ loads the config file. }
procedure M_Init;

{ Called by intro code to force menu up upon a keypress, }
{ does nothing if menu is already up. }
procedure M_StartControlPanel;

var
//
// defaulted values
//
  mouseSensitivity: integer;  // has default

// Show messages has default, 0 = off, 1 = on
  showMessages: integer;

  menuactive: boolean;

  inhelpscreens: boolean;

// Blocky mode, has default, 0 = high, 1 = normal
  detailLevel: integer;
  screenblocks: integer;  // has default

procedure M_InitMenus;


(* Unit f_finale interface *)


                        

{
    f_finale.h, f_finale.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// Game completion, final screen animation. 
// 
//----------------------------------------------------------------------------- 

function F_Responder(ev: Pevent_t): boolean;

{ Called by main loop. }
procedure F_Ticker;

{ Called by main loop. }
procedure F_Drawer;

procedure F_StartFinale;


(* Unit doomdef interface *)


{
    doomdef.h, doomdef.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Internally used data structures for virtually everything, 
//   key definitions, lots of other stuff. 
// 
//-----------------------------------------------------------------------------

//
// Global parameters/defines.
//
// DOOM version
const
  VERSION =  110;

type
// Game mode handling - identify IWAD version
//  to handle IWAD dependend animations etc.
  GameMode_t = (
    shareware,   // DOOM 1 shareware, E1, M9
    registered,  // DOOM 1 registered, E3, M27
    commercial,  // DOOM 2 retail, E1 M34
    ///////////  // DOOM 2 german edition not handled
    retail,      // DOOM 1 retail, E4, M36
    indetermined // Well, no IWAD found.
  );

// Mission packs - might be useful for TC stuff?
  GameMission_t = (
    doom,       // DOOM 1
    doom2,      // DOOM 2
    pack_tnt,   // TNT mission pack
    pack_plut,  // Plutonia pack
    none
  );

// Identify language to use, software localization.
  Language_t = (
    english,
    french,
    german,
    unknown
  );

const
// ?
  MAXWIDTH = 1600;
  MAXHEIGHT = 1200;

var
  SCREENWIDTH: integer;
  SCREENHEIGHT: integer;

const
// The maximum number of players, multiplayer/networking.
  MAXPLAYERS = 4;

// State updates, number of tics / second.
  TICRATE = 35;

// The current state of the game: whether we are
// playing, gazing at the intermission screen,
// the game final animation, or a demo.
type
  gamestate_t = (
    GS_LEVEL,
    GS_INTERMISSION,
    GS_FINALE,
    GS_DEMOSCREEN
  );

const
//
// Difficulty/skill settings/filters.
//

// Skill flags.
  MTF_EASY = 1;
  MTF_NORMAL = 2;
  MTF_HARD = 4;

// Deaf monsters/do not react to sound.
  MTF_AMBUSH = 8;

type
  skill_t = (
    sk_baby,
    sk_easy,
    sk_medium,
    sk_hard,
    sk_nightmare
  );

//
// Key cards.
//
  card_t = (
    it_bluecard,
    it_yellowcard,
    it_redcard,
    it_blueskull,
    it_yellowskull,
    it_redskull,
    NUMCARDS
  );

// The defined weapons,
//  including a marker indicating
//  user has not changed weapon.
  weapontype_t = (
    wp_fist,
    wp_pistol,
    wp_shotgun,
    wp_chaingun,
    wp_missile,
    wp_plasma,
    wp_bfg,
    wp_chainsaw,
    wp_supershotgun,
    NUMWEAPONS,
    // No pending weapon change.
    wp_nochange
  );

// Ammunition types defined.
  ammotype_t = (
    am_clip,  // Pistol / chaingun ammo.
    am_shell, // Shotgun / double barreled shotgun.
    am_cell,  // Plasma rifle, BFG.
    am_misl,  // Missile launcher.
    NUMAMMO,
    am_noammo // Unlimited for chainsaw / fist.
  );

// Power up artifacts.
  powertype_t = (
    pw_invulnerability,
    pw_strength,
    pw_invisibility,
    pw_ironfeet,
    pw_allmap,
    pw_infrared,
    NUMPOWERS
  );
  Ppowertype_t = ^powertype_t;

//
// Power up durations,
//  how many seconds till expiration,
//  assuming TICRATE is 35 ticks/second.
//
const
  INVULNTICS = 30 * TICRATE;
  INVISTICS = 60 * TICRATE;
  INFRATICS = 120 * TICRATE;
  IRONTICS = 60 * TICRATE;

//
// DOOM keyboard definition.
// This is the stuff configured by Setup.Exe.
// Most key data are simple ascii (uppercased).
//
const
  KEY_RIGHTARROW = $ae;
  KEY_LEFTARROW = $ac;
  KEY_UPARROW = $ad;
  KEY_DOWNARROW = $af;
  KEY_ESCAPE = 27;
  KEY_ENTER = 13;
  KEY_TAB = 9;

  KEY_F1 = $80 + $3b;
  KEY_F2 = $80 + $3c;
  KEY_F3 = $80 + $3d;
  KEY_F4 = $80 + $3e;
  KEY_F5 = $80 + $3f;
  KEY_F6 = $80 + $40;
  KEY_F7 = $80 + $41;
  KEY_F8 = $80 + $42;
  KEY_F9 = $80 + $43;
  KEY_F10 = $80 + $44;
  KEY_F11 = $80 + $57;
  KEY_F12 = $80 + $58;

  KEY_BACKSPACE = 127;
  KEY_PAUSE = $ff;

  KEY_EQUALS = $3d;
  KEY_MINUS = $2d;

  KEY_RSHIFT = $80 + $36;
  KEY_RCTRL = $80 + $1d;
  KEY_RALT = $80 + $38;

  KEY_PAGEDOWN = $80 + $45;
  KEY_PAGEUP = $80 + $46;
  KEY_INS = $80 + $47;


  KEY_LALT = KEY_RALT;



(* Unit doomstat interface *)


              

{
    doomstat.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// $Log:$ 
// 
// DESCRIPTION: 
//  Put all global tate variables here. 
// 
//-----------------------------------------------------------------------------

var
// Game Mode - identify IWAD as shareware, retail etc.
  gamemode: GameMode_t;
  gamemission: GameMission_t;

// Language.
  language: Language_t;

// Set if homebrew PWAD stuff has been added.
  modifiedgame : boolean;


(* Unit st_stuff interface *)


                                 

{
    st_stuff.h, st_stuff.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Status bar code.
//	Does the face/direction indicator animatin.
//	Does palette indicators as well (red pain/berserk, bright pickup)
//
//-----------------------------------------------------------------------------

// Size of statusbar.
// Now sensitive for scaling.
const
  ST_HEIGHT = 32;
  ST_WIDTH = 320; //SCREENWIDTH;
  ST_Y = 200 - ST_HEIGHT; //SCREENHEIGHT - ST_HEIGHT;


//
// STATUS BAR
//

// Called by main loop.
function ST_Responder(ev: Pevent_t): boolean;

// Called by main loop.
procedure ST_Ticker;

// Called by main loop.
procedure ST_Drawer(full: boolean; refresh: boolean);

// Called when the console player is spawned on each level.
procedure ST_Start;

// Called by startup code.
procedure ST_Init;



// States for status bar code.
type
  st_stateenum_t = (
    AutomapState,
    FirstPersonState
  );

// States for the chat code.
  st_chatstateenum_t = (
    StartChatState,
    WaitDestState,
    GetChatState
  );



(* Unit doomdata interface *)


                                                                                                                                                    

{
    doomdata.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  all external data is defined here 
//  most of the data is loaded into different structures at run time 
//  some internal structures shared by many modules are here 
// 
//-----------------------------------------------------------------------------

type
//
// Map level types.
// The following data structures define the persistent format
// used in the lumps of the WAD files.
//

// Lump order in a map WAD: each map needs a couple of lumps
// to provide a complete scene geometry description.
  maplumpdesc_t = (
    ML_LABEL,     // A separator, name, ExMx or MAPxx
    ML_THINGS,    // Monsters, items..
    ML_LINEDEFS,  // LineDefs, from editing
    ML_SIDEDEFS,  // SideDefs, from editing
    ML_VERTEXES,  // Vertices, edited and BSP splits generated
    ML_SEGS,      // LineSegs, from LineDefs split by BSP
    ML_SSECTORS,  // SubSectors, list of LineSegs
    ML_NODES,     // BSP nodes
    ML_SECTORS,   // Sectors, from editing
    ML_REJECT,    // LUT, sector-sector visibility
    ML_BLOCKMAP   // LUT, motion clipping, walls/grid element
  );

  mapvertex_t = record
    x : smallint;
    y : smallint;
  end;
  Pmapvertex_t = ^mapvertex_t;

// A SideDef, defining the visual appearance of a wall,
// by setting textures and offsets.
  mapsidedef_t = record
    textureoffset : smallint;
    rowoffset : smallint;
    toptexture : char8_t;
    bottomtexture : char8_t;
    midtexture : char8_t;
  // Front sector, towards viewer.
    sector : smallint;
  end;
  Pmapsidedef_t = ^mapsidedef_t;

// A LineDef, as used for editing, and as input
// to the BSP builder.
  maplinedef_t = record
    v1 : smallint;
    v2 : smallint;
    flags : smallint;
    special : smallint;
    tag : smallint;
  // sidenum[1] will be -1 if one sided
    sidenum : array[0..1] of smallint;
  end;
  Pmaplinedef_t = ^maplinedef_t;

//
// LineDef attributes.
//

const
// Solid, is an obstacle.
  ML_BLOCKING = 1;

// Blocks monsters only.
  ML_BLOCKMONSTERS = 2;

// Backside will not be present at all
//  if not two sided.
  ML_TWOSIDED = 4;

// If a texture is pegged, the texture will have
// the end exposed to air held constant at the
// top or bottom of the texture (stairs or pulled
// down things) and will move with a height change
// of one of the neighbor sectors.
// Unpegged textures allways have the first row of
// the texture at the top pixel of the line for both
// top and bottom textures (use next to windows).

// upper texture unpegged
  ML_DONTPEGTOP = 8;

// lower texture unpegged
  ML_DONTPEGBOTTOM = 16;

// In AutoMap: don't map as two sided: IT'S A SECRET!
  ML_SECRET = 32;

// Sound rendering: don't let sound cross two of these.
  ML_SOUNDBLOCK = 64;

// Don't draw on the automap at all.
  ML_DONTDRAW = 128;

// Set if already seen, thus drawn in automap.
  ML_MAPPED = 256;


type
// Sector definition, from editing.
  mapsector_t = record
    floorheight : smallint;
    ceilingheight : smallint;
    floorpic : char8_t;
    ceilingpic : char8_t;
    lightlevel : smallint;
    special : smallint;
    tag : smallint;
  end;
  Pmapsector_t = ^mapsector_t;

// SubSector, as generated by BSP.
  mapsubsector_t = record
    numsegs : smallint;
  // Index of first one, segs are stored sequentially.
    firstseg : smallint;
  end;
  Pmapsubsector_t = ^mapsubsector_t;

// LineSeg, generated by splitting LineDefs
// using partition lines selected by BSP builder.
  mapseg_t = record
    v1 : smallint;
    v2 : smallint;
    angle : smallint;
    linedef : smallint;
    side : smallint;
    offset : smallint;
  end;
  Pmapseg_t = ^mapseg_t;

// BSP node structure.

// Indicate a leaf.
const
  NF_SUBSECTOR = $8000;

type
  mapnode_t = record
  // Partition line from (x,y) to x+dx,y+dy)
    x : smallint;
    y : smallint;
    dx : smallint;
    dy : smallint;

  // Bounding box for each child,
  // clip against view frustum.
    bbox : packed array[0..1] of packed array[0..3] of smallint;

  // If NF_SUBSECTOR its a subsector,
  // else it's a node of another subtree.
    children : packed array[0..1] of word;
  end;
  Pmapnode_t = ^mapnode_t;
  
// Thing definition, position, orientation and type,
// plus skill/visibility flags and attributes.
  mapthing_t = record
    x : smallint;
    y : smallint;
    angle : smallint;
    _type : smallint;
    options : smallint;
  end;
  Pmapthing_t = ^mapthing_t;


(* Unit d_items interface *)


              

{
    d_items.h, d_items.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Items: key cards, artifacts, weapon, ammunition. 
// 
//----------------------------------------------------------------------------- 


type
  { Weapon info: sprite frames, ammunition use. }
  weaponinfo_t = record
    ammo: ammotype_t;
    upstate: longint;
    downstate: longint;
    readystate: longint;
    atkstate: longint;
    flashstate: longint;
  end;

var
  weaponinfo: array[0..Ord(NUMWEAPONS)-1] of weaponinfo_t;


(* Unit p_mobj_h interface *)


                                                                        // We need the thinker_t stuff.

{
    p_mobj.h
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Map Objects, MObj, definition and handling.
//
//-----------------------------------------------------------------------------

//
// NOTES: mobj_t
//
// mobj_ts are used to tell the refresh where to draw an image,
// tell the world simulation when objects are contacted,
// and tell the sound driver how to position a sound.
//
// The refresh uses the next and prev links to follow
// lists of things in sectors as they are being drawn.
// The sprite, frame, and angle elements determine which patch_t
// is used to draw the sprite if it is visible.
// The sprite and frame values are allmost allways set
// from state_t structures.
// The statescr.exe utility generates the states.h and states.c
// files that contain the sprite/frame numbers from the
// statescr.txt source file.
// The xyz origin point represents a point at the bottom middle
// of the sprite (between the feet of a biped).
// This is the default origin position for patch_ts grabbed
// with lumpy.exe.
// A walking creature will have its z equal to the floor
// it is standing on.
//
// The sound code uses the x,y, and subsector fields
// to do stereo positioning of any sound effited by the mobj_t.
//
// The play simulation uses the blocklinks, x,y,z, radius, height
// to determine when mobj_ts are touching each other,
// touching lines in the map, or hit by trace lines (gunshots,
// lines of sight, etc).
// The mobj_t->flags element has various bit flags
// used by the simulation.
//
// Every mobj_t is linked into a single sector
// based on its origin coordinates.
// The subsector_t is found with R_PointInSubsector(x,y),
// and the sector_t can be found with subsector->sector.
// The sector links are only used by the rendering code,
// the play simulation does not care about them at all.
//
// Any mobj_t that needs to be acted upon by something else
// in the play world (block movement, be shot, etc) will also
// need to be linked into the blockmap.
// If the thing has the MF_NOBLOCK flag set, it will not use
// the block links. It can still interact with other things,
// but only as the instigator (missiles will run into other
// things, but nothing can run into a missile).
// Each block in the grid is 128*128 units, and knows about
// every line_t that it contains a piece of, and every
// interactable mobj_t that has its origin contained.
//
// A valid mobj_t is a mobj_t that has the proper subsector_t
// filled in for its xy coordinates and is linked into the
// sector from which the subsector was made, or has the
// MF_NOSECTOR flag set (the subsector_t needs to be valid
// even if MF_NOSECTOR is set), and is linked into a blockmap
// block or has the MF_NOBLOCKMAP flag set.
// Links should only be modified by the P_[Un]SetThingPosition()
// functions.
// Do not change the MF_NO? flags while a thing is valid.
//
// Any questions?
//

//
// Misc. mobj flags
//

const
  // Call P_SpecialThing when touched.
  MF_SPECIAL = 1;
  // Blocks.
  MF_SOLID = 2;
  // Can be hit.
  MF_SHOOTABLE = 4;
  // Don't use the sector links (invisible but touchable).
  MF_NOSECTOR = 8;
  // Don't use the blocklinks (inert but displayable)
  MF_NOBLOCKMAP = 16;

  // Not to be activated by sound, deaf monster.
  MF_AMBUSH = 32;
  // Will try to attack right back.
  MF_JUSTHIT = 64;
  // Will take at least one step before attacking.
  MF_JUSTATTACKED = 128;
  // On level spawning (initial position),
  //  hang from ceiling instead of stand on floor.
  MF_SPAWNCEILING = 256;
  // Don't apply gravity (every tic),
  //  that is, object will float, keeping current height
  //  or changing it actively.
  MF_NOGRAVITY = 512;

  // Movement flags.
  // This allows jumps from high places.
  MF_DROPOFF = $400;
  // For players, will pick up items.
  MF_PICKUP = $800;
  // Player cheat. ???
  MF_NOCLIP = $1000;
  // Player: keep info about sliding along walls.
  MF_SLIDE = $2000;
  // Allow moves to any height, no gravity.
  // For active floaters, e.g. cacodemons, pain elementals.
  MF_FLOAT = $4000;
  // Don't cross lines
  //   ??? or look at heights on teleport.
  MF_TELEPORT = $8000;
  // Don't hit same species, explode on block.
  // Player missiles as well as fireballs of various kinds.
  MF_MISSILE = $10000;
  // Dropped by a demon, not level spawned.
  // E.g. ammo clips dropped by dying former humans.
  MF_DROPPED = $20000;
  // Use fuzzy draw (shadow demons or spectres),
  //  temporary player invisibility powerup.
  MF_SHADOW = $40000;
  // Flag: don't bleed when shot (use puff),
  //  barrels and shootable furniture shall not bleed.
  MF_NOBLOOD = $80000;
  // Don't stop moving halfway off a step,
  //  that is, have dead bodies slide down all the way.
  MF_CORPSE = $100000;
  // Floating to a height for a move, ???
  //  don't auto float to target's height.
  MF_INFLOAT = $200000;

  // On kill, count this enemy object
  //  towards intermission kill total.
  // Happy gathering.
  MF_COUNTKILL = $400000;

  // On picking up, count this item object
  //  towards intermission item total.
  MF_COUNTITEM = $800000;

  // Special handling: skull in flight.
  // Neither a cacodemon nor a missile.
  MF_SKULLFLY = $1000000;

  // Don't spawn this object
  //  in death match mode (e.g. key cards).
  MF_NOTDMATCH = $2000000;

  // Player sprites in multiplayer modes are modified
  //  using an internal color lookup table for re-indexing.
  // If 0x4 0x8 or 0xc,
  //  use a translation table for player colormaps
  MF_TRANSLATION = $c000000;
  // Hmm ???.
  MF_TRANSSHIFT = 26;

type
// Map Object definition.
  Pmobj_t = ^mobj_t;
  mobj_t = record
    // List: thinker links.
    thinker: thinker_t;

    // Info for drawing: position.
    x: fixed_t;
    y: fixed_t;
    z: fixed_t;

    // More list: links in sector (if needed)
    snext: Pmobj_t;
    sprev: Pmobj_t;

    //More drawing info: to determine current sprite.
    angle: angle_t; // orientation
    sprite: spritenum_t; // used to find patch_t and flip value
    frame: integer; // might be ORed with FF_FULLBRIGHT

    // Interaction info, by BLOCKMAP.
    // Links in blocks (if needed).
    bnext: Pmobj_t;
    bprev: Pmobj_t;

    subsector: pointer; //Psubsector_t;

    // The closest interval over all contacted Sectors.
    floorz: fixed_t;
    ceilingz: fixed_t;

    // For movement checking.
    radius: fixed_t;
    height: fixed_t;

    // Momentums, used to update position.
    momx: fixed_t;
    momy: fixed_t;
    momz: fixed_t;

    // If == validcount, already checked.
    validcount: integer;

    _type: mobjtype_t;
    info: Pmobjinfo_t; // &mobjinfo[mobj->type]

    tics: integer; // state tic counter
    state: Pstate_t;
    flags: integer;
    health: integer;

    // Movement direction, movement generation (zig-zagging).
    movedir: integer; // 0-7
    movecount: integer; // when 0, select a new dir

    // Thing being chased/attacked (or NULL),
    // also the originator for missiles.
    target: Pmobj_t;

    // Reaction time: if non 0, don't attack yet.
    // Used by player to freeze a bit after teleporting.
    reactiontime: integer;

    // If >0, the target will be chased
    // no matter what (even if shot)
    threshold: integer;

    // Additional info record for player avatars only.
    // Only valid if type == MT_PLAYER
    player: pointer; //Pplayer_t;

    // Player number last looked for.
    lastlook: integer;

    // For nightmare respawn.
    spawnpoint: mapthing_t;

    // Thing being chased/attacked for tracers.
    tracer: Pmobj_t;
  end;
  Tmobj_tPArray = array[0..$FFFF] of Pmobj_t;
  Pmobj_tPArray = ^Tmobj_tPArray;



(* Unit r_defs interface *)


                                                                                                                                                                                                                                                                          

{
    r_defs.h
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//      Refresh/rendering module, shared data struct definitions.
//
//-----------------------------------------------------------------------------

// Silhouette, needed for clipping Segs (mainly)
// and sprites representing things.
const
  SIL_NONE = 0;
  SIL_BOTTOM = 1;
  SIL_TOP = 2;
  SIL_BOTH = 3;

  MAXDRAWSEGS = 1024; //was 256; // VJ maybe bigger ????

type
//
// INTERNAL MAP TYPES
//  used by play and refresh
//

//
// Your plain vanilla vertex.
// Note: transformed values not buffered locally,
//  like some DOOM-alikes ("wt", "WebView") did.
//
  vertex_t = packed record
    x: fixed_t;
    y: fixed_t;
  end;
  Pvertex_t = ^vertex_t;
  vertex_tArray = packed array[0..$FFFF] of vertex_t;
  Pvertex_tArray = ^vertex_tArray;

{// Forward of LineDefs, for Sectors.
  line_t = record;}


// Each sector has a degenmobj_t in its center
//  for sound origin purposes.
// I suppose this does not handle sound from
//  moving objects (doppler), because
//  position is prolly just buffered, not
//  updated.
  degenmobj_t = packed record
    thinker: thinker_t; // not used for anything
    x: fixed_t;
    y: fixed_t;
    z: fixed_t;
  end;
  Pdegenmobj_t = ^degenmobj_t;

  Pline_tArray = ^line_tArray;
  Pline_tPArray = ^line_tPArray;
//
// The SECTORS record, at runtime.
// Stores things/mobjs.
//
  sector_t = packed record
    floorheight: fixed_t;
    ceilingheight: fixed_t;
    floorpic: smallint;
    ceilingpic: smallint;
    lightlevel: smallint;
    special: smallint;
    tag: smallint;

    // 0 = untraversed, 1,2 = sndlines -1
    soundtraversed: integer;

    // thing that made a sound (or null)
    soundtarget: Pmobj_t;

    // mapblock bounding box for height changes
    blockbox: packed array[0..3] of integer;

    // origin for any sounds played by the sector
    soundorg: degenmobj_t;

    // if == validcount, already checked
    validcount: integer;

    // list of mobjs in sector
    thinglist: Pmobj_t;

    // thinker_t for reversable actions
    specialdata: pointer;

    linecount: integer;
    lines: Pline_tPArray;	// [linecount] size
  end;
  Psector_t = ^sector_t;
  sector_tArray = packed array[0..$FFFF] of sector_t;
  Psector_tArray = ^sector_tArray;

//
// The SideDef.
//

  side_t = packed record
    // add this to the calculated texture column
    textureoffset: fixed_t;

    // add this to the calculated texture top
    rowoffset: fixed_t;

    // Texture indices.
    // We do not maintain names here.
    toptexture: smallint;
    bottomtexture: smallint;
    midtexture: smallint;

    // Sector the SideDef is facing.
    sector: Psector_t;
  end;
  Pside_t = ^side_t;
  side_tArray = packed array[0..$FFFF] of side_t;
  Pside_tArray = ^side_tArray;

//
// Move clipping aid for LineDefs.
//
  slopetype_t = (
    ST_HORIZONTAL,
    ST_VERTICAL,
    ST_POSITIVE,
    ST_NEGATIVE
  );

  line_t = packed record
    // Vertices, from v1 to v2.
    v1: Pvertex_t;
    v2: Pvertex_t;

    // Precalculated v2 - v1 for side checking.
    dx: fixed_t;
    dy: fixed_t;

    // Animation related.
    flags: smallint;
    special: smallint;
    tag: smallint;

    // Visual appearance: SideDefs.
    //  sidenum[1] will be -1 if one sided
    sidenum: packed array[0..1] of smallint;

    // Neat. Another bounding box, for the extent
    //  of the LineDef.
    bbox: packed array[0..3] of fixed_t;

    // To aid move clipping.
    slopetype: slopetype_t;

    // Front and back sector.
    // Note: redundant? Can be retrieved from SideDefs.
    frontsector: Psector_t;
    backsector: Psector_t;

    // if == validcount, already checked
    validcount: integer;

    // thinker_t for reversable actions
    specialdata: pointer;
  end;
  Pline_t = ^line_t;
  line_tArray = packed array[0..$FFFF] of line_t;
  line_tPArray = packed array[0..$FFFF] of Pline_t;


//
// A SubSector.
// References a Sector.
// Basically, this is a list of LineSegs,
//  indicating the visible walls that define
//  (all or some) sides of a convex BSP leaf.
//
  subsector_t = packed record
    sector: Psector_t;
    numlines: smallint;
    firstline: smallint;
  end;
  Psubsector_t = ^subsector_t;
  subsector_tArray = packed array[0..$FFFF] of subsector_t;
  Psubsector_tArray = ^subsector_tArray;

//
// The LineSeg.
//
  seg_t = packed record
    v1: Pvertex_t;
    v2: Pvertex_t;

    offset: fixed_t;

    angle: angle_t;

    sidedef: Pside_t;
    linedef: Pline_t;

    // Sector references.
    // Could be retrieved from linedef, too.
    // backsector is NULL for one sided lines
    frontsector: Psector_t;
    backsector: Psector_t;
  end;
  Pseg_t = ^seg_t;
  seg_tArray = packed array[0..$FFFF] of seg_t;
  Pseg_tArray = ^seg_tArray;

//
// BSP node.
//
  node_t = packed record
    // Partition line.
    x: fixed_t;
    y: fixed_t;
    dx: fixed_t;
    dy: fixed_t;

    // Bounding box for each child.
    bbox: packed array[0..1, 0..3] of fixed_t;

    // If NF_SUBSECTOR its a subsector.
    children: packed array[0..1] of word;
  end;
  Pnode_t = ^node_t;
  node_tArray = packed array[0..$FFFF] of node_t;
  Pnode_tArray = ^node_tArray;

// posts are runs of non masked source pixels
  post_t = packed record
    topdelta: byte; // -1 is the last post in a column
    length: byte;   // length data bytes follows
  end;
{  post_t = packed record
    topdelta: smallint; // -1 is the last post in a column
    length: smallint;   // length data bytes follows
  end;}
  Ppost_t = ^post_t;

// column_t is a list of 0 or more post_t, (byte)-1 terminated
  column_t = post_t;
  Pcolumn_t = ^column_t;

//
// OTHER TYPES
//

// This could be wider for >8 bit display.
// Indeed, true color support is posibble
//  precalculating 24bpp lightmap/colormap LUT.
//  from darkening PLAYPAL to all black.
// Could even us emore than 32 levels.
  lighttable_t = byte;
  Plighttable_t = ^lighttable_t;
  lighttable_tArray = packed array[0..$FFFF] of lighttable_t;
  Plighttable_tArray = ^lighttable_tArray;
  lighttable_tPArray = packed array[0..$FFFF] of Plighttable_tArray;
  Plighttable_tPArray = ^lighttable_tPArray;

//
// ?
//
  drawseg_t = packed record
    curline: Pseg_t;
    x1: integer;
    x2: integer;

    scale1: fixed_t;
    scale2: fixed_t;
    scalestep: fixed_t;

    // 0=none, 1=bottom, 2=top, 3=both
    silhouette: integer;

    // do not clip sprites above this
    bsilheight: fixed_t;

    // do not clip sprites below this
    tsilheight: fixed_t;

    // Pointers to lists for sprite clipping,
    //  all three adjusted so [x1] is first value.
    sprtopclip: PSmallIntArray;
    sprbottomclip: PSmallIntArray;
    maskedtexturecol: PSmallIntArray;
  end;
  Pdrawseg_t = ^drawseg_t;

// Patches.
// A patch holds one or more columns.
// Patches are used for sprites and all masked pictures,
// and we compose textures from the TEXTURE1/2 lists
// of patches.
  patch_t = packed record
    width: smallint; // bounding box size
    height: smallint;
    leftoffset: smallint; // pixels to the left of origin
    topoffset: smallint;  // pixels below the origin
    columnofs: array[0..7] of integer; // only [width] used
    // the [0] is &columnofs[width]
  end;
  Ppatch_t = ^patch_t;
  patch_tArray = packed array[0..$FFFF] of patch_t;
  Ppatch_tArray = ^patch_tArray;
  patch_tPArray = packed array[0..$FFFF] of Ppatch_t;
  Ppatch_tPArray = ^patch_tPArray;

// A vissprite_t is a thing
//  that will be drawn during a refresh.
// I.e. a sprite object that is partly visible.
  Pvissprite_t = ^vissprite_t;
  vissprite_t = packed record
    // Doubly linked list.
    prev: Pvissprite_t;
    next: Pvissprite_t;

    x1: integer;
    x2: integer;

    // for line side calculation
    gx: fixed_t;
    gy: fixed_t;

    // global bottom / top for silhouette clipping
    gz: fixed_t;
    gzt: fixed_t;

    // horizontal position of x1
    startfrac: fixed_t;

    scale: fixed_t;

    // negative if flipped
    xiscale: fixed_t;

    texturemid: fixed_t;
    patch: integer;

    // for color translation and shadow draw,
    //  maxbright frames as well
    colormap: Plighttable_tArray;

    mobjflags: integer;
  end;

//
// Sprites are patches with a special naming convention
//  so they can be recognized by R_InitSprites.
// The base name is NNNNFx or NNNNFxFx, with
//  x indicating the rotation, x = 0, 1-7.
// The sprite and frame specified by a thing_t
//  is range checked at run time.
// A sprite is a patch_t that is assumed to represent
//  a three dimensional object and may have multiple
//  rotations pre drawn.
// Horizontal flipping is used to save space,
//  thus NNNNF2F5 defines a mirrored patch.
// Some sprites will only have one picture used
// for all views: NNNNF0
//
  spriteframe_t = packed record
    // If false use 0 for any position.
    // Note: as eight entries are available,
    //  we might as well insert the same name eight times.
    rotate: integer;

    // Lump to use for view angles 0-7.
//    lump: packed array[0..7] of smallint;
    lump: array[0..7] of integer;

    // Flip bit (1 = flip) to use for view angles 0-7.
//    flip: packed array[0..7] of byte;
    flip: array[0..7] of boolean;
  end;
  Pspriteframe_t = ^spriteframe_t;
  spriteframe_tArray = packed array[0..$FFFF] of spriteframe_t;
  Pspriteframe_tArray = ^spriteframe_tArray;


//
// A sprite definition:
//  a number of animation frames.
//
  spritedef_t = packed record
    numframes: integer;
    spriteframes: Pspriteframe_tArray;
  end;
  Pspritedef_t = ^spritedef_t;
  spritedef_tArray = packed array[0..$FFFF] of spritedef_t;
  Pspritedef_tArray = ^spritedef_tArray;

const
  VISEND = $FFFF;

type
  visindex_t = word;
  visindex_tArray = packed array[-1..$FFFF] of visindex_t;
  Pvisindex_tArray = ^visindex_tArray;

//
// Now what is a visplane, anyway?
//
  visplane_t = packed record
    height: fixed_t;
    picnum: integer;
    lightlevel: integer;
    minx: integer;
    maxx: integer;

    // leave pads for [minx-1] and [maxx+1]
    top: Pvisindex_tArray;    // Now allocated dinamically!

    // See above.
    bottom: Pvisindex_tArray; // Now allocated dinamically!
  end;
  Pvisplane_t = ^visplane_t;


(* Unit r_segs interface *)


                                              

{
    r_segs.h, r_segs.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Refresh module, drawing LineSegs from BSP.
//	All the clipping: columns, horizontal spans, sky columns.
//
//-----------------------------------------------------------------------------

procedure R_RenderMaskedSegRange(ds: Pdrawseg_t; x1, x2: integer);

procedure R_StoreWallRange(start: integer; stop: integer);

var
  maskedtexturecol: PSmallIntArray; // VJ : declared in r_defs
// angle to line origin
  rw_angle1: angle_t; // VJ was integer

var
// True if any of the segs textures might be visible.
  segtextured: boolean;

// False if the back side is the same plane.
  markfloor: boolean;
  markceiling: boolean;

  maskedtexture: boolean;
  toptexture: integer;
  bottomtexture: integer;
  midtexture: integer;


//
// regular wall
//
  rw_x: integer;
  rw_stopx: integer;
  rw_centerangle: angle_t;
  rw_offset: fixed_t;
  rw_distance: fixed_t;
  rw_normalangle: angle_t;
  rw_scale: fixed_t;
  rw_scalestep: fixed_t;
  rw_midtexturemid: fixed_t;
  rw_toptexturemid: fixed_t;
  rw_bottomtexturemid: fixed_t;

  worldtop: integer;
  worldbottom: integer;
  worldhigh: integer;
  worldlow: integer;

  pixhigh: fixed_t;
  pixlow: fixed_t;
  pixhighstep: fixed_t;
  pixlowstep: fixed_t;

  topfrac: fixed_t;
  topstep: fixed_t;

  bottomfrac: fixed_t;
  bottomstep: fixed_t;


  walllights: Plighttable_tPArray;



(* Unit v_video interface *)


                                                                                                  

{
    v_video.h, v_video.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Gamma correction LUT.
//	Functions to draw patches (by post) directly to screen.
//	Functions to blit a block to the screen.
//
//-----------------------------------------------------------------------------

function _CENTERY: integer;

const
//
// VIDEO
//

// drawing stuff
//
// Background and foreground screen numbers
//
  _FG = 0;
  _SCRF = 4; // Finale Screen Buffer 320x200
  _TMP = 5;  // Temporary Screen Buffer 320x200
  _BG = 6;   // Status Bar Screen Buffer


var
// Screen 0 is the screen updated by I_Update screen.
// Screen 1 is an extra buffer.
// Screen 4 is an extra buffer for finale.
// Screen 5 is used by status line
  screens: array[0.._BG] of PByteArray;


// Allocates buffer screens, call before R_Init.
procedure V_Init;


procedure V_CopyRect(
  srcx: integer;
  srcy: integer;
  srcscrn: integer;
  width: integer;
  height: integer;
  destx: integer;
  desty: integer;
  destscrn: integer;
  preserve: boolean);

procedure V_DrawPatch(x, y: integer; scrn: integer; patch: Ppatch_t; preserve: boolean);
procedure V_DrawPatchFlipped(x, y: integer; scrn: integer; patch: Ppatch_t; preserve: boolean);

// Draw a linear block of pixels into the view buffer.
procedure V_DrawBlock(x, y: integer; scrn: integer; width, height: integer; src: PByteArray);

// Reads a linear block of pixels into the view buffer.
procedure V_GetBlock(x, y: integer; scrn: integer; width, height: integer; dest: PByteArray);


procedure V_MarkRect(x, y: integer; width, height: integer; preserve: boolean);

function V_PreserveX(x: integer): integer;

function V_PreserveY(y: integer): integer;

function V_PreserveW(x: integer; w: integer): integer;

function V_PreserveH(y: integer; h: integer): integer;

function V_NeedsPreserve: boolean; overload;

function V_NeedsPreserve(preserve: boolean): boolean; overload;

var
  dirtybox: array[0..3] of integer;

const
// Now where did these came from?
  gammatable: array[0..4, 0..255] of byte = (
    (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
     17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,
     33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,
     49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,
     65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,
     81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,
     97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,
     113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255),

    (2,4,5,7,8,10,11,12,14,15,16,18,19,20,21,23,24,25,26,27,29,30,31,
     32,33,34,36,37,38,39,40,41,42,44,45,46,47,48,49,50,51,52,54,55,
     56,57,58,59,60,61,62,63,64,65,66,67,69,70,71,72,73,74,75,76,77,
     78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,
     99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,
     115,116,117,118,119,120,121,122,123,124,125,126,127,128,129,129,
     130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,
     146,147,148,148,149,150,151,152,153,154,155,156,157,158,159,160,
     161,162,163,163,164,165,166,167,168,169,170,171,172,173,174,175,
     175,176,177,178,179,180,181,182,183,184,185,186,186,187,188,189,
     190,191,192,193,194,195,196,196,197,198,199,200,201,202,203,204,
     205,205,206,207,208,209,210,211,212,213,214,214,215,216,217,218,
     219,220,221,222,222,223,224,225,226,227,228,229,230,230,231,232,
     233,234,235,236,237,237,238,239,240,241,242,243,244,245,245,246,
     247,248,249,250,251,252,252,253,254,255),

    (4,7,9,11,13,15,17,19,21,22,24,26,27,29,30,32,33,35,36,38,39,40,42,
     43,45,46,47,48,50,51,52,54,55,56,57,59,60,61,62,63,65,66,67,68,69,
     70,72,73,74,75,76,77,78,79,80,82,83,84,85,86,87,88,89,90,91,92,93,
     94,95,96,97,98,100,101,102,103,104,105,106,107,108,109,110,111,112,
     113,114,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,
     129,130,131,132,133,133,134,135,136,137,138,139,140,141,142,143,144,
     144,145,146,147,148,149,150,151,152,153,153,154,155,156,157,158,159,
     160,160,161,162,163,164,165,166,166,167,168,169,170,171,172,172,173,
     174,175,176,177,178,178,179,180,181,182,183,183,184,185,186,187,188,
     188,189,190,191,192,193,193,194,195,196,197,197,198,199,200,201,201,
     202,203,204,205,206,206,207,208,209,210,210,211,212,213,213,214,215,
     216,217,217,218,219,220,221,221,222,223,224,224,225,226,227,228,228,
     229,230,231,231,232,233,234,235,235,236,237,238,238,239,240,241,241,
     242,243,244,244,245,246,247,247,248,249,250,251,251,252,253,254,254,
     255),

    (8,12,16,19,22,24,27,29,31,34,36,38,40,41,43,45,47,49,50,52,53,55,
     57,58,60,61,63,64,65,67,68,70,71,72,74,75,76,77,79,80,81,82,84,85,
     86,87,88,90,91,92,93,94,95,96,98,99,100,101,102,103,104,105,106,107,
     108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,
     125,126,127,128,129,130,131,132,133,134,135,135,136,137,138,139,140,
     141,142,143,143,144,145,146,147,148,149,150,150,151,152,153,154,155,
     155,156,157,158,159,160,160,161,162,163,164,165,165,166,167,168,169,
     169,170,171,172,173,173,174,175,176,176,177,178,179,180,180,181,182,
     183,183,184,185,186,186,187,188,189,189,190,191,192,192,193,194,195,
     195,196,197,197,198,199,200,200,201,202,202,203,204,205,205,206,207,
     207,208,209,210,210,211,212,212,213,214,214,215,216,216,217,218,219,
     219,220,221,221,222,223,223,224,225,225,226,227,227,228,229,229,230,
     231,231,232,233,233,234,235,235,236,237,237,238,238,239,240,240,241,
     242,242,243,244,244,245,246,246,247,247,248,249,249,250,251,251,252,
     253,253,254,254,255),

    (16,23,28,32,36,39,42,45,48,50,53,55,57,60,62,64,66,68,69,71,73,75,76,
     78,80,81,83,84,86,87,89,90,92,93,94,96,97,98,100,101,102,103,105,106,
     107,108,109,110,112,113,114,115,116,117,118,119,120,121,122,123,124,
     125,126,128,128,129,130,131,132,133,134,135,136,137,138,139,140,141,
     142,143,143,144,145,146,147,148,149,150,150,151,152,153,154,155,155,
     156,157,158,159,159,160,161,162,163,163,164,165,166,166,167,168,169,
     169,170,171,172,172,173,174,175,175,176,177,177,178,179,180,180,181,
     182,182,183,184,184,185,186,187,187,188,189,189,190,191,191,192,193,
     193,194,195,195,196,196,197,198,198,199,200,200,201,202,202,203,203,
     204,205,205,206,207,207,208,208,209,210,210,211,211,212,213,213,214,
     214,215,216,216,217,217,218,219,219,220,220,221,221,222,223,223,224,
     224,225,225,226,227,227,228,228,229,229,230,230,231,232,232,233,233,
     234,234,235,235,236,236,237,237,238,239,239,240,240,241,241,242,242,
     243,243,244,244,245,245,246,246,247,247,248,248,249,249,250,250,251,
     251,252,252,253,254,254,255,255)
  );

var
  usegamma: integer;


(* Unit hu_lib interface *)


                                   

{
    hu_lib.h, hu_lib.c
}


// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION:  none 
// 
//-----------------------------------------------------------------------------


const
// background and foreground screen numbers
// different from other modules.
  FG = 0;

// font stuff
  HU_CHARERASE = KEY_BACKSPACE;

  HU_MAXLINES = 4;
  HU_MAXLINELENGTH = 80;

//
// Typedefs of widgets
//

// Text Line widget
//  (parent of Scrolling Text and Input Text widgets)
type
  hu_textline_t = record
    // left-justified position of scrolling text window
    x: integer;
    y: integer;

    f: Ppatch_tPArray; // font
    sc: integer; // start character
    l: array[0..(HU_MAXLINELENGTH + 1) - 1] of char; // line of text
    len: integer; // current line length

    // whether this line needs to be udpated
    needsupdate: integer;
  end;
  Phu_textline_t = ^hu_textline_t;

// Scrolling Text window widget
//  (child of Text Line widget)

  hu_stext_t = record
    l: array[0..HU_MAXLINES - 1] of hu_textline_t; // text lines to draw
    h: integer; // height in lines
    cl: integer; // current line number

    // pointer to boolean stating whether to update window
    _on: Pboolean;
    laston: boolean; // last value of *._on.
  end;
  Phu_stext_t = ^hu_stext_t;

// Input Text Line widget
//  (child of Text Line widget)
  hu_itext_t = record
    l: hu_textline_t; // text line to input on

    // left margin past which I am not to delete characters
    lm: integer;

    // pointer to boolean stating whether to update window
    _on: Pboolean;
    laston: boolean; // last value of *->on;
  end;
  Phu_itext_t = ^hu_itext_t;

//
// Widget creation, access, and update routines
//

// initializes heads-up widget library
procedure HUlib_init;

//
// textline code
//

// clear a line of text
procedure HUlib_clearTextLine(t: Phu_textline_t);

procedure HUlib_initTextLine(t: Phu_textline_t; x: integer; y: integer; f: Ppatch_tPArray; sc: integer);

// returns success
function HUlib_addCharToTextLine(t: Phu_textline_t; ch: char): boolean;

// returns success
function HUlib_delCharFromTextLine(t: Phu_textline_t): boolean;

// draws tline
procedure HUlib_drawTextLine(l: Phu_textline_t; drawcursor: boolean);

// erases text line
procedure HUlib_eraseTextLine(l: Phu_textline_t);

//
// Scrolling Text window widget routines
//

// ?
procedure HUlib_initSText(s: Phu_stext_t; x: integer; y: integer; h: integer;
  font: Ppatch_tPArray; startchar: integer; _on: Pboolean);

// add a new line
procedure HUlib_addLineToSText(s: Phu_stext_t);

// ?
procedure HUlib_addMessageToSText(s: Phu_stext_t; prefix: string; msg: string);

// draws stext
procedure HUlib_drawSText(s: Phu_stext_t);

// erases all stext lines
procedure HUlib_eraseSText(s: Phu_stext_t);

// Input Text Line widget routines
procedure HUlib_initIText(it: Phu_itext_t; x: integer; y: integer; font: Ppatch_tPArray;
  startchar: integer; _on: Pboolean);

// enforces left margin
procedure HUlib_delCharFromIText(it: Phu_itext_t);

// enforces left margin
procedure HUlib_eraseLineFromIText(it: Phu_itext_t);

// resets line and left margin
procedure HUlib_resetIText(it: Phu_itext_t);

// left of left-margin
procedure HUlib_addPrefixToIText(it: Phu_itext_t; str: string);

// whether eaten
function HUlib_keyInIText(it: Phu_itext_t; ch: byte): boolean;

procedure HUlib_drawIText(it: Phu_itext_t);

// erases all itext lines
procedure HUlib_eraseIText(it: Phu_itext_t);


(* Unit r_bsp interface *)


             

{
    r_bsp.h, r_bsp.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Refresh module, BSP traversal and handling.
//	BSP traversal, handling of LineSegs for rendering.
//
//-----------------------------------------------------------------------------

// BSP?
procedure R_ClearClipSegs;
procedure R_ClearDrawSegs;


procedure R_RenderBSPNode(bspnum: integer);

type
  drawfunc_t = procedure(start: integer; stop: integer);

var
  ds_p: integer; // VJ was: Pdrawseg_t
  max_ds_p: integer;

  curline: Pseg_t;
  frontsector: Psector_t;
  backsector: Psector_t;

  sidedef: Pside_t;
  linedef: Pline_t;
//  drawsegs: array[0..MAXDRAWSEGS - 1] of drawseg_t;  
  drawsegs: array[0..MAXDRAWSEGS - 1] of Pdrawseg_t;  // test!



(* Unit r_data interface *)


                                   

{
    r_data.h, r_data.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//  Refresh module, data I/O, caching, retrieval of graphics
//  by name.
//	Preparation of data for rendering,
//	generation of lookups, caching, retrieval by name.
//
//-----------------------------------------------------------------------------

// Retrieve column data for span blitting.
function R_GetColumn(tex: integer; col: integer): PByteArray;

// I/O, setting up the stuff.
procedure R_InitData;
procedure R_PrecacheLevel;

// Retrieval.
// Floor/ceiling opaque texture tiles,
// lookup by name. For animation?
function R_FlatNumForName(const name: string): integer;

// Called by P_Ticker for switches and animations,
// returns the texture number for the texture name.
function R_TextureNumForName(const name: string): integer;
function R_CheckTextureNumForName(const name: string): integer;

var
// for global animation
  flattranslation: PIntegerArray;
  texturetranslation: PIntegerArray;

// needed for texture pegging
  textureheight: Pfixed_tArray;
  texturecompositesize: PIntegerArray;

  firstspritelump: integer;
  lastspritelump: integer;

// needed for pre rendering
  spritewidth: Pfixed_tArray;
  spriteoffset: Pfixed_tArray;
  spritetopoffset: Pfixed_tArray;

  colormaps: Plighttable_tArray;

var
  firstflat: integer;
  lastflat: integer;
  numflats: integer;
  maxvisplane: integer;

procedure R_SetupLevel;


(* Unit r_plane interface *)


                                              

{
    r_plane.h, r_plane.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Refresh, visplane stuff (floor, ceilings).
//	Here is a core component: drawing the floors and ceilings,
//	 while maintaining a per column clipping list only.
//	Moreover, the sky areas have to be determined.
//
//-----------------------------------------------------------------------------

procedure R_InitPlanes;
procedure R_ClearPlanes;

procedure R_MapPlane(y: integer; x1: integer; x2: integer);

procedure R_MakeSpans(x: integer; t1: integer; b1: integer; t2: integer; b2: integer);

procedure R_DrawPlanes;

function R_FindPlane(height: fixed_t; picnum: integer; lightlevel: integer): Pvisplane_t;

function R_CheckPlane(pl: Pvisplane_t; start: integer; stop: integer): Pvisplane_t;

var
//
// Clip values are the solid pixel bounding the range.
//  floorclip starts out SCREENHEIGHT
//  ceilingclip starts out -1
//
  floorclip: packed array[0..MAXWIDTH - 1] of smallint;
  ceilingclip: packed array[0..MAXWIDTH - 1] of smallint;

var
  floorplane: Pvisplane_t;
  ceilingplane: Pvisplane_t;

//
// opening
//

// ?
const
  MAXOPENINGS = MAXWIDTH * 64;

var
  openings: packed array[0..MAXOPENINGS - 1] of smallint; 
  lastopening: integer;

  yslope: array[0..MAXHEIGHT - 1] of fixed_t;
  distscale: array[0..MAXWIDTH - 1] of fixed_t;


(* Unit hu_stuff interface *)


                                  

{
    hu_stuff.h, hu_stuff.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION:  Head up display 
// 
//-----------------------------------------------------------------------------

const
//
// Globally visible constants.
//
  HU_FONTSTART = '!'; // the first font characters
  HU_FONTEND = '_'; // the last font characters

// Calculate # of glyphs in font.
  HU_FONTSIZE = (Ord(HU_FONTEND) - Ord(HU_FONTSTART)) + 1;
  HU_BROADCAST = 5;

  HU_MSGREFRESH = KEY_ENTER;
  HU_MSGX = 0;
  HU_MSGY = 0;
  HU_MSGWIDTH = 64; // in characters
  HU_MSGHEIGHT = 1; // in lines

  HU_MSGTIMEOUT = 4 * TICRATE;

//
// HEADS UP TEXT
//
procedure HU_Init;

procedure HU_Start;

function HU_Responder(ev: Pevent_t): boolean;

procedure HU_Ticker;

procedure HU_Drawer;

function HU_dequeueChatChar: char;

procedure HU_Erase;

var
  hu_font: array[0..HU_FONTSIZE - 1] of Ppatch_t;

  chat_on: boolean;

  message_on: boolean;
  message_dontfuckwithme: boolean;
  message_nottobefuckedwith: boolean;

var
// DOOM shareware/registered/retail (Ultimate) names.
  mapnames: array[0..44] of string;

// DOOM 2 map names.
  mapnames2: array[0..31] of string;

// Plutonia WAD map names.
  mapnamesp: array[0..31] of string;

// TNT WAD map names.
  mapnamest: array[0..31] of string;

  player_names: array[0..3] of string;

var
  chat_macros: array[0..9] of string;


(* Unit r_draw interface *)


                                   

{
    r_draw.h, r_draw.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	System specific interface stuff.
//	The actual span/column drawing functions.
//	Here find the main potential for optimization,
//	 e.g. inline assembly, different algorithms.
//
//-----------------------------------------------------------------------------

// The span blitting interface.
// Hook in assembler or system specific BLT
//  here.
procedure R_DrawColumn;
procedure R_DrawColumnLow;
procedure R_DrawSkyColumn;

// The Spectre/Invisibility effect.
procedure R_DrawFuzzColumn;
//procedure R_DrawFuzzColumnLow; // VJ ??? not implement ???

// Draw with color translation tables,
//  for player sprite rendering,
//  Green/Red/Blue/Indigo shirts.
procedure R_DrawTranslatedColumn;

//procedure R_DrawTranslatedColumnLow; // VJ ??? not implement ???

procedure R_VideoErase(const ofs: integer; const count: integer);

// Span blitting for rows, floor/ceiling.
// No Sepctre effect needed.
procedure R_DrawSpan;

// Low resolution mode, 160x200?
procedure R_DrawSpanLow;


procedure R_InitBuffer(width, height: integer);


// Initialize color translation tables,
//  for player rendering etc.
procedure R_InitTranslationTables;



// Rendering function.
procedure R_FillBackScreen;

// If the view size is not full screen, draws a border around it.
procedure R_DrawViewBorder;

var
//
// R_DrawColumn
// Source is the top of the column to scale.
//
  dc_colormap: Plighttable_tArray;
  dc_iscale: fixed_t;
  dc_texturemid: fixed_t;
  dc_x: integer;
  dc_yl: integer;
  dc_yh: integer;

// first pixel in a column (possibly virtual)
  dc_source: PByteArray;

  translationtables: PByteArray;
  dc_translation: PByteArray;

  viewwidth: integer;

var
  ds_y: integer;
  ds_x1: integer;
  ds_x2: integer;

  ds_colormap: Plighttable_tArray;

  ds_xfrac: fixed_t;
  ds_yfrac: fixed_t;
  ds_xstep: fixed_t;
  ds_ystep: fixed_t;

// start of a 64*64 tile image
  ds_source: PByteArray;

// just for profiling
  dscount: integer;

  viewheight: integer;
  scaledviewwidth: integer;

  viewwindowx: integer;
  viewwindowy: integer;
  

(* Unit st_lib interface *)


                                                           

{
    st_lib.h, st_lib.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
// 	The status bar widget code.
//
//-----------------------------------------------------------------------------

type
//
// Typedefs of widgets
//

// Number widget

  st_number_t = record
    // upper right-hand corner
    //  of the number (right-justified)
    x: integer;
    y: integer;

    // max # of digits in number
    width: integer;

    // last number value
    oldnum: integer;

    // pointer to current value
    num: PInteger;

    // pointer to boolean stating
    //  whether to update number
    _on: PBoolean;

    // list of patches for 0-9
    p: Ppatch_tPArray;

    // user data
    data: integer;
  end;
  Pst_number_t = ^st_number_t;

// Percent widget ("child" of number widget,
//  or, more precisely, contains a number widget.)
  st_percent_t = record
    // number information
    n: st_number_t;

    // percent sign graphic
    p: Ppatch_t;
  end;
  Pst_percent_t = ^st_percent_t;

// Multiple Icon widget
  st_multicon_t = record
    // center-justified location of icons
    x: integer;
    y: integer;

    // last icon number
    oldinum: integer;

    // pointer to current icon
    inum: PInteger;

    // pointer to boolean stating
    //  whether to update icon
    _on: PBoolean;

    // list of icons
    p: Ppatch_tPArray;

    // user data
    data: integer;
  end;
  Pst_multicon_t = ^st_multicon_t;

// Binary Icon widget

  st_binicon_t = record
    // center-justified location of icon
    x: integer;
    y: integer;

    // last icon value
    oldval: boolean;

    // pointer to current icon status
    val: PBoolean;

    // pointer to boolean
    //  stating whether to update icon
    _on: PBoolean;


    p: Ppatch_t;   // icon
    data: integer; // user data
  end;
  Pst_binicon_t = ^st_binicon_t;

//
// Widget creation, access, and update routines
//

// Initializes widget library.
// More precisely, initialize STMINUS,
//  everything else is done somewhere else.
//
procedure STlib_init;

// Number widget routines
procedure STlib_initNum(n: Pst_number_t; x, y: integer; pl: Ppatch_tPArray;
  num: PInteger; _on: PBoolean; width: integer);

procedure STlib_updateNum(n: Pst_number_t; refresh: boolean);

// Percent widget routines
procedure STlib_initPercent(p: Pst_percent_t; x, y: integer; pl: Ppatch_tPArray;
  num: PInteger; _on: PBoolean; percent: Ppatch_t);

procedure STlib_updatePercent(per: Pst_percent_t; refresh: integer);

// Multiple Icon widget routines
procedure STlib_initMultIcon(i: Pst_multicon_t; x, y: integer; il: Ppatch_tPArray;
  inum: PInteger; _on: PBoolean);

procedure STlib_updateMultIcon(mi: Pst_multicon_t; refresh: boolean);

// Binary Icon widget routines
procedure STlib_initBinIcon(b: Pst_binicon_t; x, y: integer; i: Ppatch_t;
  val: PBoolean; _on: PBoolean);

procedure STlib_updateBinIcon(bi: Pst_binicon_t; refresh: boolean);


(* Unit p_sight interface *)


                           

{
    p_sight.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	LineOfSight/Visibility checks, uses REJECT Lookup Table.
//
//-----------------------------------------------------------------------------

function P_CheckSight(t1: Pmobj_t; t2: Pmobj_t): boolean;

var
  bottomslope: fixed_t; // slopes to top and bottom of target
  topslope: fixed_t;


(* Unit p_local interface *)


                                   

{
    p_local.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//	Play functions, animation, global header. 
// 
//-----------------------------------------------------------------------------

const
  TOCENTER = -8;

  FLOATSPEED = FRACUNIT * 4;

  MAXHEALTH = 100;
// Player VIEWHEIGHT
  PVIEWHEIGHT = 41 * FRACUNIT;

// mapblocks are used to check movement
// against lines and things
  MAPBLOCKUNITS = 128;
  MAPBLOCKSIZE = MAPBLOCKUNITS * FRACUNIT;

  MAPBLOCKSHIFT = FRACBITS + 7;
  MAPBMASK = MAPBLOCKSIZE - 1;
  MAPBTOFRAC = MAPBLOCKSHIFT - FRACBITS;

// player radius for movement checking
  PLAYERRADIUS = 16 * FRACUNIT;

// MAXRADIUS is for precalculated sector block boxes
// the spider demon is larger,
// but we do not have any moving sectors nearby
  MAXRADIUS	= 32 * FRACUNIT;

  GRAVITY = FRACUNIT;
  MAXMOVE = 30 * FRACUNIT;

  USERANGE = 64 * FRACUNIT;
  MELEERANGE = 64 * FRACUNIT;
  MISSILERANGE = (32 * 64) * FRACUNIT;

// follow a player exlusively for 3 seconds
  BASETHRESHOLD = 100;

function ONFLOORZ: integer;

function ONCEILINGZ: integer;

const
// Time interval for item respawning.
  ITEMQUESIZE = 128;


type
  divline_t = record
    x: fixed_t;
    y: fixed_t;
    dx: fixed_t;
    dy: fixed_t;
  end;
  Pdivline_t = ^divline_t;

  thingORline_t = record
    case longint of
      0: (thing: Pmobj_t);
      1: (line: Pline_t);
    end;

  intercept_t = record
    frac : fixed_t; // along trace line
    isaline : boolean;
    d: thingORline_t;
  end;
  Pintercept_t = ^intercept_t;

const
  MAXINTERCEPTS = 128;

type
  traverser_t = function (f: Pintercept_t): boolean;
  ltraverser_t = function (p: Pline_t): boolean;
  ttraverser_t = function (p: Pmobj_t): boolean;

const
  PT_ADDLINES = 1;
  PT_ADDTHINGS = 2;
  PT_EARLYOUT = 4;


(* Unit p_tick interface *)


                                 

{
    p_tick.h, p_tick.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Archiving: SaveGame I/O.
//	Thinker, Ticker.
//
//-----------------------------------------------------------------------------

//
// THINKERS
// All thinkers should be allocated by Z_Malloc
// so they can be operated on uniformly.
// The actual structures will vary in size,
// but the first element must be thinker_t.
//

var
// Both the head and tail of the thinker list.
  thinkercap: thinker_t;

procedure P_InitThinkers;

procedure P_AddThinker(thinker: Pthinker_t);

procedure P_RemoveThinker(thinker: Pthinker_t);

procedure P_Ticker;

var
  leveltime: integer;


(* Unit p_maputl interface *)


                                                          

{
    p_maputl.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Movement/collision utility functions,
//	as used by function in p_map.c.
//	BLOCKMAP Iterator functions,
//	and some PIT_* functions to use for iteration.
//
//-----------------------------------------------------------------------------

function P_AproxDistance(dx: fixed_t; dy: fixed_t): fixed_t;

function P_PointOnLineSide(x: fixed_t; y: fixed_t; line: Pline_t): integer;

function P_BoxOnLineSide(tmbox: Pfixed_tArray; ld: Pline_t): integer;

function P_InterceptVector(v2: Pdivline_t; v1: Pdivline_t): fixed_t;

procedure P_LineOpening(linedef: Pline_t);

procedure P_UnsetThingPosition(thing: Pmobj_t);

procedure P_SetThingPosition(thing: Pmobj_t);

function P_BlockLinesIterator(x, y: integer; func: ltraverser_t): boolean;

function P_BlockThingsIterator(x, y: integer; func: ttraverser_t): boolean;

function P_PathTraverse(x1, y1, x2, y2: fixed_t; flags: integer;
  trav: traverser_t): boolean;

var
  opentop: fixed_t;
  openbottom: fixed_t;
  openrange: fixed_t;
  lowfloor: fixed_t;

  trace: divline_t;
  

(* Unit p_setup interface *)


                                                                         

{
    p_setup.h, p_setup.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//   Setup a game, startup stuff.
//	Do all the WAD I/O, get map description,
//	set up initial state and misc. LUTs.
//
//-----------------------------------------------------------------------------

// NOT called by W_Ticker. Fixme.
procedure P_SetupLevel(episode, map, playermask: integer; skill: skill_t);

// Called by startup code.
procedure P_Init;

var
// origin of block map
  bmaporgx: fixed_t;
  bmaporgy: fixed_t;

  numvertexes: integer;
  vertexes: Pvertex_tArray;

  numsegs: integer;
  segs: Pseg_tArray;

  numsectors: integer;
  sectors: Psector_tArray;

  numsubsectors: integer;
  subsectors: Psubsector_tArray;

  numnodes: integer;
  nodes: Pnode_tArray;

  numlines: integer;
  lines: Pline_tArray;

  numsides: integer;
  sides: Pside_tArray;

//
// MAP related Lookup tables.
// Store VERTEXES, LINEDEFS, SIDEDEFS, etc.
//
var

// BLOCKMAP
// Created from axis aligned bounding box
// of the map, a rectangular array of
// blocks of size ...
// Used to speed up collision detection
// by spatial subdivision in 2D.
//
// Blockmap size.
  bmapwidth: integer;
  bmapheight: integer; // size in mapblocks
  blockmap: PSmallIntArray; // int for larger maps
// offsets in blockmap are from here
  blockmaplump: PSmallIntArray;
// for thing chains
  blocklinks: Pmobj_tPArray;

// REJECT
// For fast sight rejection.
// Speeds up enemy AI by skipping detailed
//  LineOf Sight calculation.
// Without special effect, this could be
//  used as a PVS lookup as well.
//
  rejectmatrix: PByteArray;

const
// Maintain single and multi player starting spots.
  MAX_DEATHMATCH_STARTS = 10;

var
  deathmatchstarts: array[0..MAX_DEATHMATCH_STARTS - 1] of mapthing_t;
  deathmatch_p: integer; // VJ

  playerstarts: array[0..MAXPLAYERS - 1] of mapthing_t;


(* Unit p_mobj interface *)


{
    p_mobj.h, p_mobj.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Map Objects, MObj, definition and handling.
//  Moving object handling. Spawn functions.
//
//-----------------------------------------------------------------------------

                                                                                                                                                                                                                                                                   

function P_SetMobjState(mobj: Pmobj_t; state: statenum_t): boolean;

procedure P_MobjThinker(mobj: Pmobj_t);

function P_SpawnMobj(x, y, z: fixed_t; _type: mobjtype_t): Pmobj_t;

procedure P_RemoveMobj(mobj: Pmobj_t);

procedure P_SpawnPlayer(mthing: Pmapthing_t);

procedure P_SpawnMapThing(mthing: Pmapthing_t);

procedure P_SpawnPuff(x, y, z: fixed_t);

function P_SpawnMissile(source: Pmobj_t; dest: Pmobj_t; _type: mobjtype_t): Pmobj_t;

procedure P_SpawnPlayerMissile(source: Pmobj_t; _type: mobjtype_t);

procedure P_RespawnSpecials;

procedure P_SpawnBlood(x, y, z: fixed_t; damage: integer);

var
  iquehead: integer; // Initialized at p_setup
  iquetail: integer; // Initialized at p_setup


(* Unit p_telept interface *)


                                      

{
    p_telept.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Teleportation.
//
//-----------------------------------------------------------------------------


//
// TELEPORTATION
//
function EV_Teleport(line: Pline_t; side: integer; thing: Pmobj_t): integer;


(* Unit d_player interface *)


                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    

{
    d_player.h
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
// 
// 
//----------------------------------------------------------------------------- 

// 
// Player states. 
//

type
  playerstate_t = (
  { Playing or camping. }
    PST_LIVE,
  { Dead on the ground, view follows killer. }
    PST_DEAD,
  { Ready to restart/respawn??? }
    PST_REBORN);

  { }
  { Player internal flags, for cheats and debug. }
  { }

const
  { No clipping, walk through barriers. }
    CF_NOCLIP = 1;
  { No damage, no health loss. }
    CF_GODMODE = 2;
  { Not really a cheat, just a debug aid. }
    CF_NOMOMENTUM = 4;

type
  { }
  { Extended player object info: player_t }
  { }
  player_t = record
    mo : Pmobj_t;
    playerstate : playerstate_t;
    cmd : ticcmd_t;

    // Determine POV,
    //  including viewpoint bobbing during movement.
    // Focal origin above r.z
    viewz : fixed_t;
    // Base height above floor for viewz.
    viewheight : fixed_t;
    // Bob/squat speed.
    deltaviewheight: fixed_t;
    // bounded/scaled total momentum.
    bob: fixed_t;

    lookdir: integer;
    centering: boolean;

    // This is only used between levels,
    // mo->health is used during levels.
    health : integer;
    armorpoints : integer;
    // Armor type is 0-2.
    armortype : integer;

    // Power ups. invinc and invis are tic counters.
    powers: array[0..Ord(NUMPOWERS) - 1] of integer;
    cards: array[0..Ord(NUMCARDS) - 1] of boolean;
    backpack: boolean;

    // Frags, kills of other players.
    frags: array[0..(MAXPLAYERS)-1] of integer;
    readyweapon: weapontype_t;

    // Is wp_nochange if not changing.
    pendingweapon : weapontype_t;

    weaponowned: array[0..Ord(NUMWEAPONS) - 1] of integer;
    ammo: array[0..Ord(NUMAMMO) - 1] of integer;
    maxammo: array[0..Ord(NUMAMMO) - 1] of integer;

    // True if button down last tic.
    attackdown : boolean;
    usedown : boolean;

    // Bit flags, for cheats and debug.
    // See cheat_t, above.
    cheats : integer;

    // Refired shots are less accurate.
    refire : integer;

    // For intermission stats.
    killcount : integer;
    itemcount : integer;
    secretcount : integer;

    // Hint messages.
    _message : string[255];

    // For screen flashing (red or bright).
    damagecount : integer;
    bonuscount : integer;

    // Who did damage (NULL for floors/ceilings).
    attacker : Pmobj_t;

    // So gun flashes light up areas.
    extralight : integer;

    // Current PLAYPAL, ???
    //  can be set to REDCOLORMAP for pain, etc.
    fixedcolormap : integer;

    // Player skin colorshift,
    //  0-3 for which color to draw player.
    colormap : integer;

    // Overlay view sprites (gun, etc).
    psprites : array[0..Ord(NUMPSPRITES) - 1] of pspdef_t;

    // True if secret level has been done.
    didsecret : boolean;
  end;

  player_s = player_t;

  Pplayer_s = ^player_s;
  Pplayer_t = ^player_t;

  { }
  { INTERMISSION }
  { Structure passed e.g. to WI_Start(wb) }
  { }

  wbplayerstruct_t = record
    _in : boolean; // whether the player is in game

    // Player stats, kills, collected items etc.
    skills : integer;
    sitems : integer;
    ssecret : integer;
    stime : integer;
    frags : array[0..3] of integer;
    score : integer; // current score on entry, modified on return
  end;
  Pwbplayerstruct_t = ^wbplayerstruct_t;
  wbplayerstruct_tArray = packed array[0..$FFFF] of wbplayerstruct_t;
  Pwbplayerstruct_tArray = ^wbplayerstruct_tArray;

  wbstartstruct_t = record
    epsd: integer; // episode # (0-2)

    // if true, splash the secret level
    didsecret: boolean;

    // previous and next levels, origin 0
    last: integer;
    next: integer;

    maxkills: integer;
    maxitems: integer;
    maxsecret: integer;
    maxfrags: integer;

    // the par time
    partime: integer;

    // index of this player in game
    pnum: integer;

    plyr: array[0..MAXPLAYERS-1] of wbplayerstruct_t;
  end;
  Pwbstartstruct_t = ^wbstartstruct_t;

var
// VJ -> moved from g_game
  players: array[0..MAXPLAYERS - 1] of player_t;


(* Unit p_spec interface *)


                                                                     

{
    p_spec.h, p_spec.c
}


// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:  none
//	Implements special effects:
//	Texture animation, height or lighting changes
//	 according to adjacent sectors, respective
//	 utility functions, etc.
//	Line Tag handling. Line and Sector triggers.
//
//-----------------------------------------------------------------------------

//
// End-level timer (-TIMER option)
//

const
//      Define values for map objects
  MO_TELEPORTMAN = 14;

// at game start
procedure P_InitPicAnims;

// at map load
procedure P_SpawnSpecials;

// every tic
procedure P_UpdateSpecials;

// when needed
procedure P_ShootSpecialLine(thing: Pmobj_t; line: Pline_t);

procedure P_CrossSpecialLine(linenum: integer; side: integer; thing: Pmobj_t);

procedure P_PlayerInSpecialSector(player: Pplayer_t);

function twoSided(sector: integer; line: integer): integer;

function getSector(currentSector: integer; line: integer; side: integer): Psector_t;

function getSide(currentSector: integer; line: integer; side: integer): Pside_t;

function P_FindLowestFloorSurrounding(sec: Psector_t): fixed_t;

function P_FindHighestFloorSurrounding(sec: Psector_t): fixed_t;

function P_FindNextHighestFloor(sec: Psector_t; currentheight: integer): fixed_t;

function P_FindLowestCeilingSurrounding(sec: Psector_t): fixed_t;

function P_FindHighestCeilingSurrounding(sec: Psector_t): fixed_t;

function P_FindSectorFromLineTag(line: Pline_t; start: integer): integer;

function P_FindMinSurroundingLight(sector: Psector_t; max: integer): integer;

function getNextSector(line: Pline_t; sec: Psector_t): Psector_t;

//
// SPECIAL
//
function EV_DoDonut(line: Pline_t): integer;

//
// P_LIGHTS
//
type
  fireflicker_t = record
    thinker: thinker_t;
    sector: Psector_t;
    count: integer;
    maxlight: integer;
    minlight: integer;
  end;
  Pfireflicker_t = ^fireflicker_t;

  lightflash_t = record
    thinker: thinker_t;
    sector: Psector_t;
    count: integer;
    maxlight: integer;
    minlight: integer;
    maxtime: integer;
    mintime: integer;
  end;
  Plightflash_t = ^lightflash_t;

  strobe_t = record
    thinker: thinker_t;
    sector: Psector_t;
    count: integer;
    minlight: integer;
    maxlight: integer;
    darktime: integer;
    brighttime: integer;
  end;
  Pstrobe_t = ^strobe_t;

  glow_t = record
    thinker: thinker_t;
    sector: Psector_t;
    minlight: integer;
    maxlight: integer;
    direction: integer;
  end;
  Pglow_t = ^glow_t;

const
  GLOWSPEED = 8;
  STROBEBRIGHT = 5;
  FASTDARK = 15;
  SLOWDARK = 35;

type
//
// P_SWITCH
//
  switchlist_t = record
    name1: string[8];
    name2: string[8];
    episode: smallint;
  end;
  Pswitchlist_t = ^switchlist_t;

  bwhere_e = (
    top,
    middle,
    bottom
  );

  button_t = record
    line: Pline_t;
    where: bwhere_e;
    btexture: integer;
    btimer: integer;
    soundorg: Pmobj_t;
  end;

const
 // max # of wall switches in a level
  MAXSWITCHES = 50;

 // 4 players, 4 buttons each at once, max.
   MAXBUTTONS = 16;

 // 1 second, in ticks. 
   BUTTONTIME = 35; // VJ 1000 / TICKRATE ???


type
//
// P_PLATS
//
  plat_e = (
    up,
    down,
    waiting,
    in_stasis
  );

  plattype_e = (
    perpetualRaise,
    downWaitUpStay,
    raiseAndChange,
    raiseToNearestAndChange,
    blazeDWUS
  );

  plat_t = record
    thinker: thinker_t;
    sector: Psector_t;
    speed: fixed_t;
    low: fixed_t;
    high: fixed_t;
    wait: integer;
    count: integer;
    status: plat_e;
    oldstatus: plat_e;
    crush: boolean;
    tag: integer;
    _type: plattype_e;
  end;
  Pplat_t = ^plat_t;

const
  PLATWAIT = 3;
  PLATSPEED = FRACUNIT;
  MAXPLATS = 30;

type
//
// P_DOORS
//
  vldoor_e = (
    normal,
    close30ThenOpen,
    close,
    open,
    raiseIn5Mins,
    blazeRaise,
    blazeOpen,
    blazeClose
  );

  vldoor_t = record
    thinker: thinker_t;
    _type: vldoor_e;
    sector: Psector_t;
    topheight: fixed_t;
    speed: fixed_t;

    // 1 = up, 0 = waiting at top, -1 = down
    direction: integer;

    // tics to wait at the top
    topwait: integer;
    // (keep in case a door going down is reset)
    // when it reaches 0, start going down
    topcountdown: integer;
  end;
  Pvldoor_t = ^vldoor_t;

const
  VDOORSPEED = FRACUNIT * 2;
  VDOORWAIT = 150;

type
//
// P_CEILNG
//
  ceiling_e = (
    lowerToFloor,
    raiseToHighest,
    lowerAndCrush,
    crushAndRaise,
    fastCrushAndRaise,
    silentCrushAndRaise
  );

  ceiling_t = record
    thinker: thinker_t;
    _type: ceiling_e;
    sector: Psector_t;
    bottomheight: fixed_t;
    topheight: fixed_t;
    speed: fixed_t;
    crush: boolean;
    // 1 = up, 0 = waiting, -1 = down
    direction: integer;

    // ID
    tag: integer;
    olddirection: integer;
  end;
  Pceiling_t = ^ceiling_t;

const
  CEILSPEED = FRACUNIT;
  CEILWAIT = 150;
  MAXCEILINGS = 30;

type
//
// P_FLOOR
//
  floor_e = (
    // lower floor to highest surrounding floor
    lowerFloor,

    // lower floor to lowest surrounding floor
    lowerFloorToLowest,

    // lower floor to highest surrounding floor VERY FAST
    turboLower,

    // raise floor to lowest surrounding CEILING
    raiseFloor,

    // raise floor to next highest surrounding floor
    raiseFloorToNearest,

    // raise floor to shortest height texture around it
    raiseToTexture,

    // lower floor to lowest surrounding floor
    //  and change floorpic
    lowerAndChange,

    raiseFloor24,
    raiseFloor24AndChange,
    raiseFloorCrush,

     // raise to next highest floor, turbo-speed
    raiseFloorTurbo,
    donutRaise,
    raiseFloor512
  );

  stair_e = (
    build8, // slowly build by 8
    turbo16 // quickly build by 16
  );

  floormove_t = record
    thinker: thinker_t;
    _type: floor_e;
    crush: boolean;
    sector: Psector_t;
    direction: integer;
    newspecial: integer;
    texture: smallint;
    floordestheight: fixed_t;
    speed: fixed_t;
  end;
  Pfloormove_t = ^floormove_t;

const
  FLOORSPEED = FRACUNIT;

type
  result_e = (
    ok,
    crushed,
    pastdest
  );


(* Unit p_plats interface *)


                                                                                                                               

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Plats (i.e. elevator platforms) code, raising/lowering.
//
//-----------------------------------------------------------------------------

var
  activeplats: array[0..MAXPLATS - 1] of Pplat_t;

procedure T_PlatRaise(plat: Pplat_t);

function EV_DoPlat(line: Pline_t; _type: plattype_e; amount: integer): integer;

procedure P_ActivateInStasis(tag: integer);

procedure EV_StopPlat(line: Pline_t);

procedure P_AddActivePlat(plat: Pplat_t);

procedure P_RemoveActivePlat(plat: Pplat_t);


(* Unit p_floor interface *)


                                                                                                        

{
    p_floor.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//  Floor animation: raising stairs.
//
//-----------------------------------------------------------------------------

//
// FLOORS
//

function T_MovePlane(sector: Psector_t; speed: fixed_t; dest: fixed_t;
  crush: boolean; floorOrCeiling: integer; direction: integer): result_e;

procedure T_MoveFloor(floor: Pfloormove_t);

function EV_DoFloor(line: Pline_t; floortype: floor_e): integer;

function EV_BuildStairs(line: Pline_t; _type: stair_e): integer;


(* Unit p_ceilng interface *)


                                                                                                              

{
    p_ceiling.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:  Ceiling aninmation (lowering, crushing, raising)
//
//-----------------------------------------------------------------------------

var
  activeceilings: array[0..MAXCEILINGS - 1] of Pceiling_t;

procedure T_MoveCeiling(ceiling: Pceiling_t);

function EV_DoCeiling(line: Pline_t; _type: ceiling_e): integer;

procedure P_AddActiveCeiling(c: Pceiling_t);

function EV_CeilingCrushStop(line: Pline_t): integer;


(* Unit p_lights interface *)


                                                         

{
    p_lights.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Handle Sector base lighting effects.
//	Muzzle flash?
//
//-----------------------------------------------------------------------------

procedure P_SpawnFireFlicker(sector: Psector_t);

procedure P_SpawnLightFlash(sector: Psector_t);

procedure T_StrobeFlash(flash: Pstrobe_t);

procedure P_SpawnStrobeFlash(sector: Psector_t; fastOrSlow, inSync: integer);

procedure EV_StartLightStrobing(line: Pline_t);

procedure EV_TurnTagLightsOff(line: Pline_t);

procedure EV_LightTurnOn(line: Pline_t; bright: integer);

procedure T_Glow(g: Pglow_t);

procedure P_SpawnGlowingLight(sector: Psector_t);

procedure T_LightFlash(flash: Plightflash_t);


(* Unit p_doors interface *)


                                                                                                                                            

{
    p_doors.c
}


// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION: Door animation code (opening/closing)
//
//-----------------------------------------------------------------------------

procedure T_VerticalDoor(door: Pvldoor_t);

function EV_DoLockedDoor(line: Pline_t; _type: vldoor_e; thing: Pmobj_t): integer;

function EV_DoDoor(line: Pline_t; _type: vldoor_e): integer;

procedure EV_VerticalDoor(line: Pline_t; thing: Pmobj_t);

procedure P_SpawnDoorCloseIn30(sec: Psector_t);

procedure P_SpawnDoorRaiseIn5Mins(sec: Psector_t; secnum: integer);


(* Unit p_switch interface *)


                                  

{
    p_switch.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
//
// $Log:$
//
// DESCRIPTION:
//	Switches, buttons. Two-state animation. Exits.
//
//-----------------------------------------------------------------------------

procedure P_InitSwitchList;

procedure P_ChangeSwitchTexture(line: Pline_t; useAgain: integer);

function P_UseSpecialLine(thing: Pmobj_t; line: Pline_t; side: integer): boolean;

var
  buttonlist: array[0..MAXBUTTONS - 1] of button_t;


(* Unit r_main interface *)


                                                                               

{
    r_main.h, r_main.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	System specific interface stuff.
//	Rendering main loop and setup functions,
//	 utility functions (BSP, geometry, trigonometry).
//	See tables.c, too.
//
//-----------------------------------------------------------------------------

const
//
// Lighting LUT.
// Used for z-depth cuing per column/row,
//  and other lighting effects (sector ambient, flash).
//

// Lighting constants.
// Now why not 32 levels here?
  LIGHTLEVELS = 16;
  LIGHTSEGSHIFT = 4;

  MAXLIGHTSCALE = 48;
  LIGHTSCALESHIFT = 12;
  MAXLIGHTZ = 128;
  LIGHTZSHIFT = 20;

// Number of diminishing brightness levels.
// There a 0-31, i.e. 32 LUT in the COLORMAP lump.
  NUMCOLORMAPS = 32;

//
// Utility functions.
function R_PointOnSide(x: fixed_t; y: fixed_t; node: Pnode_t): integer;

function R_PointOnSegSide(x: fixed_t; y: fixed_t; line: Pseg_t): boolean;

function R_PointToAngle(x: fixed_t; y: fixed_t): angle_t;

function R_PointToAngle2(x1: fixed_t; y1: fixed_t; x2: fixed_t; y2: fixed_t): angle_t;

function R_PointToDist(x: fixed_t; y: fixed_t): fixed_t;

function R_ScaleFromGlobalAngle(visangle: angle_t): fixed_t;

function R_PointInSubsector(x: fixed_t; y: fixed_t): Psubsector_t;

procedure R_AddPointToBox(const x: integer; const y: integer; box: Pfixed_tArray);


//
// REFRESH - the actual rendering functions.
//

// Called by G_Drawer.
procedure R_RenderPlayerView(player: Pplayer_t);

// Called by startup code.
procedure R_Init;

// Called by M_Responder.
procedure R_SetViewSize(blocks: integer; detail: integer);

procedure R_ExecuteSetViewSize;

var
  colfunc: PProcedure;
  skycolfunc: PProcedure;
  basecolfunc: PProcedure;
  fuzzcolfunc: PProcedure;
  spanfunc: PProcedure;

// 0 = high, 1 = low
  detailshift: integer;

  centerxfrac: fixed_t;
  centeryfrac: fixed_t;

  viewx: fixed_t;
  viewy: fixed_t;
  viewz: fixed_t;

  viewangle: angle_t;

  viewcos: fixed_t;
  viewsin: fixed_t;

  projection: fixed_t;

  centerx: integer;
  centery: integer;

  fixedcolormap: Plighttable_tArray;

// increment every time a check is made
  validcount: integer = 1;

// bumped light from gun blasts
  extralight: integer;

  scalelight: array[0..LIGHTLEVELS - 1, 0..MAXLIGHTSCALE - 1] of Plighttable_tArray;
  scalelightfixed: array[0..MAXLIGHTSCALE - 1] of Plighttable_tArray;
  zlight: array[0..LIGHTLEVELS - 1, 0..MAXLIGHTZ - 1] of Plighttable_tArray;

  viewplayer: Pplayer_t;

// The viewangletox[viewangle + FINEANGLES/4] lookup
// maps the visible view angles to screen X coordinates,
// flattening the arc to a flat projection plane.
// There will be many angles mapped to the same X.
  viewangletox: array[0..FINEANGLES div 2 - 1] of integer;

// The xtoviewangleangle[] table maps a screen pixel
// to the lowest viewangle that maps back to x ranges
// from clipangle to -clipangle.
  xtoviewangle: array[0..MAXWIDTH] of angle_t;

//
// precalculated math tables
//
  clipangle: angle_t;

// UNUSED.
// The finetangentgent[angle+FINEANGLES/4] table
// holds the fixed_t tangent values for view angles,
// ranging from MININT to 0 to MAXINT.
// fixed_t		finetangent[FINEANGLES/2];

// fixed_t		finesine[5*FINEANGLES/4];
// fixed_t*		finecosine = &finesine[FINEANGLES/4]; // VJ -> moved to tables.pas


  sscount: integer;
  linecount: integer;
  loopcount: integer;

  viewangleoffset: angle_t = 0; // never a value assigned to this variable!

  setsizeneeded: boolean;


(* Unit d_net interface *)


                                                  

{ 
    d_net.h, d_net.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  Networking stuff. 
// 
//----------------------------------------------------------------------------- 

//
// Network play related stuff.
// There is a data struct that stores network
//  communication related stuff, and another
//  one that defines the actual packets to
//  be transmitted.
//

const
  DOOMCOM_ID = $12345678;

// Max computers/players in a game.
  MAXNETNODES = 8;

// Networking and tick handling related.
  BACKUPTICS = 12;

  CMD_SEND = 1;
  CMD_GET = 2;

type
  Tcmds = packed array[0..(BACKUPTICS)-1] of ticcmd_t;
//
// Network packet data.
//
  doomdata_t = record
    // High bit is retransmit request.
    checksum : LongWord;
    // Only valid if NCMD_RETRANSMIT.
    retransmitfrom : byte;

    starttic : byte;
    player : byte;
    numtics : byte;
    cmds: Tcmds;
  end;
  Pdoomdata_t = ^doomdata_t;

  doomcom_t = record
    // Supposed to be DOOMCOM_ID?
    id : integer;

    // DOOM executes an int to execute commands.
    intnum : smallint;

    // Communication between DOOM and the driver.
    // Is CMD_SEND or CMD_GET.
    command : smallint;
    // Is dest for send, set by get (-1 = no packet).
    remotenode : smallint;

    // Number of bytes in doomdata to be sent
    datalength : smallint;

    // Info common to all nodes.
    // Console is allways node 0.
    numnodes: smallint;
    // Flag: 1 = no duplication, 2-5 = dup for slow nets.
    ticdup: smallint;
    // Flag: 1 = send a backup tic in every packet.
    extratics: smallint;
    // Flag: 1 = deathmatch.
    deathmatch: smallint;
    // Flag: -1 = new game, 0-5 = load savegame
    savegame: smallint;
    episode: smallint; // 1-3
    map: smallint; // 1-9
    skill: smallint; // 1-5

    // Info specific to this node.
    consoleplayer: smallint;
    numplayers: smallint;

    // These are related to the 3-display mode,
    //  in which two drones looking left and right
    //  were used to render two additional views
    //  on two additional computers.
    // Probably not operational anymore.
    // 1 = left, 0 = center, -1 = right
    angleoffset: smallint;

    // 1 = drone
    drone : smallint;
    // The packet data to be sent.
    data : doomdata_t;
  end;
  Pdoomcom_t = ^doomcom_t;

{ Create any new ticcmds and broadcast to other players. }
procedure NetUpdate;

{ Broadcasts special packets to other players }
{  to notify of game exit }
procedure D_QuitNetGame;

{? how many ticks to run? }
procedure TryRunTics;

procedure CheckAbort;

procedure D_CheckNetGame;

var
  netcmds: array[0..MAXPLAYERS - 1] of array[0..(BACKUPTICS)-1] of ticcmd_t;

  maketic: integer;

  doomcom: Pdoomcom_t;
  ticdup: integer;


(* Unit wi_stuff interface *)


               

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Intermission screens.
//
//-----------------------------------------------------------------------------

const
// States for the intermission
  NoState = -1;
  StatCount = 0;
  ShowNextLoc = 1;

type
  stateenum_t = integer;

// Called by main loop, animate the intermission.
procedure WI_Ticker;

// Called by main loop,
// draws the intermission directly into the screen buffer.
procedure WI_Drawer;

// Setup for an intermission screen.
procedure WI_Start(wbstartstruct: Pwbstartstruct_t);


(* Unit p_enemy interface *)


                                                                                                           

{
    p_enemy.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//  Enemy thinking, AI.
//  Action Pointer Functions
//  that are associated with states/frames.
//
//-----------------------------------------------------------------------------

procedure A_Fall(actor: Pmobj_t);

procedure A_KeenDie(mo: Pmobj_t);

procedure A_Look(actor: Pmobj_t);

procedure A_Chase(actor: Pmobj_t);

procedure A_FaceTarget(actor: Pmobj_t);

procedure A_PosAttack(actor: Pmobj_t);

procedure A_SPosAttack(actor: Pmobj_t);

procedure A_CPosAttack(actor: Pmobj_t);

procedure A_CPosRefire(actor: Pmobj_t);

procedure A_SpidRefire(actor: Pmobj_t);

procedure A_BspiAttack(actor: Pmobj_t);

procedure A_TroopAttack(actor: Pmobj_t);

procedure A_SargAttack(actor: Pmobj_t);

procedure A_HeadAttack(actor: Pmobj_t);

procedure A_CyberAttack(actor: Pmobj_t);

procedure A_BruisAttack(actor: Pmobj_t);

procedure A_SkelMissile(actor: Pmobj_t);

procedure A_Tracer(actor: Pmobj_t);

procedure A_SkelWhoosh(actor: Pmobj_t);

procedure A_SkelFist(actor: Pmobj_t);

procedure A_VileChase(actor: Pmobj_t);

procedure A_VileStart(actor: Pmobj_t);

procedure A_Fire(actor: Pmobj_t);

procedure A_StartFire(actor: Pmobj_t);

procedure A_FireCrackle(actor: Pmobj_t);

procedure A_VileTarget(actor: Pmobj_t);

procedure A_VileAttack(actor: Pmobj_t);

procedure A_FatRaise(actor: Pmobj_t);

procedure A_FatAttack1(actor: Pmobj_t);

procedure A_FatAttack2(actor: Pmobj_t);

procedure A_FatAttack3(actor: Pmobj_t);

procedure A_SkullAttack(actor: Pmobj_t);

procedure A_PainAttack(actor: Pmobj_t);

procedure A_PainDie(actor: Pmobj_t);

procedure A_Scream(actor: Pmobj_t);

procedure A_XScream(actor: Pmobj_t);

procedure A_Pain(actor: Pmobj_t);

procedure A_Explode(thingy: Pmobj_t);

procedure A_BossDeath(mo: Pmobj_t);

procedure A_Hoof(mo: Pmobj_t);

procedure A_Metal(mo: Pmobj_t);

procedure A_BabyMetal(mo: Pmobj_t);

procedure A_OpenShotgun2(player: Pplayer_t; psp: Pplayer_t);

procedure A_LoadShotgun2(player: Pplayer_t; psp: Ppspdef_t);

procedure A_CloseShotgun2(player: Pplayer_t; psp: Ppspdef_t);

procedure A_BrainAwake(mo: Pmobj_t);

procedure A_BrainPain(mo: Pmobj_t);

procedure A_BrainScream(mo: Pmobj_t);

procedure A_BrainExplode(mo: Pmobj_t);

procedure A_BrainDie(mo: Pmobj_t);

procedure A_BrainSpit(mo: Pmobj_t);

procedure A_SpawnFly(mo: Pmobj_t);

procedure A_SpawnSound(mo: Pmobj_t);

procedure A_PlayerScream(mo: Pmobj_t);


procedure P_NoiseAlert(target: Pmobj_t; emmiter: Pmobj_t);


(* Unit am_map interface *)


                                                                                                                                                                   

{
    am_map.h, am_map.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  AutoMap module. 
// 
//-----------------------------------------------------------------------------

const
  AM_MSGHEADER = (ord('a') shl 24) + (ord('m') shl 16);
  AM_MSGENTERED = AM_MSGHEADER or (ord('e') shl 8);
  AM_MSGEXITED = AM_MSGHEADER or (ord('x') shl 8);

  rcsid = '$Id: am_map.c,v 1.4 1997/02/03 21:24:33 b1 Exp $';

  { Called by main loop. }
const
  REDS = 256 - (5 * 16);
  REDRANGE = 16;
  BLUES = (256 - (4 * 16)) + 8;
  BLUERANGE = 8;
  GREENS = 7 * 16;
  GREENRANGE = 16;
  GRAYS = 6 * 16;
  GRAYSRANGE = 16;
  BROWNS = 4 * 16;
  BROWNRANGE = 16;
  YELLOWS = (256 - 32) + 7;
  YELLOWRANGE = 1;
  BLACK = 0;
  WHITE = 256 - 47;

  { Automap colors }
  BACKGROUND = BLACK;
  YOURCOLORS = WHITE;
  YOURRANGE = 0;
  WALLCOLORS = REDS;
  WALLRANGE = REDRANGE;
  TSWALLCOLORS = GRAYS;
  TSWALLRANGE = GRAYSRANGE;
  FDWALLCOLORS = BROWNS;
  FDWALLRANGE = BROWNRANGE;
  CDWALLCOLORS = YELLOWS;
  CDWALLRANGE = YELLOWRANGE;
  THINGCOLORS = GREENS;
  THINGRANGE = GREENRANGE;
  SECRETWALLCOLORS = WALLCOLORS;
  SECRETWALLRANGE = WALLRANGE;
  GRIDCOLORS = GRAYS + (GRAYSRANGE div 2);
  GRIDRANGE = 0;
  XHAIRCOLORS = GRAYS;

  AM_PANDOWNKEY = KEY_DOWNARROW;
  AM_PANUPKEY = KEY_UPARROW;
  AM_PANRIGHTKEY = KEY_RIGHTARROW;
  AM_PANLEFTKEY = KEY_LEFTARROW;
  AM_ZOOMINKEY = '=';
  AM_ZOOMOUTKEY = '-';
  AM_STARTKEY = KEY_TAB;
  AM_ENDKEY = KEY_TAB;
  AM_GOBIGKEY = '0';
  AM_FOLLOWKEY = 'f';
  AM_GRIDKEY = 'g';
  AM_MARKKEY = 'm';
  AM_CLEARMARKKEY = 'c';

  AM_NUMMARKPOINTS = 10;

// scale on entry
  INITSCALEMTOF = FRACUNIT div 5;

// how much the automap moves window per tic in frame-buffer coordinates }
// moves 140 pixels in 1 second }
  F_PANINC = 4;

{ how much zoom-in per tic }
function M_ZOOMIN: integer;

{ how much zoom-out per tic }
function M_ZOOMOUT: integer;

{ translates between frame-buffer and map distances }
function FTOM(x : integer) : integer;
function MTOF(x : integer) : integer;

{ translates between frame-buffer and map coordinates }
function CXMTOF(x : integer) : integer;
function CYMTOF(y : integer) : integer;

{ the following is crap }

const
  LINE_NEVERSEE = ML_DONTDRAW;

type
  fpoint_t = record
    x : integer;
    y : integer;
  end;
  Pfpoint_t = ^fpoint_t;

  fline_t = record
    a : fpoint_t;
    b : fpoint_t;
  end;
  Pfline_t = ^fline_t;

  mpoint_t = record
    x : fixed_t;
    y : fixed_t;
  end;
  Pmpoint_t = ^mpoint_t;

  mline_t = record
    a : mpoint_t;
    b : mpoint_t;
  end;
  Pmline_t = ^mline_t;
  mline_tArray = packed array[0..$FFFF] of mline_t;
  Pmline_tArray = ^mline_tArray;

  islope_t = record
    slp : fixed_t;
    islp : fixed_t;
  end;
  Pislope_t = ^islope_t;

//
// The vector graphics for the automap.
//  A line drawing of the player pointing right,
//   starting from the middle.
//
const
  NUMPLYRLINES = 7;

var
  player_arrow: array[0..NUMPLYRLINES - 1] of mline_t;

const
  NUMCHEATPLYRLINES = 16;

var
  cheat_player_arrow: array[0..NUMCHEATPLYRLINES - 1] of mline_t;

const
  NUMTRIANGLEGUYLINES = 3;

var
  triangle_guy: array[0..NUMTRIANGLEGUYLINES - 1] of mline_t;

const
  NUMTHINTRIANGLEGUYLINES = 3;

var
  thintriangle_guy: array[0..NUMTHINTRIANGLEGUYLINES - 1] of mline_t;

var
  cheating: integer = 0;
  grid: boolean = false;

  leveljuststarted: integer = 1;   // kluge until AM_LevelInit() is called

  automapactive: boolean = false;

// location of window on screen
  f_x: integer;
  f_y: integer;

// size of window on screen
  f_w: integer;
  f_h: integer;

  lightlev: integer;     // used for funky strobing effect
  fb: PByteArray;       // pseudo-frame buffer
  amclock: integer;

  m_paninc: mpoint_t;    // how far the window pans each tic (map coords)
  mtof_zoommul: fixed_t; // how far the window zooms in each tic (map coords)
  ftom_zoommul: fixed_t; // how far the window zooms in each tic (fb coords)

  m_x, m_y: fixed_t;    // LL x,y where the window is on the map (map coords)
  m_x2, m_y2: fixed_t;  // UR x,y where the window is on the map (map coords)

//
// width/height of window on map (map coords)
//
  m_w: fixed_t;
  m_h: fixed_t;

// based on level size
  min_x: fixed_t;
  min_y: fixed_t;
  max_x: fixed_t;
  max_y: fixed_t;

  max_w: fixed_t; // max_x-min_x,
  max_h: fixed_t; // max_y-min_y

// based on player size
  min_w: fixed_t;
  min_h: fixed_t;


  min_scale_mtof: fixed_t; // used to tell when to stop zooming out
  max_scale_mtof: fixed_t; // used to tell when to stop zooming in

// old stuff for recovery later
  old_m_w, old_m_h: fixed_t;
  old_m_x, old_m_y: fixed_t;

// old location used by the Follower routine
  f_oldloc: mpoint_t;

// used by MTOF to scale from map-to-frame-buffer coords
  scale_mtof: fixed_t = INITSCALEMTOF;
// used by FTOM to scale from frame-buffer-to-map coords (=1/scale_mtof)
  scale_ftom: fixed_t;

  plr: Pplayer_t; // the player represented by an arrow

type
  patch_t10 = packed array[0..9] of Ppatch_t;

var
  marknums: patch_t10; // numbers used for marking by the automap

  markpoints: array[0..AM_NUMMARKPOINTS-1] of mpoint_t; // where the points are

  markpointnum: integer = 0; // next point to be assigned

  followplayer: boolean = true; // specifies whether to follow the player around

const
  cheat_amap_seq: string = Chr($b2) + Chr($26) + Chr($26) + Chr($2e) + Chr($ff);

var
  cheat_amap: cheatseq_t;

  stopped: boolean = true;

function AM_Responder(ev: Pevent_t): boolean;

// Called by main loop. 
procedure AM_Ticker;

// Called by main loop, 
// called instead of view drawer if automap active. 
procedure AM_Drawer;

// Called to force the automap to quit 
// if the level is completed while it is up.
procedure AM_Stop;


(* Unit p_inter interface *)


                                                                                                                      

{
    p_inter.h, p_inter.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Handling interactions (i.e., collisions).
//
//-----------------------------------------------------------------------------

function P_GivePower(player: Pplayer_t; power: (*powertype_t*)integer): boolean;

procedure P_TouchSpecialThing(special: Pmobj_t; toucher: Pmobj_t);

procedure P_DamageMobj(target, inflictor, source: Pmobj_t; damage: integer);

const
// a weapon is found with two clip loads,
// a big item has five clip loads
  maxammo: array[0..Ord(NUMAMMO) - 1] of integer = (200, 50, 300, 50);
  clipammo: array[0..Ord(NUMAMMO) - 1] of integer = (10, 4, 20, 1);


(* Unit p_user interface *)


{
    p_user.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Player related stuff.
//	Bobbing POV/weapon, movement.
//	Pending weapon.
//
//-----------------------------------------------------------------------------

               

procedure P_PlayerThink(player: Pplayer_t);


(* Unit p_pspr interface *)


                                                                                                                                        

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//  Sprite animation.
//	Weapon sprite animation, weapon objects.
//	Action functions for weapons.
//
//-----------------------------------------------------------------------------

const
//
// Frame flags:
// handles maximum brightness (torches, muzzle flare, light sources)
//
  FF_FULLBRIGHT = $8000; // flag in thing->frame
  FF_FRAMEMASK = $7fff;

procedure P_DropWeapon(player: Pplayer_t);

procedure A_WeaponReady(player: Pplayer_t; psp: Ppspdef_t);

procedure A_ReFire(player: Pplayer_t; psp: Ppspdef_t);

procedure A_CheckReload(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Lower(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Raise(player: Pplayer_t; psp: Ppspdef_t);

procedure A_GunFlash(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Punch(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Saw(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FireMissile(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FireBFG(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FirePlasma(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FirePistol(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FireShotgun(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FireShotgun2(player: Pplayer_t; psp: Ppspdef_t);

procedure A_FireCGun(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Light0(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Light1(player: Pplayer_t; psp: Ppspdef_t);

procedure A_Light2(player: Pplayer_t; psp: Ppspdef_t);

procedure A_BFGSpray(mo: Pmobj_t);

procedure A_BFGsound(player: Pplayer_t; psp: Ppspdef_t);

procedure P_SetupPsprites(player: Pplayer_t);

procedure P_MovePsprites(player: Pplayer_t);


(* Unit d_main interface *)


                       

{
    d_main.h, d_main.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// $Log:$ 
// 
// DESCRIPTION (d_main.h: 
//  System specific interface stuff. 
// 
// DESCRIPTION (d_main.c: 
// DOOM main program (D_DoomMain) and game loop (D_DoomLoop), 
// plus functions to determine game mode (shareware, registered), 
// parse command line parameters, configure game parameters (turbo), 
// and call the startup functions. 
// 
//-----------------------------------------------------------------------------

const
   MAXWADFILES = 20;

var
  wadfiles : array[0..(MAXWADFILES)-1] of string;
  numwadfiles: integer = 0;

procedure D_ProcessEvents;
procedure D_DoAdvanceDemo;


procedure D_AddFile(_file: string);

//
// D_DoomMain()
// Not a globally visible function, just included for source reference,
// calls all startup code, parses command line options.
// If not overrided by user input, calls N_AdvanceDemo.
//
procedure D_DoomMain;

// Called by IO functions when input is detected.
procedure D_PostEvent(ev: Pevent_t);

//
// BASE LEVEL
//
procedure D_PageTicker;

procedure D_PageDrawer;

procedure D_AdvanceDemo;

procedure D_StartTitle;

// wipegamestate can be set to -1 to force a wipe on the next draw
var
  wipegamestate: integer; // VJ was gamestate_t = GS_DEMOSCREEN;

  nomonsters: boolean;        // checkparm of -nomonsters
  fastparm: boolean;          // checkparm of -fast
  devparm: boolean;       // started game with -devparm
  singletics: boolean = false; // debug flag to cancel adaptiveness
  autostart: boolean;
  startskill: skill_t;
  respawnparm: boolean;   // checkparm of -respawn

  startepisode: integer;
  startmap: integer;
  advancedemo: boolean;

  basedefault: string;  // default file



(* Unit g_game interface *)


                                                          

{
    g_game.h,  g_game.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//   Duh. 
//  
//----------------------------------------------------------------------------- 

// 
// GAME 
//

procedure G_DeathMatchSpawnPlayer(playernum:integer);

procedure G_InitNew(skill:skill_t; episode:integer; map:integer);

{ Can be called by the startup code or M_Responder. }
{ A normal game starts at map 1, }
{ but a warp test can start elsewhere }
procedure G_DeferedInitNew(skill:skill_t; episode: integer; map: integer);

procedure G_DeferedPlayDemo(const name: string);

{ Can be called by the startup code or M_Responder, }
{ calls P_SetupLevel or W_EnterWorld. }
procedure G_LoadGame(const name: string);

procedure G_DoLoadGame;

{ Called by M_Responder. }
procedure G_SaveGame(slot: integer; const description: string);

{ Only called by startup code. }
procedure G_RecordDemo(const name: string);

procedure G_BeginRecording;

procedure G_TimeDemo(const name: string);

function G_CheckDemoStatus: boolean;

procedure G_ExitLevel;

procedure G_SecretExitLevel;

procedure G_WorldDone;

procedure G_Ticker;

function G_Responder(ev: Pevent_t): boolean;

procedure G_ScreenShot;

function G_IsDemoRecording: boolean;

var
  sendpause: boolean;        // send a pause event next tic

  paused: boolean;
  
//
// controls (have defaults)
//
  key_right: integer;
  key_left: integer;

  key_up: integer;
  key_down: integer;

  key_lookup: integer;
  key_lookdown: integer;
  key_lookcenter: integer;

  key_strafeleft: integer;
  key_straferight: integer;
  key_fire: integer;
  key_use: integer;
  key_strafe: integer;
  key_speed: integer;

  usemouse: boolean;
  mousebfire: integer;
  mousebstrafe: integer;
  mousebforward: integer;

  usejoystick: boolean;
  joybfire: integer;
  joybstrafe: integer;
  joybuse: integer;
  joybspeed: integer;

  demoplayback: boolean;

  gameepisode: integer;
  gamemap: integer;

  deathmatch: integer; // only if started as net death
  netgame: boolean; // only true if packets are broadcast
  playeringame: array[0..MAXPLAYERS - 1] of boolean;

  consoleplayer: integer; // player taking events and displaying
  displayplayer: integer; // view being displayed
  gametic: integer;

  totalkills, totalitems, totalsecret: integer; // for intermission

  wminfo: wbstartstruct_t; // parms for world map / intermission

  gameskill: skill_t;

  bodyqueslot: integer;


  precache: boolean; // if true, load all graphics at start

  respawnmonsters: boolean;

  viewactive: boolean;

  singledemo: boolean; // quit after playing a demo from cmdline

  nodrawers: boolean; // for comparative timing purposes

  gamestate: gamestate_t;

  demorecording: boolean;

  gameaction: gameaction_t;

  usergame: boolean; // ok to save / end game

procedure G_PlayerReborn(player: integer);

procedure G_BuildTiccmd(cmd: Pticcmd_t);

var
  forwardmove: array[0..1] of fixed_t;
  sidemove: array[0..1] of fixed_t;
  angleturn: array[0..2] of fixed_t;


(* Unit i_video interface *)


                                           

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	System specific interface stuff.
//	DirectX DOOM graphics
//
//-----------------------------------------------------------------------------

// Called by D_DoomMain,
// determines the hardware configuration
// and sets up the video mode
procedure I_InitGraphics;


procedure I_ShutdownGraphics;

// Takes full 8 bit values.
procedure I_SetPalette(palette: PByteArray);

procedure I_UpdateNoBlit;
procedure I_FinishUpdate;

procedure I_ReadScreen(scr: PByteArray);

procedure I_ReadScreen32bit(scr: PByteArray);

var
  fullscreen: boolean = true;


(* Unit p_map interface *)


                                                                                                                                                            

{
    p_map.c
}

// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Movement, collision handling.
//	Shooting and aiming.
//
//-----------------------------------------------------------------------------

function P_TeleportMove(thing: Pmobj_t; x, y: fixed_t): boolean;

function P_CheckPosition(thing: Pmobj_t; x, y: fixed_t): boolean;

function P_TryMove(thing: Pmobj_t; x, y: fixed_t): boolean;

function P_AimLineAttack(t1: Pmobj_t; angle: angle_t; distance: fixed_t): fixed_t;

procedure P_LineAttack(t1: Pmobj_t; angle: angle_t;
  distance: fixed_t; slope: fixed_t; damage: integer);

procedure P_UseLines(player: Pplayer_t);

procedure P_RadiusAttack(spot: Pmobj_t; source: Pmobj_t; damage: integer);

function P_ChangeSector(sector: Psector_t; crunch: boolean): boolean;

procedure P_SlideMove(mo: Pmobj_t);

var
  linetarget: Pmobj_t;	// who got hit (or NULL)

// If "floatok" true, move would be ok
// if within "tmfloorz - tmceilingz".
  floatok: boolean;

  tmfloorz: fixed_t;
  tmceilingz: fixed_t;
  tmdropoffz: fixed_t;

const
// keep track of special lines as they are hit,
// but don't process them until the move is proven valid
  MAXSPECIALCROSS = 64;

var
  spechit: array[0..MAXSPECIALCROSS - 1] of Pline_t;
  numspechit: integer;

// keep track of the line that lowers the ceiling,
// so missiles don't explode against sky hack walls
  ceilingline: Pline_t;                                    

  attackrange: fixed_t;
  


(* Unit r_things interface *)


                                                           

{
    r_things.h, r_things.c
}

// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	Rendering of moving objects, sprites.
//	Refresh of things, i.e. objects represented by sprites.
//
//-----------------------------------------------------------------------------

const
  MAXVISSPRITES = 128;

procedure R_DrawMaskedColumn(column: Pcolumn_t);


procedure R_SortVisSprites;

procedure R_AddSprites(sec: Psector_t);
procedure R_InitSprites(namelist: PsprnamesArray_t);
procedure R_ClearSprites;
procedure R_DrawMasked;

var
  pspritescale: fixed_t;
  pspriteiscale: fixed_t;

var
  mfloorclip: PSmallIntArray;
  mceilingclip: PSmallIntArray;
  spryscale: fixed_t;
  sprtopscreen: fixed_t;

// constant arrays
//  used for psprite clipping and initializing clipping
  negonearray: packed array[0..MAXWIDTH - 1] of smallint;
  screenheightarray: packed array[0..MAXWIDTH - 1] of smallint;

// variables used to look up
//  and range check thing_t sprites patches
  sprites: Pspritedef_tArray;
  r_things_numsprites: integer;
  

(* Unit i_net interface *)


{
    i_net.h, i_net.c
}

// Emacs style mode select   -*- C++ -*-  
//----------------------------------------------------------------------------- 
// 
// $Id:$ 
// 
// Copyright (C) 1993-1996 by id Software, Inc. 
// 
// This source is available for distribution and/or modification 
// only under the terms of the DOOM Source Code License as 
// published by id Software. All rights reserved. 
// 
// The source is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License 
// for more details. 
// 
// DESCRIPTION: 
//  System specific network interface stuff. 
// 
//----------------------------------------------------------------------------- 

{ Called by D_DoomMain. }

procedure I_InitNetwork;

procedure I_NetCmd;




implementation

(* Unit m_rnd implementation *)


const
  rndtable: array[0..255] of byte = (
    0,   8, 109, 220, 222, 241, 149, 107,  75, 248, 254, 140,  16,  66 ,
    74,  21, 211,  47,  80, 242, 154,  27, 205, 128, 161,  89,  77,  36 ,
    95, 110,  85,  48, 212, 140, 211, 249,  22,  79, 200,  50,  28, 188 ,
    52, 140, 202, 120,  68, 145,  62,  70, 184, 190,  91, 197, 152, 224 ,
    149, 104,  25, 178, 252, 182, 202, 182, 141, 197,   4,  81, 181, 242 ,
    145,  42,  39, 227, 156, 198, 225, 193, 219,  93, 122, 175, 249,   0 ,
    175, 143,  70, 239,  46, 246, 163,  53, 163, 109, 168, 135,   2, 235 ,
    25,  92,  20, 145, 138,  77,  69, 166,  78, 176, 173, 212, 166, 113 ,
    94, 161,  41,  50, 239,  49, 111, 164,  70,  60,   2,  37, 171,  75 ,
    136, 156,  11,  56,  42, 146, 138, 229,  73, 146,  77,  61,  98, 196 ,
    135, 106,  63, 197, 195,  86,  96, 203, 113, 101, 170, 247, 181, 113 ,
    80, 250, 108,   7, 255, 237, 129, 226,  79, 107, 112, 166, 103, 241 ,
    24, 223, 239, 120, 198,  58,  60,  82, 128,   3, 184,  66, 143, 224 ,
    145, 224,  81, 206, 163,  45,  63,  90, 168, 114,  59,  33, 159,  95 ,
    28, 139, 123,  98, 125, 196,  15,  70, 194, 253,  54,  14, 109, 226 ,
    71,  17, 161,  93, 186,  87, 244, 138,  20,  52, 123, 251,  26,  36 ,
    17,  46,  52, 231, 232,  76,  31, 221,  84,  37, 216, 165, 212, 106 ,
    197, 242,  98,  43,  39, 175, 254, 145, 190,  84, 118, 222, 187, 136 ,
    120, 163, 236, 249
  );

// Which one is deterministic?
function P_Random: integer;
begin
  prndindex := (prndindex + 1) and $ff;
  result := rndtable[prndindex];
end;

function M_Random: integer;
begin
  rndindex := (rndindex + 1) and $ff;
  result := rndtable[rndindex];
end;

procedure M_ClearRandom;
begin
  rndindex := 0;
  prndindex := 0;
end;



(* Unit f_wipe implementation *)


                                                                           

//
//                       SCREEN WIPE PACKAGE
//

// when zero, stop the wipe
var
  go: boolean = false;

var  
  wipe_scr_start: PByteArray;
  wipe_scr_end: PByteArray;
  wipe_scr: PByteArray;

procedure wipe_shittyColMajorXform(_array: PByteArray; width, height: integer);
var
  x, y: integer;
  dest: PByteArray;
begin
  dest := PByteArray(Z_Malloc(width * height, PU_STATIC, nil));

  for y := 0 to height - 1 do
    for x := 0 to width - 1 do
      dest[x * height + y] := _array[y * width + x];

  memcpy(_array, dest, width * height);

  Z_Free(dest);
end;

function wipe_initColorXForm(width, height, ticks: integer): integer;
begin
  memcpy(wipe_scr, wipe_scr_start, width * height);
  result := 0;
end;

function wipe_doColorXForm(width, height, ticks: integer): integer;
var
  changed: boolean;
  w: integer;
  e: integer;
  newval: integer;
begin
  changed := false;
  w := 0;
  e := 0;

  while w < width * height do
  begin
    if wipe_scr[w] <> wipe_scr_end[e] then
    begin
      if wipe_scr[w] > wipe_scr_end[e] then
      begin
        newval := wipe_scr[w] - ticks;
        if newval < wipe_scr_end[e] then
          wipe_scr[w] := wipe_scr_end[e]
        else
          wipe_scr[w] := newval;
      end
      else if wipe_scr[w] < wipe_scr_end[e] then
      begin
        newval := wipe_scr[w] + ticks;
        if newval > wipe_scr_end[e] then
          wipe_scr[w] := wipe_scr_end[e]
        else
          wipe_scr[w] := newval;
      end;
      changed := true;
    end;
    inc(w);
    inc(e);
  end;

  result := intval(not changed);
end;

function wipe_exitColorXForm(width, height, ticks: integer): integer;
begin
  result := 0;
end;

var
  y: Pfixed_tArray;
  vy: fixed_t;

function wipe_initMelt(width, height, ticks: integer): integer;
var
  i, r: integer;
begin
  // copy start screen to main screen
  memcpy(wipe_scr, wipe_scr_start, width * height);

  wipe_shittyColMajorXform(wipe_scr_start, width, height);
  wipe_shittyColMajorXform(wipe_scr_end, width, height);

  // setup initial column positions
  // (y<0 => not ready to scroll yet)
  y := Pfixed_tArray(Z_Malloc(width * SizeOf(integer), PU_STATIC, nil));
  y[0] := -(M_Random mod 16);
  for i := 1 to width - 1 do
  begin
    r := (M_Random mod 3) - 1;
    y[i] := y[i - 1] + r;
    if y[i] > 0 then
      y[i] := 0
    else if y[i] = -16 then
      y[i] := -15;
  end;

  // VJ change wipe timing
  vy := FRACUNIT * SCREENWIDTH div 200;
  for i := 0 to width - 1 do
    y[i] := y[i] * vy;

  result := 0;
end;

function wipe_doMelt(width, height, ticks: integer): integer;
var
  i: integer;
  j: integer;
  dy: fixed_t;
  idx: integer;
  s: PByteArray;
  d: PByteArray;
begin
  result := 1;

  while ticks > 0 do
  begin
    for i := 0 to width - 1 do
    begin
      if y[i] < 0 then
      begin
        y[i] := y[i] + vy;
        result := 0;
      end
      else if y[i] < height * FRACUNIT then
      begin
        if y[i] <= 15 * vy then
          dy := y[i] + vy
        else
          dy := 8 * vy;
        if (y[i] + dy) div FRACUNIT >= height then
          dy := height * FRACUNIT - y[i];
        s := PByteArray(integer(wipe_scr_end) + i * height + y[i] div FRACUNIT);
        d := PByteArray(integer(wipe_scr) + y[i] div FRACUNIT * width + i);
        idx := 0;
        for j := 0 to dy div FRACUNIT do //- 1 do
        begin
          d[idx] := s[j];
          idx := idx + width;
        end;
        y[i] := y[i] + dy;
        s := PByteArray(integer(wipe_scr_start) + i * height);
        d := PByteArray(integer(wipe_scr) + y[i] div FRACUNIT * width + i);

        idx := 0;
        for j := 0 to height - y[i] div FRACUNIT - 1 do
        begin
          d[idx] := s[j];
          idx := idx + width;
        end;
        result := 0;
      end;
    end;
    dec(ticks);
  end;
end;

function wipe_exitMelt(width, height, ticks: integer): integer;
begin
  Z_Free(y);
  result := 0;
end;

function wipe_StartScreen(x, y, width, height: integer): integer;
begin
  wipe_scr_start := screens[2];
  I_ReadScreen(wipe_scr_start);
  result := 0;
end;

function wipe_EndScreen(x, y, width, height: integer): integer;
begin
  wipe_scr_end := screens[3];
  I_ReadScreen(wipe_scr_end);
  V_DrawBlock(x, y, 0, width, height, wipe_scr_start); // restore start scr.
  result := 0;
end;

function wipe_ScreenWipe(wipeno, x, y, width, height, ticks: integer): boolean;
var
  rc: integer;

  function WIPES(index: integer): integer;
  begin
    case index of
      0: result := wipe_initColorXForm(width, height, ticks);
      1: result := wipe_doColorXForm(width, height, ticks);
      2: result := wipe_exitColorXForm(width, height, ticks);
      3: result := wipe_initMelt(width, height, ticks);
      4: result := wipe_doMelt(width, height, ticks);
      5: result := wipe_exitMelt(width, height, ticks);
    else
      result := 0; // Ouch
    end;
  end;

begin
  // initial stuff
  if not go then
  begin
    go := true;
  // wipe_scr = (byte *) Z_Malloc(width*height, PU_STATIC, 0); // DEBUG
    wipe_scr := screens[_FG];

    WIPES(wipeno * 3)
  end;

  // do a piece of wipe-in
  V_MarkRect(0, 0, width, height, false);
  rc := WIPES(wipeno * 3 + 1);
    //  V_DrawBlock(x, y, 0, width, height, wipe_scr); // DEBUG

  // final stuff
  if rc <> 0 then
  begin
    go := false;
    WIPES(wipeno * 3 + 2);
  end;

  result := not go;
end;


(* Unit d_textur implementation *)



(* Unit i_music implementation *)


                                                                                                                                                             

type
  music_t = (m_none, m_mus, m_midi);

const
  MAX_MIDI_EVENTS = 512;
  MUSMAGIC = $1A53554D; //"MUS"<EOF>

var
  hMidiStream: HMIDISTRM = 0;
  MidiDevice: LongWord;
  midicaps: MIDIOUTCAPS;
  m_type: music_t = m_none;
  MidiFileName: string;

type
  MidiEvent_t = packed record
    time: LongWord;                  { Ticks since last event }
    ID: LongWord;                    { Reserved, must be zero }
    case integer of
      1: (data: packed array[0..2] of byte;
         _type: byte);
      2: (mevent: LongWord);
  end;
  PMidiEvent_t = ^MidiEvent_t;
  MidiEvent_tArray = array[0..$FFFF] of MidiEvent_t;
  PMidiEvent_tArray = ^MidiEvent_tArray;

  Pmidiheader_t = ^midiheader_t;
  midiheader_t = record
    lpData: pointer;             { pointer to locked data block }
    dwBufferLength: LongWord;    { length of data in data block }
    dwBytesRecorded: LongWord;   { used for input only }
    dwUser: LongWord;            { for client's use }
    dwFlags: LongWord;           { assorted flags (see defines) }
    lpNext: Pmidiheader_t;       { reserved for driver }
    reserved: LongWord;          { reserved for driver }
    dwOffset: LongWord;          { Callback offset into buffer }
    dwReserved: array[0..7] of LongWord; { Reserved for MMSYSTEM }
  end;

  musheader_t = packed record
    ID: LongWord;       // identifier "MUS" 0x1A
    scoreLen: word;
    scoreStart: word;
    channels: word;     // count of primary channels
    sec_channels: word; // count of secondary channels
    instrCnt: word;
    dummy: word;
  end;
  Pmusheader_t = ^musheader_t;

const
  NUMMIDIHEADERS = 2;

type
  songinfo_t = record
    numevents: integer;
    nextevent: integer;
    midievents: PMidiEvent_tArray;
    header: array[0..NUMMIDIHEADERS - 1] of midiheader_t;
  end;
  Psonginfo_t = ^songinfo_t;

const
  MidiControlers: packed array[0..9] of byte =
    (0, 0, 1, 7, 10, 11, 91, 93, 64, 67);

var
  started: boolean = false;
  CurrentSong: Psonginfo_t = nil;
  loopsong: boolean = false;

function XLateMUSControl(control: byte): byte;
begin
  case control of
    10: result := 120;
    11: result := 123;
    12: result := 126;
    13: result := 127;
    14: result := 121;
  else
    begin
      I_Error('XLateMUSControl(): Unknown control %d', [control]);
      result := 0;
    end;
  end;
end;

const
  NUMTEMPOEVENTS = 2;

function GetSongLength(data: PByteArray): integer;
var
  done: boolean;
  events: integer;
  header: Pmusheader_t;
  time: boolean;
  i: integer;
begin
  header := Pmusheader_t(data);
  i := header.scoreStart;
  events := 0;
  done := header.ID <> MUSMAGIC;
  time := false;
  while not done do
  begin
    if (data[i] and $80) <> 0 then
      time := true;
    inc(i);
    case _SHR(data[i - 1], 4) and 7 of
      1:
        begin
          if (data[i] and $80) <> 0 then
            inc(i);
          inc(i);
        end;
      0,
      2,
      3: inc(i);
      4: inc(i, 2);
    else
      done := true;
    end;
    inc(events);
    if time then
    begin
      while (data[i] and $80) <> 0 do
        inc(i);
      inc(i);
      time := false;
    end;
  end;
  result := events + NUMTEMPOEVENTS;
end;

function I_MusToMidi(MusData: PByteArray; MidiEvents: PMidiEvent_tArray): boolean;
var
  header: Pmusheader_t;
  score: PByteArray;
  spos: integer;
  event: PMidiEvent_tArray;
  channel: byte;
  etype: integer;
  delta: integer;
  finished: boolean;
  channelvol: array[0..15] of byte;
  count: integer;
  i: integer;
begin
  header := Pmusheader_t(MusData);
  result := header.ID = MUSMAGIC;
  if not result then
  begin
    printf('I_MusToMidi(): Not a MUS file' + #13#10);
    exit;
  end;

  count := GetSongLength(MusData);
  score := PByteArray(@MusData[header.scoreStart]);
  event := MidiEvents;

  i := 0;
  while i < NUMTEMPOEVENTS do
  begin
    event[i].time := 0;
    event[i].ID := 0;
    event[i]._type := MEVT_TEMPO;
    event[i].data[0] := $00;
    event[i].data[1] := $80; //not sure how to work this out, should be 140bpm
    event[i].data[2] := $02; //but it's guessed so it sounds about right
    inc(i);
  end;

  delta := 0;
  spos := 0;
  ZeroMemory(channelvol, SizeOf(channelvol));

  finished := false;
  while true do
  begin
    event[i].time := delta;
    delta := 0;
    event[i].ID := 0;
    etype := _SHR(score[spos], 4) and 7;
    event[i]._type := MEVT_SHORTMSG;
    channel := score[spos] and 15;
    if channel = 9 then
      channel := 15
    else if channel = 15 then
      channel := 9;
    if (score[spos] and $80) <> 0 then
      delta := -1;
		inc(spos);
    case etype of
      0:
        begin
          event[i].data[0] := channel or $80;
          event[i].data[1] := score[spos];
          inc(spos);
          event[i].data[2] := channelvol[channel];
        end;
      1:
        begin
          event[i].data[0] := channel or $90;
          event[i].data[1] := score[spos] and 127;
          if (score[spos] and 128) <> 0 then
          begin
            inc(spos);
            channelvol[channel] := score[spos];
          end;
          inc(spos);
          event[i].data[2] := channelvol[channel];
        end;
      2:
        begin
          event[i].data[0] := channel or $e0;
          event[i].data[1] := _SHL(score[spos], 7) and $7f;
          event[i].data[2] := _SHR(score[spos], 7);
          inc(spos);
        end;
      3:
        begin
          event[i].data[0] := channel or $b0;
          event[i].data[1] := XLateMUSControl(score[spos]);
          inc(spos);
          event[i].data[2] := 0;
        end;
      4:
        begin
          if score[spos] <> 0 then
          begin
            event[i].data[0] := channel or $b0;
            event[i].data[1] := MidiControlers[score[spos]];
            inc(spos);
            event[i].data[2] := score[spos];
            inc(spos);
          end
          else
          begin
            event[i].data[0] := channel or $c0;
            inc(spos);
            event[i].data[1] := score[spos];
            inc(spos);
            event[i].data[2] := 64;
          end;
        end;
    else
      finished := true;
    end;
    if finished then
      break;
    inc(i);
    dec(count);
    if count < 3 then
      I_Error('I_MusToMidi(): Overflow');
    if delta = -1 then
    begin
      delta := 0;
      while (score[spos] and 128) <> 0 do
      begin
        delta := _SHL(delta, 7);
        delta := delta + score[spos] and 127;
        inc(spos);
      end;
      delta := delta + score[spos];
      inc(spos);
    end;
  end;
end;

//
// MUSIC API.
//
procedure I_InitMus;
var
  rc: MMRESULT;
  numdev: LongWord;
  i: integer;
begin
  if M_CheckParm('-nomusic') <> 0 then
    exit;

  if hMidiStream <> 0 then
    exit;

  ZeroMemory(midicaps, SizeOf(midicaps));
  MidiDevice := MIDI_MAPPER;

  // First try midi mapper
  rc := midiOutGetDevCaps(MidiDevice, @midicaps, SizeOf(midicaps));
  if rc <> MMSYSERR_NOERROR then
    I_Error('I_InitMusic(): midiOutGetDevCaps failed, return value = %d', [rc]);

  // midiStreamOut not supported (should not happen with MIDI MAPPER...)
  // Try to enumurate all midi devices
  if (midicaps.dwSupport and MIDICAPS_STREAM) = 0 then
  begin
    numdev := midiOutGetNumDevs;
    if numdev = 0 then // fatal
      exit;

    for i := -1 to numdev - 1 do
    begin
      rc := midiOutGetDevCaps(i, @midicaps, SizeOf(midicaps));
      if rc <> MMSYSERR_NOERROR then
        I_Error('I_InitMusic(): midiOutGetDevCaps failed, return value = %d', [rc]);

      if (midicaps.dwSupport and MIDICAPS_STREAM) <> 0 then
      begin
        MidiDevice := i;
        break;
      end;
    end;
  end;

  if MidiDevice = MIDI_MAPPER then
    printf('Using midi mapper' + #13#10)
  else
    printf('Using midi device %d' + #13#10, [MidiDevice]);

  rc := midiStreamOpen(@hMidiStream, @MidiDevice, 1, 0, 0, CALLBACK_NULL);
  if rc <> MMSYSERR_NOERROR then
  begin
    hMidiStream := 0;
    printf('I_InitMusic(): midiStreamOpen failed, result = %d' + #13#10, [rc]);
  end;
  started := false;
end;

procedure I_InitMusic;
begin
  I_InitMus;
end;


//
// I_StopMusic
//
procedure I_StopMusicMus(song: Psonginfo_t);
var
  i: integer;
  rc: MMRESULT;
begin
  if not (boolval(song) and (hMidiStream <> 0)) then
    exit;

  loopsong := false;
  rc := midiOutReset(HMIDIOUT(hMidiStream));
  if rc <> MMSYSERR_NOERROR then
    printf('I_StopMusic(): midiOutReset failed, result = %d' + #13#10, [rc]);

  started := false;

  for i := 0 to NUMMIDIHEADERS - 1 do
  begin
    if song.header[i].lpData <> nil then
    begin
      rc := midiOutUnprepareHeader(HMIDIOUT(hMidiStream), @song.header[i], SizeOf(midiheader_t));
      if rc <> MMSYSERR_NOERROR then
        printf('I_StopMusic(): midiOutUnprepareHeader failed, result = %d' + #13#10, [rc]);

      song.header[i].lpData := nil;
      song.header[i].dwFlags := MHDR_DONE or MHDR_ISSTRM;
    end;
  end;
  song.nextevent := 0;

{  rc := midiOutClose(HMIDIOUT(hMidiStream));
  if rc <> MMSYSERR_NOERROR then
    printf('I_StopMusic(): midiOutReset failed, result = %d' + #13#10, [rc]);

  hMidiStream := 0;}
end;

procedure I_StopMusic(song: Psonginfo_t);
begin
  case m_type of
    m_midi: I_StopMidi;
    m_mus: I_StopMusicMus(song);
  end;
end;

procedure I_StopMus;
var
  rc: MMRESULT;
begin
	if hMidiStream <> 0 then
  begin
    rc := midiStreamStop(hMidiStream);
    if rc <> MMSYSERR_NOERROR then
      printf('I_ShutdownMusic(): midiStreamStop failed, result = %d' + #13#10, [rc]);

    started := false;
    rc := midiStreamClose(hMidiStream);
    if rc <> MMSYSERR_NOERROR then
      printf('I_ShutdownMusic(): midiStreamClose failed, result = %d' + #13#10, [rc]);

    hMidiStream := 0;
  end;
end;

//
// I_ShutdownMusic
//
procedure I_ShutdownMusic;
begin
  I_StopMus;
  I_StopMidi;
  fdelete(MidiFileName);
end;

//
// I_PlaySong
//
procedure I_PlaySong(handle: integer; looping: boolean);
begin
	if (handle <> 0) and (hMidiStream <> 0) then
  begin
    loopsong := looping;
    CurrentSong := Psonginfo_t(handle);
  end;
end;

//
// I_PauseSong
//
procedure I_PauseSongMus(handle: integer);
var
  rc: MMRESULT;
begin
  if hMidiStream = 0 then
    exit;

  rc := midiStreamPause(hMidiStream);
  if rc <> MMSYSERR_NOERROR then
    I_Error('I_PauseSong(): midiStreamRestart failed, return value = %d', [rc]);
end;

procedure I_PauseSong(handle: integer);
begin
  case m_type of
    m_midi: I_PauseMidi;
    m_mus: I_PauseSongMus(handle);
  end;
end;

//
// I_ResumeSong
//
procedure I_ResumeSongMus(handle: integer);
var
  rc: MMRESULT;
begin
  if hMidiStream = 0 then
    exit;

  rc := midiStreamRestart(hMidiStream);
  if rc <> MMSYSERR_NOERROR then
    I_Error('I_ResumeSong(): midiStreamRestart failed, return value = %d', [rc]);
end;

procedure I_ResumeSong(handle: integer);
begin
  case m_type of
    m_midi: I_ResumeMidi;
    m_mus: I_ResumeSongMus(handle);
  end;
end;

// Stops a song over 3 seconds.
procedure I_StopSong(handle: integer);
var
  song: Psonginfo_t;
begin
  if (handle <> 0) and (hMidiStream <> 0) then
  begin

    song := Psonginfo_t(handle);

    I_StopMusic(song);

    if song = CurrentSong then
      CurrentSong := nil;
  end;
end;

procedure I_UnRegisterSong(handle: integer);
var
  song: Psonginfo_t;
begin
  if (handle <> 0) and (hMidiStream <> 0) then
  begin

    I_StopSong(handle);

    song := Psonginfo_t(handle);
    Z_Free(song.midievents);
    Z_Free(song);
  end;
end;

function I_RegisterSong(data: pointer; size: integer): integer;
var
  song: Psonginfo_t;
  i: integer;
  f: file;
begin
//  song := Psonginfo_t(malloc(SizeOf(songinfo_t)));
  song := Z_Malloc(SizeOf(songinfo_t), PU_STATIC, nil);
  song.numevents := GetSongLength(PByteArray(data));
  song.nextevent := 0;
  song.midievents := Z_Malloc(song.numevents * SizeOf(MidiEvent_t), PU_STATIC, nil);

  if m_type = m_midi then
    I_StopMidi;

  if I_MusToMidi(PByteArray(data), song.midievents) then
  begin
{    if m_type <> m_mus then
    begin}
      I_InitMus;
      m_type := m_mus;
//    end;

    if hMidiStream = 0 then
    begin
      printf('I_RegisterSong(): Could not initialize midi stream' + #13#10);
      m_type := m_none;
      result := 0;
      exit;
    end;

    for i := 0 to NUMMIDIHEADERS - 1 do
    begin
      song.header[i].lpData := nil;
      song.header[i].dwFlags := MHDR_ISSTRM or MHDR_DONE;
    end;
  end
  else
  begin
    if m_type <> m_midi then
    begin
      I_StopMus;
      m_type := m_midi;
    end;

    if M_CheckParmCDROM then
      MidiFileName := 'c:\doomdata\doom32.mid'
    else
      MidiFileName := 'doom32.mid';

    assign(f, MidiFileName);
    {$I-}
    rewrite(f, 1);
    BlockWrite(f, data^, size);
    closefile(f);
    {$I+}
    if IOResult <> 0 then
    begin
      printf('I_RegisterSong(): Could not initialize MCI' + #13#10);
      m_type := m_none;
      result := 0;
      exit;
    end;
    ClearMidiFilePlayList;
    AddMidiFileToPlayList(MidiFileName);
    I_PlayMidi(0);
  end;
  result := integer(song);
end;

// Is the song playing?
function I_QrySongPlaying(handle: integer): boolean;
begin
  result := CurrentSong <> nil;
end;

//
// I_SetMusicVolume
//
procedure I_SetMusicVolumeMus(volume: integer);
var
  rc: MMRESULT;
begin
  snd_MusicVolume := volume;
// Now set volume on output device.
// Whatever( snd_MusciVolume );
  if (CurrentSong <> nil) and (snd_MusicVolume = 0) and started then
    I_StopMusic(CurrentSong);

  if (midicaps.dwSupport and MIDICAPS_VOLUME) <> 0 then
  begin
    rc := midiOutSetVolume(hMidiStream,
      _SHLW($FFFF * snd_MusicVolume div 16, 16) or _SHLW(($FFFF * snd_MusicVolume div 16), 0));
    if rc <> MMSYSERR_NOERROR then
      printf('I_SetMusicVolume(): midiOutSetVolume failed, return value = %d' + #13#10, [rc]);
  end
  else
    printf('I_SetMusicVolume(): Midi device dos not support volume control' + #13#10);
end;

procedure I_SetMusicVolume(volume: integer);
begin
  case m_type of
    m_midi: ; // unsupported
    m_mus: I_SetMusicVolumeMus(volume);
  end;
end;

procedure I_ProcessMusic;
var
  header: Pmidiheader_t;
  length: integer;
  i: integer;
  rc: MMRESULT;
begin
  if m_type <> m_mus then
    exit;

  if (snd_MusicVolume = 0) or (CurrentSong = nil) then
    exit;

  for i := 0 to NUMMIDIHEADERS - 1 do
  begin
    header := @CurrentSong.header[i];
    if boolval(header.dwFlags and MHDR_DONE) then
    begin
      if header.lpData <> nil then
      begin
        rc := midiOutUnprepareHeader(HMIDIOUT(hMidiStream), PMidiHdr(header), SizeOf(midiheader_t));
        if rc <> MMSYSERR_NOERROR then
          printf('I_ProcessMusic(): midiOutUnprepareHeader failed, result = %d' + #13#10, [rc]);
      end;
      header.lpData := @CurrentSong.midievents[CurrentSong.nextevent];
      length := CurrentSong.numevents - CurrentSong.nextevent;
      if length > MAX_MIDI_EVENTS then
      begin
        length := MAX_MIDI_EVENTS;
        CurrentSong.nextevent := CurrentSong.nextevent + MAX_MIDI_EVENTS;
      end
      else
        CurrentSong.nextevent := 0;
      length := length * SizeOf(MidiEvent_t);
      header.dwBufferLength := length;
      header.dwBytesRecorded := length;
      header.dwFlags := MHDR_ISSTRM;
      rc := midiOutPrepareHeader(HMIDIOUT(hMidiStream), PMidiHdr(header), SizeOf(midiheader_t));
      if rc <> MMSYSERR_NOERROR then
        I_Error('I_ProcessMusic(): midiOutPrepareHeader failed, return value = %d', [rc]);
      if not started then
      begin
        rc := midiStreamRestart(hMidiStream);
        if rc <> MMSYSERR_NOERROR then
          I_Error('I_ProcessMusic(): midiStreamRestart failed, return value = %d', [rc]);
        started := true;
      end;
      rc := midiStreamOut(hMidiStream, PMidiHdr(header), SizeOf(midiheader_t));
      if rc <> MMSYSERR_NOERROR then
        I_Error('I_ProcessMusic(): midiStreamOut failed, return value = %d', [rc]);
    end;
  end;
end;


(* Unit m_fixed implementation *)


                                      

function FixedMul(a: fixed_t; b: fixed_t): fixed_t;
var
  c: extended;
begin
//  result := (a * b) div FRACUNIT;
  c := a / FRACUNIT;
  result := trunc(c * b);
end;

function FixedDiv(a: fixed_t; b: fixed_t): fixed_t;
begin
  if _SHR(abs(a), 14) >= abs(b) then
  begin
    if a xor b < 0 then
      result := MININT
    else
      result := MAXINT;
  end
  else
    result := FixedDiv2(a, b);
end;

function FixedDiv2(a: fixed_t; b: fixed_t): fixed_t;
var
  c: extended;
begin
  c := (a / b) * FRACUNIT;

  if (c >= 2147483648.0) or (c < -2147483648.0) then
    I_Error('FixedDiv(): divide by zero' + #13#10);

  result := trunc(c);
end;


(* Unit m_bbox implementation *)


               

procedure M_ClearBox(box: Pfixed_tArray);
begin
  box[BOXTOP] := MININT;
  box[BOXRIGHT] := MININT;
  box[BOXBOTTOM] := MAXINT;
  box[BOXLEFT] := MAXINT;
end;

procedure M_AddToBox(box: Pfixed_tArray; x: fixed_t; y: fixed_t);
begin
  if x < box[BOXLEFT] then
    box[BOXLEFT] := x
  else if x > box[BOXRIGHT] then
    box[BOXRIGHT] := x;
  if y < box[BOXBOTTOM] then
    box[BOXBOTTOM] := y
  else if y > box[BOXTOP] then
    box[BOXTOP] := y;
end;


(* Unit i_midi implementation *)


                                                                    

var
  wDeviceID: DWORD;
  MidiFileNames: TStringList;
  MidiID: integer;
  Window: HWnd;
  fIsPlaying: boolean;
  WindowClass: TWndClass;

const
  rsAppName = 'MIDIPLAYERWNDNOTIFY';
  rsSequencer = 'sequencer';
  rsWndTitle = 'Notify Window';
  rsErrNoMIDIMapper = 'MIDI mapper unavailable';
  rsErrOutOfRange = 'I_PlayMidi(index): index out of playlist range.';

// Plays a specified MIDI file by using MCI_OPEN and MCI_PLAY. Returns
// as soon as playback begins. The window procedure function for the
// specified window will be notified when playback is complete.
// Returns 0L on success; otherwise, it returns an MCI error code.
function playMIDIFile(hWndNotify: HWnd; lpszMIDIFileName: string; doCheckMidiMapper: boolean = false): DWORD;
var mciOpenParms: MCI_OPEN_PARMS;
    mciPlayParms: MCI_PLAY_PARMS;
    mciStatusParms: MCI_STATUS_PARMS;
begin
  // Open the device by specifying the device and filename.
  // MCI will attempt to choose the MIDI mapper as the output port.
  FillChar(mciOpenParms, SizeOf(mciOpenParms), Chr(0));
  mciOpenParms.lpstrDeviceType := PAnsiChar(rssequencer);
  mciOpenParms.lpstrElementName := PChar(lpszMIDIFileName);
  result := mciSendCommand(0, MCI_OPEN, MCI_OPEN_TYPE or MCI_OPEN_ELEMENT, DWORD(@mciOpenParms));
  // Failed to open device. Don't close it; just return error.
  if result <> 0 then exit;

  // The device opened successfully; get the device ID.
  wDeviceID := mciOpenParms.wDeviceID;

  if doCheckMidiMapper then
  begin
    // Check if the output port is the MIDI mapper.
    mciStatusParms.dwItem := MCI_SEQ_STATUS_PORT;
    result := mciSendCommand(wDeviceID, MCI_STATUS, MCI_STATUS_ITEM, DWORD(@mciStatusParms));
    if result <> 0 then
    begin
      mciSendCommand(wDeviceID, MCI_CLOSE, 0, 0);
      exit;
    end
    else if LOWORD(mciStatusParms.dwReturn) <> WORD(MIDI_MAPPER) then
    // The output port is not the MIDI mapper.
    begin
      printf(rsErrNoMIDIMapper);
      exit;
    end;
  end;

  // Begin playback. The window procedure function for the parent
  // window will be notified with an MM_MCINOTIFY message when
  // playback is complete. At this time, the window procedure closes
  // the device.
  FillChar(mciPlayParms, SizeOf(mciPlayParms), Chr(0));
  mciPlayParms.dwCallback := DWORD(hWndNotify);
  result := mciSendCommand(wDeviceID, MCI_PLAY, MCI_NOTIFY, DWORD(@mciPlayParms));
  if result > 0 then
    mciSendCommand(wDeviceID, MCI_CLOSE, 0, 0);
end;

procedure StopPlaying;
begin
  mciSendCommand(wDeviceID, MCI_STOP, 0, 0);
  mciSendCommand(wDeviceID, MCI_CLOSE, 0, 0);
end;

function i_midi_WindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT; stdcall; export;
begin
  case Msg of
    MM_MCINOTIFY:
      if (wParam = MCI_NOTIFY_SUCCESSFUL) then
      begin
        StopPlaying;
        if MidiID >= MidiFileNames.Count - 1 then
          MidiID := 0
        else
          inc(MidiID);
        if MidiID < MidiFileNames.Count then
          playMIDIFile(Window, MidiFileNames.Strings[MidiID]);
      end;
    WM_CLOSE:
      begin
        DestroyWindow(hWnd);
        Window := 0;
        result := 0;
        exit;
      end;
  end;
  result := DefWindowProc(hWnd, Msg, WParam, LParam);
end;

procedure AddMidiFileToPlayList(MidiFile: string);
begin
  if MidiFileNames = nil then
    MidiFileNames := TStringList.Create;
  MidiFileNames.Add(MidiFile);
end;

procedure DeleteMidiFileFromPlayList(MidiFile: string);
var i: integer;
begin
  if MidiFileNames <> nil then
  begin
    if MidiFileNames.IndexOf(MidiFile) > -1 then
      MidiFileNames.Delete(MidiFileNames.IndexOf(MidiFile))
    else
    begin
      for i := 0 to MidiFileNames.Count - 1 do
        if strupper(MidiFileNames.Strings[i]) = strupper(MidiFile) then
        begin
          MidiFileNames.Delete(i);
          exit;
        end;
    end;
  end;
end;

procedure ClearMidiFilePlayList;
begin
  if MidiFileNames <> nil then
    MidiFileNames.Clear;
end;

procedure I_PlayMidi(index: integer);
begin
  I_StopMidi;
  FillChar(WindowClass, SizeOf(WindowClass), Chr(0));
  if Window = 0 then
  begin
    WindowClass.style := CS_DBLCLKS;
    WindowClass.lpfnWndProc := @i_midi_WindowProc;
    WindowClass.lpszClassName := PChar(rsAppName);
    if HPrevInst = 0 then
    begin
      WindowClass.hInstance := HInstance;
      WindowClass.hCursor := LoadCursor(0, idc_Arrow);
      RegisterClass(WindowClass);
    end;
    Window := CreateWindowEx(
      0,
      WindowClass.lpszClassName,
      PChar(rsWndTitle),
      ws_OverlappedWindow,
      integer(CW_USEDEFAULT),
      integer(CW_USEDEFAULT),
      integer(CW_USEDEFAULT),
      integer(CW_USEDEFAULT),
      0,
      0,
      HInstance,
      nil);
    ShowWindow(Window, SW_HIDE);
  end;

  if (index < MidiFileNames.Count) and (index > -1) then
  begin
    MidiID := index;
    playMIDIFile(Window, MidiFileNames.Strings[MidiID], true);
    fIsPlaying := true;
  end
  else
    printf(rsErrNoMIDIMapper);
end;

procedure I_StopMidi;
begin
  if Window <> 0 then
  begin
    StopPlaying;
    SendMessage(Window, WM_CLOSE, 0, 0);
    Window := 0;
    fIsPlaying := false;
  end;
end;

function I_IsMidiPlaying: boolean;
begin
  result := fIsPlaying;
end;

procedure I_ResumeMidi;
begin
  mciSendCommand(wDeviceID, MCI_RESUME, 0, 0);
end;

procedure I_PauseMidi;
begin
  mciSendCommand(wDeviceID, MCI_PAUSE, 0, 0);
end;

function _mciGetErrorString(const code: LongWord): string;
var
  buf: array[0..127] of char;
  i:  integer;
begin
  result := '';
  FillChar(buf, 128, Chr(0));
  if mciGetErrorString(code, buf, 128) then
    for i := 0 to 127 do
    begin
      if buf[i] = #0 then
        break;
      result := result + buf[i];
    end;
end;


(* Unit tables implementation *)


               

function SlopeDiv(const num: integer; const den: integer): integer;
// VJ was function SlopeDiv(num: LongWord; den: LongWord): integer;
var ans: LongWord;
begin
  if den < 512 then
    result := SLOPERANGE
  else
  begin
    ans := _SHL(num, 3) div _SHR(den, 8);
    if ans < SLOPERANGE then
      result := ans
    else
      result := SLOPERANGE;
  end;
end;


(* Unit d_delphi implementation *)


                        

procedure sprintf(var s: string; const Fmt: string; const Args: array of const);
begin
  s := Format(Fmt, Args);
end;

procedure printf(const str: string);
begin
  if Assigned(outproc) then
    outproc(str)
  else if IsConsole then
    write(str);
end;

procedure printf(const Fmt: string; const Args: array of const);
var s: string;
begin
  sprintf(s, Fmt, Args);
  printf(s);
end;

procedure fprintf(var f: file; const str: string);
begin
  BlockWrite(f, (@str[1])^, Length(str));
end;

procedure fprintf(var f: file; const Fmt: string; const Args: array of const);
var s: string;
begin
  sprintf(s, Fmt, Args);
  fprintf(f, s);
end;

procedure fprintf(var f: TFile; const str: string);
begin
  fprintf(f.f, str);
end;

procedure fprintf(var f: TFile; const Fmt: string; const Args: array of const);
begin
  fprintf(f.f, Fmt, Args);
end;

function itoa(i: integer): string;
begin
  result := IntToStr(i);
end;

function atoi(const s: string): integer;
begin
  result := StrToIntDef(s, 0);
end;

function StrToFloatDef(const s: string; def: single): single;
var
  code: integer;
begin
  val(s, result, code);
  if code <> 0 then
    result := def;
end;

function atof(const s: string): single;
var
  s2: string;
  i: Integer;
begin
  s2 := s;
  for i := 1 to length(s2) do
  begin
    if s2[i] in ['.', ','] then
      s2[i] := SysUtils.DecimalSeparator;
  end;
  result := StrToFloatDef(s2, 0.0);
end;

function memcpy(dst: pointer; const src: pointer; len: integer): pointer;
begin
  move(src^, dst^, len);
  result := dst;
end;

function memcpy(dst: pointer; const src: string; len: integer): pointer;
var i: integer;
begin
  if len >= Length(src) then
  begin
    for i := 1 to Length(src) do
      PByteArray(dst)[i] := ord(src[i]);
    for i := Length(src) + 1 to len do
      PByteArray(dst)[i] := 0;
  end
  else
    for i := 1 to len do
      PByteArray(dst)[i] := ord(src[i]);
  result := dst;
end;

function memmove(dst: pointer; const src: pointer; len: integer): pointer;
begin
  move(src^, dst^, len);
  result := dst;
end;

function memset(buf: pointer; c: integer; len: integer): pointer;
begin
  FillChar(buf^, len, c);
  Result := buf;
end;

function malloc(size: integer): Pointer;
begin
  GetMem(Result, size);
  ZeroMemory(Result^, size);
end;

procedure free(var p: pointer);
begin
  FreeMem(p);
  p := nil;
end;

function IntToStrZfill(const z: integer; const x: integer): string;
var
  i: integer;
  len: integer;
begin
  result := IntToStr(x);
  len := Length(result);
  for i := len + 1 to z do
    result := '0' + result;
end;

function boolval(const x: integer): boolean;
begin
  result := x <> 0;
end;

function boolval(const c: char): boolean;
begin
  result := c <> #0;
end;

function boolval(const p: pointer): boolean;
begin
  result := p <> nil;
end;

function intval(const b: boolean): integer;
begin
  if b then
    result := 1
  else
    result := 0;
end;

function decide(const contition: boolean;
  const iftrue: integer; const iffalse: integer): integer;
begin
  if contition then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: boolean;
  const iftrue: boolean; const iffalse: boolean): boolean;
begin
  if contition then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: boolean;
  const iftrue: string; const iffalse: string): string;
begin
  if contition then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: boolean;
  const iftrue: pointer; const iffalse: pointer): pointer;
begin
  if contition then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: integer;
  const iftrue: integer; const iffalse: integer): integer;
begin
  if contition <> 0 then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: integer;
  const iftrue: boolean; const iffalse: boolean): boolean;
begin
  if contition <> 0 then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: integer;
  const iftrue: string; const iffalse: string): string;
begin
  if contition <> 0 then
    result := iftrue
  else
    result := iffalse;
end;

function decide(const contition: integer;
  const iftrue: pointer; const iffalse: pointer): pointer;
begin
  if contition <> 0 then
    result := iftrue
  else
    result := iffalse;
end;

function incp(var p: pointer; const size: integer = 1): pointer;
begin
  result := Pointer(integer(p) + size);
  p := result;
end;

function pOperation(const p1, p2: pointer; const op: char; size: integer): integer;
begin
  case op of
    '+': result := (Integer(p1) + Integer(p2)) div size;
    '-': result := (Integer(p1) - Integer(p2)) div size;
  else
    result := 0;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
constructor TFile.Create(const FileName: string; mode: word);
begin
  assign(f, FileName);
  {$I-}
  if mode = fCreate then
    rewrite(f, 1)
  else
    reset(f, 1);
  {$I+}
end;

destructor TFile.Destroy;
begin
  closefile(f);
  Inherited;
end;

function TFile.Read(var Buffer; Count: Longint): Longint;
begin
  BlockRead(f, Buffer, Count, result);
end;

function TFile.Write(const Buffer; Count: Longint): Longint;
begin
  BlockWrite(f, Buffer, Count, result);
end;

function TFile.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    sFromBeginning:
      result := Offset;
    sFromCurrent:
      result := FilePos(f) + Offset;
    sFromEnd:
      result := FileSize(f) - Offset;
  else
    result := 0;
  end;
  system.Seek(f, result);
end;

function TFile.Size: Longint;
begin
  result := FileSize(f);
end;

constructor TCachedFile.Create(const FileName: string; mode: word; ABufSize: integer = $FFFF);
begin
  fInitialized := false;
  Inherited Create(FileName, mode);
  fBufSize := ABufSize;
  GetMem(fBuffer, fBufSize);
  fPosition := 0;
  ResetBuffer;
  fSize := Inherited Size;
  fInitialized := true;
end;

procedure TCachedFile.ResetBuffer;
begin
  fBufferStart := -1;
  fBufferEnd := -1;
end;

destructor TCachedFile.Destroy;
begin
  FreeMem(fBuffer, fBufSize);
  Inherited;
end;

function TCachedFile.Read(var Buffer; Count: Longint): Longint;
var x: Longint;
begin
// Buffer hit
  if (fPosition >= fBufferStart) and (fPosition + Count <= fBufferEnd) then
  begin
    x := LongInt(fBuffer) + fPosition - fBufferStart;
    Move(Pointer(x)^, Buffer, Count);
    fPosition := fPosition + Count;
    result := Count;
  end
// Non Buffer hit, cache buffer
  else if Count <= fBufSize then
  begin
    fPosition := Inherited Seek(fPosition, sFromBeginning);
    x := Inherited Read(fBuffer^, fBufSize);
    if x < Count then
      result := x
    else
      result := Count;
    Move(fBuffer^, Buffer, Count);
    fBufferStart := fPosition;
    fBufferEnd := fPosition + x;
    fPosition := fPosition + result;
  end
// Keep old buffer
  else
  begin
    fPosition := Inherited Seek(fPosition, sFromBeginning);
    result := Inherited Read(Buffer, Count);
    fPosition := fPosition + result;
  end;
end;

function TCachedFile.Write(const Buffer; Count: Longint): Longint;
begin
  fPosition := Inherited Seek(fPosition, sFromBeginning);
  result := Inherited Write(Buffer, Count);
  fPosition := fPosition + result;
  if fSize < fPosition then
    fSize := fPosition;
end;

function TCachedFile.Seek(Offset: Longint; Origin: Word): Longint;
begin
  if fInitialized then
  begin
    case Origin of
      sFromBeginning: fPosition := Offset;
      sFromCurrent: Inc(fPosition, Offset);
      sFromEnd: fPosition := fSize + Offset;
    end;
    result := fPosition;
  end
  else
    result := Inherited Seek(Offset, Origin);
end;

procedure TCachedFile.SetSize(NewSize: Longint);
begin
  Inherited;
  fSize := NewSize;
end;

function getenv(const env: string): string;
var
  buf: array[0..255] of char;
begin
  ZeroMemory(buf, SizeOf(buf));
  GetEnvironmentVariable(PChar(env), buf, 255);
  result := Trim(StringVal(buf));
end;

function fexists(const filename: string): boolean;
begin
  result := FileExists(filename);
end;

procedure fdelete(const filename: string);
begin
  if fexists(filename) then
    DeleteFile(filename);
end;

function fext(const filename: string): string;
begin
  result := ExtractFileExt(filename);
end;

function tan(const x: single): single;
var
  a: single;
  b: single;
begin
  b := cos(x);
  if b <> 0 then
  begin
    a := sin(x);
    result := a / b;
  end
  else
    result := 0.0;
end;


function strupper(const S: string): string;
var
  Ch: Char;
  L: Integer;
  Source, Dest: PChar;
begin
  L := Length(S);
  SetLength(Result, L);
  Source := Pointer(S);
  Dest := Pointer(Result);
  while L <> 0 do
  begin
    Ch := Source^;
    if (Ch >= 'a') and (Ch <= 'z') then Dec(Ch, 32);
    Dest^ := Ch;
    Inc(Source);
    Inc(Dest);
    Dec(L);
  end;
end;


function toupper(ch: Char): Char;
asm
{ ->    AL      Character       }
{ <-    AL      Result          }

        CMP     AL,'a'
        JB      @@exit
        CMP     AL,'z'
        JA      @@exit
        SUB     AL,'a' - 'A'
@@exit:
end;

function tolower(ch: Char): Char;
asm
{ ->    AL      Character       }
{ <-    AL      Result          }

        CMP     AL,'A'
        JB      @@exit
        CMP     AL,'Z'
        JA      @@exit
        SUB     AL,'A' - 'a'
@@exit:
end;

function _SHL(const x: integer; const bits: integer): integer;
begin
  result := x * (1 shl bits);
end;

function _SHLW(const x: LongWord; const bits: LongWord): LongWord;
begin
//  result := x * (1 shl bits);
  result := x shl bits;
end;

function _SHR(const x: integer; const bits: integer): integer;
begin
  result := x div (1 shl bits);
end;

function _SHRW(const x: LongWord; const bits: LongWord): LongWord;
begin
//  result := x div (1 shl bits);
  result := x shr bits;
end;

function StringVal(const Str: PChar): string;
begin
  Result := Str;
end;

procedure ZeroMemory(var X; Count: Integer);
begin
  FillChar(X, Count, Chr(0));
end;


(* Unit w_wad implementation *)


                                       

const
  IWAD = integer(Ord('I') or
                (Ord('W') shl 8) or
                (Ord('A') shl 16) or
                (Ord('D') shl 24));

  PWAD = integer(Ord('P') or
                (Ord('W') shl 8) or
                (Ord('A') shl 16) or
                (Ord('D') shl 24));

//
// GLOBALS
//

// Location of each lump on disk.
var
  numlumps: integer;

  lumpcache: PPointerArray;

function char8tostring(src: char8_t): string;
var
  i: integer;
begin
  result := '';
  i := 0;
  while (i < 8) and (src[i] <> #0) do
  begin
    result := result + src[i];
    inc(i);
  end;
end;

function stringtochar8(src: string): char8_t;
var
  i: integer;
  len: integer;
begin
  len := length(src);
  if len > 8 then
    I_Error('stringtochar8(): length of %s is > 8', [src]);

  i := 1;
  while (i <= len) do
  begin
    result[i - 1] := src[i];
    inc(i);
  end;

  for i := len to 7 do
    result[i] := #0;
end;

function filelength(handle: TCachedFile): integer;
begin
  try
    result := handle.Size;
  except
    result := 0;
    I_Error('filelength(): Error fstating');
  end;
end;

procedure ExtractFileBase(const path: string; var dest: string);
var
  i: integer;
  len: integer;
begin
  len := Length(path);
  i := len;
  while i > 0 do
  begin
    if path[i] in ['/', '\'] then
      break;
    dec(i);
  end;
  dest := '';
  while (i < len) do
  begin
    inc(i);
    if path[i] = '.' then
      break
    else
      dest := dest + toupper(path[i]);
  end;
  if Length(dest) > 8 then
    I_Error('ExtractFileBase(): Filename base of %s >8 chars', [path]);
end;

procedure ExtractFileBase8(const path: string; var dest: char8_t);
var
  dst: string;
begin
  dst := char8tostring(dest);
  ExtractFileBase(path, dst);
  dest := stringtochar8(dst);
end;

//
// LUMP BASED ROUTINES.
//

//
// W_AddFile
// All files are optional, but at least one file must be
//  found (PWAD, if all required lumps are present).
// Files with a .wad extension are wadlink files
//  with multiple lumps.
// Other files are single lumps with the base filename
//  for the lump name.
//
// If filename starts with a tilde, the file is handled
//  specially to allow map reloads.
// But: the reload feature is a fragile hack...
var
  reloadlump: integer;
  reloadname: string;

procedure W_AddFile(var filename: string);
var
  header: wadinfo_t;
  lump_p: Plumpinfo_t;
  i: integer;
  j: integer;
  handle: TCachedFile;
  len: integer;
  startlump: integer;
  fileinfo: Pfilelump_tArray;
  singleinfo: filelump_t;
  storehandle: TCachedFile;
  ext: string;
  c: char;
begin
  // open the file and add to directory
  // handle reload indicator.
  if filename[1] = '~' then
  begin
    reloadname := Copy(filename, 2, Length(filename) - 1);
    Delete(filename, 1, 1);
    reloadlump := numlumps;
  end
  else
    reloadname := '';

  if not fexists(filename) then
  begin
    printf('W_AddFile(): File %s does not exist' + #13#10, [filename]);
    exit;
  end;

  try
    handle := TCachedFile.Create(filename, fOpen);
  except
    printf('W_AddFile(): couldn''t open %s' + #13#10, [filename]);
    exit;
  end;

  printf(' adding %s' + #13#10, [filename]);
  startlump := numlumps;

  ext := strupper(fext(filename));
  if ext <> '.WAD' then
  begin
    // single lump file
    fileinfo := @singleinfo;
    singleinfo.filepos := 0;
    singleinfo.size := filelength(handle);
    ExtractFileBase8(filename, singleinfo.name);
    inc(numlumps);
  end
  else
  begin
    // WAD file
	  handle.Read(header, SizeOf(header));
    if header.identification <> IWAD then
      // Homebrew levels?
	    if header.identification <> PWAD then
        I_Error('W_AddFile(): Wad file %s doesn''t have IWAD or PWAD id' + #13#10, [filename]);

	    // ???modifiedgame = true;
//    header.numlumps := header.numlumps;
//    header.infotableofs = header.infotableofs;
    len := header.numlumps * SizeOf(filelump_t);
    GetMem(fileinfo, len);
    handle.Seek(header.infotableofs, sFromBeginning);
    handle.Read(fileinfo^, len);
    numlumps := numlumps + header.numlumps;
  end;


  // Fill in lumpinfo
  ReAllocMem(lumpinfo, numlumps * SizeOf(lumpinfo_t));

  if not boolval(lumpinfo) then
    I_Error('W_AddFile(): Couldn''t realloc lumpinfo');

  if reloadname <> '' then
    storehandle := nil
  else
    storehandle := handle;

  for i := startlump to numlumps - 1 do
  begin
    lump_p := @lumpinfo[i];
    lump_p.handle := storehandle;
    lump_p.position := fileinfo[i - startlump].filepos;
    lump_p.size := fileinfo[i - startlump].size;
    c := #255;
    for j := 0 to 7 do
    begin
      // Prevent non null charactes after ending #0
      if c <> #0 then
        c := fileinfo[i - startlump].name[j];
      lump_p.name[j] := c;
    end;
  end;

  if reloadname <> '' then
    handle.Free;
end;

//
// W_Reload
// Flushes any of the reloadable lumps in memory
//  and reloads the directory.
//
procedure W_Reload;
var
  header: wadinfo_t;
  lumpcount: integer;
  lump_p: Plumpinfo_t;
  i: integer;
  handle: TCachedFile;
  length: integer;
  fileinfo: Pfilelump_tArray;
begin
  if reloadname = '' then
    exit;

  if not fexists(reloadname) then
    I_Error('W_Reload(): File %s does not exist' + #13#10, [reloadname]);

  try
    handle := TCachedFile.Create(reloadname, fOpen);
  except
    handle := nil;
    I_Error('W_Reload(): couldn''t open %s', [reloadname]);
  end;

  handle.Read(header, SizeOf(header));
  lumpcount := header.numlumps;
  length := lumpcount * SizeOf(filelump_t);
  GetMem(fileinfo, length);
  handle.Seek(header.infotableofs, sFromBeginning);
  handle.Read(fileinfo^, length);

  // Fill in lumpinfo

  for i := reloadlump to reloadlump + lumpcount - 1 do
  begin
    lump_p := @lumpinfo[reloadlump];
    if boolval(lumpcache[i]) then
      Z_Free(lumpcache[i]);

    lump_p.position := fileinfo[i - reloadlump].filepos;
    lump_p.size := fileinfo[i - reloadlump].size;
  end;

  handle.Free;
end;

//
// W_InitMultipleFiles
// Pass a null terminated list of files to use.
// All files are optional, but at least one file
//  must be found.
// Files with a .wad extension are idlink files
//  with multiple lumps.
// Other files are single lumps with the base filename
//  for the lump name.
// Lump names can appear multiple times.
// The name searcher looks backwards, so a later file
//  does override all earlier ones.
//
procedure W_InitMultipleFiles(filenames: PStringArray);
var
  size: integer;
  i: integer;
begin
  // open all the files, load headers, and count lumps
  numlumps := 0;

  // will be realloced as lumps are added
  GetMem(lumpinfo, 1);

  i := 0;
  while filenames[i] <> '' do
  begin
    W_AddFile(filenames[i]);
    inc(i);
  end;

  if numlumps = 0 then
    I_Error('W_InitMultipleFiles(): no files found');

  // set up caching
  size := numlumps * SizeOf(pointer);
  GetMem(lumpcache, size);

  if not boolval(lumpcache) then
    I_Error('W_InitMultipleFiles(): Couldn''t allocate lumpcache');

  memset(lumpcache, 0, size);
end;

//
// W_InitFile
// Just initialize from a single file.
//
procedure W_InitFile(const filename: string);
var
  names: array[0..1] of string;
begin
  names[0] := filename;
  names[1] := '';
  W_InitMultipleFiles(@names);
end;

//
// W_NumLumps
//
function W_NumLumps: integer;
begin
  result := numlumps;
end;

//
// W_CheckNumForName
// Returns -1 if name not found.
//
type
  name8_t = record
    case integer of
      0 : (s: char8_t);
      1 : (x: array[0..1] of integer);
    end;

function W_CheckNumForName(const name: string): integer;
var
  name8: name8_t;
  v1: integer;
  v2: integer;
  lump_p: Plumpinfo_t;
  len: integer;
  i: integer;
begin
  len := Length(name);
  if len > 8 then
    I_Error('W_CheckNumForName(): name string has more that 8 characters: %s', [name]);

  // make the name into two integers for easy compares
  for i := 1 to len do
    name8.s[i - 1] := toupper(name[i]); // case insensitive
  for i := len to 7 do
    name8.s[i] := #0;

  v1 := name8.x[0];
  v2 := name8.x[1];

  // scan backwards so patch lump files take precedence
  for i := numlumps - 1 downto 0 do
  begin
    lump_p := @lumpinfo[i];
    if (lump_p.v1 = v1) and (lump_p.v2 = v2) then
    begin
      result := i;
      exit;
    end;
  end;

  // TFB. Not found.
  result := -1;
end;

//
// W_GetNumForName
// Calls W_CheckNumForName, but bombs out if not found.
//
function W_GetNumForName(const name: string): integer;
begin
  result := W_CheckNumForName(name);
  if result = -1 then
    I_Error('W_GetNumForName(): %s not found!', [name]);
end;

//
// W_LumpLength
// Returns the buffer size needed to load the given lump.
//
function W_LumpLength(lump: integer): integer;
begin
  if lump >= numlumps then
    I_Error('W_LumpLength(): %d >= numlumps', [lump]);

  result := lumpinfo[lump].size;
end;

//
// W_ReadLump
// Loads the lump into the given buffer,
//  which must be >= W_LumpLength().
//
procedure W_ReadLump(lump: integer; dest: pointer);
var
  c: integer;
  l: Plumpinfo_t;
  handle: TCachedFile;
begin
  if lump >= numlumps then
    I_Error('W_ReadLump(): %d >= numlumps', [lump]);

  l := @lumpinfo[lump];

  if l.handle = nil then
  begin
    // reloadable file, so use open / read / close
    if not fexists(reloadname) then
      I_Error('W_ReadLump(): couldn''t open %s', [reloadname]);

    try
      handle := TCachedFile.Create(reloadname, fOpen);
    except
      handle := nil;
      I_Error('W_ReadLump(): couldn''t open %s', [reloadname]);
    end
  end
  else
    handle := l.handle;

  handle.Seek(l.position, sFromBeginning);
  c := handle.Read(dest^, l.size);

  if c < l.size then
    I_Error('W_ReadLump(): only read %d of %d on lump %d', [c, l.size, lump]);

  if l.handle = nil then
    handle.Free;

end;

//
// W_CacheLumpNum
//
function W_CacheLumpNum(lump: integer; tag: integer): pointer;
begin
  if lump >= numlumps then
    I_Error('W_CacheLumpNum(): lumn = %d, >= numlumps', [lump]);

  if lump < 0 then
    I_Error('W_CacheLumpNum(): lumn = %d, < 0', [lump]);

  if not boolval(lumpcache[lump]) then
  begin
    // read the lump in
    //printf ("cache miss on lump %i\n",lump);
    Z_Malloc(W_LumpLength(lump), tag, @lumpcache[lump]); // VJ     ptr := Z_Malloc(W_LumpLength(lump), tag, { @ ???)lumpcache[lump]);
    W_ReadLump(lump, lumpcache[lump]);
  end
  else
  begin
    //printf ("cache hit on lump %i\n",lump);
    Z_ChangeTag(lumpcache[lump],tag);
  end;

  result := lumpcache[lump];
end;

//
// W_CacheLumpName
//
function W_CacheLumpName(const name: string; tag: integer): pointer;
begin
  result := W_CacheLumpNum(W_GetNumForName(name), tag);
end;


(* Unit i_io implementation *)


                                               

procedure I_IOMessageBox(const s: string);
begin
  MessageBox(hMainWnd, PChar(s), AppTitle, MB_OK);
end;

procedure I_IOErrorMessageBox(const s: string);
begin
  MessageBox(hMainWnd, PChar(s), AppTitle, MB_OK or MB_ICONERROR);
end;

procedure I_IOprintf(const s: string);
begin
end;

procedure I_InitializeIO;
var
  filename: string;
begin
  if M_CheckParm('-debugfile') <> 0 then
    sprintf(filename, 'debug%d.txt', [consoleplayer])
  else
    filename := 'debug.txt';

  if M_CheckParmCDROM then
  begin
    stderr := TFile.Create('c:\dooomdata\stderr.txt', fCreate);
    filename := 'c:\dooomdata\' + filename;
  end
  else
    stderr := TFile.Create('stderr.txt', fCreate);

  printf('debug output to: %s' + #13#10, [filename]);
  debugfile := TFile.Create(filename, fCreate);
end;


procedure I_ShutdownIO;
begin
  stderr.Free;
  debugfile.Free;
end;


(* Unit d_think implementation *)



(* Unit m_argv implementation *)


                                              

var
  cdchecked: integer = -1;

{$IFDEF DEBUG}
var
  cmdparams: TStringList;
{$ENDIF}

function M_CheckParm(const check: string): integer;
var
  i: integer;
begin
{$IFDEF DEBUG}
  if cmdparams.IndexOf(check) < 0 then
    cmdparams.Add(check);
{$ENDIF}
  for i := 1 to myargc - 1 do
    if strupper(check) = myargv[i] then
    begin
      result := i;
      exit;
    end;
  result := 0;
end;

function M_CheckParmCDROM: boolean;
begin
  if cdchecked = -1 then
  begin
    cdchecked := M_CheckParm('-cdrom');
    {$I-}
    if cdchecked > 0 then
      MkDir('c:\doomdata');
    {$I+}
  end;
  result := cdchecked > 0;
end;

procedure M_InitArgv;
var
  i: integer;
begin
  myargc := ParamCount + 1;
  for i := 0 to myargc - 1 do
    myargv[i] := strupper(ParamStr(i));
  for i := myargc to MAXARGS do
    myargv[i] := '';
end;


(* Unit p_saveg implementation *)


                                                                                                                                                                                                                                                      

// Pads save_p to a 4-byte boundary
//  so that the load/save works on SGI&Gecko.

procedure PADSAVEP;
begin
  save_p := PByteArray(integer(save_p) + ((4 - (integer(save_p) and 3) and 3)));
end;

//
// P_ArchivePlayers
//
procedure P_ArchivePlayers;
var
  i: integer;
  j: integer;
  dest: Pplayer_t;
begin
  for i := 0 to MAXPLAYERS - 1 do
  begin
    if not playeringame[i] then
      continue;

    PADSAVEP;

    dest := Pplayer_t(save_p);
    memcpy(dest, @players[i], SizeOf(player_t));
    save_p := PByteArray(integer(save_p) + SizeOf(player_t));
    for j := 0 to Ord(NUMPSPRITES) - 1 do
      if boolval(dest.psprites[j].state) then
        dest.psprites[j].state := Pstate_t(pOperation(dest.psprites[j].state, @states[0], '-', SizeOf(dest.psprites[j].state^)));
  end;
end;

//
// P_UnArchivePlayers
//
procedure P_UnArchivePlayers;
var
  i: integer;
  j: integer;
begin
  for i := 0 to MAXPLAYERS - 1 do
  begin
    if not playeringame[i] then
      continue;

    PADSAVEP;

    memcpy(@players[i], save_p, SizeOf(player_t));
    incp(pointer(save_p), SizeOf(player_t));

    // will be set when unarc thinker
    players[i].mo := nil;
    players[i]._message := '';
    players[i].attacker := nil;

    for j := 0 to Ord(NUMPSPRITES) - 1 do
      if boolval(players[i].psprites[j].state) then
        players[i].psprites[j].state := @states[integer(players[i].psprites[j].state)];
  end;
end;

//
// P_ArchiveWorld
//
procedure P_ArchiveWorld;
var
  i: integer;
  j: integer;
  sec: Psector_t;
  li: Pline_t;
  si: Pside_t;
  put: PSmallIntArray;
begin
  put := PSmallIntArray(save_p);

  // do sectors
  i := 0;
  while i < numsectors do
  begin
    sec := Psector_t(@sectors[i]);
	  put[0] := sec.floorheight div FRACUNIT;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.ceilingheight div FRACUNIT;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.floorpic;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.ceilingpic;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.lightlevel;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.special; // needed?
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := sec.tag;  // needed?
    incp(pointer(put), SizeOf(SmallInt));
    inc(i);
  end;

  // do lines
  i := 0;
  while i < numlines do
  begin
    li := Pline_t(@lines[i]);
    put[0] := li.flags;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := li.special;
    incp(pointer(put), SizeOf(SmallInt));
    put[0] := li.tag;
    incp(pointer(put), SizeOf(SmallInt));
    for j := 0 to 1 do
    begin
      if li.sidenum[j] = -1 then
        continue;

      si := @sides[li.sidenum[j]];

      put[0] := si.textureoffset div FRACUNIT;
      incp(pointer(put), SizeOf(SmallInt));
      put[0] := si.rowoffset div FRACUNIT;;
      incp(pointer(put), SizeOf(SmallInt));
      put[0] := si.toptexture;
      incp(pointer(put), SizeOf(SmallInt));
      put[0] := si.bottomtexture;
      incp(pointer(put), SizeOf(SmallInt));
      put[0] := si.midtexture;
      incp(pointer(put), SizeOf(SmallInt));
    end;
    inc(i);
  end;

  save_p := PByteArray(put);
end;

//
// P_UnArchiveWorld
//
procedure P_UnArchiveWorld;
var
  i: integer;
  j: integer;
  sec: Psector_t;
  li: Pline_t;
  si: Pside_t;
  get: PSmallIntArray;
begin
  get := PSmallIntArray(save_p);

  // do sectors
  i := 0;
  while i < numsectors do
  begin
    sec := Psector_t(@sectors[i]);
    sec.floorheight := get[0] * FRACUNIT;
    incp(pointer(get), SizeOf(SmallInt));
    sec.ceilingheight := get[0] * FRACUNIT;
    incp(pointer(get), SizeOf(SmallInt));
    sec.floorpic := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    sec.ceilingpic := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    sec.lightlevel := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    sec.special := get[0]; // needed?
    incp(pointer(get), SizeOf(SmallInt));
    sec.tag := get[0]; // needed?
    incp(pointer(get), SizeOf(SmallInt));
    sec.specialdata := nil;
    sec.soundtarget := nil;
    inc(i);
  end;

  // do lines
  i := 0;
  while i < numlines do
  begin
    li := Pline_t(@lines[i]);
    li.flags := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    li.special := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    li.tag := get[0];
    incp(pointer(get), SizeOf(SmallInt));
    for j := 0 to 1 do
    begin
      if li.sidenum[j] = -1 then
        continue;
      si := @sides[li.sidenum[j]];
      si.textureoffset := get[0] * FRACUNIT;
      incp(pointer(get), SizeOf(SmallInt));
      si.rowoffset := get[0] * FRACUNIT;
      incp(pointer(get), SizeOf(SmallInt));
      si.toptexture := get[0];
      incp(pointer(get), SizeOf(SmallInt));
      si.bottomtexture := get[0];
      incp(pointer(get), SizeOf(SmallInt));
      si.midtexture := get[0];
      incp(pointer(get), SizeOf(SmallInt));
    end;
    inc(i);
  end;
  save_p := PByteArray(get);
end;

//
// Thinkers
//
type
  thinkerclass_t = (tc_end, tc_mobj);

//
// P_ArchiveThinkers
//
procedure P_ArchiveThinkers;
var
  th: Pthinker_t;
  mobj: Pmobj_t;
begin
  // save off the current thinkers
  th := thinkercap.next;
  while th <> @thinkercap do
  begin
    if @th._function.acp1 = @P_MobjThinker then
    begin
      save_p[0] := Ord(tc_mobj);
      incp(pointer(save_p));
	    PADSAVEP;
      mobj := Pmobj_t(save_p);
	    memcpy(mobj, th, SizeOf(mobj^));
      incp(pointer(save_p), SizeOf(mobj^));
      mobj.state := Pstate_t(pOperation(mobj.state, @states[0], '-', SizeOf(mobj.state^)));

      if boolval(mobj.player) then
        mobj.player := Pplayer_t(pOperation(mobj.player, @players[0], '-', SizeOf(players[0])) + 1);
    end;
	// I_Error ("P_ArchiveThinkers: Unknown thinker function");
    th := th.next;
  end;

  // add a terminating marker
  save_p[0] := Ord(tc_end);
  incp(pointer(save_p));
end;


// P_UnArchiveThinkers
//
procedure P_UnArchiveThinkers;
var
  tclass: byte;
  currentthinker: Pthinker_t;
  next: Pthinker_t;
  mobj: Pmobj_t;
begin
  // remove all the current thinkers
  currentthinker := thinkercap.next;
  while currentthinker <> @thinkercap do
  begin
    next := currentthinker.next;

    if @currentthinker._function.acp1 = @P_MobjThinker then
      P_RemoveMobj(Pmobj_t(currentthinker))
    else
      Z_Free(currentthinker);

    currentthinker := next;
  end;
  P_InitThinkers;

  // read in saved thinkers
  while true do
  begin
    tclass := save_p[0];
    incp(pointer(save_p));
    case tclass of
      Ord(tc_end):
        exit; // end of list

      Ord(tc_mobj):
        begin
          PADSAVEP;
          mobj := Z_Malloc(SizeOf(mobj^), PU_LEVEL, nil);
          memcpy(mobj, save_p, SizeOf(mobj^));
          incp(pointer(save_p), SizeOf(mobj^));
          mobj.state := @states[integer(mobj.state)];
          mobj.target := nil;
          if boolval(mobj.player) then
          begin
            mobj.player := @players[integer(mobj.player) - 1];

            Pplayer_t(mobj.player).mo := mobj;
          end;
          P_SetThingPosition(mobj);
          mobj.info := @mobjinfo[Ord(mobj._type)];
          mobj.floorz := Psubsector_t(mobj.subsector).sector.floorheight;
          mobj.ceilingz := Psubsector_t(mobj.subsector).sector.ceilingheight;
          @mobj.thinker._function.acp1 := @P_MobjThinker;
          P_AddThinker(@mobj.thinker);
        end;
      else
        I_Error('P_UnArchiveThinkers(): Unknown tclass %d in savegame', [tclass]);
    end;
  end;
end;

//
// P_ArchiveSpecials
//
type
  specials_e = (
    tc_ceiling,
    tc_door,
    tc_floor,
    tc_plat,
    tc_flash,
    tc_strobe,
    tc_glow,
    tc_endspecials
  );



//
// Things to handle:
//
// T_MoveCeiling, (ceiling_t: sector_t * swizzle), - active list
// T_VerticalDoor, (vldoor_t: sector_t * swizzle),
// T_MoveFloor, (floormove_t: sector_t * swizzle),
// T_LightFlash, (lightflash_t: sector_t * swizzle),
// T_StrobeFlash, (strobe_t: sector_t *),
// T_Glow, (glow_t: sector_t *),
// T_PlatRaise, (plat_t: sector_t *), - active list
//
procedure P_ArchiveSpecials;
var
  th: Pthinker_t;
  th1: Pthinker_t;
  ceiling: Pceiling_t;
  door: Pvldoor_t;
  floor: Pfloormove_t;
  plat: Pplat_t;
  flash: Plightflash_t;
  strobe: Pstrobe_t;
  glow: Pglow_t;
  i: integer;
begin
  // save off the current thinkers
  th1 := thinkercap.next;
  while th1 <> @thinkercap do
  begin
    th := th1;
    th1 := th1.next;
    if not Assigned(th._function.acv) then
    begin
      i := 0;
      while i < MAXCEILINGS do
      begin
        if activeceilings[i] = Pceiling_t(th) then
          break;
        inc(i);
      end;

	    if i < MAXCEILINGS then
      begin
        save_p[0] := Ord(tc_ceiling);
        incp(pointer(save_p));
        PADSAVEP;
        ceiling := Pceiling_t(save_p);
        memcpy(ceiling, th, SizeOf(ceiling^));
        incp(pointer(save_p), SizeOf(ceiling^));
        ceiling.sector := Psector_t(pOperation(ceiling.sector, @sectors[0], '-', SizeOf(sectors[0])));
      end;
      continue;
    end;

    if @th._function.acp1 = @T_MoveCeiling then
    begin
      save_p[0] := Ord(tc_ceiling);
      incp(pointer(save_p));
      PADSAVEP;
      ceiling := Pceiling_t(save_p);
      memcpy(ceiling, th, SizeOf(ceiling^));
      incp(pointer(save_p), SizeOf(ceiling^));
      ceiling.sector := Psector_t(pOperation(ceiling.sector, @sectors[0], '-', SizeOf(sectors[0])));
      continue;
    end;

    if @th._function.acp1 = @T_VerticalDoor then
    begin
      save_p[0] := Ord(tc_door);
      incp(pointer(save_p));
      PADSAVEP;
      door := Pvldoor_t(save_p);
      memcpy(door, th, SizeOf(door^));
      incp(pointer(save_p), SizeOf(door^));
      door.sector := Psector_t(pOperation(door.sector, @sectors[0], '-', SizeOf(door.sector^)));
      continue;
    end;

    if @th._function.acp1 = @T_MoveFloor then
    begin
	    save_p[0] := Ord(tc_floor);
      incp(pointer(save_p));
      PADSAVEP;
      floor := Pfloormove_t(save_p);
      memcpy(floor, th, SizeOf(floor^));
      incp(pointer(save_p), SizeOf(floor^));
      floor.sector := Psector_t(pOperation(floor.sector, @sectors[0], '-', SizeOf(floor.sector^)));
      continue;
    end;

    if @th._function.acp1 = @T_PlatRaise then
    begin
      save_p[0] := Ord(tc_plat);
	    incp(pointer(save_p));
	    PADSAVEP;
	    plat := Pplat_t(save_p);
	    memcpy(plat, th, SizeOf(plat^));
	    incp(pointer(save_p), SizeOf(plat^));
	    plat.sector := Psector_t(pOperation(plat.sector, @sectors[0], '-', SizeOf(plat.sector^)));
	    continue;
    end;

    if @th._function.acp1 = @T_LightFlash then
    begin
      save_p[0] := Ord(tc_flash);
      incp(pointer(save_p));
      PADSAVEP;
      flash := Plightflash_t(save_p);
      memcpy(flash, th, SizeOf(flash^));
      incp(pointer(save_p), SizeOf(flash^));
      flash.sector := Psector_t(pOperation(flash.sector, @sectors[0], '-', SizeOf(flash.sector^)));
      continue;
    end;

    if @th._function.acp1 = @T_StrobeFlash then
    begin
      save_p[0] := Ord(tc_strobe);
      incp(pointer(save_p));
      PADSAVEP;
      strobe := Pstrobe_t(save_p);
      memcpy(strobe, th, SizeOf(strobe^));
      incp(pointer(save_p), SizeOf(strobe^));
      strobe.sector := Psector_t(POperation(strobe.sector, @sectors[0], '-', SizeOf(strobe.sector^)));
      continue;
    end;

    if @th._function.acp1 = @T_Glow then
    begin
      save_p[0] := Ord(tc_glow);
      incp(pointer(save_p));
      PADSAVEP;
      glow := Pglow_t(save_p);
      memcpy(glow, th, SizeOf(glow^));
      incp(pointer(save_p), SizeOf(glow^));
      glow.sector := Psector_t(pOperation(glow.sector, @sectors[0], '-', SizeOf(glow.sector^)));
      continue;
    end;
  end;

  // add a terminating marker
  save_p[0] := Ord(tc_endspecials);
  incp(pointer(save_p));
end;

//
// P_UnArchiveSpecials
//
procedure P_UnArchiveSpecials;
var
  tclass: byte;
  ceiling: Pceiling_t;
  door: Pvldoor_t;
  floor: Pfloormove_t;
  plat: Pplat_t;
  flash: Plightflash_t;
  strobe: Pstrobe_t;
  glow: Pglow_t;
begin
  // read in saved thinkers
  while true do
  begin
    tclass := save_p[0];
    incp(pointer(save_p));
    case tclass of
      Ord(tc_endspecials):
        exit; // end of list

      Ord(tc_ceiling):
        begin
          PADSAVEP;
          ceiling := Z_Malloc(SizeOf(ceiling^), PU_LEVEL, nil);
          memcpy(ceiling, save_p, SizeOf(ceiling^));
          incp(pointer(save_p), SizeOf(ceiling^));
          ceiling.sector := @sectors[integer(ceiling.sector)];
          ceiling.sector.specialdata := ceiling;

          if Assigned(ceiling.thinker._function.acp1) then // VJ works ???
            @ceiling.thinker._function.acp1 := @T_MoveCeiling;

          P_AddThinker(@ceiling.thinker);
          P_AddActiveCeiling(ceiling);
        end;

      Ord(tc_door):
        begin
          PADSAVEP;
          door := Z_Malloc(SizeOf(door^), PU_LEVEL, nil);
          memcpy(door, save_p, SizeOf(door^));
          incp(pointer(save_p), SizeOf(door^));
          door.sector := @sectors[integer(door.sector)];
          door.sector.specialdata := door;
          @door.thinker._function.acp1 := @T_VerticalDoor;
          P_AddThinker(@door.thinker);
        end;

      Ord(tc_floor):
        begin
          PADSAVEP;
          floor := Z_Malloc(SizeOf(floor^), PU_LEVEL, nil);
          memcpy(floor, save_p, SizeOf(floor^));
          incp(pointer(save_p), SizeOf(floor^));
          floor.sector := @sectors[integer(floor.sector)];
          floor.sector.specialdata := floor;
          @floor.thinker._function.acp1 := @T_MoveFloor;
          P_AddThinker(@floor.thinker);
        end;

      Ord(tc_plat):
        begin
          PADSAVEP;
          plat := Z_Malloc(SizeOf(plat^), PU_LEVEL, nil);
          memcpy(plat, save_p, SizeOf(plat^));
          incp(pointer(save_p), SizeOf(plat^));
          plat.sector := @sectors[integer(plat.sector)];
          plat.sector.specialdata := plat;

          if Assigned(plat.thinker._function.acp1) then  // VJ ??? from serialization
            @plat.thinker._function.acp1 := @T_PlatRaise;

          P_AddThinker(@plat.thinker);
          P_AddActivePlat(plat);
        end;

      Ord(tc_flash):
        begin
          PADSAVEP;
          flash := Z_Malloc(Sizeof(flash^), PU_LEVEL, nil);
          memcpy(flash, save_p, SizeOf(flash^));
          incp(pointer(save_p), SizeOf(flash^));
          flash.sector := @sectors[integer(flash.sector)];
          @flash.thinker._function.acp1 := @T_LightFlash;
          P_AddThinker(@flash.thinker);
        end;

      Ord(tc_strobe):
        begin
          PADSAVEP;
          strobe := Z_Malloc(SizeOf(strobe^), PU_LEVEL, nil);
          memcpy(strobe, save_p, SizeOf(strobe^));
          incp(pointer(save_p), SizeOf(strobe^));
          strobe.sector := @sectors[integer(strobe.sector)];
          @strobe.thinker._function.acp1 := @T_StrobeFlash;
          P_AddThinker(@strobe.thinker);
        end;

      Ord(tc_glow):
        begin
          PADSAVEP;
          glow := Z_Malloc(SizeOf(glow^), PU_LEVEL, nil);
          memcpy(glow, save_p, SizeOf(glow^));
          incp(pointer(save_p), SizeOf(glow^));
          glow.sector := @sectors[integer(glow.sector)];
          @glow.thinker._function.acp1 := @T_Glow;
          P_AddThinker(@glow.thinker);
        end;

      else
        I_Error('P_UnarchiveSpecials(): Unknown tclass %d in savegame', [tclass]);
    end;
  end;
end;


(* Unit m_cheat implementation *)


                            

function get_cheatseq_string(const A: array of char): string; // VJ
var i: integer;
begin
  result := '';
  i := 0;
  repeat
    result := result + A[i];
    inc(i);
  until A[i] = Chr($FF);
end;

function get_cheatseq_string(const A: string): string;  // VJ
var i: integer;
begin
  result := '';
  i := 1;
  repeat
    result := result + A[i];
    inc(i);
  until A[i] = Chr($FF);
end;

function get_cheatseq_string(const x: integer): string; // VJ
begin
  result := '';
  if x <> 0 then
    I_Error('get_cheatseq_string(): invalid parameter: %d', [x]);
end;

function SCRAMBLE(a: integer): integer;
begin
  result := _SHL(a and 1, 7) +
            _SHL(a and 2, 5) +
            (a and 4) +
            _SHL(a and 8, 1) +
            _SHR(a and 16, 1) +
            (a and 32) +
            _SHR(a and 64, 5) +
            _SHR(a and 128, 7);
end;

var
  firsttime: boolean = true;
  cheat_xlate_table: array[0..255] of char;

//
// Called in st_stuff module, which handles the input.
// Returns a 1 if the cheat was successful, 0 if failed.
//
function cht_CheckCheat(cht: Pcheatseq_t; key: char): boolean;
var
  i: integer;
begin
  result := false;

  if firsttime then
  begin
    firsttime := false;
    for i := 0 to 255 do
      cheat_xlate_table[i] := Chr(SCRAMBLE(i));
  end;

  if (cht.p = '') then
    cht.p := cht.sequence; // initialize if first time

  if length(cht.p) = 0 then
    cht.p := key
  else if (cht.p[1] = #0) then
    cht.p[1] := key
  else if (length(cht.p) > 1) and (cht.p[2] = #0) then
  begin
    cht.p[2] := key;
    result := true;
  end
  else if cheat_xlate_table[Ord(key)] = cht.p[1] then
    Delete(cht.p, 1, 1)
  else
    cht.p := cht.sequence;

  if length(cht.p) > 0 then
  begin
    if cht.p[1] = #1 then
      Delete(cht.p, 1, 1)
    else if cht.p[1] = Chr($FF) then // end of sequence character
    begin
      cht.p := cht.sequence;
      result := true;
    end;
  end
  else
    result := true;
end;

procedure cht_GetParam(cht: Pcheatseq_t; var buffer: string);
begin
  buffer := cht.p;
end;


(* Unit d_ticcmd implementation *)




(* Unit z_zone implementation *)


               

//
// ZONE MEMORY ALLOCATION
//
// There is never any space between memblocks,
//  and there will never be two contiguous free memblocks.
// The rover can be left pointing at a non-empty block.
//
// It is of no value to free a cachable block,
//  because it will get overwritten automatically if needed.
//

const
  ZONEID = $1d4a11;

type
  memzone_t = record
    // total bytes malloced, including header
    size: integer;
    // start / end cap for linked list
    blocklist: memblock_t;
    rover: Pmemblock_t;
  end;
  Pmemzone_t = ^memzone_t;

var
  mainzone: Pmemzone_t;

//
// Z_ClearZone
//
procedure Z_ClearZone(zone: Pmemzone_t);
var
  block: Pmemblock_t;
begin
  // set the entire zone to one free block
	block := Pmemblock_t(integer(zone) + SizeOf(memzone_t));
  zone.blocklist.next := block;
	zone.blocklist.prev := block;

  zone.blocklist.user := PPointer(zone);
  zone.blocklist.tag := PU_STATIC;
  zone.rover := block;

  block.prev := @zone.blocklist;
  block.next := block.prev;

  // NULL indicates a free block.
  block.user := nil;

  block.size := zone.size - SizeOf(memzone_t);
end;

//
// Z_Init
//
procedure Z_Init;
var
  block: Pmemblock_t;
  size: integer;
begin
  mainzone := Pmemzone_t(I_ZoneBase(size));
  mainzone.size := size;

  // set the entire zone to one free block
	block := Pmemblock_t(integer(mainzone) + SizeOf(memzone_t));
  mainzone.blocklist.next := block;
  mainzone.blocklist.prev := block;

  mainzone.blocklist.user := PPointer(mainzone);
  mainzone.blocklist.tag := PU_STATIC;
  mainzone.rover := block;

  block.prev := @mainzone.blocklist;
  block.next := block.prev;

  // NULL indicates a free block.
  block.user := nil;

  block.size := mainzone.size - SizeOf(memzone_t);
end;

//
// Z_Free
//
procedure Z_Free(ptr: pointer);
var
  block: Pmemblock_t;
  other: Pmemblock_t;
begin
  block := Pmemblock_t(integer(ptr) - SizeOf(memblock_t));

  if block.id <> ZONEID then
    I_Error('Z_Free(): freed a pointer without ZONEID');

  if integer(block.user) > $100 then
  begin
    // smaller values are not pointers
    // Note: OS-dependend?

    // clear the user's mark
    block.user^ := nil;
  end;

  // mark as free
  block.user := nil;
  block.tag := 0;
  block.id := 0;

  other := block.prev;

  if not boolval(other.user) then
  begin
    // merge with previous free block
    other.size := other.size + block.size;
    other.next := block.next;
    other.next.prev := other;

    if block = mainzone.rover then
      mainzone.rover := other;

    block := other;
  end;

  other := block.next;
  if not boolval(other.user) then
  begin
    // merge the next free block onto the end
    block.size := block.size + other.size;
    block.next := other.next;
    block.next.prev := block;

    if other = mainzone.rover then
      mainzone.rover := block;
  end;
end;

//
// Z_Malloc
// You can pass a NULL user if the tag is < PU_PURGELEVEL.
//
const
  MINFRAGMENT = 64;

function Z_Malloc(size: integer; tag: integer; user: pointer): pointer;
var
  extra: integer;
  start: Pmemblock_t;
  rover: Pmemblock_t;
  newblock: Pmemblock_t;
  base: Pmemblock_t;
begin

  size := (size + 3) and (not 3);

  // scan through the block list,
  // looking for the first free block
  // of sufficient size,
  // throwing out any purgable blocks along the way.

  // account for size of block header
  size := size + SizeOf(memblock_t);

  // if there is a free block behind the rover,
  //  back up over them
  base := mainzone.rover;

  if not boolval(base.prev.user) then
    base := base.prev;

  rover := base;
  start := base.prev;

  repeat
    if rover = start then
    begin
      // scanned all the way around the list
      I_Error('Z_Malloc(): failed on allocation of %d bytes', [size]);
    end;

    if boolval(rover.user) then
    begin
      if rover.tag < PU_PURGELEVEL then
      begin
        // hit a block that can't be purged,
        //  so move base past it
        rover := rover.next;
        base := rover;
	    end
	    else
      begin
        // free the rover block (adding the size to base)
        // the rover can be the base block
        base := base.prev;
        Z_Free(pointer(integer(rover) + SizeOf(memblock_t)));
        base := base.next;
        rover := base.next;
      end;
    end
    else
      rover := rover.next;
  until not (boolval(base.user) or (base.size < size));


  // found a block big enough
  extra := base.size - size;

  if extra > MINFRAGMENT then
  begin
    // there will be a free fragment after the allocated block
    newblock := Pmemblock_t(integer(base) + size);
    newblock.size := extra;

    // NULL indicates free block.
    newblock.user := nil;
    newblock.tag := 0;
    newblock.prev := base;
    newblock.next := base.next;
    newblock.next.prev := newblock;

    base.next := newblock;
    base.size := size;
  end;

  if boolval(user) then
  begin
    // mark as an in use block
    base.user := user;
    PPointer(user)^ := Pointer(integer(base) + SizeOf(memblock_t));
//	*(void **)user = (void *) ((byte *)base + sizeof(memblock_t));
  end
  else
  begin
    if tag >= PU_PURGELEVEL then
      I_Error('Z_Malloc(): an owner is required for purgable blocks');

    // mark as in use, but unowned
    base.user := Pointer(2);
  end;

  base.tag := tag;

  // next allocation will start looking here
  mainzone.rover := base.next;

  base.id := ZONEID;

  result := Pointer(integer(base) + SizeOf(memblock_t));
end;

//
// Z_FreeTags
//
procedure Z_FreeTags(lowtag: integer; hightag: integer);
var
  block: Pmemblock_t;
  next: Pmemblock_t;
begin
  block := mainzone.blocklist.next;
  while block <> @mainzone.blocklist do
  begin
    // get link before freeing
    next := block.next;

    // not free block?
    if boolval(block.user) then
      if (block.tag >= lowtag) and (block.tag <= hightag) then
        Z_Free(Pointer(integer(block) + SizeOf(memblock_t)));
    block := next;
  end;
end;

//
// Z_DumpHeap
// Note: TFileDumpHeap( stdout ) ?
//
procedure Z_DumpHeap(lowtag: integer; hightag: integer);
var
  block: Pmemblock_t;
begin
  printf('zone size: %d  location: %s' + #13#10,
    [mainzone.size, IntToStrZfill(8, integer(mainzone))]);

  printf('tag range: %s to %s' + #13#10,
    [IntToStrZfill(3, lowtag), IntToStrZfill(3, hightag)]);

  block := mainzone.blocklist.next;
  while true do
  begin
    if (block.tag >= lowtag) and (block.tag <= hightag) then
      printf('block:%s    size:%s    user:%s    tag:%s' + #13#10,
		    [IntToStrZfill(8, integer(block)), IntToStrZfill(7, block.size),
         IntToStrZfill(8, integer(block.user)), IntToStrZfill(3, block.tag)]);
    if block.next = @mainzone.blocklist then
    begin
      // all blocks have been hit
      break;
    end;

    if integer(block) + block.size <> integer(block.next) then
      printf('ERROR: block size does not touch the next block' + #13#10);

    if block.next.prev <> block then
      printf('ERROR: next block doesn''t have proper back link' + #13#10);

    if (not boolval(block.user)) and (not boolval(block.next.user)) then
      printf('ERROR: two consecutive free blocks' + #13#10);

    block := block.next;
  end;
end;

procedure Z_FileDumpHeap(var f: file);
var
  block: Pmemblock_t;
begin
  fprintf(f, 'zone size: %d  location: %s' + #13#10,
    [mainzone.size, IntToStrZfill(8, integer(mainzone))]);

  block := mainzone.blocklist.next;
  while true do
  begin
    fprintf(f, 'block:%s    size:%s    user:%s    tag:%s' + #13#10,
      [IntToStrZfill(8, integer(block)), IntToStrZfill(7, block.size),
       IntToStrZfill(8, integer(block.user)), IntToStrZfill(3, block.tag)]);
    if block.next = @mainzone.blocklist then
    begin
      // all blocks have been hit
      break;
    end;

    if integer(block) + block.size <> integer(block.next) then
      fprintf(f, 'ERROR: block size does not touch the next block' + #13#10);

    if block.next.prev <> block then
      fprintf(f, 'ERROR: next block doesn''t have proper back link' + #13#10);

    if (not boolval(block.user)) and (not boolval(block.next.user)) then
      fprintf(f, 'ERROR: two consecutive free blocks' + #13#10);

    block := block.next;
  end;
end;

procedure Z_FileDumpHeap(const filename: string);
var f: file;
begin
  assign(f, filename);
  {$I-}
  rewrite(f, 1);
  {$I+}
  if IOResult = 0 then
  begin
    Z_FileDumpHeap(f);
    closefile(f);
  end
  else
    I_Error('Z_FileDumpHeap(): Can not create output file: %s', [filename]);
end;

//
// Z_CheckHeap
//
procedure Z_CheckHeap;
var
  block: Pmemblock_t;
begin
  block := mainzone.blocklist.next;
  while true do
  begin
    if block.next = @mainzone.blocklist then
    begin
      // all blocks have been hit
	    break;
    end;

    if integer(block) + block.size <> integer(block.next) then
      I_Error('Z_CheckHeap(): block size does not touch the next block');

    if block.next.prev <> block then
      I_Error('Z_CheckHeap(): next block doesn''t have proper back link');

    if (not boolval(block.user)) and (not boolval(block.next.user)) then
      I_Error('Z_CheckHeap(): two consecutive free blocks');

    block := block.next
  end;
end;

//
// Z_ChangeTag
//
procedure Z_ChangeTag(ptr: pointer; tag: integer);
begin
  Z_ChangeTag2(ptr, tag);
end;

procedure Z_ChangeTag2(ptr: pointer; tag: integer);
var
  block: Pmemblock_t;
begin
  block := Pmemblock_t(integer(ptr) - SizeOf(memblock_t));

  if block.id <> ZONEID then
    I_Error('Z_ChangeTag(): freed a pointer without ZONEID');

  if (tag >= PU_PURGELEVEL) and (LongWord(block.user) < $100) then
    I_Error('Z_ChangeTag(): an owner is required for purgable blocks');

  block.tag := tag;
end;

//
// Z_FreeMemory
//
function Z_FreeMemory: integer;
var
  block: Pmemblock_t;
begin
  result := 0;

  block := mainzone.blocklist.next;
  while block <> @mainzone.blocklist do
  begin
    if (not boolval(block.user)) or (block.tag >= PU_PURGELEVEL) then
      result := result + block.size;
    block := block.next;
  end;
end;


(* Unit i_input implementation *)


                                                                                         

function TranslateKey(keycode: integer): integer;
begin
  case keycode of
    VK_LEFT,
    VK_NUMPAD4: result := KEY_LEFTARROW;
    VK_RIGHT,
    VK_NUMPAD6: result := KEY_RIGHTARROW;
    VK_DOWN,
    VK_NUMPAD2: result := KEY_DOWNARROW;
    VK_UP,
    VK_NUMPAD8: result := KEY_UPARROW;
    VK_ESCAPE: result := KEY_ESCAPE;
    VK_RETURN: result := KEY_ENTER;
    VK_TAB: result := KEY_TAB;
    VK_F1: result := KEY_F1;
    VK_F2: result := KEY_F2;
    VK_F3: result := KEY_F3;
    VK_F4: result := KEY_F4;
    VK_F5: result := KEY_F5;
    VK_F6: result := KEY_F6;
    VK_F7: result := KEY_F7;
    VK_F8: result := KEY_F8;
    VK_F9: result := KEY_F9;
    VK_F10: result := KEY_F10;
    VK_F11: result := KEY_F11;
    VK_F12: result := KEY_F12;
    189: result := KEY_MINUS;
    187: result := KEY_EQUALS;
    VK_BACK: result := KEY_BACKSPACE;
    VK_PAUSE: result := KEY_PAUSE;
    VK_NUMPAD3: result := KEY_PAGEDOWN;
    VK_NUMPAD9: result := KEY_PAGEUP;
    VK_NUMPAD0: result := KEY_INS;
  else
    if (keycode >= Ord('A')) and (keycode <= Ord('Z')) then
      result := Ord(tolower(Chr(keycode)))
    else if keycode < 128 then
      result := keycode
    else
      result := 0;
  end;
end;

function TranslateSysKey(keycode: integer): integer;
begin
  case keycode of
    VK_SHIFT: result := KEY_RSHIFT;
    VK_CONTROL: result := KEY_RCTRL;
    VK_MENU: result := KEY_RALT;
  else
    result := 0;
  end;
end;

var
  g_pDI: IDirectInputA = nil;
  g_pdidKeyboard: IDirectInputDevice = nil;
  dikb: TDIKeyboardState; // DirectInput keyboard state buffer
  curkeys: PKeyboardState;
  oldkeys: PKeyboardState;

//-----------------------------------------------------------------------------
// Name: CreateDInput()
// Desc: Initialize the DirectInput variables using:
//           DirectInputCreate
//           IDirectInput::CreateDevice
//           IDirectInputDevice::SetDataFormat
//           IDirectInputDevice::SetCooperativeLevel
//-----------------------------------------------------------------------------
procedure I_InitInput;
var
  hres: HRESULT;

  procedure I_ErrorInitInput(const msg: string);
  begin
    I_Error('I_InitInput(): %s failed, result = %d', [msg, hres]);
  end;

begin
  if usedirectinput then
  begin
    // Register with the DirectInput subsystem and get a pointer
    // to a IDirectInput interface we can use
    hres := DirectInputCreate(hInstance, DIRECTINPUT_VERSION, g_pDI, nil);
    if hres <> DD_OK then
      I_ErrorInitInput('DirectInputCreate');

    // Obtain an interface to the keybord device
    hres := g_pDI.CreateDevice(GUID_SysKeyboard, g_pdidKeyboard, nil);
    if hres <> DD_OK then
      I_ErrorInitInput('CreateDevice');

    // Set the data format to "keyboard format". A data format specifies which
    // controls on a device we are interested in, and how they should be
    // reported. This tells DirectInput that we will be passing an array of 256
    // bytes to IDirectInputDevice::GetDeviceState().
    hres := g_pdidKeyboard.SetDataFormat(c_dfDIKeyboard);
    if hres <> DD_OK then
      I_ErrorInitInput('SetDataFormat');

    // Set the cooperative level to let DirectInput know how this device
    // should interact with the system and with other DirectInput applications.
    // Use DISCL_NONEXCLUSIVE to retrieve device data when acquired, not
    // interfering with any other applications which are reading mouse data.
    // Use DISCL_FOREGROUND so that if the user switches away from our app,
    // automatically release the device back to the system.
    hres := g_pdidKeyboard.SetCooperativeLevel(hMainWnd, DISCL_NONEXCLUSIVE or DISCL_FOREGROUND);
    if hres <> DD_OK then
      I_ErrorInitInput('SetCooperativeLevel');

  end;

  curkeys := malloc(SizeOf(TKeyboardState));
  oldkeys := malloc(SizeOf(TKeyboardState));

end;

//-----------------------------------------------------------------------------
// Name: I_ShutdownInput
// Desc: Terminate our usage of DirectInput
//-----------------------------------------------------------------------------
procedure I_ShutdownInput;
begin
  if usedirectinput then
  begin
    if g_pDI <> nil then
    begin
      // Destroy the keyboard object
      if g_pdidKeyboard <> nil then
      begin
        // Unacquire the device (just in case) before exitting.
        g_pdidKeyboard.Unacquire;
        g_pdidKeyboard._Release;
      end;

      // Destroy the DInput object
      g_pDI._Release;
    end;
  end;

  freemem(curkeys);
  freemem(oldkeys);
end;

//-----------------------------------------------------------------------------
// Name: I_ProcessInput;
// Desc: The game plays here. Read keyboard data and displaying it.
//-----------------------------------------------------------------------------
procedure I_ProcessInput;

  function DIKEYtoVK(Key: Byte): Integer;
  begin
    Result := 0;
    case Key of
      DIK_ESCAPE       : Result := VK_ESCAPE;
      DIK_1            : Result := Ord('1');
      DIK_2            : Result := Ord('2');
      DIK_3            : Result := Ord('3');
      DIK_4            : Result := Ord('4');
      DIK_5            : Result := Ord('5');
      DIK_6            : Result := Ord('6');
      DIK_7            : Result := Ord('7');
      DIK_8            : Result := Ord('8');
      DIK_9            : Result := Ord('9');
      DIK_0            : Result := Ord('0');
      DIK_EQUALS       : Result := Ord('=');
      DIK_BACK         : Result := VK_BACK;
      DIK_TAB          : Result := VK_TAB;
      DIK_Q            : Result := Ord('Q');
      DIK_W            : Result := Ord('W');
      DIK_E            : Result := Ord('E');
      DIK_R            : Result := Ord('R');
      DIK_T            : Result := Ord('T');
      DIK_Y            : Result := Ord('Y');
      DIK_U            : Result := Ord('U');
      DIK_I            : Result := Ord('I');
      DIK_O            : Result := Ord('O');
      DIK_P            : Result := Ord('P');
      DIK_LBRACKET     : Result := Ord('[');
      DIK_RBRACKET     : Result := Ord(']');
      DIK_RETURN       : Result := VK_RETURN;
      DIK_LCONTROL     : Result := VK_CONTROL;
      DIK_A            : Result := Ord('A');
      DIK_S            : Result := Ord('S');
      DIK_D            : Result := Ord('D');
      DIK_F            : Result := Ord('F');
      DIK_G            : Result := Ord('G');
      DIK_H            : Result := Ord('H');
      DIK_J            : Result := Ord('J');
      DIK_K            : Result := Ord('K');
      DIK_L            : Result := Ord('L');
      DIK_SEMICOLON    : Result := Ord(';');
      DIK_APOSTROPHE   : Result := Ord('''');
      DIK_LSHIFT       : Result := VK_SHIFT;
      DIK_BACKSLASH    : Result := Ord('\');
      DIK_Z            : Result := Ord('Z');
      DIK_X            : Result := Ord('X');
      DIK_C            : Result := Ord('C');
      DIK_V            : Result := Ord('V');
      DIK_B            : Result := Ord('B');
      DIK_N            : Result := Ord('N');
      DIK_M            : Result := Ord('M');
      DIK_COMMA        : Result := Ord(',');
      DIK_PERIOD       : Result := Ord('.');
      DIK_SLASH        : Result := Ord('/');
      DIK_RSHIFT       : Result := VK_SHIFT;
      DIK_MULTIPLY     : Result := Ord('*');
      DIK_LMENU        : Result := VK_MENU;
      DIK_SPACE        : Result := VK_SPACE;
      DIK_CAPITAL      : Result := VK_CAPITAL;
      DIK_F1           : Result := VK_F1;
      DIK_F2           : Result := VK_F2;
      DIK_F3           : Result := VK_F3;
      DIK_F4           : Result := VK_F4;
      DIK_F5           : Result := VK_F5;
      DIK_F6           : Result := VK_F6;
      DIK_F7           : Result := VK_F7;
      DIK_F8           : Result := VK_F8;
      DIK_F9           : Result := VK_F9;
      DIK_F10          : Result := VK_F10;
      DIK_NUMLOCK      : Result := VK_NUMLOCK;
      DIK_SCROLL       : Result := VK_SCROLL;
      DIK_NUMPAD7      : Result := VK_NUMPAD7;
      DIK_NUMPAD8      : Result := VK_NUMPAD8;
      DIK_NUMPAD9      : Result := VK_NUMPAD9;
      DIK_SUBTRACT     : Result := VK_SUBTRACT;
      DIK_NUMPAD4      : Result := VK_NUMPAD4;
      DIK_NUMPAD5      : Result := VK_NUMPAD5;
      DIK_NUMPAD6      : Result := VK_NUMPAD6;
      DIK_ADD          : Result := VK_ADD;
      DIK_NUMPAD1      : Result := VK_NUMPAD1;
      DIK_NUMPAD2      : Result := VK_NUMPAD2;
      DIK_NUMPAD3      : Result := VK_NUMPAD3;
      DIK_NUMPAD0      : Result := VK_NUMPAD0;
      DIK_DECIMAL      : Result := VK_DECIMAL;
      DIK_F11          : Result := VK_F11;
      DIK_F12          : Result := VK_F12;
      DIK_NUMPADENTER  : Result := VK_RETURN;
      DIK_RCONTROL     : Result := VK_CONTROL;
      DIK_DIVIDE       : Result := VK_DIVIDE;
      DIK_RMENU        : Result := VK_MENU;
      DIK_HOME         : Result := VK_HOME;
      DIK_UP           : Result := VK_UP;
      DIK_PRIOR        : Result := VK_PRIOR;
      DIK_LEFT         : Result := VK_LEFT;
      DIK_RIGHT        : Result := VK_RIGHT;
      DIK_END          : Result := VK_END;
      DIK_DOWN         : Result := VK_DOWN;
      DIK_NEXT         : Result := VK_NEXT;
      DIK_INSERT       : Result := VK_INSERT;
      DIK_DELETE       : Result := VK_DELETE;
      DIK_LWIN         : Result := VK_LWIN;
      DIK_RWIN         : Result := VK_RWIN;
      DIK_APPS         : Result := VK_APPS;
    end;
  end;

var
  hres: HRESULT;
  i: integer;
  ev: event_t;
  key: integer;
  p: PKeyboardState;
begin
  if I_GameFinished then
    exit;

  // VJ -> DirectInput does not work
  if usedirectinput and (g_pdidKeyboard <> nil) then
  begin
    hres := g_pdidKeyboard.GetDeviceState(SizeOf(dikb), dikb);
    if hres = DIERR_INPUTLOST then
    begin
      // DirectInput is telling us that the input stream has been
      // interrupted. Re-acquire and try again.
      hres := g_pdidKeyboard.Acquire;
      if hres = DD_OK then
        I_ProcessInput;
      exit;
    end;

//  The DirectInput key code is converted into the Windows virtual key code.
    for i := Low(dikb) to High(dikb) do
      if dikb[i] and $80 <> 0 then
        curkeys[Byte(DIKEYtoVK(i))] := $80;
  end
  else
    GetKeyboardState(curkeys^);

  ZeroMemory(ev, SizeOf(ev));

  for i := 0 to SizeOf(curkeys^) - 1 do
  begin

    if (oldkeys[i] and $80) <> (curkeys[i] and $80) then
    begin
      key := TranslateKey(i);
      if key <> 0 then
      begin
        if curkeys[i] and $80 <> 0 then
          ev._type := ev_keydown
        else
          ev._type := ev_keyup;
        ev.data1 := key;
        D_PostEvent(@ev);
      end;

      key := TranslateSysKey(i);
      if key <> 0 then
      begin
        if curkeys[i] and $80 <> 0 then
          ev._type := ev_keydown
        else
          ev._type := ev_keyup;
        ev.data1 := key;
        D_PostEvent(@ev);
      end;
    end;

  end;

  p := oldkeys;
  oldkeys := curkeys;
  curkeys := p;
end;

procedure I_SynchronizeInput(active: boolean);
begin
  if usedirectinput and (g_pdidKeyboard <> nil) then
  begin
    if active then
      g_pdidKeyboard.Acquire
    else
      g_pdidKeyboard.Unacquire;
  end;
end;


(* Unit doomtype implementation *)


function MAXCHAR : char;
begin
  MAXCHAR := char($7f);
end;

function MAXSHORT : smallint;
begin
  MAXSHORT := smallint($7fff);
end;

function MAXINT : longint;
begin
  MAXINT := longint($7fffffff);
end;

function MAXLONG : longint;
begin
  MAXLONG := longint($7fffffff);
end;

function MINCHAR : char;
begin
  MINCHAR := char($80);
end;

function MINSHORT : smallint;
begin
  MINSHORT := smallint($8000);
end;

function MININT : longint;
begin
  MININT := longint($80000000);
end;

function MINLONG : longint;
begin
  MINLONG := longint($80000000);
end;


(* Unit info_h implementation *)



(* Unit info implementation *)


                                                       

var
  states1: array[0..ord(NUMSTATES) - 1] of state_t = (
   (
    sprite: SPR_TROO;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_NULL

   (
    sprite: SPR_SHTG;         // sprite
    frame: 4;                 // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_LIGHTDONE

   (
    sprite: SPR_PUNG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH

   (
    sprite: SPR_PUNG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCHDOWN;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCHDOWN

   (
    sprite: SPR_PUNG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCHUP;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCHUP

   (
    sprite: SPR_PUNG;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH1

   (
    sprite: SPR_PUNG;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH2

   (
    sprite: SPR_PUNG;         // sprite
    frame: 3;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH4;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH3

   (
    sprite: SPR_PUNG;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH5;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH4

   (
    sprite: SPR_PUNG;         // sprite
    frame: 1;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUNCH;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUNCH5

   (
    sprite: SPR_PISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOL;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0),                // S_PISTOL  // misc2

   (
    sprite: SPR_PISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOLDOWN;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOLDOWN

   (
    sprite: SPR_PISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOLUP;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOLUP

   (
    sprite: SPR_PISG;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOL2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOL1

   (
    sprite: SPR_PISG;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOL3;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0),                // S_PISTOL2  // misc2

   (
    sprite: SPR_PISG;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOL4;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOL3

   (
    sprite: SPR_PISG;         // sprite
    frame: 1;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PISTOL;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOL4

   (
    sprite: SPR_PISF;         // sprite
    frame: 32768;             // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PISTOLFLASH

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUNDOWN;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUNDOWN

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUNUP;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUNUP

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN1

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN2

   (
    sprite: SPR_SHTG;         // sprite
    frame: 1;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN3

   (
    sprite: SPR_SHTG;         // sprite
    frame: 2;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN4

   (
    sprite: SPR_SHTG;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN6;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN5

   (
    sprite: SPR_SHTG;         // sprite
    frame: 2;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN7;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN6

   (
    sprite: SPR_SHTG;         // sprite
    frame: 1;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN8;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN7

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN9;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN8

   (
    sprite: SPR_SHTG;         // sprite
    frame: 0;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUN;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUN9

   (
    sprite: SPR_SHTF;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SGUNFLASH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUNFLASH1

   (
    sprite: SPR_SHTF;         // sprite
    frame: 32769;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SGUNFLASH2

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUNDOWN;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUNDOWN

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUNUP;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUNUP

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN1

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN2

   (
    sprite: SPR_SHT2;         // sprite
    frame: 1;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN4;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN3

   (
    sprite: SPR_SHT2;         // sprite
    frame: 2;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN5;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN4

   (
    sprite: SPR_SHT2;         // sprite
    frame: 3;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN6;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN5

   (
    sprite: SPR_SHT2;         // sprite
    frame: 4;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN7;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN6

   (
    sprite: SPR_SHT2;         // sprite
    frame: 5;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN8;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN7

   (
    sprite: SPR_SHT2;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN9;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN8

   (
    sprite: SPR_SHT2;         // sprite
    frame: 7;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN10;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN9

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUN;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUN10

   (
    sprite: SPR_SHT2;         // sprite
    frame: 1;                 // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSNR2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSNR1

   (
    sprite: SPR_SHT2;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUNDOWN;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSNR2

   (
    sprite: SPR_SHT2;         // sprite
    frame: 32776;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_DSGUNFLASH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUNFLASH1

   (
    sprite: SPR_SHT2;         // sprite
    frame: 32777;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DSGUNFLASH2

   (
    sprite: SPR_CHGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAIN;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAIN

   (
    sprite: SPR_CHGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAINDOWN;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAINDOWN

   (
    sprite: SPR_CHGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAINUP;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAINUP

   (
    sprite: SPR_CHGG;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAIN2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAIN1

   (
    sprite: SPR_CHGG;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAIN3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAIN2

   (
    sprite: SPR_CHGG;         // sprite
    frame: 1;                 // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CHAIN;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAIN3

   (
    sprite: SPR_CHGF;         // sprite
    frame: 32768;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAINFLASH1

   (
    sprite: SPR_CHGF;         // sprite
    frame: 32769;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CHAINFLASH2

   (
    sprite: SPR_MISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILE;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILE

   (
    sprite: SPR_MISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILEDOWN;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEDOWN

   (
    sprite: SPR_MISG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILEUP;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEUP

   (
    sprite: SPR_MISG;         // sprite
    frame: 1;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILE2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILE1

   (
    sprite: SPR_MISG;         // sprite
    frame: 1;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILE3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILE2

   (
    sprite: SPR_MISG;         // sprite
    frame: 1;                 // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILE;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILE3

   (
    sprite: SPR_MISF;         // sprite
    frame: 32768;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILEFLASH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEFLASH1

   (
    sprite: SPR_MISF;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILEFLASH3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEFLASH2

   (
    sprite: SPR_MISF;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MISSILEFLASH4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEFLASH3

   (
    sprite: SPR_MISF;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MISSILEFLASH4

   (
    sprite: SPR_SAWG;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAWB;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAW

   (
    sprite: SPR_SAWG;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAW;         // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAWB

   (
    sprite: SPR_SAWG;         // sprite
    frame: 2;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAWDOWN;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAWDOWN

   (
    sprite: SPR_SAWG;         // sprite
    frame: 2;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAWUP;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAWUP

   (
    sprite: SPR_SAWG;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAW2;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAW1

   (
    sprite: SPR_SAWG;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAW3;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAW2

   (
    sprite: SPR_SAWG;         // sprite
    frame: 1;                 // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SAW;         // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SAW3

   (
    sprite: SPR_PLSG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASMA;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMA

   (
    sprite: SPR_PLSG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASMADOWN;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMADOWN

   (
    sprite: SPR_PLSG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASMAUP;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMAUP

   (
    sprite: SPR_PLSG;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASMA2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMA1

   (
    sprite: SPR_PLSG;         // sprite
    frame: 1;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASMA;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMA2

   (
    sprite: SPR_PLSF;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMAFLASH1

   (
    sprite: SPR_PLSF;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASMAFLASH2

   (
    sprite: SPR_BFGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFG;         // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFG

   (
    sprite: SPR_BFGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGDOWN;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGDOWN

   (
    sprite: SPR_BFGG;         // sprite
    frame: 0;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGUP;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGUP

   (
    sprite: SPR_BFGG;         // sprite
    frame: 0;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFG2;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFG1

   (
    sprite: SPR_BFGG;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFG3;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFG2

   (
    sprite: SPR_BFGG;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFG4;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFG3

   (
    sprite: SPR_BFGG;         // sprite
    frame: 1;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFG;         // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFG4

   (
    sprite: SPR_BFGF;         // sprite
    frame: 32768;             // frame
    tics: 11;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGFLASH2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGFLASH1

   (
    sprite: SPR_BFGF;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIGHTDONE;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGFLASH2

   (
    sprite: SPR_BLUD;         // sprite
    frame: 2;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOOD2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOOD1

   (
    sprite: SPR_BLUD;         // sprite
    frame: 1;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOOD3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOOD2

   (
    sprite: SPR_BLUD;         // sprite
    frame: 0;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOOD3

   (
    sprite: SPR_PUFF;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUFF2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUFF1

   (
    sprite: SPR_PUFF;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUFF3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUFF2

   (
    sprite: SPR_PUFF;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PUFF4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUFF3

   (
    sprite: SPR_PUFF;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PUFF4

   (
    sprite: SPR_BAL1;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TBALL2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TBALL1

   (
    sprite: SPR_BAL1;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TBALL1;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TBALL2

   (
    sprite: SPR_BAL1;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TBALLX2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TBALLX1

   (
    sprite: SPR_BAL1;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TBALLX3;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TBALLX2

   (
    sprite: SPR_BAL1;         // sprite
    frame: 32772;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TBALLX3

   (
    sprite: SPR_BAL2;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RBALL2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RBALL1

   (
    sprite: SPR_BAL2;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RBALL1;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RBALL2

   (
    sprite: SPR_BAL2;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RBALLX2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RBALLX1

   (
    sprite: SPR_BAL2;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RBALLX3;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RBALLX2

   (
    sprite: SPR_BAL2;         // sprite
    frame: 32772;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RBALLX3

   (
    sprite: SPR_PLSS;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASBALL2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASBALL

   (
    sprite: SPR_PLSS;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASBALL;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASBALL2

   (
    sprite: SPR_PLSE;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASEXP2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASEXP

   (
    sprite: SPR_PLSE;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASEXP3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASEXP2

   (
    sprite: SPR_PLSE;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASEXP4;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASEXP3

   (
    sprite: SPR_PLSE;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLASEXP5;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASEXP4

   (
    sprite: SPR_PLSE;         // sprite
    frame: 32772;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLASEXP5

   (
    sprite: SPR_MISL;         // sprite
    frame: 32768;             // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ROCKET;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ROCKET

   (
    sprite: SPR_BFS1;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGSHOT2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGSHOT

   (
    sprite: SPR_BFS1;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGSHOT;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGSHOT2

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32768;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGLAND2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32769;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGLAND3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND2

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32770;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGLAND4;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND3

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32771;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGLAND5;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND4

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32772;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGLAND6;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND5

   (
    sprite: SPR_BFE1;         // sprite
    frame: 32773;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGLAND6

   (
    sprite: SPR_BFE2;         // sprite
    frame: 32768;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGEXP2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGEXP

   (
    sprite: SPR_BFE2;         // sprite
    frame: 32769;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGEXP3;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGEXP2

   (
    sprite: SPR_BFE2;         // sprite
    frame: 32770;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BFGEXP4;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGEXP3

   (
    sprite: SPR_BFE2;         // sprite
    frame: 32771;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFGEXP4

   (
    sprite: SPR_MISL;         // sprite
    frame: 32769;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EXPLODE2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EXPLODE1

   (
    sprite: SPR_MISL;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EXPLODE3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EXPLODE2

   (
    sprite: SPR_MISL;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EXPLODE3

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG01;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG02;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG01

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG02

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG2

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG3

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG4

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32772;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG6;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG5

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32773;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG7;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG6

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32774;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG8;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG7

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32775;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG9;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG8

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32776;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TFOG10;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG9

   (
    sprite: SPR_TFOG;         // sprite
    frame: 32777;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TFOG10

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG01;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG02;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG01

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG02

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG2

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG3

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_IFOG5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG4

   (
    sprite: SPR_IFOG;         // sprite
    frame: 32772;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_IFOG5

   (
    sprite: SPR_PLAY;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY

   (
    sprite: SPR_PLAY;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_RUN1

   (
    sprite: SPR_PLAY;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_RUN2

   (
    sprite: SPR_PLAY;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_RUN3

   (
    sprite: SPR_PLAY;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_RUN4

   (
    sprite: SPR_PLAY;         // sprite
    frame: 4;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_ATK1

   (
    sprite: SPR_PLAY;         // sprite
    frame: 32773;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_ATK1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_ATK2

   (
    sprite: SPR_PLAY;         // sprite
    frame: 6;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_PAIN

   (
    sprite: SPR_PLAY;         // sprite
    frame: 6;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_PAIN2

   (
    sprite: SPR_PLAY;         // sprite
    frame: 7;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE1

   (
    sprite: SPR_PLAY;         // sprite
    frame: 8;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE2

   (
    sprite: SPR_PLAY;         // sprite
    frame: 9;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE3

   (
    sprite: SPR_PLAY;         // sprite
    frame: 10;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE4

   (
    sprite: SPR_PLAY;         // sprite
    frame: 11;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE5

   (
    sprite: SPR_PLAY;         // sprite
    frame: 12;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE6

   (
    sprite: SPR_PLAY;         // sprite
    frame: 13;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_DIE7

   (
    sprite: SPR_PLAY;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE1

   (
    sprite: SPR_PLAY;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE2

   (
    sprite: SPR_PLAY;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE3

   (
    sprite: SPR_PLAY;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE4

   (
    sprite: SPR_PLAY;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE5

   (
    sprite: SPR_PLAY;         // sprite
    frame: 19;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE6

   (
    sprite: SPR_PLAY;         // sprite
    frame: 20;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE7

   (
    sprite: SPR_PLAY;         // sprite
    frame: 21;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PLAY_XDIE9;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE8

   (
    sprite: SPR_PLAY;         // sprite
    frame: 22;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAY_XDIE9

   (
    sprite: SPR_POSS;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_STND

   (
    sprite: SPR_POSS;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_STND2

   (
    sprite: SPR_POSS;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN1

   (
    sprite: SPR_POSS;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN2

   (
    sprite: SPR_POSS;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN3

   (
    sprite: SPR_POSS;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN4

   (
    sprite: SPR_POSS;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN5

   (
    sprite: SPR_POSS;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN6

   (
    sprite: SPR_POSS;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN7

   (
    sprite: SPR_POSS;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RUN8

   (
    sprite: SPR_POSS;         // sprite
    frame: 4;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_ATK1

   (
    sprite: SPR_POSS;         // sprite
    frame: 5;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_ATK2

   (
    sprite: SPR_POSS;         // sprite
    frame: 4;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_ATK3

   (
    sprite: SPR_POSS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_PAIN

   (
    sprite: SPR_POSS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_PAIN2

   (
    sprite: SPR_POSS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_DIE1

   (
    sprite: SPR_POSS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_DIE2

   (
    sprite: SPR_POSS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_DIE3

   (
    sprite: SPR_POSS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_DIE4

   (
    sprite: SPR_POSS;         // sprite
    frame: 11;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_DIE5

   (
    sprite: SPR_POSS;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE1

   (
    sprite: SPR_POSS;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE2

   (
    sprite: SPR_POSS;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE3

   (
    sprite: SPR_POSS;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE4

   (
    sprite: SPR_POSS;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE5

   (
    sprite: SPR_POSS;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE6

   (
    sprite: SPR_POSS;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE7

   (
    sprite: SPR_POSS;         // sprite
    frame: 19;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_XDIE9;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE8

   (
    sprite: SPR_POSS;         // sprite
    frame: 20;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_XDIE9

   (
    sprite: SPR_POSS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RAISE1

   (
    sprite: SPR_POSS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RAISE2

   (
    sprite: SPR_POSS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RAISE3

   (
    sprite: SPR_POSS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_POSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_POSS_RAISE4

   (
    sprite: SPR_SPOS;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_STND

   (
    sprite: SPR_SPOS;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_STND2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN1

   (
    sprite: SPR_SPOS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN3

   (
    sprite: SPR_SPOS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN4

   (
    sprite: SPR_SPOS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN5

   (
    sprite: SPR_SPOS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN6

   (
    sprite: SPR_SPOS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN7

   (
    sprite: SPR_SPOS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RUN8

   (
    sprite: SPR_SPOS;         // sprite
    frame: 4;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_ATK1

   (
    sprite: SPR_SPOS;         // sprite
    frame: 32773;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_ATK2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 4;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_ATK3

   (
    sprite: SPR_SPOS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_PAIN

   (
    sprite: SPR_SPOS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_PAIN2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_DIE1

   (
    sprite: SPR_SPOS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_DIE2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_DIE3

   (
    sprite: SPR_SPOS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_DIE4

   (
    sprite: SPR_SPOS;         // sprite
    frame: 11;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_DIE5

   (
    sprite: SPR_SPOS;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE1

   (
    sprite: SPR_SPOS;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE3

   (
    sprite: SPR_SPOS;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE4

   (
    sprite: SPR_SPOS;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE5

   (
    sprite: SPR_SPOS;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE6

   (
    sprite: SPR_SPOS;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE7

   (
    sprite: SPR_SPOS;         // sprite
    frame: 19;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_XDIE9;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE8

   (
    sprite: SPR_SPOS;         // sprite
    frame: 20;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_XDIE9

   (
    sprite: SPR_SPOS;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RAISE1

   (
    sprite: SPR_SPOS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RAISE2

   (
    sprite: SPR_SPOS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RAISE3

   (
    sprite: SPR_SPOS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RAISE4

   (
    sprite: SPR_SPOS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPOS_RAISE5

   (
    sprite: SPR_VILE;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_STND

   (
    sprite: SPR_VILE;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_STND2

   (
    sprite: SPR_VILE;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN1

   (
    sprite: SPR_VILE;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN2

   (
    sprite: SPR_VILE;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN3

   (
    sprite: SPR_VILE;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN4

   (
    sprite: SPR_VILE;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN5

   (
    sprite: SPR_VILE;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN6

   (
    sprite: SPR_VILE;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN7

   (
    sprite: SPR_VILE;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN8

   (
    sprite: SPR_VILE;         // sprite
    frame: 4;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN9

   (
    sprite: SPR_VILE;         // sprite
    frame: 4;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN10

   (
    sprite: SPR_VILE;         // sprite
    frame: 5;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN11

   (
    sprite: SPR_VILE;         // sprite
    frame: 5;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_RUN12

   (
    sprite: SPR_VILE;         // sprite
    frame: 32774;             // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK1

   (
    sprite: SPR_VILE;         // sprite
    frame: 32774;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK2

   (
    sprite: SPR_VILE;         // sprite
    frame: 32775;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK3

   (
    sprite: SPR_VILE;         // sprite
    frame: 32776;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK4

   (
    sprite: SPR_VILE;         // sprite
    frame: 32777;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK5

   (
    sprite: SPR_VILE;         // sprite
    frame: 32778;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK6

   (
    sprite: SPR_VILE;         // sprite
    frame: 32779;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK7

   (
    sprite: SPR_VILE;         // sprite
    frame: 32780;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK8

   (
    sprite: SPR_VILE;         // sprite
    frame: 32781;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK9

   (
    sprite: SPR_VILE;         // sprite
    frame: 32782;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_ATK11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK10

   (
    sprite: SPR_VILE;         // sprite
    frame: 32783;             // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_ATK11

   (
    sprite: SPR_VILE;         // sprite
    frame: 32794;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_HEAL2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_HEAL1

   (
    sprite: SPR_VILE;         // sprite
    frame: 32795;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_HEAL3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_HEAL2

   (
    sprite: SPR_VILE;         // sprite
    frame: 32796;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_HEAL3

   (
    sprite: SPR_VILE;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_PAIN

   (
    sprite: SPR_VILE;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_PAIN2

   (
    sprite: SPR_VILE;         // sprite
    frame: 16;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE1

   (
    sprite: SPR_VILE;         // sprite
    frame: 17;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE2

   (
    sprite: SPR_VILE;         // sprite
    frame: 18;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE3

   (
    sprite: SPR_VILE;         // sprite
    frame: 19;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE4

   (
    sprite: SPR_VILE;         // sprite
    frame: 20;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE5

   (
    sprite: SPR_VILE;         // sprite
    frame: 21;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE6

   (
    sprite: SPR_VILE;         // sprite
    frame: 22;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE7

   (
    sprite: SPR_VILE;         // sprite
    frame: 23;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE8

   (
    sprite: SPR_VILE;         // sprite
    frame: 24;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_VILE_DIE10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE9

   (
    sprite: SPR_VILE;         // sprite
    frame: 25;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_VILE_DIE10

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32768;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE1

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32769;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE2

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32768;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE3

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32769;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE4

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE6;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE5

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32769;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE7;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE6

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE8;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE7

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32769;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE9;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE8

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE10;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE9

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE11;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE10

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE12;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE11

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE13;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE12

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE14;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE13

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE15;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE14

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE16;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE15

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE17;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE16

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE18;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE17

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE19;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE18

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE20;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE19

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32773;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE21;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE20

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE22;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE21

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32773;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE23;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE22

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE24;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE23

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32773;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE25;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE24

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32774;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE26;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE25

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32775;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE27;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE26

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32774;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE28;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE27

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32775;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE29;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE28

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32774;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FIRE30;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE29

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32775;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FIRE30

   (
    sprite: SPR_PUFF;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SMOKE2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMOKE1

   (
    sprite: SPR_PUFF;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SMOKE3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMOKE2

   (
    sprite: SPR_PUFF;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SMOKE4;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMOKE3

   (
    sprite: SPR_PUFF;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SMOKE5;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMOKE4

   (
    sprite: SPR_PUFF;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMOKE5

   (
    sprite: SPR_FATB;         // sprite
    frame: 32768;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TRACER2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TRACER

   (
    sprite: SPR_FATB;         // sprite
    frame: 32769;             // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TRACER;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TRACER2

   (
    sprite: SPR_FBXP;         // sprite
    frame: 32768;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TRACEEXP2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TRACEEXP1

   (
    sprite: SPR_FBXP;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TRACEEXP3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TRACEEXP2

   (
    sprite: SPR_FBXP;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TRACEEXP3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_STND

   (
    sprite: SPR_SKEL;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_STND2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN1

   (
    sprite: SPR_SKEL;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN4

   (
    sprite: SPR_SKEL;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN5

   (
    sprite: SPR_SKEL;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN6

   (
    sprite: SPR_SKEL;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN7

   (
    sprite: SPR_SKEL;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN8

   (
    sprite: SPR_SKEL;         // sprite
    frame: 4;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN9

   (
    sprite: SPR_SKEL;         // sprite
    frame: 4;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN10

   (
    sprite: SPR_SKEL;         // sprite
    frame: 5;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN11

   (
    sprite: SPR_SKEL;         // sprite
    frame: 5;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RUN12

   (
    sprite: SPR_SKEL;         // sprite
    frame: 6;                 // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_FIST2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_FIST1

   (
    sprite: SPR_SKEL;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_FIST3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_FIST2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 7;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_FIST4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_FIST3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 8;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_FIST4

   (
    sprite: SPR_SKEL;         // sprite
    frame: 32777;             // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_MISS2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_MISS1

   (
    sprite: SPR_SKEL;         // sprite
    frame: 32777;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_MISS3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_MISS2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 10;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_MISS4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_MISS3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 10;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_MISS4

   (
    sprite: SPR_SKEL;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_PAIN

   (
    sprite: SPR_SKEL;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_PAIN2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 11;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE1

   (
    sprite: SPR_SKEL;         // sprite
    frame: 12;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 13;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 14;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE4

   (
    sprite: SPR_SKEL;         // sprite
    frame: 15;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE5

   (
    sprite: SPR_SKEL;         // sprite
    frame: 16;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_DIE6

   (
    sprite: SPR_SKEL;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE1

   (
    sprite: SPR_SKEL;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE2

   (
    sprite: SPR_SKEL;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE3

   (
    sprite: SPR_SKEL;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE4

   (
    sprite: SPR_SKEL;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE5

   (
    sprite: SPR_SKEL;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKEL_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKEL_RAISE6

   (
    sprite: SPR_MANF;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATSHOT2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATSHOT1

   (
    sprite: SPR_MANF;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATSHOT1;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATSHOT2

   (
    sprite: SPR_MISL;         // sprite
    frame: 32769;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATSHOTX2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATSHOTX1

   (
    sprite: SPR_MISL;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATSHOTX3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATSHOTX2

   (
    sprite: SPR_MISL;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATSHOTX3

   (
    sprite: SPR_FATT;         // sprite
    frame: 0;                 // frame
    tics: 15;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_STND

   (
    sprite: SPR_FATT;         // sprite
    frame: 1;                 // frame
    tics: 15;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_STND2

   (
    sprite: SPR_FATT;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN1

   (
    sprite: SPR_FATT;         // sprite
    frame: 0;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN2

   (
    sprite: SPR_FATT;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN3

   (
    sprite: SPR_FATT;         // sprite
    frame: 1;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN4

   (
    sprite: SPR_FATT;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN5

   (
    sprite: SPR_FATT;         // sprite
    frame: 2;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN6

   (
    sprite: SPR_FATT;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN7

   (
    sprite: SPR_FATT;         // sprite
    frame: 3;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN8

   (
    sprite: SPR_FATT;         // sprite
    frame: 4;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN9

   (
    sprite: SPR_FATT;         // sprite
    frame: 4;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN10

   (
    sprite: SPR_FATT;         // sprite
    frame: 5;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN11

   (
    sprite: SPR_FATT;         // sprite
    frame: 5;                 // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RUN12

   (
    sprite: SPR_FATT;         // sprite
    frame: 6;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK1

   (
    sprite: SPR_FATT;         // sprite
    frame: 32775;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK2

   (
    sprite: SPR_FATT;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK3

   (
    sprite: SPR_FATT;         // sprite
    frame: 6;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK4

   (
    sprite: SPR_FATT;         // sprite
    frame: 32775;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK5

   (
    sprite: SPR_FATT;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK6

   (
    sprite: SPR_FATT;         // sprite
    frame: 6;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK7

   (
    sprite: SPR_FATT;         // sprite
    frame: 32775;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK8

   (
    sprite: SPR_FATT;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_ATK10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK9

   (
    sprite: SPR_FATT;         // sprite
    frame: 6;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_ATK10

   (
    sprite: SPR_FATT;         // sprite
    frame: 9;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_PAIN

   (
    sprite: SPR_FATT;         // sprite
    frame: 9;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_PAIN2

   (
    sprite: SPR_FATT;         // sprite
    frame: 10;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE1

   (
    sprite: SPR_FATT;         // sprite
    frame: 11;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE2

   (
    sprite: SPR_FATT;         // sprite
    frame: 12;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE3

   (
    sprite: SPR_FATT;         // sprite
    frame: 13;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE4

   (
    sprite: SPR_FATT;         // sprite
    frame: 14;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE5

   (
    sprite: SPR_FATT;         // sprite
    frame: 15;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE6

   (
    sprite: SPR_FATT;         // sprite
    frame: 16;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE7

   (
    sprite: SPR_FATT;         // sprite
    frame: 17;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE8

   (
    sprite: SPR_FATT;         // sprite
    frame: 18;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_DIE10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE9

   (
    sprite: SPR_FATT;         // sprite
    frame: 19;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_DIE10

   (
    sprite: SPR_FATT;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE1

   (
    sprite: SPR_FATT;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE2

   (
    sprite: SPR_FATT;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE3

   (
    sprite: SPR_FATT;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE4

   (
    sprite: SPR_FATT;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE5

   (
    sprite: SPR_FATT;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE6

   (
    sprite: SPR_FATT;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RAISE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE7

   (
    sprite: SPR_FATT;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FATT_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FATT_RAISE8

   (
    sprite: SPR_CPOS;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_STND

   (
    sprite: SPR_CPOS;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_STND2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN1

   (
    sprite: SPR_CPOS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN3

   (
    sprite: SPR_CPOS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN4

   (
    sprite: SPR_CPOS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN5

   (
    sprite: SPR_CPOS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN6

   (
    sprite: SPR_CPOS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN7

   (
    sprite: SPR_CPOS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RUN8

   (
    sprite: SPR_CPOS;         // sprite
    frame: 4;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_ATK1

   (
    sprite: SPR_CPOS;         // sprite
    frame: 32773;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_ATK2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 32772;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_ATK3

   (
    sprite: SPR_CPOS;         // sprite
    frame: 5;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_ATK4

   (
    sprite: SPR_CPOS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_PAIN

   (
    sprite: SPR_CPOS;         // sprite
    frame: 6;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_PAIN2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE1

   (
    sprite: SPR_CPOS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE3

   (
    sprite: SPR_CPOS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE4

   (
    sprite: SPR_CPOS;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE5

   (
    sprite: SPR_CPOS;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE6

   (
    sprite: SPR_CPOS;         // sprite
    frame: 13;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_DIE7

   (
    sprite: SPR_CPOS;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE1

   (
    sprite: SPR_CPOS;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE3

   (
    sprite: SPR_CPOS;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE4

   (
    sprite: SPR_CPOS;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE5

   (
    sprite: SPR_CPOS;         // sprite
    frame: 19;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_XDIE6

   (
    sprite: SPR_CPOS;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE1

   (
    sprite: SPR_CPOS;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE2

   (
    sprite: SPR_CPOS;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE3

   (
    sprite: SPR_CPOS;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE4

   (
    sprite: SPR_CPOS;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE5

   (
    sprite: SPR_CPOS;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RAISE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE6

   (
    sprite: SPR_CPOS;         // sprite
    frame: 7;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CPOS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CPOS_RAISE7

   (
    sprite: SPR_TROO;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_STND

   (
    sprite: SPR_TROO;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_STND2

   (
    sprite: SPR_TROO;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN1

   (
    sprite: SPR_TROO;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN2

   (
    sprite: SPR_TROO;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN3

   (
    sprite: SPR_TROO;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN4

   (
    sprite: SPR_TROO;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN5

   (
    sprite: SPR_TROO;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN6

   (
    sprite: SPR_TROO;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN7

   (
    sprite: SPR_TROO;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RUN8

   (
    sprite: SPR_TROO;         // sprite
    frame: 4;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_ATK1

   (
    sprite: SPR_TROO;         // sprite
    frame: 5;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_ATK2

   (
    sprite: SPR_TROO;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_ATK3

   (
    sprite: SPR_TROO;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_PAIN

   (
    sprite: SPR_TROO;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_PAIN2

   (
    sprite: SPR_TROO;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_DIE1

   (
    sprite: SPR_TROO;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_DIE2

   (
    sprite: SPR_TROO;         // sprite
    frame: 10;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_DIE3

   (
    sprite: SPR_TROO;         // sprite
    frame: 11;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_DIE4

   (
    sprite: SPR_TROO;         // sprite
    frame: 12;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_DIE5

   (
    sprite: SPR_TROO;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE1

   (
    sprite: SPR_TROO;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE2

   (
    sprite: SPR_TROO;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE3

   (
    sprite: SPR_TROO;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE4

   (
    sprite: SPR_TROO;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE5

   (
    sprite: SPR_TROO;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE6

   (
    sprite: SPR_TROO;         // sprite
    frame: 19;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_XDIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE7

   (
    sprite: SPR_TROO;         // sprite
    frame: 20;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_XDIE8

   (
    sprite: SPR_TROO;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RAISE1

   (
    sprite: SPR_TROO;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RAISE2

   (
    sprite: SPR_TROO;         // sprite
    frame: 10;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RAISE3

   (
    sprite: SPR_TROO;         // sprite
    frame: 9;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RAISE4

   (
    sprite: SPR_TROO;         // sprite
    frame: 8;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TROO_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TROO_RAISE5

   (
    sprite: SPR_SARG;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_STND

   (
    sprite: SPR_SARG;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_STND2

   (
    sprite: SPR_SARG;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN1

   (
    sprite: SPR_SARG;         // sprite
    frame: 0;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN2

   (
    sprite: SPR_SARG;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN3

   (
    sprite: SPR_SARG;         // sprite
    frame: 1;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN4

   (
    sprite: SPR_SARG;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN5

   (
    sprite: SPR_SARG;         // sprite
    frame: 2;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN6

   (
    sprite: SPR_SARG;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN7

   (
    sprite: SPR_SARG;         // sprite
    frame: 3;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RUN8

   (
    sprite: SPR_SARG;         // sprite
    frame: 4;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_ATK1

   (
    sprite: SPR_SARG;         // sprite
    frame: 5;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_ATK2

   (
    sprite: SPR_SARG;         // sprite
    frame: 6;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_ATK3

   (
    sprite: SPR_SARG;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_PAIN

   (
    sprite: SPR_SARG;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_PAIN2

   (
    sprite: SPR_SARG;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE1

   (
    sprite: SPR_SARG;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE2

   (
    sprite: SPR_SARG;         // sprite
    frame: 10;                // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE3

   (
    sprite: SPR_SARG;         // sprite
    frame: 11;                // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE4

   (
    sprite: SPR_SARG;         // sprite
    frame: 12;                // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE5

   (
    sprite: SPR_SARG;         // sprite
    frame: 13;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_DIE6

   (
    sprite: SPR_SARG;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE1

   (
    sprite: SPR_SARG;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE2

   (
    sprite: SPR_SARG;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE3

   (
    sprite: SPR_SARG;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE4

   (
    sprite: SPR_SARG;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE5

   (
    sprite: SPR_SARG;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SARG_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SARG_RAISE6

   (
    sprite: SPR_HEAD;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_STND

   (
    sprite: SPR_HEAD;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RUN1

   (
    sprite: SPR_HEAD;         // sprite
    frame: 1;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_ATK1

   (
    sprite: SPR_HEAD;         // sprite
    frame: 2;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_ATK2

   (
    sprite: SPR_HEAD;         // sprite
    frame: 32771;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_ATK3

   (
    sprite: SPR_HEAD;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_PAIN

   (
    sprite: SPR_HEAD;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_PAIN3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_PAIN2

   (
    sprite: SPR_HEAD;         // sprite
    frame: 5;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_PAIN3

   (
    sprite: SPR_HEAD;         // sprite
    frame: 6;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE1

   (
    sprite: SPR_HEAD;         // sprite
    frame: 7;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE2

   (
    sprite: SPR_HEAD;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE3

   (
    sprite: SPR_HEAD;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE4

   (
    sprite: SPR_HEAD;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE5

   (
    sprite: SPR_HEAD;         // sprite
    frame: 11;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_DIE6

   (
    sprite: SPR_HEAD;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE1

   (
    sprite: SPR_HEAD;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE2

   (
    sprite: SPR_HEAD;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE3

   (
    sprite: SPR_HEAD;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE4

   (
    sprite: SPR_HEAD;         // sprite
    frame: 7;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE5

   (
    sprite: SPR_HEAD;         // sprite
    frame: 6;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEAD_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEAD_RAISE6

   (
    sprite: SPR_BAL7;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRBALL2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRBALL1

   (
    sprite: SPR_BAL7;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRBALL1;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRBALL2

   (
    sprite: SPR_BAL7;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRBALLX2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRBALLX1

   (
    sprite: SPR_BAL7;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRBALLX3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRBALLX2

   (
    sprite: SPR_BAL7;         // sprite
    frame: 32772;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRBALLX3

   (
    sprite: SPR_BOSS;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_STND

   (
    sprite: SPR_BOSS;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_STND2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN1

   (
    sprite: SPR_BOSS;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN3

   (
    sprite: SPR_BOSS;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN4

   (
    sprite: SPR_BOSS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN5

   (
    sprite: SPR_BOSS;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN6

   (
    sprite: SPR_BOSS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN7

   (
    sprite: SPR_BOSS;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RUN8

   (
    sprite: SPR_BOSS;         // sprite
    frame: 4;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_ATK1

   (
    sprite: SPR_BOSS;         // sprite
    frame: 5;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_ATK2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 6;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_ATK3

   (
    sprite: SPR_BOSS;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_PAIN

   (
    sprite: SPR_BOSS;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_PAIN2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE1

   (
    sprite: SPR_BOSS;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE3

   (
    sprite: SPR_BOSS;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE4

   (
    sprite: SPR_BOSS;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE5

   (
    sprite: SPR_BOSS;         // sprite
    frame: 13;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE6

   (
    sprite: SPR_BOSS;         // sprite
    frame: 14;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_DIE7

   (
    sprite: SPR_BOSS;         // sprite
    frame: 14;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE1

   (
    sprite: SPR_BOSS;         // sprite
    frame: 13;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE2

   (
    sprite: SPR_BOSS;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE3

   (
    sprite: SPR_BOSS;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE4

   (
    sprite: SPR_BOSS;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE5

   (
    sprite: SPR_BOSS;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RAISE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE6

   (
    sprite: SPR_BOSS;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOSS_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOSS_RAISE7

   (
    sprite: SPR_BOS2;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_STND

   (
    sprite: SPR_BOS2;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_STND2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN1

   (
    sprite: SPR_BOS2;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN3

   (
    sprite: SPR_BOS2;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN4

   (
    sprite: SPR_BOS2;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN5

   (
    sprite: SPR_BOS2;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN6

   (
    sprite: SPR_BOS2;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN7

   (
    sprite: SPR_BOS2;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RUN8

   (
    sprite: SPR_BOS2;         // sprite
    frame: 4;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_ATK1

   (
    sprite: SPR_BOS2;         // sprite
    frame: 5;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_ATK2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 6;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_ATK3

   (
    sprite: SPR_BOS2;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_PAIN

   (
    sprite: SPR_BOS2;         // sprite
    frame: 7;                 // frame
    tics: 2;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_PAIN2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE1

   (
    sprite: SPR_BOS2;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE3

   (
    sprite: SPR_BOS2;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE4

   (
    sprite: SPR_BOS2;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE5

   (
    sprite: SPR_BOS2;         // sprite
    frame: 13;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE6

   (
    sprite: SPR_BOS2;         // sprite
    frame: 14;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_DIE7

   (
    sprite: SPR_BOS2;         // sprite
    frame: 14;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE1

   (
    sprite: SPR_BOS2;         // sprite
    frame: 13;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE2

   (
    sprite: SPR_BOS2;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE3

   (
    sprite: SPR_BOS2;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE4

   (
    sprite: SPR_BOS2;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE5

   (
    sprite: SPR_BOS2;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RAISE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE6

   (
    sprite: SPR_BOS2;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BOS2_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BOS2_RAISE7

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32768;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_STND

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_STND;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_STND2

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_RUN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_RUN1

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_RUN1;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_RUN2

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32770;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_ATK2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_ATK1

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_ATK3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_ATK2

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_ATK4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_ATK3

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_ATK3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_ATK4

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32772;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_PAIN

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32772;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_RUN1;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_PAIN2

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32773;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_DIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE1

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32774;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_DIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE2

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32775;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_DIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE3

   (
    sprite: SPR_SKUL;         // sprite
    frame: 32776;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_DIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE4

   (
    sprite: SPR_SKUL;         // sprite
    frame: 9;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SKULL_DIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE5

   (
    sprite: SPR_SKUL;         // sprite
    frame: 10;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULL_DIE6

   (
    sprite: SPR_SPID;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_STND

   (
    sprite: SPR_SPID;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_STND2

   (
    sprite: SPR_SPID;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN1

   (
    sprite: SPR_SPID;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN2

   (
    sprite: SPR_SPID;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN3

   (
    sprite: SPR_SPID;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN4

   (
    sprite: SPR_SPID;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN5

   (
    sprite: SPR_SPID;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN6

   (
    sprite: SPR_SPID;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN7

   (
    sprite: SPR_SPID;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN8

   (
    sprite: SPR_SPID;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN9

   (
    sprite: SPR_SPID;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN10

   (
    sprite: SPR_SPID;         // sprite
    frame: 5;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN11

   (
    sprite: SPR_SPID;         // sprite
    frame: 5;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_RUN12

   (
    sprite: SPR_SPID;         // sprite
    frame: 32768;             // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_ATK1

   (
    sprite: SPR_SPID;         // sprite
    frame: 32774;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_ATK2

   (
    sprite: SPR_SPID;         // sprite
    frame: 32775;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_ATK3

   (
    sprite: SPR_SPID;         // sprite
    frame: 32775;             // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_ATK4

   (
    sprite: SPR_SPID;         // sprite
    frame: 8;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_PAIN

   (
    sprite: SPR_SPID;         // sprite
    frame: 8;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_PAIN2

   (
    sprite: SPR_SPID;         // sprite
    frame: 9;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE1

   (
    sprite: SPR_SPID;         // sprite
    frame: 10;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE2

   (
    sprite: SPR_SPID;         // sprite
    frame: 11;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE3

   (
    sprite: SPR_SPID;         // sprite
    frame: 12;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE4

   (
    sprite: SPR_SPID;         // sprite
    frame: 13;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE5

   (
    sprite: SPR_SPID;         // sprite
    frame: 14;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE6

   (
    sprite: SPR_SPID;         // sprite
    frame: 15;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE7

   (
    sprite: SPR_SPID;         // sprite
    frame: 16;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE8

   (
    sprite: SPR_SPID;         // sprite
    frame: 17;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE9

   (
    sprite: SPR_SPID;         // sprite
    frame: 18;                // frame
    tics: 30;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPID_DIE11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE10

   (
    sprite: SPR_SPID;         // sprite
    frame: 18;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPID_DIE11

   (
    sprite: SPR_BSPI;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_STND

   (
    sprite: SPR_BSPI;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_STND2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 0;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_SIGHT

   (
    sprite: SPR_BSPI;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN1

   (
    sprite: SPR_BSPI;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN3

   (
    sprite: SPR_BSPI;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN4

   (
    sprite: SPR_BSPI;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN5

   (
    sprite: SPR_BSPI;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN6

   (
    sprite: SPR_BSPI;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN7

   (
    sprite: SPR_BSPI;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN8

   (
    sprite: SPR_BSPI;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN9

   (
    sprite: SPR_BSPI;         // sprite
    frame: 4;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN10

   (
    sprite: SPR_BSPI;         // sprite
    frame: 5;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN11

   (
    sprite: SPR_BSPI;         // sprite
    frame: 5;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RUN12

   (
    sprite: SPR_BSPI;         // sprite
    frame: 32768;             // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_ATK1

   (
    sprite: SPR_BSPI;         // sprite
    frame: 32774;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_ATK2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 32775;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_ATK3

   (
    sprite: SPR_BSPI;         // sprite
    frame: 32775;             // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_ATK4

   (
    sprite: SPR_BSPI;         // sprite
    frame: 8;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_PAIN

   (
    sprite: SPR_BSPI;         // sprite
    frame: 8;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_PAIN2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 9;                 // frame
    tics: 20;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE1

   (
    sprite: SPR_BSPI;         // sprite
    frame: 10;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 11;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE3

   (
    sprite: SPR_BSPI;         // sprite
    frame: 12;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE4

   (
    sprite: SPR_BSPI;         // sprite
    frame: 13;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE5

   (
    sprite: SPR_BSPI;         // sprite
    frame: 14;                // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_DIE7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE6

   (
    sprite: SPR_BSPI;         // sprite
    frame: 15;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_DIE7

   (
    sprite: SPR_BSPI;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE1

   (
    sprite: SPR_BSPI;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE2

   (
    sprite: SPR_BSPI;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE3

   (
    sprite: SPR_BSPI;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE4

   (
    sprite: SPR_BSPI;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE5

   (
    sprite: SPR_BSPI;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RAISE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE6

   (
    sprite: SPR_BSPI;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSPI_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSPI_RAISE7

   (
    sprite: SPR_APLS;         // sprite
    frame: 32768;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLAZ2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLAZ

   (
    sprite: SPR_APLS;         // sprite
    frame: 32769;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLAZ;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLAZ2

   (
    sprite: SPR_APBX;         // sprite
    frame: 32768;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLEX2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLEX

   (
    sprite: SPR_APBX;         // sprite
    frame: 32769;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLEX3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLEX2

   (
    sprite: SPR_APBX;         // sprite
    frame: 32770;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLEX4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLEX3

   (
    sprite: SPR_APBX;         // sprite
    frame: 32771;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARACH_PLEX5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLEX4

   (
    sprite: SPR_APBX;         // sprite
    frame: 32772;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARACH_PLEX5

   (
    sprite: SPR_CYBR;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_STND

   (
    sprite: SPR_CYBR;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_STND;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_STND2

   (
    sprite: SPR_CYBR;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN1

   (
    sprite: SPR_CYBR;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN2

   (
    sprite: SPR_CYBR;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN3

   (
    sprite: SPR_CYBR;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN4

   (
    sprite: SPR_CYBR;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN5

   (
    sprite: SPR_CYBR;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN6

   (
    sprite: SPR_CYBR;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN7

   (
    sprite: SPR_CYBR;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN1;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_RUN8

   (
    sprite: SPR_CYBR;         // sprite
    frame: 4;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_ATK2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK1

   (
    sprite: SPR_CYBR;         // sprite
    frame: 5;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_ATK3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK2

   (
    sprite: SPR_CYBR;         // sprite
    frame: 4;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_ATK4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK3

   (
    sprite: SPR_CYBR;         // sprite
    frame: 5;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_ATK5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK4

   (
    sprite: SPR_CYBR;         // sprite
    frame: 4;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_ATK6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK5

   (
    sprite: SPR_CYBR;         // sprite
    frame: 5;                 // frame
    tics: 12;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN1;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_ATK6

   (
    sprite: SPR_CYBR;         // sprite
    frame: 6;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_RUN1;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_PAIN

   (
    sprite: SPR_CYBR;         // sprite
    frame: 7;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE1

   (
    sprite: SPR_CYBR;         // sprite
    frame: 8;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE2

   (
    sprite: SPR_CYBR;         // sprite
    frame: 9;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE3

   (
    sprite: SPR_CYBR;         // sprite
    frame: 10;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE4

   (
    sprite: SPR_CYBR;         // sprite
    frame: 11;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE5

   (
    sprite: SPR_CYBR;         // sprite
    frame: 12;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE6

   (
    sprite: SPR_CYBR;         // sprite
    frame: 13;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE7

   (
    sprite: SPR_CYBR;         // sprite
    frame: 14;                // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE9;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE8

   (
    sprite: SPR_CYBR;         // sprite
    frame: 15;                // frame
    tics: 30;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_CYBER_DIE10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE9

   (
    sprite: SPR_CYBR;         // sprite
    frame: 15;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CYBER_DIE10

   (
    sprite: SPR_PAIN;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_STND

   (
    sprite: SPR_PAIN;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN1

   (
    sprite: SPR_PAIN;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN2

   (
    sprite: SPR_PAIN;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN3

   (
    sprite: SPR_PAIN;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN4

   (
    sprite: SPR_PAIN;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN5

   (
    sprite: SPR_PAIN;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RUN6

   (
    sprite: SPR_PAIN;         // sprite
    frame: 3;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_ATK1

   (
    sprite: SPR_PAIN;         // sprite
    frame: 4;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_ATK2

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32773;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_ATK3

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32773;             // frame
    tics: 0;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_ATK4

   (
    sprite: SPR_PAIN;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_PAIN

   (
    sprite: SPR_PAIN;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_PAIN2

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32775;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE1

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32776;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE2

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32777;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE3

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32778;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE4

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32779;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_DIE6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE5

   (
    sprite: SPR_PAIN;         // sprite
    frame: 32780;             // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_DIE6

   (
    sprite: SPR_PAIN;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE1

   (
    sprite: SPR_PAIN;         // sprite
    frame: 11;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE2

   (
    sprite: SPR_PAIN;         // sprite
    frame: 10;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE3

   (
    sprite: SPR_PAIN;         // sprite
    frame: 9;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE4

   (
    sprite: SPR_PAIN;         // sprite
    frame: 8;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RAISE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE5

   (
    sprite: SPR_PAIN;         // sprite
    frame: 7;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PAIN_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PAIN_RAISE6

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_STND2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_STND

   (
    sprite: SPR_SSWV;         // sprite
    frame: 1;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_STND;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_STND2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN1

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN3

   (
    sprite: SPR_SSWV;         // sprite
    frame: 1;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN5

   (
    sprite: SPR_SSWV;         // sprite
    frame: 2;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN6

   (
    sprite: SPR_SSWV;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN7

   (
    sprite: SPR_SSWV;         // sprite
    frame: 3;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RUN8

   (
    sprite: SPR_SSWV;         // sprite
    frame: 4;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK1

   (
    sprite: SPR_SSWV;         // sprite
    frame: 5;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 32774;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK3

   (
    sprite: SPR_SSWV;         // sprite
    frame: 5;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 32774;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK5

   (
    sprite: SPR_SSWV;         // sprite
    frame: 5;                 // frame
    tics: 1;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_ATK2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_ATK6

   (
    sprite: SPR_SSWV;         // sprite
    frame: 7;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_PAIN2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_PAIN

   (
    sprite: SPR_SSWV;         // sprite
    frame: 7;                 // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_PAIN2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_DIE2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_DIE1

   (
    sprite: SPR_SSWV;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_DIE3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_DIE2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_DIE4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_DIE3

   (
    sprite: SPR_SSWV;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_DIE5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_DIE4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 12;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_DIE5

   (
    sprite: SPR_SSWV;         // sprite
    frame: 13;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE1

   (
    sprite: SPR_SSWV;         // sprite
    frame: 14;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 15;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE3

   (
    sprite: SPR_SSWV;         // sprite
    frame: 16;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 17;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE5

   (
    sprite: SPR_SSWV;         // sprite
    frame: 18;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE6

   (
    sprite: SPR_SSWV;         // sprite
    frame: 19;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE7

   (
    sprite: SPR_SSWV;         // sprite
    frame: 20;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_XDIE9;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE8

   (
    sprite: SPR_SSWV;         // sprite
    frame: 21;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_XDIE9

   (
    sprite: SPR_SSWV;         // sprite
    frame: 12;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RAISE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RAISE1

   (
    sprite: SPR_SSWV;         // sprite
    frame: 11;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RAISE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RAISE2

   (
    sprite: SPR_SSWV;         // sprite
    frame: 10;                // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RAISE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RAISE3

   (
    sprite: SPR_SSWV;         // sprite
    frame: 9;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RAISE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RAISE4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 8;                 // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SSWV_RUN1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SSWV_RAISE5

   (
    sprite: SPR_KEEN;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_KEENSTND;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_KEENSTND

   (
    sprite: SPR_KEEN;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN

   (
    sprite: SPR_KEEN;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN2

   (
    sprite: SPR_KEEN;         // sprite
    frame: 2;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN3

   (
    sprite: SPR_KEEN;         // sprite
    frame: 3;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN5;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN4

   (
    sprite: SPR_KEEN;         // sprite
    frame: 4;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN6;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN5

   (
    sprite: SPR_KEEN;         // sprite
    frame: 5;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN7;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN6

   (
    sprite: SPR_KEEN;         // sprite
    frame: 6;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN8;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN7

   (
    sprite: SPR_KEEN;         // sprite
    frame: 7;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN9;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN8

   (
    sprite: SPR_KEEN;         // sprite
    frame: 8;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN10;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN9

   (
    sprite: SPR_KEEN;         // sprite
    frame: 9;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN11;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN10

   (
    sprite: SPR_KEEN;         // sprite
    frame: 10;                // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_COMMKEEN12;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0),                // S_COMMKEEN11  // misc2

   (
    sprite: SPR_KEEN;         // sprite
    frame: 11;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COMMKEEN12

   (
    sprite: SPR_KEEN;         // sprite
    frame: 12;                // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_KEENPAIN2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_KEENPAIN

   (
    sprite: SPR_KEEN;         // sprite
    frame: 12;                // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_KEENSTND;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_KEENPAIN2

   (
    sprite: SPR_BBRN;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN

   (
    sprite: SPR_BBRN;         // sprite
    frame: 1;                 // frame
    tics: 36;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAIN;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN_PAIN

   (
    sprite: SPR_BBRN;         // sprite
    frame: 0;                 // frame
    tics: 100;                // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAIN_DIE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN_DIE1

   (
    sprite: SPR_BBRN;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAIN_DIE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN_DIE2

   (
    sprite: SPR_BBRN;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAIN_DIE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN_DIE3

   (
    sprite: SPR_BBRN;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAIN_DIE4

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAINEYE;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEYE

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 181;                // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAINEYE1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEYESEE

   (
    sprite: SPR_SSWV;         // sprite
    frame: 0;                 // frame
    tics: 150;                // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAINEYE1;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEYE1

   (
    sprite: SPR_BOSF;         // sprite
    frame: 32768;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWN2;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWN1

   (
    sprite: SPR_BOSF;         // sprite
    frame: 32769;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWN3;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWN2

   (
    sprite: SPR_BOSF;         // sprite
    frame: 32770;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWN4;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWN3

   (
    sprite: SPR_BOSF;         // sprite
    frame: 32771;             // frame
    tics: 3;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWN1;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWN4

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE1

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE2

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE3

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE5;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE4

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32772;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE6;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE5

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32773;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE7;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE6

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32774;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SPAWNFIRE8;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE7

   (
    sprite: SPR_FIRE;         // sprite
    frame: 32775;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SPAWNFIRE8

   (
    sprite: SPR_MISL;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAINEXPLODE2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEXPLODE1

   (
    sprite: SPR_MISL;         // sprite
    frame: 32770;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BRAINEXPLODE3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEXPLODE2

   (
    sprite: SPR_MISL;         // sprite
    frame: 32771;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINEXPLODE3

   (
    sprite: SPR_ARM1;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARM1A;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARM1

   (
    sprite: SPR_ARM1;         // sprite
    frame: 32769;             // frame
    tics: 7;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARM1;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARM1A

   (
    sprite: SPR_ARM2;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARM2A;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARM2

   (
    sprite: SPR_ARM2;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_ARM2;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ARM2A

   (
    sprite: SPR_BAR1;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BAR2;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BAR1

   (
    sprite: SPR_BAR1;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BAR1;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BAR2

   (
    sprite: SPR_BEXP;         // sprite
    frame: 32768;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BEXP2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BEXP

   (
    sprite: SPR_BEXP;         // sprite
    frame: 32769;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BEXP3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BEXP2

   (
    sprite: SPR_BEXP;         // sprite
    frame: 32770;             // frame
    tics: 5;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BEXP4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BEXP3

   (
    sprite: SPR_BEXP;         // sprite
    frame: 32771;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BEXP5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BEXP4

   (
    sprite: SPR_BEXP;         // sprite
    frame: 32772;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BEXP5

   (
    sprite: SPR_FCAN;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BBAR2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BBAR1

   (
    sprite: SPR_FCAN;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BBAR3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BBAR2

   (
    sprite: SPR_FCAN;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BBAR1;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BBAR3

   (
    sprite: SPR_BON1;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1A;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1

   (
    sprite: SPR_BON1;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1B;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1A

   (
    sprite: SPR_BON1;         // sprite
    frame: 2;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1C;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1B

   (
    sprite: SPR_BON1;         // sprite
    frame: 3;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1D;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1C

   (
    sprite: SPR_BON1;         // sprite
    frame: 2;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1E;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1D

   (
    sprite: SPR_BON1;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON1;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON1E

   (
    sprite: SPR_BON2;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2A;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2

   (
    sprite: SPR_BON2;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2B;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2A

   (
    sprite: SPR_BON2;         // sprite
    frame: 2;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2C;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2B

   (
    sprite: SPR_BON2;         // sprite
    frame: 3;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2D;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2C

   (
    sprite: SPR_BON2;         // sprite
    frame: 2;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2E;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2D

   (
    sprite: SPR_BON2;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BON2;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BON2E

   (
    sprite: SPR_BKEY;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BKEY2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BKEY

   (
    sprite: SPR_BKEY;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BKEY;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BKEY2

   (
    sprite: SPR_RKEY;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RKEY2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RKEY

   (
    sprite: SPR_RKEY;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RKEY;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RKEY2

   (
    sprite: SPR_YKEY;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_YKEY2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_YKEY

   (
    sprite: SPR_YKEY;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_YKEY;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_YKEY2

   (
    sprite: SPR_BSKU;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSKULL2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSKULL

   (
    sprite: SPR_BSKU;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BSKULL;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BSKULL2

   (
    sprite: SPR_RSKU;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RSKULL2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RSKULL

   (
    sprite: SPR_RSKU;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RSKULL;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RSKULL2

   (
    sprite: SPR_YSKU;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_YSKULL2;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_YSKULL

   (
    sprite: SPR_YSKU;         // sprite
    frame: 32769;             // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_YSKULL;      // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_YSKULL2

   (
    sprite: SPR_STIM;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_STIM

   (
    sprite: SPR_MEDI;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEDI

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL2

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL3

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL4

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL6;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL5

   (
    sprite: SPR_SOUL;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_SOUL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SOUL6

   (
    sprite: SPR_PINV;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINV2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINV

   (
    sprite: SPR_PINV;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINV3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINV2

   (
    sprite: SPR_PINV;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINV4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINV3

   (
    sprite: SPR_PINV;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINV;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINV4

   (
    sprite: SPR_PSTR;         // sprite
    frame: 32768;             // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PSTR

   (
    sprite: SPR_PINS;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINS2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINS

   (
    sprite: SPR_PINS;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINS3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINS2

   (
    sprite: SPR_PINS;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINS4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINS3

   (
    sprite: SPR_PINS;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PINS;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PINS4

   (
    sprite: SPR_MEGA;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MEGA2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEGA

   (
    sprite: SPR_MEGA;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MEGA3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEGA2

   (
    sprite: SPR_MEGA;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MEGA4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEGA3

   (
    sprite: SPR_MEGA;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_MEGA;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEGA4

   (
    sprite: SPR_SUIT;         // sprite
    frame: 32768;             // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SUIT

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP3;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP2

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP4;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP3

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32771;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP5;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP4

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP6;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP5

   (
    sprite: SPR_PMAP;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PMAP;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PMAP6

   (
    sprite: SPR_PVIS;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PVIS2;       // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PVIS

   (
    sprite: SPR_PVIS;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_PVIS;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PVIS2

   (
    sprite: SPR_CLIP;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CLIP

   (
    sprite: SPR_AMMO;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_AMMO

   (
    sprite: SPR_ROCK;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_ROCK

   (
    sprite: SPR_BROK;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BROK

   (
    sprite: SPR_CELL;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CELL

   (
    sprite: SPR_CELP;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CELP

   (
    sprite: SPR_SHEL;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SHEL

   (
    sprite: SPR_SBOX;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SBOX

   (
    sprite: SPR_BPAK;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BPAK

   (
    sprite: SPR_BFUG;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BFUG

   (
    sprite: SPR_MGUN;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MGUN

   (
    sprite: SPR_CSAW;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CSAW

   (
    sprite: SPR_LAUN;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_LAUN

   (
    sprite: SPR_PLAS;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_PLAS

   (
    sprite: SPR_SHOT;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SHOT

   (
    sprite: SPR_SGN2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SHOT2

   (
    sprite: SPR_COLU;         // sprite
    frame: 32768;             // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COLU

   (
    sprite: SPR_SMT2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_STALAG

   (
    sprite: SPR_GOR1;         // sprite
    frame: 0;                 // frame
    tics: 10;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOODYTWITCH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOODYTWITCH

   (
    sprite: SPR_GOR1;         // sprite
    frame: 1;                 // frame
    tics: 15;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOODYTWITCH3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOODYTWITCH2

   (
    sprite: SPR_GOR1;         // sprite
    frame: 2;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOODYTWITCH4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOODYTWITCH3

   (
    sprite: SPR_GOR1;         // sprite
    frame: 1;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLOODYTWITCH;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLOODYTWITCH4

   (
    sprite: SPR_PLAY;         // sprite
    frame: 13;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DEADTORSO

   (
    sprite: SPR_PLAY;         // sprite
    frame: 18;                // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DEADBOTTOM

   (
    sprite: SPR_POL2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEADSONSTICK

   (
    sprite: SPR_POL5;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GIBS

   (
    sprite: SPR_POL4;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEADONASTICK

   (
    sprite: SPR_POL3;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEADCANDLES2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEADCANDLES

   (
    sprite: SPR_POL3;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEADCANDLES;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEADCANDLES2

   (
    sprite: SPR_POL1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_DEADSTICK

   (
    sprite: SPR_POL6;         // sprite
    frame: 0;                 // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIVESTICK2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_LIVESTICK

   (
    sprite: SPR_POL6;         // sprite
    frame: 1;                 // frame
    tics: 8;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_LIVESTICK;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_LIVESTICK2

   (
    sprite: SPR_GOR2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEAT2

   (
    sprite: SPR_GOR3;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEAT3

   (
    sprite: SPR_GOR4;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEAT4

   (
    sprite: SPR_GOR5;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_MEAT5

   (
    sprite: SPR_SMIT;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_STALAGTITE

   (
    sprite: SPR_COL1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TALLGRNCOL

   (
    sprite: SPR_COL2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SHRTGRNCOL

   (
    sprite: SPR_COL3;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TALLREDCOL

   (
    sprite: SPR_COL4;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SHRTREDCOL

   (
    sprite: SPR_CAND;         // sprite
    frame: 32768;             // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CANDLESTIK

   (
    sprite: SPR_CBRA;         // sprite
    frame: 32768;             // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_CANDELABRA

   (
    sprite: SPR_COL6;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SKULLCOL

   (
    sprite: SPR_TRE1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TORCHTREE

   (
    sprite: SPR_TRE2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BIGTREE

   (
    sprite: SPR_ELEC;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECHPILLAR

   (
    sprite: SPR_CEYE;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EVILEYE2;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EVILEYE

   (
    sprite: SPR_CEYE;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EVILEYE3;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EVILEYE2

   (
    sprite: SPR_CEYE;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EVILEYE4;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EVILEYE3

   (
    sprite: SPR_CEYE;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_EVILEYE;     // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_EVILEYE4

   (
    sprite: SPR_FSKU;         // sprite
    frame: 32768;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FLOATSKULL2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FLOATSKULL

   (
    sprite: SPR_FSKU;         // sprite
    frame: 32769;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FLOATSKULL3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FLOATSKULL2

   (
    sprite: SPR_FSKU;         // sprite
    frame: 32770;             // frame
    tics: 6;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_FLOATSKULL;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_FLOATSKULL3

   (
    sprite: SPR_COL5;         // sprite
    frame: 0;                 // frame
    tics: 14;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEARTCOL2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEARTCOL

   (
    sprite: SPR_COL5;         // sprite
    frame: 1;                 // frame
    tics: 14;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_HEARTCOL;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HEARTCOL2

   (
    sprite: SPR_TBLU;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLUETORCH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLUETORCH

   (
    sprite: SPR_TBLU;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLUETORCH3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLUETORCH2

   (
    sprite: SPR_TBLU;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLUETORCH4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLUETORCH3

   (
    sprite: SPR_TBLU;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BLUETORCH;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BLUETORCH4

   (
    sprite: SPR_TGRN;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GREENTORCH2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GREENTORCH

   (
    sprite: SPR_TGRN;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GREENTORCH3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GREENTORCH2

   (
    sprite: SPR_TGRN;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GREENTORCH4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GREENTORCH3

   (
    sprite: SPR_TGRN;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GREENTORCH;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GREENTORCH4

   (
    sprite: SPR_TRED;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_REDTORCH2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_REDTORCH

   (
    sprite: SPR_TRED;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_REDTORCH3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_REDTORCH2

   (
    sprite: SPR_TRED;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_REDTORCH4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_REDTORCH3

   (
    sprite: SPR_TRED;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_REDTORCH;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_REDTORCH4

   (
    sprite: SPR_SMBT;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BTORCHSHRT2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BTORCHSHRT

   (
    sprite: SPR_SMBT;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BTORCHSHRT3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BTORCHSHRT2

   (
    sprite: SPR_SMBT;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BTORCHSHRT4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BTORCHSHRT3

   (
    sprite: SPR_SMBT;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_BTORCHSHRT;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BTORCHSHRT4

   (
    sprite: SPR_SMGT;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GTORCHSHRT2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GTORCHSHRT

   (
    sprite: SPR_SMGT;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GTORCHSHRT3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GTORCHSHRT2

   (
    sprite: SPR_SMGT;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GTORCHSHRT4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GTORCHSHRT3

   (
    sprite: SPR_SMGT;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_GTORCHSHRT;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_GTORCHSHRT4

   (
    sprite: SPR_SMRT;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RTORCHSHRT2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RTORCHSHRT

   (
    sprite: SPR_SMRT;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RTORCHSHRT3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RTORCHSHRT2

   (
    sprite: SPR_SMRT;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RTORCHSHRT4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RTORCHSHRT3

   (
    sprite: SPR_SMRT;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_RTORCHSHRT;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_RTORCHSHRT4

   (
    sprite: SPR_HDB1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGNOGUTS

   (
    sprite: SPR_HDB2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGBNOBRAIN

   (
    sprite: SPR_HDB3;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGTLOOKDN

   (
    sprite: SPR_HDB4;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGTSKULL

   (
    sprite: SPR_HDB5;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGTLOOKUP

   (
    sprite: SPR_HDB6;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_HANGTNOBRAIN

   (
    sprite: SPR_POB1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_COLONGIBS

   (
    sprite: SPR_POB2;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_SMALLPOOL

   (
    sprite: SPR_BRS1;         // sprite
    frame: 0;                 // frame
    tics: -1;                 // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_NULL;        // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_BRAINSTEM

   (
    sprite: SPR_TLMP;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECHLAMP2;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECHLAMP

   (
    sprite: SPR_TLMP;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECHLAMP3;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECHLAMP2

   (
    sprite: SPR_TLMP;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECHLAMP4;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECHLAMP3

   (
    sprite: SPR_TLMP;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECHLAMP;    // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECHLAMP4

   (
    sprite: SPR_TLP2;         // sprite
    frame: 32768;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECH2LAMP2;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECH2LAMP

   (
    sprite: SPR_TLP2;         // sprite
    frame: 32769;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECH2LAMP3;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECH2LAMP2

   (
    sprite: SPR_TLP2;         // sprite
    frame: 32770;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECH2LAMP4;  // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // misc2
   ),                         // S_TECH2LAMP3

   (
    sprite: SPR_TLP2;         // sprite
    frame: 32771;             // frame
    tics: 4;                  // tics
    action: (acp1: nil);      // action, will be set after
    nextstate: S_TECH2LAMP;   // nextstate
    misc1: 0;                 // misc1
    misc2: 0                  // S_TECH2LAMP4  // misc2
   )
  );

const
  sprnames1: array[0..ord(NUMSPRITES) - 1 + 1] of string[4] = (
    'TROO', 'SHTG', 'PUNG', 'PISG', 'PISF', 'SHTF', 'SHT2', 'CHGG', 'CHGF', 'MISG',
    'MISF', 'SAWG', 'PLSG', 'PLSF', 'BFGG', 'BFGF', 'BLUD', 'PUFF', 'BAL1', 'BAL2',
    'PLSS', 'PLSE', 'MISL', 'BFS1', 'BFE1', 'BFE2', 'TFOG', 'IFOG', 'PLAY', 'POSS',
    'SPOS', 'VILE', 'FIRE', 'FATB', 'FBXP', 'SKEL', 'MANF', 'FATT', 'CPOS', 'SARG',
    'HEAD', 'BAL7', 'BOSS', 'BOS2', 'SKUL', 'SPID', 'BSPI', 'APLS', 'APBX', 'CYBR',
    'PAIN', 'SSWV', 'KEEN', 'BBRN', 'BOSF', 'ARM1', 'ARM2', 'BAR1', 'BEXP', 'FCAN',
    'BON1', 'BON2', 'BKEY', 'RKEY', 'YKEY', 'BSKU', 'RSKU', 'YSKU', 'STIM', 'MEDI',
    'SOUL', 'PINV', 'PSTR', 'PINS', 'MEGA', 'SUIT', 'PMAP', 'PVIS', 'CLIP', 'AMMO',
    'ROCK', 'BROK', 'CELL', 'CELP', 'SHEL', 'SBOX', 'BPAK', 'BFUG', 'MGUN', 'CSAW',
    'LAUN', 'PLAS', 'SHOT', 'SGN2', 'COLU', 'SMT2', 'GOR1', 'POL2', 'POL5', 'POL4',
    'POL3', 'POL1', 'POL6', 'GOR2', 'GOR3', 'GOR4', 'GOR5', 'SMIT', 'COL1', 'COL2',
    'COL3', 'COL4', 'CAND', 'CBRA', 'COL6', 'TRE1', 'TRE2', 'ELEC', 'CEYE', 'FSKU',
    'COL5', 'TBLU', 'TGRN', 'TRED', 'SMBT', 'SMGT', 'SMRT', 'HDB1', 'HDB2', 'HDB3',
    'HDB4', 'HDB5', 'HDB6', 'POB1', 'POB2', 'BRS1', 'TLMP', 'TLP2', ''
  );

const
  mobjinfo1: array[0..ord(NUMMOBJTYPES) - 1] of mobjinfo_t = (
   (    // MT_PLAYER
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_PLAY);          // spawnstate
    spawnhealth: 100;                 // spawnhealth
    seestate: Ord(S_PLAY_RUN1);       // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 0;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_PLAY_PAIN);      // painstate
    painchance: 255;                  // painchance
    painsound: Ord(sfx_plpain);       // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_PLAY_ATK1);    // missilestate
    deathstate: Ord(S_PLAY_DIE1);     // deathstate
    xdeathstate: Ord(S_PLAY_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_pldeth);      // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_DROPOFF or MF_PICKUP or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_POSSESSED
    doomednum: 3004;                  // doomednum
    spawnstate: Ord(S_POSS_STND);     // spawnstate
    spawnhealth: 20;                  // spawnhealth
    seestate: Ord(S_POSS_RUN1);       // seestate
    seesound: Ord(sfx_posit1);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_pistol);     // attacksound
    painstate: Ord(S_POSS_PAIN);      // painstate
    painchance: 200;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_POSS_ATK1);    // missilestate
    deathstate: Ord(S_POSS_DIE1);     // deathstate
    xdeathstate: Ord(S_POSS_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_podth1);      // deathsound
    speed: 8;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_posact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_POSS_RAISE1)    // raisestate
   ),
   (    // MT_SHOTGUY
    doomednum: 9;                     // doomednum
    spawnstate: Ord(S_SPOS_STND);     // spawnstate
    spawnhealth: 30;                  // spawnhealth
    seestate: Ord(S_SPOS_RUN1);       // seestate
    seesound: Ord(sfx_posit2);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_SPOS_PAIN);      // painstate
    painchance: 170;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_SPOS_ATK1);    // missilestate
    deathstate: Ord(S_SPOS_DIE1);     // deathstate
    xdeathstate: Ord(S_SPOS_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_podth2);      // deathsound
    speed: 8;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_posact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_SPOS_RAISE1)    // raisestate
   ),
   (    // MT_VILE
    doomednum: 64;                    // doomednum
    spawnstate: Ord(S_VILE_STND);     // spawnstate
    spawnhealth: 700;                 // spawnhealth
    seestate: Ord(S_VILE_RUN1);       // seestate
    seesound: Ord(sfx_vilsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_VILE_PAIN);      // painstate
    painchance: 10;                   // painchance
    painsound: Ord(sfx_vipain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_VILE_ATK1);    // missilestate
    deathstate: Ord(S_VILE_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_vildth);      // deathsound
    speed: 15;                        // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 500;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_vilact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_FIRE
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_FIRE1);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_UNDEAD
    doomednum: 66;                    // doomednum
    spawnstate: Ord(S_SKEL_STND);     // spawnstate
    spawnhealth: 300;                 // spawnhealth
    seestate: Ord(S_SKEL_RUN1);       // seestate
    seesound: Ord(sfx_skesit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_SKEL_PAIN);      // painstate
    painchance: 100;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(S_SKEL_FIST1);    // meleestate
    missilestate: Ord(S_SKEL_MISS1);    // missilestate
    deathstate: Ord(S_SKEL_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_skedth);      // deathsound
    speed: 10;                        // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 500;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_skeact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_SKEL_RAISE1)    // raisestate
   ),
   (    // MT_TRACER
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_TRACER);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_skeatk);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_TRACEEXP1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_barexp);      // deathsound
    speed: 10 * FRACUNIT;             // speed
    radius: 11 * FRACUNIT;            // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 10;                       // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SMOKE
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_SMOKE1);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_FATSO
    doomednum: 67;                    // doomednum
    spawnstate: Ord(S_FATT_STND);     // spawnstate
    spawnhealth: 600;                 // spawnhealth
    seestate: Ord(S_FATT_RUN1);       // seestate
    seesound: Ord(sfx_mansit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_FATT_PAIN);      // painstate
    painchance: 80;                   // painchance
    painsound: Ord(sfx_mnpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_FATT_ATK1);    // missilestate
    deathstate: Ord(S_FATT_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_mandth);      // deathsound
    speed: 8;                         // speed
    radius: 48 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 1000;                       // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_posact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_FATT_RAISE1)    // raisestate
   ),
   (    // MT_FATSHOT
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_FATSHOT1);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_firsht);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_FATSHOTX1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 20 * FRACUNIT;             // speed
    radius: 6 * FRACUNIT;             // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 8;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_CHAINGUY
    doomednum: 65;                    // doomednum
    spawnstate: Ord(S_CPOS_STND);     // spawnstate
    spawnhealth: 70;                  // spawnhealth
    seestate: Ord(S_CPOS_RUN1);       // seestate
    seesound: Ord(sfx_posit2);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_CPOS_PAIN);      // painstate
    painchance: 170;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_CPOS_ATK1);    // missilestate
    deathstate: Ord(S_CPOS_DIE1);     // deathstate
    xdeathstate: Ord(S_CPOS_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_podth2);      // deathsound
    speed: 8;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_posact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_CPOS_RAISE1)    // raisestate
   ),
   (    // MT_TROOP
    doomednum: 3001;                  // doomednum
    spawnstate: Ord(S_TROO_STND);     // spawnstate
    spawnhealth: 60;                  // spawnhealth
    seestate: Ord(S_TROO_RUN1);       // seestate
    seesound: Ord(sfx_bgsit1);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_TROO_PAIN);      // painstate
    painchance: 200;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(S_TROO_ATK1);     // meleestate
    missilestate: Ord(S_TROO_ATK1);    // missilestate
    deathstate: Ord(S_TROO_DIE1);     // deathstate
    xdeathstate: Ord(S_TROO_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_bgdth1);      // deathsound
    speed: 8;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_bgact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_TROO_RAISE1)    // raisestate
   ),
   (    // MT_SERGEANT
    doomednum: 3002;                  // doomednum
    spawnstate: Ord(S_SARG_STND);     // spawnstate
    spawnhealth: 150;                 // spawnhealth
    seestate: Ord(S_SARG_RUN1);       // seestate
    seesound: Ord(sfx_sgtsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_sgtatk);     // attacksound
    painstate: Ord(S_SARG_PAIN);      // painstate
    painchance: 180;                  // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(S_SARG_ATK1);     // meleestate
    missilestate: Ord(0);             // missilestate
    deathstate: Ord(S_SARG_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_sgtdth);      // deathsound
    speed: 10;                        // speed
    radius: 30 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 400;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_SARG_RAISE1)    // raisestate
   ),
   (    // MT_SHADOWS
    doomednum: 58;                    // doomednum
    spawnstate: Ord(S_SARG_STND);     // spawnstate
    spawnhealth: 150;                 // spawnhealth
    seestate: Ord(S_SARG_RUN1);       // seestate
    seesound: Ord(sfx_sgtsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_sgtatk);     // attacksound
    painstate: Ord(S_SARG_PAIN);      // painstate
    painchance: 180;                  // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(S_SARG_ATK1);     // meleestate
    missilestate: Ord(0);             // missilestate
    deathstate: Ord(S_SARG_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_sgtdth);      // deathsound
    speed: 10;                        // speed
    radius: 30 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 400;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_SHADOW or MF_COUNTKILL;    // flags
    raisestate: Ord(S_SARG_RAISE1)    // raisestate
   ),
   (    // MT_HEAD
    doomednum: 3005;                  // doomednum
    spawnstate: Ord(S_HEAD_STND);     // spawnstate
    spawnhealth: 400;                 // spawnhealth
    seestate: Ord(S_HEAD_RUN1);       // seestate
    seesound: Ord(sfx_cacsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_HEAD_PAIN);      // painstate
    painchance: 128;                  // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_HEAD_ATK1);    // missilestate
    deathstate: Ord(S_HEAD_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_cacdth);      // deathsound
    speed: 8;                         // speed
    radius: 31 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 400;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_FLOAT or MF_NOGRAVITY or MF_COUNTKILL;    // flags
    raisestate: Ord(S_HEAD_RAISE1)    // raisestate
   ),
   (    // MT_BRUISER
    doomednum: 3003;                  // doomednum
    spawnstate: Ord(S_BOSS_STND);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_BOSS_RUN1);       // seestate
    seesound: Ord(sfx_brssit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_BOSS_PAIN);      // painstate
    painchance: 50;                   // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(S_BOSS_ATK1);     // meleestate
    missilestate: Ord(S_BOSS_ATK1);    // missilestate
    deathstate: Ord(S_BOSS_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_brsdth);      // deathsound
    speed: 8;                         // speed
    radius: 24 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 1000;                       // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_BOSS_RAISE1)    // raisestate
   ),
   (    // MT_BRUISERSHOT
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_BRBALL1);       // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_firsht);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_BRBALLX1);      // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 15 * FRACUNIT;             // speed
    radius: 6 * FRACUNIT;             // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 8;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_KNIGHT
    doomednum: 69;                    // doomednum
    spawnstate: Ord(S_BOS2_STND);     // spawnstate
    spawnhealth: 500;                 // spawnhealth
    seestate: Ord(S_BOS2_RUN1);       // seestate
    seesound: Ord(sfx_kntsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_BOS2_PAIN);      // painstate
    painchance: 50;                   // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(S_BOS2_ATK1);     // meleestate
    missilestate: Ord(S_BOS2_ATK1);    // missilestate
    deathstate: Ord(S_BOS2_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_kntdth);      // deathsound
    speed: 8;                         // speed
    radius: 24 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 1000;                       // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_BOS2_RAISE1)    // raisestate
   ),
   (    // MT_SKULL
    doomednum: 3006;                  // doomednum
    spawnstate: Ord(S_SKULL_STND);    // spawnstate
    spawnhealth: 100;                 // spawnhealth
    seestate: Ord(S_SKULL_RUN1);      // seestate
    seesound: Ord(0);                 // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_sklatk);     // attacksound
    painstate: Ord(S_SKULL_PAIN);     // painstate
    painchance: 256;                  // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_SKULL_ATK1);    // missilestate
    deathstate: Ord(S_SKULL_DIE1);    // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 8;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 50;                         // mass
    damage: 3;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_FLOAT or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SPIDER
    doomednum: 7;                     // doomednum
    spawnstate: Ord(S_SPID_STND);     // spawnstate
    spawnhealth: 3000;                // spawnhealth
    seestate: Ord(S_SPID_RUN1);       // seestate
    seesound: Ord(sfx_spisit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_shotgn);     // attacksound
    painstate: Ord(S_SPID_PAIN);      // painstate
    painchance: 40;                   // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_SPID_ATK1);    // missilestate
    deathstate: Ord(S_SPID_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_spidth);      // deathsound
    speed: 12;                        // speed
    radius: 128 * FRACUNIT;           // radius
    height: 100 * FRACUNIT;           // height
    mass: 1000;                       // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BABY
    doomednum: 68;                    // doomednum
    spawnstate: Ord(S_BSPI_STND);     // spawnstate
    spawnhealth: 500;                 // spawnhealth
    seestate: Ord(S_BSPI_SIGHT);      // seestate
    seesound: Ord(sfx_bspsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_BSPI_PAIN);      // painstate
    painchance: 128;                  // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_BSPI_ATK1);    // missilestate
    deathstate: Ord(S_BSPI_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_bspdth);      // deathsound
    speed: 12;                        // speed
    radius: 64 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 600;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_bspact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_BSPI_RAISE1)    // raisestate
   ),
   (    // MT_CYBORG
    doomednum: 16;                    // doomednum
    spawnstate: Ord(S_CYBER_STND);    // spawnstate
    spawnhealth: 4000;                // spawnhealth
    seestate: Ord(S_CYBER_RUN1);      // seestate
    seesound: Ord(sfx_cybsit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_CYBER_PAIN);     // painstate
    painchance: 20;                   // painchance
    painsound: Ord(sfx_dmpain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_CYBER_ATK1);    // missilestate
    deathstate: Ord(S_CYBER_DIE1);    // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_cybdth);      // deathsound
    speed: 16;                        // speed
    radius: 40 * FRACUNIT;            // radius
    height: 110 * FRACUNIT;           // height
    mass: 1000;                       // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_PAIN
    doomednum: 71;                    // doomednum
    spawnstate: Ord(S_PAIN_STND);     // spawnstate
    spawnhealth: 400;                 // spawnhealth
    seestate: Ord(S_PAIN_RUN1);       // seestate
    seesound: Ord(sfx_pesit);         // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_PAIN_PAIN);      // painstate
    painchance: 128;                  // painchance
    painsound: Ord(sfx_pepain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_PAIN_ATK1);    // missilestate
    deathstate: Ord(S_PAIN_DIE1);     // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_pedth);       // deathsound
    speed: 8;                         // speed
    radius: 31 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 400;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_dmact);      // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_FLOAT or MF_NOGRAVITY or MF_COUNTKILL;    // flags
    raisestate: Ord(S_PAIN_RAISE1)    // raisestate
   ),
   (    // MT_WOLFSS
    doomednum: 84;                    // doomednum
    spawnstate: Ord(S_SSWV_STND);     // spawnstate
    spawnhealth: 50;                  // spawnhealth
    seestate: Ord(S_SSWV_RUN1);       // seestate
    seesound: Ord(sfx_sssit);         // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(0);              // attacksound
    painstate: Ord(S_SSWV_PAIN);      // painstate
    painchance: 170;                  // painchance
    painsound: Ord(sfx_popain);       // painsound
    meleestate: Ord(0);               // meleestate
    missilestate: Ord(S_SSWV_ATK1);    // missilestate
    deathstate: Ord(S_SSWV_DIE1);     // deathstate
    xdeathstate: Ord(S_SSWV_XDIE1);    // xdeathstate
    deathsound: Ord(sfx_ssdth);       // deathsound
    speed: 8;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 56 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_posact);     // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_SSWV_RAISE1)    // raisestate
   ),
   (    // MT_KEEN
    doomednum: 72;                    // doomednum
    spawnstate: Ord(S_KEENSTND);      // spawnstate
    spawnhealth: 100;                 // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_KEENPAIN);       // painstate
    painchance: 256;                  // painchance
    painsound: Ord(sfx_keenpn);       // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_COMMKEEN);      // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_keendt);      // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 72 * FRACUNIT;            // height
    mass: 10000000;                   // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY or MF_SHOOTABLE or MF_COUNTKILL;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BOSSBRAIN
    doomednum: 88;                    // doomednum
    spawnstate: Ord(S_BRAIN);         // spawnstate
    spawnhealth: 250;                 // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_BRAIN_PAIN);     // painstate
    painchance: 255;                  // painchance
    painsound: Ord(sfx_bospn);        // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_BRAIN_DIE1);    // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_bosdth);      // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 10000000;                   // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SHOOTABLE;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BOSSSPIT
    doomednum: 89;                    // doomednum
    spawnstate: Ord(S_BRAINEYE);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_BRAINEYESEE);     // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 32 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOSECTOR;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BOSSTARGET
    doomednum: 87;                    // doomednum
    spawnstate: Ord(S_NULL);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 32 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOSECTOR;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SPAWNSHOT
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_SPAWN1);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_bospit);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 10 * FRACUNIT;             // speed
    radius: 6 * FRACUNIT;             // radius
    height: 32 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 3;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY or MF_NOCLIP;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SPAWNFIRE
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_SPAWNFIRE1);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BARREL
    doomednum: 2035;                  // doomednum
    spawnstate: Ord(S_BAR1);          // spawnstate
    spawnhealth: 20;                  // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_BEXP);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_barexp);      // deathsound
    speed: 0;                         // speed
    radius: 10 * FRACUNIT;            // radius
    height: 42 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SHOOTABLE or MF_NOBLOOD;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_TROOPSHOT
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_TBALL1);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_firsht);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_TBALLX1);       // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 10 * FRACUNIT;             // speed
    radius: 6 * FRACUNIT;             // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 3;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_HEADSHOT
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_RBALL1);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_firsht);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_RBALLX1);       // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 10 * FRACUNIT;             // speed
    radius: 6 * FRACUNIT;             // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 5;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_ROCKET
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_ROCKET);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_rlaunc);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_EXPLODE1);      // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_barexp);      // deathsound
    speed: 20 * FRACUNIT;             // speed
    radius: 11 * FRACUNIT;            // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 20;                       // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_PLASMA
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_PLASBALL);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_plasma);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_PLASEXP);       // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 25 * FRACUNIT;             // speed
    radius: 13 * FRACUNIT;            // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 5;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BFG
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_BFGSHOT);       // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(0);                 // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_BFGLAND);       // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_rxplod);      // deathsound
    speed: 25 * FRACUNIT;             // speed
    radius: 13 * FRACUNIT;            // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 100;                      // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_ARACHPLAZ
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_ARACH_PLAZ);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_plasma);        // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_ARACH_PLEX);    // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_firxpl);      // deathsound
    speed: 25 * FRACUNIT;             // speed
    radius: 13 * FRACUNIT;            // radius
    height: 8 * FRACUNIT;             // height
    mass: 100;                        // mass
    damage: 5;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_MISSILE or MF_DROPOFF or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_PUFF
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_PUFF1);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_BLOOD
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_BLOOD1);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP;             // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_TFOG
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_TFOG);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_IFOG
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_IFOG);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_TELEPORTMAN
    doomednum: 14;                    // doomednum
    spawnstate: Ord(S_NULL);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOSECTOR;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_EXTRABFG
    doomednum: -1;                    // doomednum
    spawnstate: Ord(S_BFGEXP);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC0
    doomednum: 2018;                  // doomednum
    spawnstate: Ord(S_ARM1);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC1
    doomednum: 2019;                  // doomednum
    spawnstate: Ord(S_ARM2);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC2
    doomednum: 2014;                  // doomednum
    spawnstate: Ord(S_BON1);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC3
    doomednum: 2015;                  // doomednum
    spawnstate: Ord(S_BON2);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC4
    doomednum: 5;                     // doomednum
    spawnstate: Ord(S_BKEY);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC5
    doomednum: 13;                    // doomednum
    spawnstate: Ord(S_RKEY);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC6
    doomednum: 6;                     // doomednum
    spawnstate: Ord(S_YKEY);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC7
    doomednum: 39;                    // doomednum
    spawnstate: Ord(S_YSKULL);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC8
    doomednum: 38;                    // doomednum
    spawnstate: Ord(S_RSKULL);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC9
    doomednum: 40;                    // doomednum
    spawnstate: Ord(S_BSKULL);        // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_NOTDMATCH;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC10
    doomednum: 2011;                  // doomednum
    spawnstate: Ord(S_STIM);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC11
    doomednum: 2012;                  // doomednum
    spawnstate: Ord(S_MEDI);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC12
    doomednum: 2013;                  // doomednum
    spawnstate: Ord(S_SOUL);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_INV
    doomednum: 2022;                  // doomednum
    spawnstate: Ord(S_PINV);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC13
    doomednum: 2023;                  // doomednum
    spawnstate: Ord(S_PSTR);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_INS
    doomednum: 2024;                  // doomednum
    spawnstate: Ord(S_PINS);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC14
    doomednum: 2025;                  // doomednum
    spawnstate: Ord(S_SUIT);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC15
    doomednum: 2026;                  // doomednum
    spawnstate: Ord(S_PMAP);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC16
    doomednum: 2045;                  // doomednum
    spawnstate: Ord(S_PVIS);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MEGA
    doomednum: 83;                    // doomednum
    spawnstate: Ord(S_MEGA);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL or MF_COUNTITEM;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_CLIP
    doomednum: 2007;                  // doomednum
    spawnstate: Ord(S_CLIP);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC17
    doomednum: 2048;                  // doomednum
    spawnstate: Ord(S_AMMO);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC18
    doomednum: 2010;                  // doomednum
    spawnstate: Ord(S_ROCK);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC19
    doomednum: 2046;                  // doomednum
    spawnstate: Ord(S_BROK);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC20
    doomednum: 2047;                  // doomednum
    spawnstate: Ord(S_CELL);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC21
    doomednum: 17;                    // doomednum
    spawnstate: Ord(S_CELP);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC22
    doomednum: 2008;                  // doomednum
    spawnstate: Ord(S_SHEL);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC23
    doomednum: 2049;                  // doomednum
    spawnstate: Ord(S_SBOX);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC24
    doomednum: 8;                     // doomednum
    spawnstate: Ord(S_BPAK);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC25
    doomednum: 2006;                  // doomednum
    spawnstate: Ord(S_BFUG);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_CHAINGUN
    doomednum: 2002;                  // doomednum
    spawnstate: Ord(S_MGUN);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC26
    doomednum: 2005;                  // doomednum
    spawnstate: Ord(S_CSAW);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC27
    doomednum: 2003;                  // doomednum
    spawnstate: Ord(S_LAUN);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC28
    doomednum: 2004;                  // doomednum
    spawnstate: Ord(S_PLAS);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SHOTGUN
    doomednum: 2001;                  // doomednum
    spawnstate: Ord(S_SHOT);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_SUPERSHOTGUN
    doomednum: 82;                    // doomednum
    spawnstate: Ord(S_SHOT2);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPECIAL;                // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC29
    doomednum: 85;                    // doomednum
    spawnstate: Ord(S_TECHLAMP);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC30
    doomednum: 86;                    // doomednum
    spawnstate: Ord(S_TECH2LAMP);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC31
    doomednum: 2028;                  // doomednum
    spawnstate: Ord(S_COLU);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC32
    doomednum: 30;                    // doomednum
    spawnstate: Ord(S_TALLGRNCOL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC33
    doomednum: 31;                    // doomednum
    spawnstate: Ord(S_SHRTGRNCOL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC34
    doomednum: 32;                    // doomednum
    spawnstate: Ord(S_TALLREDCOL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC35
    doomednum: 33;                    // doomednum
    spawnstate: Ord(S_SHRTREDCOL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC36
    doomednum: 37;                    // doomednum
    spawnstate: Ord(S_SKULLCOL);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC37
    doomednum: 36;                    // doomednum
    spawnstate: Ord(S_HEARTCOL);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC38
    doomednum: 41;                    // doomednum
    spawnstate: Ord(S_EVILEYE);       // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC39
    doomednum: 42;                    // doomednum
    spawnstate: Ord(S_FLOATSKULL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC40
    doomednum: 43;                    // doomednum
    spawnstate: Ord(S_TORCHTREE);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC41
    doomednum: 44;                    // doomednum
    spawnstate: Ord(S_BLUETORCH);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC42
    doomednum: 45;                    // doomednum
    spawnstate: Ord(S_GREENTORCH);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC43
    doomednum: 46;                    // doomednum
    spawnstate: Ord(S_REDTORCH);      // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC44
    doomednum: 55;                    // doomednum
    spawnstate: Ord(S_BTORCHSHRT);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC45
    doomednum: 56;                    // doomednum
    spawnstate: Ord(S_GTORCHSHRT);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC46
    doomednum: 57;                    // doomednum
    spawnstate: Ord(S_RTORCHSHRT);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC47
    doomednum: 47;                    // doomednum
    spawnstate: Ord(S_STALAGTITE);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC48
    doomednum: 48;                    // doomednum
    spawnstate: Ord(S_TECHPILLAR);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC49
    doomednum: 34;                    // doomednum
    spawnstate: Ord(S_CANDLESTIK);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC50
    doomednum: 35;                    // doomednum
    spawnstate: Ord(S_CANDELABRA);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC51
    doomednum: 49;                    // doomednum
    spawnstate: Ord(S_BLOODYTWITCH);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 68 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC52
    doomednum: 50;                    // doomednum
    spawnstate: Ord(S_MEAT2);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 84 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC53
    doomednum: 51;                    // doomednum
    spawnstate: Ord(S_MEAT3);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 84 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC54
    doomednum: 52;                    // doomednum
    spawnstate: Ord(S_MEAT4);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 68 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC55
    doomednum: 53;                    // doomednum
    spawnstate: Ord(S_MEAT5);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 52 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC56
    doomednum: 59;                    // doomednum
    spawnstate: Ord(S_MEAT2);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 84 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC57
    doomednum: 60;                    // doomednum
    spawnstate: Ord(S_MEAT4);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 68 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC58
    doomednum: 61;                    // doomednum
    spawnstate: Ord(S_MEAT3);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 52 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC59
    doomednum: 62;                    // doomednum
    spawnstate: Ord(S_MEAT5);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 52 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC60
    doomednum: 63;                    // doomednum
    spawnstate: Ord(S_BLOODYTWITCH);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 68 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC61
    doomednum: 22;                    // doomednum
    spawnstate: Ord(S_HEAD_DIE6);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC62
    doomednum: 15;                    // doomednum
    spawnstate: Ord(S_PLAY_DIE7);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC63
    doomednum: 18;                    // doomednum
    spawnstate: Ord(S_POSS_DIE5);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC64
    doomednum: 21;                    // doomednum
    spawnstate: Ord(S_SARG_DIE6);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC65
    doomednum: 23;                    // doomednum
    spawnstate: Ord(S_SKULL_DIE6);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC66
    doomednum: 20;                    // doomednum
    spawnstate: Ord(S_TROO_DIE5);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC67
    doomednum: 19;                    // doomednum
    spawnstate: Ord(S_SPOS_DIE5);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC68
    doomednum: 10;                    // doomednum
    spawnstate: Ord(S_PLAY_XDIE9);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC69
    doomednum: 12;                    // doomednum
    spawnstate: Ord(S_PLAY_XDIE9);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC70
    doomednum: 28;                    // doomednum
    spawnstate: Ord(S_HEADSONSTICK);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC71
    doomednum: 24;                    // doomednum
    spawnstate: Ord(S_GIBS);          // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: 0;                         // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC72
    doomednum: 27;                    // doomednum
    spawnstate: Ord(S_HEADONASTICK);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC73
    doomednum: 29;                    // doomednum
    spawnstate: Ord(S_HEADCANDLES);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC74
    doomednum: 25;                    // doomednum
    spawnstate: Ord(S_DEADSTICK);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC75
    doomednum: 26;                    // doomednum
    spawnstate: Ord(S_LIVESTICK);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC76
    doomednum: 54;                    // doomednum
    spawnstate: Ord(S_BIGTREE);       // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 32 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC77
    doomednum: 70;                    // doomednum
    spawnstate: Ord(S_BBAR1);         // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID;                  // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC78
    doomednum: 73;                    // doomednum
    spawnstate: Ord(S_HANGNOGUTS);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 88 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC79
    doomednum: 74;                    // doomednum
    spawnstate: Ord(S_HANGBNOBRAIN);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 88 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC80
    doomednum: 75;                    // doomednum
    spawnstate: Ord(S_HANGTLOOKDN);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC81
    doomednum: 76;                    // doomednum
    spawnstate: Ord(S_HANGTSKULL);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC82
    doomednum: 77;                    // doomednum
    spawnstate: Ord(S_HANGTLOOKUP);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC83
    doomednum: 78;                    // doomednum
    spawnstate: Ord(S_HANGTNOBRAIN);    // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 16 * FRACUNIT;            // radius
    height: 64 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_SOLID or MF_SPAWNCEILING or MF_NOGRAVITY;    // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC84
    doomednum: 79;                    // doomednum
    spawnstate: Ord(S_COLONGIBS);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP;             // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC85
    doomednum: 80;                    // doomednum
    spawnstate: Ord(S_SMALLPOOL);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP;             // flags
    raisestate: Ord(S_NULL)           // raisestate
   ),
   (    // MT_MISC86
    doomednum: 81;                    // doomednum
    spawnstate: Ord(S_BRAINSTEM);     // spawnstate
    spawnhealth: 1000;                // spawnhealth
    seestate: Ord(S_NULL);            // seestate
    seesound: Ord(sfx_None);          // seesound
    reactiontime: 8;                  // reactiontime
    attacksound: Ord(sfx_None);       // attacksound
    painstate: Ord(S_NULL);           // painstate
    painchance: 0;                    // painchance
    painsound: Ord(sfx_None);         // painsound
    meleestate: Ord(S_NULL);          // meleestate
    missilestate: Ord(S_NULL);        // missilestate
    deathstate: Ord(S_NULL);          // deathstate
    xdeathstate: Ord(S_NULL);         // xdeathstate
    deathsound: Ord(sfx_None);        // deathsound
    speed: 0;                         // speed
    radius: 20 * FRACUNIT;            // radius
    height: 16 * FRACUNIT;            // height
    mass: 100;                        // mass
    damage: 0;                        // damage
    activesound: Ord(sfx_None);       // activesound
    flags: MF_NOBLOCKMAP;             // flags
    raisestate: Ord(S_NULL)           // raisestate
   )
  );

procedure I_InitInfo;
begin
  states := @states1;
  sprnames := @sprnames1;
  mobjinfo := @mobjinfo1;

  states[1].action.acp1 := @A_Light0; // S_LIGHTDONE
  states[2].action.acp1 := @A_WeaponReady; // S_PUNCH
  states[3].action.acp1 := @A_Lower; // S_PUNCHDOWN
  states[4].action.acp1 := @A_Raise; // S_PUNCHUP
  states[6].action.acp1 := @A_Punch; // S_PUNCH2
  states[9].action.acp1 := @A_ReFire; // S_PUNCH5
  states[10].action.acp1 := @A_WeaponReady; // S_PISTOL
  states[11].action.acp1 := @A_Lower; // S_PISTOLDOWN
  states[12].action.acp1 := @A_Raise; // S_PISTOLUP
  states[14].action.acp1 := @A_FirePistol; // S_PISTOL2
  states[16].action.acp1 := @A_ReFire; // S_PISTOL4
  states[17].action.acp1 := @A_Light1; // S_PISTOLFLASH
  states[18].action.acp1 := @A_WeaponReady; // S_SGUN
  states[19].action.acp1 := @A_Lower; // S_SGUNDOWN
  states[20].action.acp1 := @A_Raise; // S_SGUNUP
  states[22].action.acp1 := @A_FireShotgun; // S_SGUN2
  states[29].action.acp1 := @A_ReFire; // S_SGUN9
  states[30].action.acp1 := @A_Light1; // S_SGUNFLASH1
  states[31].action.acp1 := @A_Light2; // S_SGUNFLASH2
  states[32].action.acp1 := @A_WeaponReady; // S_DSGUN
  states[33].action.acp1 := @A_Lower; // S_DSGUNDOWN
  states[34].action.acp1 := @A_Raise; // S_DSGUNUP
  states[36].action.acp1 := @A_FireShotgun2; // S_DSGUN2
  states[38].action.acp1 := @A_CheckReload; // S_DSGUN4
  states[39].action.acp1 := @A_OpenShotgun2; // S_DSGUN5
  states[41].action.acp1 := @A_LoadShotgun2; // S_DSGUN7
  states[43].action.acp1 := @A_CloseShotgun2; // S_DSGUN9
  states[44].action.acp1 := @A_ReFire; // S_DSGUN10
  states[47].action.acp1 := @A_Light1; // S_DSGUNFLASH1
  states[48].action.acp1 := @A_Light2; // S_DSGUNFLASH2
  states[49].action.acp1 := @A_WeaponReady; // S_CHAIN
  states[50].action.acp1 := @A_Lower; // S_CHAINDOWN
  states[51].action.acp1 := @A_Raise; // S_CHAINUP
  states[52].action.acp1 := @A_FireCGun; // S_CHAIN1
  states[53].action.acp1 := @A_FireCGun; // S_CHAIN2
  states[54].action.acp1 := @A_ReFire; // S_CHAIN3
  states[55].action.acp1 := @A_Light1; // S_CHAINFLASH1
  states[56].action.acp1 := @A_Light2; // S_CHAINFLASH2
  states[57].action.acp1 := @A_WeaponReady; // S_MISSILE
  states[58].action.acp1 := @A_Lower; // S_MISSILEDOWN
  states[59].action.acp1 := @A_Raise; // S_MISSILEUP
  states[60].action.acp1 := @A_GunFlash; // S_MISSILE1
  states[61].action.acp1 := @A_FireMissile; // S_MISSILE2
  states[62].action.acp1 := @A_ReFire; // S_MISSILE3
  states[63].action.acp1 := @A_Light1; // S_MISSILEFLASH1
  states[65].action.acp1 := @A_Light2; // S_MISSILEFLASH3
  states[66].action.acp1 := @A_Light2; // S_MISSILEFLASH4
  states[67].action.acp1 := @A_WeaponReady; // S_SAW
  states[68].action.acp1 := @A_WeaponReady; // S_SAWB
  states[69].action.acp1 := @A_Lower; // S_SAWDOWN
  states[70].action.acp1 := @A_Raise; // S_SAWUP
  states[71].action.acp1 := @A_Saw; // S_SAW1
  states[72].action.acp1 := @A_Saw; // S_SAW2
  states[73].action.acp1 := @A_ReFire; // S_SAW3
  states[74].action.acp1 := @A_WeaponReady; // S_PLASMA
  states[75].action.acp1 := @A_Lower; // S_PLASMADOWN
  states[76].action.acp1 := @A_Raise; // S_PLASMAUP
  states[77].action.acp1 := @A_FirePlasma; // S_PLASMA1
  states[78].action.acp1 := @A_ReFire; // S_PLASMA2
  states[79].action.acp1 := @A_Light1; // S_PLASMAFLASH1
  states[80].action.acp1 := @A_Light1; // S_PLASMAFLASH2
  states[81].action.acp1 := @A_WeaponReady; // S_BFG
  states[82].action.acp1 := @A_Lower; // S_BFGDOWN
  states[83].action.acp1 := @A_Raise; // S_BFGUP
  states[84].action.acp1 := @A_BFGsound; // S_BFG1
  states[85].action.acp1 := @A_GunFlash; // S_BFG2
  states[86].action.acp1 := @A_FireBFG; // S_BFG3
  states[87].action.acp1 := @A_ReFire; // S_BFG4
  states[88].action.acp1 := @A_Light1; // S_BFGFLASH1
  states[89].action.acp1 := @A_Light2; // S_BFGFLASH2
  states[119].action.acp1 := @A_BFGSpray; // S_BFGLAND3
  states[127].action.acp1 := @A_Explode; // S_EXPLODE1
  states[157].action.acp1 := @A_Pain; // S_PLAY_PAIN2
  states[159].action.acp1 := @A_PlayerScream; // S_PLAY_DIE2
  states[160].action.acp1 := @A_Fall; // S_PLAY_DIE3
  states[166].action.acp1 := @A_XScream; // S_PLAY_XDIE2
  states[167].action.acp1 := @A_Fall; // S_PLAY_XDIE3
  states[174].action.acp1 := @A_Look; // S_POSS_STND
  states[175].action.acp1 := @A_Look; // S_POSS_STND2
  states[176].action.acp1 := @A_Chase; // S_POSS_RUN1
  states[177].action.acp1 := @A_Chase; // S_POSS_RUN2
  states[178].action.acp1 := @A_Chase; // S_POSS_RUN3
  states[179].action.acp1 := @A_Chase; // S_POSS_RUN4
  states[180].action.acp1 := @A_Chase; // S_POSS_RUN5
  states[181].action.acp1 := @A_Chase; // S_POSS_RUN6
  states[182].action.acp1 := @A_Chase; // S_POSS_RUN7
  states[183].action.acp1 := @A_Chase; // S_POSS_RUN8
  states[184].action.acp1 := @A_FaceTarget; // S_POSS_ATK1
  states[185].action.acp1 := @A_PosAttack; // S_POSS_ATK2
  states[188].action.acp1 := @A_Pain; // S_POSS_PAIN2
  states[190].action.acp1 := @A_Scream; // S_POSS_DIE2
  states[191].action.acp1 := @A_Fall; // S_POSS_DIE3
  states[195].action.acp1 := @A_XScream; // S_POSS_XDIE2
  states[196].action.acp1 := @A_Fall; // S_POSS_XDIE3
  states[207].action.acp1 := @A_Look; // S_SPOS_STND
  states[208].action.acp1 := @A_Look; // S_SPOS_STND2
  states[209].action.acp1 := @A_Chase; // S_SPOS_RUN1
  states[210].action.acp1 := @A_Chase; // S_SPOS_RUN2
  states[211].action.acp1 := @A_Chase; // S_SPOS_RUN3
  states[212].action.acp1 := @A_Chase; // S_SPOS_RUN4
  states[213].action.acp1 := @A_Chase; // S_SPOS_RUN5
  states[214].action.acp1 := @A_Chase; // S_SPOS_RUN6
  states[215].action.acp1 := @A_Chase; // S_SPOS_RUN7
  states[216].action.acp1 := @A_Chase; // S_SPOS_RUN8
  states[217].action.acp1 := @A_FaceTarget; // S_SPOS_ATK1
  states[218].action.acp1 := @A_SPosAttack; // S_SPOS_ATK2
  states[221].action.acp1 := @A_Pain; // S_SPOS_PAIN2
  states[223].action.acp1 := @A_Scream; // S_SPOS_DIE2
  states[224].action.acp1 := @A_Fall; // S_SPOS_DIE3
  states[228].action.acp1 := @A_XScream; // S_SPOS_XDIE2
  states[229].action.acp1 := @A_Fall; // S_SPOS_XDIE3
  states[241].action.acp1 := @A_Look; // S_VILE_STND
  states[242].action.acp1 := @A_Look; // S_VILE_STND2
  states[243].action.acp1 := @A_VileChase; // S_VILE_RUN1
  states[244].action.acp1 := @A_VileChase; // S_VILE_RUN2
  states[245].action.acp1 := @A_VileChase; // S_VILE_RUN3
  states[246].action.acp1 := @A_VileChase; // S_VILE_RUN4
  states[247].action.acp1 := @A_VileChase; // S_VILE_RUN5
  states[248].action.acp1 := @A_VileChase; // S_VILE_RUN6
  states[249].action.acp1 := @A_VileChase; // S_VILE_RUN7
  states[250].action.acp1 := @A_VileChase; // S_VILE_RUN8
  states[251].action.acp1 := @A_VileChase; // S_VILE_RUN9
  states[252].action.acp1 := @A_VileChase; // S_VILE_RUN10
  states[253].action.acp1 := @A_VileChase; // S_VILE_RUN11
  states[254].action.acp1 := @A_VileChase; // S_VILE_RUN12
  states[255].action.acp1 := @A_VileStart; // S_VILE_ATK1
  states[256].action.acp1 := @A_FaceTarget; // S_VILE_ATK2
  states[257].action.acp1 := @A_VileTarget; // S_VILE_ATK3
  states[258].action.acp1 := @A_FaceTarget; // S_VILE_ATK4
  states[259].action.acp1 := @A_FaceTarget; // S_VILE_ATK5
  states[260].action.acp1 := @A_FaceTarget; // S_VILE_ATK6
  states[261].action.acp1 := @A_FaceTarget; // S_VILE_ATK7
  states[262].action.acp1 := @A_FaceTarget; // S_VILE_ATK8
  states[263].action.acp1 := @A_FaceTarget; // S_VILE_ATK9
  states[264].action.acp1 := @A_VileAttack; // S_VILE_ATK10
  states[270].action.acp1 := @A_Pain; // S_VILE_PAIN2
  states[272].action.acp1 := @A_Scream; // S_VILE_DIE2
  states[273].action.acp1 := @A_Fall; // S_VILE_DIE3
  states[281].action.acp1 := @A_StartFire; // S_FIRE1
  states[282].action.acp1 := @A_Fire; // S_FIRE2
  states[283].action.acp1 := @A_Fire; // S_FIRE3
  states[284].action.acp1 := @A_Fire; // S_FIRE4
  states[285].action.acp1 := @A_FireCrackle; // S_FIRE5
  states[286].action.acp1 := @A_Fire; // S_FIRE6
  states[287].action.acp1 := @A_Fire; // S_FIRE7
  states[288].action.acp1 := @A_Fire; // S_FIRE8
  states[289].action.acp1 := @A_Fire; // S_FIRE9
  states[290].action.acp1 := @A_Fire; // S_FIRE10
  states[291].action.acp1 := @A_Fire; // S_FIRE11
  states[292].action.acp1 := @A_Fire; // S_FIRE12
  states[293].action.acp1 := @A_Fire; // S_FIRE13
  states[294].action.acp1 := @A_Fire; // S_FIRE14
  states[295].action.acp1 := @A_Fire; // S_FIRE15
  states[296].action.acp1 := @A_Fire; // S_FIRE16
  states[297].action.acp1 := @A_Fire; // S_FIRE17
  states[298].action.acp1 := @A_Fire; // S_FIRE18
  states[299].action.acp1 := @A_FireCrackle; // S_FIRE19
  states[300].action.acp1 := @A_Fire; // S_FIRE20
  states[301].action.acp1 := @A_Fire; // S_FIRE21
  states[302].action.acp1 := @A_Fire; // S_FIRE22
  states[303].action.acp1 := @A_Fire; // S_FIRE23
  states[304].action.acp1 := @A_Fire; // S_FIRE24
  states[305].action.acp1 := @A_Fire; // S_FIRE25
  states[306].action.acp1 := @A_Fire; // S_FIRE26
  states[307].action.acp1 := @A_Fire; // S_FIRE27
  states[308].action.acp1 := @A_Fire; // S_FIRE28
  states[309].action.acp1 := @A_Fire; // S_FIRE29
  states[310].action.acp1 := @A_Fire; // S_FIRE30
  states[316].action.acp1 := @A_Tracer; // S_TRACER
  states[317].action.acp1 := @A_Tracer; // S_TRACER2
  states[321].action.acp1 := @A_Look; // S_SKEL_STND
  states[322].action.acp1 := @A_Look; // S_SKEL_STND2
  states[323].action.acp1 := @A_Chase; // S_SKEL_RUN1
  states[324].action.acp1 := @A_Chase; // S_SKEL_RUN2
  states[325].action.acp1 := @A_Chase; // S_SKEL_RUN3
  states[326].action.acp1 := @A_Chase; // S_SKEL_RUN4
  states[327].action.acp1 := @A_Chase; // S_SKEL_RUN5
  states[328].action.acp1 := @A_Chase; // S_SKEL_RUN6
  states[329].action.acp1 := @A_Chase; // S_SKEL_RUN7
  states[330].action.acp1 := @A_Chase; // S_SKEL_RUN8
  states[331].action.acp1 := @A_Chase; // S_SKEL_RUN9
  states[332].action.acp1 := @A_Chase; // S_SKEL_RUN10
  states[333].action.acp1 := @A_Chase; // S_SKEL_RUN11
  states[334].action.acp1 := @A_Chase; // S_SKEL_RUN12
  states[335].action.acp1 := @A_FaceTarget; // S_SKEL_FIST1
  states[336].action.acp1 := @A_SkelWhoosh; // S_SKEL_FIST2
  states[337].action.acp1 := @A_FaceTarget; // S_SKEL_FIST3
  states[338].action.acp1 := @A_SkelFist; // S_SKEL_FIST4
  states[339].action.acp1 := @A_FaceTarget; // S_SKEL_MISS1
  states[340].action.acp1 := @A_FaceTarget; // S_SKEL_MISS2
  states[341].action.acp1 := @A_SkelMissile; // S_SKEL_MISS3
  states[342].action.acp1 := @A_FaceTarget; // S_SKEL_MISS4
  states[344].action.acp1 := @A_Pain; // S_SKEL_PAIN2
  states[347].action.acp1 := @A_Scream; // S_SKEL_DIE3
  states[348].action.acp1 := @A_Fall; // S_SKEL_DIE4
  states[362].action.acp1 := @A_Look; // S_FATT_STND
  states[363].action.acp1 := @A_Look; // S_FATT_STND2
  states[364].action.acp1 := @A_Chase; // S_FATT_RUN1
  states[365].action.acp1 := @A_Chase; // S_FATT_RUN2
  states[366].action.acp1 := @A_Chase; // S_FATT_RUN3
  states[367].action.acp1 := @A_Chase; // S_FATT_RUN4
  states[368].action.acp1 := @A_Chase; // S_FATT_RUN5
  states[369].action.acp1 := @A_Chase; // S_FATT_RUN6
  states[370].action.acp1 := @A_Chase; // S_FATT_RUN7
  states[371].action.acp1 := @A_Chase; // S_FATT_RUN8
  states[372].action.acp1 := @A_Chase; // S_FATT_RUN9
  states[373].action.acp1 := @A_Chase; // S_FATT_RUN10
  states[374].action.acp1 := @A_Chase; // S_FATT_RUN11
  states[375].action.acp1 := @A_Chase; // S_FATT_RUN12
  states[376].action.acp1 := @A_FatRaise; // S_FATT_ATK1
  states[377].action.acp1 := @A_FatAttack1; // S_FATT_ATK2
  states[378].action.acp1 := @A_FaceTarget; // S_FATT_ATK3
  states[379].action.acp1 := @A_FaceTarget; // S_FATT_ATK4
  states[380].action.acp1 := @A_FatAttack2; // S_FATT_ATK5
  states[381].action.acp1 := @A_FaceTarget; // S_FATT_ATK6
  states[382].action.acp1 := @A_FaceTarget; // S_FATT_ATK7
  states[383].action.acp1 := @A_FatAttack3; // S_FATT_ATK8
  states[384].action.acp1 := @A_FaceTarget; // S_FATT_ATK9
  states[385].action.acp1 := @A_FaceTarget; // S_FATT_ATK10
  states[387].action.acp1 := @A_Pain; // S_FATT_PAIN2
  states[389].action.acp1 := @A_Scream; // S_FATT_DIE2
  states[390].action.acp1 := @A_Fall; // S_FATT_DIE3
  states[397].action.acp1 := @A_BossDeath; // S_FATT_DIE10
  states[406].action.acp1 := @A_Look; // S_CPOS_STND
  states[407].action.acp1 := @A_Look; // S_CPOS_STND2
  states[408].action.acp1 := @A_Chase; // S_CPOS_RUN1
  states[409].action.acp1 := @A_Chase; // S_CPOS_RUN2
  states[410].action.acp1 := @A_Chase; // S_CPOS_RUN3
  states[411].action.acp1 := @A_Chase; // S_CPOS_RUN4
  states[412].action.acp1 := @A_Chase; // S_CPOS_RUN5
  states[413].action.acp1 := @A_Chase; // S_CPOS_RUN6
  states[414].action.acp1 := @A_Chase; // S_CPOS_RUN7
  states[415].action.acp1 := @A_Chase; // S_CPOS_RUN8
  states[416].action.acp1 := @A_FaceTarget; // S_CPOS_ATK1
  states[417].action.acp1 := @A_CPosAttack; // S_CPOS_ATK2
  states[418].action.acp1 := @A_CPosAttack; // S_CPOS_ATK3
  states[419].action.acp1 := @A_CPosRefire; // S_CPOS_ATK4
  states[421].action.acp1 := @A_Pain; // S_CPOS_PAIN2
  states[423].action.acp1 := @A_Scream; // S_CPOS_DIE2
  states[424].action.acp1 := @A_Fall; // S_CPOS_DIE3
  states[430].action.acp1 := @A_XScream; // S_CPOS_XDIE2
  states[431].action.acp1 := @A_Fall; // S_CPOS_XDIE3
  states[442].action.acp1 := @A_Look; // S_TROO_STND
  states[443].action.acp1 := @A_Look; // S_TROO_STND2
  states[444].action.acp1 := @A_Chase; // S_TROO_RUN1
  states[445].action.acp1 := @A_Chase; // S_TROO_RUN2
  states[446].action.acp1 := @A_Chase; // S_TROO_RUN3
  states[447].action.acp1 := @A_Chase; // S_TROO_RUN4
  states[448].action.acp1 := @A_Chase; // S_TROO_RUN5
  states[449].action.acp1 := @A_Chase; // S_TROO_RUN6
  states[450].action.acp1 := @A_Chase; // S_TROO_RUN7
  states[451].action.acp1 := @A_Chase; // S_TROO_RUN8
  states[452].action.acp1 := @A_FaceTarget; // S_TROO_ATK1
  states[453].action.acp1 := @A_FaceTarget; // S_TROO_ATK2
  states[454].action.acp1 := @A_TroopAttack; // S_TROO_ATK3
  states[456].action.acp1 := @A_Pain; // S_TROO_PAIN2
  states[458].action.acp1 := @A_Scream; // S_TROO_DIE2
  states[460].action.acp1 := @A_Fall; // S_TROO_DIE4
  states[463].action.acp1 := @A_XScream; // S_TROO_XDIE2
  states[465].action.acp1 := @A_Fall; // S_TROO_XDIE4
  states[475].action.acp1 := @A_Look; // S_SARG_STND
  states[476].action.acp1 := @A_Look; // S_SARG_STND2
  states[477].action.acp1 := @A_Chase; // S_SARG_RUN1
  states[478].action.acp1 := @A_Chase; // S_SARG_RUN2
  states[479].action.acp1 := @A_Chase; // S_SARG_RUN3
  states[480].action.acp1 := @A_Chase; // S_SARG_RUN4
  states[481].action.acp1 := @A_Chase; // S_SARG_RUN5
  states[482].action.acp1 := @A_Chase; // S_SARG_RUN6
  states[483].action.acp1 := @A_Chase; // S_SARG_RUN7
  states[484].action.acp1 := @A_Chase; // S_SARG_RUN8
  states[485].action.acp1 := @A_FaceTarget; // S_SARG_ATK1
  states[486].action.acp1 := @A_FaceTarget; // S_SARG_ATK2
  states[487].action.acp1 := @A_SargAttack; // S_SARG_ATK3
  states[489].action.acp1 := @A_Pain; // S_SARG_PAIN2
  states[491].action.acp1 := @A_Scream; // S_SARG_DIE2
  states[493].action.acp1 := @A_Fall; // S_SARG_DIE4
  states[502].action.acp1 := @A_Look; // S_HEAD_STND
  states[503].action.acp1 := @A_Chase; // S_HEAD_RUN1
  states[504].action.acp1 := @A_FaceTarget; // S_HEAD_ATK1
  states[505].action.acp1 := @A_FaceTarget; // S_HEAD_ATK2
  states[506].action.acp1 := @A_HeadAttack; // S_HEAD_ATK3
  states[508].action.acp1 := @A_Pain; // S_HEAD_PAIN2
  states[511].action.acp1 := @A_Scream; // S_HEAD_DIE2
  states[514].action.acp1 := @A_Fall; // S_HEAD_DIE5
  states[527].action.acp1 := @A_Look; // S_BOSS_STND
  states[528].action.acp1 := @A_Look; // S_BOSS_STND2
  states[529].action.acp1 := @A_Chase; // S_BOSS_RUN1
  states[530].action.acp1 := @A_Chase; // S_BOSS_RUN2
  states[531].action.acp1 := @A_Chase; // S_BOSS_RUN3
  states[532].action.acp1 := @A_Chase; // S_BOSS_RUN4
  states[533].action.acp1 := @A_Chase; // S_BOSS_RUN5
  states[534].action.acp1 := @A_Chase; // S_BOSS_RUN6
  states[535].action.acp1 := @A_Chase; // S_BOSS_RUN7
  states[536].action.acp1 := @A_Chase; // S_BOSS_RUN8
  states[537].action.acp1 := @A_FaceTarget; // S_BOSS_ATK1
  states[538].action.acp1 := @A_FaceTarget; // S_BOSS_ATK2
  states[539].action.acp1 := @A_BruisAttack; // S_BOSS_ATK3
  states[541].action.acp1 := @A_Pain; // S_BOSS_PAIN2
  states[543].action.acp1 := @A_Scream; // S_BOSS_DIE2
  states[545].action.acp1 := @A_Fall; // S_BOSS_DIE4
  states[548].action.acp1 := @A_BossDeath; // S_BOSS_DIE7
  states[556].action.acp1 := @A_Look; // S_BOS2_STND
  states[557].action.acp1 := @A_Look; // S_BOS2_STND2
  states[558].action.acp1 := @A_Chase; // S_BOS2_RUN1
  states[559].action.acp1 := @A_Chase; // S_BOS2_RUN2
  states[560].action.acp1 := @A_Chase; // S_BOS2_RUN3
  states[561].action.acp1 := @A_Chase; // S_BOS2_RUN4
  states[562].action.acp1 := @A_Chase; // S_BOS2_RUN5
  states[563].action.acp1 := @A_Chase; // S_BOS2_RUN6
  states[564].action.acp1 := @A_Chase; // S_BOS2_RUN7
  states[565].action.acp1 := @A_Chase; // S_BOS2_RUN8
  states[566].action.acp1 := @A_FaceTarget; // S_BOS2_ATK1
  states[567].action.acp1 := @A_FaceTarget; // S_BOS2_ATK2
  states[568].action.acp1 := @A_BruisAttack; // S_BOS2_ATK3
  states[570].action.acp1 := @A_Pain; // S_BOS2_PAIN2
  states[572].action.acp1 := @A_Scream; // S_BOS2_DIE2
  states[574].action.acp1 := @A_Fall; // S_BOS2_DIE4
  states[585].action.acp1 := @A_Look; // S_SKULL_STND
  states[586].action.acp1 := @A_Look; // S_SKULL_STND2
  states[587].action.acp1 := @A_Chase; // S_SKULL_RUN1
  states[588].action.acp1 := @A_Chase; // S_SKULL_RUN2
  states[589].action.acp1 := @A_FaceTarget; // S_SKULL_ATK1
  states[590].action.acp1 := @A_SkullAttack; // S_SKULL_ATK2
  states[594].action.acp1 := @A_Pain; // S_SKULL_PAIN2
  states[596].action.acp1 := @A_Scream; // S_SKULL_DIE2
  states[598].action.acp1 := @A_Fall; // S_SKULL_DIE4
  states[601].action.acp1 := @A_Look; // S_SPID_STND
  states[602].action.acp1 := @A_Look; // S_SPID_STND2
  states[603].action.acp1 := @A_Metal; // S_SPID_RUN1
  states[604].action.acp1 := @A_Chase; // S_SPID_RUN2
  states[605].action.acp1 := @A_Chase; // S_SPID_RUN3
  states[606].action.acp1 := @A_Chase; // S_SPID_RUN4
  states[607].action.acp1 := @A_Metal; // S_SPID_RUN5
  states[608].action.acp1 := @A_Chase; // S_SPID_RUN6
  states[609].action.acp1 := @A_Chase; // S_SPID_RUN7
  states[610].action.acp1 := @A_Chase; // S_SPID_RUN8
  states[611].action.acp1 := @A_Metal; // S_SPID_RUN9
  states[612].action.acp1 := @A_Chase; // S_SPID_RUN10
  states[613].action.acp1 := @A_Chase; // S_SPID_RUN11
  states[614].action.acp1 := @A_Chase; // S_SPID_RUN12
  states[615].action.acp1 := @A_FaceTarget; // S_SPID_ATK1
  states[616].action.acp1 := @A_SPosAttack; // S_SPID_ATK2
  states[617].action.acp1 := @A_SPosAttack; // S_SPID_ATK3
  states[618].action.acp1 := @A_SpidRefire; // S_SPID_ATK4
  states[620].action.acp1 := @A_Pain; // S_SPID_PAIN2
  states[621].action.acp1 := @A_Scream; // S_SPID_DIE1
  states[622].action.acp1 := @A_Fall; // S_SPID_DIE2
  states[631].action.acp1 := @A_BossDeath; // S_SPID_DIE11
  states[632].action.acp1 := @A_Look; // S_BSPI_STND
  states[633].action.acp1 := @A_Look; // S_BSPI_STND2
  states[635].action.acp1 := @A_BabyMetal; // S_BSPI_RUN1
  states[636].action.acp1 := @A_Chase; // S_BSPI_RUN2
  states[637].action.acp1 := @A_Chase; // S_BSPI_RUN3
  states[638].action.acp1 := @A_Chase; // S_BSPI_RUN4
  states[639].action.acp1 := @A_Chase; // S_BSPI_RUN5
  states[640].action.acp1 := @A_Chase; // S_BSPI_RUN6
  states[641].action.acp1 := @A_BabyMetal; // S_BSPI_RUN7
  states[642].action.acp1 := @A_Chase; // S_BSPI_RUN8
  states[643].action.acp1 := @A_Chase; // S_BSPI_RUN9
  states[644].action.acp1 := @A_Chase; // S_BSPI_RUN10
  states[645].action.acp1 := @A_Chase; // S_BSPI_RUN11
  states[646].action.acp1 := @A_Chase; // S_BSPI_RUN12
  states[647].action.acp1 := @A_FaceTarget; // S_BSPI_ATK1
  states[648].action.acp1 := @A_BspiAttack; // S_BSPI_ATK2
  states[650].action.acp1 := @A_SpidRefire; // S_BSPI_ATK4
  states[652].action.acp1 := @A_Pain; // S_BSPI_PAIN2
  states[653].action.acp1 := @A_Scream; // S_BSPI_DIE1
  states[654].action.acp1 := @A_Fall; // S_BSPI_DIE2
  states[659].action.acp1 := @A_BossDeath; // S_BSPI_DIE7
  states[674].action.acp1 := @A_Look; // S_CYBER_STND
  states[675].action.acp1 := @A_Look; // S_CYBER_STND2
  states[676].action.acp1 := @A_Hoof; // S_CYBER_RUN1
  states[677].action.acp1 := @A_Chase; // S_CYBER_RUN2
  states[678].action.acp1 := @A_Chase; // S_CYBER_RUN3
  states[679].action.acp1 := @A_Chase; // S_CYBER_RUN4
  states[680].action.acp1 := @A_Chase; // S_CYBER_RUN5
  states[681].action.acp1 := @A_Chase; // S_CYBER_RUN6
  states[682].action.acp1 := @A_Metal; // S_CYBER_RUN7
  states[683].action.acp1 := @A_Chase; // S_CYBER_RUN8
  states[684].action.acp1 := @A_FaceTarget; // S_CYBER_ATK1
  states[685].action.acp1 := @A_CyberAttack; // S_CYBER_ATK2
  states[686].action.acp1 := @A_FaceTarget; // S_CYBER_ATK3
  states[687].action.acp1 := @A_CyberAttack; // S_CYBER_ATK4
  states[688].action.acp1 := @A_FaceTarget; // S_CYBER_ATK5
  states[689].action.acp1 := @A_CyberAttack; // S_CYBER_ATK6
  states[690].action.acp1 := @A_Pain; // S_CYBER_PAIN
  states[692].action.acp1 := @A_Scream; // S_CYBER_DIE2
  states[696].action.acp1 := @A_Fall; // S_CYBER_DIE6
  states[700].action.acp1 := @A_BossDeath; // S_CYBER_DIE10
  states[701].action.acp1 := @A_Look; // S_PAIN_STND
  states[702].action.acp1 := @A_Chase; // S_PAIN_RUN1
  states[703].action.acp1 := @A_Chase; // S_PAIN_RUN2
  states[704].action.acp1 := @A_Chase; // S_PAIN_RUN3
  states[705].action.acp1 := @A_Chase; // S_PAIN_RUN4
  states[706].action.acp1 := @A_Chase; // S_PAIN_RUN5
  states[707].action.acp1 := @A_Chase; // S_PAIN_RUN6
  states[708].action.acp1 := @A_FaceTarget; // S_PAIN_ATK1
  states[709].action.acp1 := @A_FaceTarget; // S_PAIN_ATK2
  states[710].action.acp1 := @A_FaceTarget; // S_PAIN_ATK3
  states[711].action.acp1 := @A_PainAttack; // S_PAIN_ATK4
  states[713].action.acp1 := @A_Pain; // S_PAIN_PAIN2
  states[715].action.acp1 := @A_Scream; // S_PAIN_DIE2
  states[718].action.acp1 := @A_PainDie; // S_PAIN_DIE5
  states[726].action.acp1 := @A_Look; // S_SSWV_STND
  states[727].action.acp1 := @A_Look; // S_SSWV_STND2
  states[728].action.acp1 := @A_Chase; // S_SSWV_RUN1
  states[729].action.acp1 := @A_Chase; // S_SSWV_RUN2
  states[730].action.acp1 := @A_Chase; // S_SSWV_RUN3
  states[731].action.acp1 := @A_Chase; // S_SSWV_RUN4
  states[732].action.acp1 := @A_Chase; // S_SSWV_RUN5
  states[733].action.acp1 := @A_Chase; // S_SSWV_RUN6
  states[734].action.acp1 := @A_Chase; // S_SSWV_RUN7
  states[735].action.acp1 := @A_Chase; // S_SSWV_RUN8
  states[736].action.acp1 := @A_FaceTarget; // S_SSWV_ATK1
  states[737].action.acp1 := @A_FaceTarget; // S_SSWV_ATK2
  states[738].action.acp1 := @A_CPosAttack; // S_SSWV_ATK3
  states[739].action.acp1 := @A_FaceTarget; // S_SSWV_ATK4
  states[740].action.acp1 := @A_CPosAttack; // S_SSWV_ATK5
  states[741].action.acp1 := @A_CPosRefire; // S_SSWV_ATK6
  states[743].action.acp1 := @A_Pain; // S_SSWV_PAIN2
  states[745].action.acp1 := @A_Scream; // S_SSWV_DIE2
  states[746].action.acp1 := @A_Fall; // S_SSWV_DIE3
  states[750].action.acp1 := @A_XScream; // S_SSWV_XDIE2
  states[751].action.acp1 := @A_Fall; // S_SSWV_XDIE3
  states[766].action.acp1 := @A_Scream; // S_COMMKEEN3
  states[774].action.acp1 := @A_KeenDie; // S_COMMKEEN11
  states[777].action.acp1 := @A_Pain; // S_KEENPAIN2
  states[779].action.acp1 := @A_BrainPain; // S_BRAIN_PAIN
  states[780].action.acp1 := @A_BrainScream; // S_BRAIN_DIE1
  states[783].action.acp1 := @A_BrainDie; // S_BRAIN_DIE4
  states[784].action.acp1 := @A_Look; // S_BRAINEYE
  states[785].action.acp1 := @A_BrainAwake; // S_BRAINEYESEE
  states[786].action.acp1 := @A_BrainSpit; // S_BRAINEYE1
  states[787].action.acp1 := @A_SpawnSound; // S_SPAWN1
  states[788].action.acp1 := @A_SpawnFly; // S_SPAWN2
  states[789].action.acp1 := @A_SpawnFly; // S_SPAWN3
  states[790].action.acp1 := @A_SpawnFly; // S_SPAWN4
  states[791].action.acp1 := @A_Fire; // S_SPAWNFIRE1
  states[792].action.acp1 := @A_Fire; // S_SPAWNFIRE2
  states[793].action.acp1 := @A_Fire; // S_SPAWNFIRE3
  states[794].action.acp1 := @A_Fire; // S_SPAWNFIRE4
  states[795].action.acp1 := @A_Fire; // S_SPAWNFIRE5
  states[796].action.acp1 := @A_Fire; // S_SPAWNFIRE6
  states[797].action.acp1 := @A_Fire; // S_SPAWNFIRE7
  states[798].action.acp1 := @A_Fire; // S_SPAWNFIRE8
  states[801].action.acp1 := @A_BrainExplode; // S_BRAINEXPLODE3
  states[809].action.acp1 := @A_Scream; // S_BEXP2
  states[811].action.acp1 := @A_Explode; // S_BEXP4

end;


(* Unit r_sky implementation *)


              // Needed for FRACUNIT.

//
// R_InitSkyMap
// Called whenever the view size changes.
//
procedure R_InitSkyMap;
begin
  // skyflatnum = R_FlatNumForName ( SKYFLATNAME ); // VJ is placed to g_game.pas
  skytexturemid := 100 * FRACUNIT;
end;


(* Unit s_sound implementation *)


                                                                                                                                                                                           

// Purpose?
const
  snd_prefixen: array[0..11] of char =
    ('P', 'P', 'A', 'S', 'S', 'S', 'M', 'M', 'M', 'S', 'S', 'S' );

  S_MAX_VOLUME = 127;

// when to clip out sounds
// Does not fit the large outdoor areas.
  S_CLIPPING_DIST = 1200 * $10000;

// Distance tp origin when sounds should be maxed out.
// This should relate to movement clipping resolution
// (see BLOCKMAP handling).
// Originally: (200*0x10000).
  S_CLOSE_DIST = 160 * $10000;

  S_ATTENUATOR = (S_CLIPPING_DIST - S_CLOSE_DIST) div FRACUNIT;

// Adjustable by menu.
  NORM_PITCH = 128;
  NORM_PRIORITY = 64;
  NORM_SEP = 128;

  S_PITCH_PERTURB = 1;
  S_STEREO_SWING = 96 * $10000;

// percent attenuation from front to back
  S_IFRACVOL = 30;

  NA = 0;
  S_NUMCHANNELS = 2;

type
  channel_t = record
    // sound information (if null, channel avail.)
    sfxinfo: Psfxinfo_t;

    // origin of sound
    origin: pointer;

    // handle of the sound being played
    handle: integer;
  end;
  Pchannel_t = ^channel_t;
  channel_tArray = packed array[0..$FFFF] of channel_t;
  Pchannel_tArray = ^channel_tArray;

// the set of channels available
var
  channels: Pchannel_tArray;

var
// whether songs are mus_paused
  mus_paused: boolean;

// music currently being played
  mus_playing: Pmusicinfo_t;

  nextcleanup: integer;

//
// Internals.
//
function S_getChannel(origin: pointer; sfxinfo: Psfxinfo_t): integer; forward;

function S_AdjustSoundParams(listener: Pmobj_t; source:Pmobj_t;
  vol: Pinteger; sep: Pinteger; pitch:Pinteger): boolean; forward;

procedure S_StopChannel(cnum: integer); forward;

//
// Initializes sound stuff, including volume
// Sets channels, SFX and music volume,
//  allocates channel buffer, sets S_sfx lookup.
//
procedure S_Init(sfxVolume: integer; musicVolume: integer);
var
  i: integer;
begin
  printf('S_Init: default sfx volume %d' + #13#10, [sfxVolume]);

  // Whatever these did with DMX, these are rather dummies now.
  I_SetChannels;

  S_SetSfxVolume(sfxVolume);
  // No music with Linux - another dummy.
  S_SetMusicVolume(musicVolume);

  // Allocating the internal channels for mixing
  // (the maximum numer of sounds rendered
  // simultaneously) within zone memory.
  channels := Pchannel_tArray(Z_Malloc(numChannels * SizeOf(channel_t), PU_STATIC, nil));

  // Free all channels for use
  for i := 0 to numChannels - 1 do
    channels[i].sfxinfo := nil;

  // no sounds are playing, and they are not mus_paused
  mus_paused := boolval(0);

  // Note that sounds have not been cached (yet).
  for i := 1 to Ord(NUMSFX) - 1 do
  begin
    S_sfx[i].lumpnum := -1;
    S_sfx[i].usefulness := -1;
  end;
end;

//
// Per level startup code.
// Kills playing sounds at start of level,
//  determines music if any, changes music.
//
procedure S_Start;
var
  cnum: integer;
  mnum: integer;
begin
  // kill all playing sounds at start of level
  //  (trust me - a good idea)
  for cnum := 0 to numChannels - 1 do
    if boolval(channels[cnum].sfxinfo) then
      S_StopChannel(cnum);

  // start new music for the level
  mus_paused := boolval(0);

  if gamemode = commercial then
    mnum := Ord(mus_runnin) + gamemap - 1
  else
  begin
    if gameepisode < 4 then
      mnum := Ord(mus_e1m1) + (gameepisode - 1) * 9 + gamemap - 1
    else
    begin
      case gamemap of
        1: mnum := Ord(mus_e3m4);  // American	e4m1
        2: mnum := Ord(mus_e3m2);  // Romero	e4m2
        3: mnum := Ord(mus_e3m3);  // Shawn	e4m3
        4: mnum := Ord(mus_e1m5);  // American	e4m4
        5: mnum := Ord(mus_e2m7);  // Tim 	e4m5
        6: mnum := Ord(mus_e2m4);  // Romero	e4m6
        7: mnum := Ord(mus_e2m6);  // J.Anderson	e4m7 CHIRON.WAD
        8: mnum := Ord(mus_e2m5);  // Shawn	e4m8
        9: mnum := Ord(mus_e1m9);  // Tim		e4m9
      else
        mnum := Ord(mus_e1m1); // VJ ?????
      end;
    end;
  end;

  // HACK FOR COMMERCIAL
  //  if (commercial && mnum > mus_e3m9)
  //      mnum -= mus_e3m9;

  S_ChangeMusic(mnum, true);

  nextcleanup := 15;
end;

procedure S_StartSoundAtVolume(origin_p: pointer; sfx_id: integer; volume: integer);
var
  rc: boolean;
  sep: integer;
  pitch: integer;
  priority: integer;
  sfx: Psfxinfo_t;
  cnum: integer;
  origin: Pmobj_t;
begin
  origin := Pmobj_t(origin_p);


  // Debug.
  (*fprintf( stderr,
  	   "S_StartSoundAtVolume: playing sound %d (%s)\n",
  	   sfx_id, S_sfx[sfx_id].name );*)

  // check for bogus sound #
  if (sfx_id < 1) or (sfx_id > Ord(NUMSFX)) then
    I_Error('Bad sfx #: %d', [sfx_id]);

  sfx := @S_sfx[sfx_id];

  // Initialize sound parameters
  if boolval(sfx.link) then
  begin
    pitch := sfx.pitch;
    priority := sfx.priority;
    volume := volume + sfx.volume;

    if volume < 1 then
      exit;

    if volume > snd_SfxVolume then
      volume := snd_SfxVolume;
  end
  else
  begin
    pitch := NORM_PITCH;
    priority := NORM_PRIORITY;
  end;

  // Check to see if it is audible,
  //  and if not, modify the params
  if boolval(origin) and (origin <> players[consoleplayer].mo) then
  begin
    rc := S_AdjustSoundParams(players[consoleplayer].mo, origin,
			     @volume,
			     @sep,
			     @pitch);

    if (origin.x = players[consoleplayer].mo.x) and
       (origin.y = players[consoleplayer].mo.y) then
      sep := NORM_SEP;

    if not rc then
      exit;
  end
  else
    sep := NORM_SEP;

  // hacks to vary the sfx pitches
  if (sfx_id >= Ord(sfx_sawup)) and
     (sfx_id <= Ord(sfx_sawhit)) then
  begin
    pitch := pitch + 8 - (M_Random and 15);

    if pitch < 0 then
      pitch := 0
    else if pitch > 255 then
      pitch := 255
  end
  else if (sfx_id <> Ord(sfx_itemup)) and
          (sfx_id <> Ord(sfx_tink)) then
  begin
    pitch := pitch + 16 - (M_Random and 31);

    if pitch < 0 then
      pitch := 0
    else if pitch > 255 then
      pitch := 255;
  end;

  // kill old sound
  S_StopSound(origin);

  // try to find a channel
  cnum := S_getChannel(origin, sfx);

  if cnum < 0 then
    exit;

  //
  // This is supposed to handle the loading/caching.
  // For some odd reason, the caching is done nearly
  //  each time the sound is needed?
  //

  // get lumpnum if necessary
  if sfx.lumpnum < 0 then
    sfx.lumpnum := I_GetSfxLumpNum(sfx);

//#ifndef SNDSRV
//  // cache data if necessary
//  if (!sfx->data)
//  {
//    fprintf( stderr,
//	     "S_StartSoundAtVolume: 16bit and not pre-cached - wtf?\n");
//
//    // DOS remains, 8bit handling
//    //sfx->data = (void *) W_CacheLumpNum(sfx->lumpnum, PU_MUSIC);
//    // fprintf( stderr,
//    //	     "S_StartSoundAtVolume: loading %d (lump %d) : 0x%x\n",
//    //       sfx_id, sfx->lumpnum, (int)sfx->data );
//
//  }
//#endif

  // increase the usefulness
  if sfx.usefulness < 0 then
    sfx.usefulness := 0; // VJ ??? original was := 1
  sfx.usefulness := sfx.usefulness + 1;

  // Assigns the handle to one of the channels in the
  //  mix/output buffer.
  channels[cnum].handle := I_StartSound(sfx_id, volume, sep, pitch, priority);
end;

procedure S_StartSound(origin: pointer; sfx_id: integer);
begin
//#ifdef SAWDEBUG
//    // if (sfx_id == sfx_sawful)
//    // sfx_id = sfx_itemup;
//#endif

  S_StartSoundAtVolume(origin, sfx_id, snd_SfxVolume);


    // UNUSED. We had problems, had we not?
//#ifdef SAWDEBUG
//{
//    int i;
//    int n;
//
//    static mobj_t*      last_saw_origins[10] = {1,1,1,1,1,1,1,1,1,1};
//    static int		first_saw=0;
//    static int		next_saw=0;
//
//    if (sfx_id == sfx_sawidl
//	|| sfx_id == sfx_sawful
//	|| sfx_id == sfx_sawhit)
//    {
//	for (i=first_saw;i!=next_saw;i=(i+1)%10)
//	    if (last_saw_origins[i] != origin)
//		fprintf(stderr, "old origin 0x%lx != "
//			"origin 0x%lx for sfx %d\n",
//			last_saw_origins[i],
//			origin,
//			sfx_id);
//
//	last_saw_origins[next_saw] = origin;
//	next_saw = (next_saw + 1) % 10;
//	if (next_saw == first_saw)
//	    first_saw = (first_saw + 1) % 10;
//
//	for (n=i=0; i<numChannels ; i++)
//	{
//	    if (channels[i].sfxinfo == &S_sfx[sfx_sawidl]
//		|| channels[i].sfxinfo == &S_sfx[sfx_sawful]
//		|| channels[i].sfxinfo == &S_sfx[sfx_sawhit]) n++;
//	}
//
//	if (n>1)
//	{
//	    for (i=0; i<numChannels ; i++)
//	    {
//		if (channels[i].sfxinfo == &S_sfx[sfx_sawidl]
//		    || channels[i].sfxinfo == &S_sfx[sfx_sawful]
//		    || channels[i].sfxinfo == &S_sfx[sfx_sawhit])
//		{
//		    fprintf(stderr,
//			    "chn: sfxinfo=0x%lx, origin=0x%lx, "
//			    "handle=%d\n",
//			    channels[i].sfxinfo,
//			    channels[i].origin,
//			    channels[i].handle);
//		}
//	    }
//	    fprintf(stderr, "\n");
//	}
//    }
//}
//#endif
end;

procedure S_StopSound(origin: pointer);
var
  cnum: integer;
begin
  for cnum := 0 to numChannels - 1 do
  begin
    if boolval(channels[cnum].sfxinfo) and (channels[cnum].origin = origin) then
    begin
      S_StopChannel(cnum);
      break;
    end;
  end;
end;

//
// Stop and resume music, during game PAUSE.
//
procedure S_PauseSound;
begin
  if (mus_playing <> nil) and (not mus_paused) then
  begin
    I_PauseSong(mus_playing.handle);
    mus_paused := true;
  end;
end;

procedure S_ResumeSound;
begin
  if (mus_playing <> nil) and mus_paused then
  begin
    I_ResumeSong(mus_playing.handle);
    mus_paused := false;
  end;
end;

//
// Updates music & sounds
//
procedure S_UpdateSounds(listener_p: pointer);
var
  audible: boolean;
  cnum: integer;
  volume: integer;
  sep: integer;
  pitch: integer;
  sfx: Psfxinfo_t;
  c: Pchannel_t;
  listener: Pmobj_t;
begin
  listener := Pmobj_t(listener_p);
  // Clean up unused data.
  // This is currently not done for 16bit (sounds cached static).
  // DOS 8bit remains.
    (*if (gametic > nextcleanup)
    {
	for (i=1 ; i<NUMSFX ; i++)
	{
	    if (S_sfx[i].usefulness < 1
		&& S_sfx[i].usefulness > -1)
	    {
		if (--S_sfx[i].usefulness == -1)
		{
		    Z_ChangeTag(S_sfx[i].data, PU_CACHE);
		    S_sfx[i].data = 0;
		}
	    }
	}
	nextcleanup = gametic + 15;
    }*)

  for cnum := 0 to numChannels - 1 do
  begin
    c := @channels[cnum];
    sfx := c.sfxinfo;

    if boolval(c.sfxinfo) then
    begin
      if I_SoundIsPlaying(c.handle) then
      begin
        // initialize parameters
        volume := snd_SfxVolume;
        pitch := NORM_PITCH;
        sep := NORM_SEP;

        if boolval(sfx.link) then
        begin
          pitch := sfx.pitch;
          volume := volume + sfx.volume;
          if volume < 1 then
          begin
            S_StopChannel(cnum);
            continue;
          end
          else if volume > snd_SfxVolume then
          begin
            volume := snd_SfxVolume;
          end;
        end;

        // check non-local sounds for distance clipping
        //  or modify their params
        if boolval(c.origin) and (integer(listener_p) <> integer(c.origin)) then
        begin
          audible := S_AdjustSoundParams(listener, c.origin, @volume, @sep, @pitch);

          if not audible then
          begin
            S_StopChannel(cnum);
          end
          else
            I_UpdateSoundParams(c.handle, volume, sep, pitch);
        end
      end
      else
      begin
        // if channel is allocated but sound has stopped,
        //  free it
        S_StopChannel(cnum);
      end;
    end;
  end;
  // kill music if it is a single-play && finished
  // if (	mus_playing
  //      && !I_QrySongPlaying(mus_playing->handle)
  //      && !mus_paused )
  // S_StopMusic();
end;

procedure S_SetMusicVolume(volume: integer);
begin
  if (volume < 0) or (volume > 15) then
    I_Error('S_SetMusicVolume(): Attempt to set music volume at %d', [volume]);

//  I_SetMusicVolume(15); // VJ ????
  I_SetMusicVolume(volume);
  snd_MusicVolume := volume;
end;

procedure S_SetSfxVolume(volume: integer);
begin
  if (volume < 0) or (volume > 127) then
    I_Error('S_SetSfxVolume(): Attempt to set sfx volume at %d', [volume]);

  snd_SfxVolume := volume;
end;

//
// Starts some music with the music id found in sounds.h.
//
procedure S_StartMusic(music_id: integer);
begin
  S_ChangeMusic(music_id, false);
end;

procedure S_ChangeMusic(musicnum: integer; looping: boolean);
var
  music: Pmusicinfo_t;
  namebuf: char8_t;
begin
  if (musicnum <= Ord(mus_None)) or
     (musicnum >= Ord(NUMMUSIC)) then
    I_Error('S_ChangeMusic(): Bad music number %d', [musicnum]);

  music := @S_music[musicnum];

  if mus_playing = music then
    exit;

  // shutdown old music
  S_StopMusic;

  // get lumpnum if neccessary
  if not boolval(music.lumpnum) then
  begin
    namebuf := stringtochar8('d_' + music.name);
    music.lumpnum := W_GetNumForName(namebuf);
  end;

  // load & register it
  music.data := W_CacheLumpNum(music.lumpnum, PU_MUSIC);
  music.handle := I_RegisterSong(music.data, W_LumpLength(music.lumpnum));

  // play it
  I_PlaySong(music.handle, looping);

  mus_playing := music;
end;

procedure S_StopMusic;
begin
  if mus_playing <> nil then
  begin
    if mus_paused then
      I_ResumeSong(mus_playing.handle);

    I_UnRegisterSong(mus_playing.handle);
    Z_ChangeTag(mus_playing.data, PU_CACHE);

    mus_playing.data := nil;
    mus_playing := nil;
  end;
end;

procedure S_StopChannel(cnum: integer);
var
  i: integer;
  c: Pchannel_t;
begin
  c := @channels[cnum];

  if boolval(c.sfxinfo) then
  begin
    // stop the sound playing
    if I_SoundIsPlaying(c.handle) then
    begin
//#ifdef SAWDEBUG
//	    if (c->sfxinfo == &S_sfx[sfx_sawful])
//		fprintf(stderr, "stopped\n");
//#endif
      I_StopSound(c.handle);
    end;

    // check to see
    //  if other channels are playing the sound
    for i := 0 to numChannels - 1 do
    begin
	    if (cnum <> i) and
         (c.sfxinfo = channels[i].sfxinfo) then
        break;
    end;

    // degrade usefulness of sound data
    c.sfxinfo.usefulness := c.sfxinfo.usefulness - 1;

    c.sfxinfo := nil;
  end;
end;

//
// Changes volume, stereo-separation, and pitch variables
//  from the norm of a sound effect to be played.
// If the sound is not audible, returns a 0.
// Otherwise, modifies parameters and returns 1.
//
function S_AdjustSoundParams(listener: Pmobj_t; source:Pmobj_t;
  vol: Pinteger; sep: Pinteger; pitch:Pinteger): boolean;
var
  approx_dist: fixed_t;
  adx: fixed_t;
  ady: fixed_t;
  angle: angle_t;
begin
  // calculate the distance to sound origin
  //  and clip it if necessary
  adx := abs(listener.x - source.x);
  ady := abs(listener.y - source.y);

  // From _GG1_ p.428. Appox. eucledian distance fast.
  approx_dist := adx + ady - _SHR(decide(adx < ady, adx, ady), 1);

  if (gamemap <> 8) and
     (approx_dist > S_CLIPPING_DIST) then
  begin
    result := false;
    exit;
  end;

  // angle of source to listener
  angle := R_PointToAngle2(listener.x, listener.y, source.x, source.y);

  if angle > listener.angle then
    angle := angle - listener.angle
  else
    angle := angle + ($ffffffff - listener.angle);

  angle := _SHRW(angle, ANGLETOFINESHIFT);

  // stereo separation
  sep^ := 128 - (FixedMul(S_STEREO_SWING, finesine[angle]) div FRACUNIT);

  // volume calculation
  if approx_dist < S_CLOSE_DIST then
  begin
    vol^ := snd_SfxVolume;
  end
  else if gamemap = 8 then
  begin
    if approx_dist > S_CLIPPING_DIST then
      approx_dist := S_CLIPPING_DIST;

    vol^ := 15 + ((snd_SfxVolume - 15) *
      ((S_CLIPPING_DIST - approx_dist) div FRACUNIT)) div S_ATTENUATOR;
  end
  else
  begin
    // distance effect
    vol^ := (snd_SfxVolume * ((S_CLIPPING_DIST - approx_dist) div FRACUNIT)) div
              S_ATTENUATOR;
  end;

  result := vol^ > 0;
end;

//
// S_getChannel :
//   If none available, return -1.  Otherwise channel #.
//
function S_getChannel(origin: pointer; sfxinfo: Psfxinfo_t): integer;
var
  // channel number to use
  cnum: integer;
  c: Pchannel_t;
begin
  // Find an open channel
  cnum := 0;
  while cnum < numChannels do
  begin
    if not boolval(channels[cnum].sfxinfo) then
      break
    else if boolval(origin) and (channels[cnum].origin = origin) then
    begin
      S_StopChannel(cnum);
      break;
    end;
    inc(cnum);
  end;

  // None available
  if cnum = numChannels then
  begin
    // Look for lower priority
    cnum := 0;
    while cnum < numChannels do
    begin
	    if channels[cnum].sfxinfo.priority >= sfxinfo.priority then
        break;
      inc(cnum);
    end;

    if cnum = numChannels then
    begin
      // FUCK!  No lower priority.  Sorry, Charlie.
      result := -1;
      exit;
    end
    else
    begin
      // Otherwise, kick out lower priority.
      S_StopChannel(cnum);
    end;
  end;

  c := @channels[cnum];

  // channel is decided to be cnum.
  c.sfxinfo := sfxinfo;
  c.origin := origin;

  result := cnum;
end;






(* Unit m_misc implementation *)


                                                                                                                                                                                                                     

function M_WriteFile(const name: string; source: pointer; length: integer): boolean;
var
  handle: file;
  count: integer;
begin
  assign(handle, name);
  {$I-}
  rewrite(handle, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    result := false;
    exit;
  end;

  BlockWrite(handle, source^, length, count);
  closefile(handle);

  result := count > 0;
end;

function M_ReadFile(const name: string; var buffer: Pointer): integer;
var
  handle: file;
  count: integer;
begin
  assign(handle, name);
  {$I-}
  reset(handle, 1);
  {$I+}
  if IOResult <> 0 then
    I_Error('M_ReadFile(): Could not read file %s', [name]);

  result := FileSize(handle);
  buffer := Z_Malloc(result, PU_STATIC, nil);
  BlockRead(handle, buffer^, result, count);
  closefile(handle);

  if count < result then
    I_Error('M_ReadFile(): Could not read file %s', [name]);
end;

type
  ttype_t = (tString, tInteger, tBoolean);

  default_t = record
    name: string;
    location: pointer;
    defaultsvalue: string;
    defaultivalue: integer;
    defaultbvalue: boolean;
    _type: ttype_t;
  end;

const
  NUMDEFAULTS = 43;

  defaults: array[0..NUMDEFAULTS - 1] of default_t = (
    (name: 'mouse_sensitivity';
     location: @mouseSensitivity;
     defaultivalue: 5;
     _type: tInteger),

    (name: 'screenwidth';
     location: @SCREENWIDTH;
     defaultivalue: 640;
     _type: tInteger),
    (name: 'screenheight';
     location: @SCREENHEIGHT;
     defaultivalue: 480;
     _type: tInteger),
    (name: 'fullscreen';
     location: @fullscreen;
     defaultbvalue: true;
     _type: tBoolean),

    (name: 'sfx_volume';
     location: @snd_SfxVolume;
     defaultivalue: 8;
     _type: tInteger),
    (name: 'music_volume';
     location: @snd_MusicVolume;
     defaultivalue: 8;
     _type: tInteger),

    (name: 'show_messages';
     location: @showMessages;
     defaultivalue: 1;
     _type: tInteger),

    (name: 'key_right';
     location: @key_right;
     defaultivalue: KEY_RIGHTARROW;
     _type: tInteger),
    (name: 'key_left';
     location: @key_left;
     defaultivalue: KEY_LEFTARROW;
     _type: tInteger),
    (name: 'key_up';
     location: @key_up;
     defaultivalue: KEY_UPARROW;
     _type: tInteger),
    (name: 'key_down';
     location: @key_down;
     defaultivalue: KEY_DOWNARROW;
     _type: tInteger),
    (name: 'key_strafeleft';
     location: @key_strafeleft;
     defaultivalue: Ord(',');
     _type: tInteger),
    (name: 'key_straferight';
     location: @key_straferight;
     defaultivalue: Ord('.');
     _type: tInteger),
    (name: 'key_fire';
     location: @key_fire;
     defaultivalue: KEY_RCTRL;
     _type: tInteger),
    (name: 'key_use';
     location: @key_use;
     defaultivalue: Ord(' ');
     _type: tInteger),
    (name: 'key_strafe';
     location: @key_strafe;
     defaultivalue: KEY_RALT;
     _type: tInteger),
    (name: 'key_speed';
     location: @key_speed;
     defaultivalue: KEY_RSHIFT;
     _type: tInteger),
    (name: 'key_lookup';
     location: @key_lookup;
     defaultivalue: KEY_PAGEDOWN;
     _type: tInteger),
    (name: 'key_lookdown';
     location: @key_lookdown;
     defaultivalue: KEY_PAGEUP;
     _type: tInteger),
    (name: 'key_lookcenter';
     location: @key_lookcenter;
     defaultivalue: KEY_INS;
     _type: tInteger),


    (name: 'use_mouse';
     location: @usemouse;
     defaultbvalue: false;
     _type: tBoolean),
    (name: 'mouseb_fire';
     location: @mousebfire;
     defaultivalue: 0;
     _type: tInteger),
    (name: 'mouseb_strafe';
     location: @mousebstrafe;
     defaultivalue: 1;
     _type: tInteger),
    (name: 'mouseb_forward';
     location: @mousebforward;
     defaultivalue: 2;
     _type: tInteger),

    (name: 'use_joystick';
     location: @usejoystick;
     defaultbvalue: false;
     _type: tBoolean),
    (name: 'joyb_fire';
     location: @joybfire;
     defaultivalue: 0;
     _type: tInteger),
    (name: 'joyb_strafe';
     location: @joybstrafe;
     defaultivalue: 1;
     _type: tInteger),
    (name: 'joyb_use';
     location: @joybuse;
     defaultivalue: 3;
     _type: tInteger),
    (name: 'joyb_speed';
     location: @joybspeed;
     defaultivalue: 2;
     _type: tInteger),

    (name: 'screenblocks';
     location: @screenblocks;
     defaultivalue: 9;
     _type: tInteger),
    (name: 'detaillevel';
     location: @detailLevel;
     defaultivalue: 0;
     _type: tInteger),

    (name: 'snd_channels';
     location: @numChannels;
     defaultivalue: 3;
     _type: tInteger),

    (name: 'usegamma';
     location: @usegamma;
     defaultivalue: 0;
     _type: tInteger),

    (name: 'chatmacro0';
     location: @chat_macros[0];
     defaultsvalue: HUSTR_CHATMACRO0;
     _type: tString),
    (name: 'chatmacro1';
     location: @chat_macros[1];
     defaultsvalue: HUSTR_CHATMACRO1;
     _type: tString),
    (name: 'chatmacro2';
     location: @chat_macros[2];
     defaultsvalue: HUSTR_CHATMACRO2;
     _type: tString),
    (name: 'chatmacro3';
     location: @chat_macros[3];
     defaultsvalue: HUSTR_CHATMACRO3;
     _type: tString),
    (name: 'chatmacro4';
     location: @chat_macros[4];
     defaultsvalue: HUSTR_CHATMACRO4;
     _type: tString),
    (name: 'chatmacro5';
     location: @chat_macros[5];
     defaultsvalue: HUSTR_CHATMACRO5;
     _type: tString),
    (name: 'chatmacro6';
     location: @chat_macros[6];
     defaultsvalue: HUSTR_CHATMACRO6;
     _type: tString),
    (name: 'chatmacro7';
     location: @chat_macros[7];
     defaultsvalue: HUSTR_CHATMACRO7;
     _type: tString),
    (name: 'chatmacro8';
     location: @chat_macros[8];
     defaultsvalue: HUSTR_CHATMACRO8;
     _type: tString),
    (name: 'chatmacro9';
     location: @chat_macros[9];
     defaultsvalue: HUSTR_CHATMACRO9;
     _type: tString)
  );

var
  defaultfile: string;

procedure M_SaveDefaults;
var
  i: integer;
  s: TStringList;
begin
  s := TStringList.Create;
  try
    for i := 0 to NUMDEFAULTS - 1 do
      if defaults[i]._type = tInteger then
        s.Add(defaults[i].name + '=' + itoa(PInteger(defaults[i].location)^))
      else if defaults[i]._type = tString then
        s.Add(defaults[i].name + '=' + PString(defaults[i].location)^)
      else if defaults[i]._type = tBoolean then
        s.Add(defaults[i].name + '=' + itoa(IntVal(PBoolean(defaults[i].location)^)));

    s.SaveToFile(defaultfile);

  finally
    s.Free;
  end;
end;

procedure M_LoadDefaults;
var
  i: integer;
  j: integer;
  idx: integer;
  s: TStringList;
  n: string;
begin
  // set everything to base values
  for i := 0 to NUMDEFAULTS - 1 do
    if defaults[i]._type = tInteger then
      PInteger(defaults[i].location)^ := defaults[i].defaultivalue
    else if defaults[i]._type = tBoolean then
      PBoolean(defaults[i].location)^ := defaults[i].defaultbvalue
    else if defaults[i]._type = tString then
      PString(defaults[i].location)^ := defaults[i].defaultsvalue;

  // check for a custom default file
  i := M_CheckParm('-config');
  if (i > 0) and (i < myargc - 1) then
  begin
    defaultfile := myargv[i + 1];
    printf(' default file: %s' + #13#10, [defaultfile]);
  end
  else
    defaultfile := basedefault;

  s := TStringList.Create;
  try
    // read the file in, overriding any set defaults
    if fexists(defaultfile) then
      s.LoadFromFile(defaultfile);

    for i := 0 to s.Count - 1 do
    begin
      idx := -1;
      n := s.Names[i];
      for j := 0 to NUMDEFAULTS - 1 do
        if defaults[j].name = n then
        begin
          idx := j;
          break;
        end;

      if idx > -1 then
      begin
        if defaults[idx]._type = tInteger then
          PInteger(defaults[idx].location)^ := atoi(s.Values[n])
        else if defaults[idx]._type = tBoolean then
          PBoolean(defaults[idx].location)^ := boolval(atoi(s.Values[n]))
        else if defaults[idx]._type = tString then
          PString(defaults[idx].location)^ := s.Values[n];
      end;
    end;

  finally
    s.Free;
  end;
end;


type
  TargaHeader = record
    id_length, colormap_type, image_type: byte;
    colormap_index, colormap_length: word;
    colormap_size: byte;
    x_origin, y_origin, width, height: word;
    pixel_size, attributes: byte;
  end;

//
// M_ScreenShot
//
procedure M_ScreenShot;
var
  buffer: PByteArray;
  tganame: string;
  i: integer;
  buf: array[0..MAXWIDTH - 1] of LongWord;
  src: PByteArray;
begin
  tganame := 'DOOM00.tga';
//
// find a file name to save it to
//

  i := 0;
  while i <= 99 do
  begin
    tganame[5] := Chr((i div 10) + Ord('0'));
    tganame[6] := Chr((i mod 10) + Ord('0'));
    if not fexists(tganame) then
      break;  // file doesn't exist
    inc(i);
  end;
  if i = 100 then
  begin
    players[consoleplayer]._message := 'Couldn''t create a screenshot';
    exit;
  end;

  buffer := malloc(SCREENWIDTH * SCREENHEIGHT * 4 + 18);
  memset(buffer, 0, 18);
  buffer[2] := 2;    // uncompressed type
  buffer[12] := SCREENWIDTH and 255;
  buffer[13] := SCREENWIDTH div 256;
  buffer[14] := SCREENHEIGHT and 255;
  buffer[15] := SCREENHEIGHT div 256;
  buffer[16] := 32;  // pixel size

  src := @buffer[18];

  I_ReadScreen32bit(src);

  for i := 0 to (SCREENHEIGHT div 2) - 1 do
  begin
    memcpy(@buf, @src[i * SCREENWIDTH * 4], SCREENWIDTH * 4);
    memcpy(@src[i * SCREENWIDTH * 4], @src[(SCREENHEIGHT - i) * SCREENWIDTH * 4], SCREENWIDTH * 4);
    memcpy(@src[(SCREENHEIGHT - i) * SCREENWIDTH * 4], @buf, SCREENWIDTH * 4);
  end;

  if not M_WriteFile(tganame, buffer, SCREENWIDTH * SCREENHEIGHT * 4 + 18) then
    players[consoleplayer]._message := 'Couldn''t create a screenshot'
  else
    players[consoleplayer]._message := 'screen shot';

  freemem(buffer);

end;



(* Unit d_englsh implementation *)




(* Unit i_main implementation *)


                                                                    

function WindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT; stdcall; export;
begin
  if not I_GameFinished then
  begin
    case Msg of
      WM_SETCURSOR:
        begin
          SetCursor(0);
        end;
      WM_SYSCOMMAND:
        begin
          if (wParam = SC_SCREENSAVE) then
          begin
            result := 0;
            exit;
          end;
        end;
      WM_ACTIVATE:
        begin
          I_SynchronizeInput(wparam <> WA_INACTIVE);
        end;
      WM_CLOSE:
        begin
          result := 0; // Preserve closing window by pressing Alt + F4
          exit;
        end;
      WM_DESTROY:
        begin
          I_Destroy(0);
        end;
    end;
  end;

  result := DefWindowProc(hWnd, Msg, WParam, LParam);
end;

procedure DoomMain;
var
  WindowClass: TWndClass;
begin
  ZeroMemory(WindowClass, SizeOf(WindowClass));
  WindowClass.lpfnWndProc := @WindowProc;
  WindowClass.hbrBackground := GetStockObject(WHITE_BRUSH);
  WindowClass.lpszClassName := 'Doom32';
  if HPrevInst = 0 then
  begin
    WindowClass.hInstance := HInstance;
    WindowClass.hIcon := LoadIcon(HInstance, 'MAINICON');
    WindowClass.hCursor := LoadCursor(0, nil);
    if RegisterClass(WindowClass) = 0 then
      halt(1);
  end;
  hMainWnd := CreateWindowEx(
    CS_HREDRAW or CS_VREDRAW,
    WindowClass.lpszClassName,
    AppTitle,
    WS_OVERLAPPED,
    0,
    0,
    0,
    0,
    0,
    0,
    HInstance,
    nil);
  ShowWindow(hMainWnd, SW_SHOW{CmdShow});
  UpdateWindow(hMainWnd);
  D_DoomMain;
end;


(* Unit sounds implementation *)



(* Unit i_sound implementation *)


                                                                                                                                 

// The number of internal mixing channels,
//  the samples calculated for each mixing step,
//  the size of the 16bit, 2 hardware channel (stereo)
//  mixing buffer, and the samplerate of the raw data.

// Needed for calling the actual sound output.
const
  NUM_CHANNELS = 16;

type
  LPDIRECTSOUND = IDirectSound;
  LPDIRECTSOUNDBUFFER = IDirectSoundBuffer;

var
  pDS: LPDIRECTSOUND;
  pDSBPrimary: LPDIRECTSOUNDBUFFER;
  HandleCount: integer;

  SampleFormat: TWAVEFORMATEX;

// The actual lengths of all sound effects.
  SoundLengths: array[0..Ord(NUMSFX) - 1] of integer;
  HighSound: boolean = false;

// The sound in channel handles,
//  determined on registration,
//  might be used to unregister/stop/modify,
//  currently unused.
  channelhandles: array[0..NUM_CHANNELS - 1] of integer;

// SFX id of the playing sound effect.
// Used to catch duplicates (like chainsaw).
  channelids: array[0..NUM_CHANNELS - 1] of integer;

//actual data buffers
  ChannelBuffers: array[0..NUM_CHANNELS - 1] of LPDIRECTSOUNDBUFFER;
  ChannelActive: packed array[0..NUM_CHANNELS - 1] of boolean;

//
// Retrieve the raw data lump index
//  for a given SFX name.
//
function I_GetSfxLumpNum(sfxinfo: Psfxinfo_t): integer;
var
  namebuf: string;
begin
  sprintf(namebuf, 'ds%s', [sfxinfo.name]);
  result := W_GetNumForName(namebuf);
end;

// This function loads the sound data from the WAD lump,
//  for single sound.
//
procedure CacheSFX(sfxid: integer);
var
  name: string;
  sfx: Psfxinfo_t;
begin
  sfx := @S_sfx[sfxid];
  if sfx.data <> nil then
    exit;
  // Get the sound data from the WAD, allocate lump
  //  in zone memory.
  sprintf(name, 'ds%s', [sfx.name]);

  // Now, there is a severe problem with the
  //  sound handling, in it is not (yet/anymore)
  //  gamemode aware. That means, sounds from
  //  DOOM II will be requested even with DOOM
  //  shareware.
  // The sound list is wired into sounds.c,
  //  which sets the external variable.
  // I do not do runtime patches to that
  //  variable. Instead, we will use a
  //  default sound for replacement.
  if W_CheckNumForName(name) = -1 then
    sfx.lumpnum := W_GetNumForName('dspistol')
  else
    sfx.lumpnum := W_GetNumForName(name);

  SoundLengths[sfxid] := W_LumpLength(sfx.lumpnum);

  sfx.data := W_CacheLumpNum(sfx.lumpnum, PU_STATIC);
end;

//
// SFX API
// Note: this was called by S_Init.
// However, whatever they did in the
// old DPMS based DOS version, this
// were simply dummies in the Linux
// version.
// See soundserver initdata().
//
procedure I_SetChannels;
begin
end;

procedure I_SetSfxVolume(volume: integer);
begin
  // Identical to DOS.
  // Basically, this should propagate
  //  the menu/config file setting
  //  to the state variable used in
  //  the mixing.
  snd_SfxVolume := volume;
end;

function I_ChannelPlaying(channel: integer): boolean;
var
  status: LongWord;
begin
  if pDS = nil then
  begin
    result := false;
    exit;
  end;

  if not boolval(pointer(ChannelBuffers[channel])) then
  begin
    result := false;
    exit;
  end;

  if not ChannelActive[channel] then
  begin
    result := false;
    exit;
  end;

  ChannelBuffers[channel].GetStatus(status);
  if boolval(status and DSBSTATUS_PLAYING) then
    result := true
  else
  begin
    ChannelActive[channel] := false;
    result := false;
  end;
end;

procedure I_KillChannel(channel: integer);
begin
  if pDS <> nil then
  begin
    if ChannelBuffers[channel] <> nil then
    begin
      ChannelBuffers[channel].Stop;
      ChannelBuffers[channel]._Release;
    end;
  end;
end;

const
  vulumetrans: array[0..15] of integer = (
      0,  96, 128, 168, 186, 200, 212, 222,
    230, 237, 243, 248, 250, 252, 254, 256
  );

  vulumetransshift = 8;

function I_RestartChannel(channel: integer; volume: integer): integer;
var
  dsb: LPDIRECTSOUNDBUFFER;
begin
  if not boolval(pointer(pDS)) then
  begin
    result := HandleCount;
    inc(HandleCount);
    exit;
  end;

  ChannelActive[channel] := true;
  dsb := ChannelBuffers[channel];
  if not boolval(pointer(dsb)) then
    I_Error('I_RestartChannel(): Restarting dead sound at channel %d', [channel]);

  dsb.Stop;
  dsb.SetCurrentPosition(0);
  dsb.SetVolume(DSBVOLUME_MIN + _SHR((DSBVOLUME_MAX - DSBVOLUME_MIN) * (vulumetrans[volume] + 1), vulumetransshift));
  dsb.Play(0, 0, 0);
  channelhandles[channel] := HandleCount;
  result := HandleCount;
  inc(HandleCount);
end;

//
// Starting a sound means adding it
//  to the current list of active sounds
//  in the internal channels.
// As the SFX info struct contains
//  e.g. a pointer to the raw data,
//  it is ignored.
// As our sound handling does not handle
//  priority, it is ignored.
// Pitching (that is, increased speed of playback)
//  is set, but currently not used by mixing.
//
function I_StartSound(id: integer; vol: integer; sep: integer;
  pitch: integer; priority: integer): integer;
var
  channel: integer;
  dsb: LPDIRECTSOUNDBUFFER;
  hres: HRESULT;
  dsbd: DSBUFFERDESC;
  oldchannel: integer;
  oldhandle: integer;
  freechannel: integer;
  p: pointer;
  p2: pointer;
  s: LongWord;
  s2: LongWord;

  procedure I_ErrorStartSound(const procname: string);
  begin
    I_Error('I_StartSound(): %s failed, result = %d', [procname, hres]);
  end;

begin
  if not boolval(pointer(pDS)) then
  begin
    result := HandleCount;
    inc(HandleCount);
    exit;
  end;

  oldhandle := 0;
  oldchannel := 0;
  freechannel := NUM_CHANNELS;
  for channel := 0 to NUM_CHANNELS - 1 do
  begin
		if boolval(pointer(ChannelBuffers[channel])) then
    begin
      if (channelids[channel] = id) and (not boolval(pointer(I_ChannelPlaying(channel)))) then
      begin
        result := I_RestartChannel(channel, vol);
        exit;
      end;
      if HandleCount - channelhandles[channel] > oldhandle then
      begin
        oldhandle := HandleCount - channelhandles[channel];
        oldchannel := channel;
      end
    end
    else
      freechannel := channel;
  end;

  if boolval(freechannel) then
    channel := freechannel
  else
    channel := oldchannel;
  CacheSFX(id);
  ZeroMemory(dsbd, SizeOf(DSBUFFERDESC));
  dsbd.dwSize := Sizeof(DSBUFFERDESC);
    //dsbd.dwFlags=DSBCAPS_CTRLDEFAULT|DSBCAPS_GETCURRENTPOSITION2|DSBCAPS_STATIC;
  dsbd.dwFlags := DSBCAPS_CTRLVOLUME or DSBCAPS_CTRLFREQUENCY or
                  DSBCAPS_CTRLPAN or DSBCAPS_GETCURRENTPOSITION2 or DSBCAPS_STATIC;
  dsbd.dwBufferBytes := SoundLengths[id];
  dsbd.lpwfxFormat := @SampleFormat;

  hres := pDS.CreateSoundBuffer(dsbd, dsb, nil);
  if hres <> DS_OK then
    I_ErrorStartSound('CreateSoundBuffer()');

  hres := dsb.Lock(0, SoundLengths[id] - 8, p, s, p2, s2, 0);
  if hres <> DS_OK then
    I_ErrorStartSound('SoundBuffer.Lock()');

  memcpy(p, pointer(integer(S_sfx[id].data) + 8), s);
  hres := dsb.Unlock(p, s, p2, s2);
  if hres <> DS_OK then
    I_ErrorStartSound('SoundBuffer.Unlock()');

  ChannelBuffers[channel] := dsb;
  channelids[channel] := id;
  result := I_RestartChannel(channel, vol);
end;

procedure I_StopSound(handle: integer);
var
  channel: integer;
begin
  if not boolval(pointer(pDS)) then
    exit;

  for channel := 0 to NUM_CHANNELS - 1 do
  begin
    if I_ChannelPlaying(channel) and (channelhandles[channel]=handle) then
    begin
      ChannelBuffers[channel].Stop;
      ChannelActive[channel] := false;
    end;
  end;
end;

function I_SoundIsPlaying(handle: integer): boolean;
var
  channel: integer;
begin
  if not boolval(pointer(pDS)) then
  begin
    result := false;
    exit;
  end;

  for channel := 0 to NUM_CHANNELS - 1 do
  begin
    if (channelhandles[channel] = handle) and I_ChannelPlaying(channel) then
    begin
      result := true;
      exit;
    end;
  end;
  result := false;
end;

//
// to mix the sounds into buffers, called every frame
//
procedure I_ProcessSound;
begin
end;

procedure I_UpdateSoundParams(handle: integer; vol: integer; sep: integer;
  pitch: integer);
begin
  // I fail too see that this is used.
  // Would be using the handle to identify
  //  on which channel the sound might be active,
  //  and resetting the channel parameters.
end;

procedure I_ShutdownSound1;
var
  i: integer;
begin
  if pDS <> nil then
  begin
    for i := 0 to NUM_CHANNELS - 1 do
      I_KillChannel(i);
  end;

  if pDSBPrimary <> nil then
    pDSBPrimary._Release;

  if pDS <> nil then
    pDS._Release;
end;

procedure I_ShutdownSound;
var
  i: integer;
begin
  for i := 0 to NUM_CHANNELS - 1 do
    I_KillChannel(i);
end;

procedure I_InitSound;
var
  hres: HRESULT;
  dsbd: DSBUFFERDESC;
  i: integer;
begin
  if boolval(M_CheckParm('-nosound')) then
    exit;

  hres := DirectSoundCreate(nil, pDS, nil);
  if hres <> DS_OK then
  begin
    pDS := nil;
    printf('I_InitSound(): DirectSoundCreate Failed, result = %d', [hres]);
    exit;
  end;

  hres := pDS.SetCooperativeLevel(hMainWnd, DSSCL_PRIORITY);
  if hres <> DS_OK then
    I_Error('I_InitSound(): DirectSound.SetCooperativeLevel Failed, result = %d', [hres]);

  SampleFormat.wFormatTag := WAVE_FORMAT_PCM;
  SampleFormat.nChannels := 1;
  SampleFormat.nSamplesPerSec := 11025;
  SampleFormat.cbSize := 0;
  if HighSound then
  begin
    SampleFormat.nBlockAlign := 2;
    SampleFormat.nAvgBytesPerSec := 22050;
    SampleFormat.wBitsPerSample := 16;
  end
  else
  begin
    SampleFormat.nBlockAlign := 1;
    SampleFormat.nAvgBytesPerSec := 11025;
    SampleFormat.wBitsPerSample := 8;
  end;

  ZeroMemory(dsbd, SizeOf(DSBUFFERDESC));
  dsbd.dwSize := SizeOf(DSBUFFERDESC);
  dsbd.dwFlags := DSBCAPS_PRIMARYBUFFER;
  dsbd.dwBufferBytes := 0;
  dsbd.lpwfxFormat := nil;

  hres := pDS.CreateSoundBuffer(dsbd, pDSBPrimary, nil);
  if hres <> DS_OK then
  begin
    printf('I_InitSound(): Unable to access primary sound buffer, result = %d', [hres]);
    pDSBPrimary := nil;
  end
  else
  begin
    hres := pDSBPrimary.SetFormat(SampleFormat);
    if hres <> DS_OK then
      printf('I_InitSound(): Unable to set primary sound buffer format, result = %d', [hres]);
    pDSBPrimary.Play(0, 0, DSBPLAY_LOOPING);
  end;

  for i := 0 to NUM_CHANNELS - 1 do
    ChannelBuffers[i] := nil;

  for i := 0 to Ord(NUMSFX) - 1 do
    S_sfx[i].data := nil;

  SampleFormat.wFormatTag := WAVE_FORMAT_PCM;
  SampleFormat.nChannels := 1;
  SampleFormat.nSamplesPerSec := 11025;
  SampleFormat.nBlockAlign := 1;
  SampleFormat.nAvgBytesPerSec := 11025;
  SampleFormat.wBitsPerSample := 8;
  SampleFormat.cbSize := 0;
end;

// ... update sound buffer and audio device at runtime...
procedure I_UpdateSound;
begin
end;

procedure I_SubmitSound;
begin
end;



(* Unit dstrings implementation *)


               


(* Unit p_pspr_h implementation *)



(* Unit d_event implementation *)




(* Unit i_system implementation *)


                                                                                                                   

var
  i_system_finished: boolean = false;

function I_GameFinished: boolean;
begin
  result := i_system_finished;
end;

procedure I_ProcessWindows;
var
  msg: TMsg;
begin
  while PeekMessage(msg, 0, 0, 0, PM_REMOVE) do
  begin
    if msg.message <> WM_QUIT then
    begin
      TranslateMessage(msg);
      DispatchMessage(msg);
    end;
  end;
end;

//
// I_StartFrame
//
procedure I_StartFrame;
begin
  I_ProcessWindows;
  I_ProcessMusic;
  I_ProcessInput;
end;

//
// I_StartTic
//
procedure I_StartTic;
begin
//  if not InBackground then   // VJ ?
//    I_ProcessInput;          // VJ ?
end;

var
  emptycmd: ticcmd_t;

function I_BaseTiccmd: Pticcmd_t;
begin
  result := @emptycmd;
end;

function I_GetHeapSize: integer;
begin
  result := mb_used * 1024 * 1024;
end;

function I_ZoneBase(var size: integer): pointer;
begin
  size := I_GetHeapSize;
  result := malloc(size);
end;

//
// I_GetTime
// returns time in 1/70th second tics
//
var
  basetime: int64;
  Freq: int64;

function I_GetTime: integer;
var
  _time: int64;
begin
  if Freq = 1000 then
    _time := GetTickCount
  else
  begin
    if not QueryPerformanceCounter(_time) then
      _time := GetTickCount;
  end;
  if basetime = 0 then
    basetime := _time;
  result := round(((_time - basetime) / Freq) * TICRATE);
end;

//
// I_Init
//
procedure I_Init;
begin
  printf('I_InitSound: Initializing DirectSound.' + #13#10);
  I_InitSound;
  printf('I_InitMusic: Initializing music.' + #13#10);
  I_InitMusic;
  printf('I_InitInput: Initializing DirectInput.' + #13#10);
  I_InitInput;
end;

//
// I_Quit
//
procedure I_Quit;
begin
  PostMessage(hMainWnd, WM_DESTROY, 0, 0);
end;

procedure I_RestoreDesktop;
begin
  InvalidateRect(0, nil, true)
end;

procedure I_Destroy(const code: integer);
begin
  i_system_finished := true;
  D_QuitNetGame;
//  I_ShutdownSound;
  I_ShutdownMusic;
  I_ShutdownInput;
  M_SaveDefaults;
//  I_ShutdownGraphics;
  I_ShutdownIO;
  I_RestoreDesktop;
  Halt(code);
end;

// Wait for vertical retrace or pause a bit.
procedure I_WaitVBL(const count: integer);
begin
  sleep(count);
end;

//
// I_Error
//
procedure I_Error(const error: string; const Args: array of const);
var
  soutproc: TOutProc;
begin
  fprintf(stderr, 'I_Error: ');
  fprintf(stderr, error, Args);
  fprintf(stderr, #13#10);

  // Shutdown. Here might be other errors.
  if G_IsDemoRecording then
    G_CheckDemoStatus;

  soutproc := outproc;
  outproc := I_IOErrorMessageBox;
  printf(error, Args);
  outproc := soutproc;

  I_Destroy(1);
end;

procedure I_Error(const error: string);
begin
  I_Error(error, []);
end;


(* Unit m_menu implementation *)


                                                                                                                                                                                                                                           

var
// temp for screenblocks (0-9)
  screenSize: integer;

// -1 = no quicksave slot picked!
  quickSaveSlot: integer;

 // 1 = message to be printed
  messageToPrint: integer;
// ...and here is the message string!
  messageString: string;

  messageLastMenuActive: boolean;

// timed message = no input from user
  messageNeedsInput: boolean;

type
  PmessageRoutine = function(i: integer): pointer;

var
  messageRoutine: PmessageRoutine;

const
  SAVESTRINGSIZE = 24;

var
  gammamsg: array[0..4] of string;

// we are going to be entering a savegame string
  saveStringEnter: integer;
  saveSlot: integer;  // which slot to save in
  saveCharIndex: integer; // which char we're editing
// old save description before edit
  saveOldString: string;

const
  SKULLXOFF = -32;
  LINEHEIGHT = 16;

var
  savegamestrings: array[0..9] of string;
  endstring: string;

type
  menuitem_t = record
    // 0 = no cursor here, 1 = ok, 2 = arrows ok
    status: smallint;

    name: string[10];

    // choice = menu item #.
    // if status = 2,
    //   choice=0:leftarrow,1:rightarrow
    routine: PmessageRoutine;

    // hotkey in menu
    alphaKey: char;
  end;
  Pmenuitem_t = ^menuitem_t;
  menuitem_tArray = packed array[0..$FFFF] of menuitem_t;
  Pmenuitem_tArray = ^menuitem_tArray;

  Pmenu_t = ^menu_t;
  menu_t = record
    numitems: smallint;         // # of menu items
    prevMenu:Pmenu_t;           // previous menu
    menuitems: Pmenuitem_tArray;// menu items
    routine: PProcedure;        // draw routine
    x: smallint;
    y: smallint;                // x,y of menu
    lastOn: smallint;           // last item user was on in menu
  end;

var
  itemOn: smallint;             // menu item skull is on
  skullAnimCounter: smallint;   // skull animation counter
  whichSkull: smallint;         // which skull to draw

// graphic name of skulls
// warning: initializer-string for array of chars is too long
  skullName: array[0..1] of string;

// current menudef
  currentMenu: Pmenu_t;

//
// PROTOTYPES
//
procedure M_NewGame(choice: integer); forward;
procedure M_Episode(choice: integer); forward;
procedure M_ChooseSkill(choice: integer); forward;
procedure M_LoadGame(choice: integer); forward;
procedure M_SaveGame(choice: integer); forward;
procedure M_Options(choice: integer); forward;
procedure M_EndGame(choice: integer); forward;
procedure M_ReadThis(choice: integer); forward;
procedure M_ReadThis2(choice: integer); forward;
procedure M_QuitDOOM(choice: integer); forward;

procedure M_ChangeMessages(choice: integer); forward;
procedure M_ChangeSensitivity(choice: integer); forward;
procedure M_SfxVol(choice: integer); forward;
procedure M_MusicVol(choice: integer); forward;
procedure M_ChangeDetail(choice: integer); forward;
procedure M_SizeDisplay(choice: integer); forward;
procedure M_Sound(choice: integer); forward;

procedure M_FinishReadThis(choice: integer); forward;
procedure M_LoadSelect(choice: integer); forward;
procedure M_SaveSelect(choice: integer); forward;
procedure M_ReadSaveStrings; forward;
procedure M_QuickSave; forward;
procedure M_QuickLoad; forward;

procedure M_DrawMainMenu; forward;
procedure M_DrawReadThis1; forward;
procedure M_DrawReadThis2; forward;
procedure M_DrawNewGame; forward;
procedure M_DrawEpisode; forward;
procedure M_DrawOptions; forward;
procedure M_DrawSound; forward;
procedure M_DrawLoad; forward;
procedure M_DrawSave; forward;

procedure M_DrawSaveLoadBorder(x, y: integer); forward;
procedure M_SetupNextMenu(menudef: Pmenu_t); forward;
procedure M_DrawThermo(x, y, thermWidth, thermDot: integer); forward;
procedure M_DrawEmptyCell(menu: Pmenu_t; item: integer); forward;
procedure M_DrawSelCell(menu: Pmenu_t; item: integer); forward;
procedure M_WriteText(x, y: integer;const _string: string); forward;
function  M_StringWidth(const _string: string): integer; forward;
function  M_StringHeight(const _string: string): integer; forward;
procedure M_StartMessage(const _string: string; routine: PmessageRoutine; input: boolean); forward;
procedure M_StopMessage; forward;
procedure M_ClearMenus; forward;

type
//
// DOOM MENU
//
  main_e = (
    newgame,
    options,
    loadgame,
    savegame,
    readthis,
    quitdoom,
    main_end
  );

var
  MainMenu: array[0..5] of menuitem_t;
  MainDef: menu_t;

type
//
// EPISODE SELECT
//
  episodes_e = (
    ep1,
    ep2,
    ep3,
    ep4,
    ep_end
  );

var
  EpisodeMenu: array[0..3] of menuitem_t;
  EpiDef: menu_t;

type
//
// NEW GAME
//
  newgame_e = (
    killthings,
    toorough,
    hurtme,
    violence,
    nightmare,
    newg_end
  );

var
  NewGameMenu: array[0..4] of menuitem_t;
  NewDef: menu_t;

type
//
// OPTIONS MENU
//
  options_e = (
    endgame,
    enum_messages,
    detail,
    scrnsize,
    option_empty1,
    mousesens,
    option_empty2,
    soundvol,
    opt_end
  );

var
  OptionsMenu: array[0..7] of menuitem_t;
  OptionsDef: menu_t;

type
//
// Read This! MENU 1 & 2
//
  read_e = (
    rdthsempty1,
    read1_end
  );

var
  ReadMenu1: array[0..0] of menuitem_t;
  ReadDef1: menu_t;

type
  read_e2 = (
    rdthsempty2,
    read2_end
  );

var
  ReadMenu2: array[0..0] of menuitem_t;
  ReadDef2: menu_t;

type
//
// SOUND VOLUME MENU
//
  sound_e = (
    sfx_vol,
    sfx_empty1,
    music_vol,
    sfx_empty2,
    sound_end
  );

var
  SoundMenu: array[0..3] of menuitem_t;
  SoundDef: menu_t;

type
//
// LOAD GAME MENU
//
  load_e = (
    load1,
    load2,
    load3,
    load4,
    load5,
    load6,
    load_end
  );

var
  LoadMenu: array[0..5] of menuitem_t;
  LoadDef: menu_t;
  SaveMenu: array[0..5] of menuitem_t;
  SaveDef: menu_t;

//
// M_ReadSaveStrings
//  read the strings from the savegame files
//
procedure M_ReadSaveStrings;
var
  handle: file;
  i: integer;
  name: string;
begin
  for i := 0 to Ord(load_end) - 1 do
  begin
    if M_CheckParmCDROM then
      sprintf(name, 'c:\doomdata\' + SAVEGAMENAME + '%d.dsg', [i])
    else
      sprintf(name, SAVEGAMENAME + '%d.dsg', [i]);

    assign(handle, name);
    {$I-}
    reset(handle, 1);
    {$I+}
    if IOResult <> 0 then
    begin
      savegamestrings[i] := '';
      LoadMenu[i].status := 0;
      continue;
    end;
    SetLength(savegamestrings[i], SAVESTRINGSIZE);
    BlockRead(handle, (@savegamestrings[i][1])^, SAVESTRINGSIZE);
    closefile(handle);
    LoadMenu[i].status := 1;
  end;
end;

//
// M_LoadGame & Cie.
//
procedure M_DrawLoad;
var
  i: integer;
begin
  V_DrawPatch(72, 28, _FG, W_CacheLumpName('M_LOADG', PU_CACHE), true);
  for i := 0 to Ord(load_end) - 1 do
  begin
    M_DrawSaveLoadBorder(LoadDef.x, LoadDef.y + LINEHEIGHT * i);
    M_WriteText(LoadDef.x, LoadDef.y + LINEHEIGHT * i, savegamestrings[i]);
  end;
end;

//
// Draw border for the savegame description
//
procedure M_DrawSaveLoadBorder(x, y: integer);
var
  i: integer;
begin
  V_DrawPatch(x - 8, y + 7, _FG, W_CacheLumpName('M_LSLEFT', PU_CACHE), true);

  for i := 0 to 23 do
  begin
    V_DrawPatch (x, y + 7, _FG, W_CacheLumpName('M_LSCNTR', PU_CACHE), true);
    x := x + 8;
  end;

  V_DrawPatch(x, y + 7, _FG, W_CacheLumpName('M_LSRGHT', PU_CACHE), true);
end;

//
// User wants to load this game
//
procedure M_LoadSelect(choice: integer);
var
  name: string;
begin
  if M_CheckParmCDROM then
    sprintf(name, 'c:\doomdata\' + SAVEGAMENAME + '%d.dsg', [choice])
  else
    sprintf(name,SAVEGAMENAME + '%d.dsg', [choice]);
  G_LoadGame(name);
  M_ClearMenus;
end;

//
// Selected from DOOM menu
//
procedure M_LoadGame(choice: integer);
begin
  if netgame then
  begin
    M_StartMessage(LOADNET, nil, false);
    exit;
  end;

  M_SetupNextMenu(@LoadDef);
  M_ReadSaveStrings;
end;

//
//  M_SaveGame & Cie.
//
procedure M_DrawSave;
var
  i: integer;
begin
  V_DrawPatch(72, 28, _FG, W_CacheLumpName('M_SAVEG', PU_CACHE), true);
  for i := 0 to Ord(load_end) - 1 do
  begin
    M_DrawSaveLoadBorder(LoadDef.x, LoadDef.y + LINEHEIGHT * i);
    M_WriteText(LoadDef.x, LoadDef.y + LINEHEIGHT * i, savegamestrings[i]);
  end;

  if saveStringEnter <> 0 then
  begin
    i := M_StringWidth(savegamestrings[saveSlot]);
    M_WriteText(LoadDef.x + i, LoadDef.y + LINEHEIGHT * saveSlot, '_');
  end;
end;

//
// M_Responder calls this when user is finished
//
procedure M_DoSave(slot: integer);
begin
  G_SaveGame(slot, savegamestrings[slot]);
  M_ClearMenus;

  // PICK QUICKSAVE SLOT YET?
  if (quickSaveSlot = -2) then
    quickSaveSlot := slot;
end;

//
// User wants to save. Start string input for M_Responder
//
procedure M_SaveSelect(choice: integer);
begin
  // we are going to be intercepting all chars
  saveStringEnter := 1;

  saveSlot := choice;
  saveOldString := savegamestrings[choice];
  if savegamestrings[choice] <> '' then
    savegamestrings[choice] := '';
  saveCharIndex := Length(savegamestrings[choice]);
end;

//
// Selected from DOOM menu
//
procedure M_SaveGame(choice: integer);
begin
  if not usergame then
  begin
    M_StartMessage(SAVEDEAD, nil, false);
    exit;
  end;

  if gamestate <> GS_LEVEL then
    exit;

  M_SetupNextMenu(@SaveDef);
  M_ReadSaveStrings;
end;

//
//      M_QuickSave
//
var
  tempstring: string;

procedure M_QuickSaveResponse(ch: integer);
begin
  if ch = Ord('y') then
  begin
    M_DoSave(quickSaveSlot);
    S_StartSound(nil, Ord(sfx_swtchx));
  end;
end;

procedure M_QuickSave;
begin
  if not usergame then
  begin
    S_StartSound(nil, Ord(sfx_oof));
    exit;
  end;

  if gamestate <> GS_LEVEL then
    exit;

  if quickSaveSlot < 0 then
  begin
    M_StartControlPanel;
    M_ReadSaveStrings;
    M_SetupNextMenu(@SaveDef);
    quickSaveSlot := -2;  // means to pick a slot now
    exit;
  end;
  sprintf(tempstring, QSPROMPT, [savegamestrings[quickSaveSlot]]);
  M_StartMessage(tempstring, @M_QuickSaveResponse, true);
end;

//
// M_QuickLoad
//
procedure M_QuickLoadResponse(ch: integer);
begin
  if ch = ord('y') then
  begin
    M_LoadSelect(quickSaveSlot);
    S_StartSound(nil, Ord(sfx_swtchx));
  end;
end;

procedure M_QuickLoad;
begin
  if netgame then
  begin
    M_StartMessage(QLOADNET, nil, false);
    exit;
  end;

  if quickSaveSlot < 0 then
  begin
    M_StartMessage(QSAVESPOT, nil, false);
    exit;
  end;
  sprintf(tempstring, QLPROMPT, [savegamestrings[quickSaveSlot]]);
  M_StartMessage(tempstring, @M_QuickLoadResponse, true);
end;

//
// Read This Menus
// Had a "quick hack to fix romero bug"
//
procedure M_DrawReadThis1;
begin
  inhelpscreens := true;
  case gamemode of
    commercial:
      V_DrawPatch(0, 0, _FG, W_CacheLumpName('HELP', PU_CACHE), true);
    shareware,
    registered,
    retail:
      V_DrawPatch(0, 0, _FG, W_CacheLumpName('HELP1', PU_CACHE), true);
  end;
end;

//
// Read This Menus - optional second page.
//
procedure M_DrawReadThis2;
begin
  inhelpscreens := true;
  case gamemode of
    retail,
    commercial:
      // This hack keeps us from having to change menus.
      V_DrawPatch(0, 0, _FG, W_CacheLumpName('CREDIT', PU_CACHE), true);
    shareware,
    registered:
      V_DrawPatch(0, 0, _FG, W_CacheLumpName('HELP2', PU_CACHE), true);
  end;
end;

//
// Change Sfx & Music volumes
//
procedure M_DrawSound;
begin
  V_DrawPatch(60, 38, _FG, W_CacheLumpName('M_SVOL', PU_CACHE), true);

  M_DrawThermo(
    SoundDef.x, SoundDef.y + LINEHEIGHT * (Ord(sfx_vol) + 1), 16, snd_SfxVolume);

  M_DrawThermo(
    SoundDef.x, SoundDef.y + LINEHEIGHT * (Ord(music_vol) + 1), 16, snd_MusicVolume);
end;

procedure M_Sound(choice: integer);
begin
  M_SetupNextMenu(@SoundDef);
end;

procedure M_SfxVol(choice: integer);
begin
  case choice of
    0: if snd_SfxVolume <> 0 then dec(snd_SfxVolume);
    1: if snd_SfxVolume < 15 then inc(snd_SfxVolume);
  end;
  S_SetSfxVolume(snd_SfxVolume);
end;

procedure M_MusicVol(choice: integer);
begin
  case choice of
    0: if snd_MusicVolume <> 0 then dec(snd_MusicVolume);
    1: if snd_MusicVolume < 15 then inc(snd_MusicVolume);
  end;
  S_SetMusicVolume(snd_MusicVolume);
end;

//
// M_DrawMainMenu
//
procedure M_DrawMainMenu;
begin
  V_DrawPatch(94, 2, _FG, W_CacheLumpName('M_DOOM', PU_CACHE), true);
end;

//
// M_NewGame
//
procedure M_DrawNewGame;
begin
  V_DrawPatch(96, 14, _FG, W_CacheLumpName('M_NEWG', PU_CACHE), true);
  V_DrawPatch(54, 38, _FG, W_CacheLumpName('M_SKILL', PU_CACHE), true);
end;

procedure M_NewGame(choice: integer);
begin
  if netgame and (not demoplayback) then
  begin
    M_StartMessage(SNEWGAME, nil, false);
    exit;
  end;

  if gamemode = commercial then
    M_SetupNextMenu(@NewDef)
  else
    M_SetupNextMenu(@EpiDef);
end;

//
//      M_Episode
//
var
  epi: integer;

procedure M_DrawEpisode;
begin
  V_DrawPatch(54, 38, _FG, W_CacheLumpName('M_EPISOD', PU_CACHE), true);
end;

procedure M_VerifyNightmare(ch: integer);
begin
  if ch <> Ord('y') then
    exit;

  G_DeferedInitNew(sk_nightmare, epi + 1, 1); // VJ nightmare become sk_nightmare
  M_ClearMenus;
end;

procedure M_ChooseSkill(choice: integer);
begin
  if choice = Ord(nightmare) then
  begin
    M_StartMessage(SNIGHTMARE, @M_VerifyNightmare, true);
    exit;
  end;

  G_DeferedInitNew(skill_t(choice), epi + 1, 1);
  M_ClearMenus;
end;

procedure M_Episode(choice: integer);
begin
  if (gamemode = shareware) and (choice <> 0) then
  begin
    M_StartMessage(SWSTRING, nil, false);
    M_SetupNextMenu(@ReadDef1);
    exit;
  end;

  // Yet another hack...
  if (gamemode = registered) and (choice > 2) then
  begin
    fprintf(stderr, 'M_Episode(): 4th episode requires UltimateDOOM' + #13#10);
    choice := 0;
  end;

  epi := choice;
  M_SetupNextMenu(@NewDef);
end;

//
// M_Options
//
var
  detailNames: array[0..1] of string;
  msgNames: array[0..1] of string;

procedure M_DrawOptions;
begin
  V_DrawPatch(108, 15, _FG, W_CacheLumpName('M_OPTTTL', PU_CACHE), true);

  V_DrawPatch(OptionsDef.x + 175, OptionsDef.y + LINEHEIGHT * Ord(detail), _FG,
      W_CacheLumpName(detailNames[detailLevel], PU_CACHE), true);

  V_DrawPatch(OptionsDef.x + 120, OptionsDef.y + LINEHEIGHT * Ord(enum_messages), _FG,
      W_CacheLumpName(msgNames[showMessages], PU_CACHE), true);

  M_DrawThermo(
    OptionsDef.x, OptionsDef.y + LINEHEIGHT * (Ord(mousesens) + 1), 10, mouseSensitivity);

  M_DrawThermo(
    OptionsDef.x, OptionsDef.y + LINEHEIGHT * (Ord(scrnsize) + 1), 9, screenSize);
end;

procedure M_Options(choice: integer);
begin
  M_SetupNextMenu(@OptionsDef);
end;

//
//      Toggle messages on/off
//
procedure M_ChangeMessages(choice: integer);
begin
  showMessages := 1 - showMessages;

  if showMessages = 0 then
    players[consoleplayer]._message := MSGOFF
  else
    players[consoleplayer]._message := MSGON;

  message_dontfuckwithme := true;
end;

//
// M_EndGame
//
procedure M_EndGameResponse(ch: integer);
begin
  if ch <> Ord('y') then
    exit;

  currentMenu.lastOn := itemOn;
  M_ClearMenus;
  D_StartTitle;
end;

procedure M_EndGame(choice: integer);
begin
  if not usergame then
  begin
    S_StartSound(nil, Ord(sfx_oof));
    exit;
  end;

  if netgame then
  begin
    M_StartMessage(NETEND, nil, false);
    exit;
  end;

  M_StartMessage(SENDGAME, @M_EndGameResponse, true);
end;

//
// M_ReadThis
//
procedure M_ReadThis(choice: integer);
begin
  M_SetupNextMenu(@ReadDef1);
end;

procedure M_ReadThis2(choice: integer);
begin
  M_SetupNextMenu(@ReadDef2);
end;

procedure M_FinishReadThis(choice: integer);
begin
  M_SetupNextMenu(@MainDef);
end;

//
// M_QuitDOOM
//
const
  quitsounds: array[0..7] of integer = (
    Ord(sfx_pldeth),
    Ord(sfx_dmpain),
    Ord(sfx_popain),
    Ord(sfx_slop),
    Ord(sfx_telept),
    Ord(sfx_posit1),
    Ord(sfx_posit3),
    Ord(sfx_sgtatk)
  );

  quitsounds2: array[0..7] of integer = (
    Ord(sfx_vilact),
    Ord(sfx_getpow),
    Ord(sfx_boscub),
    Ord(sfx_slop),
    Ord(sfx_skeswg),
    Ord(sfx_kntdth),
    Ord(sfx_bspact),
    Ord(sfx_sgtatk)
  );


procedure M_QuitResponse(ch: integer);
begin
  if ch <> Ord('y') then
    exit;
  if not netgame then
  begin
    if gamemode = commercial then
      S_StartSound(nil, quitsounds2[_SHR(gametic, 2) and 7])
    else
      S_StartSound(nil, quitsounds[_SHR(gametic, 2) and 7]);
//    I_WaitVBL(105);
    I_WaitVBL(1000);
  end;
  I_Quit;
end;

procedure M_QuitDOOM(choice: integer);
begin
  // We pick index 0 which is language sensitive,
  //  or one at random, between 1 and maximum number.
  if language <> english then
    sprintf(endstring, '%s' + #13#10#13#10 + DOSY, [endmsg[0]])
  else
    sprintf(endstring,'%s' + #13#10#13#10 + DOSY, [endmsg[(gametic mod (NUM_QUITMESSAGES - 2)) + 1]]);

  M_StartMessage(endstring, @M_QuitResponse, true);
end;

procedure M_ChangeSensitivity(choice: integer);
begin
  case choice of
    0:
      if mouseSensitivity > 0 then
        dec(mouseSensitivity);
    1:
      if mouseSensitivity < 9 then
        inc(mouseSensitivity);
  end;
end;

procedure M_ChangeDetail(choice: integer);
begin
// FIXME - does not work. Remove anyway?
{  fprintf(stderr, 'M_ChangeDetail(): low detail mode n.a.' + #13#10);
  exit;}

  detailLevel := 1 - detailLevel;


  R_SetViewSize(screenblocks, detailLevel);

  if detailLevel = 0 then
    players[consoleplayer]._message := DETAILHI
  else
    players[consoleplayer]._message := DETAILLO;

end;

procedure M_SizeDisplay(choice: integer);
begin
  case choice of
    0:
      begin
        if screenSize > 0 then
        begin
          dec(screenblocks);
          dec(screenSize);
        end;
      end;
    1:
      begin
        if screenSize < 8 then
        begin
          inc(screenblocks);
          inc(screenSize);
        end;
      end;
  end;

  R_SetViewSize(screenblocks, detailLevel);
end;

//
//      Menu Functions
//
procedure M_DrawThermo(x, y, thermWidth, thermDot: integer);
var
  xx: integer;
  i: integer;
begin
  xx := x;
  V_DrawPatch(xx, y, _FG, W_CacheLumpName('M_THERML', PU_CACHE), true);
  xx := xx + 8;
  for i := 0 to thermWidth - 1 do
  begin
    V_DrawPatch(xx, y, _FG, W_CacheLumpName('M_THERMM', PU_CACHE), true);
    xx := xx + 8;
  end;
  V_DrawPatch(xx, y, _FG, W_CacheLumpName('M_THERMR', PU_CACHE), true);

  V_DrawPatch((x + 8) + thermDot * 8, y, _FG,
    W_CacheLumpName('M_THERMO', PU_CACHE), true);
end;

procedure M_DrawEmptyCell(menu: Pmenu_t; item: integer);
begin
  V_DrawPatch(menu.x - 10, menu.y + item * LINEHEIGHT - 1, _FG,
    W_CacheLumpName('M_CELL1', PU_CACHE), true);
end;

procedure M_DrawSelCell(menu: Pmenu_t; item: integer);
begin
  V_DrawPatch(menu.x - 10, menu.y + item * LINEHEIGHT - 1, _FG,
    W_CacheLumpName('M_CELL2', PU_CACHE), true);
end;

procedure M_StartMessage(const _string: string; routine: PmessageRoutine; input: boolean);
begin
  messageLastMenuActive := menuactive;
  messageToPrint := 1;
  messageString := _string;
  if Assigned(routine) then
    @messageRoutine := @routine
  else
    messageRoutine := nil;
  messageNeedsInput := input;
  menuactive := true;
end;

procedure M_StopMessage;
begin
  menuactive := messageLastMenuActive;
  messageToPrint := 0;
end;

//
// Find string width from hu_font chars
//
function  M_StringWidth(const _string: string): integer;
var
  i: integer;
  c: integer;
begin
  result := 0;
  for i := 1 to Length(_string) do
  begin
    c := Ord(toupper(_string[i])) - Ord(HU_FONTSTART);
    if (c < 0) or (c >= HU_FONTSIZE) then
      result := result + 4
    else
      result := result + hu_font[c].width;
  end;
end;

//
//      Find string height from hu_font chars
//
function  M_StringHeight(const _string: string): integer;
var
  i: integer;
  height: integer;
begin
  height := hu_font[0].height;

  result := height;
  for i := 1 to Length(_string) do
    if _string[i] = #13 then
      result := result + height;
end;

//
//      Write a string using the hu_font
//
procedure M_WriteText(x, y: integer;const _string: string);
var
  w: integer;
  ch: integer;
  c: integer;
  cx: integer;
  cy: integer;
  len: integer;
begin
  len := Length(_string);
  if len = 0 then
    exit;

  ch := 1;
  cx := x;
  cy := y;

  while true do
  begin
    if ch > len then
      break;
      
    c := Ord(_string[ch]);
    inc(ch);

    if c = 0 then
      break;

    if c = 10 then
    begin
      cx := x;
      continue;
    end;

    if c = 13 then
    begin
      cy := cy + 12;
      continue;
    end;

    c := Ord(toupper(Chr(c))) - Ord(HU_FONTSTART);
    if (c < 0) or (c >= HU_FONTSIZE) then
    begin
      cx := cx + 4;
      continue;
    end;

    w := hu_font[c].width;
//    if (cx + w) > SCREENWIDTH then
    if (cx + w) > 320 then
      break;
    V_DrawPatch(cx, cy, _FG, hu_font[c], true);
    cx := cx + w;
  end;
end;

//
// CONTROL PANEL
//

//
// M_Responder
//
var
  joywait: integer;
  mousewait: integer;
  mousey: integer;
  lasty: integer;
  mousex: integer;
  lastx: integer;

function M_Responder(ev: Pevent_t): boolean;
var
  ch: integer;
  i: integer;
begin
  ch := -1;

  if (ev._type = ev_joystick) and (joywait < I_GetTime) then
  begin
    if ev.data3 = -1 then
    begin
      ch := KEY_UPARROW;
      joywait := I_GetTime + 5;
    end
    else if ev.data3 = 1 then
    begin
      ch := KEY_DOWNARROW;
      joywait := I_GetTime + 5;
    end;

    if ev.data2 = -1 then
    begin
      ch := KEY_LEFTARROW;
      joywait := I_GetTime + 2;
    end
    else if ev.data2 = 1 then
    begin
      ch := KEY_RIGHTARROW;
      joywait := I_GetTime + 2;
    end;

    if (ev.data1 and 1) <> 0 then
    begin
      ch := KEY_ENTER;
      joywait := I_GetTime + 5;
    end;
    if (ev.data1 and 2) <> 0 then
    begin
      ch := KEY_BACKSPACE;
      joywait := I_GetTime + 5;
    end;
  end
  else if (ev._type = ev_mouse) and (mousewait < I_GetTime) then
  begin
    mousey := mousey + ev.data3;
    if mousey < lasty - 30 then
    begin
      ch := KEY_DOWNARROW;
      mousewait := I_GetTime + 5;
      lasty := lasty - 30;
      mousey := lasty;
    end
    else if mousey > lasty + 30 then
    begin
      ch := KEY_UPARROW;
      mousewait := I_GetTime + 5;
      lasty := lasty + 30;
      mousey := lasty;
    end;

    mousex := mousex + ev.data2;
    if mousex < lastx - 30 then
    begin
      ch := KEY_LEFTARROW;
      mousewait := I_GetTime + 5;
      lastx := lastx - 30;
      mousex := lastx;
    end
    else if mousex > lastx + 30 then
    begin
      ch := KEY_RIGHTARROW;
      mousewait := I_GetTime + 5;
      lastx := lastx + 30;
      mousex := lastx;
    end;

    if (ev.data1 and 1) <> 0 then
    begin
      ch := KEY_ENTER;
      mousewait := I_GetTime + 15;
    end;

    if (ev.data1 and 2) <> 0 then
    begin
      ch := KEY_BACKSPACE;
      mousewait := I_GetTime + 15;
    end
  end
  else if ev._type = ev_keydown then
    ch := ev.data1;

  if ch = -1 then
  begin
    result := false;
    exit;
  end;

  // Save Game string input
  if saveStringEnter <> 0 then
  begin
    case ch of
      KEY_BACKSPACE:
        begin
          if saveCharIndex > 0 then
          begin
            dec(saveCharIndex);
            SetLength(savegamestrings[saveSlot], saveCharIndex);
          end;
        end;
      KEY_ESCAPE:
        begin
          saveStringEnter := 0;
          savegamestrings[saveSlot] := saveOldString;
        end;
      KEY_ENTER:
        begin
          saveStringEnter := 0;
          if savegamestrings[saveSlot] <> '' then
            M_DoSave(saveSlot);
        end
    else
      begin
        ch := Ord(toupper(Chr(ch)));
        if ch <> 32 then
        if (ch - Ord(HU_FONTSTART) < 0) or (ch - Ord(HU_FONTSTART) >= HU_FONTSIZE) then
        else
        begin
          if (ch >= 32) and (ch <= 127) and
             (saveCharIndex < SAVESTRINGSIZE - 1) and
             (M_StringWidth(savegamestrings[saveSlot]) < (SAVESTRINGSIZE - 2) * 8) then
          begin
            inc(saveCharIndex);
            savegamestrings[saveSlot] := savegamestrings[saveSlot] + Chr(ch);
          end;
        end;
      end;
    end;
    result := true;
    exit;
  end;

  // Take care of any messages that need input
  if messageToPrint <> 0 then
  begin
    if messageNeedsInput and ( not(
      (ch = Ord(' ')) or (ch = Ord('n')) or (ch = Ord('y')) or (ch = KEY_ESCAPE))) then
    begin
      result := false;
      exit;
    end;

    menuactive := messageLastMenuActive;
    messageToPrint := 0;
    if Assigned(messageRoutine) then
      messageRoutine(ch);

    result := true;

    if I_GameFinished then
      exit;
      
    menuactive := false;
    S_StartSound(nil, Ord(sfx_swtchx));
    exit;
  end;

  if devparm and (ch = KEY_F1) then
  begin
    G_ScreenShot;
    result := true;
    exit;
  end;

  // F-Keys
  if not menuactive then
    case ch of
      KEY_MINUS:    // Screen size down
        begin
          if automapactive or chat_on then
          begin
            result := false;
            exit;
          end;
          M_SizeDisplay(0);
          S_StartSound(nil, Ord(sfx_stnmov));
          result := true;
          exit;
        end;
      KEY_EQUALS:   // Screen size up
        begin
          if automapactive or chat_on then
          begin
            result := false;
            exit;
          end;
          M_SizeDisplay(1);
          S_StartSound(nil, Ord(sfx_stnmov));
          result := true;
          exit;
        end;
      KEY_F1:      // Help key
        begin
          M_StartControlPanel;
          if gamemode = retail then
            currentMenu := @ReadDef2
          else
            currentMenu := @ReadDef1;

          itemOn := 0;
          S_StartSound(nil, Ord(sfx_swtchn));
          result := true;
          exit;
        end;
      KEY_F2:  // Save
        begin
          M_StartControlPanel;
          S_StartSound(nil, Ord(sfx_swtchn));
          M_SaveGame(0);
          result := true;
          exit;
        end;
      KEY_F3:  // Load
        begin
          M_StartControlPanel;
          S_StartSound(nil, Ord(sfx_swtchn));
          M_LoadGame(0);
          result := true;
          exit;
        end;
      KEY_F4:   // Sound Volume
        begin
          M_StartControlPanel;
          currentMenu := @SoundDef;
          itemOn := Ord(sfx_vol);
          S_StartSound(nil, Ord(sfx_swtchn));
          result := true;
          exit;
        end;
      KEY_F5:   // Detail toggle
        begin
          M_ChangeDetail(0);
          S_StartSound(nil, Ord(sfx_swtchn));
          result := true;
          exit;
        end;
      KEY_F6:   // Quicksave
        begin
          S_StartSound(nil, Ord(sfx_swtchn));
          M_QuickSave;
          result := true;
          exit;
        end;
      KEY_F7:   // End game
        begin
          S_StartSound(nil, Ord(sfx_swtchn));
          M_EndGame(0);
          result := true;
          exit;
        end;
      KEY_F8:   // Toggle messages
        begin
          M_ChangeMessages(0);
          S_StartSound(nil, Ord(sfx_swtchn));
          result := true;
          exit;
        end;
      KEY_F9:   // Quickload
        begin
          S_StartSound(nil, Ord(sfx_swtchn));
          M_QuickLoad;
          result := true;
          exit;
        end;
      KEY_F10:  // Quit DOOM
        begin
          S_StartSound(nil, Ord(sfx_swtchn));
          M_QuitDOOM(0);
          result := true;
          exit;
        end;
      KEY_F11:  // gamma toggle
        begin
          inc(usegamma);
          if usegamma > 4 then
            usegamma := 0;
          players[consoleplayer]._message := gammamsg[usegamma];
          I_SetPalette(W_CacheLumpName('PLAYPAL', PU_CACHE));
          result := true;
          exit;
        end;
    end;

  // Pop-up menu?
  if not menuactive then
  begin
    if ch = KEY_ESCAPE then
    begin
      M_StartControlPanel;
      S_StartSound(nil, Ord(sfx_swtchn));
      result := true;
      exit;
    end;
    result := false;
    exit;
  end;

  // Keys usable within menu
  case ch of
    KEY_DOWNARROW:
      begin
        repeat
          if itemOn + 1 > currentMenu.numitems - 1 then
            itemOn := 0
          else
            inc(itemOn);
          S_StartSound(nil, Ord(sfx_pstop));
        until currentMenu.menuitems[itemOn].status <> -1;
        result := true;
        exit;
      end;
    KEY_UPARROW:
      begin
        repeat
          if itemOn = 0 then
            itemOn := currentMenu.numitems - 1
          else
            dec(itemOn);
          S_StartSound(nil, Ord(sfx_pstop));
        until currentMenu.menuitems[itemOn].status <> -1;
        result := true;
        exit;
      end;
    KEY_LEFTARROW:
      begin
        if Assigned(currentMenu.menuitems[itemOn].routine) and
          (currentMenu.menuitems[itemOn].status = 2) then
        begin
          S_StartSound(nil, Ord(sfx_stnmov));
          currentMenu.menuitems[itemOn].routine(0);
        end;
        result := true;
        exit;
      end;
    KEY_RIGHTARROW:
      begin
        if Assigned(currentMenu.menuitems[itemOn].routine) and
          (currentMenu.menuitems[itemOn].status = 2) then
        begin
          S_StartSound(nil, Ord(sfx_stnmov));
          currentMenu.menuitems[itemOn].routine(1);
        end;
        result := true;
        exit;
      end;
    KEY_ENTER:
      begin
        if Assigned(currentMenu.menuitems[itemOn].routine) and
          (currentMenu.menuitems[itemOn].status <> 0) then
        begin
          currentMenu.lastOn := itemOn;
          if currentMenu.menuitems[itemOn].status = 2 then
          begin
            currentMenu.menuitems[itemOn].routine(1); // right arrow
            S_StartSound(nil, Ord(sfx_stnmov));
          end
          else
          begin
            currentMenu.menuitems[itemOn].routine(itemOn);
            S_StartSound(nil, Ord(sfx_pistol));
          end;
        end;
        result := true;
        exit;
      end;
    KEY_ESCAPE:
      begin
        currentMenu.lastOn := itemOn;
        M_ClearMenus;
        S_StartSound(nil, Ord(sfx_swtchx));
        result := true;
        exit;
      end;
    KEY_BACKSPACE:
      begin
        currentMenu.lastOn := itemOn;
        if currentMenu.prevMenu <> nil then
        begin
          currentMenu := currentMenu.prevMenu;
          itemOn := currentMenu.lastOn;
          S_StartSound(nil, Ord(sfx_swtchn));
        end;
        result := true;
        exit;
      end;
  else
    begin
      for i := itemOn + 1 to currentMenu.numitems - 1 do
        if currentMenu.menuitems[i].alphaKey = Chr(ch) then
        begin
          itemOn := i;
          S_StartSound(nil, Ord(sfx_pstop));
          result := true;
          exit;
        end;
      for i := 0 to itemOn do
        if currentMenu.menuitems[i].alphaKey = Chr(ch) then
        begin
          itemOn := i;
          S_StartSound(nil, Ord(sfx_pstop));
          result := true;
          exit;
        end;
    end;
  end;
  
  result := false;
end;

//
// M_StartControlPanel
//
procedure M_StartControlPanel;
begin
  // intro might call this repeatedly
  if menuactive then
    exit;

  menuactive := true;
  currentMenu := @MainDef;// JDC
  itemOn := currentMenu.lastOn; // JDC
end;

//
// M_Drawer
// Called after the view has been rendered,
// but before it has been blitted.
//
var
  x, m_menu_y: smallint;

procedure M_Drawer;
var
  i: smallint;
  max: smallint;
  _string: string;
  len: integer;
begin
  inhelpscreens := false;

  // Horiz. & Vertically center string and print it.
  if messageToPrint <> 0 then
  begin
//    y := (SCREENHEIGHT - M_StringHeight(messageString)) div 2;
    m_menu_y := (200 - M_StringHeight(messageString)) div 2;
    len := Length(messageString);
    _string := '';
    for i := 1 to len do
    begin
      if messageString[i] = #13 then
        m_menu_y := m_menu_y + hu_font[0].height
      else if messageString[i] = #10 then
      begin
//        x := (SCREENWIDTH - M_StringWidth(_string)) div 2;
        x := (320 - M_StringWidth(_string)) div 2;
        M_WriteText(x, m_menu_y, _string);
        _string := '';
      end
      else
        _string := _string + messageString[i];
    end;
    if _string <> '' then
    begin
//      x := (SCREENWIDTH - M_StringWidth(_string)) div 2;
      x := (320 - M_StringWidth(_string)) div 2;
      m_menu_y := m_menu_y + hu_font[0].height;
      M_WriteText(x, m_menu_y, _string);
    end;
    exit;
  end;

  if not menuactive then
    exit;

  if Assigned(currentMenu.routine) then
    currentMenu.routine; // call Draw routine

  // DRAW MENU
  x := currentMenu.x;
  m_menu_y := currentMenu.y;
  max := currentMenu.numitems;

  for i := 0 to max - 1 do
  begin
    if currentMenu.menuitems[i].name <> '' then
      V_DrawPatch(x, m_menu_y, _FG,
        W_CacheLumpName(currentMenu.menuitems[i].name, PU_CACHE), true);
    m_menu_y := m_menu_y + LINEHEIGHT;
  end;

  // DRAW SKULL
  V_DrawPatch(x + SKULLXOFF, currentMenu.y - 5 + itemOn * LINEHEIGHT, _FG,
    W_CacheLumpName(skullName[whichSkull], PU_CACHE), true);
end;

//
// M_ClearMenus
//
procedure M_ClearMenus;
begin
  menuactive := false;
    // if (!netgame && usergame && paused)
    //       sendpause = true;
end;

//
// M_SetupNextMenu
//
procedure M_SetupNextMenu(menudef: Pmenu_t);
begin
  currentMenu := menudef;
  itemOn := currentMenu.lastOn;
end;

//
// M_Ticker
//
procedure M_Ticker;
begin
  dec(skullAnimCounter);
  if skullAnimCounter <= 0 then
  begin
    whichSkull := whichSkull xor 1;
    skullAnimCounter := 8;
  end;
end;

//
// M_Init
//
procedure M_Init;
begin
  currentMenu := @MainDef;
  menuactive := false;
  itemOn := currentMenu.lastOn;
  whichSkull := 0;
  skullAnimCounter := 10;
  screenSize := screenblocks - 3;
  messageToPrint := 0;
  messageString := '';
  messageLastMenuActive := menuactive;
  quickSaveSlot := -1;

  // Here we could catch other version dependencies,
  //  like HELP1/2, and four episodes.

  case gamemode of
    commercial:
      begin
        // This is used because DOOM 2 had only one HELP
        //  page. I use CREDIT as second page now, but
        //  kept this hack for educational purposes.
        MainMenu[Ord(readthis)] := MainMenu[Ord(quitdoom)];
        dec(MainDef.numitems);
        MainDef.y := MainDef.y + 8;
        NewDef.prevMenu := @MainDef;
        ReadDef1.routine := M_DrawReadThis1;
        ReadDef1.x := 330;
        ReadDef1.y := 165;
        ReadMenu1[0].routine := @M_FinishReadThis;
      end;
    shareware,
      // Episode 2 and 3 are handled,
      // branching to an ad screen.
    registered:
      begin
        // We need to remove the fourth episode.
        dec(EpiDef.numitems);
      end;
  end;
end;


procedure M_InitMenus;
begin
////////////////////////////////////////////////////////////////////////////////
//gammamsg
  gammamsg[0] := GAMMALVL0;
  gammamsg[1] := GAMMALVL1;
  gammamsg[2] := GAMMALVL2;
  gammamsg[3] := GAMMALVL3;
  gammamsg[4] := GAMMALVL4;

////////////////////////////////////////////////////////////////////////////////
//skullName
  skullName[0] := 'M_SKULL1';
  skullName[1] := 'M_SKULL2';

////////////////////////////////////////////////////////////////////////////////
// MainMenu
  MainMenu[0].status := 1;
  MainMenu[0].name := 'M_NGAME';
  MainMenu[0].routine := @M_NewGame;
  MainMenu[0].alphaKey := 'n';

  MainMenu[1].status := 1;
  MainMenu[1].name := 'M_OPTION';
  MainMenu[1].routine := @M_Options;
  MainMenu[1].alphaKey := 'o';

  MainMenu[2].status := 1;
  MainMenu[2].name := 'M_LOADG';
  MainMenu[2].routine := @M_LoadGame;
  MainMenu[2].alphaKey := 'l';

  MainMenu[3].status := 1;
  MainMenu[3].name := 'M_SAVEG';
  MainMenu[3].routine := @M_SaveGame;
  MainMenu[3].alphaKey := 's';

  // Another hickup with Special edition.
  MainMenu[4].status := 1;
  MainMenu[4].name := 'M_RDTHIS';
  MainMenu[4].routine := @M_ReadThis;
  MainMenu[4].alphaKey := 'r';

  MainMenu[5].status := 1;
  MainMenu[5].name := 'M_QUITG';
  MainMenu[5].routine := @M_QuitDOOM;
  MainMenu[5].alphaKey := 'q';

////////////////////////////////////////////////////////////////////////////////
//MainDef
  MainDef.numitems := ord(main_end);
  MainDef.prevMenu := nil;
  MainDef.menuitems := Pmenuitem_tArray(@MainMenu);
  MainDef.routine := @M_DrawMainMenu;  // draw routine
  MainDef.x := 97;
  MainDef.y := 64;
  MainDef.lastOn := 0;

////////////////////////////////////////////////////////////////////////////////
//EpisodeMenu
  EpisodeMenu[0].status := 1;
  EpisodeMenu[0].name := 'M_EPI1';
  EpisodeMenu[0].routine := @M_Episode;
  EpisodeMenu[0].alphaKey := 'k';

  EpisodeMenu[1].status := 1;
  EpisodeMenu[1].name := 'M_EPI2';
  EpisodeMenu[1].routine := @M_Episode;
  EpisodeMenu[1].alphaKey := 't';

  EpisodeMenu[2].status := 1;
  EpisodeMenu[2].name := 'M_EPI3';
  EpisodeMenu[2].routine := @M_Episode;
  EpisodeMenu[2].alphaKey := 'i';

  EpisodeMenu[3].status := 1;
  EpisodeMenu[3].name := 'M_EPI4';
  EpisodeMenu[3].routine := @M_Episode;
  EpisodeMenu[3].alphaKey := 't';

////////////////////////////////////////////////////////////////////////////////
//EpiDef
  EpiDef.numitems := Ord(ep_end); // # of menu items
  EpiDef.prevMenu := @MainDef; // previous menu
  EpiDef.menuitems := Pmenuitem_tArray(@EpisodeMenu);  // menu items
  EpiDef.routine := @M_DrawEpisode;  // draw routine
  EpiDef.x := 48;
  EpiDef.y := 63; // x,y of menu
  EpiDef.lastOn := Ord(ep1); // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//NewGameMenu
  NewGameMenu[0].status := 1;
  NewGameMenu[0].name := 'M_JKILL';
  NewGameMenu[0].routine := @M_ChooseSkill;
  NewGameMenu[0].alphaKey := 'i';

  NewGameMenu[1].status := 1;
  NewGameMenu[1].name := 'M_ROUGH';
  NewGameMenu[1].routine := @M_ChooseSkill;
  NewGameMenu[1].alphaKey := 'h';

  NewGameMenu[2].status := 1;
  NewGameMenu[2].name := 'M_HURT';
  NewGameMenu[2].routine := @M_ChooseSkill;
  NewGameMenu[2].alphaKey := 'h';

  NewGameMenu[3].status := 1;
  NewGameMenu[3].name := 'M_ULTRA';
  NewGameMenu[3].routine := @M_ChooseSkill;
  NewGameMenu[3].alphaKey := 'u';

  NewGameMenu[4].status := 1;
  NewGameMenu[4].name := 'M_NMARE';
  NewGameMenu[4].routine := @M_ChooseSkill;
  NewGameMenu[4].alphaKey := 'n';

////////////////////////////////////////////////////////////////////////////////
//NewDef
  NewDef.numitems := Ord(newg_end); // # of menu items
  NewDef.prevMenu := @EpiDef; // previous menu
  NewDef.menuitems := Pmenuitem_tArray(@NewGameMenu);  // menu items
  NewDef.routine := @M_DrawNewGame;  // draw routine
  NewDef.x := 48;
  NewDef.y := 63; // x,y of menu
  NewDef.lastOn := Ord(hurtme); // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//OptionsMenu
  OptionsMenu[0].status := 1;
  OptionsMenu[0].name := 'M_ENDGAM';
  OptionsMenu[0].routine := @M_EndGame;
  OptionsMenu[0].alphaKey := 'e';

  OptionsMenu[1].status := 1;
  OptionsMenu[1].name := 'M_MESSG';
  OptionsMenu[1].routine := @M_ChangeMessages;
  OptionsMenu[1].alphaKey := 'm';

  OptionsMenu[2].status := 1;
  OptionsMenu[2].name := 'M_DETAIL';
  OptionsMenu[2].routine := @M_ChangeDetail;
  OptionsMenu[2].alphaKey := 'g';

  OptionsMenu[3].status := 2;
  OptionsMenu[3].name := 'M_SCRNSZ';
  OptionsMenu[3].routine := @M_SizeDisplay;
  OptionsMenu[3].alphaKey := 's';

  OptionsMenu[4].status := -1;
  OptionsMenu[4].name := '';
  OptionsMenu[4].routine := nil;
  OptionsMenu[4].alphaKey := #0;

  OptionsMenu[5].status := 2;
  OptionsMenu[5].name := 'M_MSENS';
  OptionsMenu[5].routine := @M_ChangeSensitivity;
  OptionsMenu[5].alphaKey := 'm';

  OptionsMenu[6].status := -1;
  OptionsMenu[6].name := '';
  OptionsMenu[6].routine := nil;
  OptionsMenu[6].alphaKey := #0;

  OptionsMenu[7].status := 1;
  OptionsMenu[7].name := 'M_SVOL';
  OptionsMenu[7].routine := @M_Sound;
  OptionsMenu[7].alphaKey := 's';

////////////////////////////////////////////////////////////////////////////////
//OptionsDef
  OptionsDef.numitems := Ord(opt_end); // # of menu items
  OptionsDef.prevMenu := @MainDef; // previous menu
  OptionsDef.menuitems := Pmenuitem_tArray(@OptionsMenu);  // menu items
  OptionsDef.routine := @M_DrawOptions;  // draw routine
  OptionsDef.x := 60;
  OptionsDef.y := 37; // x,y of menu
  OptionsDef.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//ReadMenu1
  ReadMenu1[0].status := 1;
  ReadMenu1[0].name := '';
  ReadMenu1[0].routine := @M_ReadThis2;
  ReadMenu1[0].alphaKey := #0;

////////////////////////////////////////////////////////////////////////////////
//ReadDef1
  ReadDef1.numitems := Ord(read1_end); // # of menu items
  ReadDef1.prevMenu := @MainDef; // previous menu
  ReadDef1.menuitems := Pmenuitem_tArray(@ReadMenu1);  // menu items
  ReadDef1.routine := @M_DrawReadThis1;  // draw routine
  ReadDef1.x := 280;
  ReadDef1.y := 185; // x,y of menu
  ReadDef1.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//ReadMenu2
  ReadMenu2[0].status := 1;
  ReadMenu2[0].name := '';
  ReadMenu2[0].routine := @M_FinishReadThis;
  ReadMenu2[0].alphaKey := #0;

////////////////////////////////////////////////////////////////////////////////
//ReadDef2
  ReadDef2.numitems := Ord(read2_end); // # of menu items
  ReadDef2.prevMenu := @ReadDef1; // previous menu
  ReadDef2.menuitems := Pmenuitem_tArray(@ReadMenu2);  // menu items
  ReadDef2.routine := @M_DrawReadThis2;  // draw routine
  ReadDef2.x := 330;
  ReadDef2.y := 175; // x,y of menu
  ReadDef2.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//SoundMenu
  SoundMenu[0].status := 2;
  SoundMenu[0].name := 'M_SFXVOL';
  SoundMenu[0].routine := @M_SfxVol;
  SoundMenu[0].alphaKey := 's';

  SoundMenu[1].status := -1;
  SoundMenu[1].name := '';
  SoundMenu[1].routine := nil;
  SoundMenu[1].alphaKey := #0;

  SoundMenu[2].status := 2;
  SoundMenu[2].name := 'M_MUSVOL';
  SoundMenu[2].routine := @M_MusicVol;
  SoundMenu[2].alphaKey := 'm';

  SoundMenu[3].status := -1;
  SoundMenu[3].name := '';
  SoundMenu[3].routine := nil;
  SoundMenu[3].alphaKey := #0;

////////////////////////////////////////////////////////////////////////////////
//SoundDef
  SoundDef.numitems := Ord(sound_end); // # of menu items
  SoundDef.prevMenu := @OptionsDef; // previous menu
  SoundDef.menuitems := Pmenuitem_tArray(@SoundMenu);  // menu items
  SoundDef.routine := @M_DrawSound;  // draw routine
  SoundDef.x := 80;
  SoundDef.y := 64; // x,y of menu
  SoundDef.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//LoadMenu
  LoadMenu[0].status := 1;
  LoadMenu[0].name := '';
  LoadMenu[0].routine := @M_LoadSelect;
  LoadMenu[0].alphaKey := '1';

  LoadMenu[1].status := 1;
  LoadMenu[1].name := '';
  LoadMenu[1].routine := @M_LoadSelect;
  LoadMenu[1].alphaKey := '2';

  LoadMenu[2].status := 1;
  LoadMenu[2].name := '';
  LoadMenu[2].routine := @M_LoadSelect;
  LoadMenu[2].alphaKey := '3';

  LoadMenu[3].status := 1;
  LoadMenu[3].name := '';
  LoadMenu[3].routine := @M_LoadSelect;
  LoadMenu[3].alphaKey := '4';

  LoadMenu[4].status := 1;
  LoadMenu[4].name := '';
  LoadMenu[4].routine := @M_LoadSelect;
  LoadMenu[4].alphaKey := '5';

  LoadMenu[5].status := 1;
  LoadMenu[5].name := '';
  LoadMenu[5].routine := @M_LoadSelect;
  LoadMenu[5].alphaKey := '6';

////////////////////////////////////////////////////////////////////////////////
//LoadDef
  LoadDef.numitems := Ord(load_end); // # of menu items
  LoadDef.prevMenu := @MainDef; // previous menu
  LoadDef.menuitems := Pmenuitem_tArray(@LoadMenu);  // menu items
  LoadDef.routine := @M_DrawLoad;  // draw routine
  LoadDef.x := 80;
  LoadDef.y := 54; // x,y of menu
  LoadDef.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
//SaveMenu
  SaveMenu[0].status := 1;
  SaveMenu[0].name := '';
  SaveMenu[0].routine := @M_SaveSelect;
  SaveMenu[0].alphaKey := '1';

  SaveMenu[1].status := 1;
  SaveMenu[1].name := '';
  SaveMenu[1].routine := @M_SaveSelect;
  SaveMenu[1].alphaKey := '2';

  SaveMenu[2].status := 1;
  SaveMenu[2].name := '';
  SaveMenu[2].routine := @M_SaveSelect;
  SaveMenu[2].alphaKey := '3';

  SaveMenu[3].status := 1;
  SaveMenu[3].name := '';
  SaveMenu[3].routine := @M_SaveSelect;
  SaveMenu[3].alphaKey := '4';

  SaveMenu[4].status := 1;
  SaveMenu[4].name := '';
  SaveMenu[4].routine := @M_SaveSelect;
  SaveMenu[4].alphaKey := '5';

  SaveMenu[5].status := 1;
  SaveMenu[5].name := '';
  SaveMenu[5].routine := @M_SaveSelect;
  SaveMenu[5].alphaKey := '6';

////////////////////////////////////////////////////////////////////////////////
//SaveDef
  SaveDef.numitems := Ord(load_end); // # of menu items
  SaveDef.prevMenu := @MainDef; // previous menu
  SaveDef.menuitems := Pmenuitem_tArray(@SaveMenu);  // menu items
  SaveDef.routine := M_DrawSave;  // draw routine
  SaveDef.x := 80;
  SaveDef.y := 54; // x,y of menu
  SaveDef.lastOn := 0; // last item user was on in menu

////////////////////////////////////////////////////////////////////////////////
  detailNames[0] := 'M_GDHIGH';
  detailNames[1] := 'M_GDLOW';
  msgNames[0] := 'M_MSGOFF';
  msgNames[1] := 'M_MSGON';

////////////////////////////////////////////////////////////////////////////////
  joywait := 0;
  mousewait := 0;
  mousey := 0;
  lasty := 0;
  mousex := 0;
  lastx := 0;

end;


(* Unit f_finale implementation *)


                                                                                                                                                                                                                                                                        

var
// Stage of animation:
//  0 = text, 1 = art screen, 2 = character cast
  finalestage: integer;

  finalecount: integer;

const
  TEXTSPEED = 3;
  TEXTWAIT = 250;

var
  finaletext: string;
  finaleflat: string;

procedure F_StartCast; forward;

procedure F_CastTicker; forward;

function F_CastResponder(ev: Pevent_t): boolean; forward;

procedure F_CastDrawer; forward;

procedure F_StartFinale;
begin
  gameaction := ga_nothing;
  gamestate := GS_FINALE;
  viewactive := false;
  automapactive := false;

  // Okay - IWAD dependend stuff.
  // This has been changed severly, and
  //  some stuff might have changed in the process.
  case gamemode of
    // DOOM 1 - E1, E3 or E4, but each nine missions
    shareware,
    registered,
    retail:
      begin
        S_ChangeMusic(Ord(mus_victor), true);
        case gameepisode of
          1:
            begin
              finaleflat := 'FLOOR4_8';
              finaletext := E1TEXT;
            end;
          2:
            begin
              finaleflat := 'SFLR6_1';
              finaletext := E2TEXT;
            end;
          3:
            begin
              finaleflat := 'MFLR8_4';
              finaletext := E3TEXT;
            end;
          4:
            begin
              finaleflat := 'MFLR8_3';
              finaletext := E4TEXT;
            end;
        else
          // Ouch.
        end;
      end;
    // DOOM II and missions packs with E1, M34
    commercial:
      begin
        S_ChangeMusic(Ord(mus_read_m), true);
        case gamemap of
          6:
            begin
              finaleflat := 'SLIME16';
              finaletext := C1TEXT;
            end;
         11:
            begin
              finaleflat := 'RROCK14';
              finaletext := c2text;
            end;
         20:
            begin
              finaleflat := 'RROCK07';
              finaletext := C3TEXT;
            end;
         30:
            begin
              finaleflat := 'RROCK17';
              finaletext := C4TEXT;
            end;
         15:
            begin
              finaleflat := 'RROCK13';
              finaletext := C5TEXT;
            end;
         31:
            begin
              finaleflat := 'RROCK19';
              finaletext := C6TEXT;
            end;
        else
        // Ouch.
        end;
      end;
  else
    begin
      S_ChangeMusic(Ord(mus_read_m), true);
      finaleflat := 'F_SKY1'; // Not used anywhere else.
      finaletext := C1TEXT;   // FIXME - other text, music?
    end;
  end;
  finalestage := 0;
  finalecount := 0;
end;

function F_Responder(ev: Pevent_t): boolean;
begin
  result := false;
  if finalestage = 2 then
    result := F_CastResponder(ev);
end;

//
// F_Ticker
//
procedure F_Ticker;
var
  i: integer;
begin
  // check for skipping
  if (gamemode = commercial) and (finalecount > 50) then
  begin
    // go on to the next level
    i := 0;
    while i < MAXPLAYERS do
    begin
      if boolval(players[i].cmd.buttons) then
        break;
      inc(i);
    end;
    if i < MAXPLAYERS then
    begin
      if gamemap = 30 then
        F_StartCast
      else
        gameaction := ga_worlddone;
    end;
  end;

  // advance animation
  inc(finalecount);

  if finalestage = 2 then
  begin
    F_CastTicker;
    exit;
  end;

  if gamemode = commercial then
    exit;

  if (not boolval(finalestage)) and (finalecount > Length(finaletext) * TEXTSPEED + TEXTWAIT) then
  begin
    finalecount := 0;
    finalestage := 1;
    wipegamestate := -1;    // force a wipe
    if gameepisode = 3 then
      S_StartMusic(Ord(mus_bunny));
  end;
end;

procedure F_TextWrite;
var
  src: PByteArray;
  dest: integer;
  x, y, w: integer;
  count: integer;
  ch: string;
  c: char;
  c1: integer;
  i: integer;
  len: integer;
  cx: integer;
  cy: integer;
  dstscr: integer;
begin
  // erase the entire screen to a tiled background
  src := W_CacheLumpName(finaleflat, PU_CACHE);
  dest := 0;
  if V_NeedsPreserve then
    dstscr := _SCRF
  else
    dstscr := _FG;

  for y := 0 to 200 - 1 do
  begin
    for x := 0 to (320 div 64) - 1 do
    begin
      memcpy(@screens[dstscr, dest], @src[_SHL(y and 63, 6)], 64);
      dest := dest + 64;
    end;

    if (320 and 63) <> 0 then
    begin
      memcpy(@screens[dstscr, dest], @src[_SHL(y and 63, 6)], 320 and 63);
      dest := dest + (320 and 63);
    end;
  end;

//  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT, false);

  // draw some of the text onto the screen
  cx := 10;
  cy := 10;
  ch := finaletext;
  len := Length(ch);

  count := (finalecount - 10) div TEXTSPEED;
  if count < 0 then
    count := 0;

  i := 1;
  while count > 0 do
  begin

    if i > len then
      break;

    c := ch[i];
    inc(i);
    if c = #13 then
    begin
      cy := cy + 11;
      continue;
    end;
    if c = #10 then
    begin
      cx := 10;
      continue;
    end;

    c1 := Ord(toupper(c)) - Ord(HU_FONTSTART);
    if (c1 < 0) or (c1 > HU_FONTSIZE) then
    begin
      cx := cx + 4;
      continue;
    end;

    w := hu_font[c1].width;
    if cx + w > 320 then
      break;
    V_DrawPatch(cx, cy, dstscr, hu_font[c1], false);
    cx := cx + w;
    dec(count);
  end;
  if dstscr = _SCRF then
    V_CopyRect(0, 0, _SCRF, 320, 200, 0, 0, _FG, true);
end;


//
// Final DOOM 2 animation
// Casting by id Software.
//   in order of appearance
//
type
  castinfo_t = record
    name: string;
    _type: mobjtype_t;
  end;
  Pcastinfo_t = ^castinfo_t;

const
  NUM_CASTS = 18;

var
  castorder: array[0..NUM_CASTS - 1] of castinfo_t;

  castnum: integer;
  casttics: integer;
  caststate: Pstate_t;
  castdeath: boolean;
  castframes: integer;
  castonmelee: integer;
  castattacking: boolean;

//
// F_StartCast
//
procedure F_StartCast;
begin
  wipegamestate := -1;    // force a screen wipe
  castnum := 0;
  caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].seestate];
  casttics := caststate.tics;
  castdeath := false;
  finalestage := 2;
  castframes := 0;
  castonmelee := 0;
  castattacking := false;
  S_ChangeMusic(Ord(mus_evil), true);
end;

//
// F_CastTicker
//
procedure F_CastTicker;
var
  st: integer;
  sfx: integer;
begin
  dec(casttics);
  if casttics > 0 then
    exit; // not time to change state yet

  if (caststate.tics = -1) or (caststate.nextstate = S_NULL) then
  begin
    // switch from deathstate to next monster
    inc(castnum);
    castdeath := false;
    if castorder[castnum].name = '' then
      castnum := 0;
    if boolval(mobjinfo[Ord(castorder[castnum]._type)].seesound) then
      S_StartSound(nil, mobjinfo[Ord(castorder[castnum]._type)].seesound);
    caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].seestate];
    castframes := 0;
  end
  else
  begin
  // just advance to next state in animation
    if caststate = @states[Ord(S_PLAY_ATK1)] then
    begin
      castattacking := false;
      castframes := 0;
      caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].seestate];
      casttics := caststate.tics;
      if casttics = -1 then
        casttics := 15;
      exit;
    end;
    st := Ord(caststate.nextstate);
    caststate := @states[st];
    inc(castframes);

    // sound hacks....
    case statenum_t(st) of
      S_PLAY_ATK1:  sfx := Ord(sfx_dshtgn);
      S_POSS_ATK2:  sfx := Ord(sfx_pistol);
      S_SPOS_ATK2:  sfx := Ord(sfx_shotgn);
      S_VILE_ATK2:  sfx := Ord(sfx_vilatk);
      S_SKEL_FIST2:  sfx := Ord(sfx_skeswg);
      S_SKEL_FIST4:  sfx := Ord(sfx_skepch);
      S_SKEL_MISS2:  sfx := Ord(sfx_skeatk);
      S_FATT_ATK8,
      S_FATT_ATK5,
      S_FATT_ATK2:  sfx := Ord(sfx_firsht);
      S_CPOS_ATK2,
      S_CPOS_ATK3,
      S_CPOS_ATK4:  sfx := Ord(sfx_shotgn);
      S_TROO_ATK3:  sfx := Ord(sfx_claw);
      S_SARG_ATK2:  sfx := Ord(sfx_sgtatk);
      S_BOSS_ATK2,
      S_BOS2_ATK2,
      S_HEAD_ATK2:  sfx := Ord(sfx_firsht);
      S_SKULL_ATK2:  sfx := Ord(sfx_sklatk);
      S_SPID_ATK2,
      S_SPID_ATK3:  sfx := Ord(sfx_shotgn);
      S_BSPI_ATK2:  sfx := Ord(sfx_plasma);
      S_CYBER_ATK2,
      S_CYBER_ATK4,
      S_CYBER_ATK6:  sfx := Ord(sfx_rlaunc);
      S_PAIN_ATK3:  sfx := Ord(sfx_sklatk);
    else
      sfx := 0;
    end;
    if sfx <> 0 then
      S_StartSound(nil, sfx);
  end;

  if castframes = 12 then
  begin
    // go into attack frame
    castattacking := true;
    if castonmelee <> 0 then
      caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].meleestate]
    else
      caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].missilestate];
    castonmelee := castonmelee xor 1;
    if caststate = @states[Ord(S_NULL)] then
    begin
      if castonmelee <> 0 then
        caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].meleestate]
      else
        caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].missilestate];
    end;
  end;

  if castattacking then
  begin
    if (castframes = 24) or
       (caststate = @states[mobjinfo[Ord(castorder[castnum]._type)].seestate]) then
    begin
      castattacking := false;
      castframes := 0;
      caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].seestate];
    end;
  end;

  casttics := caststate.tics;
  if casttics = -1 then
    casttics := 15;
end;

//
// F_CastResponder
//
function F_CastResponder(ev: Pevent_t): boolean;
begin
  if ev._type <> ev_keydown then
  begin
    result := false;
    exit;
  end;

  if castdeath then
  begin
    result := true; // already in dying frames
    exit;
  end;

  // go into death frame
  castdeath := true;
  caststate := @states[mobjinfo[Ord(castorder[castnum]._type)].deathstate];
  casttics := caststate.tics;
  castframes := 0;
  castattacking := false;
  if mobjinfo[Ord(castorder[castnum]._type)].deathsound <> 0 then
    S_StartSound(nil, mobjinfo[Ord(castorder[castnum]._type)].deathsound);

  result := true;
end;

procedure F_CastPrint(text: string);
var
  ch: string;
  i: integer;
  c: char;
  c1: integer;
  len: integer;
  cx: integer;
  w: integer;
  width: integer;
begin
  // find width
  ch := text;
  width := 0;

  len := Length(ch);
  for i := 1 to len do
  begin
    c := ch[i];
    if c = #0 then
      break;
    c1 := Ord(toupper(c)) - Ord(HU_FONTSTART);
    if (c1 < 0) or (c1 > HU_FONTSIZE) then
      width := width + 4
    else
    begin
      w := hu_font[c1].width;
      width := width + w;
    end;
  end;

  // draw it
  cx := (320 - width) div 2;
  for i := 1 to len do
  begin
    c := ch[i];
    if c = #0 then
      break;
    c1 := Ord(toupper(c)) - Ord(HU_FONTSTART);
    if (c1 < 0) or (c1 > HU_FONTSIZE) then
      cx := cx + 4
    else
    begin
      w := hu_font[c1].width;
      V_DrawPatch(cx, 180, 0, hu_font[c1], true);
      cx := cx + w;
    end;
  end;
end;

//
// F_CastDrawer
//
procedure F_CastDrawer;
var
  sprdef: Pspritedef_t;
  sprframe: Pspriteframe_t;
  lump: integer;
  flip: boolean;
  patch: Ppatch_t;
begin
  // erase the entire screen to a background
  V_DrawPatch(0, 0, 0, W_CacheLumpName('BOSSBACK', PU_CACHE), true);

  F_CastPrint(castorder[castnum].name);

  // draw the current frame in the middle of the screen
  sprdef := @sprites[Ord(caststate.sprite)];
  sprframe := @sprdef.spriteframes[caststate.frame and FF_FRAMEMASK];
  lump := sprframe.lump[0];
  flip := boolean(sprframe.flip[0]);

  patch := W_CacheLumpNum(lump + firstspritelump, PU_CACHE);
{  if flip then
    V_DrawPatchFlipped(SCREENWIDTH div 2, SCREENHEIGHT - 30, 0, patch)
  else
    V_DrawPatch(SCREENWIDTH div 2, SCREENHEIGHT - 30, 0, patch);}
  if flip then
    V_DrawPatchFlipped(160, 170, 0, patch, true)
  else
    V_DrawPatch(160, 170, 0, patch, true);
end;

//
// F_DrawPatchCol
//
procedure F_DrawPatchCol(scr: integer; x: integer; patch: Ppatch_t; col: integer);
var
  column: Pcolumn_t;
  source: PByte;
  dest: PByte;
  desttop: PByte;
  count: integer;
begin
  column := Pcolumn_t(integer(patch) + patch.columnofs[col]);
  desttop := PByte(integer(screens[scr]) + x);

  // step through the posts in a column
  while column.topdelta <> $ff do
  begin
    source := PByte(integer(column) + 3);
    dest := PByte(integer(desttop) + column.topdelta * 320);
    count := column.length;

    while count > 0 do
    begin
      dest^ := source^;
      inc(source);
      inc(dest, 320);
      dec(count);
    end;
    column := Pcolumn_t(integer(column) + column.length + 4);
  end;
end;

//
// F_BunnyScroll
//
var
  laststage: integer;

procedure F_BunnyScroll;
var
  scrolled: integer;
  x: integer;
  p1: Ppatch_t;
  p2: Ppatch_t;
  name: string;
  stage: integer;
  dstscr: integer;
begin
  p1 := W_CacheLumpName('PFUB2', PU_LEVEL);
  p2 := W_CacheLumpName('PFUB1', PU_LEVEL);

  if V_NeedsPreserve then
    dstscr := _SCRF
  else
    dstscr := _FG;

  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT, false);

  scrolled := 320 - (finalecount - 230) div 2;
  if scrolled > 320 then
    scrolled := 320
  else if scrolled < 0 then
    scrolled := 0;

  for x := 0 to 320 - 1 do
  begin
    if x + scrolled < 320 then
      F_DrawPatchCol(dstscr, x, p1, x + scrolled)
    else
      F_DrawPatchCol(dstscr, x, p2, x + scrolled - 320);
  end;

  if finalecount >= 1130 then
  begin
    if finalecount < 1180 then
    begin
      V_DrawPatch((320 - 13 * 8) div 2,
                  (200 - 8 * 8) div 2,
                   dstscr, W_CacheLumpName('END0', PU_CACHE), false);
      laststage := 0;
    end
    else
    begin
      stage := (finalecount - 1180) div 5;
      if stage > 6 then
        stage := 6;
      if stage > laststage then
      begin
        S_StartSound(nil, Ord(sfx_pistol));
        laststage := stage;
      end;

      sprintf(name,'END%d', [stage]);
      V_DrawPatch((320 - 13 * 8) div 2,
                  (200 - 8 * 8) div 2,
                   dstscr, W_CacheLumpName(name, PU_CACHE), false);
    end;
  end;

  if dstscr = _SCRF then
    V_CopyRect(0, 0, _SCRF, 320, 200, 0, 0, _FG, true);
end;

//
// F_Drawer
//
procedure F_Drawer;
begin
  if finalestage = 2 then
  begin
    F_CastDrawer;
    exit;
  end;

  if finalestage = 0 then
    F_TextWrite
  else
  begin
    case gameepisode of
      1:
        begin
          if gamemode = retail then
            V_DrawPatch(0, 0, 0,
              W_CacheLumpName('CREDIT', PU_CACHE), true)
          else
            V_DrawPatch(0, 0, 0,
              W_CacheLumpName('HELP2', PU_CACHE), true);
        end;
      2:
        begin
          V_DrawPatch(0, 0, 0,
            W_CacheLumpName('VICTORY2', PU_CACHE), true);
        end;
      3:
        begin
          F_BunnyScroll;
        end;
      4:
        begin
          V_DrawPatch(0, 0, 0,
            W_CacheLumpName('ENDPIC', PU_CACHE), true);
        end;
    end;
  end;
end;


(* Unit doomdef implementation *)



(* Unit doomstat implementation *)



(* Unit st_stuff implementation *)


                                                                                                                                                                                                                                                                                                                                             

//
// STATUS BAR DATA
//

const
// Palette indices.
// For damage/bonus red-/gold-shifts
  STARTREDPALS = 1;
  STARTBONUSPALS = 9;
  NUMREDPALS = 8;
  NUMBONUSPALS = 4;
// Radiation suit, green shift.
  RADIATIONPAL = 13;

// N/256*100% probability
//  that the normal face state will change
  ST_FACEPROBABILITY = 96;

// For Responder
  ST_TOGGLECHAT = KEY_ENTER;

// Location of status bar
  ST_X = 0;
  ST_X2 = 104;

  ST_FX = 143;
  ST_FY = ST_Y + 1; // VJ was 169;

// Number of status faces.
  ST_NUMPAINFACES = 5;
  ST_NUMSTRAIGHTFACES = 3;
  ST_NUMTURNFACES = 2;
  ST_NUMSPECIALFACES = 3;

  ST_FACESTRIDE = ST_NUMSTRAIGHTFACES + ST_NUMTURNFACES + ST_NUMSPECIALFACES;

  ST_NUMEXTRAFACES = 2;

  ST_NUMFACES = ST_FACESTRIDE * ST_NUMPAINFACES + ST_NUMEXTRAFACES;

  ST_TURNOFFSET = ST_NUMSTRAIGHTFACES;
  ST_OUCHOFFSET = ST_TURNOFFSET + ST_NUMTURNFACES;
  ST_EVILGRINOFFSET = ST_OUCHOFFSET + 1;
  ST_RAMPAGEOFFSET = ST_EVILGRINOFFSET + 1;
  ST_GODFACE = ST_NUMPAINFACES * ST_FACESTRIDE;
  ST_DEADFACE = ST_GODFACE + 1;

  ST_FACESX = 143;
  ST_FACESY = ST_Y; // VJ was 168;

  ST_EVILGRINCOUNT = 2 * TICRATE;
  ST_STRAIGHTFACECOUNT = TICRATE div 2;
  ST_TURNCOUNT = 1 * TICRATE;
  ST_OUCHCOUNT = 1 * TICRATE;
  ST_RAMPAGEDELAY = 2 * TICRATE;

  ST_MUCHPAIN = 20;


// Location and size of statistics,
//  justified according to widget type.
// Problem is, within which space? STbar? Screen?
// Note: this could be read in by a lump.
//       Problem is, is the stuff rendered
//       into a buffer,
//       or into the frame buffer?

// AMMO number pos.
  ST_AMMOWIDTH = 3;
  ST_AMMOX = 44;
  ST_AMMOY = ST_Y + 3; // VJ was 171;

// HEALTH number pos.
  ST_HEALTHWIDTH = 3;
  ST_HEALTHX = 90;
  ST_HEALTHY = ST_Y + 3; // VJ was 171;

// Weapon pos.
  ST_ARMSX = 111;
  ST_ARMSY = ST_Y + 4; // VJ was 172;
  ST_ARMSBGX = 104;
  ST_ARMSBGY = ST_Y; // VJ was 168;
  ST_ARMSXSPACE = 12;
  ST_ARMSYSPACE = 10;

// Frags pos.
  ST_FRAGSX = 138;
  ST_FRAGSY = ST_Y + 3; // VJ was 171;
  ST_FRAGSWIDTH = 2;

// ARMOR number pos.
  ST_ARMORWIDTH = 3;
  ST_ARMORX = 221;
  ST_ARMORY = ST_Y + 3; // VJ was 171;

// Key icon positions.
  ST_KEY0WIDTH = 8;
  ST_KEY0HEIGHT = 5;
  ST_KEY0X = 239;
  ST_KEY0Y = ST_Y + 3; // VJ was 171;
  ST_KEY1WIDTH = ST_KEY0WIDTH;
  ST_KEY1X = 239;
  ST_KEY1Y = ST_Y + 13; // VJ was 181;
  ST_KEY2WIDTH = ST_KEY0WIDTH;
  ST_KEY2X = 239;
  ST_KEY2Y = ST_Y + 23; // VJ was 191;

// Ammunition counter.
  ST_AMMO0WIDTH = 3;
  ST_AMMO0HEIGHT = 6;
  ST_AMMO0X = 288;
  ST_AMMO0Y = ST_Y + 5; // VJ was 173;
  ST_AMMO1WIDTH = ST_AMMO0WIDTH;
  ST_AMMO1X = 288;
  ST_AMMO1Y = ST_Y + 11; // VJ was 179;
  ST_AMMO2WIDTH = ST_AMMO0WIDTH;
  ST_AMMO2X = 288;
  ST_AMMO2Y = ST_Y + 23; // VJ was 191;
  ST_AMMO3WIDTH = ST_AMMO0WIDTH;
  ST_AMMO3X = 288;
  ST_AMMO3Y = ST_Y + 17; // VJ was 185;

// Indicate maximum ammunition.
// Only needed because backpack exists.
  ST_MAXAMMO0WIDTH = 3;
  ST_MAXAMMO0HEIGHT = 5;
  ST_MAXAMMO0X = 314;
  ST_MAXAMMO0Y = ST_Y + 5; // VJ was 173;
  ST_MAXAMMO1WIDTH = ST_MAXAMMO0WIDTH;
  ST_MAXAMMO1X = 314;
  ST_MAXAMMO1Y = ST_Y + 11; // VJ was 179;
  ST_MAXAMMO2WIDTH = ST_MAXAMMO0WIDTH;
  ST_MAXAMMO2X = 314;
  ST_MAXAMMO2Y = ST_Y + 23; // VJ was 191;
  ST_MAXAMMO3WIDTH = ST_MAXAMMO0WIDTH;
  ST_MAXAMMO3X = 314;
  ST_MAXAMMO3Y = ST_Y + 17; // VJ was 185;

// pistol
  ST_WEAPON0X = 110;
  ST_WEAPON0Y = ST_Y + 4; // VJ was 172;

// shotgun
  ST_WEAPON1X = 122;
  ST_WEAPON1Y = ST_Y + 4; // VJ was 172;

// chain gun
  ST_WEAPON2X = 134;
  ST_WEAPON2Y = ST_Y + 4; // VJ was 172;

// missile launcher
  ST_WEAPON3X = 110;
  ST_WEAPON3Y = ST_Y + 13; // VJ was 181;

// plasma gun
  ST_WEAPON4X = 122;
  ST_WEAPON4Y = ST_Y + 13; // VJ was 181;

 // bfg
  ST_WEAPON5X = 134;
  ST_WEAPON5Y = ST_Y + 13; // VJ was 181;

// WPNS title
  ST_WPNSX = 109;
  ST_WPNSY = ST_Y + 23; // VJ was 191;

 // DETH title
  ST_DETHX = 109;
  ST_DETHY = ST_Y + 23; // VJ was 191;

//Incoming messages window location
//UNUSED
//   ST_MSGTEXTX	   (viewwindowx)
//   ST_MSGTEXTY	   (viewwindowy+viewheight-18)
  ST_MSGTEXTX = 0;
  ST_MSGTEXTY = 0;
// Dimensions given in characters.
  ST_MSGWIDTH = 52;
// Or shall I say, in lines?
  ST_MSGHEIGHT = 1;

  ST_OUTTEXTX = 0;
  ST_OUTTEXTY = 6;

// Width, in characters again.
  ST_OUTWIDTH = 52;
 // Height, in lines.
  ST_OUTHEIGHT = 1;


  ST_MAPTITLEY = 0;
  ST_MAPHEIGHT = 1;

var

// main player in game
  plyr: Pplayer_t;

// ST_Start() has just been called
  st_firsttime: boolean;

// used to execute ST_Init() only once
  veryfirsttime: integer;

// lump number for PLAYPAL
  lu_palette: integer;

// used for timing
  st_clock: LongWord;

// used for making messages go away
  st_msgcounter: integer;

// used when in chat
  st_chatstate: st_chatstateenum_t;

// whether in automap or first-person
  st_gamestate: st_stateenum_t;

// whether left-side main status bar is active
  st_statusbaron: boolean;

// whether status bar chat is active
  st_chat: boolean;

// value of st_chat before message popped up
  st_oldchat: boolean;

// whether chat window has the cursor on
  st_cursoron: boolean;

// !deathmatch
  st_notdeathmatch: boolean;

// !deathmatch && st_statusbaron
  st_armson: boolean;

// !deathmatch
  st_fragson: boolean;

// main bar left
  sbar: Ppatch_t;

// 0-9, tall numbers
  tallnum: array[0..9] of Ppatch_t;

// tall % sign
  tallpercent: Ppatch_t;

// 0-9, short, yellow (,different!) numbers
  shortnum: array[0..9] of Ppatch_t;

// 3 key-cards, 3 skulls
  keys: array[0..Ord(NUMCARDS) - 1] of Ppatch_t;

// face status patches
  faces: array[0..ST_NUMFACES - 1] of Ppatch_t;

// face background
  faceback: Ppatch_t;

 // main bar right
  armsbg: Ppatch_t;

// weapon ownership patches
  arms: array[0..5, 0..1] of Ppatch_t;

// ready-weapon widget
  w_ready: st_number_t;

 // in deathmatch only, summary of frags stats
  w_frags: st_number_t;

// health widget
  w_health: st_percent_t;

// arms background
  w_armsbg: st_binicon_t;


// weapon ownership widgets
  w_arms: array[0..5] of st_multicon_t;

// face status widget
  w_faces: st_multicon_t;

// keycard widgets
  w_keyboxes: array[0..2] of st_multicon_t;

// armor widget
  w_armor: st_percent_t;

// ammo widgets
  w_ammo: array[0..3] of st_number_t;

// max ammo widgets
  w_maxammo: array[0..3] of st_number_t;



// number of frags so far in deathmatch
  st_fragscount: integer;

// used to use appopriately pained face
  st_oldhealth: integer;

// used for evil grin
  oldweaponsowned: array[0..Ord(NUMWEAPONS) - 1] of integer;

 // count until face changes
  st_facecount: integer;

// current face index, used by w_faces
  st_faceindex: integer;

// holds key-type for each key box on bar
  keyboxes: array[0..2] of integer;

// a random number per tick
  st_randomnumber: integer;  


const
// Massive bunches of cheat shit
//  to keep it from being easy to figure them out.
// Yeah, right...
  cheat_mus_seq: array[0..8] of char = (
    Chr($b2), Chr($26), Chr($b6), Chr($ae), Chr($ea),
    Chr($1),  Chr($0),  Chr($0),  Chr($ff)
  );

  cheat_choppers_seq: array[0..10] of char = (
    Chr($b2), Chr($26), Chr($e2), Chr($32), Chr($f6),
    Chr($2a), Chr($2a), Chr($a6), Chr($6a), Chr($ea),
    Chr($ff) // id...
  );

  cheat_god_seq: array[0..5] of char = (
    Chr($b2), Chr($26), Chr($26), Chr($aa), Chr($26),
    Chr($ff)  // iddqd
  );

  cheat_ammo_seq: array[0..5] of char = (
    Chr($b2), Chr($26), Chr($f2), Chr($66), Chr($a2),
    Chr($ff)	// idkfa
  );

  cheat_ammonokey_seq: array[0..4] of char = (
    Chr($b2), Chr($26), Chr($66), Chr($a2), Chr($ff) // idfa
  );

// Smashing Pumpkins Into Samml Piles Of Putried Debris.
  cheat_noclip_seq: array[0..10] of char = (
    Chr($b2), Chr($26), Chr($ea), Chr($2a), Chr($b2), // idspispopd
    Chr($ea), Chr($2a), Chr($f6), Chr($2a), Chr($26),
    Chr($ff)
  );

//
  cheat_commercial_noclip_seq: array[0..6] of char = (
    Chr($b2), Chr($26), Chr($e2), Chr($36), Chr($b2),
    Chr($2a), Chr($ff)	// idclip
  );



  cheat_powerup_seq0: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($6e), Chr($ff)  // beholdv
  );

  cheat_powerup_seq1: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($ea), Chr($ff)  // beholds
  );

  cheat_powerup_seq2: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($b2), Chr($ff)  // beholdi
  );

  cheat_powerup_seq3: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($6a), Chr($ff)  // beholdr
  );

  cheat_powerup_seq4: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($a2), Chr($ff)  // beholda
  );

  cheat_powerup_seq5: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($36), Chr($ff)  // beholdl
  );

  cheat_powerup_seq6: array[0..8] of char = (
    Chr($b2), Chr($26), Chr($62), Chr($a6), Chr($32),
    Chr($f6), Chr($36), Chr($26), Chr($ff)  // behold
  );


  cheat_clev_seq: array[0..9] of char = (
    Chr($b2), Chr($26), Chr($e2), Chr($36), Chr($a6),
    Chr($6e), Chr($1),  Chr($0),  Chr($0),  Chr($ff)  // idclev
  );

// my position cheat
  cheat_mypos_seq: array[0..7] of char = (
    Chr($b2), Chr($26), Chr($b6), Chr($ba), Chr($2a),
    Chr($f6), Chr($ea), Chr($ff) // idmypos
  );

var
// Now what?
  cheat_mus: cheatseq_t;
  cheat_god: cheatseq_t;
  cheat_ammo: cheatseq_t;
  cheat_ammonokey: cheatseq_t;
  cheat_noclip: cheatseq_t;
  cheat_commercial_noclip: cheatseq_t;

  cheat_powerup: array[0..6] of cheatseq_t;

  cheat_choppers: cheatseq_t;
  cheat_clev: cheatseq_t;
  cheat_mypos: cheatseq_t;


// Should be set to patch width
//  for tall numbers later on
function ST_TALLNUMWIDTH: integer;
begin
  result := tallnum[0].width;
end;

function ST_MAPWIDTH: integer;
begin
  result := Length(mapnames[(gameepisode - 1) * 9 + (gamemap - 1)]);
end;

//
// STATUS BAR CODE
//
procedure ST_RefreshBackground;
begin
  if st_statusbaron then
  begin
    V_DrawPatch(ST_X, 0, _BG, sbar, false);

    if netgame then
      V_DrawPatch(ST_FX, 0, _BG, faceback, false);
  end;
end;

procedure ST_FinishRefresh;
begin
  V_CopyRect(ST_X,
             0,
             _BG,
             ST_WIDTH,
             ST_HEIGHT,
             ST_X,
             ST_Y,
             _FG,
             true);
end;

// Respond to keyboard input events,
//  intercept cheats.
function ST_Responder(ev: Pevent_t): boolean;
var
  i: integer;
  buf: string;
  musnum: integer;
  epsd: integer;
  map: integer;
begin
  result := false;
  // Filter automap on/off.
  if (ev._type = ev_keyup) and
     ((ev.data1 and $ffff0000) = AM_MSGHEADER) then
  begin
    case ev.data1 of
      AM_MSGENTERED:
        begin
          st_gamestate := AutomapState;
          st_firsttime := true;
        end;

      AM_MSGEXITED:
        begin
          //	fprintf(stderr, "AM exited\n");
          st_gamestate := FirstPersonState;
        end;
    end;
  end
  // if a user keypress...
  else if ev._type = ev_keydown then
  begin
    if not netgame then
    begin
      // b. - enabled for more debug fun.
      // if (gameskill != sk_nightmare) {

      // 'dqd' cheat for toggleable god mode
      if cht_CheckCheat(@cheat_god, Chr(ev.data1)) then
      begin
        plyr.cheats := plyr.cheats xor CF_GODMODE;
        if boolval(plyr.cheats and CF_GODMODE) then
        begin
          if boolval(plyr.mo) then
            plyr.mo.health := 100;

          plyr.health := 100;
          plyr._message := STSTR_DQDON;
        end
        else
          plyr._message := STSTR_DQDOFF;
      end
      // 'fa' cheat for killer fucking arsenal
      else if cht_CheckCheat(@cheat_ammonokey, Chr(ev.data1)) then
      begin
        plyr.armorpoints := 200;
        plyr.armortype := 2;

        for i := 0 to Ord(NUMWEAPONS) - 1 do
          plyr.weaponowned[i] := 1;

        for i := 0 to Ord(NUMAMMO) - 1 do
          plyr.ammo[i] := plyr.maxammo[i];

        plyr._message := STSTR_FAADDED;
      end
      // 'kfa' cheat for key full ammo
      else if cht_CheckCheat(@cheat_ammo, Chr(ev.data1)) then
      begin
        plyr.armorpoints := 200;
        plyr.armortype := 2;

        for i := 0 to Ord(NUMWEAPONS) - 1 do
          plyr.weaponowned[i] := 1;

        for i := 0 to Ord(NUMAMMO) - 1 do
          plyr.ammo[i] := plyr.maxammo[i];

        for i := 0 to Ord(NUMCARDS) - 1 do
          plyr.cards[i] := true;

        plyr._message := STSTR_KFAADDED;
      end
      // 'mus' cheat for changing music
      else if cht_CheckCheat(@cheat_mus, Chr(ev.data1)) then
      begin
        plyr._message := STSTR_MUS;
        cht_GetParam(@cheat_mus, buf);

        if gamemode = commercial then
        begin
          musnum := Ord(mus_runnin) + (Ord(buf[1]) - Ord('0')) * 10 + Ord(buf[2]) - Ord('0') - 1;

          if (Ord(buf[1]) - Ord('0')) * 10 + Ord(buf[2]) - Ord('0') > 35 then
            plyr._message := STSTR_NOMUS
          else
            S_ChangeMusic(musnum, boolval(1));
        end
        else
        begin
          musnum := Ord(mus_e1m1) + (Ord(buf[1]) - Ord('1')) * 9 + Ord(buf[2]) - Ord('1');

          if (Ord(buf[1]) - Ord('1')) * 9 + Ord(buf[2]) - Ord('1') > 31 then
            plyr._message := STSTR_NOMUS
          else
            S_ChangeMusic(musnum, boolval(1));
        end;
      end
      // Simplified, accepting both "noclip" and "idspispopd".
      // no clipping mode cheat
      else if cht_CheckCheat(@cheat_noclip, Chr(ev.data1)) or
              cht_CheckCheat(@cheat_commercial_noclip, Chr(ev.data1)) then
      begin
        plyr.cheats := plyr.cheats xor CF_NOCLIP;

        if boolval(plyr.cheats and CF_NOCLIP) then
          plyr._message := STSTR_NCON
        else
          plyr._message := STSTR_NCOFF;
      end;
      // 'behold?' power-up cheats
      for i := 0 to 5 do
      begin
        if cht_CheckCheat(@cheat_powerup[i], Chr(ev.data1)) then
        begin
          if not boolval(plyr.powers[i]) then
            P_GivePower(plyr, i)
          else if i <> Ord(pw_strength) then
            plyr.powers[i] := 1
          else
            plyr.powers[i] := 0;

          plyr._message := STSTR_BEHOLDX;
        end;
      end;

      // 'behold' power-up menu
      if cht_CheckCheat(@cheat_powerup[6], Chr(ev.data1)) then
      begin
        plyr._message := STSTR_BEHOLD;
      end
      // 'choppers' invulnerability & chainsaw
      else if cht_CheckCheat(@cheat_choppers, Chr(ev.data1)) then
      begin
        plyr.weaponowned[Ord(wp_chainsaw)] := 1;
        plyr.powers[Ord(pw_invulnerability)] := 1;
        plyr._message := STSTR_CHOPPERS;
      end
      // 'mypos' for player position
      else if cht_CheckCheat(@cheat_mypos, Chr(ev.data1)) then
      begin
        sprintf(buf, 'ang = %d, (x, y) = (%d, %d)', [
          plyr.mo.angle div $B60B60,
          plyr.mo.x div FRACUNIT,
          plyr.mo.y div FRACUNIT]);
        plyr._message := buf;
      end;
    end;

    // 'clev' change-level cheat
    if cht_CheckCheat(@cheat_clev, Chr(ev.data1)) then
    begin
      cht_GetParam(@cheat_clev, buf);

      if gamemode = commercial then
      begin
        epsd := 0;
        map := (Ord(buf[1]) - Ord('0')) * 10 + Ord(buf[2]) - Ord('0');
      end
      else
      begin
        epsd := Ord(buf[1]) - Ord('0');
        map := Ord(buf[2]) - Ord('0');
        // Catch invalid maps.
        if epsd < 1 then
          exit;
      end;

      if map < 1 then
        exit;

      // Ohmygod - this is not going to work.
      if (gamemode = retail) and
         ((epsd > 4) or (map > 9)) then
        exit;

      if (gamemode = registered) and
         ((epsd > 3) or (map > 9)) then
        exit;

      if (gamemode = shareware) and
         ((epsd > 1) or (map > 9)) then
        exit;

      if (gamemode = commercial) and
         ((epsd > 1) or (map > 34)) then
        exit;

      // So be it.
      plyr._message := STSTR_CLEV;
      G_DeferedInitNew(gameskill, epsd, map);
      result := true;
    end;
  end;
end;

var
  lastcalc: integer;
  oldhealth: integer;

function ST_calcPainOffset: integer;
var
  health: integer;
begin
  health := decide(plyr.health > 100, 100, plyr.health);

  if health <> oldhealth then
  begin
    lastcalc := ST_FACESTRIDE * (((100 - health) * ST_NUMPAINFACES) div 101);
    oldhealth := health;
  end;
  result := lastcalc;
end;

//
// This is a not-very-pretty routine which handles
//  the face states and their timing.
// the precedence of expressions is:
//  dead > evil grin > turned head > straight ahead
//
var
  lastattackdown: integer;
  priority: integer;

procedure ST_updateFaceWidget;
var
  i: integer;
  badguyangle: angle_t;
  diffang: angle_t;
  doevilgrin: boolean;
begin
  if priority < 10 then
  begin
    // dead
    if not boolval(plyr.health) then
    begin
      priority := 9;
      st_faceindex := ST_DEADFACE;
      st_facecount := 1;
    end;
  end;

  if priority < 9 then
  begin
    if boolval(plyr.bonuscount) then
    begin
      // picking up bonus
      doevilgrin := false;

      for i := 0 to Ord(NUMWEAPONS) - 1 do
      begin
        if oldweaponsowned[i] <> plyr.weaponowned[i] then
        begin
          doevilgrin := true;
          oldweaponsowned[i] := plyr.weaponowned[i];
        end;
      end;
	    if doevilgrin then
      begin
        // evil grin if just picked up weapon
        priority := 8;
        st_facecount := ST_EVILGRINCOUNT;
        st_faceindex := ST_calcPainOffset + ST_EVILGRINOFFSET;
      end;
    end;
  end;

  if priority < 8 then
  begin
    if boolval(plyr.damagecount) and
       boolval(plyr.attacker) and
       (plyr.attacker <> plyr.mo) then
    begin
      // being attacked
      priority := 7;

      if plyr.health - st_oldhealth > ST_MUCHPAIN then
      begin
        st_facecount := ST_TURNCOUNT;
        st_faceindex := ST_calcPainOffset + ST_OUCHOFFSET;
      end
      else
      begin
        badguyangle :=
          R_PointToAngle2(plyr.mo.x, plyr.mo.y, plyr.attacker.x, plyr.attacker.y);

        if badguyangle > plyr.mo.angle then
        begin
          // whether right or left
          diffang := badguyangle - plyr.mo.angle;
          i := intval(diffang > ANG180);
        end
        else
        begin
          // whether left or right
          diffang := plyr.mo.angle - badguyangle;
          i := intval(diffang <= ANG180);
        end; // confusing, aint it?


        st_facecount := ST_TURNCOUNT;
        st_faceindex := ST_calcPainOffset;

        if diffang < ANG45 then
        begin
          // head-on
          st_faceindex := st_faceindex + ST_RAMPAGEOFFSET;
        end
        else if boolval(i) then
        begin
          // turn face right
          st_faceindex := st_faceindex + ST_TURNOFFSET;
        end
        else
        begin
          // turn face left
          st_faceindex := st_faceindex + ST_TURNOFFSET + 1;
        end;
      end;
    end;
  end;

  if priority < 7 then
  begin
    // getting hurt because of your own damn stupidity
    if boolval(plyr.damagecount) then
    begin
      if plyr.health - st_oldhealth > ST_MUCHPAIN then
      begin
        priority := 7;
        st_facecount := ST_TURNCOUNT;
        st_faceindex := ST_calcPainOffset + ST_OUCHOFFSET;
      end
      else
      begin
        priority := 6;
        st_facecount := ST_TURNCOUNT;
        st_faceindex := ST_calcPainOffset + ST_RAMPAGEOFFSET;
      end;
    end;
  end;
    
  if priority < 6 then
  begin
    // rapid firing
    if plyr.attackdown then
    begin
      if lastattackdown = -1 then
        lastattackdown := ST_RAMPAGEDELAY
      else
      begin
        dec(lastattackdown);
        if not boolval(lastattackdown) then
        begin
          priority := 5;
          st_faceindex := ST_calcPainOffset + ST_RAMPAGEOFFSET;
          st_facecount := 1;
          lastattackdown := 1;
        end;
      end;
    end
    else
	    lastattackdown := -1;
  end;

  if priority < 5 then
  begin
    // invulnerability
    if boolval(plyr.cheats and CF_GODMODE) or
       boolval(plyr.powers[Ord(pw_invulnerability)]) then
    begin
      priority := 4;

      st_faceindex := ST_GODFACE;
      st_facecount := 1;
    end;
  end;

  // look left or look right if the facecount has timed out
  if not boolval(st_facecount) then
  begin
    st_faceindex := ST_calcPainOffset + (st_randomnumber mod 3);
    st_facecount := ST_STRAIGHTFACECOUNT;
    priority := 0;
  end;

  dec(st_facecount);
end;

var
  largeammo: integer; // means "n/a"

procedure ST_updateWidgets;
var
  i: integer;
begin
  // must redirect the pointer if the ready weapon has changed.
  //  if (w_ready.data != plyr->readyweapon)
  //  {
  if weaponinfo[Ord(plyr.readyweapon)].ammo = am_noammo then
    w_ready.num := @largeammo
  else
    w_ready.num := @plyr.ammo[Ord(weaponinfo[Ord(plyr.readyweapon)].ammo)];

    //{
    // static int tic=0;
    // static int dir=-1;
    // if (!(tic&15))
    //   plyr->ammo[weaponinfo[plyr->readyweapon].ammo]+=dir;
    // if (plyr->ammo[weaponinfo[plyr->readyweapon].ammo] == -100)
    //   dir = 1;
    // tic++;
    // }

  w_ready.data := Ord(plyr.readyweapon);

    // if (*w_ready.on)
    //  STlib_updateNum(&w_ready, true);
    // refresh weapon change
    //  }

  // update keycard multiple widgets
  for i := 0 to 2 do
  begin
    keyboxes[i] := decide(plyr.cards[i], i, -1);

    if plyr.cards[i + 3] then
      keyboxes[i] := i + 3;
  end;

  // refresh everything if this is him coming back to life
  ST_updateFaceWidget;

  // used by the w_armsbg widget
  st_notdeathmatch := not boolval(deathmatch);

  // used by w_arms[] widgets
  st_armson := st_statusbaron and st_notdeathmatch;

  // used by w_frags widget
  st_fragson := boolval(deathmatch) and st_statusbaron;
  st_fragscount := 0;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if i <> consoleplayer then
      st_fragscount := st_fragscount + plyr.frags[i]
    else
      st_fragscount := st_fragscount - plyr.frags[i];
  end;

  // get rid of chat window if up because of message
  dec(st_msgcounter);
  if not boolval(st_msgcounter) then
    st_chat := st_oldchat;
end;

procedure ST_Ticker;
begin
  inc(st_clock);
  st_randomnumber := M_Random;
  ST_updateWidgets;
  st_oldhealth := plyr.health;
end;

var
  st_palette: integer;

procedure ST_doPaletteStuff;
var
  palette: integer;
  pal: PByteArray;
  cnt: integer;
  bzc: integer;
begin
  cnt := plyr.damagecount;

  if boolval(plyr.powers[Ord(pw_strength)]) then
  begin
    // slowly fade the berzerk out
    bzc := 12 - _SHR(plyr.powers[Ord(pw_strength)], 6);

    if bzc > cnt then
      cnt := bzc;
  end;

  if boolval(cnt) then
  begin
    palette := _SHR(cnt + 7, 3);
	
    if palette >= NUMREDPALS then
      palette := NUMREDPALS - 1;

    palette := palette + STARTREDPALS;
  end
  else if boolval(plyr.bonuscount) then
  begin
    palette := _SHR(plyr.bonuscount + 7, 3);

    if palette >= NUMBONUSPALS then
      palette := NUMBONUSPALS - 1;

    palette := palette + STARTBONUSPALS;
  end
  else if (plyr.powers[Ord(pw_ironfeet)] > 4 * 32) or
          boolval(plyr.powers[Ord(pw_ironfeet)] and 8) then
    palette := RADIATIONPAL
  else
    palette := 0;

  if palette <> st_palette then
  begin
    st_palette := palette;
    pal := PByteArray(integer(W_CacheLumpNum(lu_palette, PU_CACHE)) + palette * 768);
    I_SetPalette(pal);
  end;
end;

procedure ST_DrawWidgets(refresh: boolean);
var
  i: integer;
begin
  // used by w_arms[] widgets
  st_armson := st_statusbaron and (not boolval(deathmatch));

  // used by w_frags widget
  st_fragson := boolval(deathmatch) and st_statusbaron;

  STlib_updateNum(@w_ready, refresh);

  for i := 0 to 3 do
  begin
    STlib_updateNum(@w_ammo[i], refresh);
    STlib_updateNum(@w_maxammo[i], refresh);
  end;

  STlib_updatePercent(@w_health, intval(refresh));
  STlib_updatePercent(@w_armor, intval(refresh));

  STlib_updateBinIcon(@w_armsbg, refresh);

  for i := 0 to 5 do
    STlib_updateMultIcon(@w_arms[i], refresh);

  STlib_updateMultIcon(@w_faces, refresh);

  for i := 0 to 2 do
    STlib_updateMultIcon(@w_keyboxes[i], refresh);

  STlib_updateNum(@w_frags, refresh);
end;

procedure ST_Refresh(refresh: boolean);
begin
{  if refresh then
  begin
    st_firsttime := false;
    // draw status bar background to off-screen buff}
    ST_RefreshBackground;
{//    ST_FinishRefresh;
  end;}
  // and refresh all widgets
  ST_DrawWidgets(refresh);
  ST_FinishRefresh;
end;

procedure ST_doRefresh;
begin
  st_firsttime := false;

  // draw status bar background to off-screen buff
  ST_RefreshBackground;

  // and refresh all widgets
  ST_DrawWidgets(true);

  ST_FinishRefresh;
end;

procedure ST_diffDraw;
begin
  // update all widgets
  ST_DrawWidgets(true);

  ST_FinishRefresh;
end;

procedure ST_Drawer(full: boolean; refresh: boolean);
begin
  st_statusbaron := (not full) or automapactive;
  st_firsttime := st_firsttime or refresh;

  // Do red-/gold-shifts from damage/items
  ST_doPaletteStuff;

  ST_Refresh(st_firsttime);
{  // If just after ST_Start(), refresh all
  if st_firsttime then
    ST_doRefresh
  // Otherwise, update as little as possible
  else
    ST_diffDraw;}
end;

procedure ST_loadGraphics;
var
  i: integer;
  j: integer;
  facenum: integer;
  namebuf: string;
begin
  // Load the numbers, tall and short
  for i := 0 to 9 do
  begin
    sprintf(namebuf, 'STTNUM%d', [i]);
    tallnum[i] := Ppatch_t(W_CacheLumpName(namebuf, PU_STATIC));

    sprintf(namebuf, 'STYSNUM%d', [i]);
    shortnum[i] := Ppatch_t(W_CacheLumpName(namebuf, PU_STATIC));
  end;

  // Load percent key.
  //Note: why not load STMINUS here, too?
  tallpercent := Ppatch_t(W_CacheLumpName('STTPRCNT', PU_STATIC));

  // key cards
  for i := 0 to Ord(NUMCARDS) - 1 do
  begin
    sprintf(namebuf, 'STKEYS%d', [i]);
    keys[i] := Ppatch_t(W_CacheLumpName(namebuf, PU_STATIC));
  end;

  // arms background
  armsbg := Ppatch_t(W_CacheLumpName('STARMS', PU_STATIC));

  // arms ownership widgets
  for i := 0 to 5 do
  begin
    sprintf(namebuf, 'STGNUM%d', [i + 2]);

    // gray #
    arms[i][0] := Ppatch_t(W_CacheLumpName(namebuf, PU_STATIC));

    // yellow #
    arms[i][1] := shortnum[i + 2];
  end;

  // face backgrounds for different color players
  sprintf(namebuf, 'STFB%d', [consoleplayer]);
  faceback := Ppatch_t(W_CacheLumpName(namebuf, PU_STATIC));

  // status bar background bits
  sbar := Ppatch_t(W_CacheLumpName('STBAR', PU_STATIC));

  // face states
  facenum := 0;
  for i := 0 to ST_NUMPAINFACES - 1 do
  begin
    for j := 0 to ST_NUMSTRAIGHTFACES - 1 do
    begin
      sprintf(namebuf, 'STFST%d%d', [i, j]);
      faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
      inc(facenum);
    end;
    sprintf(namebuf, 'STFTR%d0', [i]); // turn right
    faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
    inc(facenum);
    sprintf(namebuf, 'STFTL%d0', [i]); // turn left
    faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
    inc(facenum);
    sprintf(namebuf, 'STFOUCH%d', [i]); // ouch!
    faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
    inc(facenum);
    sprintf(namebuf, 'STFEVL%d', [i]); // evil grin ;)
    faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
    inc(facenum);
    sprintf(namebuf, 'STFKILL%d', [i]); // pissed off
    faces[facenum] := W_CacheLumpName(namebuf, PU_STATIC);
    inc(facenum);
  end;
  faces[facenum] := W_CacheLumpName('STFGOD0', PU_STATIC);
  inc(facenum);
  faces[facenum] := W_CacheLumpName('STFDEAD0', PU_STATIC);
end;

procedure ST_loadData;
begin
  lu_palette := W_GetNumForName('PLAYPAL');
  ST_loadGraphics;
end;

procedure ST_unloadGraphics;
var
  i: integer;
begin
  // unload the numbers, tall and short
  for i := 0 to 9 do
  begin
    Z_ChangeTag(tallnum[i], PU_CACHE);
    Z_ChangeTag(shortnum[i], PU_CACHE);
  end;
  // unload tall percent
  Z_ChangeTag(tallpercent, PU_CACHE);

  // unload arms background
  Z_ChangeTag(armsbg, PU_CACHE);

  // unload gray #'s
  for i := 0 to 5 do
    Z_ChangeTag(arms[i][0], PU_CACHE);

  // unload the key cards
  for i := 0 to Ord(NUMCARDS) - 1 do
    Z_ChangeTag(keys[i], PU_CACHE);

  Z_ChangeTag(sbar, PU_CACHE);
  Z_ChangeTag(faceback, PU_CACHE);

  for i := 0 to ST_NUMFACES - 1 do
    Z_ChangeTag(faces[i], PU_CACHE);

  // Note: nobody ain't seen no unloading    // VJ ???
  //   of stminus yet. Dude.
end;

procedure ST_unloadData;
begin
  ST_unloadGraphics;
end;

procedure ST_initData;
var
  i: integer;
begin
  st_firsttime := true;
  plyr := @players[consoleplayer];

  st_clock := 0;
  st_chatstate := StartChatState;
  st_gamestate := FirstPersonState;

  st_statusbaron := true;
  st_oldchat := false;
  st_chat := false;
  st_cursoron := false;

  st_faceindex := 0;
  st_palette := -1;

  st_oldhealth := -1;

  for i := 0 to Ord(NUMWEAPONS) - 1 do
    oldweaponsowned[i] := plyr.weaponowned[i];

  for i := 0 to 2 do
    keyboxes[i] := -1;

  STlib_init;
end;

procedure ST_createWidgets;
var
  i: integer;
begin
  // ready weapon ammo
  STlib_initNum(
    @w_ready,
    ST_AMMOX,
    ST_AMMOY,
    @tallnum,
    @plyr.ammo[Ord(weaponinfo[Ord(plyr.readyweapon)].ammo)],
    @st_statusbaron,
    ST_AMMOWIDTH);

  // the last weapon type
  w_ready.data := Ord(plyr.readyweapon);

  // health percentage
  STlib_initPercent(
    @w_health,
    ST_HEALTHX,
    ST_HEALTHY,
    @tallnum,
    @plyr.health,
    @st_statusbaron,
    tallpercent);

  // arms background
  STlib_initBinIcon(
    @w_armsbg,
    ST_ARMSBGX,
    ST_ARMSBGY,
    armsbg,
    @st_notdeathmatch,
    @st_statusbaron);

  // weapons owned
  for i := 0 to 5 do
  begin
    STlib_initMultIcon(
      @w_arms[i],
      ST_ARMSX + (i mod 3) * ST_ARMSXSPACE,
      ST_ARMSY + (i div 3) * ST_ARMSYSPACE,
      @arms[i],
      @plyr.weaponowned[i + 1],
      @st_armson);
  end;

  // frags sum
  STlib_initNum(
    @w_frags,
    ST_FRAGSX,
    ST_FRAGSY,
    @tallnum,
    @st_fragscount,
    @st_fragson,
    ST_FRAGSWIDTH);

  // faces
  STlib_initMultIcon(
    @w_faces,
    ST_FACESX,
    ST_FACESY,
    @faces,
    @st_faceindex,
    @st_statusbaron);

  // armor percentage - should be colored later
  STlib_initPercent(
    @w_armor,
    ST_ARMORX,
    ST_ARMORY,
    @tallnum,
    @plyr.armorpoints,
    @st_statusbaron,
    tallpercent);

  // keyboxes 0-2
  STlib_initMultIcon(
    @w_keyboxes[0],
    ST_KEY0X,
    ST_KEY0Y,
    @keys,
    @keyboxes[0],
    @st_statusbaron);

  STlib_initMultIcon(
    @w_keyboxes[1],
    ST_KEY1X,
    ST_KEY1Y,
    @keys,
    @keyboxes[1],
    @st_statusbaron);

  STlib_initMultIcon(
    @w_keyboxes[2],
    ST_KEY2X,
    ST_KEY2Y,
    @keys,
    @keyboxes[2],
    @st_statusbaron);

  // ammo count (all four kinds)
  STlib_initNum(
    @w_ammo[0],
    ST_AMMO0X,
    ST_AMMO0Y,
    @shortnum,
    @plyr.ammo[0],
    @st_statusbaron,
    ST_AMMO0WIDTH);

  STlib_initNum(
    @w_ammo[1],
    ST_AMMO1X,
    ST_AMMO1Y,
    @shortnum,
    @plyr.ammo[1],
    @st_statusbaron,
    ST_AMMO1WIDTH);

  STlib_initNum(
    @w_ammo[2],
    ST_AMMO2X,
    ST_AMMO2Y,
    @shortnum,
    @plyr.ammo[2],
    @st_statusbaron,
    ST_AMMO2WIDTH);

  STlib_initNum(
    @w_ammo[3],
    ST_AMMO3X,
    ST_AMMO3Y,
    @shortnum,
    @plyr.ammo[3],
    @st_statusbaron,
    ST_AMMO3WIDTH);

  // max ammo count (all four kinds)
  STlib_initNum(
    @w_maxammo[0],
    ST_MAXAMMO0X,
    ST_MAXAMMO0Y,
    @shortnum,
    @plyr.maxammo[0],
    @st_statusbaron,
    ST_MAXAMMO0WIDTH);

  STlib_initNum(
    @w_maxammo[1],
    ST_MAXAMMO1X,
    ST_MAXAMMO1Y,
    @shortnum,
    @plyr.maxammo[1],
    @st_statusbaron,
    ST_MAXAMMO1WIDTH);

  STlib_initNum(
    @w_maxammo[2],
    ST_MAXAMMO2X,
    ST_MAXAMMO2Y,
    @shortnum,
    @plyr.maxammo[2],
    @st_statusbaron,
    ST_MAXAMMO2WIDTH);

  STlib_initNum(
    @w_maxammo[3],
    ST_MAXAMMO3X,
    ST_MAXAMMO3Y,
    @shortnum,
    @plyr.maxammo[3],
    @st_statusbaron,
    ST_MAXAMMO3WIDTH);
end;

var
  st_stopped: boolean;

procedure ST_Stop;
begin
  if st_stopped then
    exit;

  I_SetPalette(W_CacheLumpNum (lu_palette, PU_CACHE));

  st_stopped := true;
end;

procedure ST_Start;
begin
  if not st_stopped then
    ST_Stop;

  ST_initData;
  ST_createWidgets;
  st_stopped := false;
end;

procedure ST_Init;
begin
  veryfirsttime := 0;
  ST_loadData;
  screens[_BG] := PByteArray(Z_Malloc(ST_WIDTH * ST_HEIGHT, PU_STATIC, nil));
end;



(* Unit doomdata implementation *)



(* Unit d_items implementation *)


             


(* Unit p_mobj_h implementation *)



(* Unit r_defs implementation *)



(* Unit r_segs implementation *)


                                                                                                                                

// OPTIMIZE: closed two sided lines as single sided

//
// R_RenderMaskedSegRange
//
procedure R_RenderMaskedSegRange(ds: Pdrawseg_t; x1, x2: integer);
var
  index: LongWord;
  col: Pcolumn_t;
  lightnum: integer;
  texnum: integer;
  i: integer;
begin
  // Calculate light table.
  // Use different light tables
  //   for horizontal / vertical / diagonal. Diagonal?
  // OPTIMIZE: get rid of LIGHTSEGSHIFT globally
  curline := ds.curline;
  frontsector := curline.frontsector;
  backsector := curline.backsector;
  texnum := texturetranslation[curline.sidedef.midtexture];

  lightnum := _SHR(frontsector.lightlevel, LIGHTSEGSHIFT) + extralight;

  if curline.v1.y = curline.v2.y then
    dec(lightnum)
  else if curline.v1.x = curline.v2.x then
    inc(lightnum);

  if lightnum < 0 then
    walllights := @scalelight[0, 0] // VJ ??? maybe @scalelight[0] ????
  else if lightnum >= LIGHTLEVELS then
    walllights := @scalelight[LIGHTLEVELS - 1, 0] // VJ ???
  else
    walllights := @scalelight[lightnum]; // VJ ???

  maskedtexturecol := ds.maskedtexturecol;

  rw_scalestep := ds.scalestep;
  spryscale := ds.scale1 + (x1 - ds.x1) * rw_scalestep;
  mfloorclip := ds.sprbottomclip;
  mceilingclip := ds.sprtopclip;

  // find positioning
  if boolval(curline.linedef.flags and ML_DONTPEGBOTTOM) then
  begin
    if frontsector.floorheight > backsector.floorheight then
      dc_texturemid := frontsector.floorheight
    else
      dc_texturemid := backsector.floorheight;
    dc_texturemid := dc_texturemid + textureheight[texnum] - viewz;
  end
  else
  begin
    if frontsector.ceilingheight < backsector.ceilingheight then
      dc_texturemid := frontsector.ceilingheight
     else
      dc_texturemid := backsector.ceilingheight;
    dc_texturemid := dc_texturemid - viewz;
  end;
  dc_texturemid := dc_texturemid + curline.sidedef.rowoffset;

  if boolval(fixedcolormap) then
    dc_colormap := fixedcolormap;

  // draw the columns
  for i := x1 to x2 do
  begin
    dc_x := i;
    // calculate lighting
    if maskedtexturecol[dc_x] <> MAXSHORT then
    begin
      if not boolval(fixedcolormap) then
      begin
        index := _SHR(spryscale, LIGHTSCALESHIFT);

        if index >=  MAXLIGHTSCALE then
          index := MAXLIGHTSCALE - 1;

        dc_colormap := walllights[index];
      end;

      sprtopscreen := centeryfrac - FixedMul(dc_texturemid, spryscale);
      dc_iscale := LongWord($ffffffff) div LongWord(spryscale);

      // draw the texture
      col := Pcolumn_t(integer(R_GetColumn(texnum, maskedtexturecol[dc_x])) - 3);

      R_DrawMaskedColumn(col);
      maskedtexturecol[dc_x] := MAXSHORT;
    end;
    spryscale := spryscale + rw_scalestep;
  end;
end;

//
// R_RenderSegLoop
// Draws zero, one, or two textures (and possibly a masked
//  texture) for walls.
// Can draw or mark the starting pixel of floor and ceiling
//  textures.
// CALLED: CORE LOOPING ROUTINE.
//
const
  HEIGHTBITS = 12;
  HEIGHTUNIT = 1 shl HEIGHTBITS;

procedure R_RenderSegLoop;
var
  angle: angle_t;
  index: LongWord;
  yl: integer;
  yh: integer;
  mid: integer;
  texturecolumn: fixed_t;
  top: integer;
  bottom: integer;
begin
  texturecolumn := 0; // shut up compiler warning
  while rw_x < rw_stopx do
  begin
    // mark floor / ceiling areas
    yl := _SHR(topfrac + HEIGHTUNIT - 1, HEIGHTBITS);

    // no space above wall?
    if yl < ceilingclip[rw_x] + 1 then
      yl := ceilingclip[rw_x] + 1;

    if markceiling then
    begin
      top := ceilingclip[rw_x] + 1;
      bottom := yl - 1;

      if bottom >= floorclip[rw_x] then
        bottom := floorclip[rw_x] - 1;

      if top <= bottom then
      begin
        ceilingplane.top[rw_x] := top;
        ceilingplane.bottom[rw_x] := bottom;
      end;
    end;


    yh := _SHR(bottomfrac, HEIGHTBITS);

    if yh >= floorclip[rw_x] then
      yh := floorclip[rw_x] - 1;

    if markfloor then
    begin
      top := yh + 1;
      bottom := floorclip[rw_x] - 1;
      if top <= ceilingclip[rw_x] then
        top := ceilingclip[rw_x] + 1;
      if top <= bottom then
      begin
        floorplane.top[rw_x] := top;
        floorplane.bottom[rw_x] := bottom;
      end;
    end;

    // texturecolumn and lighting are independent of wall tiers
    if segtextured then
    begin
      // calculate texture offset
      angle := _SHRW(rw_centerangle + xtoviewangle[rw_x], ANGLETOFINESHIFT);
      texturecolumn := rw_offset - FixedMul(finetangent[angle], rw_distance);
      texturecolumn := texturecolumn div FRACUNIT;
      // calculate lighting
      index := _SHR(rw_scale, LIGHTSCALESHIFT);

      if index >=  MAXLIGHTSCALE then
        index := MAXLIGHTSCALE - 1;

      dc_colormap := walllights[index];
      dc_x := rw_x;
      dc_iscale := LongWord($ffffffff) div LongWord(rw_scale);
    end;

    // draw the wall tiers
    if boolval(midtexture) then
    begin
      // single sided line
      dc_yl := yl;
      dc_yh := yh;
      dc_texturemid := rw_midtexturemid;
      dc_source := R_GetColumn(midtexture, texturecolumn);
      colfunc;
      ceilingclip[rw_x] := viewheight;
      floorclip[rw_x] := -1;
    end
    else
    begin
      // two sided line
      if boolval(toptexture) then
      begin
        // top wall
        mid := _SHR(pixhigh, HEIGHTBITS);
        pixhigh := pixhigh + pixhighstep;

        if mid >= floorclip[rw_x] then
          mid := floorclip[rw_x] - 1;

        if mid >= yl then
        begin
          dc_yl := yl;
          dc_yh := mid;
          dc_texturemid := rw_toptexturemid;
          dc_source := R_GetColumn(toptexture, texturecolumn);
          colfunc;
          ceilingclip[rw_x] := mid;
        end
        else
          ceilingclip[rw_x] := yl - 1;
      end
      else
      begin
        // no top wall
        if markceiling then
          ceilingclip[rw_x] := yl - 1;
      end;

	    if boolval(bottomtexture) then
      begin
        // bottom wall
        mid := _SHR(pixlow + HEIGHTUNIT - 1, HEIGHTBITS);
        pixlow := pixlow + pixlowstep;

        // no space above wall?
        if mid <= ceilingclip[rw_x] then
          mid := ceilingclip[rw_x] + 1;

        if mid <= yh then
        begin
          dc_yl := mid;
          dc_yh := yh;
          dc_texturemid := rw_bottomtexturemid;
          dc_source := R_GetColumn(bottomtexture, texturecolumn);
          colfunc;
          floorclip[rw_x] := mid;
        end
        else
          floorclip[rw_x] := yh + 1;
      end
      else
      begin
        // no bottom wall
        if markfloor then
          floorclip[rw_x] := yh + 1;
      end;

      if maskedtexture then
      begin
        // save texturecol
        // for backdrawing of masked mid texture
        maskedtexturecol[rw_x] := texturecolumn;
      end;
    end;

    rw_scale := rw_scale + rw_scalestep;
    topfrac := topfrac + topstep;
    bottomfrac := bottomfrac + bottomstep;
    inc(rw_x);
  end;
end;

//
// R_StoreWallRange
// A wall segment will be drawn
//  between start and stop pixels (inclusive).
//
procedure R_StoreWallRange(start: integer; stop: integer);
var
  hyp: fixed_t;
  sineval: fixed_t;
  distangle,
  offsetangle: angle_t;
  vtop: fixed_t;
  lightnum: integer;
begin
  // don't overflow and crash
  if ds_p = MAXDRAWSEGS then
    exit;

  // Now drawsergs is an array of pointer to drawseg_t
  // Dynamically allocation using zone
  if ds_p > max_ds_p then
  begin
    drawsegs[ds_p] := Z_Malloc(SizeOf(drawseg_t), PU_LEVEL, nil); // test!
    max_ds_p := ds_p;
  end;

  sidedef := curline.sidedef;
  linedef := curline.linedef;

  // mark the segment as visible for auto map
  linedef.flags := linedef.flags or ML_MAPPED;

  // calculate rw_distance for scale calculation
  rw_normalangle := curline.angle + ANG90;
  offsetangle := abs(rw_normalangle - rw_angle1);

  if offsetangle > ANG90 then
    offsetangle := ANG90;

  distangle := ANG90 - offsetangle;
  hyp := R_PointToDist(curline.v1.x, curline.v1.y);
  sineval := finesine[_SHRW(distangle, ANGLETOFINESHIFT)];
  rw_distance := FixedMul(hyp, sineval);


  rw_x := start;
  drawsegs[ds_p].x1 := rw_x;
  drawsegs[ds_p].x2 := stop;
  drawsegs[ds_p].curline := curline;
  rw_stopx := stop + 1;

  // calculate scale at both ends and step
  rw_scale := R_ScaleFromGlobalAngle(viewangle + xtoviewangle[start]);
  drawsegs[ds_p].scale1 := rw_scale;

  if stop > start then
  begin
    drawsegs[ds_p].scale2 := R_ScaleFromGlobalAngle(viewangle + xtoviewangle[stop]);
    rw_scalestep := (drawsegs[ds_p].scale2 - rw_scale) div (stop - start);
    drawsegs[ds_p].scalestep := rw_scalestep
  end
  else
  begin
    // UNUSED: try to fix the stretched line bug
(*
#if 0
	if (rw_distance < FRACUNIT/2)
	{
	    fixed_t		trx,try;
	    fixed_t		gxt,gyt;

	    trx = curline->v1->x - viewx;
	    try = curline->v1->y - viewy;

	    gxt = FixedMul(trx,viewcos);
	    gyt = -FixedMul(try,viewsin);
	    ds_p->scale1 = FixedDiv(projection, gxt-gyt)<<detailshift;
	}
#endif
*)
    drawsegs[ds_p].scale2 := drawsegs[ds_p].scale1;
  end;

  // calculate texture boundaries
  //  and decide if floor / ceiling marks are needed
  worldtop := frontsector.ceilingheight - viewz;
  worldbottom := frontsector.floorheight - viewz;

  midtexture := 0;
  toptexture := 0;
  bottomtexture := 0;
  maskedtexture := boolval(0);
  drawsegs[ds_p].maskedtexturecol := nil;

  if not boolval(backsector) then
  begin
    // single sided line
    midtexture := texturetranslation[sidedef.midtexture];
    // a single sided line is terminal, so it must mark ends
    markfloor := true;
    markceiling := true;
    if boolval(linedef.flags and ML_DONTPEGBOTTOM) then
    begin
	    vtop := frontsector.floorheight + textureheight[sidedef.midtexture];
	    // bottom of texture at bottom
	    rw_midtexturemid := vtop - viewz;
    end
    else
    begin
	    // top of texture at top
	    rw_midtexturemid := worldtop;
    end;
    rw_midtexturemid := rw_midtexturemid + sidedef.rowoffset;

    drawsegs[ds_p].silhouette := SIL_BOTH;
    drawsegs[ds_p].sprtopclip := @screenheightarray;
    drawsegs[ds_p].sprbottomclip := @negonearray;
    drawsegs[ds_p].bsilheight := MAXINT;
    drawsegs[ds_p].tsilheight := MININT;
  end
  else
  begin
    // two sided line
    drawsegs[ds_p].sprtopclip := nil;
    drawsegs[ds_p].sprbottomclip := nil;
    drawsegs[ds_p].silhouette := 0;

    if frontsector.floorheight > backsector.floorheight then
    begin
      drawsegs[ds_p].silhouette := SIL_BOTTOM;
      drawsegs[ds_p].bsilheight := frontsector.floorheight;
    end
    else if backsector.floorheight > viewz then
    begin
      drawsegs[ds_p].silhouette := SIL_BOTTOM;
	    drawsegs[ds_p].bsilheight := MAXINT;
	    // ds_p->sprbottomclip = negonearray;
    end;

    if frontsector.ceilingheight < backsector.ceilingheight then
    begin
      drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_TOP;
      drawsegs[ds_p].tsilheight := frontsector.ceilingheight;
    end
    else if backsector.ceilingheight < viewz then
    begin
      drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_TOP;
      drawsegs[ds_p].tsilheight := MININT;
      // ds_p->sprtopclip = screenheightarray;
    end;

    if backsector.ceilingheight <= frontsector.floorheight then
    begin
      drawsegs[ds_p].sprbottomclip := @negonearray;
      drawsegs[ds_p].bsilheight := MAXINT;
      drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_BOTTOM;
    end;

    if backsector.floorheight >= frontsector.ceilingheight then
    begin
      drawsegs[ds_p].sprtopclip := @screenheightarray;
	    drawsegs[ds_p].tsilheight := MININT;
	    drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_TOP;
    end;

    worldhigh := backsector.ceilingheight - viewz;
    worldlow := backsector.floorheight - viewz;

    // hack to allow height changes in outdoor areas
    if (frontsector.ceilingpic = skyflatnum) and
       (backsector.ceilingpic = skyflatnum) then
    begin
      worldtop := worldhigh;
    end;


    if (worldlow <> worldbottom) or
       (backsector.floorpic <> frontsector.floorpic) or
       (backsector.lightlevel <> frontsector.lightlevel) then
    begin
      markfloor := true;
    end
    else
    begin
      // same plane on both sides
      markfloor := false;
    end;


    if (worldhigh <> worldtop) or
       (backsector.ceilingpic <> frontsector.ceilingpic) or
       (backsector.lightlevel <> frontsector.lightlevel) then
    begin
      markceiling := true;
    end
    else
    begin
      // same plane on both sides
      markceiling := false;
    end;

    if (backsector.ceilingheight <= frontsector.floorheight) or
       (backsector.floorheight >= frontsector.ceilingheight) then
    begin
      // closed door
      markceiling := true;
      markfloor := true;
    end;


    if worldhigh < worldtop then
    begin
      // top texture
      toptexture := texturetranslation[sidedef.toptexture];
      if boolval(linedef.flags and ML_DONTPEGTOP) then
      begin
        // top of texture at top
        rw_toptexturemid := worldtop;
      end
      else
      begin
        vtop := backsector.ceilingheight + textureheight[sidedef.toptexture];

        // bottom of texture
        rw_toptexturemid := vtop - viewz;
      end
    end;

    if worldlow > worldbottom then
    begin
      // bottom texture
      bottomtexture := texturetranslation[sidedef.bottomtexture];

      if boolval(linedef.flags and ML_DONTPEGBOTTOM) then
      begin
        // bottom of texture at bottom
        // top of texture at top
        rw_bottomtexturemid := worldtop;
      end
      else // top of texture at top
        rw_bottomtexturemid := worldlow;
    end;
    rw_toptexturemid := rw_toptexturemid + sidedef.rowoffset;
    rw_bottomtexturemid := rw_bottomtexturemid + sidedef.rowoffset;

    // allocate space for masked texture tables
    if boolval(sidedef.midtexture) then
    begin
      // masked midtexture
      maskedtexture := true;
      maskedtexturecol := PSmallIntArray(@openings[lastopening - rw_x]);
      drawsegs[ds_p].maskedtexturecol := maskedtexturecol;
	    lastopening := lastopening + rw_stopx - rw_x;
    end;
  end;

  // calculate rw_offset (only needed for textured lines)
  segtextured := boolval(midtexture or toptexture or bottomtexture) or maskedtexture;

  if segtextured then
  begin
    offsetangle := rw_normalangle - rw_angle1;

    if offsetangle > ANG180 then
      offsetangle := -offsetangle;

    if offsetangle > ANG90 then
      offsetangle := ANG90;

    sineval := finesine[_SHRW(offsetangle, ANGLETOFINESHIFT)];
    rw_offset := FixedMul(hyp, sineval);

    if rw_normalangle - rw_angle1 < ANG180 then
      rw_offset := -rw_offset;

    rw_offset := rw_offset + sidedef.textureoffset + curline.offset;
    rw_centerangle := ANG90 + viewangle - rw_normalangle;

    // calculate light table
    //  use different light tables
    //  for horizontal / vertical / diagonal
    // OPTIMIZE: get rid of LIGHTSEGSHIFT globally
    if not boolval(fixedcolormap) then
    begin
      lightnum := _SHR(frontsector.lightlevel, LIGHTSEGSHIFT) + extralight;

      if curline.v1.y = curline.v2.y then
        dec(lightnum)
      else if curline.v1.x = curline.v2.x then
        inc(lightnum);

      if lightnum < 0 then
        walllights := @scalelight[0, 0] // VJ ???
      else if lightnum >= LIGHTLEVELS then
        walllights := @scalelight[LIGHTLEVELS - 1, 0]
      else
        walllights := @scalelight[lightnum, 0];
    end;
  end;

  // if a floor / ceiling plane is on the wrong side
  //  of the view plane, it is definitely invisible
  //  and doesn't need to be marked.


  if frontsector.floorheight >= viewz then
  begin
    // above view plane
    markfloor := false;
  end;

  if (frontsector.ceilingheight <= viewz) and
     (frontsector.ceilingpic <> skyflatnum) then
  begin
    // below view plane
    markceiling := false;
  end;


  // calculate incremental stepping values for texture edges
  worldtop := _SHR(worldtop, 4);
  worldbottom := _SHR(worldbottom, 4);

  topstep := -FixedMul(rw_scalestep, worldtop);
  topfrac := _SHR(centeryfrac, 4) - FixedMul(worldtop, rw_scale);

  bottomstep := -FixedMul(rw_scalestep, worldbottom);
  bottomfrac := _SHR(centeryfrac, 4) - FixedMul(worldbottom, rw_scale);

  if boolval(backsector) then
  begin
    worldhigh := _SHR(worldhigh, 4);
    worldlow := _SHR(worldlow, 4);

    if worldhigh < worldtop then
    begin
      pixhigh := _SHR(centeryfrac, 4) - FixedMul(worldhigh, rw_scale);
      pixhighstep := -FixedMul(rw_scalestep, worldhigh);
    end;

    if worldlow > worldbottom then
    begin
      pixlow := _SHR(centeryfrac, 4) - FixedMul(worldlow, rw_scale);
	    pixlowstep := -FixedMul(rw_scalestep, worldlow);
    end;
  end;

  // render it
  if markceiling then
    ceilingplane := R_CheckPlane(ceilingplane, rw_x, rw_stopx - 1);

  if markfloor then
    floorplane := R_CheckPlane(floorplane, rw_x, rw_stopx - 1);

  R_RenderSegLoop;


  // save sprite clipping info
  if (boolval(drawsegs[ds_p].silhouette and SIL_TOP) or maskedtexture) and
	   (not boolval(drawsegs[ds_p].sprtopclip)) then
  begin
    memcpy(@openings[lastopening], @ceilingclip[start], SizeOf(ceilingclip[0]) * (rw_stopx - start));
    drawsegs[ds_p].sprtopclip := PSmallIntArray(@openings[lastopening - start]);
    lastopening := lastopening + rw_stopx - start;
  end;

  if (boolval(drawsegs[ds_p].silhouette and SIL_BOTTOM) or maskedtexture) and
     (not boolval(drawsegs[ds_p].sprbottomclip)) then
  begin
    memcpy(@openings[lastopening], @floorclip[start], SizeOf(floorclip[0]) * (rw_stopx - start));
    drawsegs[ds_p].sprbottomclip := PSmallIntArray(@openings[lastopening - start]);
    lastopening := lastopening + rw_stopx - start;
  end;

  if maskedtexture and (not boolval(drawsegs[ds_p].silhouette and SIL_TOP)) then
  begin
    drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_TOP;
    drawsegs[ds_p].tsilheight := MININT;
  end;
  if maskedtexture and (not boolval(drawsegs[ds_p].silhouette and SIL_BOTTOM)) then
  begin
    drawsegs[ds_p].silhouette := drawsegs[ds_p].silhouette or SIL_BOTTOM;
    drawsegs[ds_p].bsilheight := MAXINT;
  end;
  inc(ds_p);
end;



(* Unit v_video implementation *)


                         

function _CENTERY: integer;
begin
  result := SCREENHEIGHT div 2;
end;

// x and y translation tables for stretcing
var
  preserveX: array[0..319] of integer;
  preserveY: array[0..199] of integer;

function V_NeedsPreserve: boolean;
begin
  result := (SCREENWIDTH <> 320) or (SCREENHEIGHT <> 200);
end;

function V_NeedsPreserve(preserve: boolean): boolean;
begin
  result := preserve and V_NeedsPreserve;
end;

// preserve x coordinates
function V_PreserveX(x: integer): integer;
begin
  if x <= 0 then
    result := 0
  else if x >= 320 then
    result := SCREENWIDTH
  else
  begin
    result := preserveX[x];
    if result - preserveX[x - 1] > 1 then
      result := preserveX[x - 1] + 1;
  end;
end;

// preserve y coordinates
function V_PreserveY(y: integer): integer;
begin
  if y <= 0 then
    result := 0
  else if y >= 200 then
    result := SCREENHEIGHT
  else
  begin
    result := preserveY[y];
    if result - preserveY[y - 1] > 1 then
      result := preserveY[y - 1] + 1;
  end;
end;

// preserve width coordinates
function V_PreserveW(x: integer; w: integer): integer;
begin
  result := V_PreserveX(x + w) - V_PreserveX(x);
//  result := preserveX[x + w] - V_PreserveX(x);
//  result := V_PreserveX(320 - x - w) - V_PreserveX(320 - x);
end;

// preserve height coordinates
function V_PreserveH(y: integer; h: integer): integer;
begin
  result := V_PreserveY(y + h) - V_PreserveY(y);
//  result := preserveY[y + h] - V_PreserveY(y);
end;


//
// V_MarkRect
//
procedure V_MarkRect(x, y: integer; width, height: integer; preserve: boolean);
begin
  if V_NeedsPreserve(preserve) then
  begin
    width := V_PreserveW(x, width);
    height := V_PreserveH(y, height);
    x := V_PreserveX(x);
    y := V_PreserveY(y);
  end;
  M_AddToBox(@dirtybox, x, y);
  M_AddToBox(@dirtybox, x + width - 1, y + height - 1);
end;

//
// V_CopyRect
//
procedure V_CopyRect(
  srcx: integer;
  srcy: integer;
  srcscrn: integer;
  width: integer;
  height: integer;
  destx: integer;
  desty: integer;
  destscrn: integer;
  preserve: boolean);
var
  src: PByteArray;
  dest: PByteArray;
  destw: integer;
  desth: integer;
  fracxstep: fixed_t;
  fracystep: fixed_t;
  fracx: fixed_t;
  fracy: fixed_t;
  col: integer;
  row: integer;
begin
  if V_NeedsPreserve(preserve) then
  begin
    destw := V_PreserveW(destx, width);

    desth := V_PreserveH(desty, height);

    destx := V_PreserveX(destx);

    desty := V_PreserveY(desty);

    V_MarkRect(destx, desty, destw, desth, false);

    fracy := srcy * FRACUNIT;
    fracxstep := FRACUNIT * width div destw;
    fracystep := FRACUNIT * height div desth;

    for row := desty to desty + desth - 1 do
    begin
      fracx := 0;
      dest := PByteArray(integer(screens[destscrn]) + SCREENWIDTH * row + destx);
      // Source is a 320 width screen
      src := PByteArray(integer(screens[srcscrn]) + 320 * (fracy div FRACUNIT) + srcx);
      for col := 0 to destw - 1 do
      begin
        dest[col] := src[fracx div FRACUNIT];
        fracx := fracx + fracxstep;
      end;
      fracy := fracy + fracystep;
    end;

  end
  else
  begin
    V_MarkRect(destx, desty, width, height, false);

    src := PByteArray(integer(screens[srcscrn]) + SCREENWIDTH * srcy + srcx);
    dest := PByteArray(integer(screens[destscrn]) + SCREENWIDTH * desty + destx);

    while height > 0 do
    begin
      memcpy(dest, src, width);
      src := PByteArray(integer(src) + SCREENWIDTH);
      dest := PByteArray(integer(dest) + SCREENWIDTH);
      dec(height);
    end;
  end;
end;

function V_GetScreenWidth(scrn: integer): integer;
begin
  // VJ
  // Hack to adjust screen coordinates!
  if  scrn in [0..3] then
    result := SCREENWIDTH
  else
    result := 320;
end;

procedure V_DrawPatch(x, y: integer; scrn: integer; patch: Ppatch_t; preserve: boolean);
var
  count: integer;
  col: integer;
  column: Pcolumn_t;
  desttop: PByte;
  dest: PByte;
  source: PByte;
  w: integer;
  pw: integer;
  ph: integer;
  fracx: fixed_t;
  fracy: fixed_t;
  fracxstep: fixed_t;
  fracystep: fixed_t;
  lasty: integer;
  swidth: integer;
begin
  if not V_NeedsPreserve(preserve) then
  begin
    y := y - patch.topoffset;
    x := x - patch.leftoffset;

    if scrn = _FG then
      V_MarkRect(x, y, patch.width, patch.height, false);

    col := 0;

    // VJ
    // Hack to adjust screen coordinates!
    swidth := V_GetScreenWidth(scrn);
    desttop := PByte(integer(screens[scrn]) + y * swidth + x);

    w := patch.width;

    while col < w do
    begin
      column := Pcolumn_t(integer(patch) + patch.columnofs[col]);

    // step through the posts in a column
      while column.topdelta <> $ff do
      begin
        source := PByte(integer(column) + 3);
        dest := PByte(integer(desttop) + column.topdelta * swidth);
  	    count := column.length;

        while count > 0 do
        begin
          dest^ := source^;
          inc(source);
          dest := PByte(integer(dest) + swidth);
          dec(count);
        end;
        column := Pcolumn_t(integer(column) + column.length + 4);
      end;
      inc(col);
      inc(desttop);
    end;
  end
////////////////////////////////////////////////////
// Streching Draw, preserving original dimentions
////////////////////////////////////////////////////
  else
  begin

    y := y - patch.topoffset;
    x := x - patch.leftoffset;

    pw := V_PreserveW(x, patch.width);
    ph := V_PreserveH(y, patch.height);

    x := V_PreserveX(x);
    y := V_PreserveY(y);

    if scrn = _FG then
      V_MarkRect(x, y, pw, ph, false);

    fracx := 0;
    fracxstep := FRACUNIT * patch.width div pw;
    fracystep := FRACUNIT * patch.height div ph;

    col := 0;
    // VJ
    // Hack to adjust screen coordinates!
    swidth := V_GetScreenWidth(scrn);
    desttop := PByte(integer(screens[scrn]) + y * swidth + x);

    while col < pw do
    begin
      column := Pcolumn_t(integer(patch) + patch.columnofs[fracx div FRACUNIT]);

    // step through the posts in a column
      while column.topdelta <> $ff do
      begin
        source := PByte(integer(column) + 3);
        dest := PByte(integer(desttop) + ((column.topdelta * SCREENHEIGHT) div 200) * SCREENWIDTH);
  	    count := column.length;
        fracy := 0;
        lasty := 0;

        while count > 0 do
        begin
          dest^ := source^;
          inc(dest, SCREENWIDTH);
          fracy := fracy + fracystep;
          if fracy div FRACUNIT > lasty then
          begin
            lasty := fracy div FRACUNIT;
            inc(source);
            dec(count);
          end;
        end;
        column := Pcolumn_t(integer(column) + column.length + 4);
      end;
      inc(col);
      inc(desttop);

      fracx := fracx + fracxstep;
    end;
  end;
end;


//
// V_DrawPatchFlipped
// Masks a column based masked pic to the screen.
// Flips horizontally, e.g. to mirror face.
//
procedure V_DrawPatchFlipped(x, y: integer; scrn: integer; patch: Ppatch_t; preserve: boolean);
var
  count: integer;
  col: integer;
  column: Pcolumn_t;
  desttop: PByteArray;
  dest: PByte;
  source: PByte;
  w: integer;
  pw: integer;
  ph: integer;
  fracx: fixed_t;
  fracy: fixed_t;
  fracxstep: fixed_t;
  fracystep: fixed_t;
  lasty: integer;
  swidth: integer;
begin
  if not V_NeedsPreserve(preserve) then
  begin
    y := y - patch.topoffset;
    x := x - patch.leftoffset;

    if scrn = _FG then
      V_MarkRect(x, y, patch.width, patch.height, false);

    col := 0;

    // VJ
    // Hack to adjust screen coordinates!
    swidth := V_GetScreenWidth(scrn);
    desttop := PByteArray(integer(screens[scrn]) + y * swidth + x);

    w := patch.width;

    while col < w do
    begin
      column := Pcolumn_t(integer(patch) + patch.columnofs[w - 1 - col]);

    // step through the posts in a column
      while column.topdelta <> $ff do
      begin
        source := PByte(integer(column) + 3);
        dest := PByte(integer(desttop) + column.topdelta * swidth);
  	    count := column.length;

        while count > 0 do
        begin
          dest^ := source^;
          inc(source);
          inc(dest, swidth);
          dec(count);
        end;
        column := Pcolumn_t(integer(column) + column.length + 4);
      end;
      inc(col);
      inc(desttop);
//      incp(pointer(desttop));
    end;
  end
////////////////////////////////////////////////////
// Streching Draw, preserving original dimentions
////////////////////////////////////////////////////
  else
  begin
    y := y - patch.topoffset;
    x := x - patch.leftoffset;

    pw := V_PreserveW(x, patch.width);
    ph := V_PreserveH(y, patch.height);

    x := V_PreserveX(x);
    y := V_PreserveY(y);

    if scrn = _FG then
      V_MarkRect(x, y, pw, ph, false);

    fracx := 0;
    fracxstep := FRACUNIT * patch.width div pw;
    fracystep := FRACUNIT * patch.height div ph;

    col := 0;
    desttop := PByteArray(integer(screens[scrn]) + y * SCREENWIDTH + x);

    w := patch.width;

    while col < pw do
    begin
      column := Pcolumn_t(integer(patch) + patch.columnofs[w - 1 - (fracx div FRACUNIT)]);

    // step through the posts in a column
      while column.topdelta <> $ff do
      begin
        source := PByte(integer(column) + 3);
        dest := PByte(integer(desttop) + ((column.topdelta * SCREENHEIGHT) div 200) * SCREENWIDTH);
  	    count := column.length;
        fracy := 0;
        lasty := 0;

        while count > 0 do
        begin
          dest^ := source^;
          dest := PByte(integer(dest) + SCREENWIDTH);
          fracy := fracy + fracystep;
          if fracy div FRACUNIT > lasty then
          begin
            lasty := fracy div FRACUNIT;
            inc(source);
            dec(count);
          end;
        end;
        column := Pcolumn_t(integer(column) + column.length + 4);
      end;
      inc(col);
      inc(desttop);
      fracx := fracx + fracxstep;
    end;
  end;
end;

//
// V_DrawBlock
// Draw a linear block of pixels into the view buffer.
//
procedure V_DrawBlock(x, y: integer; scrn: integer; width, height: integer; src: PByteArray);
var
  dest: PByteArray;
begin
  V_MarkRect(x, y, width, height, false);

  dest := PByteArray(integer(screens[scrn]) + y * SCREENWIDTH + x);

  while height <> 0 do
  begin
    memcpy(dest, src, width);
    src := PByteArray(integer(src) + width);
    dest := PByteArray(integer(dest) + SCREENWIDTH);
    dec(height);
  end;
end;

//
// V_GetBlock
// Gets a linear block of pixels from the view buffer.
//
procedure V_GetBlock(x, y: integer; scrn: integer; width, height: integer; dest: PByteArray);
var
  src: PByteArray;
begin
  src := PByteArray(integer(screens[scrn]) + y * SCREENWIDTH + x);

  while height <> 0 do
  begin
    memcpy(dest, src, width);
    src := PByteArray(integer(src) + SCREENWIDTH);
    dest := PByteArray(integer(dest) + width);
    dec(height);
  end;
end;

//
// V_Init
//
procedure V_Init;
var
  i: integer;
  base: PByteArray;
  st: integer;
begin
  // stick these in low dos memory on PCs
  base := PByteArray(malloc(SCREENWIDTH * SCREENHEIGHT * 4 + 2 * 320 * 200));

  st := 0;
  for i := 0 to 5 do
  begin
    screens[i] := @base[st];
    if i < 4 then
      st := st + SCREENWIDTH * SCREENHEIGHT
    else
      st := st + 320 * 200;
  end;

  // initialize translation tables
  for i := 0 to 319 do
    preserveX[i] := round(i * SCREENWIDTH / 320);

  for i := 0 to 199 do
    preserveY[i] := round(i * SCREENHEIGHT / 200);

end;


(* Unit hu_lib implementation *)


                                                 

procedure HUlib_init;
begin
end;

procedure HUlib_clearTextLine(t: Phu_textline_t);
begin
  t.len := 0;
  t.l[0] := Chr(0);
  t.needsupdate := 1; //true;
end;

procedure HUlib_initTextLine(t: Phu_textline_t; x: integer; y: integer; f: Ppatch_tPArray; sc: integer);
begin
  t.x := x;
  t.y := y;
  t.f := f;
  t.sc := sc;
  HUlib_clearTextLine(t);
end;

function HUlib_addCharToTextLine(t: Phu_textline_t; ch: char): boolean;
begin
  if t.len = HU_MAXLINELENGTH then
    result := false
  else
  begin
    t.l[t.len] := ch;
    inc(t.len);
    t.l[t.len] := Chr(0);
    t.needsupdate := 4;
    result := true;
  end;
end;

function HUlib_delCharFromTextLine(t: Phu_textline_t): boolean;
begin
  if t.len = 0 then
    result := false
  else
  begin
    dec(t.len);
    t.l[t.len] := Chr(0);
    t.needsupdate := 4;
    result := true;
  end;
end;

procedure HUlib_drawTextLine(l: Phu_textline_t; drawcursor: boolean);
var
  i: integer;
  w: integer;
  x: integer;
  c: char;
begin
  // draw the new stuff
  x := l.x;
  for i := 0 to l.len - 1 do
  begin
    c := toupper(l.l[i]);
    if (c <> ' ') and (Ord(c) >= l.sc) and (c <= '_') then
    begin
      w := l.f[Ord(c) - l.sc].width;
      if x + w > SCREENWIDTH then
        break;
      V_DrawPatch(x, l.y, FG, l.f[Ord(c) - l.sc], false);
      x := x + w;
    end
    else
    begin
      x := x + 4;
      if x >= SCREENWIDTH then
        break;
    end;
  end;

  // draw the cursor if requested
  if drawcursor and (x + l.f[Ord('_') - l.sc].width <= SCREENWIDTH) then
    V_DrawPatch(x, l.y, FG, l.f[Ord('_') - l.sc], false);
end;

// sorta called by HU_Erase and just better darn get things straight
var
  lastautomapactive: boolean;

procedure HUlib_eraseTextLine(l: Phu_textline_t);
var
  lh: integer;
  y: integer;
  yoffset: integer;
begin
  // Only erases when NOT in automap and the screen is reduced,
  // and the text must either need updating or refreshing
  // (because of a recent change back from the automap)
  if (not automapactive) and (viewwindowx <> 0) and (l.needsupdate <> 0) then
  begin
    lh := l.f[0].height + 1;
    y := l.y;
    yoffset := y * SCREENWIDTH;
    while y < l.y + lh do
    begin
      if (y < viewwindowy) or (y >= viewwindowy + viewheight) then
        R_VideoErase(yoffset, SCREENWIDTH) // erase entire line
      else
      begin
        // erase left border
        R_VideoErase(yoffset, viewwindowx);
        // erase right border
        R_VideoErase(yoffset + viewwindowx + viewwidth, viewwindowx);
      end;
      inc(y);
      yoffset := yoffset + SCREENWIDTH;
    end;
  end;

  lastautomapactive := automapactive;
  if l.needsupdate > 0 then
    l.needsupdate := l.needsupdate - 1;
end;

procedure HUlib_initSText(s: Phu_stext_t; x: integer; y: integer; h: integer;
  font: Ppatch_tPArray; startchar: integer; _on: Pboolean);
var
  i: integer;
begin
  s.h := h;
  s._on := _on;
  s.laston := true;
  s.cl := 0;
  for i := 0 to h - 1 do
    HUlib_initTextLine(@s.l[i], x, y - i * (font[0].height + 1),
      font, startchar);
end;

procedure HUlib_addLineToSText(s: Phu_stext_t);
var
  i: integer;
begin
  // add a clear line
  inc(s.cl);
  if s.cl = s.h then
    s.cl := 0;
  HUlib_clearTextLine(@s.l[s.cl]);

  // everything needs updating
  for i := 0 to s.h - 1 do
    s.l[i].needsupdate := 4;
end;

procedure HUlib_addMessageToSText(s: Phu_stext_t; prefix: string; msg: string);
var
  i: integer;
begin
  HUlib_addLineToSText(s);

  for i := 1 to Length(prefix) do
    HUlib_addCharToTextLine(@s.l[s.cl], prefix[i]);

  for i := 1 to Length(msg) do
    HUlib_addCharToTextLine(@s.l[s.cl], msg[i]);
end;

procedure HUlib_drawSText(s: Phu_stext_t);
var
  i, idx: integer;
  l: Phu_textline_t;
begin
  if not s._on^ then
    exit; // if not on, don't draw

  // draw everything
  for i := 0 to s.h - 1 do
  begin
    idx := s.cl - i;
    if idx < 0 then
      idx := idx + s.h; // handle queue of lines

    l := @s.l[idx];

    // need a decision made here on whether to skip the draw
    HUlib_drawTextLine(l, false); // no cursor, please
  end;
end;

procedure HUlib_eraseSText(s: Phu_stext_t);
var
  i: integer;
begin
  for i := 0 to s.h - 1 do
  begin
    if (s.laston) and (not s._on^) then
      s.l[i].needsupdate := 4;
    HUlib_eraseTextLine(@s.l[i]);
  end;
  s.laston := s._on^;
end;

procedure HUlib_initIText(it: Phu_itext_t; x: integer; y: integer; font: Ppatch_tPArray;
  startchar: integer; _on: Pboolean);
begin
  it.lm := 0; // default left margin is start of text
  it._on := _on;
  it.laston := true;
  HUlib_initTextLine(@it.l, x, y, font, startchar);
end;

// The following deletion routines adhere to the left margin restriction
procedure HUlib_delCharFromIText(it: Phu_itext_t);
begin
  if it.l.len <> it.lm then
    HUlib_delCharFromTextLine(@it.l);
end;

procedure HUlib_eraseLineFromIText(it: Phu_itext_t);
begin
  while it.lm <> it.l.len do
    HUlib_delCharFromTextLine(@it.l);
end;

// Resets left margin as well
procedure HUlib_resetIText(it: Phu_itext_t);
begin
  it.lm := 0;
  HUlib_clearTextLine(@it.l);
end;

procedure HUlib_addPrefixToIText(it: Phu_itext_t; str: string);
var
  i: integer;
begin
  for i := 1 to Length(str) do
    HUlib_addCharToTextLine(@it.l, str[i]);
  it.lm := it.l.len;
end;

// wrapper function for handling general keyed input.
// returns true if it ate the key
function HUlib_keyInIText(it: Phu_itext_t; ch: byte): boolean;
begin
  if (ch >= Ord(' ')) and (ch <= Ord('_')) then
    HUlib_addCharToTextLine(@it.l, Chr(ch))
  else
  if ch = KEY_BACKSPACE then
    HUlib_delCharFromIText(it)
  else if ch <> KEY_ENTER then
  begin
    result := false; // did not eat key
    exit;
  end;

  result := true; // ate the key
end;

procedure HUlib_drawIText(it: Phu_itext_t);
var
  l: Phu_textline_t;
begin
  l := @it.l;

  if it._on^ then
    HUlib_drawTextLine(l, true); // draw the line w/ cursor
end;

procedure HUlib_eraseIText(it: Phu_itext_t);
begin
  if it.laston and (not it._on^) then
    it.l.needsupdate := 4;
  HUlib_eraseTextLine(@it.l);
  it.laston := it._on^;
end;



(* Unit r_bsp implementation *)


                                                                                                                                                                       

//
// R_ClearDrawSegs
//
procedure R_ClearDrawSegs;
begin
  ds_p := 0;
end;

//
// ClipWallSegment
// Clips the given range of columns
// and includes it in the new clip list.
//
type
  cliprange_t = record
    first: integer;
    last: integer;
  end;
  Pcliprange_t = ^cliprange_t;

const
  MAXSEGS = 32; // VJ maybe bigger ???

var
// newend is one past the last valid seg
  newend: integer; // VJ was Pcliprange_t
  solidsegs: array[0..MAXSEGS - 1] of cliprange_t;

//
// R_ClipSolidWallSegment
// Does handle solid walls,
//  e.g. single sided LineDefs (middle texture)
//  that entirely block the view.
//
procedure R_ClipSolidWallSegment(first, last: integer);
var
  next: integer;
  start: integer;

  procedure crunch;
  begin
    if next = start then
    // Post just extended past the bottom of one post.
      exit;
    while next <> newend do
    begin
      // Remove a post.
      inc(start);
      inc(next);       
      solidsegs[start] := solidsegs[next];
    end;
    newend := start + 1;
  end;

begin
  // Find the first range that touches the range
  //  (adjacent pixels are touching).
  start := 0;
  while solidsegs[start].last < first - 1 do
    inc(start);

  if first < solidsegs[start].first then
  begin
    if last < solidsegs[start].first - 1 then
    begin
      // Post is entirely visible (above start),
      //  so insert a new clippost.
      R_StoreWallRange(first, last);
      next := newend;
      inc(newend);

      while next <> start do
      begin
        solidsegs[next] := solidsegs[next - 1];
        dec(next);
      end;
	    solidsegs[next].first := first;
	    solidsegs[next].last := last;
      exit;
    end;

    // There is a fragment above *start.
    R_StoreWallRange(first, solidsegs[start].first - 1);
    // Now adjust the clip size.
    solidsegs[start].first := first;
  end;

  // Bottom contained in start?
  if last <= solidsegs[start].last then
    exit;

  next := start;
  while last >= solidsegs[next + 1].first - 1 do
  begin
    // There is a fragment between two posts.
    R_StoreWallRange(solidsegs[next].last + 1, solidsegs[next + 1].first - 1);
    inc(next);

    if last <= solidsegs[next].last then
    begin
	    // Bottom is contained in next.
	    // Adjust the clip size.
	    solidsegs[start].last := solidsegs[next].last;
      crunch;
      exit;
    end;
  end;

  // There is a fragment after *next.
  R_StoreWallRange(solidsegs[next].last + 1, last);
  // Adjust the clip size.
  solidsegs[start].last := last;

  // Remove start+1 to next from the clip list,
  // because start now covers their area.
  crunch;
end;

//
// R_ClipPassWallSegment
// Clips the given range of columns,
//  but does not includes it in the clip list.
// Does handle windows,
//  e.g. LineDefs with upper and lower texture.
//
procedure R_ClipPassWallSegment(first, last: integer);
var
  start: integer;
begin
  // Find the first range that touches the range
  //  (adjacent pixels are touching).
  start := 0;
  while solidsegs[start].last < first - 1 do
    inc(start);

  if first < solidsegs[start].first then
  begin
    if last < solidsegs[start].first - 1 then
    begin
	    // Post is entirely visible (above start).
	    R_StoreWallRange (first, last);
	    exit;
    end;
    // There is a fragment above *start.
    R_StoreWallRange(first, solidsegs[start].first - 1);
  end;

  // Bottom contained in start?
  if last <= solidsegs[start].last then
    exit;

  while last >= solidsegs[start + 1].first - 1 do
  begin
    // There is a fragment between two posts.
    R_StoreWallRange(solidsegs[start].last + 1, solidsegs[start + 1].first - 1);
    inc(start);

    if last <= solidsegs[start].last then
      exit;
  end;

  // There is a fragment after *next.
  R_StoreWallRange(solidsegs[start].last + 1, last);
end;

//
// R_ClearClipSegs
//
procedure R_ClearClipSegs;
begin
  solidsegs[0].first := -$7fffffff;
  solidsegs[0].last := -1;
  solidsegs[1].first := viewwidth;
  solidsegs[1].last := $7fffffff;
  newend := 2;
end;

//
// R_AddLine
// Clips the given segment
// and adds any visible pieces to the line list.
//
procedure R_AddLine(line: Pseg_t);
var
  x1: integer;
  x2: integer;
  angle1: angle_t;
  angle2: angle_t;
  span: angle_t;
  tspan: angle_t;
begin
  curline := line;

  // OPTIMIZE: quickly reject orthogonal back sides.
  angle1 := R_PointToAngle(line.v1.x, line.v1.y);
  angle2 := R_PointToAngle(line.v2.x, line.v2.y);

  // Clip to view edges.
  // OPTIMIZE: make constant out of 2*clipangle (FIELDOFVIEW).
  span := angle1 - angle2;

  // Back side? I.e. backface culling?
  if span >= ANG180 then
    exit;

  // Global angle needed by segcalc.
  rw_angle1 := angle1;
  angle1 := angle1 - viewangle;
  angle2 := angle2 - viewangle;

  tspan := angle1 + clipangle;
  if tspan > 2 * clipangle then
  begin
    tspan := tspan - 2 * clipangle;

    // Totally off the left edge?
    if tspan >= span then
      exit;

    angle1 := clipangle;
  end;

  tspan := clipangle - angle2;
  if tspan > 2 * clipangle then
  begin
    tspan := tspan - 2 * clipangle;

    // Totally off the left edge?
    if tspan >= span then
      exit;

    angle2 := -clipangle;
  end;

  // The seg is in the view range,
  // but not necessarily visible.
  angle1 := _SHRW(angle1 + ANG90, ANGLETOFINESHIFT);
  angle2 := _SHRW(angle2 + ANG90, ANGLETOFINESHIFT);
  x1 := viewangletox[angle1];
  x2 := viewangletox[angle2];

  // Does not cross a pixel?
  if x1 = x2 then
    exit;

  backsector := line.backsector;

  // Single sided line?
  if not boolval(backsector) then
  begin
    R_ClipSolidWallSegment(x1, x2 - 1);
    exit;
  end;

  // Closed door.
  if (backsector.ceilingheight <= frontsector.floorheight) or
     (backsector.floorheight >= frontsector.ceilingheight) then
  begin
    R_ClipSolidWallSegment(x1, x2 - 1);
    exit;
  end;

  // Window.
  if (backsector.ceilingheight <> frontsector.ceilingheight) or
     (backsector.floorheight <> frontsector.floorheight) then
  begin
    R_ClipPassWallSegment(x1, x2 - 1);
    exit;
  end;

  // Reject empty lines used for triggers
  //  and special events.
  // Identical floor and ceiling on both sides,
  // identical light levels on both sides,
  // and no middle texture.
  if (backsector.ceilingpic = frontsector.ceilingpic) and
     (backsector.floorpic = frontsector.floorpic) and
     (backsector.lightlevel = frontsector.lightlevel) and
     (curline.sidedef.midtexture = 0) then
    exit;

  R_ClipPassWallSegment(x1, x2 - 1);
end;

//
// R_CheckBBox
// Checks BSP node/subtree bounding box.
// Returns true
//  if some part of the bbox might be visible.
//
const
  checkcoord: array[0..11, 0..3] of integer = (
    (3, 0, 2, 1),
    (3, 0, 2, 0),
    (3, 1, 2, 0),
    (0, 0, 0, 0),
    (2, 0, 2, 1),
    (0, 0, 0, 0),
    (3, 1, 3, 0),
    (0, 0, 0, 0),
    (2, 0, 3, 1),
    (2, 1, 3, 1),
    (2, 1, 3, 0),
    (0, 0, 0, 0)
  );

function R_CheckBBox(bspcoordA: Pfixed_tArray; const side: integer): boolean;
var
  bspcoord: array[0..3] of fixed_t; // VJ
  boxx: integer;
  boxy: integer;
  boxpos: integer;
  x1: fixed_t;
  y1: fixed_t;
  x2: fixed_t;
  y2: fixed_t;
  angle1: angle_t;
  angle2: angle_t;
  span: angle_t;
  tspan: angle_t;
  start: integer;
  sx1: integer;
  sx2: integer;
begin
  if side = 0 then
  begin
    bspcoord[0] := bspcoordA[0];
    bspcoord[1] := bspcoordA[1];
    bspcoord[2] := bspcoordA[2];
    bspcoord[3] := bspcoordA[3];
  end
  else
  begin
    bspcoord[0] := bspcoordA[4];
    bspcoord[1] := bspcoordA[5];
    bspcoord[2] := bspcoordA[6];
    bspcoord[3] := bspcoordA[7];
  end;

  // Find the corners of the box
  // that define the edges from current viewpoint.
  if viewx <= bspcoord[BOXLEFT] then
    boxx := 0
  else if viewx < bspcoord[BOXRIGHT] then
    boxx := 1
  else
    boxx := 2;

  if viewy >= bspcoord[BOXTOP] then
    boxy := 0
  else if viewy > bspcoord[BOXBOTTOM] then
    boxy := 1
  else
    boxy := 2;

  boxpos := _SHL(boxy, 2) + boxx;
  if boxpos = 5 then
  begin
    result := true;
    exit;
  end;

  x1 := bspcoord[checkcoord[boxpos][0]];
  y1 := bspcoord[checkcoord[boxpos][1]];
  x2 := bspcoord[checkcoord[boxpos][2]];
  y2 := bspcoord[checkcoord[boxpos][3]];

  // check clip list for an open space
  angle1 := R_PointToAngle(x1, y1) - viewangle;
  angle2 := R_PointToAngle(x2, y2) - viewangle;

  span := angle1 - angle2;

  // Sitting on a line?
  if span >= ANG180 then
  begin
    result := true;
    exit;
  end;

  tspan := angle1 + clipangle;

  if tspan > 2 * clipangle then
  begin
    tspan := tspan - 2 * clipangle;

    // Totally off the left edge?
    if tspan >= span then
    begin
      result := false;
      exit;
    end;

    angle1 := clipangle;
  end;

  tspan := clipangle - angle2;
  if tspan > 2 * clipangle then
  begin
    tspan := tspan - 2 * clipangle;

    // Totally off the left edge?
    if tspan >= span then
    begin
      result := false;
      exit;
    end;

    angle2 := -clipangle;
  end;


  // Find the first clippost
  //  that touches the source post
  //  (adjacent pixels are touching).
  angle1 := _SHRW(angle1 + ANG90, ANGLETOFINESHIFT);
  angle2 := _SHRW(angle2 + ANG90, ANGLETOFINESHIFT);
  sx1 := viewangletox[angle1];
  sx2 := viewangletox[angle2];

  // Does not cross a pixel.
  if sx1 = sx2 then
  begin
    result := false;
    exit;
  end;

  dec(sx2);

  start := 0;
  while solidsegs[start].last < sx2 do
    inc(start);

  if (sx1 >= solidsegs[start].first) and
     (sx2 <= solidsegs[start].last) then
    // The clippost contains the new span.
    result := false
  else
    result := true;
end;

//
// R_Subsector
// Determine floor/ceiling planes.
// Add sprites of things in sector.
// Draw one or more line segments.
//
procedure R_Subsector(num: integer);
var
  count: integer;
  line: Pseg_t;
  i_line: integer;
  sub: Psubsector_t;
begin
  inc(sscount);
  sub := @subsectors[num];
  frontsector := sub.sector;
  count := sub.numlines;
  i_line := sub.firstline;
  line := @segs[i_line];

  if frontsector.floorheight < viewz then
  begin
    floorplane := R_FindPlane(frontsector.floorheight,
                              frontsector.floorpic,
                              frontsector.lightlevel);
  end
  else
    floorplane := nil;

  if (frontsector.ceilingheight > viewz) or
     (frontsector.ceilingpic = skyflatnum) then
  begin
    ceilingplane := R_FindPlane(frontsector.ceilingheight,
                                frontsector.ceilingpic,
                                frontsector.lightlevel);
  end
  else
    ceilingplane := nil;

  R_AddSprites(frontsector);

  while boolval(count) do
  begin
    R_AddLine(line);
    inc(i_line);
    line := @segs[i_line];
    dec(count);
  end;
end;

//
// RenderBSPNode
// Renders all subsectors below a given node,
//  traversing subtree recursively.
// Just call with BSP root.
procedure R_RenderBSPNode(bspnum: integer);
var
  bsp: Pnode_t;
  side: integer;
begin
  // Found a subsector?
  if boolval(bspnum and NF_SUBSECTOR) then
  begin
    if bspnum = -1 then
      R_Subsector(0)
    else
      R_Subsector(bspnum and (not NF_SUBSECTOR));
    exit;
  end;

  bsp := @nodes[bspnum];

  // Decide which side the view point is on.

  side := R_PointOnSide(viewx, viewy, bsp);

  // Recursively divide front space.
  R_RenderBSPNode(bsp.children[side]);

  // Possibly divide back space.
  if R_CheckBBox(Pfixed_tArray(@(bsp^.bbox)), side xor 1) then
    R_RenderBSPNode(bsp.children[side xor 1]);
end;



(* Unit r_data implementation *)


                                                                                                                                                        

//
// Graphics.
// DOOM graphics for walls and sprites
// is stored in vertical runs of opaque pixels (posts).
// A column is composed of zero or more posts,
// a patch or sprite is composed of zero or more columns.
//



//
// Texture definition.
// Each texture is composed of one or more patches,
// with patches being lumps stored in the WAD.
// The lumps are referenced by number, and patched
// into the rectangular texture space using origin
// and possibly other attributes.
//
type
  mappatch_t = record
    originx: smallint;
    originy: smallint;
    patch: smallint;
    stepdir: smallint;
    colormap: smallint;
  end;
  Pmappatch_t = ^mappatch_t;

//
// Texture definition.
// A DOOM wall texture is a list of patches
// which are to be combined in a predefined order.
//
  maptexture_t = packed record
    name: char8_t;
    masked: integer;
    width: smallint;
    height: smallint;
    columndirectory: PPointer; // OBSOLETE
    patchcount: smallint;
    patches: array[0..0] of mappatch_t;
  end;
  Pmaptexture_t = ^maptexture_t;

// A single patch from a texture definition,
//  basically a rectangular area within
//  the texture rectangle.
  texpatch_t = packed record
    // Block origin (allways UL),
    // which has allready accounted
    // for the internal origin of the patch.
    originx: integer;
    originy: integer;
    patch: integer;
  end;
  Ptexpatch_t = ^texpatch_t;

// A maptexturedef_t describes a rectangular texture,
//  which is composed of one or more mappatch_t structures
//  that arrange graphic patches.
  texture_t = packed record
    // Keep name for switch changing, etc.
    name: char8_t;
    width: smallint;
    height: smallint;

    // All the patches[patchcount]
    //  are drawn back to front into the cached texture.
    patchcount: smallint;
    patches: array[0..0] of texpatch_t;
  end;
  Ptexture_t = ^texture_t;
  texture_tPArray = array[0..$FFFF] of Ptexture_t;
  Ptexture_tPArray = ^texture_tPArray;


var
  numspritelumps: integer;

  numtextures: integer;
  textures: Ptexture_tPArray;

  texturewidthmask: PIntegerArray;

  texturecolumnlump: PSmallIntPArray;
  texturecolumnofs: PWordPArray;
  texturecomposite: PBytePArray;

//
// MAPTEXTURE_T CACHING
// When a texture is first needed,
//  it counts the number of composite columns
//  required in the texture and allocates space
//  for a column directory and any new columns.
// The directory will simply point inside other patches
//  if there is only one patch in a given column,
//  but any columns with multiple patches
//  will have new column_ts generated.
//



//
// R_DrawColumnInCache
// Clip and draw a column
//  from a patch into a cached post.
//
procedure R_DrawColumnInCache(patch: Pcolumn_t; cache: PByteArray;
  originy: integer; cacheheight: integer);
var
  count: integer;
  position: integer;
  source: PByteArray;
begin
  while patch.topdelta <> $ff do
  begin
    source := PByteArray(integer(patch) + 3);

    count := patch.length;
    position := originy + patch.topdelta;

    if position < 0 then
    begin
      count := count + position;
	    position := 0;
    end;

    if position + count > cacheheight then
      count := cacheheight - position;

    if count > 0 then
      memcpy (@cache[position], source, count);

    patch := Pcolumn_t(integer(patch) + patch.length + 4);
  end;
end;

//
// R_GenerateComposite
// Using the texture definition,
//  the composite texture is created from the patches,
//  and each column is cached.
//
procedure R_GenerateComposite(texnum: integer);
var
  block: PByteArray;
  texture: Ptexture_t;
  patch: Ptexpatch_t;
  realpatch: Ppatch_t;
  x: integer;
  x1: integer;
  x2: integer;
  i: integer;
  patchcol: Pcolumn_t;
  collump: PSmallIntArray;
  colofs: PWordArray;
begin
  texture := textures[texnum];

  block := Z_Malloc(texturecompositesize[texnum], PU_STATIC, @texturecomposite[texnum]);

  collump := texturecolumnlump[texnum];
  colofs := texturecolumnofs[texnum];

  // Composite the columns together.
  for i := 0 to texture.patchcount - 1 do
  begin
    patch := @texture.patches[i];

    realpatch := W_CacheLumpNum(patch.patch, PU_CACHE);
    x1 := patch.originx;
    x2 := x1 + realpatch.width;

    if x1 < 0 then
      x := 0
    else
      x := x1;

    if x2 > texture.width then
      x2 := texture.width;

    while x < x2 do
    begin
      // Column does not have multiple patches?
      if collump[x] < 0 then
      begin
        patchcol := Pcolumn_t(integer(realpatch) + realpatch.columnofs[x - x1]);
        R_DrawColumnInCache(
          patchcol, @block[colofs[x]], patch.originy, texture.height);
      end;
      inc(x);
    end;
  end;

  // Now that the texture has been built in column cache,
  //  it is purgable from zone memory.
  Z_ChangeTag(block, PU_CACHE);
end;

//
// R_GenerateLookup
//
procedure R_GenerateLookup(texnum: integer);
var
  texture: Ptexture_t;
  patchcount: PByteArray; // patchcount[texture->width]
  patch: Ptexpatch_t;
  realpatch: Ppatch_t;
  x: integer;
  x1: integer;
  x2: integer;
  i: integer;
  collump: PSmallIntArray;
  colofs: PWordArray;
begin
  texture := textures[texnum];

  // Composited texture not created yet.
  texturecomposite[texnum] := nil;

  texturecompositesize[texnum] := 0;
  collump := texturecolumnlump[texnum];
  colofs := texturecolumnofs[texnum];

  // Now count the number of columns
  //  that are covered by more than one patch.
  // Fill in the lump / offset, so columns
  //  with only a single patch are all done.
  patchcount := malloc(texture.width);
//  patchcount := Z_Malloc(texture.width, PU_LEVEL, nil);
//  GetMem(patchcount, texture.width);
//  memset(patchcount, 0, texture.width);
  patch := @texture.patches[0];

  for i := 0 to texture.patchcount - 1 do
  begin
    realpatch := W_CacheLumpNum(patch.patch, PU_CACHE);
    x1 := patch.originx;
    x2 := x1 + realpatch.width;

    if x1 < 0 then
	    x := 0
    else
      x := x1;

    if x2 > texture.width then
      x2 := texture.width;
    while x < x2 do
    begin
      patchcount[x] := patchcount[x] + 1;
      collump[x] := patch.patch;
      colofs[x] := realpatch.columnofs[x - x1] + 3;
      inc(x);
    end;
    inc(patch);
//    incp(pointer(patch), SizeOf(patch^));
  end;

  for x := 0 to texture.width - 1 do
  begin
    if not boolval(patchcount[x]) then
    begin
      printf('R_GenerateLookup(): column without a patch (%s)' + #13#10, [texture.name]);
      exit;
    end; // I_Error ("R_GenerateLookup: column without a patch");


    if patchcount[x] > 1 then
    begin
      // Use the cached block.
      collump[x] := -1;
      colofs[x] := texturecompositesize[texnum];

      if texturecompositesize[texnum] > $10000 - texture.height then
        I_Error('R_GenerateLookup(): texture %d is >64k', [texnum]);

      texturecompositesize[texnum] := texturecompositesize[texnum] + texture.height;
    end;
  end;
  freemem(patchcount);
end;

//
// R_GetColumn
//
function R_GetColumn(tex: integer; col: integer): PByteArray;
var
  lump: integer;
  ofs: integer;
begin
  col := col and texturewidthmask[tex];
  lump := texturecolumnlump[tex][col];
  ofs := texturecolumnofs[tex][col];

  if lump > 0 then
  begin
    result := PByteArray(integer(W_CacheLumpNum(lump, PU_CACHE)) + ofs); // VJ ???
    exit;
  end;

  if not boolval(texturecomposite[tex]) then
    R_GenerateComposite(tex);

  result := PByteArray(integer(texturecomposite[tex]) + ofs); // VJ ???
end;

//
// R_InitTextures
// Initializes the texture list
//  with the textures from the world map.
//
procedure R_InitTextures;
var
  mtexture: Pmaptexture_t;
  texture: Ptexture_t;
  mpatch: Pmappatch_t;
  patch: Ptexpatch_t;
  i: integer;
  j: integer;
  maptex: PIntegerArray;
  maptex2: PIntegerArray;
  maptex1: PIntegerArray;
  name: char8_t;
  names: PByteArray;
  name_p: PByteArray;
  patchlookup: PIntegerArray;
  nummappatches: integer;
  offset: integer;
  maxoff: integer;
  maxoff2: integer;
  numtextures1: integer;
  numtextures2: integer;
  directory: PIntegerArray;
  temp1: integer;
  temp2: integer;
  temp3: integer;
begin
  // Load the patch names from pnames.lmp.
  ZeroMemory(name, SizeOf(name));
  names := W_CacheLumpName('PNAMES', PU_STATIC);
  nummappatches := PInteger(names)^;
  name_p := PByteArray(integer(names) + 4);

  GetMem(patchlookup, nummappatches * SizeOf(integer));

  for i := 0 to nummappatches - 1 do
  begin
    for j := 0 to 7 do
      name[j] := Chr(name_p[i * 8 + j]);
    patchlookup[i] := W_CheckNumForName(char8tostring(name));
  end;
  Z_Free(names);

  // Load the map texture definitions from textures.lmp.
  // The data is contained in one or two lumps,
  //  TEXTURE1 for shareware, plus TEXTURE2 for commercial.
  maptex1 := W_CacheLumpName('TEXTURE1', PU_STATIC);
  maptex := maptex1;
  numtextures1 := maptex[0];
  maxoff := W_LumpLength(W_GetNumForName('TEXTURE1'));
  directory := PintegerArray(integer(maptex) + SizeOf(integer));

  if W_CheckNumForName('TEXTURE2') <> -1 then
  begin
    maptex2 := W_CacheLumpName('TEXTURE2', PU_STATIC);
    numtextures2 := maptex2[0];
    maxoff2 := W_LumpLength(W_GetNumForName('TEXTURE2'));
  end
  else
  begin
    maptex2 := nil;
    numtextures2 := 0;
    maxoff2 := 0;
  end;
  numtextures := numtextures1 + numtextures2;

  textures := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  texturecolumnlump := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  texturecolumnofs := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  texturecomposite := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  texturecompositesize := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  texturewidthmask := Z_Malloc(numtextures * 4, PU_STATIC, nil);
  textureheight := Z_Malloc(numtextures * 4, PU_STATIC, nil);

  //	Really complex printing shit...
  temp1 := W_GetNumForName('S_START');  // P_???????
  temp2 := W_GetNumForName('S_END') - 1;
  temp3 := ((temp2 - temp1 + 63) div 64) + ((numtextures + 63) div 64);
  printf('[');
  for i := 0 to temp3 - 1 do
    printf(' ');
  printf('         ]');
  for i := 0 to temp3 - 1 do
    printf(#9);             // VJ original Chr(8) not Chr(9)
  printf(#9#9#9#9#9#9#9#9#9#9);

  for i := 0 to numtextures - 1 do
  begin
    if not boolval(i and 63) then
      printf('.');

    if i = numtextures1 then
    begin
      // Start looking in second texture file.
      maptex := maptex2;
      maxoff := maxoff2;
      directory := PIntegerArray(integer(maptex) + SizeOf(integer));
    end;

    offset := directory[0];

    if offset > maxoff then
      I_Error('R_InitTextures(): bad texture directory');

//    padcpy(mtexture, pointer(integer(maptex) + offset), SizeOf(mtexture));
    mtexture := Pmaptexture_t(integer(maptex) + offset);

    // Deal with padding problem
{    mt_ofs := 0;

    memcpy(@mtexture, pointer(integer(maptex) + offset), SizeOf(mtexture));
{    mtexture.name := Pchar8_t(Pmaptexture_t(integer(maptex) + offset + mt_ofs))^;
    mt_ofs := mt_ofs + SizeOf(char8_t);

    masked: boolean;
    width: smallint;
    height: smallint;
    columndirectory: PPointer; // OBSOLETE
    patchcount: smallint;
    patches: array[0..0] of mappatch_t;}

    textures[i] :=
      Z_Malloc(
        SizeOf(texture_t) + SizeOf(texpatch_t) * (mtexture.patchcount - 1),
          PU_STATIC, nil);
    texture := textures[i];

    texture.width := mtexture.width;
    texture.height := mtexture.height;
    texture.patchcount := mtexture.patchcount;

    memcpy(@texture.name, @mtexture.name, SizeOf(texture.name));
    mpatch := @mtexture.patches[0];
    patch := @texture.patches[0];

    for j := 0 to texture.patchcount - 1 do
    begin
      patch.originx := mpatch.originx;
      patch.originy := mpatch.originy;
      patch.patch := patchlookup[mpatch.patch];
      if patch.patch = -1 then
        I_Error('R_InitTextures(): Missing patch in texture %s', [texture.name]);
      inc(mpatch);
      inc(patch);
    end;
    texturecolumnlump[i] := Z_Malloc(texture.width * 2, PU_STATIC, nil);
    texturecolumnofs[i] := Z_Malloc(texture.width * 2, PU_STATIC, nil);

    j := 1;
    while j * 2 <= texture.width do
      j := _SHL(j, 1);

    texturewidthmask[i] := j - 1;
    textureheight[i] := texture.height * FRACUNIT;

    incp(pointer(directory), SizeOf(integer));
  end;

  Z_Free(maptex1);
  if boolval(maptex2) then
    Z_Free(maptex2);

  // Precalculate whatever possible.
  for i := 0 to numtextures - 1 do
    R_GenerateLookup(i);

  // Create translation table for global animation.
  texturetranslation := Z_Malloc((numtextures + 1) * 4, PU_STATIC, nil);

  for i := 0 to numtextures - 1 do
    texturetranslation[i] := i;
end;

//
// R_InitFlats
//
procedure R_InitFlats;
var
  i: integer;
begin
  firstflat := W_GetNumForName('F_START') + 1;
  lastflat := W_GetNumForName('F_END') - 1;
  numflats := lastflat - firstflat + 1;

  // Create translation table for global animation.
  flattranslation := Z_Malloc((numflats + 1) * 4, PU_STATIC, nil);

  for i := 0 to numflats - 1 do
    flattranslation[i] := i;
end;

//
// R_InitSpriteLumps
// Finds the width and hoffset of all sprites in the wad,
//  so the sprite does not need to be cached completely
//  just for having the header info ready during rendering.
//
procedure R_InitSpriteLumps;
var
  i: integer;
  patch: Ppatch_t;
begin
  firstspritelump := W_GetNumForName('S_START') + 1;
  lastspritelump := W_GetNumForName('S_END') - 1;

  numspritelumps := lastspritelump - firstspritelump + 1;
  spritewidth := Z_Malloc(numspritelumps * 4, PU_STATIC, nil);
  spriteoffset := Z_Malloc(numspritelumps * 4, PU_STATIC, nil);
  spritetopoffset := Z_Malloc(numspritelumps * 4, PU_STATIC, nil);

  for i := 0 to numspritelumps - 1 do
  begin
    if not boolval(i and 63) then
      printf('.');

    patch := W_CacheLumpNum(firstspritelump + i, PU_CACHE);
    spritewidth[i] := patch.width * FRACUNIT;
    spriteoffset[i] := patch.leftoffset * FRACUNIT;
    spritetopoffset[i] := patch.topoffset * FRACUNIT;
  end;
end;

//
// R_InitColormaps
//
procedure R_InitColormaps;
var
  lump: integer;
  length: integer;
begin
  // Load in the light tables,
  //  256 byte align tables.
  lump := W_GetNumForName('COLORMAP');
  length := W_LumpLength(lump) + 255;
  colormaps := Z_Malloc(length, PU_STATIC, nil);
  colormaps := Plighttable_tArray((integer(colormaps) + 255)and (not $ff));
  W_ReadLump(lump, colormaps);
end;

//
// R_InitData
// Locates all the lumps
//  that will be used by all views
// Must be called after W_Init.
//
procedure R_InitData;
begin
  R_InitTextures;
  printf(#13#10 + 'InitTextures');
  R_InitFlats;
  printf(#13#10 + 'InitFlats');
  R_InitSpriteLumps;
  printf(#13#10 + 'InitSprites');
  R_InitColormaps;
  printf(#13#10 + 'InitColormaps');
end;

//
// R_FlatNumForName
// Retrieval, get a flat number for a flat name.
//
function R_FlatNumForName(const name: string): integer;
var
  i: integer;
begin
  i := W_CheckNumForName(name);

  if i = -1 then
    I_Error('R_FlatNumForName(): %s not found', [name]);

  result := i - firstflat;
end;

//
// R_CheckTextureNumForName
// Check whether texture is available.
// Filter out NoTexture indicator.
//
function R_CheckTextureNumForName(const name: string): integer;
var
  i: integer;
  check: string;
begin
  // "NoTexture" marker.
  if name[1] = '-' then
  begin
    result := 0;
    exit;
  end;

  check := strupper(name);
  for i := 0 to numtextures - 1 do
    if strupper(char8tostring(textures[i].name)) = check then
    begin
      result := i;
      exit;
    end;

  result := -1;
end;

//
// R_TextureNumForName
// Calls R_CheckTextureNumForName,
//  aborts with error message.
//
function R_TextureNumForName(const name: string): integer;
begin
  result := R_CheckTextureNumForName(name);

  if result = -1 then
    I_Error('R_TextureNumForName(): %s not found', [name]);
end;

//
// R_PrecacheLevel
// Preloads all relevant graphics for the level.
//
var
  flatmemory: integer;
  texturememory: integer;
  spritememory: integer;

procedure R_PrecacheLevel;
var
  flatpresent: PByteArray;
  texturepresent: PByteArray;
  spritepresent: PByteArray;
  i: integer;
  j: integer;
  k: integer;
  lump: integer;
  texture: Ptexture_t;
  th: Pthinker_t;
  sf: Pspriteframe_t;
begin
  if demoplayback then
    exit;

  // Precache flats.
  flatpresent := malloc(numflats);
//  GetMem(flatpresent, numflats);
//  memset(flatpresent, 0, numflats);

  for i := 0 to numsectors - 1 do
  begin
    flatpresent[sectors[i].floorpic] := 1;
    flatpresent[sectors[i].ceilingpic] := 1;
  end;

  flatmemory := 0;

  for i := 0 to numflats - 1 do
  begin
    if boolval(flatpresent[i]) then
    begin
      lump := firstflat + i;
      flatmemory := flatmemory + lumpinfo[lump].size;
      W_CacheLumpNum(lump, PU_CACHE);
    end;
  end;

  // Precache textures.
  texturepresent := malloc(numtextures);
//  GetMem(texturepresent, numtextures);
//  memset(texturepresent, 0, numtextures);

  for i := 0 to numsides - 1 do
  begin
    texturepresent[sides[i].toptexture] := 1;
    texturepresent[sides[i].midtexture] := 1;
    texturepresent[sides[i].bottomtexture] := 1;
  end;

  // Sky texture is always present.
  // Note that F_SKY1 is the name used to
  //  indicate a sky floor/ceiling as a flat,
  //  while the sky texture is stored like
  //  a wall texture, with an episode dependend
  //  name.
  texturepresent[skytexture] := 1;

  texturememory := 0;
  for i := 0 to numtextures - 1 do
  begin
    if not boolval(texturepresent[i]) then
      continue;

    texture := textures[i];

    for j := 0 to texture.patchcount - 1 do
    begin
      lump := texture.patches[j].patch;
      texturememory := texturememory + lumpinfo[lump].size;
      W_CacheLumpNum(lump, PU_CACHE);
    end;
  end;

  // Precache sprites.
  spritepresent := malloc(r_things_numsprites);
//  GetMem(spritepresent, numsprites);
//  memset(spritepresent, 0, numsprites);

  th := thinkercap.next;
  while th <> @thinkercap do
  begin
    if @th._function.acp1 = @P_MobjThinker then
      spritepresent[Ord(Pmobj_t(th).sprite)] := 1;
    th := th.next;
  end;

  spritememory := 0;
  for i := 0 to r_things_numsprites - 1 do
  begin
    if not boolval(spritepresent[i]) then
      continue;

    for j := 0 to sprites[i].numframes - 1 do
    begin
      sf := @sprites[i].spriteframes[j];
	    for k := 0 to 7 do
      begin
        lump := firstspritelump + sf.lump[k];
        spritememory := spritememory + lumpinfo[lump].size;
        W_CacheLumpNum(lump, PU_CACHE);
      end;
    end;
  end;
  freemem(flatpresent);
  freemem(texturepresent);
  freemem(spritepresent);

end;

procedure R_SetupLevel;
begin
  maxvisplane := -1;
  max_ds_p := -1;
end;



(* Unit r_plane implementation *)


                                                                                                                       

// Here comes the obnoxious "visplane".
const
  MAXVISPLANES = 1024; // VJ was = 128
// VJ - Note about visplanes:
//   Top and Bottom arrays (of visplane_t struct) are now
//   allocated dynamically (using zone memory)
//   Use -zone cmdline param to specify more zone memory allocation
//   if out of memory.
//   See also R_NewVisPlane() 

var
  visplanes: array[0..MAXVISPLANES - 1] of visplane_t;
  lastvisplane: integer;

//
// spanstart holds the start of a plane span
// initialized to 0 at start
//
  spanstart: array[0..MAXHEIGHT - 1] of integer;

//
// texture mapping
//
  planezlight: Plighttable_tPArray;
  planeheight: fixed_t;

  basexscale: fixed_t;
  baseyscale: fixed_t;

  cachedheight: array[0..MAXHEIGHT - 1] of fixed_t;
  cacheddistance: array[0..MAXHEIGHT -1] of fixed_t;
  cachedxstep: array[0..MAXHEIGHT - 1] of fixed_t;
  cachedystep: array[0..MAXHEIGHT - 1] of fixed_t;


//
// R_InitPlanes
// Only at game startup.
//
procedure R_InitPlanes;
begin
  // Doh!
end;

//
// R_MapPlane
//
// Uses global vars:
//  planeheight
//  ds_source
//  basexscale
//  baseyscale
//  viewx
//  viewy
//
// BASIC PRIMITIVE
//
procedure R_MapPlane(y: integer; x1: integer; x2: integer);
var
  angle: angle_t;
  distance: fixed_t;
  length: fixed_t;
  index: LongWord;
begin
  if planeheight <> cachedheight[y] then
  begin
    cachedheight[y] := planeheight;
    cacheddistance[y] := FixedMul(planeheight, yslope[y]);
    distance := cacheddistance[y];
    cachedxstep[y] := FixedMul(distance, basexscale);
    ds_xstep := cachedxstep[y];
    cachedystep[y] := FixedMul(distance, baseyscale);
    ds_ystep := cachedystep[y];
  end
  else
  begin
    distance := cacheddistance[y];
    ds_xstep := cachedxstep[y];
    ds_ystep := cachedystep[y];
  end;

  length := FixedMul(distance, distscale[x1]);
  angle := _SHRW(viewangle + xtoviewangle[x1], ANGLETOFINESHIFT);
  ds_xfrac := viewx + FixedMul(finecosine[angle], length);
  ds_yfrac := -viewy - FixedMul(finesine[angle], length);

  if boolval(fixedcolormap) then
    ds_colormap := fixedcolormap
  else
  begin
    index := _SHR(distance, LIGHTZSHIFT);

    if index >= MAXLIGHTZ then
      index := MAXLIGHTZ - 1;

    ds_colormap := planezlight[index];
  end;

  ds_y := y;
  ds_x1 := x1;
  ds_x2 := x2;

  // high or low detail
  spanfunc;
end;

//
// R_ClearPlanes
// At begining of frame.
//
procedure R_ClearPlanes;
var
  i: integer;
  angle: angle_t;
begin
  // opening / clipping determination
  for i := 0 to viewwidth - 1 do
  begin
    floorclip[i] := viewheight;
    ceilingclip[i] := -1;
  end;

  lastvisplane := 0;
  lastopening := 0;

  // texture calculation
  memset(@cachedheight, 0, SizeOf(cachedheight));

  // left to right mapping
  angle := _SHRW(viewangle - ANG90, ANGLETOFINESHIFT);

  // scale will be unit scale at SCREENWIDTH/2 distance
  basexscale := FixedDiv(finecosine[angle], centerxfrac);
  baseyscale := -FixedDiv(finesine[angle], centerxfrac);
end;

//
// R_NewVisPlane
//
function R_NewVisPlane: integer;
begin
// VJ
//   Create a new visplane
//   Uses zone memory to allocate top and bottom arrays
  if lastvisplane = MAXVISPLANES then
    I_Error('R_NewVisPlane(): no more visplanes');

  if lastvisplane > maxvisplane then
  begin
    visplanes[lastvisplane].top := Pvisindex_tArray(
      Z_Malloc((SCREENWIDTH + 1) * SizeOf(visindex_t), PU_LEVEL, nil));
    visplanes[lastvisplane].bottom := Pvisindex_tArray(
      Z_Malloc((SCREENWIDTH + 1) * SizeOf(visindex_t), PU_LEVEL, nil));
    maxvisplane := lastvisplane;
  end;

  inc(lastvisplane);
  result := lastvisplane;
end;

//
// R_FindPlane
//
function R_FindPlane(height: fixed_t; picnum: integer; lightlevel: integer): Pvisplane_t;
var
  check: integer;
  i: integer;
begin
  if picnum = skyflatnum then
  begin
    height := 0; // all skys map together
    lightlevel := 0;
  end;

  check := 0;
  while check < lastvisplane do
  begin
    if (height = visplanes[check].height) and
       (picnum = visplanes[check].picnum) and
       (lightlevel = visplanes[check].lightlevel) then
      break;
    inc(check);
  end;

  if check < lastvisplane then
  begin
    result := @visplanes[check];
    exit;
  end;

  if lastvisplane = MAXVISPLANES then
    I_Error('R_FindPlane(): no more visplanes');

  lastvisplane := R_NewVisPlane;

  visplanes[check].height := height;
  visplanes[check].picnum := picnum;
  visplanes[check].lightlevel := lightlevel;
  visplanes[check].minx := SCREENWIDTH;
  visplanes[check].maxx := -1;

  for i := 0 to SCREENWIDTH - 1 do
    visplanes[check].top[i] := VISEND;

  result := @visplanes[check];
end;

//
// R_CheckPlane
//
function R_CheckPlane(pl: Pvisplane_t; start: integer; stop: integer): Pvisplane_t;
var
  intrl: integer;
  intrh: integer;
  unionl: integer;
  unionh: integer;
  x: integer;
  i: integer;
begin
  if start < pl.minx then
  begin
    intrl := pl.minx;
    unionl := start;
  end
  else
  begin
    unionl := pl.minx;
    intrl := start;
  end;

  if stop > pl.maxx then
  begin
    intrh := pl.maxx;
    unionh := stop;
  end
  else
  begin
    unionh := pl.maxx;
    intrh := stop;
  end;

  x := intrl;
  while x <= intrh do
  begin
    if pl.top[x] <> VISEND then
      break
    else
      inc(x);
  end;

  if x > intrh then
  begin
    pl.minx := unionl;
    pl.maxx := unionh;

    // use the same one
    result := pl;
    exit;
  end;

  // make a new visplane

  if lastvisplane = MAXVISPLANES then
    I_Error('R_CheckPlane(): no more visplanes');

  visplanes[lastvisplane].height := pl.height;
  visplanes[lastvisplane].picnum := pl.picnum;
  visplanes[lastvisplane].lightlevel := pl.lightlevel;

  pl := @visplanes[lastvisplane];

  lastvisplane := R_NewVisPlane;

  pl.minx := start;
  pl.maxx := stop;

  for i := 0 to SCREENWIDTH - 1 do
    pl.top[i] := VISEND;

  result := pl;
end;

//
// R_MakeSpans
//
procedure R_MakeSpans(x: integer; t1: integer; b1: integer; t2: integer; b2: integer);
begin
  while (t1 < t2) and (t1 <= b1) do
  begin
  // VJ 9/7/05
    if (t1 >= 0) and (t1 <= Length(spanstart)) then
      R_MapPlane(t1, spanstart[t1], x - 1);
    inc(t1);
  end;
  while (b1 > b2) and (b1 >= t1) do
  begin
  // VJ 9/7/05
    if (b1 >= 0) and (b1 <= Length(spanstart)) then
      R_MapPlane(b1, spanstart[b1], x - 1);
    dec(b1);
  end;

  while (t2 < t1) and (t2 <= b2) do
  begin
  // VJ 9/7/05
    if (t2 >= 0) and (t2 <= Length(spanstart)) then
      spanstart[t2] := x;
    inc(t2);
  end;
  while (b2 > b1) and (b2 >= t2) do
  begin
  // VJ 9/7/05
    if (b2 >= 0) and (b2 <= Length(spanstart)) then
      spanstart[b2] := x;
    dec(b2);
  end;
end;

//
// R_DrawPlanes
// At the end of each frame.
//
procedure R_DrawPlanes;
var
  pl: Pvisplane_t;
  i: integer;
  light: integer;
  x: integer;
  stop: integer;
  angle: integer;
begin
  for i := 0 to lastvisplane - 1 do
  begin
    pl := @visplanes[i];
//    pl := visplanes[i];
    if pl.minx > pl.maxx then
      continue;


    // sky flat
    if pl.picnum = skyflatnum then
    begin
      dc_iscale := pspriteiscale;

      // Sky is allways drawn full bright,
      //  i.e. colormaps[0] is used.
      // Because of this hack, sky is not affected
      //  by INVUL inverse mapping.
      dc_colormap := Plighttable_tArray(colormaps);
      dc_texturemid := skytexturemid;
      for x := pl.minx to pl.maxx do
      begin
        dc_yl := pl.top[x];
        dc_yh := pl.bottom[x];

        if dc_yl <= dc_yh then
        begin
          angle := _SHR(viewangle + xtoviewangle[x], ANGLETOSKYSHIFT);
          dc_x := x;
          dc_source := R_GetColumn(skytexture, angle);
        // VJ
        // call skycolfunc(), not colfunc()
          skycolfunc;
        end;
      end;
	    continue;
    end;

    // regular flat
    ds_source := W_CacheLumpNum(firstflat + flattranslation[pl.picnum], PU_STATIC);

    planeheight := abs(pl.height - viewz);
    light := _SHR(pl.lightlevel, LIGHTSEGSHIFT) + extralight;

    if light >= LIGHTLEVELS then
      light := LIGHTLEVELS - 1;

    if light < 0 then
      light := 0;

    planezlight := @zlight[light]; // VJ ???

    pl.top[pl.maxx + 1] := VISEND;
    pl.top[pl.minx - 1] := VISEND;

    stop := pl.maxx + 1;

    for x := pl.minx to stop do
      R_MakeSpans(x, pl.top[x - 1], pl.bottom[x - 1], pl.top[x], pl.bottom[x]);

    Z_ChangeTag(ds_source, PU_CACHE);
  end;
end;



(* Unit hu_stuff implementation *)


                                                                                                                                                               

function HU_TITLE: string;
begin
  result := mapnames[(gameepisode - 1) * 9 + gamemap - 1];
end;

function HU_TITLE2: string;
begin
  result := mapnames2[gamemap - 1];
end;

function HU_TITLEP: string;
begin
  result := mapnamesp[gamemap - 1];
end;

function HU_TITLET: string;
begin
  result := mapnamest[gamemap - 1];
end;

var
  hu_stuff_plr: Pplayer_t;
  w_title: hu_textline_t;
  w_chat: hu_itext_t;
  always_off: boolean;
  chat_dest: array[0..MAXPLAYERS - 1] of char;
  w_inputbuffer: array[0..MAXPLAYERS - 1] of hu_itext_t;

  w_message: hu_stext_t;
  message_counter: integer;

  headsupactive: boolean;


const
  HU_TITLEHEIGHT = 1;
  HU_TITLEX = 0;

function HU_TITLEY: integer;
begin
//  result := SCREENHEIGHT - ST_HEIGHT - 1 - hu_font[0].height;
  result := (200 - ST_HEIGHT) * SCREENHEIGHT div 200 - 1 - hu_font[0].height;
end;

const
  HU_INPUTTOGGLE: char = 't';

function HU_INPUTX: integer;
begin
  result := HU_MSGX;
end;

function HU_INPUTY: integer;
begin
  result := HU_MSGY + HU_MSGHEIGHT * (hu_font[0].height + 1)
end;

const
  HU_INPUTWIDTH = 64;
  HU_INPUTHEIGHT = 1;

var
  shiftxform: array[0..127] of char;

const
  french_shiftxform: array[0..127] of char = (
    #0,
    #1, #2, #3, #4, #5, #6, #7, #8, #9, #10,
    #11, #12, #13, #14, #15, #16, #17, #18, #19, #20,
    #21, #22, #23, #24, #25, #26, #27, #28, #29, #30,
    #31,
    ' ', '!', '"', '#', '$', '%', '&',
    '"', // shift-'
    '(', ')', '*', '+',
    '?', // shift-,
    '_', // shift--
    '>', // shift-.
    '?', // shift-/
    '0', // shift-0
    '1', // shift-1
    '2', // shift-2
    '3', // shift-3
    '4', // shift-4
    '5', // shift-5
    '6', // shift-6
    '7', // shift-7
    '8', // shift-8
    '9', // shift-9
    '/',
    '.', // shift-;
    '<',
    '+', // shift-=
    '>', '?', '@',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '[', // shift-[
    '!', // shift-backslash - OH MY GOD DOES WATCOM SUCK
    ']', // shift-]
    '"', '_',
    '''', // shift-`
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '{', '|', '}', '~', #127
  );

const
  english_shiftxform: array[0..127] of char = (
    #0,
    #1, #2, #3, #4, #5, #6, #7, #8, #9, #10,
    #11, #12, #13, #14, #15, #16, #17, #18, #19, #20,
    #21, #22, #23, #24, #25, #26, #27, #28, #29, #30,
    #31,
    ' ', '!', '"', '#', '$', '%', '&',
    '"', // shift-'
    '(', ')', '*', '+',
    '<', // shift-,
    '_', // shift--
    '>', // shift-.
    '?', // shift-/
    ')', // shift-0
    '!', // shift-1
    '@', // shift-2
    '#', // shift-3
    '$', // shift-4
    '%', // shift-5
    '^', // shift-6
    '&', // shift-7
    '*', // shift-8
    '(', // shift-9
    ':',
    ':', // shift-;
    '<',
    '+', // shift-=
    '>', '?', '@',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '[', // shift-[
    '!', // shift-backslash - OH MY GOD DOES WATCOM SUCK
    ']', // shift-]
    '"', '_',
    '''', // shift-`
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '{', '|', '}', '~', #127
  );

const
  frenchKeyMap: array[0..127] of char = (
    #0,
    #1,#2,#3,#4,#5,#6,#7,#8,#9,#10,
    #11,#12,#13,#14,#15,#16,#17,#18,#19,#20,
    #21,#22,#23,#24,#25,#26,#27,#28,#29,#30,
    #31,
    ' ','!','"','#','$','%','&','%','(',')','*','+',';','-',':','!',
    '0','1','2','3','4','5','6','7','8','9',':','M','<','=','>','?',
    '@','Q','B','C','D','E','F','G','H','I','J','K','L',',','N','O',
    'P','A','R','S','T','U','V','Z','X','Y','W','^','\','$','^','_',
    '@','Q','B','C','D','E','F','G','H','I','J','K','L',',','N','O',
    'P','A','R','S','T','U','V','Z','X','Y','W','^','\','$','^',#127
  );

function ForeignTranslation(ch: char): char;
begin
  if ch < #128 then
    result := frenchKeyMap[Ord(ch)]
  else
    result := ch;
end;

procedure HU_Init;
var
  i: integer;
  j: integer;
  buffer: string;
begin
  if language = french then
  begin
    for i := 0 to 127 do
      shiftxform[i] := french_shiftxform[i];
  end
  else
  begin
    for i := 0 to 127 do
      shiftxform[i] := english_shiftxform[i];
  end;

  // load the heads-up font
  j := Ord(HU_FONTSTART);
  for i := 0 to HU_FONTSIZE - 1 do
  begin
    buffer := 'STCFN' + IntToStrZfill(3, j);
    inc(j);
    hu_font[i] := Ppatch_t(W_CacheLumpName(buffer, PU_STATIC));
  end;
end;

procedure HU_Stop;
begin
  headsupactive := false;
end;

procedure HU_Start;
var
  i: integer;
  s: string;
begin
  if headsupactive then
    HU_Stop;

  hu_stuff_plr := @players[consoleplayer];
  message_on := false;
  message_dontfuckwithme := false;
  message_nottobefuckedwith := false;
  chat_on := false;

  // create the message widget
  HUlib_initSText(@w_message,
    HU_MSGX, HU_MSGY, HU_MSGHEIGHT,
    @hu_font,
    Ord(HU_FONTSTART), @message_on);

  // create the map title widget
  HUlib_initTextLine(@w_title,
    HU_TITLEX, HU_TITLEY,
    @hu_font,
    Ord(HU_FONTSTART));

  case gamemode of
    shareware,
    registered,
    retail:  s := HU_TITLE;

{ FIXME
      case pack_plut:
  s = HU_TITLEP;
  break;
      case pack_tnt:
  s = HU_TITLET;
  break;
}

  else
    s := HU_TITLE2;
  end;

  for i := 1 to Length(s) do
    HUlib_addCharToTextLine(@w_title, s[i]);

  // create the chat widget
  HUlib_initIText(@w_chat,
    HU_INPUTX, HU_INPUTY,
    @hu_font,
    Ord(HU_FONTSTART), @chat_on);

  // create the inputbuffer widgets
  for i := 0 to MAXPLAYERS - 1 do
    HUlib_initIText(@w_inputbuffer[i], 0, 0, nil, 0, @always_off);

  headsupactive := true;
end;

procedure HU_Drawer;
begin
  HUlib_drawSText(@w_message);
  HUlib_drawIText(@w_chat);
  if automapactive then
    HUlib_drawTextLine(@w_title, false);
end;

procedure HU_Erase;
begin
  HUlib_eraseSText(@w_message);
  HUlib_eraseIText(@w_chat);
  HUlib_eraseTextLine(@w_title);
end;

procedure HU_Ticker;
var
  i, rc: integer;
  c: char;
begin
  // tick down message counter if message is up
  if message_counter = 1 then
  begin
    message_counter := 0;
    message_on := false;
    message_nottobefuckedwith := false;
  end;

  if (showMessages <> 0) or message_dontfuckwithme then
  begin
    // display message if necessary
    if ((hu_stuff_plr._message <> '') and (not message_nottobefuckedwith)) or
       ((hu_stuff_plr._message <> '') and message_dontfuckwithme) then
    begin
      HUlib_addMessageToSText(@w_message, '', hu_stuff_plr._message);
      hu_stuff_plr._message := '';
      message_on := true;
      message_counter := HU_MSGTIMEOUT;
      message_nottobefuckedwith := message_dontfuckwithme;
      message_dontfuckwithme := false;
    end;
  end; // else message_on = false;

  // check for incoming chat characters
  if netgame then
  begin
    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;
      c := Chr(players[i].cmd.chatchar);
      if (i <> consoleplayer) and boolval(c) then
      begin
        if c <= Chr(HU_BROADCAST) then
          chat_dest[i] := c
        else
        begin
          if (c >= 'a') and (c <= 'z') then
            c := shiftxform[ord(c)];
          rc := intval(HUlib_keyInIText(@w_inputbuffer[i], Ord(c)));
          if (rc <> 0) and (Ord(c) = KEY_ENTER) then
          begin
            if (w_inputbuffer[i].l.len <> 0) and
               ((Ord(chat_dest[i]) = consoleplayer + 1) or (Ord(chat_dest[i]) = HU_BROADCAST)) then
            begin
              HUlib_addMessageToSText(@w_message,
                player_names[i],
                w_inputbuffer[i].l.l);

              message_nottobefuckedwith := true;
              message_on := true;
              message_counter := HU_MSGTIMEOUT;
              if gamemode = commercial then
                S_StartSound(nil, Ord(sfx_radio))
              else
                S_StartSound(nil, Ord(sfx_tink));
            end;
            HUlib_resetIText(@w_inputbuffer[i]);
          end;
        end;
        players[i].cmd.chatchar := 0;
      end;
    end;
  end;
end;

const
  QUEUESIZE = 128;

var
  chatchars: array[0..QUEUESIZE - 1] of char;
  head: integer;
  tail: integer;

procedure HU_queueChatChar(c: char);
begin
  if ((head + 1) and (QUEUESIZE - 1)) = tail then
    hu_stuff_plr._message := HUSTR_MSGU
  else
  begin
    chatchars[head] := c;
    head := (head + 1) and (QUEUESIZE - 1);
  end;
end;

function HU_dequeueChatChar: char;
begin
  if head <> tail then
  begin
    result := chatchars[tail];
    tail := (tail + 1) and (QUEUESIZE - 1);
  end
  else
    result := #0;
end;

var
  lastmessage: string;
  shiftdown: boolean;
  altdown: boolean;
  destination_keys: array[0..MAXPLAYERS - 1] of char;
  num_nobrainers: integer;

function HU_Responder(ev: Pevent_t): boolean;
var
  macromessage: string;
  c: char;
  i: integer;
  numplayers: integer;
begin
  result := false;

  if ev.data1 = KEY_RSHIFT then
  begin
    shiftdown := Ord(ev_keydown) <> 0;
    exit;
  end
  else if (ev.data1 = KEY_RALT) or (ev.data1 = KEY_LALT) then
  begin
    altdown := Ord(ev_keydown) <> 0;
    exit;
  end;

  if ev._type <> ev_keydown then
    exit;

  numplayers := 0;
  for i := 0 to MAXPLAYERS - 1 do
    if playeringame[i] then
      inc(numplayers);

  if not chat_on then
  begin
    if ev.data1 = HU_MSGREFRESH then
    begin
      message_on := true;
      message_counter := HU_MSGTIMEOUT;
      result := true;
    end
    else if netgame and (ev.data1 = Ord(HU_INPUTTOGGLE)) then
    begin
      result := true;
      chat_on := true;
      HUlib_resetIText(@w_chat);
      HU_queueChatChar(Chr(HU_BROADCAST));
    end
    else if netgame and (numplayers > 2) then
    begin
      for i := 0 to MAXPLAYERS - 1 do
      begin
        if ev.data1 = Ord(destination_keys[i]) then
        begin
          if playeringame[i] and (i <> consoleplayer) then
          begin
            result := true;
            chat_on := true;
            HUlib_resetIText(@w_chat);
            HU_queueChatChar(Chr(i + 1));
            break;
          end
          else if i = consoleplayer then
          begin
            inc(num_nobrainers);
            if num_nobrainers < 3 then
              hu_stuff_plr._message := HUSTR_TALKTOSELF1
            else if num_nobrainers < 6 then
              hu_stuff_plr._message := HUSTR_TALKTOSELF2
            else if num_nobrainers < 9 then
              hu_stuff_plr._message := HUSTR_TALKTOSELF3
            else if num_nobrainers < 32 then
              hu_stuff_plr._message := HUSTR_TALKTOSELF4
            else
              hu_stuff_plr._message := HUSTR_TALKTOSELF5;
          end
        end;
      end;
    end;
  end
  else
  begin
    c := Chr(ev.data1);
    // send a macro
    if altdown then
    begin
      c := Chr(Ord(c) - Ord('0'));
      if c > Chr(9) then
        exit;
      // fprintf(stderr, "got here\n");
      macromessage := chat_macros[Ord(c)];

      // kill last message with a '\n'
      HU_queueChatChar(Chr(KEY_ENTER)); // DEBUG!!!

      // send the macro message
      for i := 1 to Length(macromessage) do
        HU_queueChatChar(macromessage[i]);
      HU_queueChatChar(Chr(KEY_ENTER));

      // leave chat mode and notify that it was sent
      chat_on := false;
      lastmessage := chat_macros[Ord(c)];
      hu_stuff_plr._message := lastmessage;
      result := true;
    end
    else
    begin
      if language = french then
        c := ForeignTranslation(c);
      if shiftdown or ((c >= 'a') and (c <= 'z')) then
        c := shiftxform[Ord(c)];
      result := HUlib_keyInIText(@w_chat, Ord(c));
      if result then
        HU_queueChatChar(c);
      if Ord(c) = KEY_ENTER then
      begin
        chat_on := false;
        if w_chat.l.len <> 0 then
        begin
          lastmessage := w_chat.l.l;
          hu_stuff_plr._message := lastmessage;
        end
      end
      else if Ord(c) = KEY_ESCAPE then
        chat_on := false;
    end;
  end;
end;



(* Unit r_draw implementation *)


                                                                                                                                                       

//
// All drawing to the view buffer is accomplished in this file.
// The other refresh files only know about ccordinates,
//  not the architecture of the frame buffer.
// Conveniently, the frame buffer is a linear one,
//  and we need only the base address,
//  and the total size == width*height*depth/8.,
//

var
  ylookup: array[0..MAXHEIGHT - 1] of PByteArray;
  columnofs: array[0..MAXWIDTH - 1] of integer;

{// Color tables for different players,
//  translate a limited part to another
//  (color ramps used for  suit colors).
//
  translations: array[0..2,0..255] of byte;}

{// just for profiling
  dccount: integer;}

//
// A column is a vertical slice/span from a wall texture that,
//  given the DOOM style restrictions on the view orientation,
//  will always have constant z depth.
// Thus a special case loop for very fast rendering can
//  be used. It has also been used with Wolfenstein 3D.
//
procedure R_DrawColumn;
var
  count: integer;
  i: integer;
  dest: PByte;
  frac: fixed_t;
  fracstep: fixed_t;
begin
  count := dc_yh - dc_yl;

  // Zero length, column does not exceed a pixel.
  if count < 0 then
    exit;

  // Framebuffer destination address.
  // Use ylookup LUT to avoid multiply with ScreenWidth.
  // Use columnofs LUT for subwindows?
  dest := @((ylookup[dc_yl]^)[columnofs[dc_x]]);

  // Determine scaling,
  //  which is the only mapping to be done.
  fracstep := dc_iscale;
  frac := dc_texturemid + (dc_yl - centery) * fracstep;

  // Inner loop that does the actual texture mapping,
  //  e.g. a DDA-lile scaling.
  // This is as fast as it gets.
  for i := 0 to count do
  begin
    // Re-map color indices from wall texture column
    //  using a lighting/special effects LUT.
    dest^ := dc_colormap[dc_source[(frac div FRACUNIT) and 127]];
    inc(dest, SCREENWIDTH);
    frac := frac + fracstep;
  end;
end;

procedure R_DrawColumnLow;
var
  count: integer;
  i: integer;
  dest: PByte;
  bdest: byte;
  frac: fixed_t;
  fracstep: fixed_t;
begin
  count := (dc_yh - dc_yl) div 3;

  // Zero length.
  if count < 0 then
    exit;
    // Blocky mode, need to multiply by 2.
  dest := @((ylookup[dc_yl]^)[columnofs[dc_x]]);

  frac := dc_texturemid + (dc_yl - centery) * dc_iscale;
  fracstep := 3 * dc_iscale;

  for i := 0 to count do
  begin
    // Hack. Does not work corretly.
    bdest := dc_colormap[dc_source[(frac div FRACUNIT) and 127]];
    dest^ := bdest;
  	inc(dest, SCREENWIDTH);
    dest^ := bdest;
  	inc(dest, SCREENWIDTH);
    dest^ := bdest;
  	inc(dest, SCREENWIDTH);
    frac := frac + fracstep;
  end;
end;

procedure R_DrawSkyColumn;
var
  count: integer;
  i: integer;
  dest: PByte;
  frac: fixed_t;
  fracstep: fixed_t;
begin
  count := dc_yh - dc_yl;

  // Zero length, column does not exceed a pixel.
  if count < 0 then
    exit;

  // Framebuffer destination address.
  // Use ylookup LUT to avoid multiply with ScreenWidth.
  // Use columnofs LUT for subwindows?
  dest := @((ylookup[dc_yl]^)[columnofs[dc_x]]);

  // Determine scaling,
  //  which is the only mapping to be done.
//  fracstep := dc_iscale;// * 200 div SCREENHEIGHT;
  fracstep := FRACUNIT * 200 div SCREENHEIGHT;
  frac := dc_texturemid + (dc_yl - centery) * fracstep;

  // Inner loop that does the actual texture mapping,
  //  e.g. a DDA-lile scaling.
  // This is as fast as it gets.
  for i := 0 to count do 
  begin
    // Re-map color indices from wall texture column
    //  using a lighting/special effects LUT.
    dest^ := dc_source[(frac div FRACUNIT) and 127]; // VJ removed colormap, sky is full bright

    inc(dest, SCREENWIDTH);
    frac := frac + fracstep;
  end;
end;

//
// Spectre/Invisibility.
//
const
  FUZZTABLE = 50;
  FUZZOFF = 1;


  fuzzoffset: array[0..FUZZTABLE - 1] of integer = (
    FUZZOFF,-FUZZOFF, FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF,
    FUZZOFF, FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF,
    FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,
    FUZZOFF,-FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF,
    FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF,-FUZZOFF, FUZZOFF,
    FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF,
    FUZZOFF, FUZZOFF,-FUZZOFF, FUZZOFF, FUZZOFF,-FUZZOFF, FUZZOFF
  );

var
  fuzzpos: integer;

//
// Framebuffer postprocessing.
// Creates a fuzzy image by copying pixels
//  from adjacent ones to left and right.
// Used with an all black colormap, this
//  could create the SHADOW effect,
//  i.e. spectres and invisible players.
//
procedure R_DrawFuzzColumn;
var
  count: integer;
  i: integer;
  dest: PByteArray;
begin
  // Adjust borders. Low...
  if dc_yl = 0 then
    dc_yl := 1;

  // .. and high.
  if dc_yh = viewheight - 1 then
    dc_yh := viewheight - 2;

  count := dc_yh - dc_yl;

  // Zero length.
  if count < 0 then
    exit;

  // Does not work with blocky mode.
  dest := PByteArray(integer(ylookup[dc_yl]) + columnofs[dc_x]);

  // Looks like an attempt at dithering,
  //  using the colormap #6 (of 0-31, a bit
  //  brighter than average).
  for i := 0 to count do
  begin
    // Lookup framebuffer, and retrieve
    //  a pixel that is either one column
    //  left or right of the current one.
    // Add index from colormap to index.
    dest[0] := colormaps[6 * 256 + dest[SCREENWIDTH * fuzzoffset[fuzzpos]]];

    // Clamp table lookup index.
    inc(fuzzpos);
    if fuzzpos = FUZZTABLE then
      fuzzpos := 0;

    incp(pointer(dest), SCREENWIDTH);

  end;
end;

//
// R_DrawTranslatedColumn
// Used to draw player sprites
//  with the green colorramp mapped to others.
// Could be used with different translation
//  tables, e.g. the lighter colored version
//  of the BaronOfHell, the HellKnight, uses
//  identical sprites, kinda brightened up.
//

procedure R_DrawTranslatedColumn;
var
  count: integer;
  dest: PByte;
  frac: fixed_t;
  fracstep: fixed_t;
  i: integer;
begin
  count := dc_yh - dc_yl;
  if count < 0 then
    exit;

  // FIXME. As above.
  dest := PByte(integer(ylookup[dc_yl]) + columnofs[dc_x]);

  // Looks familiar.
  fracstep := dc_iscale;
  frac := dc_texturemid + (dc_yl - centery) * fracstep;

  // Here we do an additional index re-mapping.
  for i := 0 to count do
  begin
    // Translation tables are used
    //  to map certain colorramps to other ones,
    //  used with PLAY sprites.
    // Thus the "green" ramp of the player 0 sprite
    //  is mapped to gray, red, black/indigo.
    dest^ := dc_colormap[dc_translation[dc_source[frac div FRACUNIT]]];
    inc(dest, SCREENWIDTH);

    frac := frac + fracstep;
  end;
end;

//
// R_InitTranslationTables
// Creates the translation tables to map
//  the green color ramp to gray, brown, red.
// Assumes a given structure of the PLAYPAL.
// Could be read from a lump instead.
//
procedure R_InitTranslationTables;
var
  i: integer;
begin
  translationtables := Z_Malloc (256 * 3 + 255, PU_STATIC, nil);
  translationtables := PByteArray((integer(translationtables) + 255 ) and (not 255));

  // translate just the 16 green colors
  for i := 0 to 255 do
    if (i >= $70) and (i <= $7f) then
    begin
      // map green ramp to gray, brown, red
      translationtables[i] := $60 + (i and $f);
      translationtables[i + 256] := $40 + (i and $f);
      translationtables[i + 512] := $20 + (i and $f);
    end
    else
    begin
      // Keep all other colors as is.
      translationtables[i] := i;
      translationtables[i + 256] := i;
      translationtables[i + 512] := i;
    end;
end;

//
// R_DrawSpan
// With DOOM style restrictions on view orientation,
//  the floors and ceilings consist of horizontal slices
//  or spans with constant z depth.
// However, rotation around the world z axis is possible,
//  thus this mapping, while simpler and faster than
//  perspective correct texture mapping, has to traverse
//  the texture at an angle in all but a few cases.
// In consequence, flats are not stored by column (like walls),
//  and the inner loop has to step in texture space u and v.
//

//
// Draws the actual span.
procedure R_DrawSpan;
var
  xfrac: fixed_t;
  yfrac: fixed_t;
  dest: PByte;
  count: integer;
  i: integer;
  spot: integer;
begin
  xfrac := ds_xfrac;
  yfrac := ds_yfrac;

  dest := PByte(integer(ylookup[ds_y]) + columnofs[ds_x1]);

  // We do not check for zero spans here?
  count := ds_x2 - ds_x1;

  for i := 0 to count do
  begin
    // Current texture index in u,v.
    spot := (yfrac div 1024) and (63 * 64) + (xfrac div FRACUNIT) and 63;

    // Lookup pixel from flat texture tile,
    //  re-index using light/colormap.

	  dest^ := ds_colormap[ds_source[spot]];
    inc(dest);

    // Next step in u,v.
    xfrac := xfrac + ds_xstep;
    yfrac := yfrac + ds_ystep;
  end;
end;

//
// Again..
//
procedure R_DrawSpanLow;
var
  xfrac: fixed_t;
  yfrac: fixed_t;
  dest: PByte;
  bdest: byte;
  ds_xstep2: fixed_t;
  ds_ystep2: fixed_t;
  count: integer;
  i: integer;
  spot: integer;
begin
  xfrac := ds_xfrac;
  yfrac := ds_yfrac;


  dest := PByte(integer(ylookup[ds_y]) + columnofs[ds_x1]);

  // Blocky mode, multiply by 3 (!!).
  ds_xstep2 := ds_xstep * 3;
  ds_ystep2 := ds_ystep * 3;

  count := (ds_x2 - ds_x1) div 3;
  for i := 0 to count do
  begin
//    spot := (_SHR(yfrac, 10) and (63 * 64)) + (_SHR(xfrac, 16) and 63);
    spot := (yfrac div 1024) and (63 * 64) + (xfrac div FRACUNIT) and 63;
    // Lowres/blocky mode does it twice,
    //  while scale is adjusted appropriately.
    bdest := ds_colormap[ds_source[spot]];
    dest^ := bdest;
    inc(dest);
    dest^ := bdest;
    inc(dest);
    dest^ := bdest;
    inc(dest);

    xfrac := xfrac + ds_xstep2;
    yfrac := yfrac + ds_ystep2;
  end;
end;

//
// R_InitBuffer
// Creats lookup tables that avoid
//  multiplies and other hazzles
//  for getting the framebuffer address
//  of a pixel to draw.
//
procedure R_InitBuffer(width, height: integer);
var
  i: integer;
begin
  // Handle resize,
  //  e.g. smaller view windows
  //  with border and/or status bar.
  viewwindowx := _SHR(SCREENWIDTH - width, 1);

  // Column offset. For windows.
  for i := 0 to width - 1 do
    columnofs[i] := viewwindowx + i;

  // Samw with base row offset.
  if width = SCREENWIDTH then
    viewwindowy := 0
  else
//    viewwindowy := _SHR(SCREENHEIGHT - ST_HEIGHT - height, 1);
    viewwindowy := _SHR(V_PreserveY(200 - ST_HEIGHT) - height, 1);
//    viewwindowy := _SHR(V_PreserveH(200, 200 - ST_HEIGHT) - height, 1);
//    viewwindowy := _SHR(SCREENHEIGHT - V_PreserveY(ST_HEIGHT) - height + 1, 1);
//    viewwindowy := _SHR(SCREENHEIGHT - V_PreserveH(200 - ST_HEIGHT, ST_HEIGHT) - height + 1, 1);

  // Preclaculate all row offsets.
  for i := 0 to height - 1 do
    ylookup[i] := PByteArray(integer(screens[_FG]) + (i + viewwindowy) * SCREENWIDTH);
end;

//
// R_FillBackScreen
// Fills the back screen with a pattern
//  for variable screen sizes
// Also draws a beveled edge.
//
procedure R_FillBackScreen320x200;
var
  src: PByteArray;
  dest: PByteArray;
  x: integer;
  y: integer;
  patch: Ppatch_t;
  name: string;
begin
  if scaledviewwidth = SCREENWIDTH then
  begin
    if V_NeedsPreserve then
    begin
      dest := @screens[0][viewheight * SCREENWIDTH];
      ZeroMemory(dest^, SCREENWIDTH * (V_PreserveY(200 - ST_HEIGHT) - viewheight));
    end;
    exit;
  end;

  if gamemode = commercial then
    name := 'GRNROCK'   // DOOM II border patch.
  else
    name := 'FLOOR7_2'; // DOOM border patch.

  src := W_CacheLumpName(name, PU_CACHE);
  dest := screens[1];

  for y := 0 to SCREENHEIGHT - V_PreserveY(ST_HEIGHT) - 1 do //- 1 do // do // - 1 do
  begin
    for x := 0 to SCREENWIDTH div 64 - 1 do
    begin
	    memcpy(dest, PByteArray(integer(src) + _SHL(y and 63, 6)), 64);
      incp(pointer(dest), 64);
    end;
    if (SCREENWIDTH and 63) <> 0 then
    begin
      memcpy(dest, PByteArray(integer(src) + _SHL(y and 63, 6)), SCREENWIDTH and 63);
      incp(pointer(dest), (SCREENWIDTH and 63));
    end;
  end;

  patch := W_CacheLumpName('brdr_t', PU_CACHE);
  x := 0;
  while x < scaledviewwidth do
  begin
    V_DrawPatch(viewwindowx + x, viewwindowy - 8, 1, patch, false);
    x := x + 8;
  end;

  patch := W_CacheLumpName('brdr_b', PU_CACHE);
  x := 0;
  while x < scaledviewwidth do
  begin
    V_DrawPatch(viewwindowx + x, viewwindowy + viewheight, 1, patch, false);
    x := x + 8;
  end;

  patch := W_CacheLumpName('brdr_l', PU_CACHE);
  y := 0;
  while y < viewheight do
  begin
    V_DrawPatch(viewwindowx - 8, viewwindowy + y, 1, patch, false);
    y := y + 8;
  end;

  patch := W_CacheLumpName('brdr_r', PU_CACHE);
  y := 0;
  while y < viewheight do
  begin
    V_DrawPatch (viewwindowx + scaledviewwidth, viewwindowy + y, 1, patch, false);
    y := y + 8;
  end;

  // Draw beveled edge.
  V_DrawPatch(viewwindowx - 8, viewwindowy - 8, 1,
    W_CacheLumpName('brdr_tl', PU_CACHE), false);

  V_DrawPatch(viewwindowx + scaledviewwidth, viewwindowy - 8, 1,
    W_CacheLumpName('brdr_tr', PU_CACHE), false);

  V_DrawPatch(viewwindowx - 8, viewwindowy + viewheight, 1,
    W_CacheLumpName('brdr_bl', PU_CACHE), false);

  V_DrawPatch(viewwindowx + scaledviewwidth, viewwindowy + viewheight, 1,
    W_CacheLumpName('brdr_br', PU_CACHE), false);
end;

procedure R_FillBackScreen;
var
  src: PByteArray;
  dest: PByteArray;
  x: integer;
  y: integer;
  patch: Ppatch_t;
  name: string;
  pscaledviewwidth: integer;
  pviewwindowx: integer;
  pviewwindowy: integer;
  pviewheight: integer;
begin
  if (SCREENWIDTH = 320) and (SCREENHEIGHT = 200) then
  begin
    R_FillBackScreen320x200;
    exit;
  end;

  if scaledviewwidth = SCREENWIDTH then
    exit;

  if gamemode = commercial then
    name := 'GRNROCK'   // DOOM II border patch.
  else
    name := 'FLOOR7_2'; // DOOM border patch.

  src := W_CacheLumpName(name, PU_CACHE);
  dest := screens[_TMP];

  for y := 0 to 200 - ST_HEIGHT - 1 do //- 1 do // do // - 1 do
  begin
    for x := 0 to 320 div 64 - 1 do
    begin
	    memcpy(dest, PByteArray(integer(src) + _SHL(y and 63, 6)), 64);
      incp(pointer(dest), 64);
    end;
    if (320 and 63) <> 0 then
    begin
      memcpy(dest, PByteArray(integer(src) + _SHL(y and 63, 6)), 320 and 63);
      incp(pointer(dest), (320 and 63));
    end;
  end;

  patch := W_CacheLumpName('brdr_t', PU_CACHE);
  x := 0;
  pscaledviewwidth := scaledviewwidth * 320 div SCREENWIDTH;
  pviewwindowx := viewwindowx * 320 div SCREENWIDTH;
  pviewwindowy := viewwindowy * 200 div SCREENHEIGHT;
  pviewheight := viewheight * 200 div SCREENHEIGHT;

  while x < pscaledviewwidth do
  begin
    V_DrawPatch(pviewwindowx + x, pviewwindowy - 8, _TMP, patch, false);
    x := x + 8;
  end;

  patch := W_CacheLumpName('brdr_b', PU_CACHE);
  x := 0;
  while x < pscaledviewwidth do
  begin
    V_DrawPatch(pviewwindowx + x, pviewwindowy + pviewheight, _TMP, patch, false);
    x := x + 8;
  end;

  patch := W_CacheLumpName('brdr_l', PU_CACHE);
  y := 0;
  while y < pviewheight do
  begin
    V_DrawPatch(pviewwindowx - 8, pviewwindowy + y, _TMP, patch, false);
    y := y + 8;
  end;

  patch := W_CacheLumpName('brdr_r', PU_CACHE);
  y := 0;
  while y < pviewheight do
  begin
    V_DrawPatch (pviewwindowx + pscaledviewwidth, pviewwindowy + y, _TMP, patch, false);
    y := y + 8;
  end;

  // Draw beveled edge.
  V_DrawPatch(pviewwindowx - 8, pviewwindowy - 8, _TMP,
    W_CacheLumpName('brdr_tl', PU_CACHE), false);

  V_DrawPatch(pviewwindowx + pscaledviewwidth, pviewwindowy - 8, _TMP,
    W_CacheLumpName('brdr_tr', PU_CACHE), false);

  V_DrawPatch(pviewwindowx - 8, pviewwindowy + pviewheight, _TMP,
    W_CacheLumpName('brdr_bl', PU_CACHE), false);

  V_DrawPatch(pviewwindowx + pscaledviewwidth, pviewwindowy + pviewheight, _TMP,
    W_CacheLumpName('brdr_br', PU_CACHE), false);

  V_CopyRect(0, 0, _TMP, 320, 200, 0, 0, 1, true);
end;

//
// Copy a screen buffer.
//
procedure R_VideoErase(const ofs: integer; const count: integer);
begin
  // LFB copy.
  // This might not be a good idea if memcpy
  //  is not optiomal, e.g. byte by byte on
  //  a 32bit CPU, as GNU GCC/Linux libc did
  //  at one point.
  memcpy(Pointer(integer(screens[_FG]) + ofs), Pointer(integer(screens[1]) + ofs), count);
end;

//
// R_DrawViewBorder
// Draws the border around the view
//  for different size windows?
//
procedure R_DrawViewBorder;
var
  top: integer;
  bottom: integer;
  side: integer;
  ofs: integer;
  i: integer;
begin
  if scaledviewwidth = SCREENWIDTH then
    exit;

  top := (SCREENHEIGHT - V_PreserveY(ST_HEIGHT) - viewheight) div 2;
  bottom := SCREENHEIGHT - V_PreserveY(ST_HEIGHT) - top - viewheight;
  side := (SCREENWIDTH - scaledviewwidth) div 2;

  // copy top and one line of left side
  R_VideoErase(0, top * SCREENWIDTH + side);

  // copy one line of right side and bottom
//  ofs := (viewheight + top) * SCREENWIDTH - side;
//  R_VideoErase (ofs, top * SCREENWIDTH + side);
  if odd(bottom) then
    ofs := (viewheight + bottom - 1) * SCREENWIDTH - side
  else
    ofs := (viewheight + bottom) * SCREENWIDTH - side;
  R_VideoErase (ofs, bottom * SCREENWIDTH + side);

  // copy sides using wraparound
  ofs := top * SCREENWIDTH + SCREENWIDTH - side;
  side := _SHL(side, 1);

  for i := 1 to viewheight - 1 do
  begin
    R_VideoErase(ofs, side);
    ofs := ofs + SCREENWIDTH;
  end;

  // ?
//  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT - SBARHEIGHT);
  V_MarkRect(0, 0, 320, 200 - ST_HEIGHT, true);
end;




(* Unit st_lib implementation *)


                                                                                  // automapactive

//
// Hack display negative frags.
//  Loads and store the stminus lump.
//
var
  sttminus: Ppatch_t;

procedure STlib_init;
begin
  sttminus := Ppatch_t(W_CacheLumpName('STTMINUS', PU_STATIC));
end;

// ?
procedure STlib_initNum(n: Pst_number_t; x, y: integer; pl: Ppatch_tPArray;
  num: PInteger; _on: PBoolean; width: integer);
begin
  n.x := x;
  n.y := y;
  n.oldnum := 0;
  n.width := width;
  n.num := num;
  n._on := _on;
  n.p := pl;
end;

//
// A fairly efficient way to draw a number
//  based on differences from the old number.
// Note: worth the trouble?
//
procedure STlib_drawNum(n: Pst_number_t; refresh: boolean);
var
  numdigits: integer;
  num: integer;
  w: integer;
  h: integer;
  x: integer;
  neg: boolean;
begin
  numdigits := n.width;
  num := n.num^;

  w := n.p[0].width;
  h := n.p[0].height;

  n.oldnum := num;

  neg := num < 0;

  if neg then
  begin
    if (numdigits = 2) and (num < -9) then
      num := -9
    else if (numdigits = 3) and (num < -99) then
      num := -99;

    num := -num;
  end;

  // clear the area
  x := n.x - numdigits * w;

  if n.y - ST_Y < 0 then
    I_Error('STlib_drawNum() : n.y - ST_Y < 0');

  V_CopyRect(x, n.y - ST_Y, _BG, w * numdigits, h, x, n.y, _FG, true);

  // if non-number, do not draw it
  if num = 1994 then
    exit;

  x := n.x;

  // in the special case of 0, you draw 0
  if not boolval(num) then
//    V_DrawPatch(x - w, n.y, _FG, n.p[0], true);
    V_DrawPatch(x - w, n.y - ST_Y, _BG, n.p[0], false);

  // draw the new number
  while boolval(num) and boolval(numdigits) do
  begin
    x := x - w;
//    V_DrawPatch(x, n.y, _FG, n.p[num mod 10], true);
    V_DrawPatch(x, n.y - ST_Y, _BG, n.p[num mod 10], false);
    num := num div 10;
    dec(numdigits);
  end;

  // draw a minus sign if necessary
  if neg then
//    V_DrawPatch(x - 8, n.y, _FG, sttminus, true);
    V_DrawPatch(x - 8, n.y - ST_Y, _BG, sttminus, false);
end;

//
procedure STlib_updateNum(n: Pst_number_t; refresh: boolean);
begin
  if n._on^ then
    STlib_drawNum(n, refresh);
end;

//
procedure STlib_initPercent(p: Pst_percent_t; x, y: integer; pl: Ppatch_tPArray;
  num: PInteger; _on: PBoolean; percent: Ppatch_t);
begin
  STlib_initNum(@p.n, x, y, pl, num, _on, 3);
  p.p := percent;
end;

procedure STlib_updatePercent(per: Pst_percent_t; refresh: integer);
begin
  if boolval(refresh) and per.n._on^ then
    V_DrawPatch(per.n.x, per.n.y - ST_Y, _BG, per.p, false);

  STlib_updateNum(@per.n, boolval(refresh));
end;

procedure STlib_initMultIcon(i: Pst_multicon_t; x, y: integer; il: Ppatch_tPArray;
  inum: PInteger; _on: PBoolean);
begin
  i.x	:= x;
  i.y	:= y;
  i.oldinum := -1;
  i.inum := inum;
  i._on := _on;
  i.p := il;
end;

procedure STlib_updateMultIcon(mi: Pst_multicon_t; refresh: boolean);
var
  y: integer;
begin
  if mi._on^ and ((mi.oldinum <> mi.inum^) or refresh) and (mi.inum^ <> -1) then
  begin
    if mi.oldinum <> -1 then
    begin
      y := mi.y - mi.p[mi.oldinum].topoffset;

      if y - ST_Y < 0 then
        I_Error('STlib_updateMultIcon(): y - ST_Y < 0');

    end;
    V_DrawPatch(mi.x, mi.y - ST_Y, _BG, mi.p[mi.inum^], false);
    mi.oldinum := mi.inum^;
  end;
end;

procedure STlib_initBinIcon(b: Pst_binicon_t; x, y: integer; i: Ppatch_t;
  val: PBoolean; _on: PBoolean);
begin
  b.x	:= x;
  b.y	:= y;
  b.oldval := false;
  b.val := val;
  b._on := _on;
  b.p := i;
end;

procedure STlib_updateBinIcon(bi: Pst_binicon_t; refresh: boolean);
var
  y: integer;
begin
  if bi._on^ and ((bi.oldval <> bi.val^) or refresh) then
  begin
    y := bi.y - bi.p.topoffset;

    if y - ST_Y < 0 then
      I_Error('STlib_updateBinIcon(): y - ST_Y < 0');

    if bi.val^ then
      V_DrawPatch(bi.x, bi.y - ST_Y, _BG, bi.p, false);

    bi.oldval := bi.val^;
  end;
end;



(* Unit p_sight implementation *)


                                                                                          

//
// P_CheckSight
//
var
  sightzstart: fixed_t; // eye z of looker

  strace: divline_t; // from t1 to t2
  t2x: fixed_t;
  t2y: fixed_t;

  sightcounts: array[0..1] of integer;

//
// P_DivlineSide
// Returns side 0 (front), 1 (back), or 2 (on).
//
function P_DivlineSide(x, y: fixed_t; node: Pdivline_t): integer;
var
  dx: fixed_t;
  dy: fixed_t;
  left: fixed_t;
  right: fixed_t;
begin
  if node.dx = 0 then
  begin
    if x = node.x then
    begin
      result := 2;
      exit;
    end;
    if x <= node.x then
    begin
      result := intval(node.dy > 0);
      exit;
    end;
    result := intval(node.dy < 0);
    exit;
  end;

  if node.dy = 0 then
  begin
    if x = node.y then
    begin
      result := 2;
      exit;
    end;
    if y <= node.y then
    begin
      result := intval(node.dx < 0);
      exit;
    end;
    result := intval(node.dx > 0);
    exit;
  end;

  dx := (x - node.x);
  dy := (y - node.y);

  left := (node.dy div FRACUNIT) * (dx div FRACUNIT);
  right := (dy div FRACUNIT) * (node.dx div FRACUNIT);

  if right < left then
  begin
    result := 0; // front side
    exit;
  end;

  if left = right then
    result := 2
  else
    result := 1; // back side
end;

//
// P_InterceptVector2
// Returns the fractional intercept point
// along the first divline.
// This is only called by the addthings and addlines traversers.
//
function P_InterceptVector2(v2, v1: Pdivline_t): fixed_t;
var
  num: fixed_t;
  den: fixed_t;
begin
  den := FixedMul(_SHR(v1.dy, 8), v2.dx) - FixedMul(_SHR(v1.dx, 8), v2.dy);

  if den = 0 then
  begin
    result := 0;
    exit;
    //	I_Error ("P_InterceptVector: parallel");
  end;

  num := FixedMul(_SHR(v1.x - v2.x, 8), v1.dy) +
         FixedMul(_SHR(v2.y - v1.y, 8), v1.dx);

  result := FixedDiv(num , den);
end;

//
// P_CrossSubsector
// Returns true
//  if strace crosses the given subsector successfully.
//
function P_CrossSubsector(num: integer): boolean;
var
  seg: Pseg_t;
  line: Pline_t;
  s1: integer;
  s2: integer;
  i: integer;
  sub: Psubsector_t;
  front: Psector_t;
  back: Psector_t;
  opentop: fixed_t;
  openbottom: fixed_t;
  divl: divline_t;
  v1: Pvertex_t;
  v2: Pvertex_t;
  frac: fixed_t;
  slope: fixed_t;
begin
  sub := @subsectors[num];

  // check lines
  for i := sub.firstline to sub.firstline + sub.numlines - 1 do
  begin
    seg := @segs[i];
    line := seg.linedef;

    // allready checked other side?
    if line.validcount = validcount then
      continue;

    line.validcount := validcount;

    v1 := line.v1;
    v2 := line.v2;
    s1 := P_DivlineSide(v1.x, v1.y, @strace);
    s2 := P_DivlineSide(v2.x, v2.y, @strace);

    // line isn't crossed?
    if s1 = s2 then
      continue;

    divl.x := v1.x;
    divl.y := v1.y;
    divl.dx := v2.x - v1.x;
    divl.dy := v2.y - v1.y;
    s1 := P_DivlineSide(strace.x, strace.y, @divl);
    s2 := P_DivlineSide(t2x, t2y, @divl);

    // line isn't crossed?
    if s1 = s2 then
      continue;

    // stop because it is not two sided anyway
    // might do this after updating validcount?
    if not boolval(line.flags and ML_TWOSIDED) then
    begin
      result := false;
      exit;
    end;

    // crosses a two sided line
    front := seg.frontsector;
    back := seg.backsector;

    // no wall to block sight with?
    if (front.floorheight = back.floorheight) and
       (front.ceilingheight = back.ceilingheight) then
      continue;

    // possible occluder
    // because of ceiling height differences
    if front.ceilingheight < back.ceilingheight then
      opentop := front.ceilingheight
    else
      opentop := back.ceilingheight;

    // because of ceiling height differences
    if front.floorheight > back.floorheight then
      openbottom := front.floorheight
    else
      openbottom := back.floorheight;

    // quick test for totally closed doors
    if openbottom >= opentop then
    begin
      result := false; // stop
      exit;
    end;

    frac := P_InterceptVector2(@strace, @divl);

    if front.floorheight <> back.floorheight then
    begin
      slope := FixedDiv(openbottom - sightzstart, frac);
      if slope > bottomslope then
        bottomslope := slope;
    end;

    if front.ceilingheight <> back.ceilingheight then
    begin
      slope := FixedDiv(opentop - sightzstart, frac);
      if slope < topslope then
        topslope := slope;
    end;

    if topslope <= bottomslope then
    begin
      result := false; // stop
      exit;
    end;
  end;

  // passed the subsector ok
  result := true;
end;

//
// P_CrossBSPNode
// Returns true
//  if strace crosses the given node successfully.
//
function P_CrossBSPNode(bspnum: integer): boolean;
var
  bsp: Pnode_t;
  side: integer;
begin
  if boolval(bspnum and NF_SUBSECTOR) then
  begin
    if bspnum = -1 then
      result := P_CrossSubsector(0)
    else
      result := P_CrossSubsector(bspnum and (not NF_SUBSECTOR));
    exit;
  end;

  bsp := @nodes[bspnum];

  // decide which side the start point is on
  side := P_DivlineSide(strace.x, strace.y, Pdivline_t(bsp));
  if side = 2 then
    side := 0; // an "on" should cross both sides

  // cross the starting side
  if not P_CrossBSPNode(bsp.children[side]) then
  begin
    result := false;
    exit;
  end;

  // the partition plane is crossed here
  if side = P_DivlineSide(t2x, t2y, Pdivline_t(bsp)) then
  begin
    // the line doesn't touch the other side
    result := true;
    exit;
  end;

  // cross the ending side
  result := P_CrossBSPNode(bsp.children[side xor 1]);
end;

//
// P_CheckSight
// Returns true
//  if a straight line between t1 and t2 is unobstructed.
// Uses REJECT.
//
function P_CheckSight(t1: Pmobj_t; t2: Pmobj_t): boolean;
var
  s1: integer;
  s2: integer;
  pnum: integer;
  bytenum: integer;
  bitnum: integer;
begin
  // First check for trivial rejection.

  // Determine subsector entries in REJECT table.
  s1 := pOperation(Psubsector_t(t1.subsector).sector, @sectors[0], '-', SizeOf(sectors[0]));
  s2 := pOperation(Psubsector_t(t2.subsector).sector, @sectors[0], '-', SizeOf(sectors[0]));
  pnum := s1 * numsectors + s2;
  bytenum := _SHR(pnum, 3);
  bitnum := 1 shl (pnum and 7);

  // Check in REJECT table.
  if boolval(rejectmatrix[bytenum] and bitnum) then
  begin
    sightcounts[0] := sightcounts[0] + 1;

    // can't possibly be connected
    result := false;
    exit;
  end;

  // An unobstructed LOS is possible.
  // Now look from eyes of t1 to any part of t2.
  sightcounts[1] := sightcounts[1] + 1;

  inc(validcount);

  sightzstart := t1.z + t1.height - _SHR(t1.height, 2);
  topslope := (t2.z + t2.height) - sightzstart;
  bottomslope := t2.z - sightzstart;

  strace.x := t1.x;
  strace.y := t1.y;
  t2x := t2.x;
  t2y := t2.y;
  strace.dx := t2.x - t1.x;
  strace.dy := t2.y - t1.y;

  // the head node is the last node output
  result := P_CrossBSPNode(numnodes - 1);
end;


(* Unit p_local implementation *)


               

function ONFLOORZ: integer;
begin
  result := MININT;
end;

function ONCEILINGZ: integer;
begin
  result := MAXINT;
end;


(* Unit p_tick implementation *)


                                                                                                    

procedure P_InitThinkers;
begin
  thinkercap.prev := @thinkercap;
  thinkercap.next := @thinkercap;
end;

//
// P_AddThinker
// Adds a new thinker at the end of the list.
//
procedure P_AddThinker(thinker: Pthinker_t);
begin
  thinkercap.prev.next := thinker;
  thinker.next := @thinkercap;
  thinker.prev := thinkercap.prev;
  thinkercap.prev := thinker;
end;

//
// P_RemoveThinker
// Deallocation is lazy -- it will not actually be freed
// until its thinking turn comes up.
//
procedure P_RemoveThinker(thinker: Pthinker_t);
begin
  // FIXME: NOP.
  thinker._function.acv := nil;
end;

//
// P_AllocateThinker
// Allocates memory and adds a new thinker at the end of the list.
//
procedure P_AllocateThinker(thinker: Pthinker_t);
begin
end;

//
// P_RunThinkers
//
procedure P_RunThinkers;
var
  currentthinker: Pthinker_t;
begin
  currentthinker := thinkercap.next;
  while currentthinker <> @thinkercap do
  begin
    if not Assigned(currentthinker._function.acv) then
    begin
      // time to remove it
      currentthinker.next.prev := currentthinker.prev;
      currentthinker.prev.next := currentthinker.next;
      Z_Free(currentthinker);
    end
    else
    begin
      if Assigned(currentthinker._function.acp1) then
        currentthinker._function.acp1(currentthinker);
    end;
    currentthinker := currentthinker.next;
  end;
end;

//
// P_Ticker
//
procedure P_Ticker;
var
  i: integer;
begin
  // run the tic
  if paused then
    exit;

  // pause if in menu and at least one tic has been run
  if (not netgame) and menuactive and
     (not demoplayback) and (players[consoleplayer].viewz <> 1) then
    exit;


  for i := 0 to MAXPLAYERS - 1 do
    if playeringame[i] then
      P_PlayerThink(@players[i]);

  P_RunThinkers;
  P_UpdateSpecials;
  P_RespawnSpecials;

  // for par times
  inc(leveltime);
end;


(* Unit p_maputl implementation *)


                                                   

//
// P_AproxDistance
// Gives an estimation of distance (not exact)
//
function P_AproxDistance(dx: fixed_t; dy: fixed_t): fixed_t;
begin
  dx := abs(dx);
  dy := abs(dy);
  if dx < dy then
    result := dx + dy - _SHR(dx, 1)
  else
    result := dx + dy - _SHR(dy, 1);
end;

//
// P_PointOnLineSide
// Returns 0 or 1
//
function P_PointOnLineSide(x: fixed_t; y: fixed_t; line: Pline_t): integer;
var
  dx: fixed_t;
  dy: fixed_t;
  left: fixed_t;
  right: fixed_t;
begin
  if line.dx = 0 then
  begin
    if x <= line.v1.x then
      result := intval(line.dy > 0)
    else
      result := intval(line.dy < 0);
    exit;
  end;

  if line.dy = 0 then
  begin
    if y <= line.v1.y then
      result := intval(line.dx < 0)
    else
      result := intval(line.dx > 0);
    exit;
  end;

  dx := x - line.v1.x;
  dy := y - line.v1.y;

  left := FixedMul(line.dy div FRACUNIT, dx);
  right := FixedMul(dy, line.dx div FRACUNIT);

  if right < left then
    result := 0  // front side
  else
    result := 1; // back side
end;

//
// P_BoxOnLineSide
// Considers the line to be infinite
// Returns side 0 or 1, -1 if box crosses the line.
//
function P_BoxOnLineSide(tmbox: Pfixed_tArray; ld: Pline_t): integer;
var
  p1: integer;
  p2: integer;
begin
  case ld.slopetype of
    ST_HORIZONTAL:
      begin
        p1 := intval(tmbox[BOXTOP] > ld.v1.y);
        p2 := intval(tmbox[BOXBOTTOM] > ld.v1.y);
        if ld.dx < 0 then
        begin
          p1 := p1 xor 1;
          p2 := p2 xor 1;
        end;
      end;
    ST_VERTICAL:
      begin
        p1 := intval(tmbox[BOXRIGHT] < ld.v1.x);
        p2 := intval(tmbox[BOXLEFT] < ld.v1.x);
        if ld.dy < 0 then
        begin
          p1 := p1 xor 1;
          p2 := p2 xor 1;
        end;
      end;
    ST_POSITIVE:
      begin
        p1 := P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXTOP], ld);
        p2 := P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXBOTTOM], ld);
      end;
    ST_NEGATIVE:
      begin
        p1 := P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXTOP], ld);
        p2 := P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXBOTTOM], ld);
      end;
  else
    begin
      p1 := 0;
      p2 := 0;
      I_Error('P_BoxOnLineSide(): wierd slopetype %d', [Ord(ld.slopetype)]);
    end;
  end;
  if p1 = p2 then
    result := p1
  else
    result := -1;
end;

//
// P_PointOnDivlineSide
// Returns 0 or 1.
//
function P_PointOnDivlineSide(x: fixed_t; y: fixed_t; line: Pdivline_t): integer;
var
  dx: fixed_t;
  dy: fixed_t;
  left: fixed_t;
  right: fixed_t;
begin
  if line.dx = 0 then
  begin
    if x <= line.x then
      result := intval(line.dy > 0)
    else
      result := intval(line.dy < 0);
    exit;
  end;

  if line.dy = 0 then
  begin
    if y <= line.y then
      result := intval(line.dx < 0)
    else
      result := intval(line.dx > 0);
    exit;
  end;

  dx := x - line.x;
  dy := y - line.y;

  // try to quickly decide by looking at sign bits
  if boolval((line.dy xor line.dx xor dx xor dy) and $80000000) then
  begin                                                              //(left is negative)
    result := intval(boolval((line.dy xor dx) and $80000000)); // VJ
    exit;
  end;

  left := FixedMul(_SHR(line.dy, 8), _SHR(dx, 8));
  right := FixedMul(_SHR(dy, 8), _SHR(line.dx, 8));

  if right < left then
    result := 0  // front side
  else
    result := 1; // back side
end;


//
// P_MakeDivline
//
procedure P_MakeDivline(li: Pline_t; dl: Pdivline_t);
begin
  dl.x := li.v1.x;
  dl.y := li.v1.y;
  dl.dx := li.dx;
  dl.dy := li.dy;
end;

//
// P_InterceptVector
// Returns the fractional intercept point
// along the first divline.
// This is only called by the addthings
// and addlines traversers.
//
function P_InterceptVector(v2: Pdivline_t; v1: Pdivline_t): fixed_t;
var
  num: fixed_t;
  den: fixed_t;
begin
  den := FixedMul(_SHR(v1.dy, 8), v2.dx) -
         FixedMul(_SHR(v1.dx, 8), v2.dy);

  if den = 0 then
    result := 0     // Parallel
  else
  begin
    num := FixedMul(_SHR((v1.x - v2.x), 8), v1.dy)	+
           FixedMul(_SHR((v2.y - v1.y), 8), v1.dx);
    result := FixedDiv(num, den);
  end;
end;

//
// P_LineOpening
// Sets opentop and openbottom to the window
// through a two sided line.
// OPTIMIZE: keep this precalculated
//
procedure P_LineOpening(linedef: Pline_t);
var
  front: Psector_t;
  back: Psector_t;
begin
  if linedef.sidenum[1] = -1 then
  begin
    // single sided line
    openrange := 0;
    exit;
  end;

  front := linedef.frontsector;
  back := linedef.backsector;

  if front.ceilingheight < back.ceilingheight then
    opentop := front.ceilingheight
  else
    opentop := back.ceilingheight;

  if front.floorheight > back.floorheight then
  begin
    openbottom := front.floorheight;
    lowfloor := back.floorheight;
  end
  else
  begin
    openbottom := back.floorheight;
    lowfloor := front.floorheight;
  end;

  openrange := opentop - openbottom;
end;

//
// THING POSITION SETTING
//


//
// P_UnsetThingPosition
// Unlinks a thing from block map and sectors.
// On each position change, BLOCKMAP and other
// lookups maintaining lists ot things inside
// these structures need to be updated.
//
procedure P_UnsetThingPosition(thing: Pmobj_t);
var
  blockx: integer;
  blocky: integer;
begin
  if (thing.flags and MF_NOSECTOR) = 0 then
  begin
    // inert things don't need to be in blockmap?
    // unlink from subsector
    if thing.snext <> nil then
      thing.snext.sprev := thing.sprev;

    if thing.sprev <> nil then
      thing.sprev.snext := thing.snext
    else
      Psubsector_t(thing.subsector).sector.thinglist := thing.snext;
  end;

  if (thing.flags and MF_NOBLOCKMAP) = 0 then
  begin
    // inert things don't need to be in blockmap
    // unlink from block map
    if thing.bnext <> nil then
      thing.bnext.bprev := thing.bprev;

    if thing.bprev <> nil then
      thing.bprev.bnext := thing.bnext
    else
    begin
      blockx := _SHR((thing.x - bmaporgx), MAPBLOCKSHIFT);
      blocky := _SHR((thing.y - bmaporgy), MAPBLOCKSHIFT);

      if (blockx >= 0) and (blockx < bmapwidth) and
         (blocky >= 0) and (blocky < bmapheight) then
        blocklinks[blocky * bmapwidth + blockx] := thing.bnext;
    end;
  end;
end;

//
// P_SetThingPosition
// Links a thing into both a block and a subsector
// based on it's x y.
// Sets thing->subsector properly
//
procedure P_SetThingPosition(thing: Pmobj_t);
var
  ss: Psubsector_t;
  sec: Psector_t;
  blockx: integer;
  blocky: integer;
  link: ^Pmobj_t;
begin
  // link into subsector
  ss := R_PointInSubsector(thing.x, thing.y);
  thing.subsector := ss;

  if (thing.flags and MF_NOSECTOR) = 0 then
  begin
    // invisible things don't go into the sector links
    sec := ss.sector;

    thing.sprev := nil;
    thing.snext := sec.thinglist;

    if sec.thinglist <> nil then
      sec.thinglist.sprev := thing;

    sec.thinglist := thing;
  end;

  // link into blockmap
  if (thing.flags and MF_NOBLOCKMAP) = 0 then
  begin
    // inert things don't need to be in blockmap
    blockx := _SHR((thing.x - bmaporgx), MAPBLOCKSHIFT);
    blocky := _SHR((thing.y - bmaporgy), MAPBLOCKSHIFT);

    if (blockx >= 0) and (blockx < bmapwidth) and
       (blocky >= 0) and (blocky < bmapheight) then
    begin
      link := @blocklinks[blocky * bmapwidth + blockx];
      thing.bprev := nil;
      thing.bnext := link^;
	    if boolval(link^) then
        (link^).bprev := thing;

      link^ := thing;
    end
    else
    begin
      // thing is off the map
      thing.bnext := nil;
      thing.bprev := nil;
    end;
  end;
end;

//
// BLOCK MAP ITERATORS
// For each line/thing in the given mapblock,
// call the passed PIT_* function.
// If the function returns false,
// exit with false without checking anything else.
//


//
// P_BlockLinesIterator
// The validcount flags are used to avoid checking lines
// that are marked in multiple mapblocks,
// so increment validcount before the first call
// to P_BlockLinesIterator, then make one or more calls
// to it.
//
function P_BlockLinesIterator(x, y: integer; func: ltraverser_t): boolean;
var
  offset: integer;
  ld: Pline_t;
begin
  if (x < 0) or (y < 0) or (x >= bmapwidth) or (y >= bmapheight) then
  begin
    result := true;
    exit;
  end;

  offset := blockmap[y * bmapwidth + x];
//  offset := PInteger(integer(blockmap) + y * bmapwidth + x)^;

  while blockmaplump[offset] <> - 1 do
  begin
    ld := @lines[blockmaplump[offset]];
    inc(offset);
    if ld.validcount = validcount then
      continue; // line has already been checked

    ld.validcount := validcount;

    if not func(ld) then
    begin
      result := false;
      exit;
    end;
  end;

  result := true; // everything was checked
end;

//
// P_BlockThingsIterator
//
function P_BlockThingsIterator(x, y: integer; func: ttraverser_t): boolean;
var
  mobj: Pmobj_t;
begin
  if (x < 0) or (y < 0) or (x >= bmapwidth) or (y >= bmapheight) then
  begin
    result := true;
    exit;
  end;

  mobj := blocklinks[y * bmapwidth + x];

  while mobj <> nil do
  begin
    if not func(mobj) then
    begin
      result := false;
      exit;
    end;
    mobj := mobj.bnext;
  end;

  result := true;
end;

//
// INTERCEPT ROUTINES
//
var
  intercepts: array[0..MAXINTERCEPTS - 1] of intercept_t;
  intercept_p: integer; 

  earlyout: boolean;

//
// PIT_AddLineIntercepts.
// Looks for lines in the given block
// that intercept the given trace
// to add to the intercepts list.
//
// A line is crossed if its endpoints
// are on opposite sides of the trace.
// Returns true if earlyout and a solid line hit.
//
function PIT_AddLineIntercepts(ld: Pline_t): boolean;
var
  s1: integer;
  s2: integer;
  frac: fixed_t;
  dl: divline_t;
begin
  // avoid precision problems with two routines
  if (trace.dx > FRACUNIT * 16) or (trace.dy > FRACUNIT * 16) or
     (trace.dx < -FRACUNIT * 16) or (trace.dy < -FRACUNIT * 16) then
  begin
    s1 := P_PointOnDivlineSide(ld.v1.x, ld.v1.y, @trace);
    s2 := P_PointOnDivlineSide(ld.v2.x, ld.v2.y, @trace);
  end
  else
  begin
    s1 := P_PointOnLineSide(trace.x, trace.y, ld);
    s2 := P_PointOnLineSide(trace.x + trace.dx, trace.y + trace.dy, ld);
  end;

  if s1 = s2 then
  begin
    result := true; // line isn't crossed
    exit;
  end;

  // hit the line
  P_MakeDivline(ld, @dl);
  frac := P_InterceptVector(@trace, @dl);

  if frac < 0 then
  begin
    result := true; // behind source
    exit;
  end;

  // try to early out the check
  if earlyout and (frac < FRACUNIT) and (not boolval(ld.backsector)) then
  begin
    result := false; // stop checking
    exit;
  end;

  intercepts[intercept_p].frac := frac;
  intercepts[intercept_p].isaline := true;
  intercepts[intercept_p].d.line := ld;
  inc(intercept_p);

  result := true; // continue
end;

//
// PIT_AddThingIntercepts
//
function PIT_AddThingIntercepts(thing: Pmobj_t): boolean;
var
  x1: fixed_t;
  y1: fixed_t;
  x2: fixed_t;
  y2: fixed_t;
  s1: integer;
  s2: integer;
  tracepositive: boolean;
  dl: divline_t;
  frac: fixed_t;
begin
  tracepositive := (trace.dx xor trace.dy) > 0;

  // check a corner to corner crossection for hit
  if tracepositive then
  begin
    x1 := thing.x - thing.radius;
    y1 := thing.y + thing.radius;
    x2 := thing.x + thing.radius;
    y2 := thing.y - thing.radius;
  end
  else
  begin
    x1 := thing.x - thing.radius;
    y1 := thing.y - thing.radius;
    x2 := thing.x + thing.radius;
    y2 := thing.y + thing.radius;
  end;

  s1 := P_PointOnDivlineSide(x1, y1, @trace);
  s2 := P_PointOnDivlineSide(x2, y2, @trace);

  if s1 = s2 then
  begin
    result := true; // line isn't crossed
    exit;
  end;

  dl.x := x1;
  dl.y := y1;
  dl.dx := x2 - x1;
  dl.dy := y2 - y1;

  frac := P_InterceptVector(@trace, @dl);

  if frac < 0 then
  begin
    result := true; // behind source
    exit;
  end;

  intercepts[intercept_p].frac := frac;
  intercepts[intercept_p].isaline := false;
  intercepts[intercept_p].d.thing := thing;
  inc(intercept_p);

  result := true; // keep going
end;

//
// P_TraverseIntercepts
// Returns true if the traverser function returns true
// for all lines.
//
function P_TraverseIntercepts(func: traverser_t; maxfrac: fixed_t): boolean;
var
  i: integer;
  dist: fixed_t;
  scan: integer;
  _in: Pintercept_t;
begin
  _in := nil; // shut up compiler warning

  for i := 0 to intercept_p - 1 do
  begin
    dist := MAXINT;

    for scan := 0 to intercept_p - 1 do
    begin
      if intercepts[scan].frac < dist then
      begin
    		dist := intercepts[scan].frac;
		    _in := @intercepts[scan];
      end;
    end;

    if dist > maxfrac then
    begin
      result := true; // checked everything in range
      exit;
    end;

    if not func(_in) then
    begin
      result := false; // don't bother going farther
      exit;
    end;
    _in.frac := MAXINT;
  end;

  result := true;	// everything was traversed
end;

//
// P_PathTraverse
// Traces a line from x1,y1 to x2,y2,
// calling the traverser function for each.
// Returns true if the traverser function returns true
// for all lines.
//
function P_PathTraverse(x1, y1, x2, y2: fixed_t; flags: integer;
  trav: traverser_t): boolean;
var
  xt1: fixed_t;
  yt1: fixed_t;
  xt2: fixed_t;
  yt2: fixed_t;
  xstep: fixed_t;
  ystep: fixed_t;
  partial: fixed_t;
  xintercept: fixed_t;
  yintercept: fixed_t;
  mapx: integer;
  mapy: integer;
  mapxstep: integer;
  mapystep: integer;
  count: integer;
begin		
  earlyout := boolval(flags and PT_EARLYOUT);

  inc(validcount);
  intercept_p := 0;

  if (x1 - bmaporgx) and (MAPBLOCKSIZE - 1) = 0 then
    x1 := x1 + FRACUNIT; // don't side exactly on a line

  if (y1 - bmaporgy) and (MAPBLOCKSIZE - 1) = 0 then
    y1 := y1 + FRACUNIT; // don't side exactly on a line

  trace.x := x1;
  trace.y := y1;
  trace.dx := x2 - x1;
  trace.dy := y2 - y1;

  x1 := x1 - bmaporgx;
  y1 := y1 - bmaporgy;
  xt1 := _SHR(x1, MAPBLOCKSHIFT);
  yt1 := _SHR(y1, MAPBLOCKSHIFT);

  x2 := x2 - bmaporgx;
  y2 := y2 - bmaporgy;
  xt2 := _SHR(x2, MAPBLOCKSHIFT);
  yt2 := _SHR(y2, MAPBLOCKSHIFT);

  if xt2 > xt1 then
  begin
    mapxstep := 1;
    partial := FRACUNIT - (_SHR(x1, MAPBTOFRAC) and (FRACUNIT - 1));
    ystep := FixedDiv(y2 - y1, abs(x2 - x1));
  end
  else if xt2 < xt1 then
  begin
    mapxstep := -1;
    partial := _SHR(x1, MAPBTOFRAC) and (FRACUNIT - 1);
    ystep := FixedDiv(y2 - y1, abs(x2 - x1));
  end
  else
  begin
    mapxstep := 0;
    partial := FRACUNIT;
    ystep := 256 * FRACUNIT;
  end;

  yintercept := _SHR(y1, MAPBTOFRAC) + FixedMul(partial, ystep);

  if yt2 > yt1 then
  begin
    mapystep := 1;
    partial := FRACUNIT - (_SHR(y1, MAPBTOFRAC) and (FRACUNIT - 1));
    xstep := FixedDiv(x2 - x1, abs(y2 - y1));
  end
  else if yt2 < yt1 then
  begin
    mapystep := -1;
    partial := _SHR(y1, MAPBTOFRAC) and (FRACUNIT - 1);
    xstep := FixedDiv(x2 - x1, abs(y2 - y1));
  end
  else
  begin
    mapystep := 0;
    partial := FRACUNIT;
    xstep := 256 * FRACUNIT;
  end;

  xintercept := _SHR(x1, MAPBTOFRAC) + FixedMul(partial, xstep);

  // Step through map blocks.
  // Count is present to prevent a round off error
  // from skipping the break.
  mapx := xt1;
  mapy := yt1;

  for count := 0 to 63 do
  begin
    if boolval(flags and PT_ADDLINES) then
    begin
      if not P_BlockLinesIterator(mapx, mapy, PIT_AddLineIntercepts) then
      begin
        result := false; // early out
        exit;
      end;
    end;

    if boolval(flags and PT_ADDTHINGS) then
    begin
	    if not P_BlockThingsIterator(mapx, mapy, PIT_AddThingIntercepts) then
      begin
        result := false;// early out
        exit;
      end;
    end;

    if (mapx = xt2) and (mapy = yt2) then
      break;

    if yintercept div FRACUNIT = mapy then
    begin
      yintercept := yintercept + ystep;
      mapx := mapx + mapxstep;
    end
    else if xintercept div FRACUNIT = mapx then
    begin
      xintercept := xintercept + xstep;
      mapy := mapy + mapystep;
    end;
  end;

  // go through the sorted list
  result := P_TraverseIntercepts(trav, FRACUNIT );
end;


(* Unit p_setup implementation *)


                                                                                                                                                             


//
// P_LoadVertexes
//
procedure P_LoadVertexes(lump: integer);
var
  data: pointer;
  i: integer;
  ml: Pmapvertex_t;
  li: Pvertex_t;
begin
  // Determine number of lumps:
  //  total lump length / vertex record length.
  numvertexes := W_LumpLength(lump) div SizeOf(mapvertex_t);

  // Allocate zone memory for buffer.
  vertexes := Z_Malloc(numvertexes * SizeOf(vertex_t), PU_LEVEL, nil);

  // Load data into cache.
  data := W_CacheLumpNum(lump, PU_STATIC);

  ml := Pmapvertex_t(data);

  // Copy and convert vertex coordinates,
  // internal representation as fixed.
  for i := 0 to numvertexes - 1 do
  begin
    li := @vertexes[i];
    li.x := ml.x * FRACUNIT;
    li.y := ml.y * FRACUNIT;
    inc(ml);
  end;

  // Free buffer memory.
  Z_Free(data);
end;

//
// P_LoadSegs
//
procedure P_LoadSegs(lump: integer);
var
  data: pointer;
  i: integer;
  ml: Pmapseg_t;
  li: Pseg_t;
  ldef: Pline_t;
  linedef: integer;
  side: integer;
begin
  numsegs := W_LumpLength(lump) div SizeOf(mapseg_t);
  segs := Z_Malloc(numsegs * SizeOf(seg_t), PU_LEVEL, nil);
  memset(segs, 0, numsegs * SizeOf(seg_t));
  data := W_CacheLumpNum(lump, PU_STATIC);

  ml := Pmapseg_t(data);
  for i := 0 to numsegs - 1 do
  begin
    li := @segs[i];
    li.v1 := @vertexes[ml.v1];
    li.v2 := @vertexes[ml.v2];

    li.angle := _SHL(ml.angle, 16);
    li.offset := _SHL(ml.offset, 16);
    linedef := ml.linedef;
    ldef := @lines[linedef];
    li.linedef := ldef;
    side := ml.side;
    li.sidedef := @sides[ldef.sidenum[side]];
    li.frontsector := sides[ldef.sidenum[side]].sector;
    if boolval(ldef.flags and ML_TWOSIDED) then
      li.backsector := sides[ldef.sidenum[side xor 1]].sector
    else
      li.backsector := nil;
    inc(ml);
  end;

  Z_Free(data);
end;

//
// P_LoadSubsectors
//
procedure P_LoadSubsectors(lump: integer);
var
  data: pointer;
  i: integer;
  ms: Pmapsubsector_t;
  ss: Psubsector_t;
begin
  numsubsectors := W_LumpLength(lump) div SizeOf(mapsubsector_t);
  subsectors := Z_Malloc(numsubsectors * SizeOf(subsector_t), PU_LEVEL, nil);
  data := W_CacheLumpNum(lump, PU_STATIC);

  ms := Pmapsubsector_t(data);
  memset(subsectors, 0, numsubsectors * SizeOf(subsector_t));

  for i := 0 to numsubsectors - 1 do //; i++, ss++, ms++)
  begin
    ss := @subsectors[i];
    ss.numlines := ms.numsegs;
    ss.firstline := ms.firstseg;
    inc(ms);
  end;

  Z_Free(data);
end;

//
// P_LoadSectors
//
procedure P_LoadSectors(lump: integer);
var
  data: pointer;
  i: integer;
  ms: Pmapsector_t;
  ss: Psector_t;
begin
  numsectors := W_LumpLength(lump) div SizeOf(mapsector_t);
  sectors := Z_Malloc(numsectors * SizeOf(sector_t), PU_LEVEL, nil);
  memset(sectors, 0, numsectors * SizeOf(sector_t));
  data := W_CacheLumpNum(lump, PU_STATIC);

  ms := Pmapsector_t(data);
  for i := 0 to numsectors - 1 do //; i++, ss++, ms++)
  begin
    ss := @sectors[i];
    ss.floorheight := ms.floorheight * FRACUNIT;
    ss.ceilingheight := ms.ceilingheight * FRACUNIT;
    ss.floorpic := R_FlatNumForName(ms.floorpic);
    ss.ceilingpic := R_FlatNumForName(ms.ceilingpic);
    ss.lightlevel := ms.lightlevel;
    ss.special := ms.special;
    ss.tag := ms.tag;
    ss.thinglist := nil;
    inc(ms);
  end;

  Z_Free (data);
end;

//
// P_LoadNodes
//
procedure P_LoadNodes(lump: integer);
var
  data: pointer;
  i: integer;
  j: integer;
  k: integer;
  mn: Pmapnode_t;
  no: Pnode_t;
begin
  numnodes := W_LumpLength(lump) div SizeOf(mapnode_t);
  nodes := Z_Malloc(numnodes * SizeOf(node_t), PU_LEVEL, nil);
  data := W_CacheLumpNum(lump, PU_STATIC);

  mn := Pmapnode_t(data);

  for i := 0 to numnodes - 1 do //; i++, no++, mn++)
  begin
    no := @nodes[i];
    no.x := mn.x * FRACUNIT;
    no.y := mn.y * FRACUNIT;
    no.dx := mn.dx * FRACUNIT;
    no.dy := mn.dy * FRACUNIT;
    for j := 0 to 1 do
    begin
      no.children[j] := mn.children[j];
      for k := 0 to 3 do
        no.bbox[j, k] := SmallInt(mn.bbox[j, k]) * FRACUNIT;
    end;
    inc(mn);
  end;

  Z_Free (data);
end;

//
// P_LoadThings
//
procedure P_LoadThings(lump: integer);
var
  data: pointer;
  i: integer;
  mt: Pmapthing_t;
  numthings: integer;
  spawn: boolean;
begin
  data := W_CacheLumpNum(lump, PU_STATIC);
  numthings := W_LumpLength(lump) div SizeOf(mapthing_t);

  mt := Pmapthing_t(data);
  for i := 0 to numthings - 1 do
  begin
    spawn := true;
    // Do not spawn cool, new monsters if !commercial
    if gamemode <> commercial then
    begin
      case mt._type of
        68, // Arachnotron
        64, // Archvile
        88, // Boss Brain
        89, // Boss Shooter
        69, // Hell Knight
        67, // Mancubus
        71, // Pain Elemental
        65, // Former Human Commando
        66, // Revenant
        84: // Wolf SS
          spawn := false;
      end;
    end;

    if spawn then
    begin
      // Do spawn all other stuff.
      mt.x := SmallInt(mt.x);
      mt.y := SmallInt(mt.y);
      mt.angle := SmallInt(mt.angle);
      mt._type := SmallInt(mt._type);
      mt.options := SmallInt(mt.options);

      P_SpawnMapThing(mt);
    end;
    inc(mt);
  end;

  Z_Free(data);
end;

//
// P_LoadLineDefs
// Also counts secret lines for intermissions.
//
procedure P_LoadLineDefs(lump: integer);
var
  data: pointer;
  i: integer;
  mld: Pmaplinedef_t;
  ld: Pline_t;
  v1: Pvertex_t;
  v2: Pvertex_t;
begin
  numlines := W_LumpLength(lump) div SizeOf(maplinedef_t);
  lines := Z_Malloc(numlines * SizeOf(line_t), PU_LEVEL, nil);
  memset(lines, 0, numlines * SizeOf(line_t));
  data := W_CacheLumpNum(lump, PU_STATIC);

  mld := Pmaplinedef_t(data);
  for i := 0 to numlines - 1 do
  begin
    ld := @lines[i];
    ld.flags := SmallInt(mld.flags);
    ld.special := SmallInt(mld.special);
    ld.tag := SmallInt(mld.tag);
    ld.v1 := @vertexes[SmallInt(mld.v1)];
    v1 := ld.v1;
    ld.v2 := @vertexes[SmallInt(mld.v2)];
    v2 := ld.v2;
    ld.dx := v2.x - v1.x;
    ld.dy := v2.y - v1.y;

    if not boolval(ld.dx) then
      ld.slopetype := ST_VERTICAL
    else if not boolval(ld.dy) then
      ld.slopetype := ST_HORIZONTAL
    else
    begin
      if FixedDiv(ld.dy , ld.dx) > 0 then
        ld.slopetype := ST_POSITIVE
      else
        ld.slopetype := ST_NEGATIVE;
    end;

    if v1.x < v2.x then
    begin
      ld.bbox[BOXLEFT] := v1.x;
      ld.bbox[BOXRIGHT] := v2.x;
    end
    else
    begin
      ld.bbox[BOXLEFT] := v2.x;
      ld.bbox[BOXRIGHT] := v1.x;
    end;

    if v1.y < v2.y then
    begin
      ld.bbox[BOXBOTTOM] := v1.y;
      ld.bbox[BOXTOP] := v2.y;
    end
    else
    begin
      ld.bbox[BOXBOTTOM] := v2.y;
      ld.bbox[BOXTOP] := v1.y;
    end;

    ld.sidenum[0] := SmallInt(mld.sidenum[0]);
    ld.sidenum[1] := SmallInt(mld.sidenum[1]);

    if ld.sidenum[0] <> -1 then
      ld.frontsector := sides[ld.sidenum[0]].sector
    else
      ld.frontsector := nil;

    if ld.sidenum[1] <> -1 then
      ld.backsector := sides[ld.sidenum[1]].sector
    else
      ld.backsector := nil;

    inc(mld);
  end;

  Z_Free (data);
end;

//
// P_LoadSideDefs
//
procedure P_LoadSideDefs(lump: integer);
var
  data: pointer;
  i: integer;
  msd: Pmapsidedef_t;
  sd: Pside_t;
begin
  numsides := W_LumpLength(lump) div SizeOf(mapsidedef_t);
  sides := Z_Malloc(numsides * SizeOf(side_t), PU_LEVEL, nil);
  memset(sides, 0, numsides * SizeOf(side_t));
  data := W_CacheLumpNum(lump, PU_STATIC);

  msd := Pmapsidedef_t(data);
  for i := 0 to numsides - 1 do
  begin
    sd := @sides[i];
    sd.textureoffset := SmallInt(msd.textureoffset) * FRACUNIT;
    sd.rowoffset := SmallInt(msd.rowoffset) * FRACUNIT;
    sd.toptexture := R_TextureNumForName(msd.toptexture);
    sd.bottomtexture := R_TextureNumForName(msd.bottomtexture);
    sd.midtexture := R_TextureNumForName(msd.midtexture);
    sd.sector := @sectors[SmallInt(msd.sector)];
    inc(msd);
  end;

  Z_Free (data);
end;

//
// P_LoadBlockMap
//
procedure P_LoadBlockMap(lump: integer);
var
  count: integer;
begin
  blockmaplump := W_CacheLumpNum(lump, PU_LEVEL);
//  blockmap := PSmallIntArray(integer(blockmaplump) + 4);
  blockmap := @blockmaplump[4];

  bmaporgx := blockmaplump[0] * FRACUNIT;
  bmaporgy := blockmaplump[1] * FRACUNIT;
  bmapwidth := blockmaplump[2];
  bmapheight := blockmaplump[3];

  // clear out mobj chains
  count := SizeOf(Pmobj_t) * bmapwidth * bmapheight;
  blocklinks := Z_Malloc(count, PU_LEVEL, nil);
  memset(blocklinks, 0, count);
end;

//
// P_GroupLines
// Builds sector line lists and subsector sector numbers.
// Finds block bounding boxes for sectors.
//
procedure P_GroupLines;
var
  linebuffer: Pline_tPArray; // pointer to an array of pointers Pline_t
  i: integer;
  j: integer;
  total: integer;
  li: Pline_t;
  sector: Psector_t;
  seg: Pseg_t;
  bbox: array[0..3] of fixed_t;
  block: integer;
begin
  // look up sector number for each subsector
  for i := 0 to numsubsectors - 1 do
  begin
    seg := @segs[subsectors[i].firstline];
    subsectors[i].sector := seg.sidedef.sector;
  end;

  // count number of lines in each sector
  total := 0;
  for i := 0 to numlines - 1 do
  begin
    li := @lines[i];
    inc(total);
    li.frontsector.linecount := li.frontsector.linecount + 1;

    if boolval(li.backsector) and (li.backsector <> li.frontsector) then
    begin
      li.backsector.linecount := li.backsector.linecount + 1;
      inc(total);
    end;
  end;

  // build line tables for each sector
  linebuffer := Z_Malloc(total * 4, PU_LEVEL, nil);
  for i := 0 to numsectors - 1 do
  begin
    sector := @sectors[i];
    M_ClearBox(@bbox);
    sector.lines := linebuffer;
    for j := 0 to numlines - 1 do
    begin
      li := @lines[j];
	    if (li.frontsector = sector) or (li.backsector = sector) then
      begin
  		  linebuffer[0] := li;
        incp(pointer(linebuffer), SizeOf(Pointer));
        M_AddToBox(@bbox, li.v1.x, li.v1.y);
        M_AddToBox(@bbox, li.v2.x, li.v2.y);
      end;
    end;
    if pOperation(linebuffer, sector.lines, '-', SizeOf(pointer)) <> sector.linecount then
      I_Error('P_GroupLines(): miscounted'); // VJ ?????

    // set the degenmobj_t to the middle of the bounding box
    sector.soundorg.x := (bbox[BOXRIGHT] + bbox[BOXLEFT]) div 2;
    sector.soundorg.y := (bbox[BOXTOP] + bbox[BOXBOTTOM]) div 2;

    // adjust bounding box to map blocks
    block := _SHR(bbox[BOXTOP] - bmaporgy + MAXRADIUS, MAPBLOCKSHIFT);
    if block >= bmapheight then
      block  := bmapheight - 1;
    sector.blockbox[BOXTOP] := block;

    block := _SHR(bbox[BOXBOTTOM] - bmaporgy - MAXRADIUS, MAPBLOCKSHIFT);
    if block < 0 then
      block  := 0;
    sector.blockbox[BOXBOTTOM] := block;

    block := _SHR(bbox[BOXRIGHT] - bmaporgx + MAXRADIUS, MAPBLOCKSHIFT);
    if block >= bmapwidth then
      block := bmapwidth - 1;
    sector.blockbox[BOXRIGHT] := block;

    block := _SHR(bbox[BOXLEFT] - bmaporgx - MAXRADIUS, MAPBLOCKSHIFT);
    if block < 0 then
      block := 0;
    sector.blockbox[BOXLEFT] := block;
  end;
end;

//
// P_SetupLevel
//
procedure P_SetupLevel(episode, map, playermask: integer; skill: skill_t);
var
  i: integer;
  lumpname: string;
  lumpnum: integer;
begin
  totalkills := 0;
  totalitems := 0;
  totalsecret := 0;

  wminfo.maxfrags := 0;
  wminfo.partime := 180;
  for i := 0 to MAXPLAYERS - 1 do
  begin
	  players[i].killcount := 0;
    players[i].secretcount := 0;
    players[i].itemcount := 0;
  end;

  // Initial height of PointOfView
  // will be set by player think.
  players[consoleplayer].viewz := 1;

  // Make sure all sounds are stopped before Z_FreeTags.
  S_Start;

(*
#if 0 // UNUSED
    if (debugfile)
    {
	Z_FreeTags (PU_LEVEL, MAXINT);
	Z_FileDumpHeap (debugfile);
    }
    else
#endif
*)

  Z_FreeTags(PU_LEVEL, PU_PURGELEVEL - 1);

  R_SetupLevel;

  // UNUSED W_Profile ();
  P_InitThinkers;

  // if working with a devlopment map, reload it
  W_Reload;

  // find map name
  if gamemode = commercial then
  begin
    if map < 10 then
      sprintf(lumpname,'map0%d', [map])
    else
      sprintf(lumpname,'map%d', [map]);
  end
  else
    sprintf(lumpname, 'E%dM%d', [episode, map]);

  lumpnum := W_GetNumForName(lumpname);

  leveltime := 0;

  // note: most of this ordering is important
  P_LoadBlockMap(lumpnum + Ord(ML_BLOCKMAP));
  P_LoadVertexes(lumpnum + Ord(ML_VERTEXES));
  P_LoadSectors(lumpnum + Ord(ML_SECTORS));
  P_LoadSideDefs(lumpnum + Ord(ML_SIDEDEFS));

  P_LoadLineDefs(lumpnum + Ord(ML_LINEDEFS));
  P_LoadSubsectors(lumpnum + Ord(ML_SSECTORS));
  P_LoadNodes(lumpnum + Ord(ML_NODES));
  P_LoadSegs(lumpnum + Ord(ML_SEGS));

  rejectmatrix := W_CacheLumpNum(lumpnum + Ord(ML_REJECT), PU_LEVEL);
  P_GroupLines;

  bodyqueslot := 0;
  deathmatch_p := 0;
  P_LoadThings(lumpnum + Ord(ML_THINGS));

  // if deathmatch, randomly spawn the active players
  if boolval(deathmatch) then
  begin
    for i := 0 to MAXPLAYERS - 1 do
      if playeringame[i] then
      begin
        players[i].mo := nil;
        G_DeathMatchSpawnPlayer(i);
      end;
  end;

  // clear special respawning que
  iquehead := 0;
  iquetail := 0;

  // set up world state
  P_SpawnSpecials;

  // build subsector connect matrix
  // UNUSED P_ConnectSubsectors ();

  // preload graphics
  if precache then
    R_PrecacheLevel;

    //printf ("free memory: 0x%x\n", Z_FreeMemory());
end;

//
// P_Init
//
procedure P_Init;
begin
  P_InitSwitchList;
  P_InitPicAnims;
  R_InitSprites(sprnames);
end;


(* Unit p_mobj implementation *)


                                                                                                                                                                                                                                                               

//
// P_SetMobjState
// Returns true if the mobj is still present.
//
function P_SetMobjState(mobj: Pmobj_t; state: statenum_t): boolean;
var
  st: Pstate_t;
begin
  repeat
    if state = S_NULL then
    begin
      mobj.state := nil; // VJ was mobj.state := @S_NULL;
	    P_RemoveMobj(mobj);
      result := false;
      exit;
    end;

    st := @states[Ord(state)];
    mobj.state := st;
    mobj.tics := st.tics;
    mobj.sprite := st.sprite;
    mobj.frame := st.frame;

    // Modified handling.
    // Call action functions when the state is set
    if Assigned(st.action.acp1) then
      st.action.acp1(mobj);

    state := st.nextstate;
  until (mobj.tics <> 0);

  result := true;
end;

//
// P_ExplodeMissile
//
procedure P_ExplodeMissile(mo: Pmobj_t);
begin
  mo.momx := 0;
  mo.momy := 0;
  mo.momz := 0;

  P_SetMobjState(mo, statenum_t(mobjinfo[Ord(mo._type)].deathstate));

  mo.tics := mo.tics - (P_Random and 3);

  if mo.tics < 1 then
    mo.tics := 1;

  mo.flags := mo.flags and (not MF_MISSILE);

  if mo.info.deathsound <> 0 then
    S_StartSound(mo, mo.info.deathsound);
end;

//
// P_XYMovement  
//
const
  STOPSPEED = $1000;
  FRICTION = $e800;

procedure P_XYMovement(mo: Pmobj_t);
var
  ptryx: fixed_t;
  ptryy: fixed_t;
  player: Pplayer_t;
  xmove: fixed_t;
  ymove: fixed_t;
begin
  if (mo.momx = 0) and (mo.momy = 0) then
  begin
    if (mo.flags and MF_SKULLFLY) <> 0 then
    begin
      // the skull slammed into something
      mo.flags := mo.flags and (not MF_SKULLFLY);
      mo.momx := 0;
      mo.momy := 0;
      mo.momz := 0;

      P_SetMobjState(mo, statenum_t(mo.info.spawnstate));
    end;
    exit;
  end;

  player := mo.player;

  if mo.momx > MAXMOVE then
    mo.momx := MAXMOVE
  else if mo.momx < -MAXMOVE then
    mo.momx := -MAXMOVE;

  if mo.momy > MAXMOVE then
    mo.momy := MAXMOVE
  else if mo.momy < -MAXMOVE then
    mo.momy := -MAXMOVE;

  xmove := mo.momx;
  ymove := mo.momy;

  repeat
    if (xmove > MAXMOVE div 2) or (ymove > MAXMOVE div 2) then
    begin
      ptryx := mo.x + xmove div 2;
	    ptryy := mo.y + ymove div 2;
	    xmove := _SHR(xmove, 1);
	    ymove := _SHR(ymove, 1);
    end
    else
    begin
	    ptryx := mo.x + xmove;
	    ptryy := mo.y + ymove;
	    xmove := 0;
      ymove := 0;
    end;

    if not P_TryMove(mo, ptryx, ptryy) then
    begin
      // blocked move
      if mo.player <> nil then
      begin // try to slide along it
        P_SlideMove(mo);
      end
      else if (mo.flags and MF_MISSILE) <> 0 then
      begin
        // explode a missile
        if (ceilingline <> nil) and
           boolval(ceilingline.backsector) and
           (ceilingline.backsector.ceilingpic = skyflatnum) then
        begin
          // Hack to prevent missiles exploding
          // against the sky.
          // Does not handle sky floors.
		      P_RemoveMobj(mo);
          exit;
        end;
        P_ExplodeMissile(mo);
      end
	    else
      begin
        mo.momx := 0;
        mo.momy := 0;
      end;
    end;
  until not ((xmove <> 0) or (ymove <> 0));

  // slow down
  if boolval(player) and  ((player.cheats and CF_NOMOMENTUM) <> 0) then
  begin
    // debug option for no sliding at all
    mo.momx := 0;
    mo.momy := 0;
    exit;
  end;

  if (mo.flags and (MF_MISSILE or MF_SKULLFLY)) <> 0 then
    exit; // no friction for missiles ever

  if mo.z > mo.floorz then
    exit; // no friction when airborne

  if (mo.flags and MF_CORPSE) <> 0 then
  begin
    // do not stop sliding
    //  if halfway off a step with some momentum
    if (mo.momx > FRACUNIT div 4) or
       (mo.momx < -FRACUNIT div 4) or
       (mo.momy > FRACUNIT div 4) or
       (mo.momy < -FRACUNIT div 4) then
    begin
      if mo.floorz <> Psubsector_t(mo.subsector).sector.floorheight then
        exit;
    end;
  end;

  if (mo.momx > -STOPSPEED) and
     (mo.momx < STOPSPEED) and
     (mo.momy > -STOPSPEED) and
     (mo.momy < STOPSPEED) and
     ((not boolval(player)) or
      ((player.cmd.forwardmove = 0) and
       (player.cmd.sidemove = 0))) then
  begin
    // if in a walking frame, stop moving
    if boolval(player) and
      (LongWord((pOperation(player.mo.state, @states[0], '-', SizeOf(states[0]))) - Ord(S_PLAY_RUN1)) < 4) then
      P_SetMobjState(player.mo, S_PLAY);

    mo.momx := 0;
    mo.momy := 0;
  end
  else
  begin
    mo.momx := FixedMul(mo.momx, FRICTION);
    mo.momy := FixedMul(mo.momy, FRICTION);
  end;
end;

//
// P_ZMovement
//
procedure P_ZMovement(mo: Pmobj_t);
var
  dist: fixed_t;
  delta: fixed_t;
begin
  // check for smooth step up
  if boolval(mo.player) and (mo.z < mo.floorz) then
  begin
    Pplayer_t(mo.player).viewheight := Pplayer_t(mo.player).viewheight - (mo.floorz - mo.z);
    Pplayer_t(mo.player).deltaviewheight :=
      _SHR((PVIEWHEIGHT - Pplayer_t(mo.player).viewheight), 3);
  end;

  // adjust height
  mo.z := mo.z + mo.momz;

  if ((mo.flags and MF_FLOAT) <> 0) and boolval(mo.target) then
  begin
    // float down towards target if too close
    if ((mo.flags and MF_SKULLFLY) = 0) and
       ((mo.flags and MF_INFLOAT) = 0) then
    begin
      dist := P_AproxDistance(mo.x - mo.target.x, mo.y - mo.target.y);

      delta := (mo.target.z + (_SHR(mo.height, 1))) - mo.z; // VJ is it right ???

      if (delta < 0) and (dist < -(delta * 3)) then
        mo.z := mo.z - FLOATSPEED
      else if (delta > 0) and (dist < (delta * 3)) then
        mo.z := mo.z + FLOATSPEED;
    end;
  end;

  // clip movement
  if mo.z <= mo.floorz then
  begin
    // hit the floor

    // Note (id):
    //  somebody left this after the setting momz to 0,
    //  kinda useless there.
    if boolval(mo.flags and MF_SKULLFLY) then
    begin
      // the skull slammed into something
      mo.momz := -mo.momz;
    end;

    if mo.momz < 0 then
    begin
      if boolval(mo.player) and (mo.momz < -GRAVITY * 8) then
      begin
        // Squat down.
        // Decrease viewheight for a moment
        // after hitting the ground (hard),
        // and utter appropriate sound.
        Pplayer_t(mo.player).deltaviewheight := _SHR(mo.momz, 3);
        S_StartSound (mo, Ord(sfx_oof));
      end;
      mo.momz := 0;
    end;
    mo.z := mo.floorz;

    if boolval(mo.flags and MF_MISSILE) and (not boolval(mo.flags and MF_NOCLIP)) then
    begin
      P_ExplodeMissile(mo);
      exit;
    end;
  end
  else if (mo.flags and MF_NOGRAVITY) = 0 then
  begin
    if mo.momz = 0 then
      mo.momz := - GRAVITY * 2
    else
      mo.momz := mo.momz - GRAVITY;
  end;

  if mo.z + mo.height > mo.ceilingz then
  begin
    // hit the ceiling
    if mo.momz > 0 then
      mo.momz := 0;

    mo.z := mo.ceilingz - mo.height;

    if boolval(mo.flags and MF_SKULLFLY) then
      mo.momz := -mo.momz; // the skull slammed into something

    if boolval(mo.flags and MF_MISSILE) and (not boolval(mo.flags and MF_NOCLIP)) then
      P_ExplodeMissile(mo);
  end;
end;

//
// P_NightmareRespawn
//
procedure P_NightmareRespawn(mobj: Pmobj_t);
var
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
  ss: Psubsector_t;
  mo: Pmobj_t;
  mthing: Pmapthing_t;
begin
  x := mobj.spawnpoint.x * FRACUNIT;
  y := mobj.spawnpoint.y * FRACUNIT;

  // somthing is occupying it's position?
  if not P_CheckPosition(mobj, x, y) then
    exit; // no respwan

  // spawn a teleport fog at old spot
  // because of removal of the body?
  mo := P_SpawnMobj(mobj.x, mobj.y, Psubsector_t(mobj.subsector).sector.floorheight, MT_TFOG);
  // initiate teleport sound
  S_StartSound(mo, Ord(sfx_telept));

  // spawn a teleport fog at the new spot
  ss := R_PointInSubsector(x, y);

  mo := P_SpawnMobj(x, y, ss.sector.floorheight, MT_TFOG);

  S_StartSound (mo, Ord(sfx_telept));

  // spawn the new monster
  mthing := @mobj.spawnpoint;

  // spawn it
  if boolval(mobj.info.flags and MF_SPAWNCEILING) then
    z := ONCEILINGZ
  else
    z := ONFLOORZ;

  // inherit attributes from deceased one
  mo := P_SpawnMobj(x, y, z, mobj._type);
  mo.spawnpoint := mobj.spawnpoint;
  mo.angle := ANG45 * (mthing.angle div 45);

  if (mthing.options and MTF_AMBUSH) <> 0 then
    mo.flags := mo.flags or MF_AMBUSH;

  mo.reactiontime := 18;

  // remove the old monster,
  P_RemoveMobj(mobj);
end;

//
// P_MobjThinker
//
procedure P_MobjThinker(mobj: Pmobj_t);
begin
  // momentum movement
  if boolval(mobj.momx) or
     boolval(mobj.momy) or
     boolval(mobj.flags and MF_SKULLFLY) then
  begin
    P_XYMovement(mobj);

    if not Assigned(mobj.thinker._function.acv) then
      exit; // mobj was removed
  end;

  if (mobj.z <> mobj.floorz) or boolval(mobj.momz) then
  begin
    P_ZMovement(mobj);

    if not Assigned(mobj.thinker._function.acv) then
      exit; // mobj was removed
  end;


  // cycle through states,
  // calling action functions at transitions
  if mobj.tics <> -1 then
  begin
    mobj.tics := mobj.tics - 1;

    // you can cycle through multiple states in a tic
    if not boolval(mobj.tics) then
      if not P_SetMobjState(mobj, mobj.state.nextstate) then
        exit; // freed itself
  end
  else
  begin
    // check for nightmare respawn
    if not boolval(mobj.flags and MF_COUNTKILL) then
      exit;

    if not respawnmonsters then
      exit;

    mobj.movecount := mobj.movecount + 1;

    if mobj.movecount < 12 * 35 then // ??? VJ 12 * TICKRATE
      exit;

    if boolval(leveltime and 31) then
      exit;

    if P_Random > 4 then
      exit;

    P_NightmareRespawn(mobj);
  end;
end;

//
// P_SpawnMobj
//
function P_SpawnMobj(x, y, z: fixed_t; _type: mobjtype_t): Pmobj_t;
var
  mobj: Pmobj_t;
  st: Pstate_t;
  info: Pmobjinfo_t;
begin
  mobj := Z_Malloc(SizeOf(mobj^), PU_LEVEL, nil);
  memset(mobj, 0, SizeOf(mobj^));
  info := @mobjinfo[Ord(_type)];

  mobj._type := _type;
  mobj.info := info;
  mobj.x := x;
  mobj.y := y;
  mobj.radius := info.radius;
  mobj.height := info.height;
  mobj.flags := info.flags;
  mobj.health := info.spawnhealth;

  if gameskill <> sk_nightmare then
    mobj.reactiontime := info.reactiontime;

  mobj.lastlook := P_Random mod MAXPLAYERS;
  // do not set the state with P_SetMobjState,
  // because action routines can not be called yet
  st := @states[info.spawnstate];

  mobj.state := st;
  mobj.tics := st.tics;
  mobj.sprite := st.sprite;
  mobj.frame := st.frame;

  // set subsector and/or block links
  P_SetThingPosition(mobj);

  mobj.floorz := Psubsector_t(mobj.subsector).sector.floorheight;
  mobj.ceilingz := Psubsector_t(mobj.subsector).sector.ceilingheight;

  if z = ONFLOORZ then
    mobj.z := mobj.floorz
  else if z = ONCEILINGZ then
    mobj.z := mobj.ceilingz - mobj.info.height
  else
    mobj.z := z;

  @mobj.thinker._function.acp1 := @P_MobjThinker;

  P_AddThinker(@mobj.thinker);

  result := mobj;
end;

//
// P_RemoveMobj
//
var
  itemrespawnque: array[0..ITEMQUESIZE - 1] of mapthing_t;
  itemrespawntime: array[0..ITEMQUESIZE - 1] of integer;

procedure P_RemoveMobj(mobj: Pmobj_t);
begin
  if boolval(mobj.flags and MF_SPECIAL) and
     (not boolval(mobj.flags and MF_DROPPED)) and
     (mobj._type <> MT_INV) and
     (mobj._type <> MT_INS) then
  begin
    itemrespawnque[iquehead] := mobj.spawnpoint;
    itemrespawntime[iquehead] := leveltime;
    iquehead := (iquehead + 1) and (ITEMQUESIZE - 1);

    // lose one off the end?
    if iquehead = iquetail then
      iquetail := (iquetail + 1) and (ITEMQUESIZE - 1);
  end;

  // unlink from sector and block lists
  P_UnsetThingPosition(mobj);

  // stop any playing sound
  S_StopSound(mobj);

  // free block
  P_RemoveThinker(Pthinker_t(mobj));
end;

//
// P_RespawnSpecials
//
procedure P_RespawnSpecials;
var
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
  ss: Psubsector_t;
  mo: Pmobj_t;
  mthing: Pmapthing_t;
  i: integer;
begin
  // only respawn items in deathmatch
  if deathmatch <> 2 then
    exit; //

  // nothing left to respawn?
  if iquehead = iquetail then
    exit;

  // wait at least 30 seconds
  if leveltime - itemrespawntime[iquetail] < 30 * 35 then // VJ ???? 30 * TICKRATE ???
    exit;

  mthing := @itemrespawnque[iquetail];

  x := mthing.x * FRACUNIT;
  y := mthing.y * FRACUNIT;

  // spawn a teleport fog at the new spot
  ss := R_PointInSubsector(x, y);
  mo := P_SpawnMobj(x, y, ss.sector.floorheight, MT_IFOG);
  S_StartSound(mo, Ord(sfx_itmbk));

  // find which type to spawn
  i := 0;
  while i < Ord(NUMMOBJTYPES) do
  begin
    if mthing._type = mobjinfo[i].doomednum then
      break;
    inc(i);
  end;

  // spawn it
  if boolval(mobjinfo[i].flags and MF_SPAWNCEILING) then
    z := ONCEILINGZ
  else
    z := ONFLOORZ;

  mo := P_SpawnMobj(x, y, z, mobjtype_t(i));
  mo.spawnpoint := mthing^;
  mo.angle := ANG45 * (mthing.angle div 45);

  // pull it from the que
  iquetail := (iquetail + 1) and (ITEMQUESIZE - 1);
end;

//
// P_SpawnPlayer
// Called when a player is spawned on the level.
// Most of the player structure stays unchanged
//  between levels.
//
procedure P_SpawnPlayer(mthing: Pmapthing_t);
var
  p: Pplayer_t;
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
  mobj: Pmobj_t;
  i: integer;
begin
  // not playing?
  if not playeringame[mthing._type - 1] then
    exit;

  p := @players[mthing._type - 1];

  if p.playerstate = PST_REBORN then
    G_PlayerReborn(mthing._type - 1);

  x := mthing.x * FRACUNIT;
  y := mthing.y * FRACUNIT;
  z	:= ONFLOORZ;
  mobj := P_SpawnMobj(x, y, z, MT_PLAYER);

    // set color translations for player sprites
  if mthing._type > 1 then
    mobj.flags := mobj.flags or _SHL(mthing._type - 1, MF_TRANSSHIFT);

  mobj.angle := ANG45 * (mthing.angle div 45);
  mobj.player := p;
  mobj.health := p.health;

  p.mo := mobj;
  p.playerstate := PST_LIVE;
  p.refire := 0;
  p._message := '';
  p.damagecount := 0;
  p.bonuscount := 0;
  p.extralight := 0;
  p.fixedcolormap := 0;
  p.viewheight := PVIEWHEIGHT;

  // setup gun psprite
  P_SetupPsprites(p);

  // give all cards in death match mode
  if boolval(deathmatch) then
    for i := 0 to Ord(NUMCARDS) - 1 do
      p.cards[i] := true;

  if mthing._type - 1 = consoleplayer then
  begin
    // wake up the status bar
    ST_Start;
    // wake up the heads up text
    HU_Start;
  end;
end;

//
// P_SpawnMapThing
// The fields of the mapthing should
// already be in host byte order.
//
procedure P_SpawnMapThing(mthing: Pmapthing_t);
var
  i: integer;
  bit: integer;
  mobj: Pmobj_t;
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
begin
  // count deathmatch start positions
  if mthing._type = 11 then
  begin
    if deathmatch_p < MAX_DEATHMATCH_STARTS then
    begin
      memcpy(@deathmatchstarts[deathmatch_p], mthing, SizeOf(mthing^));
	    inc(deathmatch_p);
    end;
    exit;
  end;

  // check for players specially
  if mthing._type <= 4 then
  begin
    // save spots for respawning in network games
    playerstarts[mthing._type - 1] := mthing^;
    if not boolval(deathmatch) then
      P_SpawnPlayer(mthing);
    exit;
  end;

  // check for apropriate skill level
  if (not netgame) and boolval(mthing.options and 16) then
    exit;

  if gameskill = sk_baby then
    bit := 1
  else if gameskill = sk_nightmare then
    bit := 4
  else
    bit := _SHL(1, Ord(gameskill) - 1);

  if not boolval(mthing.options and bit) then
    exit;

  // find which type to spawn
  i := 0;
  while i < Ord(NUMMOBJTYPES) do
  begin
    if mthing._type = mobjinfo[i].doomednum then
      break;
    inc(i);
  end;

  if i = Ord(NUMMOBJTYPES) then
    I_Error('P_SpawnMapThing(): Unknown type %d at (%d, %d)',
      [mthing._type, mthing.x, mthing.y]);

  // don't spawn keycards and players in deathmatch
  if boolval(deathmatch) and boolval(mobjinfo[i].flags and MF_NOTDMATCH) then
    exit;

  // don't spawn any monsters if -nomonsters
  if nomonsters and
    ((i = Ord(MT_SKULL)) or boolval(mobjinfo[i].flags and MF_COUNTKILL)) then
    exit;

  // spawn it
  x := mthing.x * FRACUNIT;
  y := mthing.y * FRACUNIT;

  if boolval(mobjinfo[i].flags and MF_SPAWNCEILING) then
    z := ONCEILINGZ
  else
    z := ONFLOORZ;

  mobj := P_SpawnMobj(x, y, z, mobjtype_t(i));
  mobj.spawnpoint := mthing^;

  if mobj.tics > 0 then
    mobj.tics := 1 + (P_Random mod mobj.tics);
  if (mobj.flags and MF_COUNTKILL) <> 0 then
    inc(totalkills);
  if (mobj.flags and MF_COUNTITEM) <> 0 then
    inc(totalitems);

  mobj.angle := ANG45 * (mthing.angle div 45);
  if (mthing.options and MTF_AMBUSH) <> 0 then
    mobj.flags := mobj.flags or MF_AMBUSH;
end;

//
// GAME SPAWN FUNCTIONS
//


//
// P_SpawnPuff
//
procedure P_SpawnPuff(x, y, z: fixed_t);
var
  th: Pmobj_t;
begin
  z := z + _SHL(P_Random - P_Random, 10);

  th := P_SpawnMobj(x, y, z, MT_PUFF);
  th.momz := FRACUNIT;
  th.tics := th.tics - (P_Random and 3);

  if th.tics < 1 then
    th.tics := 1;

  // don't make punches spark on the wall
  if attackrange = MELEERANGE then
    P_SetMobjState(th, S_PUFF3);
end;

//
// P_SpawnBlood
//
procedure P_SpawnBlood(x, y, z: fixed_t; damage: integer);
var
  th: Pmobj_t;
begin
  z := z + _SHL(P_Random - P_Random, 10);
  th := P_SpawnMobj(x, y, z, MT_BLOOD);
  th.momz := FRACUNIT * 2;
  th.tics := th.tics - (P_Random and 3);

  if th.tics < 1 then
    th.tics := 1;

  if (damage <= 12) and (damage >= 9) then
    P_SetMobjState(th, S_BLOOD2)
  else if damage < 9 then
    P_SetMobjState(th, S_BLOOD3);
end;

//
// P_CheckMissileSpawn
// Moves the missile forward a bit
//  and possibly explodes it right there.
//
procedure P_CheckMissileSpawn(th: Pmobj_t);
begin
  th.tics := th.tics - (P_Random and 3);
  if th.tics < 1 then
    th.tics := 1;

  // move a little forward so an angle can
  // be computed if it immediately explodes
  th.x := th.x + _SHR(th.momx, 1);
  th.y := th.y + _SHR(th.momy, 1);
  th.z := th.z + _SHR(th.momz, 1);

  if not P_TryMove(th, th.x, th.y) then
    P_ExplodeMissile(th);
end;

//
// P_SpawnMissile
//

function P_SpawnMissile(source: Pmobj_t; dest: Pmobj_t; _type: mobjtype_t): Pmobj_t;
var
  th: Pmobj_t;
  an: angle_t;
  dist: integer;
begin
  th := P_SpawnMobj(source.x, source.y, source.z + 4 * 8 * FRACUNIT, _type);

  if boolval(th.info.seesound) then
    S_StartSound(th, th.info.seesound);

  th.target := source;	// where it came from
  an := R_PointToAngle2(source.x, source.y, dest.x, dest.y);

  // fuzzy player
  if boolval(dest.flags and MF_SHADOW) then
    an := an + _SHLW(P_Random - P_Random, 20);

  th.angle := an;
  an := _SHRW(an, ANGLETOFINESHIFT);
  th.momx := FixedMul(th.info.speed, finecosine[an]);
  th.momy := FixedMul(th.info.speed, finesine[an]);

  dist := P_AproxDistance(dest.x - source.x, dest.y - source.y);
  dist := dist div th.info.speed;

  if dist < 1 then
    dist := 1;

  th.momz := (dest.z - source.z) div dist;
  P_CheckMissileSpawn(th);

  result := th;
end;

//
// P_SpawnPlayerMissile
// Tries to aim at a nearby monster
//
procedure P_SpawnPlayerMissile(source: Pmobj_t; _type: mobjtype_t);
var
  th: Pmobj_t;
  an: angle_t;
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
  slope: fixed_t;
begin
  // see which target is to be aimed at
  an := source.angle;
  slope := P_AimLineAttack(source, an, 16 * 64 * FRACUNIT);

  if not boolval(linetarget) then
  begin
    an := an + _SHLW(1, 26);
    slope := P_AimLineAttack(source, an, 16 * 64 * FRACUNIT);

    if not boolval(linetarget) then
    begin
      an := an - _SHLW(2, 26);
      slope := P_AimLineAttack(source, an, 16 * 64 * FRACUNIT);
    end;

    if not boolval(linetarget) then
    begin
      an := source.angle;
//      slope := 0;
//      slope := ((source.player.lookdir)<<FRACBITS)/173;
      slope := (Pplayer_t(source.player).lookdir * FRACUNIT) div 173;
    end;
  end;

  x := source.x;
  y := source.y;
//	z = source->z + 4*8*FRACUNIT+((source.player.lookdir)<<FRACBITS)/173;
	z := source.z + 4 * 8 * FRACUNIT + (Pplayer_t(source.player).lookdir * FRACUNIT) div 173;
//  z := source.z + 4 * 8 * FRACUNIT;

  th := P_SpawnMobj(x, y, z, _type);

  if boolval(th.info.seesound) then
    S_StartSound(th, th.info.seesound);

  th.target := source;
  th.angle := an;
  th.momx := FixedMul(th.info.speed, finecosine[_SHRW(an, ANGLETOFINESHIFT)]);
  th.momy := FixedMul(th.info.speed, finesine[_SHRW(an, ANGLETOFINESHIFT)]);
  th.momz := FixedMul(th.info.speed, slope);

  P_CheckMissileSpawn(th);
end;



(* Unit p_telept implementation *)


                                                                                                                                          

function EV_Teleport(line: Pline_t; side: integer; thing: Pmobj_t): integer;
var
  i: integer;
  tag: integer;
  m: Pmobj_t;
  fog: Pmobj_t;
  an: LongWord;
  thinker: Pthinker_t;
  sector: Psector_t;
  oldx: fixed_t;
  oldy: fixed_t;
  oldz: fixed_t;
begin
  // don't teleport missiles
  if boolval(thing.flags and MF_MISSILE) then
  begin
    result := 0;
    exit;
  end;

  // Don't teleport if hit back of line,
  //  so you can get out of teleporter.
  if side = 1 then
  begin
    result := 0;
    exit;
  end;

  tag := line.tag;
  for i := 0 to numsectors - 1 do
  begin
    if sectors[i].tag = tag then
    begin
      thinker := thinkercap.next;
      while thinker <> @thinkercap do
      begin
        // not a mobj
        if @thinker._function.acp1 <> @P_MobjThinker then
        begin
          thinker := thinker.next;
          continue;
        end;

        m := Pmobj_t(thinker);

        // not a teleportman
        if m._type <> MT_TELEPORTMAN then
        begin
          thinker := thinker.next;
          continue;
        end;

        sector := Psubsector_t(m.subsector).sector;
        // wrong sector
        if sector <> @sectors[i] then // VJ (was (sector-sectors != i ))
        begin
          thinker := thinker.next;
          continue;
        end;

        oldx := thing.x;
        oldy := thing.y;
        oldz := thing.z;

        if not P_TeleportMove(thing, m.x, m.y) then
        begin
          result := 0;
          exit;
        end;

        thing.z := thing.floorz;  //fixme: not needed?
        if boolval(thing.player) then
          Pplayer_t(thing.player).viewz := thing.z + Pplayer_t(thing.player).viewheight;

        // spawn teleport fog at source and destination
        fog := P_SpawnMobj(oldx, oldy, oldz, MT_TFOG);
        S_StartSound (fog, Ord(sfx_telept));
        an := _SHRW(m.angle, ANGLETOFINESHIFT);
        fog := P_SpawnMobj(m.x + 20 * finecosine[an],
                           m.y + 20 * finesine[an],
                           thing.z, MT_TFOG);

        // emit sound, where?
        S_StartSound (fog, Ord(sfx_telept));

        // don't move for a bit
        if boolval(thing.player) then
          thing.reactiontime := 18;

        thing.angle := m.angle;
        thing.momx := 0;
        thing.momy := 0;
        thing.momz := 0;
        result := 1;
        exit;
      end;
    end;
  end;
  result := 0;
end;


(* Unit d_player implementation *)




(* Unit p_spec implementation *)


                                                                                                                                                                                                                                                                              

//
// Animating textures and planes
// There is another anim_t used in wi_stuff, unrelated.
//
type
  anim_t = record
    istexture: boolean;
    picnum: integer;
    basepic: integer;
    numpics: integer;
    speed: integer;
  end;
  Panim_t = ^anim_t;

//
//      source animation definition
//
  animdef_t = record
    istexture: boolean; // if false, it is a flat
    endname: string[8];
    startname: string[8];
    speed: integer;
  end;
  Panimdef_t = ^animdef_t;

const
  MAXANIMS = 32;

//
// P_InitPicAnims
//

// Floor/ceiling animation sequences,
//  defined by first and last frame,
//  i.e. the flat (64x64 tile) name to
//  be used.
// The full animation sequence is given
//  using all the flats between the start
//  and end entry, in the order found in
//  the WAD file.
//
  animdefs: array[0..22] of animdef_t = (
    (istexture: false; endname: 'NUKAGE3';  startname: 'NUKAGE1';  speed: 8),
    (istexture: false; endname: 'FWATER4';  startname: 'FWATER1';  speed: 8),
    (istexture: false; endname: 'SWATER4';  startname: 'SWATER1';  speed: 8),
    (istexture: false; endname: 'LAVA4';    startname: 'LAVA1';    speed: 8),
    (istexture: false; endname: 'BLOOD3';   startname: 'BLOOD1';   speed: 8),

    // DOOM II flat animations.
    (istexture: false; endname: 'RROCK08';  startname: 'RROCK05';  speed: 8),
    (istexture: false; endname: 'SLIME04';  startname: 'SLIME01';  speed: 8),
    (istexture: false; endname: 'SLIME08';  startname: 'SLIME05';  speed: 8),
    (istexture: false; endname: 'SLIME12';  startname: 'SLIME09';  speed: 8),

    (istexture: true;  endname: 'BLODGR4';  startname: 'BLODGR1';  speed: 8),
    (istexture: true;  endname: 'SLADRIP3'; startname: 'SLADRIP1'; speed: 8),

    (istexture: true;  endname: 'BLODRIP4'; startname: 'BLODRIP1'; speed: 8),
    (istexture: true;  endname: 'FIREWALL'; startname: 'FIREWALA'; speed: 8),
    (istexture: true;  endname: 'GSTFONT3'; startname: 'GSTFONT1'; speed: 8),
    (istexture: true;  endname: 'FIRELAVA'; startname: 'FIRELAV3'; speed: 8),
    (istexture: true;  endname: 'FIREMAG3'; startname: 'FIREMAG1'; speed: 8),
    (istexture: true;  endname: 'FIREBLU2'; startname: 'FIREBLU1'; speed: 8),
    (istexture: true;  endname: 'ROCKRED3'; startname: 'ROCKRED1'; speed: 8),

    (istexture: true;  endname: 'BFALL4';   startname: 'BFALL1';   speed: 8),
    (istexture: true;  endname: 'SFALL4';   startname: 'SFALL1';   speed: 8),
    (istexture: true;  endname: 'WFALL4';   startname: 'WFALL1';   speed: 8),
    (istexture: true;  endname: 'DBRAIN4';  startname: 'DBRAIN1';  speed: 8),

    (istexture: false; endname: '';         startname: '';         speed: 0)
  );

var
  anims: array[0..MAXANIMS - 1] of anim_t;
  lastanim: integer;

const
//
//      Animating line specials
//
  MAXLINEANIMS = 64;

procedure P_InitPicAnims;
var
  i: integer;
begin
  //	Init animation
  lastanim := 0;
  i := 0;
  while animdefs[i].speed <> 0 do
  begin
    if animdefs[i].istexture then
    begin
	    // different episode ?
      if R_CheckTextureNumForName(animdefs[i].startname) = -1 then
      begin
        inc(i);
        continue;
      end;

      anims[lastanim].picnum := R_TextureNumForName(animdefs[i].endname);
      anims[lastanim].basepic := R_TextureNumForName(animdefs[i].startname);
    end
    else
    begin
      if W_CheckNumForName(animdefs[i].startname) = -1 then
      begin
        inc(i);
        continue;
      end;

      anims[lastanim].picnum := R_FlatNumForName(animdefs[i].endname);
      anims[lastanim].basepic := R_FlatNumForName(animdefs[i].startname);
    end;

    anims[lastanim].istexture := animdefs[i].istexture;
    anims[lastanim].numpics := anims[lastanim].picnum - anims[lastanim].basepic + 1;

    if anims[lastanim].numpics < 2 then
      I_Error('P_InitPicAnims(): bad cycle from %s to %s',
        [animdefs[i].startname, animdefs[i].endname]);

    anims[lastanim].speed := animdefs[i].speed;
    inc(lastanim);
    inc(i);
  end;
end;

//
// UTILITIES
//



//
// getSide()
// Will return a side_t*
//  given the number of the current sector,
//  the line number, and the side (0/1) that you want.
//
function getSide(currentSector: integer; line: integer; side: integer): Pside_t;
begin
  result := @sides[(sectors[currentSector].lines[line]).sidenum[side]];
end;

//
// getSector()
// Will return a sector_t*
//  given the number of the current sector,
//  the line number and the side (0/1) that you want.
//
function getSector(currentSector: integer; line: integer; side: integer): Psector_t;
begin
  result := sides[(sectors[currentSector].lines[line]).sidenum[side]].sector;
end;

//
// twoSided()
// Given the sector number and the line number,
//  it will tell you whether the line is two-sided or not.
//
function twoSided(sector: integer; line: integer): integer;
begin
  result := (sectors[sector].lines[line]).flags and ML_TWOSIDED;
end;

//
// getNextSector()
// Return sector_t * of sector next to current.
// NULL if not two-sided line
//
function getNextSector(line: Pline_t; sec: Psector_t): Psector_t;
begin
  if not boolval(line.flags and ML_TWOSIDED) then
    result := nil
  else
  begin
    if line.frontsector = sec then
      result := line.backsector
    else
      result := line.frontsector;
  end;
end;

//
// P_FindLowestFloorSurrounding()
// FIND LOWEST FLOOR HEIGHT IN SURROUNDING SECTORS
//
function P_FindLowestFloorSurrounding(sec: Psector_t): fixed_t;
var
  i: integer;
  check: Pline_t;
  other: Psector_t;
begin
  result := sec.floorheight;

  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    other := getNextSector(check, sec);

    if other <> nil then
      if other.floorheight < result then
        result := other.floorheight;
  end;
end;

//
// P_FindHighestFloorSurrounding()
// FIND HIGHEST FLOOR HEIGHT IN SURROUNDING SECTORS
//
function P_FindHighestFloorSurrounding(sec: Psector_t): fixed_t;
var
  i: integer;
  check: Pline_t;
  other: Psector_t;
begin
  result := -500 * FRACUNIT;

  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    other := getNextSector(check, sec);

    if other <> nil then
      if other.floorheight > result then
	      result := other.floorheight;
  end;
end;

//
// P_FindNextHighestFloor
// FIND NEXT HIGHEST FLOOR IN SURROUNDING SECTORS
// Note: this should be doable w/o a fixed array.

// 20 adjoining sectors max!
const
  MAX_ADJOINING_SECTORS = 20;

function P_FindNextHighestFloor(sec: Psector_t; currentheight: integer): fixed_t;
var
  i: integer;
  h: integer;
  check: Pline_t;
  other: Psector_t;
  height: fixed_t;
  heightlist: array[0..MAX_ADJOINING_SECTORS - 1] of fixed_t;
begin
  height := currentheight;

  h := 0;
  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    other := getNextSector(check, sec);

    if other <> nil then
    begin
      if other.floorheight > height then
      begin
        heightlist[h] := other.floorheight;
        inc(h);
      end;

      // Check for overflow. Exit.
      if h >= MAX_ADJOINING_SECTORS then
      begin
        fprintf(stderr, 'Sector with more than 20 adjoining sectors' + #13#10);
        break;
      end;
    end;
  end;

  // Find lowest height in list
  if not boolval(h) then
  begin
    result := currentheight;
    exit;
  end;

  result := heightlist[0];

  // Range checking?
  for i := 1 to h - 1 do
    if heightlist[i] < result then
      result := heightlist[i];
end;

//
// FIND LOWEST CEILING IN THE SURROUNDING SECTORS
//
function P_FindLowestCeilingSurrounding(sec: Psector_t): fixed_t;
var
  i: integer;
  check: Pline_t;
  other: Psector_t;
begin
  result := MAXINT;

  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    other := getNextSector(check, sec);

    if other <> nil then
      if other.ceilingheight < result then
        result := other.ceilingheight;
  end;
end;

//
// FIND HIGHEST CEILING IN THE SURROUNDING SECTORS
//
function P_FindHighestCeilingSurrounding(sec: Psector_t): fixed_t;
var
  i: integer;
  check: Pline_t;
  other: Psector_t;
begin
  result := 0;

  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    other := getNextSector(check, sec);

    if other <> nil then
      if other.ceilingheight > result then
        result := other.ceilingheight;
  end;
end;

//
// RETURN NEXT SECTOR # THAT LINE TAG REFERS TO
//
function P_FindSectorFromLineTag(line: Pline_t; start: integer): integer;
var
  i: integer;
begin
  for i := start + 1 to numsectors - 1 do
    if sectors[i].tag = line.tag then
    begin
      result := i;
      exit;
    end;

  result := -1;
end;

//
// Find minimum light from an adjacent sector
//
function P_FindMinSurroundingLight(sector: Psector_t; max: integer): integer;
var
  i: integer;
  line: Pline_t;
  check: Psector_t;
begin
  result := max;
  for i := 0 to sector.linecount - 1 do
  begin
    line := sector.lines[i];
    check := getNextSector(line, sector);

    if check <> nil then
      if check.lightlevel < result then
        result := check.lightlevel;
  end;
end;

//
// EVENTS
// Events are operations triggered by using, crossing,
// or shooting special lines, or by timed thinkers.
//

//
// P_CrossSpecialLine - TRIGGER
// Called every time a thing origin is about
//  to cross a line with a non 0 special.
//
procedure P_CrossSpecialLine(linenum: integer; side: integer; thing: Pmobj_t);
var
  line: Pline_t;
begin
  line := @lines[linenum];

  //	Triggers that other things can activate
  if not boolval(thing.player) then
  begin
    // Things that should NOT trigger specials...
    case thing._type of
      MT_ROCKET,
      MT_PLASMA,
      MT_BFG,
      MT_TROOPSHOT,
      MT_HEADSHOT,
      MT_BRUISERSHOT:
	      exit;
    end;

    case line.special of
      39, // TELEPORT TRIGGER
      97, // TELEPORT RETRIGGER
     125, // TELEPORT MONSTERONLY TRIGGER
     126, // TELEPORT MONSTERONLY RETRIGGER
       4, // RAISE DOOR
      10, // PLAT DOWN-WAIT-UP-STAY TRIGGER
      88: // PLAT DOWN-WAIT-UP-STAY RETRIGGER
        ;
      else
        exit;
    end;
  end;


  // Note: could use some const's here.
  case line.special of
	// TRIGGERS.
	// All from here to RETRIGGERS.
     2:
      begin
        // Open Door
        EV_DoDoor(line, open);
        line.special := 0;
      end;

     3:
      begin
        // Close Door
        EV_DoDoor(line, close);
        line.special := 0;
      end;

     4:
      begin
        // Raise Door
        EV_DoDoor(line, normal);
        line.special := 0;
      end;

     5:
      begin
        // Raise Floor
        EV_DoFloor(line, raiseFloor);
        line.special := 0;
      end;

     6:
      begin
        // Fast Ceiling Crush & Raise
        EV_DoCeiling(line, fastCrushAndRaise);
        line.special := 0;
      end;

     8:
      begin
        // Build Stairs
        EV_BuildStairs(line, build8);
        line.special := 0;
      end;

    10:
      begin
        // PlatDownWaitUp
        EV_DoPlat(line, downWaitUpStay,0);
        line.special := 0;
      end;

    12:
      begin
        // Light Turn On - brightest near
        EV_LightTurnOn(line, 0);
        line.special := 0;
      end;

    13:
      begin
        // Light Turn On 255
        EV_LightTurnOn(line, 255);
        line.special := 0;
      end;

    16:
      begin
        // Close Door 30
        EV_DoDoor(line, close30ThenOpen);
        line.special := 0;
      end;

    17:
      begin
        // Start Light Strobing
        EV_StartLightStrobing(line);
        line.special := 0;
      end;

    19:
      begin
        // Lower Floor
        EV_DoFloor(line, lowerFloor);
        line.special := 0;
      end;

    22:
      begin
        // Raise floor to nearest height and change texture
        EV_DoPlat(line, raiseToNearestAndChange, 0);
        line.special := 0;
      end;

    25:
      begin
        // Ceiling Crush and Raise
        EV_DoCeiling(line, crushAndRaise);
        line.special := 0;
      end;
	
    30:
      begin
        // Raise floor to shortest texture height
        //  on either side of lines.
        EV_DoFloor(line, raiseToTexture);
        line.special := 0;
      end;

    35:
      begin
        // Lights Very Dark
        EV_LightTurnOn(line, 35);
        line.special := 0;
      end;

    36:
      begin
        // Lower Floor (TURBO)
        EV_DoFloor(line, turboLower);
        line.special := 0;
      end;

    37:
      begin
        // LowerAndChange
        EV_DoFloor(line, lowerAndChange);
        line.special := 0;
      end;

    38:
      begin
        // Lower Floor To Lowest
        EV_DoFloor(line, lowerFloorToLowest);
        line.special := 0;
      end;

    39:
      begin
        // TELEPORT!
        EV_Teleport(line, side, thing );
        line.special := 0;
      end;

    40:
      begin
        // RaiseCeilingLowerFloor
        EV_DoCeiling(line, raiseToHighest);
        EV_DoFloor(line, lowerFloorToLowest);
        line.special := 0;
      end;

    44:
      begin
        // Ceiling Crush
        EV_DoCeiling(line, lowerAndCrush);
        line.special := 0;
      end;

    52:
      begin
        // EXIT!
        G_ExitLevel;
      end;

    53:
      begin
        // Perpetual Platform Raise
        EV_DoPlat(line, perpetualRaise, 0);
        line.special := 0;
      end;

    54:
      begin
        // Platform Stop
        EV_StopPlat(line);
        line.special := 0;
      end;

    56:
      begin
        // Raise Floor Crush
        EV_DoFloor(line, raiseFloorCrush);
        line.special := 0;
      end;

    57:
      begin
        // Ceiling Crush Stop
        EV_CeilingCrushStop(line);
        line.special := 0;
      end;

    58:
      begin
        // Raise Floor 24
        EV_DoFloor(line, raiseFloor24);
        line.special := 0;
      end;

    59:
      begin
        // Raise Floor 24 And Change
        EV_DoFloor(line, raiseFloor24AndChange);
        line.special := 0;
      end;

   104:
      begin
        // Turn lights off in sector(tag)
        EV_TurnTagLightsOff(line);
        line.special := 0;
      end;

   108:
      begin
        // Blazing Door Raise (faster than TURBO!)
        EV_DoDoor(line, blazeRaise);
        line.special := 0;
      end;

   109:
      begin
        // Blazing Door Open (faster than TURBO!)
        EV_DoDoor(line, blazeOpen);
        line.special := 0;
      end;

   100:
      begin
        // Build Stairs Turbo 16
        EV_BuildStairs(line, turbo16);
        line.special := 0;
      end;

   110:
      begin
        // Blazing Door Close (faster than TURBO!)
        EV_DoDoor(line, blazeClose);
        line.special := 0;
      end;

   119:
      begin
        // Raise floor to nearest surr. floor
        EV_DoFloor(line, raiseFloorToNearest);
        line.special := 0;
      end;

   121:
      begin
        // Blazing PlatDownWaitUpStay
        EV_DoPlat(line, blazeDWUS, 0);
        line.special := 0;
      end;

   124:
      begin
        // Secret EXIT
        G_SecretExitLevel;
      end;

   125:
      begin
        // TELEPORT MonsterONLY
        if not boolval(thing.player) then
        begin
          EV_Teleport(line, side, thing);
          line.special := 0;
        end;
      end;

   130:
      begin
        // Raise Floor Turbo
        EV_DoFloor(line, raiseFloorTurbo);
        line.special := 0;
      end;

   141:
      begin
        // Silent Ceiling Crush & Raise
        EV_DoCeiling(line, silentCrushAndRaise);
        line.special := 0;
      end;
	
	// RETRIGGERS.  All from here till end.
    72:
      begin
        // Ceiling Crush
        EV_DoCeiling(line, lowerAndCrush);
      end;

    73:
      begin
        // Ceiling Crush and Raise
        EV_DoCeiling(line, crushAndRaise);
      end;

    74:
      begin
        // Ceiling Crush Stop
        EV_CeilingCrushStop(line);
      end;

    75:
      begin
        // Close Door
        EV_DoDoor(line, close);
      end;

    76:
      begin
        // Close Door 30
        EV_DoDoor(line, close30ThenOpen);
      end;

    77:
      begin
        // Fast Ceiling Crush & Raise
        EV_DoCeiling(line,fastCrushAndRaise);
      end;

    79:
      begin
        // Lights Very Dark
        EV_LightTurnOn(line, 35);
      end;

    80:
      begin
        // Light Turn On - brightest near
        EV_LightTurnOn(line, 0);
      end;

    81:
      begin
        // Light Turn On 255
        EV_LightTurnOn(line, 255);
      end;

    82:
      begin
        // Lower Floor To Lowest
        EV_DoFloor(line, lowerFloorToLowest);
      end;

    83:
      begin
        // Lower Floor
        EV_DoFloor(line,lowerFloor);
      end;

    84:
      begin
        // LowerAndChange
        EV_DoFloor(line,lowerAndChange);
      end;

    86:
      begin
        // Open Door
        EV_DoDoor(line,open);
      end;

    87:
      begin
        // Perpetual Platform Raise
        EV_DoPlat(line, perpetualRaise, 0);
      end;

    88:
      begin
        // PlatDownWaitUp
        EV_DoPlat(line, downWaitUpStay, 0);
      end;
	
    89:
      begin
        // Platform Stop
        EV_StopPlat(line);
      end;
	
    90:
      begin
        // Raise Door
        EV_DoDoor(line, normal);
      end;

    91:
      begin
        // Raise Floor
        EV_DoFloor(line, raiseFloor);
      end;

    92:
      begin
        // Raise Floor 24
        EV_DoFloor(line, raiseFloor24);
      end;

    93:
      begin
        // Raise Floor 24 And Change
        EV_DoFloor(line, raiseFloor24AndChange);
      end;
	
    94:
      begin
        // Raise Floor Crush
        EV_DoFloor(line,raiseFloorCrush);
      end;

    95:
      begin
        // Raise floor to nearest height
        // and change texture.
        EV_DoPlat(line, raiseToNearestAndChange, 0);
      end;

    96:
      begin
        // Raise floor to shortest texture height
        // on either side of lines.
        EV_DoFloor(line, raiseToTexture);
      end;

    97:
      begin
        // TELEPORT!
        EV_Teleport(line, side, thing);
      end;

    98:
      begin
        // Lower Floor (TURBO)
        EV_DoFloor(line, turboLower);
      end;

   105:
      begin
        // Blazing Door Raise (faster than TURBO!)
        EV_DoDoor(line, blazeRaise);
      end;
	
   106:
      begin
        // Blazing Door Open (faster than TURBO!)
        EV_DoDoor(line, blazeOpen);
      end;

   107:
      begin
        // Blazing Door Close (faster than TURBO!)
        EV_DoDoor(line, blazeClose);
      end;

   120:
      begin
        // Blazing PlatDownWaitUpStay.
        EV_DoPlat(line, blazeDWUS, 0);
      end;

   126:
      begin
        // TELEPORT MonsterONLY.
        if not boolval(thing.player) then
          EV_Teleport(line, side, thing);
      end;

   128:
      begin
        // Raise To Nearest Floor
        EV_DoFloor(line, raiseFloorToNearest);
      end;

   129:
      begin
        // Raise Floor Turbo
        EV_DoFloor(line, raiseFloorTurbo);
      end;
  end;
end;

//
// P_ShootSpecialLine - IMPACT SPECIALS
// Called when a thing shoots a special line.
//
procedure P_ShootSpecialLine(thing: Pmobj_t; line: Pline_t);
begin
  //	Impacts that other things can activate.
  if not boolval(thing.player) then
    case line.special of
      46: ; // OPEN DOOR IMPACT
    else
      exit;
    end;

  case line.special of
    24:
      begin
        // RAISE FLOOR
        EV_DoFloor(line, raiseFloor);
        P_ChangeSwitchTexture(line, 0);
      end;

    46:
      begin
        // OPEN DOOR
        EV_DoDoor(line, open);
        P_ChangeSwitchTexture(line, 1);
      end;

    47:
      begin
        // RAISE FLOOR NEAR AND CHANGE
        EV_DoPlat(line, raiseToNearestAndChange, 0);
        P_ChangeSwitchTexture(line, 0);
      end;
  end;
end;

//
// P_PlayerInSpecialSector
// Called every tic frame
//  that the player origin is in a special sector
//
procedure P_PlayerInSpecialSector(player: Pplayer_t);
var
  sector: Psector_t;
begin
  sector := Psubsector_t(player.mo.subsector).sector;

  // Falling, not all the way down yet?
  if player.mo.z <> sector.floorheight then
    exit;

  // Has hitten ground.
  case sector.special of
     5:
      begin
        // HELLSLIME DAMAGE
        if not boolval(player.powers[Ord(pw_ironfeet)]) then
          if not boolval(leveltime and $1f) then
            P_DamageMobj(player.mo, nil, nil, 10);
      end;

     7:
      begin
        // NUKAGE DAMAGE
        if not boolval(player.powers[Ord(pw_ironfeet)]) then
          if not boolval(leveltime and $1f) then
            P_DamageMobj(player.mo, nil, nil, 5);
      end;

    16, // SUPER HELLSLIME DAMAGE
     4: // STROBE HURT
      begin
        if (not boolval(player.powers[Ord(pw_ironfeet)])) or
           (P_Random < 5) then
          if not boolval(leveltime and $1f) then
            P_DamageMobj(player.mo, nil, nil, 20);
      end;

     9:
      begin
        // SECRET SECTOR
        player.secretcount := player.secretcount + 1;
        sector.special := 0;
      end;

    11:
      begin
        // EXIT SUPER DAMAGE! (for E1M8 finale)
        player.cheats := player.cheats and (not CF_GODMODE);

        if not boolval(leveltime and $1f) then
          P_DamageMobj(player.mo, nil, nil, 20);

        if player.health <= 10 then
          G_ExitLevel;
      end;

  else
    I_Error('P_PlayerInSpecialSector(): unknown special %d', [sector.special]);
  end;
end;

var
  numlinespecials: smallint;
  linespeciallist: array[0..MAXLINEANIMS - 1] of Pline_t;


//
// P_UpdateSpecials
// Animate planes, scroll walls, etc.
//
var
  levelTimer: boolean;
  levelTimeCount: integer;

procedure P_UpdateSpecials;
var
  anim: Panim_t;
  pic: integer;
  i: integer;
  j: integer;
  line: Pline_t;
begin
  // LEVEL TIMER
  if levelTimer then
  begin
    dec(levelTimeCount);
    if not boolval(levelTimeCount) then
      G_ExitLevel;
  end;

  // ANIMATE FLATS AND TEXTURES GLOBALLY
  for j := 0 to lastanim - 1 do
  begin
    anim := @anims[j];
    for i := anim.basepic to anim.basepic + anim.numpics - 1 do
    begin
      pic := anim.basepic + ((leveltime div anim.speed + i) mod anim.numpics);
      if anim.istexture then
        texturetranslation[i] := pic
      else
        flattranslation[i] := pic;
    end;
  end;

  // ANIMATE LINE SPECIALS
  for i := 0 to numlinespecials - 1 do
  begin
    line := linespeciallist[i];
    if line.special = 48 then
      // EFFECT FIRSTCOL SCROLL +
      sides[line.sidenum[0]].textureoffset := sides[line.sidenum[0]].textureoffset + FRACUNIT;
  end;


  // DO BUTTONS
  for i := 0 to MAXBUTTONS - 1 do
  begin
    if boolval(buttonlist[i].btimer) then
    begin
      buttonlist[i].btimer := buttonlist[i].btimer - 1;

      if not boolval(buttonlist[i].btimer) then
      begin
        case buttonlist[i].where of
          top:
            sides[buttonlist[i].line.sidenum[0]].toptexture := buttonlist[i].btexture;

          middle:
            sides[buttonlist[i].line.sidenum[0]].midtexture := buttonlist[i].btexture;

          bottom:
            sides[buttonlist[i].line.sidenum[0]].bottomtexture := buttonlist[i].btexture;
        end;
        S_StartSound(Pmobj_t(@buttonlist[i].soundorg), Ord(sfx_swtchn));
        memset(@buttonlist[i], 0, SizeOf(button_t));
      end;

    end;
  end;
end;

//
// Special Stuff that can not be categorized
//
function EV_DoDonut(line: Pline_t): integer;
var
  s1: Psector_t;
  s2: Psector_t;
  s3: Psector_t;
  secnum: integer;
  i: integer;
  floor: Pfloormove_t;
begin
  result := 0;
  secnum := P_FindSectorFromLineTag(line, -1);
  while secnum >= 0 do
  begin
    s1 := @sectors[secnum];
    secnum := P_FindSectorFromLineTag(line, secnum);

    // ALREADY MOVING?  IF SO, KEEP GOING...
    if s1.specialdata <> nil then
      continue;

    result := 1;
    s2 := getNextSector(s1.lines[0], s1);
    for i := 0 to s2.linecount - 1 do
    begin
      if (not boolval(s2.lines[i].flags and ML_TWOSIDED)) or
         (s2.lines[i].backsector = s1) then
        continue;
      s3 := s2.lines[i].backsector;

      //	Spawn rising slime
      floor := Z_Malloc(SizeOf(floor^), PU_LEVSPEC, nil);
      P_AddThinker(@floor.thinker);
      s2.specialdata := floor;
      floor.thinker._function.acp1 := @T_MoveFloor;
      floor._type := donutRaise;
      floor.crush := false;
      floor.direction := 1;
      floor.sector := s2;
      floor.speed := FLOORSPEED div 2;
      floor.texture := s3.floorpic;
      floor.newspecial := 0;
      floor.floordestheight := s3.floorheight;

      //	Spawn lowering donut-hole
      floor := Z_Malloc (sizeof(floor^), PU_LEVSPEC, nil);
      P_AddThinker(@floor.thinker);
      s1.specialdata := floor;
      floor.thinker._function.acp1 := @T_MoveFloor;
      floor._type := lowerFloor;
      floor.crush := false;
      floor.direction := -1;
      floor.sector := s1;
      floor.speed := FLOORSPEED div 2;
      floor.floordestheight := s3.floorheight;
      break;
    end;
  end;
end;

//
// SPECIAL SPAWNING
//

//
// P_SpawnSpecials
// After the map has been loaded, scan for specials
//  that spawn thinkers
//
// Parses command line parameters.
procedure P_SpawnSpecials;
var
  sector: Psector_t;
  i: integer;
  time: integer;
begin
  if W_CheckNumForName('texture2') < 0 then
    gameepisode := 1; // ??? 


  // See if -TIMER needs to be used.
  levelTimer := false;

  i := M_CheckParm('-avg');
  if boolval(i) and boolval(deathmatch) then
  begin
    levelTimer := true;
    levelTimeCount := 20 * 60 * 35; // VJ 20 * 60 * TICKRATE ???
  end;

  i := M_CheckParm('-timer');
  if boolval(i) and boolval(deathmatch) then
  begin
    time := atoi(myargv[i+1]) * 60 * 35;
    levelTimer := true;
    levelTimeCount := time;
  end;

  //	Init special SECTORs.
  for i := 0 to numsectors - 1 do
  begin
    sector := @sectors[i];
    if not boolval(sector.special) then
      continue;

    case sector.special of
     1:
      begin
        // FLICKERING LIGHTS
        P_SpawnLightFlash(sector);
      end;

     2:
      begin
        // STROBE FAST
        P_SpawnStrobeFlash(sector, FASTDARK, 0);
      end;

     3:
      begin
        // STROBE SLOW
        P_SpawnStrobeFlash(sector, SLOWDARK, 0);
      end;

     4:
      begin
        // STROBE FAST/DEATH SLIME
        P_SpawnStrobeFlash(sector, FASTDARK, 0);
        sector.special := 4;
      end;

     8:
      begin
        // GLOWING LIGHT
        P_SpawnGlowingLight(sector);
      end;

     9:
      begin
        // SECRET SECTOR
        inc(totalsecret);
      end;

    10:
      begin
        // DOOR CLOSE IN 30 SECONDS
        P_SpawnDoorCloseIn30 (sector);
      end;

    12:
      begin
        // SYNC STROBE SLOW
        P_SpawnStrobeFlash (sector, SLOWDARK, 1);
      end;

    13:
      begin
        // SYNC STROBE FAST
        P_SpawnStrobeFlash (sector, FASTDARK, 1);
      end;

    14:
      begin
        // DOOR RAISE IN 5 MINUTES
        P_SpawnDoorRaiseIn5Mins (sector, i);
      end;

    17:
      begin
        P_SpawnFireFlicker(sector);
      end;
    end;
  end;


    //	Init line EFFECTs
  numlinespecials := 0;
  for i := 0 to numlines - 1 do
  begin
    if lines[i].special = 48 then
    begin
      // EFFECT FIRSTCOL SCROLL+
      linespeciallist[numlinespecials] := @lines[i];
      inc(numlinespecials);
    end;
  end;


  //	Init other misc stuff
  for i := 0 to MAXCEILINGS - 1 do
    activeceilings[i] := nil;

  for i := 0 to MAXPLATS - 1 do
    activeplats[i] := nil;

  for i := 0 to MAXBUTTONS - 1 do
    memset(@buttonlist[i], 0, SizeOf(button_t));

    // UNUSED: no horizonal sliders.
    //	P_InitSlidingDoorFrames();
end;



(* Unit p_plats implementation *)


                                                                  

procedure T_PlatRaise(plat: Pplat_t);
var
  res: result_e;
begin
  case plat.status of
    up:
      begin
        res := T_MovePlane(plat.sector, plat.speed, plat.high, plat.crush, 0, 1);

        if (plat._type = raiseAndChange) or (plat._type = raiseToNearestAndChange) then
        begin
          if leveltime and 7 = 0 then
            S_StartSound(Pmobj_t(@plat.sector.soundorg), Ord(sfx_stnmov));
        end;

        if (res = crushed) and (not plat.crush) then
        begin
          plat.count := plat.wait;
          plat.status := down;
          S_StartSound(Pmobj_t(@plat.sector.soundorg), Ord(sfx_pstart));
        end
        else
        begin
          if res = pastdest then
          begin
            plat.count := plat.wait;
            plat.status := waiting;
            S_StartSound(Pmobj_t(@plat.sector.soundorg), Ord(sfx_pstop));

            case plat._type of
              blazeDWUS,
              downWaitUpStay:
                P_RemoveActivePlat(plat);
              raiseAndChange,
              raiseToNearestAndChange:
                P_RemoveActivePlat(plat);
            end;
          end;
        end;
      end;

    down:
      begin
        res := T_MovePlane(plat.sector, plat.speed, plat.low, false, 0, -1);

        if res = pastdest then
        begin
          plat.count := plat.wait;
          plat.status := waiting;
          S_StartSound(Pmobj_t(@plat.sector.soundorg), Ord(sfx_pstop));
        end;
      end;
    waiting:
      begin
        plat.count := plat.count - 1;
        if not boolval(plat.count) then
        begin
          if plat.sector.floorheight = plat.low then
            plat.status := up
          else
            plat.status := down;
          S_StartSound(Pmobj_t(@plat.sector.soundorg), Ord(sfx_pstart));
        end;
      end;
  end;
end;

//
// Do Platforms
//  "amount" is only used for SOME platforms.
//
function EV_DoPlat(line: Pline_t; _type: plattype_e; amount: integer): integer;
var
  plat: Pplat_t;
  secnum: integer;
  sec: Psector_t;
begin
  result := 0;

  //	Activate all <type> plats that are in_stasis
  if _type = perpetualRaise then
    P_ActivateInStasis(line.tag);

  secnum := P_FindSectorFromLineTag(line, -1);
  while secnum >= 0 do
  begin
    sec := @sectors[secnum];
    secnum := P_FindSectorFromLineTag(line, secnum);

    if sec.specialdata <> nil then
      continue;

    // Find lowest & highest floors around sector
    result := 1;
    plat := Z_Malloc(SizeOf(plat^), PU_LEVSPEC, nil);
    P_AddThinker(@plat.thinker);

    plat._type := _type;
    plat.sector := sec;
    plat.sector.specialdata := plat;
    plat.thinker._function.acp1 := @T_PlatRaise;
    plat.crush := false;
    plat.tag := line.tag;

    case _type of
      raiseToNearestAndChange:
        begin
          plat.speed := PLATSPEED div 2;
          sec.floorpic := sides[line.sidenum[0]].sector.floorpic;
          plat.high := P_FindNextHighestFloor(sec, sec.floorheight);
          plat.wait := 0;
          plat.status := up;
          // NO MORE DAMAGE, IF APPLICABLE
          sec.special := 0;
          S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_stnmov));
        end;

      raiseAndChange:
        begin
          plat.speed := PLATSPEED div 2;
          sec.floorpic := sides[line.sidenum[0]].sector.floorpic;
          plat.high := sec.floorheight + amount * FRACUNIT;
          plat.wait := 0;
          plat.status := up;
          S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_stnmov));
        end;
      downWaitUpStay:
        begin
          plat.speed := PLATSPEED * 4;
          plat.low := P_FindLowestFloorSurrounding(sec);
          if plat.low > sec.floorheight then
            plat.low := sec.floorheight;
          plat.high := sec.floorheight;
          plat.wait := 35 * PLATWAIT; // VJ ??? TICKRATE * PLATWAIT
          plat.status := down;
          S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_pstart));
        end;
      blazeDWUS:
        begin
          plat.speed := PLATSPEED * 8;
          plat.low := P_FindLowestFloorSurrounding(sec);
          if plat.low > sec.floorheight then
            plat.low := sec.floorheight;
          plat.high := sec.floorheight;
          plat.wait := 35 * PLATWAIT; // VJ ??? TICKRATE * PLATWAIT
          plat.status := down;
          S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_pstart));
        end;
      perpetualRaise:
        begin
          plat.speed := PLATSPEED;
          plat.low := P_FindLowestFloorSurrounding(sec);
          if plat.low > sec.floorheight then
            plat.low := sec.floorheight;
          plat.high := P_FindHighestFloorSurrounding(sec);
          if plat.high < sec.floorheight then
            plat.high := sec.floorheight;
          plat.wait := 35 * PLATWAIT; // VJ ??? TICKRATE * PLATWAIT
          plat.status := plat_e(P_Random and 1);
          S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_pstart));
        end;
    end;
    P_AddActivePlat(plat);
  end;
end;

procedure P_ActivateInStasis(tag: integer);
var
  i: integer;
begin
  for i := 0 to MAXPLATS - 1 do
    if boolval(activeplats[i]) and
      (activeplats[i].tag = tag) and
      (activeplats[i].status = in_stasis) then
    begin
      activeplats[i].status := activeplats[i].oldstatus;
      activeplats[i].thinker._function.acp1 := @T_PlatRaise;
    end;
end;


procedure EV_StopPlat(line: Pline_t);
var
  i: integer;
begin
  for i := 0 to MAXPLATS - 1 do
    if boolval(activeplats[i]) and
      (activeplats[i].status <> in_stasis) and
      (activeplats[i].tag = line.tag) then
    begin
      activeplats[i].oldstatus := activeplats[i].status;
      activeplats[i].status := in_stasis;
      activeplats[i].thinker._function.acv := nil;
    end;
end;

procedure P_AddActivePlat(plat: Pplat_t);
var
  i: integer;
begin
  for i := 0 to MAXPLATS - 1 do
    if activeplats[i] = nil then
    begin
      activeplats[i] := plat;
      exit;
    end;

  I_Error('P_AddActivePlat(): no more plats!');
end;

procedure P_RemoveActivePlat(plat: Pplat_t);
var
  i: integer;
begin
  for i := 0 to MAXPLATS - 1 do
    if plat = activeplats[i] then
    begin
      activeplats[i].sector.specialdata := nil;
      P_RemoveThinker(@activeplats[i].thinker);
      activeplats[i] := nil;
      exit;
    end;

  I_Error('P_RemoveActivePlat(): can''t find plat!');
end;


(* Unit p_floor implementation *)


                                                                            

//
// Move a plane (floor or ceiling) and check for crushing
//
function T_MovePlane(sector: Psector_t; speed: fixed_t; dest: fixed_t;
  crush: boolean; floorOrCeiling: integer; direction: integer): result_e;
var
  lastpos: fixed_t;
begin
  case floorOrCeiling of
    0:
      begin
        // FLOOR
        if direction = -1 then
        begin
        // DOWN
          if sector.floorheight - speed < dest then
          begin
            lastpos := sector.floorheight;
            sector.floorheight := dest;
            if P_ChangeSector(sector, crush) then
            begin
              sector.floorheight := lastpos;
              P_ChangeSector(sector, crush);
            //return crushed;
            end;
            result := pastdest;
            exit;
          end
          else
          begin
            lastpos := sector.floorheight;
            sector.floorheight := sector.floorheight - speed;
            if P_ChangeSector(sector, crush) then
            begin
              sector.floorheight := lastpos;
              P_ChangeSector(sector, crush);
              result := crushed;
              exit;
            end;
          end;
        end
        else if direction = 1 then
        begin
        // UP
          if sector.floorheight + speed > dest then
          begin
            lastpos := sector.floorheight;
            sector.floorheight := dest;
            if P_ChangeSector(sector, crush) then
            begin
              sector.floorheight := lastpos;
              P_ChangeSector(sector, crush);
            //return crushed;
            end;
            result := pastdest;
            exit;
          end
          else
          begin
          // COULD GET CRUSHED
            lastpos := sector.floorheight;
            sector.floorheight := sector.floorheight + speed;
            if P_ChangeSector(sector, crush) then
            begin
              if crush then
              begin
                result := crushed;
                exit;
              end;
              sector.floorheight := lastpos;
              P_ChangeSector(sector, crush);
              result := crushed;
              exit;
            end
          end;
        end;
      end;
    1:
      begin
      // CEILING
        if direction = -1 then
        begin
        // DOWN
          if sector.ceilingheight - speed < dest then
          begin
            lastpos := sector.ceilingheight;
            sector.ceilingheight := dest;
            if P_ChangeSector(sector,crush) then
            begin
              sector.ceilingheight := lastpos;
              P_ChangeSector(sector, crush);
            //return crushed;
            end;
            result := pastdest;
            exit;
          end
          else
          begin
          // COULD GET CRUSHED
            lastpos := sector.ceilingheight;
            sector.ceilingheight := sector.ceilingheight - speed;
            if P_ChangeSector(sector, crush) then
            begin
              if crush then
              begin
                result := crushed;
                exit;
              end;
              sector.ceilingheight := lastpos;
              P_ChangeSector(sector, crush);
              result := crushed;
              exit;
            end;
          end;
        end
        else if direction = 1 then
        begin
        // UP
          if sector.ceilingheight + speed > dest then
          begin
            lastpos := sector.ceilingheight;
            sector.ceilingheight := dest;
            if P_ChangeSector(sector, crush) then
            begin
              sector.ceilingheight := lastpos;
              P_ChangeSector(sector, crush);
            //return crushed;
            end;
            result := pastdest;
            exit;
          end
          else
          begin
//            lastpos := sector.ceilingheight;
            sector.ceilingheight := sector.ceilingheight + speed;
            P_ChangeSector(sector, crush);
// UNUSED
//#if 0
//    if (flag == true)
//    {
//        sector->ceilingheight = lastpos;
//        P_ChangeSector(sector,crush);
//        return crushed;
//    }
//#endif
          end;
        end;
      end;
  end;
  result := ok;
end;

//
// MOVE A FLOOR TO IT'S DESTINATION (UP OR DOWN)
//
procedure T_MoveFloor(floor: Pfloormove_t);
var
  res: result_e;
begin
  res := T_MovePlane(floor.sector, floor.speed, floor.floordestheight,
            floor.crush, 0, floor.direction);

  if not boolval(leveltime and 7) then
    S_StartSound(Pmobj_t(@floor.sector.soundorg), Ord(sfx_stnmov));

  if res = pastdest then
  begin
    floor.sector.specialdata := nil;

    if floor.direction = 1 then
    begin
      if floor._type = donutRaise then
      begin
        floor.sector.special := floor.newspecial;
        floor.sector.floorpic := floor.texture;
      end;
    end
    else if floor.direction = -1 then
    begin
      if floor._type = lowerAndChange then
      begin
        floor.sector.special := floor.newspecial;
        floor.sector.floorpic := floor.texture;
      end;
    end;

    P_RemoveThinker(@floor.thinker);

    S_StartSound(Pmobj_t(@floor.sector.soundorg), Ord(sfx_pstop));
  end;
end;

//
// HANDLE FLOOR TYPES
//
function EV_DoFloor(line: Pline_t; floortype: floor_e): integer;
var
  secnum: integer;
  i: integer;
  sec: Psector_t;
  floor: Pfloormove_t;
  minsize: integer;
  side: Pside_t;
begin
  secnum := -1;
  result := 0;
  repeat
    secnum := P_FindSectorFromLineTag(line, secnum);
    if secnum < 0 then
      exit;

    sec := @sectors[secnum];

    // ALREADY MOVING?  IF SO, KEEP GOING...
    if boolval(sec.specialdata) then
      continue;

    // new floor thinker
     result := 1;
    floor := Z_Malloc(SizeOf(floormove_t), PU_LEVSPEC, nil);
    P_AddThinker(@floor.thinker);
    sec.specialdata := floor;
    floor.thinker._function.acp1 := @T_MoveFloor;
    floor._type := floortype;
    floor.crush := false;

    case floortype of
      lowerFloor:
        begin
          floor.direction := -1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := P_FindHighestFloorSurrounding(sec);
        end;
      lowerFloorToLowest:
        begin
          floor.direction := -1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := P_FindLowestFloorSurrounding(sec);
        end;
      turboLower:
        begin
          floor.direction := -1;
          floor.sector := sec;
          floor.speed := FLOORSPEED * 4;
          floor.floordestheight :=
          P_FindHighestFloorSurrounding(sec);
          if floor.floordestheight <> sec.floorheight then
            floor.floordestheight := floor.floordestheight + 8 * FRACUNIT;
        end;
      raiseFloorCrush,
      raiseFloor:
        begin
          if floortype = raiseFloorCrush then
            floor.crush := true;
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := P_FindLowestCeilingSurrounding(sec);
          if floor.floordestheight > sec.ceilingheight then
            floor.floordestheight := sec.ceilingheight;
          if floortype = raiseFloorCrush then // if floor.crush then
            floor.floordestheight := floor.floordestheight - 8 * FRACUNIT;
        end;
      raiseFloorTurbo:
        begin
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED * 4;
          floor.floordestheight := P_FindNextHighestFloor(sec, sec.floorheight);
        end;
      raiseFloorToNearest:
        begin
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := P_FindNextHighestFloor(sec, sec.floorheight);
        end;
      raiseFloor24:
        begin
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := floor.sector.floorheight + 24 * FRACUNIT;
        end;
      raiseFloor512:
        begin
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := floor.sector.floorheight + 512 * FRACUNIT;
        end;
      raiseFloor24AndChange:
        begin
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := floor.sector.floorheight + 24 * FRACUNIT;
          sec.floorpic := line.frontsector.floorpic;
          sec.special := line.frontsector.special;
        end;
      raiseToTexture:
        begin
          minsize := MAXINT;
          floor.direction := 1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          for i := 0 to sec.linecount - 1 do
          begin
            if boolval(twoSided(secnum, i)) then
            begin
              side := getSide(secnum, i, 0);
              if side.bottomtexture >= 0 then
                if textureheight[side.bottomtexture] < minsize then
                  minsize := textureheight[side.bottomtexture];
              side := getSide(secnum, i, 1);
              if side.bottomtexture >= 0 then
                if textureheight[side.bottomtexture] < minsize then
                  minsize := textureheight[side.bottomtexture];
            end;
          end;
          floor.floordestheight := floor.sector.floorheight + minsize;
        end;
      lowerAndChange:
        begin
          floor.direction := -1;
          floor.sector := sec;
          floor.speed := FLOORSPEED;
          floor.floordestheight := P_FindLowestFloorSurrounding(sec);
          floor.texture := sec.floorpic;
          for i := 0 to sec.linecount - 1 do
          begin
            if boolval(twoSided(secnum, i)) then
            begin
              side := getSide(secnum, i, 0);
              if pOperation(side.sector, @sectors[0], '-', SizeOf(side.sector^)) = secnum then
              begin
                sec := getSector(secnum, i, 1);
                if sec.floorheight = floor.floordestheight then
                begin
                  floor.texture := sec.floorpic;
                  floor.newspecial := sec.special;
                  break;
                end;
              end
              else
              begin
                sec := getSector(secnum, i, 0);
                if sec.floorheight = floor.floordestheight then
                begin
                  floor.texture := sec.floorpic;
                  floor.newspecial := sec.special;
                  break;
                end;
              end;
            end;
          end;
        end;
    end;
  until secnum < 0;
end;

//
// BUILD A STAIRCASE!
//
function EV_BuildStairs(line: Pline_t; _type: stair_e): integer;
var
  secnum: integer;
  height: integer;
  i: integer;
  newsecnum: integer;
  texture: integer;
  ok: boolean;
  sec: Psector_t;
  tsec: Psector_t;
  floor: Pfloormove_t;
  stairsize: fixed_t;
  speed: fixed_t;
begin
  secnum := -1;
  result := 0;

  repeat
    secnum := P_FindSectorFromLineTag(line, secnum);
    if secnum < 0 then
      exit;

    sec := @sectors[secnum];

    // ALREADY MOVING?  IF SO, KEEP GOING...
    if sec.specialdata <> nil then
      continue;

    // new floor thinker
    result := 1;
    floor := Z_Malloc(SizeOf(floormove_t), PU_LEVSPEC, nil);
    P_AddThinker(@floor.thinker);
    sec.specialdata := floor;
    floor.thinker._function.acp1 := @T_MoveFloor;
    floor.direction := 1;
    floor.sector := sec;
    case _type of
      build8:
        begin
          speed := FLOORSPEED div 4;
          stairsize := 8 * FRACUNIT;
        end;
      turbo16:
        begin
          speed := FLOORSPEED * 4;
          stairsize := 16 * FRACUNIT;
        end;
      else
      begin
        speed := 0;
        stairsize := 0;
      end;
    end;
    floor.speed := speed;

    height := sec.floorheight + stairsize;
    floor.floordestheight := height;

    texture := sec.floorpic;

    // Find next sector to raise
    // 1.  Find 2-sided line with same sector side[0]
    // 2.  Other side is the next sector to raise
    repeat
      ok := false;
      for i := 0 to sec.linecount - 1 do
      begin
        if not boolval(sec.lines[i].flags and ML_TWOSIDED) then
          continue;

        tsec := sec.lines[i].frontsector;
        newsecnum := pOperation (tsec, @sectors[0], '-', SizeOf(sector_t));
//        (integer(tsec) - integer(@sectors[0])) div ;

        if secnum <> newsecnum then
          continue;

        tsec := sec.lines[i].backsector;
        newsecnum := pOperation(tsec, @sectors[0], '-', SizeOf(sector_t));
        //integer(tsec) - integer(sectors)) div SizeOf(sector_t);

        if tsec.floorpic <> texture then
          continue;

        height := height + stairsize;

        if tsec.specialdata <> nil then
          continue;

        sec := tsec;
        secnum := newsecnum;
        floor := Z_Malloc(SizeOf(floormove_t), PU_LEVSPEC, nil);

        P_AddThinker(@floor.thinker);

        sec.specialdata := floor;
        floor.thinker._function.acp1 := @T_MoveFloor;
        floor.direction := 1;
        floor.sector := sec;
        floor.speed := speed;
        floor.floordestheight := height;
        ok := true;
        break;
      end;
    until not ok;
  until secnum < 0;
end;


(* Unit p_ceilng implementation *)


                                                                  

//
// Add an active ceiling
//
procedure P_AddActiveCeiling(c: Pceiling_t);
var
  i: integer;
begin
  for i := 0 to MAXCEILINGS - 1 do
    if activeceilings[i] = nil then
    begin
      activeceilings[i] := c;
      exit;
    end;
end;

//
// Remove a ceiling's thinker
//
procedure P_RemoveActiveCeiling(c: Pceiling_t);
var
  i: integer;
begin
  for i := 0 to MAXCEILINGS - 1 do
    if activeceilings[i] = c then
    begin
      activeceilings[i].sector.specialdata := nil;
      P_RemoveThinker(@activeceilings[i].thinker);
      activeceilings[i] := nil;
      exit;
    end;
end;

//
// Restart a ceiling that's in-stasis
//
procedure P_ActivateInStasisCeiling(line: Pline_t);
var
  i: integer;
begin
  for i := 0 to MAXCEILINGS - 1 do
    if (activeceilings[i] <> nil) and
       (activeceilings[i].tag = line.tag) and
       (activeceilings[i].direction = 0) then
    begin
      activeceilings[i].direction := activeceilings[i].olddirection;
      activeceilings[i].thinker._function.acp1 := @T_MoveCeiling;
    end;
end;

//
// T_MoveCeiling
//

procedure T_MoveCeiling(ceiling: Pceiling_t);
var
  res: result_e;
begin
  case ceiling.direction of
    0:
    // IN STASIS
      begin
      end;
    1:
    // UP
      begin
        res := T_MovePlane(ceiling.sector,
          ceiling.speed,
          ceiling.topheight,
          false, 1, ceiling.direction);

        if leveltime and 7 = 0 then
        begin
          if ceiling._type <> silentCrushAndRaise then
            S_StartSound(Pmobj_t(@ceiling.sector.soundorg),
              Ord(sfx_stnmov));
        end;

        if res = pastdest then
        begin
          case ceiling._type of
            raiseToHighest:
              P_RemoveActiveCeiling(ceiling);
            silentCrushAndRaise:
              begin
                S_StartSound(Pmobj_t(@ceiling.sector.soundorg), Ord(sfx_pstop));
                ceiling.direction := -1;
              end;
            fastCrushAndRaise,
            crushAndRaise:
              ceiling.direction := -1;
          end;
        end;
      end;
   -1:
    // DOWN
      begin
        res := T_MovePlane(ceiling.sector,
          ceiling.speed,
          ceiling.bottomheight,
          ceiling.crush, 1, ceiling.direction);

        if leveltime and 7 = 0 then
        begin
          if ceiling._type <> silentCrushAndRaise then
            S_StartSound(Pmobj_t(@ceiling.sector.soundorg), Ord(sfx_stnmov));
        end;

        if res = pastdest then
        begin
          case ceiling._type of
            silentCrushAndRaise:
              begin
                S_StartSound(Pmobj_t(@ceiling.sector.soundorg), Ord(sfx_pstop));
                ceiling.speed := CEILSPEED;
                ceiling.direction := 1;
              end;
            crushAndRaise:
              begin
                ceiling.speed := CEILSPEED;
                ceiling.direction := 1;
              end;
            fastCrushAndRaise:
              begin
                ceiling.direction := 1;
              end;
            lowerAndCrush,
            lowerToFloor:
              P_RemoveActiveCeiling(ceiling);
          end;
        end
        else // ( res <> pastdest )
        begin
          if res = crushed then
          begin
            case ceiling._type of
              silentCrushAndRaise,
              crushAndRaise,
              lowerAndCrush:
                ceiling.speed := CEILSPEED div 8;
            end;
          end;
        end;
      end;
  end;
end;

//
// EV_DoCeiling
// Move a ceiling up/down and all around!
//
function EV_DoCeiling(line: Pline_t; _type: ceiling_e): integer;
var
  initial: boolean;
  secnum: integer;
  sec: Psector_t;
  ceiling: Pceiling_t;
begin
  secnum := -1;
  result := 0;

  // Reactivate in-stasis ceilings...for certain types.
  case _type of
    fastCrushAndRaise,
    silentCrushAndRaise,
    crushAndRaise:
      P_ActivateInStasisCeiling(line);
  end;

  initial := true;
  while (secnum >= 0) or initial do
  begin
    initial := false;
    secnum := P_FindSectorFromLineTag(line,secnum);
    if secnum < 0 then
      break;

    sec := @sectors[secnum];
    if sec.specialdata <> nil then
      continue;

    // new door thinker
    result := 1;
    ceiling := Z_Malloc(sizeof(ceiling^), PU_LEVSPEC, nil);
    P_AddThinker(@ceiling.thinker);
    sec.specialdata := ceiling;
    ceiling.thinker._function.acp1 := @T_MoveCeiling;
    ceiling.sector := sec;
    ceiling.crush := false;

    case _type of
      fastCrushAndRaise:
        begin
          ceiling.crush := true;
          ceiling.topheight := sec.ceilingheight;
          ceiling.bottomheight := sec.floorheight + (8 * FRACUNIT);
          ceiling.direction := -1;
          ceiling.speed := CEILSPEED * 2;
        end;

      silentCrushAndRaise,
      crushAndRaise:
        begin
          ceiling.crush := true;
          ceiling.topheight := sec.ceilingheight;
          ceiling.bottomheight := sec.floorheight;
          if _type <> lowerToFloor then
            ceiling.bottomheight := ceiling.bottomheight + 8 * FRACUNIT;
          ceiling.direction := -1;
          ceiling.speed := CEILSPEED;
        end;
      lowerAndCrush,
      lowerToFloor:
        begin
          ceiling.bottomheight := sec.floorheight;
          if _type <> lowerToFloor then
            ceiling.bottomheight := ceiling.bottomheight + 8 * FRACUNIT;
          ceiling.direction := -1;
          ceiling.speed := CEILSPEED;
        end;

      raiseToHighest:
        begin
          ceiling.topheight := P_FindHighestCeilingSurrounding(sec);
          ceiling.direction := 1;
          ceiling.speed := CEILSPEED;
        end;
    end;

    ceiling.tag := sec.tag;
    ceiling._type := _type;
    P_AddActiveCeiling(ceiling);
  end;
end;

//
// EV_CeilingCrushStop
// Stop a ceiling from crushing!
//
function EV_CeilingCrushStop(line: Pline_t): integer;
var
  i: integer;
begin
  result := 0;
  for i := 0 to MAXCEILINGS - 1 do
    if (activeceilings[i] <> nil) and
       (activeceilings[i].tag = line.tag) and
       (activeceilings[i].direction <> 0) then
    begin
      activeceilings[i].olddirection := activeceilings[i].direction;
      activeceilings[i].thinker._function.acv := nil;
      activeceilings[i].direction := 0; // in-stasis
      result := 1;
    end;
end;


(* Unit p_lights implementation *)


                                   

//
// FIRELIGHT FLICKER
//

//
// T_FireFlicker
//
procedure T_FireFlicker(flick: Pfireflicker_t);
var
  amount: integer;
begin
  flick.count := flick.count - 1;
  if flick.count > 0  then
    exit;

  amount := (P_Random and 3) * 16;

  if flick.sector.lightlevel - amount < flick.minlight then
    flick.sector.lightlevel := flick.minlight
  else
    flick.sector.lightlevel := flick.maxlight - amount;

  flick.count := 4;
end;

//
// P_SpawnFireFlicker
//
procedure P_SpawnFireFlicker(sector: Psector_t);
var
  flick: Pfireflicker_t;
begin
  // Note that we are resetting sector attributes.
  // Nothing special about it during gameplay.
  sector.special := 0;

  flick := Z_Malloc(SizeOf(flick^), PU_LEVSPEC, nil);

  P_AddThinker(@flick.thinker);

  flick.thinker._function.acp1 := @T_FireFlicker;
  flick.sector := sector;
  flick.maxlight := sector.lightlevel;
  flick.minlight := P_FindMinSurroundingLight(sector, sector.lightlevel) + 16;
  flick.count := 4;
end;

//
// BROKEN LIGHT FLASHING
//


//
// T_LightFlash
// Do flashing lights.
//
procedure T_LightFlash(flash: Plightflash_t);
begin
  flash.count := flash.count - 1;
  if flash.count > 0 then
    exit;

  if flash.sector.lightlevel = flash.maxlight then
  begin
    flash.sector.lightlevel := flash.minlight;
    flash.count := (P_Random and flash.mintime) + 1;
  end
  else
  begin
    flash.sector.lightlevel := flash.maxlight;
    flash.count := (P_Random and flash.maxtime) + 1;
  end;
end;

//
// P_SpawnLightFlash
// After the map has been loaded, scan each sector
// for specials that spawn thinkers
//
procedure P_SpawnLightFlash(sector: Psector_t);
var
  flash: Plightflash_t;
begin
  // nothing special about it during gameplay
  sector.special := 0;

  flash := Z_Malloc(SizeOf(flash^), PU_LEVSPEC, nil);

  P_AddThinker(@flash.thinker);

  flash.thinker._function.acp1 := @T_LightFlash;
  flash.sector := sector;
  flash.maxlight := sector.lightlevel;

  flash.minlight := P_FindMinSurroundingLight(sector, sector.lightlevel);
  flash.maxtime := 64;
  flash.mintime := 7;
  flash.count := (P_Random and flash.maxtime) + 1;
end;

//
// STROBE LIGHT FLASHING
//


//
// T_StrobeFlash
//
procedure T_StrobeFlash(flash: Pstrobe_t);
begin
  flash.count := flash.count - 1;
  if flash.count > 0 then
    exit;

  if flash.sector.lightlevel = flash.minlight then
  begin
    flash.sector.lightlevel := flash.maxlight;
    flash.count := flash.brighttime;
  end
  else
  begin
    flash.sector.lightlevel := flash.minlight;
    flash.count := flash.darktime;
  end;
end;

//
// P_SpawnStrobeFlash
// After the map has been loaded, scan each sector
// for specials that spawn thinkers
//
procedure P_SpawnStrobeFlash(sector: Psector_t; fastOrSlow, inSync: integer);
var
  flash: Pstrobe_t;
begin
  flash := Z_Malloc(SizeOf(flash^), PU_LEVSPEC, nil);

  P_AddThinker(@flash.thinker);

  flash.sector := sector;
  flash.darktime := fastOrSlow;
  flash.brighttime := STROBEBRIGHT;
  flash.thinker._function.acp1 := @T_StrobeFlash;
  flash.maxlight := sector.lightlevel;
  flash.minlight := P_FindMinSurroundingLight(sector, sector.lightlevel);

  if flash.minlight = flash.maxlight then
    flash.minlight := 0;

  // nothing special about it during gameplay
  sector.special := 0;

  if not boolval(inSync) then
    flash.count := (P_Random and 7) + 1
  else
    flash.count := 1;
end;

//
// Start strobing lights (usually from a trigger)
//
procedure EV_StartLightStrobing(line: Pline_t);
var
  secnum: integer;
  sec: Psector_t;
begin
  secnum := -1;
  repeat
    secnum := P_FindSectorFromLineTag(line, secnum);

    if secnum >= 0 then
    begin
      sec := @sectors[secnum];
      if sec.specialdata = nil then
        P_SpawnStrobeFlash(sec, SLOWDARK, 0);
    end;
  until secnum < 0;
end;

//
// TURN LINE'S TAG LIGHTS OFF
//
procedure EV_TurnTagLightsOff(line: Pline_t);
var
  i: integer;
  j: integer;
  min: integer;
  sector: Psector_t;
  tsec: Psector_t;
  templine: Pline_t;
begin
  for i := 0 to numsectors - 1 do
  begin
    sector := @sectors[i];
    if sector.tag = line.tag then
    begin
      min := sector.lightlevel;
      for j := 0 to sector.linecount - 1 do
      begin
        templine := sector.lines[j];
        tsec := getNextSector(templine, sector);
        if boolval(tsec) then
        begin
          if tsec.lightlevel < min then
            min := tsec.lightlevel;
        end;
      end;
	    sector.lightlevel := min;
    end;
  end;
end;

//
// TURN LINE'S TAG LIGHTS ON
//
procedure EV_LightTurnOn(line: Pline_t; bright: integer);
var
  i: integer;
  j: integer;
  sector: Psector_t;
  temp: Psector_t;
  templine: Pline_t;
begin
  for i := 0 to numsectors - 1 do
  begin
    sector := @sectors[i];
    if sector.tag = line.tag then
    begin
      // bright = 0 means to search
      // for highest light level
      // surrounding sector
	    if not boolval(bright) then
      begin
        for j := 0 to sector.linecount - 1 do
        begin
          templine := sector.lines[j];
          temp := getNextSector(templine, sector);
          if boolval(temp) then
          begin
            if temp.lightlevel > bright then
              bright := temp.lightlevel;
          end;
        end;
      end;
	    sector.lightlevel := bright;
    end;
  end;
end;

//
// Spawn glowing light
//
procedure T_Glow(g: Pglow_t);
begin
  case g.direction of
   -1:
      begin
        // DOWN
        g.sector.lightlevel := g.sector.lightlevel - GLOWSPEED;
        if g.sector.lightlevel <= g.minlight then
        begin
          g.sector.lightlevel := g.sector.lightlevel + GLOWSPEED;
          g.direction := 1;
        end;
      end;
    1:
      begin
        // UP
        g.sector.lightlevel := g.sector.lightlevel + GLOWSPEED;
        if g.sector.lightlevel >= g.maxlight then
        begin
          g.sector.lightlevel := g.sector.lightlevel - GLOWSPEED;
          g.direction := -1;
        end;
      end;
  end;
end;

procedure P_SpawnGlowingLight(sector: Psector_t);
var
  g: Pglow_t;
begin
  g := Z_Malloc(SizeOf(g^), PU_LEVSPEC, nil);

  P_AddThinker(@g.thinker);

  g.sector := sector;
  g.minlight := P_FindMinSurroundingLight(sector, sector.lightlevel);
  g.maxlight := sector.lightlevel;
  g.thinker._function.acp1 := @T_Glow;
  g.direction := -1;

  sector.special := 0;
end;


(* Unit p_doors implementation *)


                                                                     

//
// VERTICAL DOORS
//

//
// T_VerticalDoor
//
procedure T_VerticalDoor(door: Pvldoor_t);
var
  res: result_e;
begin
  case door.direction of
    0:
      begin
  // WAITING
        door := Pvldoor_t(integer(door) - SizeOf(vldoor_t));
        if door.topcountdown = 0 then
        begin
          case door._type of
            blazeRaise:
              begin
                door.direction := -1; // time to go back down
                S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_bdcls));
              end;
            normal:
              begin
                door.direction := -1; // time to go back down
                S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_dorcls));
              end;
            close30ThenOpen:
              begin
                door.direction := 1;
                S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_doropn));
              end;
          end;
        end;
      end;
    2:
      begin
  //  INITIAL WAIT
        door := Pvldoor_t(integer(door) - SizeOf(vldoor_t));
        if door.topcountdown = 0 then
        begin
          case door._type of
            raiseIn5Mins:
              begin
                door.direction := 1;
                door._type := normal;
                S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_doropn));
              end;
          end;
        end;
      end;

   -1:
      begin
  // DOWN
        res := T_MovePlane(door.sector, door.speed, door.sector.floorheight,
                  false, 1, door.direction);
        if res = pastdest then
        begin
          case door._type of
            blazeRaise,
            blazeClose:
              begin
                door.sector.specialdata := nil;
                P_RemoveThinker(@door.thinker); // unlink and free
                S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_bdcls));
              end;
            normal,
            close:
              begin
                door.sector.specialdata := nil;
                P_RemoveThinker(@door.thinker);  // unlink and free
              end;
            close30ThenOpen:
              begin
                door.direction := 0;
                door.topcountdown := 35 * 30; // ticrate * 30 ?? VJ
              end;
          end;
        end
        else if res = crushed then
        begin
          case door._type of
            blazeClose,
            close:    // DO NOT GO BACK UP!
              begin
              end;
          else
            begin
              door.direction := 1;
              S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_doropn));
            end;
          end;
        end;
      end;
    1:
      begin
  // UP
        res := T_MovePlane(door.sector, door.speed, door.topheight,
                  false, 1, door.direction);
        if res = pastdest then
        begin
          case door._type of
            blazeRaise,
            normal:
              begin
                door.direction := 0; // wait at top
                door.topcountdown := door.topwait;
              end;
            close30ThenOpen,
            blazeOpen,
            open:
              begin
                door.sector.specialdata := nil;
                P_RemoveThinker(@door.thinker); // unlink and free
              end;
          end;
        end;
      end;
  end;
end;

//
// EV_DoLockedDoor
// Move a locked door up/down
//

function EV_DoLockedDoor(line: Pline_t; _type: vldoor_e; thing: Pmobj_t): integer;
var
  p: Pplayer_t;
begin
  p := thing.player;

  if p = nil then
  begin
    result := 0;
    exit;
  end;

  case line.special of
    99,  // Blue Lock
   133:
      begin
        if (not p.cards[Ord(it_bluecard)]) and
           (not p.cards[Ord(it_blueskull)]) then
        begin
          p._message := PD_BLUEO;
          S_StartSound(nil, Ord(sfx_oof));
          result := 0;
          exit;
        end;
      end;

   134, // Red Lock
   135:
      begin
        if (not p.cards[Ord(it_redcard)]) and
           (not p.cards[Ord(it_redskull)]) then
        begin
          p._message := PD_REDO;
          S_StartSound(nil, Ord(sfx_oof));
          result := 0;
          exit;
        end;
      end;

   136,  // Yellow Lock
   137:
      begin
        if (not p.cards[Ord(it_yellowcard)]) and
           (not p.cards[Ord(it_yellowskull)]) then
        begin
          p._message := PD_YELLOWO;
          S_StartSound(nil, Ord(sfx_oof));
          result := 0;
          exit;
        end;
      end;
  end;

  result := EV_DoDoor(line, _type);
end;

function EV_DoDoor(line: Pline_t; _type: vldoor_e): integer;
var
  initial: boolean;
  secnum: integer;
  sec: Psector_t;
  door: Pvldoor_t;
begin
  secnum := -1;
  result := 0;

  initial := true;
  while (secnum >= 0) or initial do
  begin
    initial := false;
    secnum := P_FindSectorFromLineTag(line, secnum);
    if secnum < 0 then
      break;

    sec := @sectors[secnum];
    if sec.specialdata <> nil then
      continue;

    // new door thinker
    result := 1;
    door := Z_Malloc(SizeOf(door^), PU_LEVSPEC, nil);
    P_AddThinker(@door.thinker);
    sec.specialdata := door;

    door.thinker._function.acp1 := @T_VerticalDoor;
    door.sector := sec;
    door._type := _type;
    door.topwait := VDOORWAIT;
    door.speed := VDOORSPEED;

    case _type of
      blazeClose:
        begin
          door.topheight := P_FindLowestCeilingSurrounding(sec);
          door.topheight := door.topheight - 4 * FRACUNIT;
          door.direction := -1;
          door.speed := VDOORSPEED * 4;
          S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_bdcls));
        end;

      close:
        begin
          door.topheight := P_FindLowestCeilingSurrounding(sec);
          door.topheight := door.topheight - 4 * FRACUNIT;
          door.direction := -1;
          S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_dorcls));
        end;

      close30ThenOpen:
        begin
          door.topheight := sec.ceilingheight;
          door.direction := -1;
          S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_dorcls));
        end;

      blazeRaise,
      blazeOpen:
        begin
          door.direction := 1;
          door.topheight := P_FindLowestCeilingSurrounding(sec);
          door.topheight := door.topheight - 4 * FRACUNIT;
          door.speed := VDOORSPEED * 4;
          if door.topheight <> sec.ceilingheight then
            S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_bdopn));
        end;

      normal,
      open:
        begin
          door.direction := 1;
          door.topheight := P_FindLowestCeilingSurrounding(sec);
          door.topheight := door.topheight - 4 * FRACUNIT;
          if door.topheight <> sec.ceilingheight then
            S_StartSound(Pmobj_t(@door.sector.soundorg), Ord(sfx_doropn));
        end;
    end;
  end;
end;

//
// EV_VerticalDoor : open a door manually, no tag value
//
procedure EV_VerticalDoor(line: Pline_t; thing: Pmobj_t);
var
  player: Pplayer_t;
  sec: Psector_t;
  door: Pvldoor_t;
  side: integer;
begin
  side := 0;  // only front sides can be used

  // Check for locks
  player := thing.player;

  case line.special of
    26, // Blue Lock
    32:
      begin
        if player = nil then
          exit;

        if (not player.cards[Ord(it_bluecard)]) and
           (not player.cards[Ord(it_blueskull)]) then
        begin
          player._message := PD_BLUEK;
          S_StartSound(nil, Ord(sfx_oof));
          exit;
        end;
      end;
    27, // Yellow Lock
    34:
      begin
        if player = nil then
          exit;

        if (not player.cards[Ord(it_yellowcard)]) and
           (not player.cards[Ord(it_yellowskull)]) then
        begin
          player._message := PD_YELLOWK;
          S_StartSound(nil, Ord(sfx_oof));
          exit;
        end;
      end;
    28, // Red Lock
    33:
      begin
        if player = nil then
          exit;

        if (not player.cards[Ord(it_redcard)]) and
           (not player.cards[Ord(it_redskull)]) then
        begin
          player._message := PD_REDK;
          S_StartSound(nil, Ord(sfx_oof));
          exit;
        end;
      end;
  end;

  // if the sector has an active thinker, use it
  sec := sides[line.sidenum[side xor 1]].sector;

  if sec.specialdata <> nil then
  begin
    door := sec.specialdata;
    case line.special of
       1, // ONLY FOR "RAISE" DOORS, NOT "OPEN"s
      26,
      27,
      28,
     117:
        begin
          if door.direction = -1 then
            door.direction := 1 // go back up
          else
          begin
            if thing.player = nil then
              exit; // JDC: bad guys never close doors

            door.direction := -1; // start going down immediately
          end;
          exit;
        end;
    end;
  end;

  // for proper sound
  case line.special of
   117, // BLAZING DOOR RAISE
   118: // BLAZING DOOR OPEN
      S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_bdopn));
     1, // NORMAL DOOR SOUND
    31:
      S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_doropn));
  else // LOCKED DOOR SOUND
    S_StartSound(Pmobj_t(@sec.soundorg), Ord(sfx_doropn));
  end;

  // new door thinker
  door := Z_Malloc(SizeOf(door^), PU_LEVSPEC, nil);
  P_AddThinker(@door.thinker);
  sec.specialdata := door;
  door.thinker._function.acp1 := @T_VerticalDoor;
  door.sector := sec;
  door.direction := 1;
  door.speed := VDOORSPEED;
  door.topwait := VDOORWAIT;

  case line.special of
     1,
    26,
    27,
    28:
      door._type := normal;
    31,
    32,
    33,
    34:
      begin
        door._type := open;
        line.special := 0;
      end;
   117: // blazing door raise
      begin
        door._type := blazeRaise;
        door.speed := VDOORSPEED * 4;
      end;
   118: // blazing door open
      begin
        door._type := blazeOpen;
        line.special := 0;
        door.speed := VDOORSPEED * 4;
      end;
  end;

  // find the top and bottom of the movement range
  door.topheight := P_FindLowestCeilingSurrounding(sec);
  door.topheight := door.topheight - 4 * FRACUNIT;
end;

//
// Spawn a door that closes after 30 seconds
//
procedure P_SpawnDoorCloseIn30(sec: Psector_t);
var
  door: Pvldoor_t;
begin
  door := Z_Malloc(SizeOf(door^), PU_LEVSPEC, nil);

  P_AddThinker(@door.thinker);

  sec.specialdata := door;
  sec.special := 0;

  door.thinker._function.acp1 := @T_VerticalDoor;
  door.sector := sec;
  door.direction := 0;
  door._type := normal;
  door.speed := VDOORSPEED;
  door.topcountdown := 30 * 35; // 30 * tickrate ???? VJ
end;

//
// Spawn a door that opens after 5 minutes
//
procedure P_SpawnDoorRaiseIn5Mins(sec: Psector_t; secnum: integer);
var
  door: Pvldoor_t;
begin
  door := Z_Malloc(SizeOf(door^), PU_LEVSPEC, nil);

  P_AddThinker(@door.thinker);

  sec.specialdata := door;
  sec.special := 0;

  door.thinker._function.acp1 := @T_VerticalDoor;
  door.sector := sec;
  door.direction := 2;
  door._type := raiseIn5Mins;
  door.speed := VDOORSPEED;
  door.topheight := P_FindLowestCeilingSurrounding(sec);
  door.topheight := door.topheight - 4 * FRACUNIT;
  door.topwait := VDOORWAIT;
  door.topcountdown := 5 * 60 * 35; // 5 * 60 * tickrate ???? VJ
end;



(* Unit p_switch implementation *)


                                                                                                                                                                                             

const
  alphSwitchList: array[0..40] of switchlist_t = (
    // Doom shareware episode 1 switches
    (name1: 'SW1BRCOM'; name2: 'SW2BRCOM'; episode: 1),
    (name1: 'SW1BRN1';  name2: 'SW2BRN1';  episode: 1),
    (name1: 'SW1BRN2';  name2: 'SW2BRN2';  episode: 1),
    (name1: 'SW1BRNGN'; name2: 'SW2BRNGN'; episode: 1),
    (name1: 'SW1BROWN'; name2: 'SW2BROWN'; episode: 1),
    (name1: 'SW1COMM';  name2: 'SW2COMM';  episode: 1),
    (name1: 'SW1COMP';  name2: 'SW2COMP';  episode: 1),
    (name1: 'SW1DIRT';  name2: 'SW2DIRT';  episode: 1),
    (name1: 'SW1EXIT';  name2: 'SW2EXIT';  episode: 1),
    (name1: 'SW1GRAY';  name2: 'SW2GRAY';  episode: 1),
    (name1: 'SW1GRAY1'; name2: 'SW2GRAY1'; episode: 1),
    (name1: 'SW1METAL'; name2: 'SW2METAL'; episode: 1),
    (name1: 'SW1PIPE';  name2: 'SW2PIPE';  episode: 1),
    (name1: 'SW1SLAD';  name2: 'SW2SLAD';  episode: 1),
    (name1: 'SW1STARG'; name2: 'SW2STARG'; episode: 1),
    (name1: 'SW1STON1'; name2: 'SW2STON1'; episode: 1),
    (name1: 'SW1STON2'; name2: 'SW2STON2'; episode: 1),
    (name1: 'SW1STONE'; name2: 'SW2STONE'; episode: 1),
    (name1: 'SW1STRTN'; name2: 'SW2STRTN'; episode: 1),

    // Doom registered episodes 2&3 switches
    (name1: 'SW1BLUE';  name2: 'SW2BLUE';  episode: 2),
    (name1: 'SW1CMT';   name2: 'SW2CMT';   episode: 2),
    (name1: 'SW1GARG';  name2: 'SW2GARG';  episode: 2),
    (name1: 'SW1GSTON'; name2: 'SW2GSTON'; episode: 2),
    (name1: 'SW1HOT';   name2: 'SW2HOT';   episode: 2),
    (name1: 'SW1LION';  name2: 'SW2LION';  episode: 2),
    (name1: 'SW1SATYR'; name2: 'SW2SATYR'; episode: 2),
    (name1: 'SW1SKIN';  name2: 'SW2SKIN';  episode: 2),
    (name1: 'SW1VINE';  name2: 'SW2VINE';  episode: 2),
    (name1: 'SW1WOOD';  name2: 'SW2WOOD';  episode: 2),

    // Doom II switches
    (name1: 'SW1PANEL'; name2: 'SW2PANEL'; episode: 3),
    (name1: 'SW1ROCK';  name2: 'SW2ROCK';  episode: 3),
    (name1: 'SW1MET2';  name2: 'SW2MET2';  episode: 3),
    (name1: 'SW1WDMET'; name2: 'SW2WDMET'; episode: 3),
    (name1: 'SW1BRIK';  name2: 'SW2BRIK';  episode: 3),
    (name1: 'SW1MOD1';  name2: 'SW2MOD1';  episode: 3),
    (name1: 'SW1ZIM';   name2: 'SW2ZIM';   episode: 3),
    (name1: 'SW1STON6'; name2: 'SW2STON6'; episode: 3),
    (name1: 'SW1TEK';   name2: 'SW2TEK';   episode: 3),
    (name1: 'SW1MARB';  name2: 'SW2MARB';  episode: 3),
    (name1: 'SW1SKULL'; name2: 'SW2SKULL'; episode: 3),

    (name1: '';         name2: '';         episode: 0)
  );

var
  switchlist: array[0..2 * MAXSWITCHES - 1] of integer;
  numswitches: integer;

//
// P_InitSwitchList
// Only called at game initialization.
//
procedure P_InitSwitchList;
var
  i: integer;
  index: integer;
  episode: integer;
begin

  if gamemode = registered then
    episode := 2
  else if gamemode = commercial then
    episode := 3
  else
    episode := 1;

  index := 0;
  for i := 0 to MAXSWITCHES - 1 do
  begin
    if not boolval(alphSwitchList[i].episode) then
    begin
      numswitches := index div 2;
      switchlist[index] := -1;
      break;
    end;

    if alphSwitchList[i].episode <= episode then
    begin
(**
#if 0	// UNUSED - debug?
	    int		value;

	    if (R_CheckTextureNumForName(alphSwitchList[i].name1) < 0)
	    {
		I_Error("Can't find switch texture '%s'!",
			alphSwitchList[i].name1);
		continue;
	    }

	    value = R_TextureNumForName(alphSwitchList[i].name1);
#endif
***)
      switchlist[index] := R_TextureNumForName(alphSwitchList[i].name1);
      inc(index);
      switchlist[index] := R_TextureNumForName(alphSwitchList[i].name2);
      inc(index);
    end;
  end;
end;

//
// Start a button counting down till it turns off.
//
procedure P_StartButton(line: Pline_t; w: bwhere_e; texture: integer; time: integer);
var
  i: integer;
begin
  // See if button is already pressed
  for i := 0 to MAXBUTTONS - 1 do
    if boolval(buttonlist[i].btimer) and (buttonlist[i].line = line) then
      exit;

  for i := 0 to MAXBUTTONS - 1 do
  begin
    if not boolval(buttonlist[i].btimer) then
    begin
      buttonlist[i].line := line;
      buttonlist[i].where := w;
      buttonlist[i].btexture := texture;
      buttonlist[i].btimer := time;
      buttonlist[i].soundorg := Pmobj_t(@line.frontsector.soundorg);
      exit;
    end;
  end;

  I_Error('P_StartButton(): no button slots left!');
end;

//
// Function that changes wall texture.
// Tell it if switch is ok to use again (1=yes, it's a button).
//
procedure P_ChangeSwitchTexture(line: Pline_t; useAgain: integer);
var
  texTop: integer;
  texMid: integer;
  texBot: integer;
  i: integer;
  sound: integer;
begin
  if not boolval(useAgain) then
    line.special := 0;

  texTop := sides[line.sidenum[0]].toptexture;
  texMid := sides[line.sidenum[0]].midtexture;
  texBot := sides[line.sidenum[0]].bottomtexture;

  sound := Ord(sfx_swtchn);

  // EXIT SWITCH?
  if line.special = 11 then
    sound := Ord(sfx_swtchx);

  for i := 0 to numswitches * 2 - 1 do
  begin
    if switchlist[i] = texTop then
    begin
      S_StartSound(buttonlist[0].soundorg, sound);
      sides[line.sidenum[0]].toptexture := switchlist[i xor 1];

      if boolval(useAgain) then
        P_StartButton(line, top, switchlist[i], BUTTONTIME);

      exit;
    end
    else
    begin
      if switchlist[i] = texMid then
      begin
        S_StartSound(buttonlist[0].soundorg, sound);
        sides[line.sidenum[0]].midtexture := switchlist[i xor 1];

        if boolval(useAgain) then
          P_StartButton(line, middle, switchlist[i], BUTTONTIME);

        exit;
      end
      else
      begin
        if switchlist[i] = texBot then
        begin
          S_StartSound(buttonlist[0].soundorg, sound);
          sides[line.sidenum[0]].bottomtexture := switchlist[i xor 1];

          if boolval(useAgain) then
            P_StartButton(line, bottom, switchlist[i], BUTTONTIME);

          exit;
        end;
      end;
    end;
  end;
end;

//
// P_UseSpecialLine
// Called when a thing uses a special line.
// Only the front sides of lines are usable.
//
function P_UseSpecialLine(thing: Pmobj_t; line: Pline_t; side: integer): boolean;
begin
  // Err...
  // Use the back sides of VERY SPECIAL lines...
  if boolval(side) then
  begin
    case line.special of
      124:
        // Sliding door open&close
        // UNUSED?
        ;
      else
      begin
        result := false;
        exit;
      end;
    end;
  end;


  // Switches that other things can activate.
  if not boolval(thing.player) then
  begin
    // never open secret doors
    if boolval(line.flags and ML_SECRET) then
    begin
      result := false;
      exit;
    end;

    case line.special of
       1:; // MANUAL DOOR RAISE
      32:; // MANUAL BLUE
      33:; // MANUAL RED
      34:; // MANUAL YELLOW
    else
      begin
        result := false;
        exit;
      end;
    end;
  end;

  // do something
  case line.special of
  // MANUALS
     1, // Vertical Door
    26, // Blue Door/Locked
    27, // Yellow Door /Locked
    28, // Red Door /Locked

    31, // Manual door open
    32, // Blue locked door open
    33, // Red locked door open
    34, // Yellow locked door open

   117, // Blazing door raise
   118: // Blazing door open
      EV_VerticalDoor(line, thing);

	//UNUSED - Door Slide Open&Close
	// case 124:
	// EV_SlidingDoor (line, thing);
	// break;

	// SWITCHES
     7:
      begin
        // Build Stairs
        if boolval(EV_BuildStairs(line, build8)) then
          P_ChangeSwitchTexture(line, 0);
      end;
     9:
      begin
        // Change Donut
        if boolval(EV_DoDonut(line)) then
          P_ChangeSwitchTexture(line, 0);
      end;
    11:
      begin
        // Exit level
        P_ChangeSwitchTexture(line, 0);
        G_ExitLevel;
      end;

    14:
      begin
        // Raise Floor 32 and change texture
        if boolval(EV_DoPlat(line, raiseAndChange, 32)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    15:
      begin
        // Raise Floor 24 and change texture
        if boolval(EV_DoPlat(line, raiseAndChange, 24)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    18:
      begin
        // Raise Floor to next highest floor
        if boolval(EV_DoFloor(line, raiseFloorToNearest)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    20:
      begin
        // Raise Plat next highest floor and change texture
        if boolval(EV_DoPlat(line, raiseToNearestAndChange, 0)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    21:
      begin
        // PlatDownWaitUpStay
        if boolval(EV_DoPlat(line, downWaitUpStay, 0)) then
          P_ChangeSwitchTexture(line,0);
      end;

    23:
      begin
        // Lower Floor to Lowest
        if boolval(EV_DoFloor(line, lowerFloorToLowest)) then
          P_ChangeSwitchTexture(line,0);
      end;

    29:
      begin
        // Raise Door
        if boolval(EV_DoDoor(line, normal)) then
          P_ChangeSwitchTexture(line, 0);
      end;
	
    41:
      begin
        // Lower Ceiling to Floor
        if boolval(EV_DoCeiling(line, lowerToFloor)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    71:
      begin
        // Turbo Lower Floor
        if boolval(EV_DoFloor(line, turboLower)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    49:
      begin
        // Ceiling Crush And Raise
        if boolval(EV_DoCeiling(line, crushAndRaise)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    50:
      begin
        // Close Door
        if boolval(EV_DoDoor(line, close)) then
          P_ChangeSwitchTexture(line, 0);
      end;

    51:
      begin
        // Secret EXIT
        P_ChangeSwitchTexture(line,0);
        G_SecretExitLevel;
      end;

    55:
      begin
        // Raise Floor Crush
        if boolval(EV_DoFloor(line, raiseFloorCrush)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   101:
      begin
        // Raise Floor
        if boolval(EV_DoFloor(line, raiseFloor)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   102:
      begin
        // Lower Floor to Surrounding floor height
        if boolval(EV_DoFloor(line, lowerFloor)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   103:
      begin
        // Open Door
        if boolval(EV_DoDoor(line, open)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   111:
      begin
        // Blazing Door Raise (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeRaise)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   112:
      begin
        // Blazing Door Open (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeOpen)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   113:
      begin
        // Blazing Door Close (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeClose)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   122:
      begin
        // Blazing PlatDownWaitUpStay
        if boolval(EV_DoPlat(line, blazeDWUS, 0)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   127:
      begin
        // Build Stairs Turbo 16
        if boolval(EV_BuildStairs(line, turbo16)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   131:
      begin
        // Raise Floor Turbo
        if boolval(EV_DoFloor(line, raiseFloorTurbo)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   133, // BlzOpenDoor BLUE
   135, // BlzOpenDoor RED
   137: // BlzOpenDoor YELLOW
      begin
        if boolval(EV_DoLockedDoor(line, blazeOpen, thing)) then
          P_ChangeSwitchTexture(line, 0);
      end;

   140:
      begin
        // Raise Floor 512
        if boolval(EV_DoFloor(line, raiseFloor512)) then
          P_ChangeSwitchTexture(line, 0);
      end;

	// BUTTONS
    42:
      begin
        // Close Door
        if boolval(EV_DoDoor(line, close)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    43:
      begin
        // Lower Ceiling to Floor
        if boolval(EV_DoCeiling(line, lowerToFloor)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    45:
      begin
        // Lower Floor to Surrounding floor height
        if boolval(EV_DoFloor(line, lowerFloor)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    60:
      begin
        // Lower Floor to Lowest
        if boolval(EV_DoFloor(line, lowerFloorToLowest)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    61:
      begin
        // Open Door
        if boolval(EV_DoDoor(line, open)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    62:
      begin
        // PlatDownWaitUpStay
        if boolval(EV_DoPlat(line, downWaitUpStay, 1)) then
          P_ChangeSwitchTexture(line, 1);
      end;
	
    63:
      begin
        // Raise Door
        if boolval(EV_DoDoor(line, normal)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    64:
      begin
        // Raise Floor to ceiling
        if boolval(EV_DoFloor(line, raiseFloor)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    66:
      begin
        // Raise Floor 24 and change texture
        if boolval(EV_DoPlat(line, raiseAndChange, 24)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    67:
      begin
        // Raise Floor 32 and change texture
        if boolval(EV_DoPlat(line, raiseAndChange, 32)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    65:
      begin
        // Raise Floor Crush
        if boolval(EV_DoFloor(line, raiseFloorCrush)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    68:
      begin
        // Raise Plat to next highest floor and change texture
        if boolval(EV_DoPlat(line, raiseToNearestAndChange, 0)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    69:
      begin
        // Raise Floor to next highest floor
        if boolval(EV_DoFloor(line, raiseFloorToNearest)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    70:
      begin
        // Turbo Lower Floor
        if boolval(EV_DoFloor(line, turboLower)) then
          P_ChangeSwitchTexture(line, 1);
      end;

   114:
      begin
        // Blazing Door Raise (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeRaise)) then
          P_ChangeSwitchTexture(line, 1);
      end;

   115:
      begin
        // Blazing Door Open (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeOpen)) then
          P_ChangeSwitchTexture(line, 1);
      end;
	
   116:
      begin
        // Blazing Door Close (faster than TURBO!)
        if boolval(EV_DoDoor(line, blazeClose)) then
          P_ChangeSwitchTexture(line, 1);
      end;

   123:
      begin
        // Blazing PlatDownWaitUpStay
        if boolval(EV_DoPlat(line, blazeDWUS, 0)) then
          P_ChangeSwitchTexture(line, 1);
      end;

   132:
      begin
        // Raise Floor Turbo
        if boolval(EV_DoFloor(line, raiseFloorTurbo)) then
          P_ChangeSwitchTexture(line, 1);
      end;

    99, // BlzOpenDoor BLUE
   134, // BlzOpenDoor RED
   136: // BlzOpenDoor YELLOW
      begin
        if boolval(EV_DoLockedDoor(line, blazeOpen, thing)) then
          P_ChangeSwitchTexture(line, 1);
      end;

   138:
      begin
        // Light Turn On
        EV_LightTurnOn(line, 255);
        P_ChangeSwitchTexture(line, 1);
      end;

   139:
      begin
        // Light Turn Off
        EV_LightTurnOn(line, 35);
        P_ChangeSwitchTexture(line, 1);
      end;
  end;
  result := true;
end;




(* Unit r_main implementation *)


                                                                                                                                     

const
// Fineangles in the SCREENWIDTH wide window.
  FIELDOFVIEW = 2048;

var

// just for profiling purposes
  framecount: integer;

//
// R_AddPointToBox
// Expand a given bbox
// so that it encloses a given point.
//
procedure R_AddPointToBox(const x: integer; const y: integer; box: Pfixed_tArray);
begin
  if x < box[BOXLEFT] then
    box[BOXLEFT] := x;
  if x > box[BOXRIGHT] then
    box[BOXRIGHT] := x;
  if y < box[BOXBOTTOM] then
    box[BOXBOTTOM] := y;
  if y > box[BOXTOP] then
    box[BOXTOP] := y;
end;

//
// R_PointOnSide
// Traverse BSP (sub) tree,
//  check point against partition plane.
// Returns side 0 (front) or 1 (back).
//
function R_PointOnSide(x: fixed_t; y: fixed_t; node: Pnode_t): integer;
var
  dx: fixed_t;
  dy: fixed_t;
  left: fixed_t;
  right: fixed_t;
begin
  if not boolval(node.dx) then
  begin
    if x <= node.x then
      result := intval(node.dy > 0)
    else
      result := intval(node.dy < 0);
    exit;
  end;

  if not boolval(node.dy) then
  begin
    if y <= node.y then
      result := intval(node.dx < 0)
    else
      result := intval(node.dx > 0);
    exit;
  end;

  dx := (x - node.x);
  dy := (y - node.y);

    // Try to quickly decide by looking at sign bits.
  if boolval((node.dy xor node.dx xor dx xor dy) and $80000000) then
  begin
    if boolval((node.dy xor dx) and $80000000) then
      // (left is negative)
      result := 1
    else
      result := 0;
    exit;
  end;

  left := FixedMul(node.dy div FRACUNIT, dx);
  right := FixedMul(dy, node.dx div FRACUNIT);

  if right < left then
    result := 0  // front side
  else
    result := 1; // back side
end;

function R_PointOnSegSide(x: fixed_t; y: fixed_t; line: Pseg_t): boolean;
var
  lx: fixed_t;
  ly: fixed_t;
  ldx: fixed_t;
  ldy: fixed_t;
  dx: fixed_t;
  dy: fixed_t;
  left: fixed_t;
  right: fixed_t;
begin
  lx := line.v1.x;
  ly := line.v1.y;

  ldx := line.v2.x - lx;
  ldy := line.v2.y - ly;

  if not boolval(ldx) then
  begin
    if x <= lx then
      result := ldy > 0
    else
      result := ldy < 0;
    exit;
  end;

  if not boolval(ldy) then
  begin
    if y <= ly then
      result := ldx < 0
    else
      result := ldx > 0;
    exit;
  end;

  dx := x - lx;
  dy := y - ly;

  // Try to quickly decide by looking at sign bits.
  if boolval((ldy xor ldx xor dx xor dy) and $80000000) then
  begin
    if  boolval((ldy xor dx) and $80000000) then
      // (left is negative)
      result := true
    else
      result := false;
    exit;
  end;

  left := FixedMul(ldy div FRACUNIT, dx);
  right := FixedMul(dy, ldx div FRACUNIT);

  if right < left then
    result := false // front side
  else
    result := true; // back side
end;

//
// R_PointToAngle
// To get a global angle from cartesian coordinates,
//  the coordinates are flipped until they are in
//  the first octant of the coordinate system, then
//  the y (<=x) is scaled and divided by x to get a
//  tangent (slope) value which is looked up in the
//  tantoangle[] table.

//
function R_PointToAngle(x: fixed_t; y: fixed_t): angle_t;
begin
  x := x - viewx;
  y := y - viewy;

  if (not boolval(x)) and (not boolval(y)) then
  begin
    result := 0;
    exit;
  end;

  if x >= 0 then
  begin
    // x >=0
    if y >= 0 then
    begin
	    // y>= 0
	    if x > y then
      begin
        // octant 0
        result := tantoangle[SlopeDiv(y, x)];
        exit;
      end
	    else
      begin
        // octant 1
        result := ANG90 - 1 - tantoangle[SlopeDiv(x, y)];
        exit;
      end;
    end
    else
    begin
	    // y<0
      y := -y;
      if x > y then
      begin
        // octant 8
        result := -tantoangle[SlopeDiv(y, x)];
        exit;
      end
	    else
      begin
        // octant 7
        result := ANG270 + tantoangle[SlopeDiv(x, y)];
        exit;
      end;
    end;
  end
  else
  begin
    // x<0
    x := -x;
    if y >= 0 then
    begin
      // y>= 0
      if x > y then
      begin
        // octant 3
        result := ANG180 - 1 - tantoangle[SlopeDiv(y, x)];
        exit;
      end
	    else
      begin
        // octant 2
        result := ANG90 + tantoangle[SlopeDiv(x, y)];
        exit;
      end;
    end
    else
    begin
      // y<0
      y := -y;
      if x > y then
      begin
        // octant 4
        result := ANG180 + tantoangle[SlopeDiv(y, x)];
        exit;
      end
	    else
      begin
        // octant 5
        result := ANG270 - 1 - tantoangle[SlopeDiv(x, y)];
        exit;
      end;
    end;
  end;

  result := 0;
end;

function R_PointToAngle2(x1: fixed_t; y1: fixed_t; x2: fixed_t; y2: fixed_t): angle_t;
begin
  viewx := x1;
  viewy := y1;

  result := R_PointToAngle(x2, y2);
end;

function R_PointToDist(x: fixed_t; y: fixed_t): fixed_t;
var
  angle: integer;
  dx: fixed_t;
  dy: fixed_t;
  temp: fixed_t;
begin
  dx := abs(x - viewx);
  dy := abs(y - viewy);

  if dy > dx then
  begin
    temp := dx;
    dx := dy;
    dy := temp;
  end;

  angle := _SHRW(tantoangle[_SHRW(FixedDiv(dy, dx), DBITS)] + ANG90, ANGLETOFINESHIFT); // VJ, does it work ???

  // use as cosine
  result := FixedDiv(dx, finesine[angle]);
end;

//
// R_InitPointToAngle
//
procedure R_InitPointToAngle; 
// UNUSED - now getting from tables.c
var
  i: integer;
  t: integer;
  f: single;
begin
//
// slope (tangent) to angle lookup
//
  for i := 0 to SLOPERANGE do
  begin
    f := arctan(i / SLOPERANGE) / (3.141592657 * 2); // VJ was atan() function
    t := round($ffffffff * f);
    tantoangle[i] := t;
  end;
end;

//
// R_ScaleFromGlobalAngle
// Returns the texture mapping scale
//  for the current line (horizontal span)
//  at the given angle.
// rw_distance must be calculated first.
//
function R_ScaleFromGlobalAngle(visangle: angle_t): fixed_t;
var
  anglea: angle_t;
  angleb: angle_t;
  sinea: integer;
  sineb: integer;
  num: fixed_t;
  den: integer;
begin
(*
    // UNUSED
#if 0
{
    fixed_t		dist;
    fixed_t		z;
    fixed_t		sinv;
    fixed_t		cosv;

    sinv = finesine[(visangle-rw_normalangle)>>ANGLETOFINESHIFT];
    dist = FixedDiv (rw_distance, sinv);
    cosv = finecosine[(viewangle-visangle)>>ANGLETOFINESHIFT];
    z = abs(FixedMul (dist, cosv));
    result = FixedDiv(projection, z);
    return result;
}
#endif
*)
  anglea := ANG90 + (visangle - viewangle);
  angleb := ANG90 + (visangle - rw_normalangle);

  // both sines are allways positive
  sinea := finesine[_SHRW(anglea, ANGLETOFINESHIFT)];
  sineb := finesine[_SHRW(angleb, ANGLETOFINESHIFT)];
//  num := _SHL(FixedMul(projection, sineb), detailshift);
  num := FixedMul(projection, sineb);
  den := FixedMul(rw_distance, sinea);

  if den > _SHR(num, 16) then
  begin
    result := FixedDiv(num, den);

    if result > 64 * FRACUNIT then
      result := 64 * FRACUNIT
    else if result < 256 then
      result := 256
  end
  else
    result := 64 * FRACUNIT;
end;

//
// R_InitTables
//
procedure R_InitTables;
// UNUSED: now getting from tables.c
// VJ: tables constants are now calculated here
var
  i: integer;
  a: extended; // VJ was single
  fv: single;
  t: integer;
begin
// viewangle tangent table
  for i := 0 to FINEANGLES div 2 - 1 do
  begin
    a := (i - FINEANGLES / 4 + 0.5) * PI * 2 / FINEANGLES;
    fv := FRACUNIT * tan(a);
    t := round(fv);
    finetangent[i] := t;
  end;

  // finesine table
  for i := 0 to 5 * FINEANGLES div 4 - 1 do
  begin
    // OPTIMIZE: mirror...
    a := (i + 0.5) * PI * 2 / FINEANGLES;
    t := round(FRACUNIT * sin(a));
    finesine[i] := t;
  end;
  
  finecosine := Pfixed_tArray(@finesine[FINEANGLES div 4]);
end;

//
// R_InitTextureMapping
//
procedure R_InitTextureMapping;
var
  i: integer;
  x: integer;
  t: integer;
  focallength: fixed_t;
  an: angle_t;
begin
  // Use tangent table to generate viewangletox:
  //  viewangletox will give the next greatest x
  //  after the view angle.
  //
  // Calc focallength
  //  so FIELDOFVIEW angles covers SCREENWIDTH.
  focallength := FixedDiv(centerxfrac, finetangent[FINEANGLES div 4 + FIELDOFVIEW div 2]);

  for i := 0 to FINEANGLES div 2 - 1 do
  begin
    if finetangent[i] > FRACUNIT * 2 then
      t := -1
    else if finetangent[i] < -FRACUNIT * 2 then
      t := viewwidth + 1
    else
    begin
      t := FixedMul(finetangent[i], focallength);
	    t := (centerxfrac - t + FRACUNIT - 1) div FRACUNIT;

      if t < -1 then
        t := -1
      else if t > viewwidth + 1 then
        t := viewwidth + 1;
    end;
    viewangletox[i] := t;
  end;

  // Scan viewangletox[] to generate xtoviewangle[]:
  //  xtoviewangle will give the smallest view angle
  //  that maps to x.
  for x := 0 to viewwidth do
  begin
    an := 0;
    while viewangletox[an] > x do
      inc(an);
    xtoviewangle[x] := _SHLW(an, ANGLETOFINESHIFT) - ANG90;
  end;

  // Take out the fencepost cases from viewangletox.
  for i := 0 to FINEANGLES div 2 - 1 do
  begin
    if viewangletox[i] = -1 then
      viewangletox[i] := 0
    else if viewangletox[i] = viewwidth + 1 then
      viewangletox[i] := viewwidth;
  end;
  clipangle := xtoviewangle[0];
end;

//
// R_InitLightTables
// Only inits the zlight table,
//  because the scalelight table changes with view size.
//
const
  DISTMAP = 2;

procedure R_InitLightTables;
var
  i: integer;
  j: integer;
  level: integer;
  startmap: integer;
  scale: integer;
begin
  // Calculate the light levels to use
  //  for each level / distance combination.
  for i := 0 to LIGHTLEVELS - 1 do
  begin
    startmap := ((LIGHTLEVELS - 1 - i) * 2) * NUMCOLORMAPS div LIGHTLEVELS;
    for j := 0 to MAXLIGHTZ - 1 do
    begin
//      scale := FixedDiv(160 * FRACUNIT, _SHL(j + 1, LIGHTZSHIFT));
      scale := FixedDiv(160 * FRACUNIT, _SHL(j + 1, LIGHTZSHIFT)) * 320 div SCREENWIDTH;
      scale := _SHR(scale, LIGHTSCALESHIFT);
      level := startmap - scale div DISTMAP;

      if level < 0 then
        level := 0
      else if level >= NUMCOLORMAPS then
        level := NUMCOLORMAPS - 1;

      zlight[i][j] := Plighttable_tArray(integer(colormaps) + level * 256 * SizeOf(lighttable_t));
    end;
  end;
end;

//
// R_SetViewSize
// Do not really change anything here,
//  because it might be in the middle of a refresh.
// The change will take effect next refresh.
//
var
  setblocks: integer;
  setdetail: integer;

procedure R_SetViewSize(blocks: integer; detail: integer);
begin
  setsizeneeded := true;
  setblocks := blocks;
  setdetail := detail;
end;

//
// R_ExecuteSetViewSize
//
procedure R_ExecuteSetViewSize;
var
  cosadj: fixed_t;
  dy: fixed_t;
  i: integer;
  j: integer;
  level: integer;
  startmap: integer;
begin
  setsizeneeded := false;

  if setblocks = 11 then
  begin
    scaledviewwidth := SCREENWIDTH;
    viewheight := SCREENHEIGHT;
  end
  else
  begin
    scaledviewwidth := setblocks * (SCREENWIDTH div 10);
    if setblocks = 10 then
      viewheight := V_PreserveY(200 - ST_HEIGHT)
    else
      viewheight := (setblocks * V_PreserveY(200 - ST_HEIGHT) div 10) and (not 7); //and (not 7);
  end;

  detailshift := setdetail;
  viewwidth := scaledviewwidth;

  centery := viewheight div 2;
  centerx := viewwidth div 2;
  centerxfrac := centerx * FRACUNIT;
  centeryfrac := centery * FRACUNIT;
  projection := centerxfrac;

  if not boolval(detailshift) then
  begin
    colfunc := R_DrawColumn;

    if SCREENHEIGHT = 200 then
      skycolfunc := R_DrawColumn
    else
      skycolfunc := R_DrawSkyColumn;

    basecolfunc := R_DrawColumn;

    fuzzcolfunc := R_DrawFuzzColumn;

    spanfunc := R_DrawSpan;
  end
  else
  begin
    colfunc := R_DrawColumnLow;

    if SCREENHEIGHT = 200 then
      skycolfunc := R_DrawColumn
    else
      skycolfunc := R_DrawSkyColumn;

    basecolfunc := R_DrawColumn;

    fuzzcolfunc := R_DrawFuzzColumn;

    spanfunc := R_DrawSpanLow;
  end;

  R_InitBuffer(scaledviewwidth, viewheight);

  R_InitTextureMapping;

  // psprite scales
  pspritescale := FRACUNIT * viewwidth div 320;
  pspriteiscale := FRACUNIT * 320 div viewwidth;

  // thing clipping
  for i := 0 to viewwidth - 1 do
    screenheightarray[i] := viewheight;

  // planes
  for i := 0 to viewheight - 1 do
  begin
    dy := ((i - viewheight div 2) * FRACUNIT) + FRACUNIT div 2;
    dy := abs(dy);
    yslope[i] := FixedDiv(viewwidth * FRACUNIT div 2, dy);
  end;

  for i := 0 to viewwidth - 1 do
  begin
    cosadj := abs(finecosine[_SHRW(xtoviewangle[i], ANGLETOFINESHIFT)]);
    distscale[i] := FixedDiv(FRACUNIT, cosadj);
  end;

  // Calculate the light levels to use
  //  for each level / scale combination.
  for i := 0 to LIGHTLEVELS - 1 do
  begin
    startmap := ((LIGHTLEVELS - 1 - i) * 2) * NUMCOLORMAPS div LIGHTLEVELS;
    for j := 0 to MAXLIGHTSCALE - 1 do
    begin
//      level := startmap - j * SCREENWIDTH div _SHL(viewwidth, detailshift) div DISTMAP;
      level := startmap - j * SCREENWIDTH div viewwidth div DISTMAP;

      if level < 0 then
        level := 0;

      if level >= NUMCOLORMAPS then
        level := NUMCOLORMAPS - 1;

      scalelight[i][j] := Plighttable_tArray(integer(colormaps) + level * 256);
    end;
  end;
end;

//
// R_Init
//
procedure R_Init;
begin
  R_InitData;
  printf(#13#10 + 'R_InitData');
  R_InitPointToAngle;
  printf(#13#10 + 'R_InitPointToAngle');
  R_InitTables;
  // viewwidth / viewheight / detailLevel are set by the defaults
  printf(#13#10 + 'R_InitTables');

  R_SetViewSize(screenblocks, detailLevel);
  R_InitPlanes;
  printf(#13#10 + 'R_InitPlanes');
  R_InitLightTables;
  printf(#13#10 + 'R_InitLightTables');
  R_InitSkyMap;
  printf(#13#10 + 'R_InitSkyMap');
  R_InitTranslationTables;
  printf(#13#10 + 'R_InitTranslationsTables');

  framecount := 0;
end;

//
// R_PointInSubsector
//
function R_PointInSubsector(x: fixed_t; y: fixed_t): Psubsector_t;
var
  node: Pnode_t;
  side: integer;
  nodenum: integer;
begin
  // single subsector is a special case
  if numnodes = 0 then
  begin
    result := @subsectors[0];
    exit;
  end;

  nodenum := numnodes - 1;

  while (nodenum and NF_SUBSECTOR) = 0 do
  begin
    node := @nodes[nodenum];
    side := R_PointOnSide(x, y, node);
    nodenum := node.children[side];
  end;

  result := @subsectors[nodenum and (not NF_SUBSECTOR)];
end;

//
// R_SetupFrame
//
procedure R_SetupFrame(player: Pplayer_t);
var
  i: integer;
  cy: integer;
begin
  viewplayer := player;
  viewx := player.mo.x;
  viewy := player.mo.y;
  viewangle := player.mo.angle + viewangleoffset;
  extralight := player.extralight;

  viewz := player.viewz;

//{***********************
// VJ Enable this for z axis shift (still buggy)
  cy := viewheight div 2 + player.lookdir * screenblocks div 10;
  if centery <> cy then
  begin
    centery := cy;
    centeryfrac := centery * FRACUNIT;
    for i := 0 to viewwidth - 1 do
      yslope[i] := FixedDiv(viewwidth div 2 * FRACUNIT,
        abs(((i - centery) * FRACUNIT) + FRACUNIT div 2));
  end;
//******************************}

  viewsin := finesine[_SHRW(viewangle, ANGLETOFINESHIFT)];
  viewcos := finecosine[_SHRW(viewangle, ANGLETOFINESHIFT)];

  sscount := 0;

  if boolval(player.fixedcolormap) then
  begin
    fixedcolormap := Plighttable_tArray(
      integer(colormaps) + player.fixedcolormap * 256 * SizeOf(lighttable_t));

    walllights := @scalelightfixed;

    for i := 0 to MAXLIGHTSCALE - 1 do
      scalelightfixed[i] := fixedcolormap;
  end
  else
    fixedcolormap := nil;

  inc(framecount);
  inc(validcount);
end;

//
// R_RenderView
//
procedure R_RenderPlayerView(player: Pplayer_t);
begin
  R_SetupFrame(player);

  // Clear buffers.
  R_ClearClipSegs;
  R_ClearDrawSegs;
  R_ClearPlanes;
  R_ClearSprites;

  // check for new console commands.
  NetUpdate;

  // The head node is the last node output.
  R_RenderBSPNode(numnodes - 1);

  // Check for new console commands.
  NetUpdate;

  R_DrawPlanes;

  // Check for new console commands.
  NetUpdate;

  R_DrawMasked;

  // Check for new console commands.
  NetUpdate;
end;


(* Unit d_net implementation *)


                                                                                   

const
  NCMD_EXIT  = $80000000;
  NCMD_RETRANSMIT  = $40000000;
  NCMD_SETUP= $20000000;
  NCMD_KILL  = $10000000;  // kill game
  NCMD_CHECKSUM  = $0fffffff;

var
  netbuffer: Pdoomdata_t; // points inside doomcom

//
// NETWORKING
//
// gametic is the tic about to (or currently being) run
// maketic is the tick that hasn't had control made for it yet
// nettics[] has the maketics for all players
//
// a gametic cannot be run until nettics[] > gametic for all players
//

const
  CRESENDCOUNT = 10;
  PL_DRONE = $80; // bit flag in doomdata->player

var
  localcmds: array[0..(BACKUPTICS)-1] of ticcmd_t;

  nettics: array[0..(MAXNETNODES)-1] of integer;
  nodeingame: array[0..(MAXNETNODES)-1] of boolean; // set false as nodes leave game
  remoteresend: array[0..(MAXNETNODES)-1] of boolean; // set when local needs tics
  resendto: array[0..(MAXNETNODES)-1] of integer; // set when remote needs tics
  resendcount: array[0..(MAXNETNODES)-1] of integer;

  nodeforplayer: array[0..(MAXPLAYERS)-1] of integer;

  skiptics: integer;
  maxsend: integer; // BACKUPTICS/(2*ticdup)-1

var
  reboundpacket: boolean;
  reboundstore: doomdata_t;

//
//
//
function NetbufferSize: integer;
begin
//    return (int)&(((doomdata_t *)0)->cmds[netbuffer->numtics]);
  result := SizeOf(doomdata_t) - SizeOf(Tcmds) + netbuffer.numtics * SizeOf(ticcmd_t);
end;

//
// Checksum
//
function NetbufferChecksum: LongWord;
var
  i: integer;
  p: PLongWordArray;
begin
  result := $1234567;

  p := PLongWordArray(pointer(netbuffer));

  // Hack: first position of doomdata_t is checksum (longword)
  for i := 1 to NetbufferSize div 4 do
    result := result + p[i];

  result := result and NCMD_CHECKSUM;
end;

function ExpandTics(low: integer): integer;
var
  delta: integer;
begin
  delta := low - (maketic and $ff);

  if (delta >= -64) and (delta <= 64) then
  begin
    result := (maketic and (not $ff)) + low;
    exit;
  end;

  if delta > 64 then
  begin
    result := (maketic and (not $ff)) - 256 + low;
    exit;
  end;

  if delta < -64 then
  begin
    result := (maketic and (not $ff)) + 256 + low;
    exit;
  end;

  I_Error('ExpandTics(): strange value %d at maketic %d', [low, maketic]);
  result := 0;
end;

//
// HSendPacket
//
procedure HSendPacket(node: integer; flags: LongWord);
var
  i: integer;
  realretrans: integer;
  p: PByteArray;
begin
  netbuffer.checksum := NetbufferChecksum or flags;

  if node = 0 then
  begin
    reboundstore := netbuffer^;
    reboundpacket := true;
    exit;
  end;

  if demoplayback then
    exit;

  if not netgame then
    I_Error('HSendPacket(): Tried to transmit to another node');

  doomcom.command := CMD_SEND;
  doomcom.remotenode := node;
  doomcom.datalength := NetbufferSize;

  if debugfile <> nil then
  begin
    if (netbuffer.checksum and NCMD_RETRANSMIT) <> 0 then
      realretrans := ExpandTics(netbuffer.retransmitfrom)
    else
      realretrans := -1;

    fprintf(debugfile, 'send (%d + %d, R %d) [%d] ',
       [ExpandTics(netbuffer.starttic), netbuffer.numtics,
        realretrans, doomcom.datalength]);

    p := PByteArray(pointer(netbuffer));
    for i := 0 to doomcom.datalength - 1 do
      fprintf(debugfile, '%d ', [p[i]]);

    fprintf(debugfile, #13#10);
  end;

  I_NetCmd;
end;

//
// HGetPacket
// Returns false if no packet is waiting
//
function HGetPacket: boolean;
var
  realretrans: integer;
  i: integer;
  p: PByteArray;
begin
  if reboundpacket then
  begin
    netbuffer^ := reboundstore;
    doomcom.remotenode := 0;
    reboundpacket := false;
    result := true;
    exit;
  end;

  if not netgame then
  begin
    result := false;
    exit;
  end;

  if demoplayback then
  begin
    result := false;
    exit;
  end;

  doomcom.command := CMD_GET;
  I_NetCmd;

  if doomcom.remotenode = -1 then
  begin
    result := false;
    exit;
  end;

  if doomcom.datalength <> NetbufferSize then
  begin
    fprintf(debugfile, 'bad packet length %d' + #13#10, [doomcom.datalength]);
    result := false;
    exit;
  end;

  if NetbufferChecksum <> (netbuffer.checksum and NCMD_CHECKSUM) then
  begin
    fprintf(debugfile, 'bad packet checksum' + #13#10);
    result := false;
    exit;
  end;

  if debugfile <> nil then
  begin

    if (netbuffer.checksum and NCMD_SETUP) <> 0 then
      fprintf(debugfile, 'setup packet' + #13#10)
    else
    begin
      if (netbuffer.checksum and NCMD_RETRANSMIT) <> 0 then
        realretrans := ExpandTics(netbuffer.retransmitfrom)
      else
        realretrans := -1;

      fprintf(debugfile, 'get %d = (%d + %d, R %d)[%d] ',
        [doomcom.remotenode, ExpandTics(netbuffer.starttic),
         netbuffer.numtics, realretrans, doomcom.datalength]);

      p := PByteArray(pointer(netbuffer));
      for i := 0 to doomcom.datalength - 1 do
        fprintf(debugfile, '%d ', [p[i]]);
      fprintf(debugfile, #13#10);
    end;
  end;
  result := true;
end;

//
// GetPackets
//
var exitmsg: string;

procedure GetPackets;
var
  netconsole: integer;
  netnode: integer;
  src, dest: Pticcmd_t;
  realend: integer;
  realstart: integer;
  start: integer;
begin
  while HGetPacket do
  begin
    if (netbuffer.checksum and NCMD_SETUP) <> 0 then
      continue;    // extra setup packet

    netconsole := netbuffer.player and (not PL_DRONE);
    netnode := doomcom.remotenode;

  // to save bytes, only the low byte of tic numbers are sent
  // Figure out what the rest of the bytes are
    realstart := ExpandTics(netbuffer.starttic);
    realend := realstart + netbuffer.numtics;

  // check for exiting the game
    if (netbuffer.checksum and NCMD_EXIT) <> 0 then
    begin
      if not nodeingame[netnode] then
        continue;
      nodeingame[netnode] := false;
      playeringame[netconsole] := false;
      exitmsg := 'Player 1 left the game';
      exitmsg[7] := chr(ord(exitmsg[7]) + netconsole);
      players[consoleplayer]._message := exitmsg;
      if demorecording then
        G_CheckDemoStatus;
      continue;
    end;

  // check for a remote game kill
    if  (netbuffer.checksum and NCMD_KILL) <> 0 then
      I_Error('GetPackets(): Killed by network driver');

    nodeforplayer[netconsole] := netnode;

  // check for retransmit request
    if (resendcount[netnode] <= 0) and ((netbuffer.checksum and NCMD_RETRANSMIT) <> 0) then
    begin
      resendto[netnode] := ExpandTics(netbuffer.retransmitfrom);
      if debugfile <> nil then
        fprintf(debugfile, 'retransmit from %d' + #13#10, [resendto[netnode]]);
      resendcount[netnode] := CRESENDCOUNT;
    end
    else
      resendcount[netnode] := resendcount[netnode] - 1;

  // check for out of order / duplicated packet
    if realend = nettics[netnode] then
      continue;

    if realend < nettics[netnode] then
    begin
      fprintf(debugfile, 'out of order packet (%d + %d)' + #13#10,
        [realstart, netbuffer.numtics]);
      continue;
    end;

  // check for a missed packet
    if realstart > nettics[netnode] then
    begin
      // stop processing until the other system resends the missed tics
      fprintf(debugfile, 'missed tics from %d (%d - %d)' + #13#10,
        [netnode, realstart, nettics[netnode]]);
      remoteresend[netnode] := true;
      continue;
    end;

  // update command store from the packet

    remoteresend[netnode] := false;

    start := nettics[netnode] - realstart;
    src := @netbuffer.cmds[start];

    while nettics[netnode] < realend do
    begin
      dest := @netcmds[netconsole][nettics[netnode] mod BACKUPTICS];
      nettics[netnode] := nettics[netnode] + 1;
////////  *dest = *src;
      dest^ := src^;
      src := Pticcmd_t(pointer(integer(src) + SizeOf(ticcmd_t)));
    end;
  end;
end;

//
// NetUpdate
// Builds ticcmds for console player,
// sends out a packet
//
var gametime: integer;

procedure NetUpdate;
var
  nowtime: integer;
  newtics: integer;
  i, j: integer;
  realstart: integer;
  gameticdiv: integer;
begin
  // check time
  nowtime := I_GetTime div ticdup;
  newtics := nowtime - gametime;
  gametime := nowtime;

  if newtics <= 0 then // nothing new to update
  begin
    GetPackets;
    exit;
  end;

  if skiptics <= newtics then
  begin
    newtics := newtics - skiptics;
    skiptics := 0;
  end
  else
  begin
    skiptics := skiptics - newtics;
    newtics := 0;
  end;

  netbuffer.player := consoleplayer;

  // build new ticcmds for console player
  gameticdiv := gametic div ticdup;
  for i := 0 to newtics - 1 do
  begin
    I_StartTic;
    D_ProcessEvents;
    
    if I_GameFinished then
      exit;

    if maketic - gameticdiv >= BACKUPTICS div 2 - 1 then
      break;          // can't hold any more

  //printf ("mk:%i ",maketic);
    G_BuildTiccmd(@localcmds[maketic mod BACKUPTICS]);
    inc(maketic);
  end;


  if singletics then
    exit;         // singletic update is syncronous

  // send the packet to the other nodes
  for i := 0 to doomcom.numnodes - 1 do
    if nodeingame[i] then
    begin
      netbuffer.starttic := resendto[i];
      realstart := resendto[i];
      netbuffer.numtics := maketic - realstart;
      if netbuffer.numtics > BACKUPTICS then
        I_Error('NetUpdate(): netbuffer.numtics > BACKUPTICS');

      resendto[i] := maketic - doomcom.extratics;

      for j := 0 to netbuffer.numtics - 1 do
        netbuffer.cmds[j] :=
          localcmds[(realstart + j) mod BACKUPTICS];

      if remoteresend[i] then
      begin
        netbuffer.retransmitfrom := nettics[i];
        HSendPacket(i, NCMD_RETRANSMIT);
      end
      else
      begin
        netbuffer.retransmitfrom := 0;
        HSendPacket(i, 0);
      end;
    end;

  // listen for other packets
  GetPackets;
end;

//
// CheckAbort
//
procedure CheckAbort;
var
  ev: Pevent_t;
  stoptic: integer;
begin
  stoptic := I_GetTime + 2;
  while I_GetTime < stoptic do
  begin
    I_WaitVBL(1);
    I_StartTic;
    I_ProcessWindows;
  end;

  I_StartTic;
  repeat
    ev := @events[eventtail];
    if (ev._type = ev_keydown) and (ev.data1 = KEY_ESCAPE) then
      I_Error('CheckAbort(): Network game synchronization aborted.');

    eventtail := (eventtail + 1) and (MAXEVENTS-1);
  until eventtail = eventhead;
end;

//
// D_ArbitrateNetStart
//
procedure D_ArbitrateNetStart;
var
  i: integer;
  gotinfo: array[0..MAXNETNODES - 1] of boolean;
begin
  autostart := true;
  ZeroMemory(gotinfo, SizeOf(gotinfo));

  if doomcom.consoleplayer <> 0 then
  begin
  // listen for setup info from key player
    printf('listening for network start info...' + #13#10);
    while true do
    begin
      CheckAbort;
      if not HGetPacket then
        continue;
      if (netbuffer.checksum and NCMD_SETUP) <> 0 then
      begin
        if netbuffer.player <> VERSION then
          I_Error('D_ArbitrateNetStart(): Different DOOM versions cannot play a net game!');
        startskill := skill_t(netbuffer.retransmitfrom and 15);
        deathmatch := _SHR((netbuffer.retransmitfrom and $C0), 6);
        nomonsters := (netbuffer.retransmitfrom and $20) > 0;
        respawnparm := (netbuffer.retransmitfrom and $10) > 0;
        startmap := netbuffer.starttic and $3f;
        startepisode := _SHR(netbuffer.starttic, 6);
        exit;
      end;
    end // while
  end
  else
  begin
  // key player, send the setup info
    printf('sending network start info...' + #13#10);
    repeat
      CheckAbort;
      for i := 0 to doomcom.numnodes - 1 do
      begin
        netbuffer.retransmitfrom := Ord(startskill);
        if deathmatch <> 0 then
          netbuffer.retransmitfrom := netbuffer.retransmitfrom or _SHL(deathmatch, 6);
        if nomonsters then
          netbuffer.retransmitfrom := netbuffer.retransmitfrom or $20;
        if respawnparm then
          netbuffer.retransmitfrom := netbuffer.retransmitfrom or $10;
        netbuffer.starttic := startepisode * 64 + startmap;
        netbuffer.player := VERSION;
        netbuffer.numtics := 0;
        HSendPacket (i, NCMD_SETUP);
      end;

      i := 10;
      while (i > 0) and HGetPacket do
      begin
        if (netbuffer.player and $7f) < MAXNETNODES then
          gotinfo[netbuffer.player and $7f] := true;
      end;

      i := 1;
      while (i < doomcom.numnodes) and gotinfo[i] do
        inc(i);

    until i >= doomcom.numnodes;
  end;
end;

//
// D_CheckNetGame
// Works out player numbers among the net participants
//
procedure D_CheckNetGame;
var
  i: integer;
begin
  for i := 0 to MAXNETNODES - 1 do
  begin
    nodeingame[i] := false;
    nettics[i] := 0;
    remoteresend[i] := false; // set when local needs tics
    resendto[i] := 0;         // which tic to start sending
  end;

  // I_InitNetwork sets doomcom and netgame
  I_InitNetwork;
  if doomcom.id <> DOOMCOM_ID then
    I_Error('D_CheckNetGame(): Doomcom buffer invalid!');

  netbuffer := @doomcom.data;
  consoleplayer := doomcom.consoleplayer;
  displayplayer := doomcom.consoleplayer;
  if netgame then
    D_ArbitrateNetStart;

  printf('startskill %d  deathmatch: %d  startmap: %d  startepisode: %d' + #13#10,
    [Ord(startskill), deathmatch, startmap, startepisode]);

  // read values out of doomcom
  ticdup := doomcom.ticdup;
  maxsend := BACKUPTICS div (2 * ticdup) - 1;
  if maxsend < 1 then
    maxsend := 1;

  for i := 0 to doomcom.numplayers - 1 do
    playeringame[i] := true;
  for i := 0 to doomcom.numnodes - 1 do
    nodeingame[i] := true;

  printf('player %d of %d (%d nodes)' + #13#10,
    [consoleplayer + 1, doomcom.numplayers, doomcom.numnodes]);
end;

//
// D_QuitNetGame
// Called before quitting to leave a net game
// without hanging the other players
//
procedure D_QuitNetGame;
var
  i, j: integer;
begin
  if (not netgame) or (not usergame) or (consoleplayer = -1) or (demoplayback) then
    exit;

  // send a bunch of packets for security
  netbuffer.player := consoleplayer;
  netbuffer.numtics := 0;
  for i := 0 to 3 do
  begin
    for j := 1 to doomcom.numnodes - 1 do
      if nodeingame[j] then
        HSendPacket(j, NCMD_EXIT);
    I_WaitVBL(1);
  end;
end;

//
// TryRunTics
//
var
  frameon: integer;
  frameskip: array[0..3] of boolean;
  oldnettics: integer;
  oldentertics: integer;

procedure TryRunTics;
var
  i, j: integer;
  lowtic: integer;
  entertic: integer;
  realtics: integer;
  availabletics: integer;
  counts: integer;
  cmd: Pticcmd_t;
  buf: integer;
begin
  // get real tics
  entertic := I_GetTime div ticdup;
  realtics := entertic - oldentertics;
  oldentertics := entertic;

  // get available tics
  NetUpdate;
  
  if I_GameFinished then
    exit;

  lowtic := MAXINT;
  for i := 0 to doomcom.numnodes - 1 do
  begin
    if nodeingame[i] then
    begin
      if nettics[i] < lowtic then
        lowtic := nettics[i];
    end;
  end;
  availabletics := lowtic - gametic div ticdup;

  // decide how many tics to run
  if realtics < availabletics - 1 then
    counts := realtics + 1
  else if realtics < availabletics then
    counts := realtics
  else
    counts := availabletics;

  if counts < 1 then
    counts := 1;

  inc(frameon);

  if not demoplayback then
  begin
  // ideally nettics[0] should be 1 - 3 tics above lowtic
  // if we are consistantly slower, speed up time
    i := 0;
    while (i < MAXPLAYERS) and (not playeringame[i]) do
      inc(i);
    if consoleplayer = i then
    begin
      // the key player does not adapt
    end
    else
    begin
      if nettics[0] <= nettics[nodeforplayer[i]] then
      begin
        dec(gametime);
    // printf ('-');
      end;
      frameskip[frameon and 3] := (oldnettics > nettics[nodeforplayer[i]]);
      oldnettics := nettics[0];
      if (frameskip[0] and frameskip[1] and frameskip[2] and frameskip[3]) then
      begin
        skiptics := 1;
    // printf ("+");
      end;
    end;
  end; // demoplayback

  // wait for new tics if needed
  while lowtic < gametic / ticdup + counts do
  begin
    NetUpdate;
    lowtic := MAXINT;

    for i := 0 to doomcom.numnodes - 1 do
      if nodeingame[i] and (nettics[i] < lowtic) then
        lowtic := nettics[i];

    if lowtic < gametic / ticdup then
      I_Error('TryRunTics(): lowtic < gametic');

    // don't stay in here forever -- give the menu a chance to work
    if I_GetTime / ticdup - entertic >= 20 then
    begin
      M_Ticker;
      exit;
    end;
  end;

  // run the count * ticdup dics
  while counts <> 0 do
  begin
    dec(counts);
    
    for i := 0 to ticdup - 1 do
    begin
      if gametic / ticdup > lowtic then
        I_Error('TryRunTics(): gametic > lowtic');
      if advancedemo then
        D_DoAdvanceDemo;
      M_Ticker;
      G_Ticker;
      inc(gametic);

      // modify command for duplicated tics
      if i <> ticdup - 1 then
      begin

        buf := (gametic div ticdup) mod BACKUPTICS;
        for j := 0 to MAXPLAYERS - 1 do
        begin
          cmd := @netcmds[j][buf];
          cmd.chatchar := 0;
          if (cmd.buttons and BT_SPECIAL) <> 0 then
            cmd.buttons := 0;
        end;
      end;
    end;
    NetUpdate;  // check for new console commands
  end;
end;


(* Unit wi_stuff implementation *)


                                                                                                                                                      

const
//
// Data needed to add patches to full screen intermission pics.
// Patches are statistics messages, and animations.
// Loads of by-pixel layout and placement, offsets etc.
//


//
// Different vetween registered DOOM (1994) and
//  Ultimate DOOM - Final edition (retail, 1995?).
// This is supposedly ignored for commercial
//  release (aka DOOM II), which had 34 maps
//  in one episode. So there.
  NUMEPISODES = 4;
  NUMMAPS = 9;


// in tics
//U #define PAUSELEN		(TICRATE*2)
//U #define SCORESTEP		100
//U #define ANIMPERIOD		32
// pixel distance from "(YOU)" to "PLAYER N"
//U #define STARDIST		10
//U #define WK 1


// GLOBAL LOCATIONS
  WI_TITLEY = 2;
  WI_SPACINGY = 33;

// SINGPLE-PLAYER STUFF
  SP_STATSX = 50;
  SP_STATSY = 50;

  SP_TIMEX = 16;
  SP_TIMEY = 200 - 32; // (SCREENHEIGHT - 32);

// NET GAME STUFF
  NG_STATSY = 50;

  NG_SPACINGX = 64;


// DEATHMATCH STUFF
  DM_MATRIXX = 42;
  DM_MATRIXY = 68;

  DM_SPACINGX = 40;

  DM_TOTALSX = 269;

  DM_KILLERSX = 10;
  DM_KILLERSY = 100;
  DM_VICTIMSX = 5;
  DM_VICTIMSY = 50;

type
  animenum_t = (
    ANIM_ALWAYS,
    ANIM_RANDOM,
    ANIM_LEVEL
  );


type
  point_t = record
    x: integer;
    y: integer;
  end;


//
// Animation.
// There is another anim_t used in p_spec.
//
type
  wianim_t = record // VJ renamed from anim_t -> wianim_t
    _type: animenum_t;

    // period in tics between animations
    period: integer;

    // number of animation frames
    nanims: integer;

    // location of animation
    loc: point_t;

    // ALWAYS: n/a,
    // RANDOM: period deviation (<256),
    // LEVEL: level
    data1: integer;

    // ALWAYS: n/a,
    // RANDOM: random base period,
    // LEVEL: n/a
    data2: integer;

    // actual graphics for frames of animations
    p: array[0..2] of Ppatch_t;

    // following must be initialized to zero before use!

    // next value of bcnt (used in conjunction with period)
    nexttic: integer;

    // last drawn animation frame
    lastdrawn: integer;

    // next frame number to animate
    ctr: integer;

    // used by RANDOM and LEVEL when animating
    state: integer;
  end;
  Pwianim_t = ^wianim_t;
  wianim_tArray = packed array[0..$FFFF] of wianim_t;
  Pwianim_tArray = ^wianim_tArray;


var
  lnodes: array[0..NUMEPISODES - 1, 0..NUMMAPS - 1] of point_t = (
  // Episode 0 World Map
    (
    (x: 185; y: 164),  // location of level 0 (CJ)
    (x: 148; y: 143),  // location of level 1 (CJ)
    (x:  69; y: 122),  // location of level 2 (CJ)
    (x: 209; y: 102),  // location of level 3 (CJ)
    (x: 116; y:  89),  // location of level 4 (CJ)
    (x: 166; y:  55),  // location of level 5 (CJ)
    (x:  71; y:  56),  // location of level 6 (CJ)
    (x: 135; y:  29),  // location of level 7 (CJ)
    (x:  71; y:  24)   // location of level 8 (CJ)
    ),

  // Episode 1 World Map should go here
    (
    (x: 254; y:  25),  // location of level 0 (CJ)
    (x:  97; y:  50),  // location of level 1 (CJ)
    (x: 188; y:  64),  // location of level 2 (CJ)
    (x: 128; y:  78),  // location of level 3 (CJ)
    (x: 214; y:  92),  // location of level 4 (CJ)
    (x: 133; y: 130),  // location of level 5 (CJ)
    (x: 208; y: 136),  // location of level 6 (CJ)
    (x: 148; y: 140),  // location of level 7 (CJ)
    (x: 235; y: 158)   // location of level 8 (CJ)
    ),

  // Episode 2 World Map should go here
    (
    (x: 156; y: 168),  // location of level 0 (CJ)
    (x:  48; y: 154),  // location of level 1 (CJ)
    (x: 174; y:  95),  // location of level 2 (CJ)
    (x: 265; y:  75),  // location of level 3 (CJ)
    (x: 130; y:  48),  // location of level 4 (CJ)
    (x: 279; y:  23),  // location of level 5 (CJ)
    (x: 198; y:  48),  // location of level 6 (CJ)
    (x: 140; y:  25),  // location of level 7 (CJ)
    (x: 281; y: 136)   // location of level 8 (CJ)
    ),
  ////////////////////////////////////////////////
    (
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0),
    (x: 0; y: 0)
    )
  );

//
// Animation locations for episode 0 (1).
// Using patches saves a lot of space,
//  as they replace 320x200 full screen frames.
//
  epsd0animinfo: array[0..9] of wianim_t = (
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 224; y: 104)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 184; y: 160)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 112; y: 136)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  72; y: 112)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  88; y:  96)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  64; y:  48)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 192; y:  40)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 136; y:  16)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  80; y:  16)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  64; y:  24))
  );

  epsd1animinfo: array[0..8] of wianim_t = (
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 1),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 2),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 3),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 4),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 5),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 6),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 7),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 3; loc: (x: 192; y: 144); data1: 8),
    (_type: ANIM_LEVEL; period: TICRATE div 3; nanims: 1; loc: (x: 128; y: 136); data1: 8)
  );

  epsd2animinfo: array[0..5] of wianim_t = (
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 104; y: 168)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x:  40; y: 136)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 160; y:  96)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 104; y:  80)),
    (_type: ANIM_ALWAYS; period: TICRATE div 3; nanims: 3; loc: (x: 120; y:  32)),
    (_type: ANIM_ALWAYS; period: TICRATE div 4; nanims: 3; loc: (x:  40; y:   0))
  );

const
  NUMANIMS: array[0..NUMEPISODES - 1] of integer = (10, 9, 6, 0);

var
  wi_stuff_anims: array[0..NUMEPISODES - 1] of Pwianim_tArray;


//
// GENERAL DATA
//

//
// Locally used stuff.
//
const
// States for single-player
{  SP_KILLS = 0;
  SP_ITEMS = 2;
  SP_SECRET = 4;
  SP_FRAGS = 6;
  SP_TIME = 8;}
//  SP_PAR = ST_TIME; // removed by VJ

  SP_PAUSE = 1;

// in seconds
  SHOWNEXTLOCDELAY = 4;
//#define SHOWLASTLOCDELAY	SHOWNEXTLOCDELAY


var
// used to accelerate or skip a stage
  acceleratestage: integer = 0;

// wbs->pnum
  me: integer = 0;

// specifies current state
  state: stateenum_t;

// contains information passed into intermission
  wbs: Pwbstartstruct_t = nil;

  plrs: Pwbplayerstruct_tArray = nil;  // wbs->plyr[]

// used for general timing
  cnt: integer = 0;

// used for timing of background animation
  bcnt: integer;

// signals to refresh everything for one frame
  firstrefresh: integer = 0;

  cnt_kills: array[0..MAXPLAYERS - 1] of integer;
  cnt_items: array[0..MAXPLAYERS - 1] of integer;
  cnt_secret: array[0..MAXPLAYERS - 1] of integer;
  cnt_time: integer = 0;
  cnt_par: integer = 0;
  cnt_pause: integer = 0;

// # of commercial levels
  NUMCMAPS: integer = 0;


//
//	GRAPHICS
//

// background (map of levels).
  bg: Ppatch_t = nil;

// You Are Here graphic
  yah: array[0..1] of Ppatch_t;

// splat
  splat: Ppatch_t = nil;

// %, : graphics
  percent: Ppatch_t;
  colon: Ppatch_t;

// 0-9 graphic
  num: array[0..9] of Ppatch_t;

// minus sign
  wiminus: Ppatch_t = nil;

// "Finished!" graphics
  finished: Ppatch_t = nil;

// "Entering" graphic
  entering: Ppatch_t = nil;

// "secret"
  sp_secret: Ppatch_t = nil;

 // "Kills", "Scrt", "Items", "Frags"
  kills: Ppatch_t = nil;
  secret: Ppatch_t = nil;
  items: Ppatch_t = nil;
  frags: Ppatch_t = nil;

// Time sucks.
  time: Ppatch_t = nil;
  par: Ppatch_t = nil;
  sucks: Ppatch_t = nil;

// "killers", "victims"
  killers: Ppatch_t = nil;
  victims: Ppatch_t = nil;

// "Total", your face, your dead face
  total: Ppatch_t = nil;
  star: Ppatch_t = nil;
  bstar: Ppatch_t = nil;

// "red P[1..MAXPLAYERS]"
  p: array[0..MAXPLAYERS - 1] of Ppatch_t;

// "gray P[1..MAXPLAYERS]"
  bp: array[0..MAXPLAYERS - 1] of Ppatch_t;

 // Name graphics of each level (centered)
  lnames: Ppatch_tPArray;

//
// CODE
//

// slam background
// UNUSED static unsigned char *background=0;


procedure WI_slamBackground;
begin
  memcpy(screens[_FG], screens[1], SCREENWIDTH * SCREENHEIGHT);
  V_MarkRect(0, 0, SCREENWIDTH, SCREENHEIGHT, false);
end;

// The ticker is used to detect keys
//  because of timing issues in netgames.
function WI_Responder(ev: Pevent_t): boolean;
begin
  result := false;
end;

// Draws "<Levelname> Finished!"
procedure WI_drawLF;
var
  y: integer;
begin
  y := WI_TITLEY;

  // draw <LevelName>
//  V_DrawPatch((SCREENWIDTH - lnames[wbs.last].width) div 2, y, _FB, lnames[wbs.last]);
  V_DrawPatch((320 - lnames[wbs.last].width) div 2, y, _FG, lnames[wbs.last], true);

  // draw "Finished!"
  y := y + (5 * lnames[wbs.last].height) div 4;

//  V_DrawPatch((SCREENWIDTH - finished.width) div 2, y, _FB, finished);
  V_DrawPatch((320 - finished.width) div 2, y, _FG, finished, true);
end;

// Draws "Entering <LevelName>"
procedure WI_drawEL;
var
  y: integer;
begin
  y := WI_TITLEY;

  // draw "Entering"
//  V_DrawPatch((SCREENWIDTH - entering.width) div 2, y, _FB, entering);
  V_DrawPatch((320 - entering.width) div 2, y, _FG, entering, true);

  // draw level
  y := y + (5 * lnames[wbs.next].height) div 4;

//  V_DrawPatch((SCREENWIDTH - lnames[wbs.next].width) div 2, y, _FB, lnames[wbs.next]);
  V_DrawPatch((320 - lnames[wbs.next].width) div 2, y, _FG, lnames[wbs.next], true);
end;

procedure WI_drawOnLnode(n: integer; c: Ppatch_tPArray);
var
  i: integer;
  left: integer;
  top: integer;
  right: integer;
  bottom: integer;
  fits: boolean;
begin
  fits := false;

  i := 0;
  repeat
    left := lnodes[wbs.epsd][n].x - c[i].leftoffset;
    top := lnodes[wbs.epsd][n].y - c[i].topoffset;
    right := left + c[i].width;
    bottom := top + c[i].height;

{    if (left >= 0) and
       (right < SCREENWIDTH) and
       (top >= 0) and
       (bottom < SCREENHEIGHT) then
      fits := true
    else
      inc(i);}
    if (left >= 0) and
       (right < 320) and
       (top >= 0) and
       (bottom < 200) then
      fits := true
    else
      inc(i);
  until not ((not fits) and (i <> 2));

  if fits and (i < 2) then
    V_DrawPatch(lnodes[wbs.epsd][n].x, lnodes[wbs.epsd][n].y, _FG, c[i], true)
  else
    // DEBUG
    printf('WI_drawOnLnode(): Could not place patch on level %d', [n + 1]);
end;

procedure WI_initAnimatedBack;
var
  i: integer;
  a: Pwianim_t;
begin
  if gamemode = commercial then
    exit;

  if wbs.epsd > 2 then
    exit;

  for i := 0 to NUMANIMS[wbs.epsd] - 1 do
  begin
    a := @wi_stuff_anims[wbs.epsd, i];

    // init variables
    a.ctr := -1;

    // specify the next time to draw it
    if a._type = ANIM_ALWAYS then
      a.nexttic := bcnt + 1 + (M_Random mod a.period)
    else if a._type = ANIM_RANDOM then
      a.nexttic := bcnt + 1 + a.data2 + (M_Random mod a.data1)
    else if a._type = ANIM_LEVEL then
      a.nexttic := bcnt + 1;
  end;
end;

procedure WI_updateAnimatedBack;
var
  i: integer;
  a: Pwianim_t;
begin
  if gamemode = commercial then
    exit;

  if wbs.epsd > 2 then
    exit;

  for i := 0 to NUMANIMS[wbs.epsd] - 1 do
  begin
    a := @wi_stuff_anims[wbs.epsd, i];

    if bcnt = a.nexttic then
    begin
      case a._type of
        ANIM_ALWAYS:
          begin
            a.ctr := a.ctr + 1;
            if a.ctr >= a.nanims then
              a.ctr := 0;
            a.nexttic := bcnt + a.period;
          end;
        ANIM_RANDOM:
          begin
            a.ctr := a.ctr + 1;
            if a.ctr = a.nanims then
            begin
              a.ctr := -1;
              a.nexttic := bcnt + a.data2 + (M_Random mod a.data1);
            end
            else
              a.nexttic := bcnt + a.period;
          end;
        ANIM_LEVEL:
          begin
            // gawd-awful hack for level anims
            if (not ((state = StatCount) and (i = 7))) and
               (wbs.next = a.data1) then
            begin
              a.ctr := a.ctr + 1;
              if a.ctr = a.nanims then
                a.ctr := a.ctr - 1;
              a.nexttic := bcnt + a.period;
            end;
          end;
      end;
    end;
  end;
end;

procedure WI_drawAnimatedBack;
var
  i: integer;
  a: Pwianim_t;
begin
  if gamemode = commercial then // VJ was if (commercial)
    exit;

  if wbs.epsd > 2 then
    exit;

  for i := 0 to NUMANIMS[wbs.epsd] - 1 do
  begin
    a := @wi_stuff_anims[wbs.epsd, i];

    if a.ctr >= 0 then
      V_DrawPatch(a.loc.x, a.loc.y, _FG, a.p[a.ctr], true);
  end;
end;

//
// Draws a number.
// If digits > 0, then use that many digits minimum,
//  otherwise only use as many as necessary.
// Returns new x position.
//

function WI_drawNum(x, y: integer; n: integer; digits: integer): integer;
var
  fontwidth: integer;
  neg: boolean;
  temp: integer;
begin
  fontwidth := num[0].width;
  if digits < 0 then
  begin
    if not boolval(n) then
      // make variable-length zeros 1 digit long
      digits := 1
    else
    begin
      // figure out # of digits in #
      digits := 0;
      temp := n;

      while boolval(temp) do
      begin
        temp := temp div 10;
        inc(digits);
      end;

    end;
  end;

  neg := n < 0;
  if neg then
    n := -n;

  // if non-number, do not draw it
  if n = 1994 then
  begin
    result := 0;
    exit;
  end;

  // draw the new number
  while boolval(digits) do
  begin
    x := x - fontwidth;
    V_DrawPatch(x, y, _FG, num[n mod 10], true);
    n := n div 10;
    dec(digits);
  end;

  // draw a minus sign if necessary
  if neg then
  begin
    x := x - 8;
    V_DrawPatch(x, y, _FG, wiminus, true);
  end;

  result := x;
end;

procedure WI_drawPercent(x, y: integer; p: integer);
begin
  if p < 0 then
    exit;

  V_DrawPatch(x, y, _FG, percent, true);
  WI_drawNum(x, y, p, -1);
end;

//
// Display level completion time and par,
//  or "sucks" message if overflow.
//
procedure WI_drawTime(x, y: integer; t: integer);
var
  _div: integer;
  n: integer;
begin
  if t < 0 then
    exit;

  if t <= 61 * 59 then
  begin
    _div := 1;

    repeat
      n := (t div _div) mod 60;
      x := WI_drawNum(x, y, n, 2) - colon.width;
      _div := _div * 60;

      // draw
      if (_div = 60) or boolval(t div _div) then
        V_DrawPatch(x, y, _FG, colon, true);
    until not boolval(t div _div);
  end
  else
    // "sucks"
    V_DrawPatch(x - sucks.width, y, _FG, sucks, true);
end;

procedure WI_unloadData; forward;

procedure WI_End;
begin
  WI_unloadData;
end;

procedure WI_initNoState;
begin
  state := NoState;
  acceleratestage := 0;
  cnt := 10;
end;

procedure WI_updateNoState;
begin
  WI_updateAnimatedBack;

  dec(cnt);
  if not boolval(cnt) then
  begin
    WI_End;
    G_WorldDone;
  end;
end;

var
  snl_pointeron: boolean = false;

procedure WI_initShowNextLoc;
begin
  state := ShowNextLoc;
  acceleratestage := 0;
  cnt := SHOWNEXTLOCDELAY * TICRATE;

  WI_initAnimatedBack;
end;

procedure WI_updateShowNextLoc;
begin
  WI_updateAnimatedBack;

  dec(cnt);
  if (not boolval(cnt)) or boolval(acceleratestage) then
    WI_initNoState
  else
    snl_pointeron := (cnt and 31) < 20;
end;


procedure WI_drawShowNextLoc;
var
  i: integer;
  last: integer;
begin
  WI_slamBackground;

  // draw animated background
  WI_drawAnimatedBack();

  if gamemode <> commercial then
  begin
    if wbs.epsd > 2 then
    begin
      WI_drawEL;
      exit;
    end;

    last := decide(wbs.last = 8, wbs.next - 1, wbs.last);

    // draw a splat on taken cities.
    for i := 0 to last do
      WI_drawOnLnode(i, @splat);

    // splat the secret level?
    if wbs.didsecret then
      WI_drawOnLnode(8, @splat);

    // draw flashing ptr
    if snl_pointeron then
      WI_drawOnLnode(wbs.next, @yah);
  end;

  // draws which level you are entering..
  if (gamemode <> commercial) or
     (wbs.next <> 30) then
    WI_drawEL;
end;

procedure WI_drawNoState;
begin
  snl_pointeron := true;
  WI_drawShowNextLoc;
end;

function WI_fragSum(playernum: integer): integer;
var
  i: integer;
begin
  result := 0;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] and (i <> playernum) then
      result := result + plrs[playernum].frags[i];
  end;

  // JDC hack - negative frags.
  result := result - plrs[playernum].frags[playernum];
    // UNUSED if (result < 0)
    // 	result = 0;
end;

var
  dm_state: integer;
  dm_frags: array[0..MAXPLAYERS - 1, 0..MAXPLAYERS - 1] of integer;
  dm_totals: array[0..MAXPLAYERS - 1] of integer;

procedure WI_initDeathmatchStats;
var
  i: integer;
  j: integer;
begin
  state := StatCount;
  acceleratestage := 0;
  dm_state := 1;

  cnt_pause := TICRATE;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] then
    begin
      for j := 0 to MAXPLAYERS - 1 do
        if playeringame[j] then
          dm_frags[i, j] := 0;

      dm_totals[i] := 0;
    end;
  end;

  WI_initAnimatedBack;
end;

procedure WI_updateDeathmatchStats;
var
  i: integer;
  j: integer;
  stillticking: boolean;
begin
  WI_updateAnimatedBack;

  if boolval(acceleratestage) and (dm_state <> 4) then
  begin
    acceleratestage := 0;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if playeringame[i] then
      begin
        for j := 0 to MAXPLAYERS - 1 do
          if playeringame[j] then
            dm_frags[i, j] := plrs[i].frags[j];

        dm_totals[i] := WI_fragSum(i);
      end;
    end;

    S_StartSound(nil, Ord(sfx_barexp));
    dm_state := 4;
  end;

  if dm_state = 2 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    stillticking := false;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if playeringame[i] then
      begin
        for j := 0 to MAXPLAYERS - 1 do
        begin
          if playeringame[j] and (dm_frags[i, j] <> plrs[i].frags[j]) then
          begin
            if plrs[i].frags[j] < 0 then
              dm_frags[i, j] := dm_frags[i, j] - 1
            else
              dm_frags[i, j] := dm_frags[i, j] + 1;

            if dm_frags[i, j] > 99 then
              dm_frags[i, j] := 99
            else if dm_frags[i, j] < -99 then
              dm_frags[i, j] := -99;
            stillticking := true;
          end;
        end;
        dm_totals[i] := WI_fragSum(i);

        if dm_totals[i] > 99 then
          dm_totals[i] := 99
        else if dm_totals[i] < -99 then
          dm_totals[i] := -99;
      end;
    end;
    if not stillticking then
    begin
      S_StartSound(nil, Ord(sfx_barexp));
      inc(dm_state);
    end;
  end
  else if dm_state = 4 then
  begin
    if boolval(acceleratestage) then
    begin
      S_StartSound(nil, Ord(sfx_slop));

      if gamemode = commercial then
        WI_initNoState
      else
        WI_initShowNextLoc;
    end;
  end
  else if boolval(dm_state and 1) then
  begin
    dec(cnt_pause);
    if not boolval(cnt_pause) then
    begin
      inc(dm_state);
      cnt_pause := TICRATE;
    end;
  end;
end;

procedure WI_drawDeathmatchStats;
var
  i: integer;
  j: integer;
  x: integer;
  y: integer;
  w: integer;
begin

  WI_slamBackground;

  // draw animated background
  WI_drawAnimatedBack;
  WI_drawLF;

  // draw stat titles (top line)
  V_DrawPatch(DM_TOTALSX - total.width div 2, DM_MATRIXY - WI_SPACINGY + 10, _FG, total, true);

  V_DrawPatch(DM_KILLERSX, DM_KILLERSY, _FG, killers, true);
  V_DrawPatch(DM_VICTIMSX, DM_VICTIMSY, _FG, victims, true);

  // draw P?
  x := DM_MATRIXX + DM_SPACINGX;
  y := DM_MATRIXY;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] then
    begin
      V_DrawPatch(x - p[i].width div 2, DM_MATRIXY - WI_SPACINGY, _FG, p[i], true);
      V_DrawPatch(DM_MATRIXX - p[i].width div 2, y, _FG, p[i], true);

      if i = me then
      begin
        V_DrawPatch(x - p[i].width div 2, DM_MATRIXY - WI_SPACINGY, _FG, bstar, true);
        V_DrawPatch(DM_MATRIXX - p[i].width div 2, y, _FG, star, true);
      end;
    end
    else
    begin
      // V_DrawPatch(x-SHORT(bp[i]->width)/2,
      //   DM_MATRIXY - WI_SPACINGY, FB, bp[i]);
      // V_DrawPatch(DM_MATRIXX-SHORT(bp[i]->width)/2,
      //   y, FB, bp[i]);
    end;
    x := x + DM_SPACINGX;
    y := y + WI_SPACINGY;
  end;

  // draw stats
  y := DM_MATRIXY + 10;
  w := num[0].width;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    x := DM_MATRIXX + DM_SPACINGX;

    if playeringame[i] then
    begin
      for j := 0 to MAXPLAYERS - 1 do
      begin
        if playeringame[j] then
          WI_drawNum(x + w, y, dm_frags[i, j], 2);
        x := x + DM_SPACINGX;
      end;
      WI_drawNum(DM_TOTALSX + w, y, dm_totals[i], 2);
    end;
    y := y + WI_SPACINGY;
  end;
end;

var
  cnt_frags: array[0..MAXPLAYERS - 1] of integer;
  dofrags: integer = 0;
  ng_state: integer = 0;

function NG_STATSX: integer;
begin
  result := 32 + star.width div 2 + 32 * (not intval(boolval(dofrags))); // VJ ???
end;

procedure WI_initNetgameStats;
var
  i: integer;
begin
  state := StatCount;
  acceleratestage := 0;
  ng_state := 1;

  cnt_pause := TICRATE;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if not playeringame[i] then
      continue;

    cnt_kills[i] := 0;
    cnt_items[i] := 0;
    cnt_secret[i] := 0;
    cnt_frags[i] := 0;

    dofrags := dofrags + WI_fragSum(i);
  end;

  dofrags := intval(boolval(dofrags));

  WI_initAnimatedBack;
end;

procedure WI_updateNetgameStats;
var
  i: integer;
  fsum: integer;
  stillticking: boolean;
begin
  WI_updateAnimatedBack;

  if boolval(acceleratestage) and (ng_state <> 10) then
  begin
    acceleratestage := 0;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;

      cnt_kills[i] := (plrs[i].skills * 100) div wbs.maxkills;
      cnt_items[i] := (plrs[i].sitems * 100) div wbs.maxitems;
      cnt_secret[i] := (plrs[i].ssecret * 100) div wbs.maxsecret;

      if boolval(dofrags) then
        cnt_frags[i] := WI_fragSum(i);
    end;
    S_StartSound(nil, Ord(sfx_barexp));
    ng_state := 10;
  end;

  if ng_state = 2 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    stillticking := false;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;

      cnt_kills[i] := cnt_kills[i] + 2;

      if cnt_kills[i] >= (plrs[i].skills * 100) div wbs.maxkills then
        cnt_kills[i] := (plrs[i].skills * 100) div wbs.maxkills
      else
        stillticking := true;
    end;

    if not stillticking then
    begin
      S_StartSound(nil, Ord(sfx_barexp));
      inc(ng_state);
    end;
  end
  else if ng_state = 4 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    stillticking := false;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;

      cnt_items[i] := cnt_items[i] + 2;
      if cnt_items[i] >= (plrs[i].sitems * 100) div wbs.maxitems then
        cnt_items[i] := (plrs[i].sitems * 100) div wbs.maxitems
      else
        stillticking := true;
    end;
    if not stillticking then
    begin
      S_StartSound(nil, Ord(sfx_barexp));
      inc(ng_state);
    end;
  end
  else if ng_state = 6 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    stillticking := false;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;

      cnt_secret[i] := cnt_secret[i] + 2;

      if cnt_secret[i] >= (plrs[i].ssecret * 100) div wbs.maxsecret then
        cnt_secret[i] := (plrs[i].ssecret * 100) div wbs.maxsecret
      else
        stillticking := true;
    end;
    if not stillticking then
    begin
      S_StartSound(nil, Ord(sfx_barexp));
      ng_state := ng_state + 1 + 2 * intval(boolval(dofrags));
    end;
  end
  else if ng_state = 8 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    stillticking := false;

    for i := 0 to MAXPLAYERS - 1 do
    begin
      if not playeringame[i] then
        continue;

      cnt_frags[i] := cnt_frags[i] + 1;

      fsum := WI_fragSum(i);
      if cnt_frags[i] >= fsum then
        cnt_frags[i] := fsum
      else
        stillticking := true;
    end;
    if not stillticking then
    begin
      S_StartSound(nil, Ord(sfx_pldeth));
      inc(ng_state);
    end;
  end
  else if ng_state = 10 then
  begin
    if boolval(acceleratestage) then
    begin
      S_StartSound(nil, Ord(sfx_sgcock));
      if gamemode = commercial then
        WI_initNoState
      else
        WI_initShowNextLoc;
    end;
  end
  else if boolval(ng_state and 1) then
  begin
    dec(cnt_pause);
    if not boolval(cnt_pause) then
    begin
      inc(ng_state);
      cnt_pause := TICRATE;
    end;
  end;
end;

procedure WI_drawNetgameStats;
var
  i: integer;
  x: integer;
  y: integer;
  pwidth: integer;
begin
  pwidth := percent.width;

  WI_slamBackground;

  // draw animated background
  WI_drawAnimatedBack;

  WI_drawLF;

  // draw stat titles (top line)
  V_DrawPatch(NG_STATSX + NG_SPACINGX - kills.width, NG_STATSY, _FG, kills, true);

  V_DrawPatch(NG_STATSX + 2 * NG_SPACINGX - items.width, NG_STATSY, _FG, items, true);

  V_DrawPatch(NG_STATSX + 3 * NG_SPACINGX - secret.width, NG_STATSY, _FG, secret, true);

  if boolval(dofrags) then
    V_DrawPatch(NG_STATSX + 4 * NG_SPACINGX - frags.width, NG_STATSY, _FG, frags, true);

  // draw stats
  y := NG_STATSY + kills.height;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if not playeringame[i] then
      continue;

    x := NG_STATSX;
    V_DrawPatch(x - p[i].width, y, _FG, p[i], true);

    if i = me then
      V_DrawPatch(x - p[i].width, y, _FG, star, true);

    x := x + NG_SPACINGX;
    WI_drawPercent(x - pwidth, y + 10, cnt_kills[i]);
    x := x + NG_SPACINGX;
    WI_drawPercent(x - pwidth, y + 10, cnt_items[i]);
    x := x + NG_SPACINGX;
    WI_drawPercent(x - pwidth, y + 10, cnt_secret[i]);
    x := x + NG_SPACINGX;

    if boolval(dofrags) then
      WI_drawNum(x, y + 10, cnt_frags[i], -1);

    y := y + WI_SPACINGY;
  end;
end;

var
  sp_state: integer = 0;

procedure WI_initStats;
begin
  state := StatCount;
  acceleratestage := 0;
  sp_state := 1;
  cnt_kills[0] := -1;
  cnt_items[0] := -1;
  cnt_secret[0] := -1;
  cnt_time := -1;
  cnt_par := -1;
  cnt_pause := TICRATE;

  WI_initAnimatedBack;
end;

procedure WI_updateStats;
begin
  WI_updateAnimatedBack;

  if boolval(acceleratestage) and (sp_state <> 10) then
  begin
    acceleratestage := 0;
    cnt_kills[0] := (plrs[me].skills * 100) div wbs.maxkills;
    cnt_items[0] := (plrs[me].sitems * 100) div wbs.maxitems;
    cnt_secret[0] := (plrs[me].ssecret * 100) div wbs.maxsecret;
    cnt_time := plrs[me].stime div TICRATE;
    cnt_par := wbs.partime div TICRATE;
    S_StartSound(nil, Ord(sfx_barexp));
    sp_state := 10;
  end;

  if sp_state = 2 then
  begin
    cnt_kills[0] := cnt_kills[0] + 2;

    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    if cnt_kills[0] >= (plrs[me].skills * 100) div wbs.maxkills then
    begin
      cnt_kills[0] := (plrs[me].skills * 100) div wbs.maxkills;
      S_StartSound(nil, Ord(sfx_barexp));
      inc(sp_state);
    end;
  end
  else if sp_state = 4 then
  begin
    cnt_items[0] := cnt_items[0] + 2;

    if boolval(bcnt and 3) then
      S_StartSound(nil, OrD(sfx_pistol));

    if cnt_items[0] >= (plrs[me].sitems * 100) div wbs.maxitems then
    begin
      cnt_items[0] := (plrs[me].sitems * 100) div wbs.maxitems;
      S_StartSound(nil, Ord(sfx_barexp));
      inc(sp_state);
    end;
  end
  else if sp_state = 6 then
  begin
    cnt_secret[0] := cnt_secret[0] + 2;

    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    if cnt_secret[0] >= (plrs[me].ssecret * 100) div wbs.maxsecret then
    begin
      cnt_secret[0] := (plrs[me].ssecret * 100) div wbs.maxsecret;
      S_StartSound(nil, Ord(sfx_barexp));
      inc(sp_state);
    end;
  end
  else if sp_state = 8 then
  begin
    if not boolval(bcnt and 3) then
      S_StartSound(nil, Ord(sfx_pistol));

    cnt_time := cnt_time + 3;

    if cnt_time >= plrs[me].stime div TICRATE then
      cnt_time := plrs[me].stime div TICRATE;

    cnt_par := cnt_par + 3;

    if cnt_par >= wbs.partime div TICRATE then
    begin
	    cnt_par := wbs.partime div TICRATE;

      if cnt_time >= plrs[me].stime div TICRATE then
      begin
        S_StartSound(nil, Ord(sfx_barexp));
        inc(sp_state);
      end;
    end;
  end
  else if sp_state = 10 then
  begin
    if boolval(acceleratestage) then
    begin
      S_StartSound(nil, Ord(sfx_sgcock));

      if gamemode = commercial then
        WI_initNoState
      else
        WI_initShowNextLoc;
    end;
  end
  else if boolval(sp_state and 1) then
  begin
    dec(cnt_pause);
    if not boolval(cnt_pause) then
    begin
      inc(sp_state);
      cnt_pause := TICRATE;
    end;
  end;
end;

procedure WI_drawStats;
var
  // line height
  lh: integer;
begin
  lh := (3 * num[0].height) div 2;

  WI_slamBackground;

  // draw animated background
  WI_drawAnimatedBack;

  WI_drawLF;

  V_DrawPatch(SP_STATSX, SP_STATSY, _FG, kills, true);
//  WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY, cnt_kills[0]);
  WI_drawPercent(320 - SP_STATSX, SP_STATSY, cnt_kills[0]);

  V_DrawPatch(SP_STATSX, SP_STATSY + lh, _FG, items, true);
//  WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY + lh, cnt_items[0]);
  WI_drawPercent(320 - SP_STATSX, SP_STATSY + lh, cnt_items[0]);

  V_DrawPatch(SP_STATSX, SP_STATSY + 2 * lh, _FG, sp_secret, true);
//  WI_drawPercent(SCREENWIDTH - SP_STATSX, SP_STATSY + 2 * lh, cnt_secret[0]);
  WI_drawPercent(320 - SP_STATSX, SP_STATSY + 2 * lh, cnt_secret[0]);

  V_DrawPatch(SP_TIMEX, SP_TIMEY, _FG, time, true);
//  WI_drawTime(SCREENWIDTH div 2 - SP_TIMEX, SP_TIMEY, cnt_time);
  WI_drawTime(160 - SP_TIMEX, SP_TIMEY, cnt_time);

  if wbs.epsd < 3 then
  begin
//    V_DrawPatch(SCREENWIDTH div 2 + SP_TIMEX, SP_TIMEY, _FB, par);
//    WI_drawTime(SCREENWIDTH - SP_TIMEX, SP_TIMEY, cnt_par);
    V_DrawPatch(160 + SP_TIMEX, SP_TIMEY, _FG, par, true);
    WI_drawTime(320 - SP_TIMEX, SP_TIMEY, cnt_par);
  end;
end;

procedure WI_checkForAccelerate;
var
  i: integer;
  player: Pplayer_t;
begin
  // check for button presses to skip delays
  for i := 0 to MAXPLAYERS - 1 do
  begin
    player := @players[i];

    if playeringame[i] then
    begin

      if boolval(player.cmd.buttons and BT_ATTACK) then
      begin
        if not player.attackdown then
          acceleratestage := 1;
        player.attackdown := true;
      end
      else
        player.attackdown := false;

	    if boolval(player.cmd.buttons and BT_USE) then
      begin
        if not player.usedown then
          acceleratestage := 1;
        player.usedown := true;
      end
      else
        player.usedown := false;
    end;
  end;
end;

// Updates stuff each tick
procedure WI_Ticker;
begin
  // counter for general background animation
  inc(bcnt);

  if bcnt = 1 then
  begin
    // intermission music
    if gamemode = commercial then
      S_ChangeMusic(Ord(mus_dm2int), true)
    else
      S_ChangeMusic(Ord(mus_inter), true);
  end;

  WI_checkForAccelerate;

  case state of
    StatCount:
      begin
        if boolval(deathmatch) then
          WI_updateDeathmatchStats
        else if netgame then
          WI_updateNetgameStats
        else
          WI_updateStats;
      end;

    ShowNextLoc:
      begin
        WI_updateShowNextLoc();
      end;

    NoState:
      begin
        WI_updateNoState();
      end;
  end;
end;

procedure WI_loadData;
var
  i: integer;
  j: integer;
  name: string;
  a: Pwianim_t;
begin
  if gamemode = commercial then
    name := 'INTERPIC'
  else
    sprintf(name, 'WIMAP%d', [wbs.epsd]);

  if gamemode = retail then
    if wbs.epsd = 3 then
      name := 'INTERPIC';

  // background
  bg := W_CacheLumpName(name, PU_CACHE);
  V_DrawPatch(0, 0, 1, bg, true);


    // UNUSED unsigned char *pic = screens[1];
    // if (gamemode == commercial)
    // {
    // darken the background image
    // while (pic != screens[1] + SCREENHEIGHT*SCREENWIDTH)
    // {
    //   *pic = colormaps[256*25 + *pic];
    //   pic++;
    // }
    //}

  if gamemode = commercial then
  begin
    NUMCMAPS := 32;
    lnames :=
      Ppatch_tPArray(
        Z_Malloc(SizeOf(Ppatch_t) * NUMCMAPS, PU_STATIC, nil));
    for i := 0 to NUMCMAPS - 1 do
    begin
      name := 'CWILV' + IntToStrZfill(2, i); // VJ was sprintf(name, 'CWILV%2.2d', i);
      lnames[i] := W_CacheLumpName(name, PU_STATIC);
    end;
  end
  else
  begin
    lnames :=
      Ppatch_tPArray(
        Z_Malloc(SizeOf(Ppatch_t) * NUMMAPS, PU_STATIC, nil));
    for i := 0 to NUMMAPS - 1 do
    begin
      sprintf(name, 'WILV%d%d', [wbs.epsd, i]);
      lnames[i] := W_CacheLumpName(name, PU_STATIC);
    end;

    // you are here
    yah[0] := W_CacheLumpName('WIURH0', PU_STATIC);

    // you are here (alt.)
    yah[1] := W_CacheLumpName('WIURH1', PU_STATIC);

    // splat
    splat := W_CacheLumpName('WISPLAT', PU_STATIC);

    if wbs.epsd < 3 then
    begin
      for j := 0 to NUMANIMS[wbs.epsd] - 1 do
      begin
        a := @wi_stuff_anims[wbs.epsd, j];
        for i := 0 to a.nanims - 1 do
        begin
          // MONDO HACK!
          if (wbs.epsd <> 1) or (j <> 8) then
          begin
            // animations
            sprintf(name, 'WIA%d%.2d%.2d', [wbs.epsd, j, i]); // VJ ????
            a.p[i] := W_CacheLumpName(name, PU_STATIC);
          end
          else
          begin
          // HACK ALERT!
            a.p[i] := wi_stuff_anims[1, 4].p[i];
          end;
        end;
      end;
    end;
  end;

  // More hacks on minus sign.
  wiminus := W_CacheLumpName('WIMINUS', PU_STATIC);

  for i :=0 to 9 do
  begin
    // numbers 0-9
    sprintf(name, 'WINUM%d', [i]);
    num[i] := W_CacheLumpName(name, PU_STATIC);
  end;

  // percent sign
  percent := W_CacheLumpName('WIPCNT', PU_STATIC);

  // "finished"
  finished := W_CacheLumpName('WIF', PU_STATIC);

  // "entering"
  entering := W_CacheLumpName('WIENTER', PU_STATIC);

  // "kills"
  kills := W_CacheLumpName('WIOSTK', PU_STATIC);

  // "scrt"
  secret := W_CacheLumpName('WIOSTS', PU_STATIC);

  // "secret"
  sp_secret := W_CacheLumpName('WISCRT2', PU_STATIC);

  // Yuck.
  if language = french then
  begin
    // "items"
    if netgame and (not boolval(deathmatch)) then
      items := W_CacheLumpName('WIOBJ', PU_STATIC)
    else
      items := W_CacheLumpName('WIOSTI', PU_STATIC);
  end
  else
    items := W_CacheLumpName('WIOSTI', PU_STATIC);

  // "frgs"
  frags := W_CacheLumpName('WIFRGS', PU_STATIC);

  // ":"
  colon := W_CacheLumpName('WICOLON', PU_STATIC);

  // "time"
  time := W_CacheLumpName('WITIME', PU_STATIC);

  // "sucks"
  sucks := W_CacheLumpName('WISUCKS', PU_STATIC);

  // "par"
  par := W_CacheLumpName('WIPAR', PU_STATIC);

  // "killers" (vertical)
  killers := W_CacheLumpName('WIKILRS', PU_STATIC);

  // "victims" (horiz)
  victims := W_CacheLumpName('WIVCTMS', PU_STATIC);

  // "total"
  total := W_CacheLumpName('WIMSTT', PU_STATIC);

  // your face
  star := W_CacheLumpName('STFST01', PU_STATIC);

  // dead face
  bstar := W_CacheLumpName('STFDEAD0', PU_STATIC);

  for i := 0 to MAXPLAYERS - 1 do
  begin
    // "1,2,3,4"
    sprintf(name, 'STPB%d', [i]);
    p[i] := W_CacheLumpName(name, PU_STATIC);

    // "1,2,3,4"
    sprintf(name, 'WIBP%d', [i + 1]);
    bp[i] := W_CacheLumpName(name, PU_STATIC);
  end;
end;

procedure WI_unloadData;
var
  i: integer;
  j: integer;
begin
  Z_ChangeTag(wiminus, PU_CACHE);

  for i := 0 to 9 do
    Z_ChangeTag(num[i], PU_CACHE);

  if gamemode = commercial then
  begin
    for i := 0 to NUMCMAPS - 1 do
      Z_ChangeTag(lnames[i], PU_CACHE);
  end
  else
  begin
    Z_ChangeTag(yah[0], PU_CACHE);
    Z_ChangeTag(yah[1], PU_CACHE);

    Z_ChangeTag(splat, PU_CACHE);

    for i := 0 to NUMMAPS - 1 do
      Z_ChangeTag(lnames[i], PU_CACHE);

    if wbs.epsd < 3 then
    begin
      for j := 0 to NUMANIMS[wbs.epsd] - 1 do
      begin
        if (wbs.epsd <> 1) or (j <> 8) then
          for i := 0 to wi_stuff_anims[wbs.epsd, j].nanims - 1 do
            Z_ChangeTag(wi_stuff_anims[wbs.epsd, j].p[i], PU_CACHE);
      end;
    end;
  end;

  Z_Free(lnames);

  Z_ChangeTag(percent, PU_CACHE);
  Z_ChangeTag(colon, PU_CACHE);
  Z_ChangeTag(finished, PU_CACHE);
  Z_ChangeTag(entering, PU_CACHE);
  Z_ChangeTag(kills, PU_CACHE);
  Z_ChangeTag(secret, PU_CACHE);
  Z_ChangeTag(sp_secret, PU_CACHE);
  Z_ChangeTag(items, PU_CACHE);
  Z_ChangeTag(frags, PU_CACHE);
  Z_ChangeTag(time, PU_CACHE);
  Z_ChangeTag(sucks, PU_CACHE);
  Z_ChangeTag(par, PU_CACHE);

  Z_ChangeTag(victims, PU_CACHE);
  Z_ChangeTag(killers, PU_CACHE);
  Z_ChangeTag(total, PU_CACHE);
//  Z_ChangeTag(star, PU_CACHE);
//  Z_ChangeTag(bstar, PU_CACHE);

  for i := 0 to MAXPLAYERS - 1 do
    Z_ChangeTag(p[i], PU_CACHE);

  for i := 0 to MAXPLAYERS - 1 do
    Z_ChangeTag(bp[i], PU_CACHE);
end;

procedure WI_Drawer;
begin
  case state of
    StatCount:
      begin
        if boolval(deathmatch) then
          WI_drawDeathmatchStats
        else if netgame then
          WI_drawNetgameStats
        else
          WI_drawStats;
      end;

    ShowNextLoc:
      begin
        WI_drawShowNextLoc;
      end;

    NoState:
      begin
        WI_drawNoState;
      end;
  end;
end;

procedure WI_initVariables(wbstartstruct: Pwbstartstruct_t);
begin
  wbs := wbstartstruct;

  acceleratestage := 0;
  cnt := 0;
  bcnt := 0;
  firstrefresh := 1;
  me := wbs.pnum;
  plrs := @wbs.plyr;

  if not boolval(wbs.maxkills) then
    wbs.maxkills := 1;

  if not boolval(wbs.maxitems) then
    wbs.maxitems := 1;

  if not boolval(wbs.maxsecret) then
    wbs.maxsecret := 1;

  if gamemode <> retail then
    if wbs.epsd > 2 then
      wbs.epsd := wbs.epsd - 3;
end;

procedure WI_Start(wbstartstruct: Pwbstartstruct_t);
begin
  WI_initVariables(wbstartstruct);
  WI_loadData;

  if boolval(deathmatch) then
    WI_initDeathmatchStats
  else if netgame then
    WI_initNetgameStats
  else
    WI_initStats;
end;


(* Unit p_enemy implementation *)


                                                                                                                                                                                                                                                        

type
  dirtype_t = (
    DI_EAST,
    DI_NORTHEAST,
    DI_NORTH,
    DI_NORTHWEST,
    DI_WEST,
    DI_SOUTHWEST,
    DI_SOUTH,
    DI_SOUTHEAST,
    DI_NODIR,
    NUMDIRS
  );

const
  opposite: array[0..8] of dirtype_t = (
    DI_WEST, DI_SOUTHWEST, DI_SOUTH, DI_SOUTHEAST,
    DI_EAST, DI_NORTHEAST, DI_NORTH, DI_NORTHWEST, DI_NODIR
  );

  diags: array[0..3] of dirtype_t = (
    DI_NORTHWEST, DI_NORTHEAST, DI_SOUTHWEST, DI_SOUTHEAST
  );

procedure A_Fall(actor: Pmobj_t);
begin
  // actor is on ground, it can be walked over
  actor.flags := actor.flags and (not MF_SOLID);

  // So change this if corpse objects
  // are meant to be obstacles.
end;

//
// ENEMY THINKING
// Enemies are allways spawned
// with targetplayer = -1, threshold = 0
// Most monsters are spawned unaware of all players,
// but some can be made preaware
//


//
// Called by P_NoiseAlert.
// Recursively traverse adjacent sectors,
// sound blocking lines cut off traversal.
//

var
  soundtarget: Pmobj_t;

procedure P_RecursiveSound(sec: Psector_t; soundblocks: integer);
var
  i: integer;
  check: Pline_t;
  other: Psector_t;
begin
  // wake up all monsters in this sector
  if (sec.validcount = validcount) and
     (sec.soundtraversed <= soundblocks + 1) then
    exit; // already flooded

  sec.validcount := validcount;
  sec.soundtraversed := soundblocks + 1;
  sec.soundtarget := soundtarget;

  for i := 0 to sec.linecount - 1 do
  begin
    check := sec.lines[i];
    if (check.flags and ML_TWOSIDED) = 0 then
      continue;

    P_LineOpening(check);

    if openrange <= 0 then
      continue; // closed door

    if sides[check.sidenum[0]].sector = sec then
      other := sides[check.sidenum[1]].sector
    else
      other := sides[check.sidenum[0]].sector;

    if (check.flags and ML_SOUNDBLOCK) <> 0 then
    begin
      if soundblocks = 0 then
        P_RecursiveSound(other, 1);
    end
    else
      P_RecursiveSound(other, soundblocks);
  end;
end;

//
// P_NoiseAlert
// If a monster yells at a player,
// it will alert other monsters to the player.
//
procedure P_NoiseAlert(target: Pmobj_t; emmiter: Pmobj_t);
begin
  soundtarget := target;
  inc(validcount);
  P_RecursiveSound(Psubsector_t(emmiter.subsector).sector, 0);
end;

//
// P_CheckMeleeRange
//
function P_CheckMeleeRange(actor: Pmobj_t): boolean;
var
  pl: Pmobj_t;
  dist: fixed_t;
begin
  if actor.target = nil then
  begin
    result := false;
    exit;
  end;

  pl := actor.target;
  dist := P_AproxDistance(pl.x - actor.x, pl.y - actor.y);

  if dist >= MELEERANGE - 20 * FRACUNIT + pl.info.radius then
  begin
    result := false;
    exit;
  end;

  result := P_CheckSight(actor, actor.target);
end;

//
// P_CheckMissileRange
//
function P_CheckMissileRange(actor: Pmobj_t): boolean;
var
  dist: fixed_t;
begin
  if not P_CheckSight(actor, actor.target) then
  begin
    result := false;
    exit;
  end;

  if (actor.flags and MF_JUSTHIT) <> 0 then
  begin
    // the target just hit the enemy,
    // so fight back!
    actor.flags := actor.flags and (not MF_JUSTHIT);
    result := true;
    exit;
  end;

  if actor.reactiontime <> 0 then
  begin
    result := false; // do not attack yet
    exit;
  end;

  // OPTIMIZE: get this from a global checksight
  dist := P_AproxDistance(actor.x - actor.target.x, actor.y - actor.target.y) -
            64 * FRACUNIT;

  if not boolval(actor.info.meleestate) then
    dist := dist - 128 * FRACUNIT;  // no melee attack, so fire more

  dist := _SHR(dist, 16);

  if actor._type = MT_VILE then
    if dist > 14 * 64 then
    begin
      result := false; // too far away
      exit;
    end;


  if actor._type = MT_UNDEAD then
  begin
    if dist < 196 then
    begin
      result := false; // close for fist attack
      exit;
    end;
    dist := _SHR(dist, 1);
  end;
  

  if (actor._type = MT_CYBORG) or
     (actor._type = MT_SPIDER) or
     (actor._type = MT_SKULL) then
    dist := _SHR(dist, 1);

  if dist > 200 then
    dist := 200;

  if (actor._type = MT_CYBORG) and (dist > 160) then
    dist := 160;

  if P_Random < dist then
    result := false
  else
    result := true;
end;

//
// P_Move
// Move in the current direction,
// returns false if the move is blocked.
//
const
  xspeed: array[0..7] of fixed_t =
    (FRACUNIT, 47000, 0, -47000, -FRACUNIT, -47000, 0, 47000);

  yspeed: array[0..7] of fixed_t =
    (0, 47000, FRACUNIT, 47000, 0, -47000, -FRACUNIT, -47000);

function P_Move(actor: Pmobj_t): boolean;
var
  tryx: fixed_t;
  tryy: fixed_t;
  ld: Pline_t;
  try_ok: boolean;
begin
  if actor.movedir = Ord(DI_NODIR) then
  begin
    result := false;
    exit;
  end;

  if Ord(actor.movedir) >= 8 then
    I_Error('P_Move(): Weird actor->movedir = %d', [Ord(actor.movedir)]);

  tryx := actor.x + actor.info.speed * xspeed[actor.movedir];
  tryy := actor.y + actor.info.speed * yspeed[actor.movedir];

  try_ok := P_TryMove(actor, tryx, tryy);

  if not try_ok then
  begin
    // open any specials
    if boolval(actor.flags and MF_FLOAT) and floatok then
    begin
      // must adjust height
      if actor.z < tmfloorz then
        actor.z := actor.z + FLOATSPEED
      else
        actor.z := actor.z - FLOATSPEED;

      actor.flags := actor.flags or MF_INFLOAT;
      result := true;
      exit;
    end;

    if numspechit = 0 then
    begin
      result := false;
      exit;
    end;

    actor.movedir := Ord(DI_NODIR);
    result := false;
    while boolval(numspechit) do
    begin
      dec(numspechit);
      ld := spechit[numspechit];
      // if the special is not a door
      // that can be opened,
      // return false
      if P_UseSpecialLine(actor, ld, 0) then
        result := true;
    end;
    exit;
  end
  else
    actor.flags := actor.flags and (not MF_INFLOAT);

  if not boolval(actor.flags and MF_FLOAT) then
    actor.z := actor.floorz;
  result := true;
end;

//
// TryWalk
// Attempts to move actor on
// in its current (ob->moveangle) direction.
// If blocked by either a wall or an actor
// returns FALSE
// If move is either clear or blocked only by a door,
// returns TRUE and sets...
// If a door is in the way,
// an OpenDoor call is made to start it opening.
//
function P_TryWalk(actor: Pmobj_t): boolean;
begin
  if not P_Move(actor) then
    result := false
  else
  begin
    actor.movecount := P_Random and 15;
    result := true;
  end;
end;

procedure P_NewChaseDir(actor: Pmobj_t);
var
  deltax: fixed_t;
  deltay: fixed_t;
  d: array[0..2] of dirtype_t;
  tdir: dirtype_t;
  olddir: dirtype_t;
  turnaround: dirtype_t;
  idx: integer; 
begin
  if actor.target = nil then
    I_Error('P_NewChaseDir(): called with no target');

  olddir := dirtype_t(actor.movedir);
  turnaround := opposite[Ord(olddir)];

  deltax := actor.target.x - actor.x;
  deltay := actor.target.y - actor.y;

  if deltax > 10 * FRACUNIT then
    d[1] := DI_EAST
  else if deltax < -10 * FRACUNIT then
    d[1] := DI_WEST
  else
    d[1] := DI_NODIR;

  if deltay < -10 * FRACUNIT then
    d[2] := DI_SOUTH
  else if deltay > 10 * FRACUNIT then
    d[2] := DI_NORTH
  else
    d[2] := DI_NODIR;

  // try direct route
  if (d[1] <> DI_NODIR) and (d[2] <> DI_NODIR) then
  begin
    if deltay < 0 then
      idx := 2
    else
      idx := 0;
    if deltax > 0 then
      inc(idx);
    actor.movedir := Ord(diags[idx]);
    if (actor.movedir <> Ord(turnaround)) and P_TryWalk(actor) then
      exit;
  end;

  // try other directions
  if (P_Random > 200) or (abs(deltay) > abs(deltax)) then
  begin
    tdir := d[1];
    d[1] := d[2];
    d[2] := tdir;
  end;

  if d[1] = turnaround then
    d[1] := DI_NODIR;
  if d[2] = turnaround then
    d[2] := DI_NODIR;

  if d[1] <> DI_NODIR then
  begin
    actor.movedir := Ord(d[1]);
    if P_TryWalk(actor) then
      exit; // either moved forward or attacked
  end;

  if d[2] <> DI_NODIR then
  begin
    actor.movedir := Ord(d[2]);
    if P_TryWalk(actor) then
      exit;
  end;

  // there is no direct path to the player,
  // so pick another direction.
  if olddir <> DI_NODIR then
  begin
    actor.movedir := Ord(olddir);
    if P_TryWalk(actor) then
      exit;
  end;

  // randomly determine direction of search
  if boolval(P_Random and 1) then
  begin
    for tdir := DI_EAST to DI_SOUTHEAST do
    begin
      if tdir <> turnaround then
      begin
        actor.movedir := Ord(tdir);
        if P_TryWalk(actor) then
          exit;
      end
    end
  end
  else
  begin
    for tdir := DI_SOUTHEAST downto DI_EAST do
    begin
      if tdir <> turnaround then
      begin
        actor.movedir := Ord(tdir);
        if P_TryWalk(actor) then
          exit;
      end;
    end;
  end;

  if turnaround <> DI_NODIR then
  begin
    actor.movedir := Ord(turnaround);
    if P_TryWalk(actor) then
      exit;
  end;

  actor.movedir := Ord(DI_NODIR); // can not move
end;

//
// P_LookForPlayers
// If allaround is false, only look 180 degrees in front.
// Returns true if a player is targeted.
//
function P_LookForPlayers(actor: Pmobj_t; allaround: boolean): boolean;
var
  c: integer;
  stop: integer;
  player: Pplayer_t;
  an: angle_t;
  dist: fixed_t;
  initial: boolean;
begin
  c := 0;
  stop := (actor.lastlook - 1) and 3;

  initial := true;
  while true do
  begin
    if initial then
      initial := false
    else
      actor.lastlook := (actor.lastlook + 1) and 3;

    if not playeringame[actor.lastlook] then
      continue;

    if (c = 2) or (actor.lastlook = stop) then
    begin
      // done looking
      result := false;
      exit;
    end;
    inc(c);

    player := @players[actor.lastlook];

    if player.health <= 0 then
      continue;   // dead

    if not P_CheckSight(actor, player.mo) then
      continue;   // out of sight

    if not allaround then
    begin
      an := R_PointToAngle2(
              actor.x,
              actor.y,
              player.mo.x,
              player.mo.y)
          - actor.angle;

      if (an > ANG90) and (an < ANG270) then
      begin
        dist := P_AproxDistance(player.mo.x - actor.x, player.mo.y - actor.y);
        // if real close, react anyway
        if dist > MELEERANGE then
          continue; // behind back
      end;
    end;

    actor.target := player.mo;
    result := true;
    exit;
  end;

  result := false;
end;


//
// A_KeenDie
// DOOM II special, map 32.
// Uses special tag 666.
//
procedure A_KeenDie(mo: Pmobj_t);
var
  th: Pthinker_t;
  mo2: Pmobj_t;
  junk: line_t;
begin
  A_Fall(mo);

  // scan the remaining thinkers
  // to see if all Keens are dead

  th := thinkercap.next;
  while th <> @thinkercap do
  begin
    if @th._function.acp1 = @P_MobjThinker then
    begin
      mo2 := Pmobj_t(th);
      if (mo2 <> mo) and (mo2._type = mo._type) and (mo2.health > 0) then
        exit; // other Keen not dead
    end;
    th := th.next;
  end;

  junk.tag := 666;
  EV_DoDoor(@junk, open);
end;

//
// ACTION ROUTINES
//

//
// A_Look
// Stay in state until a player is sighted.
//
procedure A_Look(actor: Pmobj_t);
var
  targ: Pmobj_t;
  seeyou: boolean;
  sound: integer;
begin
  actor.threshold := 0; // any shot will wake up
  targ := Psubsector_t(actor.subsector).sector.soundtarget;
  seeyou := false;
  if (targ <> nil) and ((targ.flags and MF_SHOOTABLE) <> 0) then
  begin
    actor.target := targ;

    if (actor.flags and MF_AMBUSH) <> 0 then
    begin
      if P_CheckSight(actor, actor.target) then
        seeyou := true;
    end
    else
      seeyou := true;
  end;

  if not seeyou then
  begin
    if not P_LookForPlayers(actor, false) then
      exit;
  end;

  // go into chase state
  if actor.info.seesound <> 0 then
  begin
    case sfxenum_t(actor.info.seesound) of
      sfx_posit1,
      sfx_posit2,
      sfx_posit3:
        sound := Ord(sfx_posit1) + P_Random mod 3;

      sfx_bgsit1,
      sfx_bgsit2:
        sound := Ord(sfx_bgsit1) + P_Random mod 2;
    else
      sound := actor.info.seesound;
    end;

    if (actor._type = MT_SPIDER) or (actor._type = MT_CYBORG) then
      // full volume
      S_StartSound(nil, sound)
    else
      S_StartSound(actor, sound);
  end;

  P_SetMobjState(actor, statenum_t(actor.info.seestate));
end;

//
// A_Chase
// Actor has a melee attack,
// so it tries to close as fast as possible
//
procedure A_Chase(actor: Pmobj_t);
var
  delta: integer;
  nomissile: boolean;
begin
  if boolval(actor.reactiontime) then
    actor.reactiontime := actor.reactiontime - 1;

  // modify target threshold
  if boolval(actor.threshold) then
  begin
    if (not boolval(actor.target)) or (actor.target.health <= 0) then
      actor.threshold := 0
    else
      actor.threshold := actor.threshold - 1;
  end;

  // turn towards movement direction if not there yet
  if actor.movedir < 8 then
  begin
    actor.angle := actor.angle and _SHLW(7, 29);
    delta := actor.angle - _SHLW(actor.movedir, 29);

    if delta > 0 then
      actor.angle := actor.angle - ANG90 div 2
    else if delta < 0 then
      actor.angle := actor.angle + ANG90 div 2;
  end;

  if (not boolval(actor.target)) or
     (not boolval(actor.target.flags and MF_SHOOTABLE)) then
  begin
    // look for a new target
    if P_LookForPlayers(actor, true) then
      exit; // got a new target

    P_SetMobjState(actor, statenum_t(actor.info.spawnstate));
    exit;
  end;

  // do not attack twice in a row
  if (actor.flags and MF_JUSTATTACKED) <> 0 then
  begin
    actor.flags := actor.flags and (not MF_JUSTATTACKED);
    if (gameskill <> sk_nightmare) and (not fastparm) then
      P_NewChaseDir(actor);
    exit;
  end;

  // check for melee attack
  if boolval(actor.info.meleestate) and P_CheckMeleeRange(actor) then
  begin
    if boolval(actor.info.attacksound) then
      S_StartSound(actor, actor.info.attacksound);

    P_SetMobjState(actor, statenum_t(actor.info.meleestate));
    exit;
  end;

  nomissile := false;
  // check for missile attack
  if boolval(actor.info.missilestate) then
  begin
    if (gameskill < sk_nightmare) and (not fastparm) and boolval(actor.movecount) then
      nomissile := true
    else if not P_CheckMissileRange(actor) then
      nomissile := true;
    if not nomissile then
    begin
      P_SetMobjState(actor, statenum_t(actor.info.missilestate));
      actor.flags := actor.flags or MF_JUSTATTACKED;
      exit;
    end;
  end;

  // possibly choose another target
  if netgame and
    (not boolval(actor.threshold)) and
    (not P_CheckSight(actor, actor.target)) then
  begin
    if P_LookForPlayers(actor, true) then
      exit;  // got a new target
  end;

  // chase towards player
  actor.movecount := actor.movecount - 1;
  if (actor.movecount < 0) or (not P_Move(actor)) then
    P_NewChaseDir(actor);

  // make active sound
  if boolval(actor.info.activesound) and (P_Random < 3) then
    S_StartSound(actor, actor.info.activesound);
end;

//
// A_FaceTarget
//
procedure A_FaceTarget(actor: Pmobj_t);
begin
  if actor.target = nil then
    exit;

  actor.flags := actor.flags and (not MF_AMBUSH);

  actor.angle :=
    R_PointToAngle2(actor.x, actor.y, actor.target.x, actor.target.y);

  if boolval(actor.target.flags and MF_SHADOW) then
    actor.angle := actor.angle + _SHLW(P_Random - P_Random, 21);
end;

//
// A_PosAttack
//
procedure A_PosAttack(actor: Pmobj_t);
var
  angle: angle_t;
  damage: integer;
  slope: integer;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  angle := actor.angle;
  slope := P_AimLineAttack (actor, angle, MISSILERANGE);

  S_StartSound(actor, Ord(sfx_pistol));
  angle := angle + _SHLW(P_Random - P_Random, 20);
  damage := ((P_Random mod 5) + 1) * 3;
  P_LineAttack(actor, angle, MISSILERANGE, slope, damage);
end;

procedure A_SPosAttack(actor: Pmobj_t);
var
  i: integer;
  angle: angle_t;
  bangle: angle_t;
  damage: integer;
  slope: integer;
begin
  if actor.target = nil then
    exit;

  S_StartSound(actor, Ord(sfx_shotgn));
  A_FaceTarget(actor);
  bangle := actor.angle;
  slope := P_AimLineAttack(actor, bangle, MISSILERANGE);

  for i := 0 to 2 do
  begin
    angle := bangle + _SHLW(P_Random - P_Random, 20);
    damage := ((P_Random mod 5) + 1) * 3;
    P_LineAttack(actor, angle, MISSILERANGE, slope, damage);
  end;
end;

procedure A_CPosAttack(actor: Pmobj_t);
var
  angle: angle_t;
  bangle: angle_t;
  damage: integer;
  slope: integer;
begin
  if actor.target = nil then
    exit;

  S_StartSound(actor, Ord(sfx_shotgn));
  A_FaceTarget(actor);
  bangle := actor.angle;
  slope := P_AimLineAttack(actor, bangle, MISSILERANGE);

  angle := bangle + _SHLW(P_Random - P_Random, 20);
  damage := ((P_Random mod 5) + 1) * 3;
  P_LineAttack(actor, angle, MISSILERANGE, slope, damage);
end;

procedure A_CPosRefire(actor: Pmobj_t);
begin
  // keep firing unless target got out of sight
  A_FaceTarget(actor);

  if P_Random < 40 then
    exit;

  if (not boolval(actor.target)) or (actor.target.health <= 0) or
     (not P_CheckSight(actor, actor.target)) then
    P_SetMobjState(actor, statenum_t(actor.info.seestate));
end;

procedure A_SpidRefire(actor: Pmobj_t);
begin
  // keep firing unless target got out of sight
  A_FaceTarget(actor);

  if P_Random < 10 then
    exit;

  if (not boolval(actor.target)) or (actor.target.health <= 0) or
     (not P_CheckSight(actor, actor.target)) then
    P_SetMobjState(actor, statenum_t(actor.info.seestate));
end;

procedure A_BspiAttack(actor: Pmobj_t);
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);

  // launch a missile
  P_SpawnMissile(actor, actor.target, MT_ARACHPLAZ);
end;

//
// A_TroopAttack
//
procedure A_TroopAttack(actor: Pmobj_t);
var
  damage: integer;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  if P_CheckMeleeRange(actor) then
  begin
    S_StartSound(actor, Ord(sfx_claw));
    damage := (P_Random mod 8 + 1) * 3;
    P_DamageMobj(actor.target, actor, actor, damage);
    exit;
  end;

  // launch a missile
  P_SpawnMissile(actor, actor.target, MT_TROOPSHOT);
end;

procedure A_SargAttack(actor: Pmobj_t);
var
  damage: integer;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  if P_CheckMeleeRange(actor) then
  begin
    damage := ((P_Random mod 10) + 1) * 4;
    P_DamageMobj(actor.target, actor, actor, damage);
  end;
end;

procedure A_HeadAttack(actor: Pmobj_t);
var
  damage: integer;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  if P_CheckMeleeRange(actor) then
  begin
    damage := (P_Random mod 6 + 1) * 10;
    P_DamageMobj(actor.target, actor, actor, damage);
    exit;
  end;

  // launch a missile
  P_SpawnMissile(actor, actor.target, MT_HEADSHOT);
end;

procedure A_CyberAttack(actor: Pmobj_t);
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  P_SpawnMissile(actor, actor.target, MT_ROCKET);
end;

procedure A_BruisAttack(actor: Pmobj_t);
var
  damage: integer;
begin
  if actor.target = nil then
    exit;

  if P_CheckMeleeRange(actor) then
  begin
    S_StartSound(actor, Ord(sfx_claw));
    damage := (P_Random mod 8 + 1) * 10;
    P_DamageMobj(actor.target, actor, actor, damage);
    exit;
  end;

  // launch a missile
  P_SpawnMissile(actor, actor.target, MT_BRUISERSHOT);
end;

//
// A_SkelMissile
//
procedure A_SkelMissile(actor: Pmobj_t);
var
  mo: Pmobj_t;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  actor.z := actor.z + 16 * FRACUNIT; // so missile spawns higher
  mo := P_SpawnMissile(actor, actor.target, MT_TRACER);
  actor.z := actor.z - 16 * FRACUNIT; // back to normal

  mo.x := mo.x + mo.momx;
  mo.y := mo.y + mo.momy;
  mo.tracer := actor.target;
end;

const
  TRACEANGLE = $c000000;

procedure A_Tracer(actor: Pmobj_t);
var
  exact: angle_t;
  dist: fixed_t;
  slope: fixed_t;
  dest: Pmobj_t;
  th: Pmobj_t;
begin
  if boolval(gametic and 3) then
    exit;

  // spawn a puff of smoke behind the rocket
  P_SpawnPuff(actor.x, actor.y, actor.z);

  th := P_SpawnMobj(actor.x - actor.momx,
                    actor.y - actor.momy,
                    actor.z, MT_SMOKE);

  th.momz := FRACUNIT;
  th.tics := th.tics - P_Random and 3;
  if th.tics < 1 then
    th.tics := 1;

  // adjust direction
  dest := actor.tracer;

  if (not boolval(dest)) or (dest.health <= 0) then
    exit;

  // change angle
  exact := R_PointToAngle2(actor.x, actor.y, dest.x, dest.y);

  if exact <> actor.angle then
  begin
    if exact - actor.angle > ANG180 then
    begin
      actor.angle := actor.angle - TRACEANGLE;
      if exact - actor.angle < ANG180 then
        actor.angle := exact;
    end
    else
    begin
      actor.angle := actor.angle + TRACEANGLE;
      if exact - actor.angle > ANG180 then
        actor.angle := exact;
    end;
  end;

  exact := _SHRW(actor.angle, ANGLETOFINESHIFT);
  actor.momx := FixedMul(actor.info.speed, finecosine[exact]);
  actor.momy := FixedMul(actor.info.speed, finesine[exact]);

  // change slope
  dist := P_AproxDistance(dest.x - actor.x, dest.y - actor.y);

  dist := dist div actor.info.speed;

  if dist < 1 then
    dist := 1;
  slope := (dest.z + 40 * FRACUNIT - actor.z) div dist;

  if slope < actor.momz then
    actor.momz := actor.momz - FRACUNIT div 8
  else
    actor.momz := actor.momz + FRACUNIT div 8;
end;

procedure A_SkelWhoosh(actor: Pmobj_t);
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  S_StartSound(actor, Ord(sfx_skeswg));
end;

procedure A_SkelFist(actor: Pmobj_t);
var
  damage: integer;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);

  if P_CheckMeleeRange(actor) then
  begin
    damage := ((P_Random mod 10) + 1) * 6;
    S_StartSound(actor, Ord(sfx_skepch));
    P_DamageMobj(actor.target, actor, actor, damage);
  end;
end;

//
// PIT_VileCheck
// Detect a corpse that could be raised.
//
var
  corpsehit: Pmobj_t;
  vileobj: Pmobj_t;
  viletryx: fixed_t;
  viletryy: fixed_t;

function PIT_VileCheck(thing: Pmobj_t): boolean;
var
  maxdist: integer;
  check: boolean;
begin
  if not boolval(thing.flags and MF_CORPSE) then
  begin
    result := true; // not a monster
    exit;
  end;

  if thing.tics <> -1 then
  begin
    result := true; // not lying still yet
    exit;
  end;

  if thing.info.raisestate = Ord(S_NULL) then
  begin
    result := true; // monster doesn't have a raise state
    exit;
  end;

  maxdist := thing.info.radius + mobjinfo[Ord(MT_VILE)].radius;

  if (abs(thing.x - viletryx) > maxdist) or
     (abs(thing.y - viletryy) > maxdist) then
  begin
    result := true; // not actually touching
    exit;
  end;

  corpsehit := thing;
  corpsehit.momx := 0;
  corpsehit.momy := 0;
  corpsehit.height := _SHL(corpsehit.height, 2);
  check := P_CheckPosition(corpsehit, corpsehit.x, corpsehit.y);
  corpsehit.height := _SHR(corpsehit.height, 2);

  if not check then
    result := true    // doesn't fit here
  else
    result := false;  // got one, so stop checking
end;

//
// A_VileChase
// Check for ressurecting a body
//
procedure A_VileChase(actor: Pmobj_t);
var
  xl: integer;
  xh: integer;
  yl: integer;
  yh: integer;

  bx: integer;
  by: integer;

  info: Pmobjinfo_t;
  temp: Pmobj_t;
begin
  if actor.movedir <> Ord(DI_NODIR) then
  begin
    // check for corpses to raise
    viletryx := actor.x + actor.info.speed * xspeed[actor.movedir];
    viletryy := actor.y + actor.info.speed * yspeed[actor.movedir];

    xl := _SHR((viletryx - bmaporgx - MAXRADIUS * 2), MAPBLOCKSHIFT);
    xh := _SHR((viletryx - bmaporgx + MAXRADIUS * 2), MAPBLOCKSHIFT);
    yl := _SHR((viletryy - bmaporgy - MAXRADIUS * 2), MAPBLOCKSHIFT);
    yh := _SHR((viletryy - bmaporgy + MAXRADIUS * 2), MAPBLOCKSHIFT);

    vileobj := actor;
    for bx := xl to xh do
    begin
      for by := yl to yh do
      begin
      // Call PIT_VileCheck to check
      // whether object is a corpse
      // that canbe raised.
        if not P_BlockThingsIterator(bx, by, PIT_VileCheck) then
        begin
        // got one!
          temp := actor.target;
          actor.target := corpsehit;
          A_FaceTarget(actor);
          actor.target := temp;

          P_SetMobjState(actor, S_VILE_HEAL1);
          S_StartSound(corpsehit, Ord(sfx_slop));
          info := corpsehit.info;

          P_SetMobjState(corpsehit, statenum_t(info.raisestate));
          corpsehit.height := _SHL(corpsehit.height, 2);
          corpsehit.flags := info.flags;
          corpsehit.health := info.spawnhealth;
          corpsehit.target := nil;
          exit;
        end;
      end;
    end;
  end;

  // Return to normal attack.
  A_Chase(actor);
end;

//
// A_VileStart
//
procedure A_VileStart(actor: Pmobj_t);
begin
  S_StartSound(actor, Ord(sfx_vilatk));
end;

//
// A_Fire
// Keep fire in front of player unless out of sight
//
procedure A_Fire(actor: Pmobj_t);
var
  dest: Pmobj_t;
  an: LongWord;
begin
  dest := actor.tracer;
  if dest = nil then
    exit;

  // don't move it if the vile lost sight
  if not P_CheckSight(actor.target, dest) then
    exit;

  an := _SHRW(dest.angle, ANGLETOFINESHIFT);

  P_UnsetThingPosition(actor);
  actor.x := dest.x + FixedMul(24 * FRACUNIT, finecosine[an]);
  actor.y := dest.y + FixedMul(24 * FRACUNIT, finesine[an]);
  actor.z := dest.z;
  P_SetThingPosition (actor);
end;

procedure A_StartFire(actor: Pmobj_t);
begin
  S_StartSound(actor, Ord(sfx_flamst));
  A_Fire(actor);
end;

procedure A_FireCrackle(actor: Pmobj_t);
begin
  S_StartSound(actor, Ord(sfx_flame));
  A_Fire(actor);
end;

//
// A_VileTarget
// Spawn the hellfire
//
procedure A_VileTarget(actor: Pmobj_t);
var
  fog: Pmobj_t;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);

  fog := P_SpawnMobj(actor.target.x, actor.target.x, actor.target.z, MT_FIRE);

  actor.tracer := fog;
  fog.target := actor;
  fog.tracer := actor.target;
  A_Fire(fog);
end;

//
// A_VileAttack
//
procedure A_VileAttack(actor: Pmobj_t);
var
  fire: Pmobj_t;
  an: angle_t;
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);

  if not P_CheckSight(actor, actor.target) then
    exit;

  S_StartSound(actor, Ord(sfx_barexp));
  P_DamageMobj(actor.target, actor, actor, 20);
  actor.target.momz := 1000 * FRACUNIT div actor.target.info.mass;

  an := _SHRW(actor.angle, ANGLETOFINESHIFT);

  fire := actor.tracer;

  if fire = nil then
    exit;

  // move the fire between the vile and the player
  fire.x := actor.target.x - FixedMul(24 * FRACUNIT, finecosine[an]);
  fire.y := actor.target.y - FixedMul(24 * FRACUNIT, finesine[an]);
  P_RadiusAttack(fire, actor, 70 );
end;

//
// Mancubus attack,
// firing three missiles (bruisers)
// in three different directions?
// Doesn't look like it. 
//
const
  FATSPREAD = ANG90 div 8;

procedure A_FatRaise(actor: Pmobj_t);
begin
  A_FaceTarget(actor);
  S_StartSound(actor, Ord(sfx_manatk));
end;

procedure A_FatAttack1(actor: Pmobj_t);
var
  mo: Pmobj_t;
  an: angle_t;
begin
  A_FaceTarget(actor);
  // Change direction  to ...
  actor.angle := actor.angle + FATSPREAD;
  P_SpawnMissile(actor, actor.target, MT_FATSHOT);

  mo := P_SpawnMissile(actor, actor.target, MT_FATSHOT);
  mo.angle := mo.angle + FATSPREAD;
  an := _SHRW(mo.angle, ANGLETOFINESHIFT);
  mo.momx := FixedMul(mo.info.speed, finecosine[an]);
  mo.momy := FixedMul(mo.info.speed, finesine[an]);
end;

procedure A_FatAttack2(actor: Pmobj_t);
var
  mo: Pmobj_t;
  an: angle_t;
begin
  A_FaceTarget(actor);
  // Now here choose opposite deviation.
  actor.angle := actor.angle - FATSPREAD;
  P_SpawnMissile(actor, actor.target, MT_FATSHOT);

  mo := P_SpawnMissile(actor, actor.target, MT_FATSHOT);
  mo.angle := mo.angle - FATSPREAD * 2;
  an := _SHRW(mo.angle, ANGLETOFINESHIFT);
  mo.momx := FixedMul(mo.info.speed, finecosine[an]);
  mo.momy := FixedMul(mo.info.speed, finesine[an]);
end;

procedure A_FatAttack3(actor: Pmobj_t);
var
  mo: Pmobj_t;
  an: angle_t;
begin
  A_FaceTarget(actor);

  mo := P_SpawnMissile(actor, actor.target, MT_FATSHOT);
  mo.angle := mo.angle - FATSPREAD div 2;
  an := _SHRW(mo.angle, ANGLETOFINESHIFT);
  mo.momx := FixedMul(mo.info.speed, finecosine[an]);
  mo.momy := FixedMul(mo.info.speed, finesine[an]);

  mo := P_SpawnMissile(actor, actor.target, MT_FATSHOT);
  mo.angle := mo.angle + FATSPREAD div 2;
  an := _SHRW(mo.angle, ANGLETOFINESHIFT);
  mo.momx := FixedMul(mo.info.speed, finecosine[an]);
  mo.momy := FixedMul(mo.info.speed, finesine[an]);
end;

//
// SkullAttack
// Fly at the player like a missile.
//
const
  SKULLSPEED = 20 * FRACUNIT;

procedure A_SkullAttack(actor: Pmobj_t);
var
  dest: Pmobj_t;
  an: angle_t;
  dist: integer;
begin
  if actor.target = nil then
    exit;

  dest := actor.target;
  actor.flags := actor.flags or MF_SKULLFLY;

  S_StartSound(actor, actor.info.attacksound);
  A_FaceTarget(actor);
  an := _SHRW(actor.angle, ANGLETOFINESHIFT);
  actor.momx := FixedMul(SKULLSPEED, finecosine[an]);
  actor.momy := FixedMul(SKULLSPEED, finesine[an]);
  dist := P_AproxDistance(dest.x - actor.x, dest.y - actor.y);
  dist := dist div SKULLSPEED;

  if dist < 1 then
    dist := 1;
  actor.momz := (dest.z + (_SHR(dest.height, 1)) - actor.z) div dist;
end;

//
// A_PainShootSkull
// Spawn a lost soul and launch it at the target
//
procedure A_PainShootSkull(actor: Pmobj_t; angle: angle_t);
var
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;

  newmobj: Pmobj_t;
  an: angle_t;
  prestep: integer;
  count: integer;
  currentthinker: Pthinker_t;
begin
  // count total number of skull currently on the level
  count := 0;

  currentthinker := thinkercap.next;
  while Pointer(currentthinker) <> Pointer(@thinkercap) do
  begin
    if (@currentthinker._function.acp1 = @P_MobjThinker) and
       (Pmobj_t(currentthinker)._type = MT_SKULL) then
      inc(count);
    currentthinker := currentthinker.next;
  end;
{  currentthinker := @thinkercap;
  repeat
    currentthinker := currentthinker.next;
    if (@currentthinker._function.acp1 = @P_MobjThinker) and
       (Pmobj_t(currentthinker)._type = MT_SKULL) then
      inc(count);
  until Pointer(currentthinker) = Pointer(@thinkercap);}


  // if there are allready 20 skulls on the level,
  // don't spit another one
  if count > 20 then
    exit;


  // okay, there's playe for another one
  an := _SHRW(angle, ANGLETOFINESHIFT);

  prestep := 4 * FRACUNIT +
             3 * (actor.info.radius + mobjinfo[Ord(MT_SKULL)].radius) div 2;

  x := actor.x + FixedMul(prestep, finecosine[an]);
  y := actor.y + FixedMul(prestep, finesine[an]);
  z := actor.z + 8 * FRACUNIT;

  newmobj := P_SpawnMobj(x, y, z, MT_SKULL);

  // Check for movements.
  if not P_TryMove(newmobj, newmobj.x, newmobj.y) then
  begin
    // kill it immediately
    P_DamageMobj(newmobj, actor, actor, 10000);
    exit;
  end;

  newmobj.target := actor.target;
  A_SkullAttack(newmobj);
end;

//
// A_PainAttack
// Spawn a lost soul and launch it at the target
// 
procedure A_PainAttack(actor: Pmobj_t);
begin
  if actor.target = nil then
    exit;

  A_FaceTarget(actor);
  A_PainShootSkull(actor, actor.angle);
end;

procedure A_PainDie(actor: Pmobj_t);
begin
  A_Fall(actor);
  A_PainShootSkull(actor, actor.angle + ANG90);
  A_PainShootSkull(actor, actor.angle + ANG180);
  A_PainShootSkull(actor, actor.angle + ANG270);
end;

procedure A_Scream(actor: Pmobj_t);
var
  sound: integer;
begin
  case actor.info.deathsound of
    0: exit;
    Ord(sfx_podth1),
    Ord(sfx_podth2),
    Ord(sfx_podth3):
      sound := Ord(sfx_podth1) + P_Random mod 3;

    Ord(sfx_bgdth1),
    Ord(sfx_bgdth2):
      sound := Ord(sfx_bgdth1) + P_Random mod 2;
  else
    sound := actor.info.deathsound;
  end;

  // Check for bosses.
  if (actor._type = MT_SPIDER) or (actor._type = MT_CYBORG) then
    // full volume
    S_StartSound(nil, sound)
  else
    S_StartSound(actor, sound);
end;

procedure A_XScream(actor: Pmobj_t);
begin
  S_StartSound(actor, Ord(sfx_slop));
end;

procedure A_Pain(actor: Pmobj_t);
begin
  if boolval(actor.info.painsound) then
    S_StartSound(actor, actor.info.painsound);
end;

//
// A_Explode
//
procedure A_Explode(thingy: Pmobj_t);
begin
  P_RadiusAttack(thingy, thingy.target, 128);
end;

//
// A_BossDeath
// Possibly trigger special effects
// if on first boss level
//
procedure A_BossDeath(mo: Pmobj_t);
var
  th: Pthinker_t;
  mo2: Pmobj_t;
  junk: line_t;
  i: integer;
begin
  if gamemode = commercial then
  begin
    if gamemap <> 7 then
      exit;

    if (mo._type <> MT_FATSO) and (mo._type <> MT_BABY) then
      exit;
  end
  else
  begin
    case gameepisode of
      1:
        begin
          if gamemap <> 8 then
            exit;
          if mo._type <> MT_BRUISER then
            exit;
        end;
      2:
        begin
          if gamemap <> 8 then
            exit;
          if mo._type <> MT_CYBORG then
            exit;
        end;
      3:
        begin
          if gamemap <> 8 then
            exit;
          if mo._type <> MT_SPIDER then
            exit;
        end;
      4:
        begin
          case gamemap of
            6: if mo._type <> MT_CYBORG then
                 exit;
            8: if mo._type <> MT_SPIDER then
                 exit;
          end;
        end;
    else
      begin
        if gamemap <> 8 then
          exit;
      end;
    end;
  end;

  // make sure there is a player alive for victory
  i := 0;
  while i < MAXPLAYERS do
  begin
    if playeringame[i] and (players[i].health > 0) then
      break;
    inc(i);
  end;

  if i = MAXPLAYERS then
    exit;  // no one left alive, so do not end game

  // scan the remaining thinkers to see
  // if all bosses are dead
{  th := thinkercap.next;
  while Pointer(th) <> Pointer(@thinkercap) do
  begin
    if th._function.acp1 <> actionf_p1(P_MobjThinker) then
    begin
      th := th.next;
      continue;
    end;

    mo2 := Pmobj_t(th);
    if (mo2 <> mo) and (mo2._type = mo._type) and (mo2.health > 0) then
    begin
      // other boss not dead
      exit;
    end;
    th := th.next;
  end;}
  th := thinkercap.next;
  while Pointer(th) <> Pointer(@thinkercap) do
  begin
    if @th._function.acp1 = @P_MobjThinker then
    begin
      mo2 := Pmobj_t(th);
      if (mo2 <> mo) and (mo2._type = mo._type) and (mo2.health > 0) then
      begin
        // other boss not dead
        exit;
      end;
    end;
    th := th.next;
  end;

{  repeat
    th := th.next;
    if @th._function.acp1 = @P_MobjThinker then
    begin
      mo2 := Pmobj_t(th);
      if (mo2 <> mo) and (mo2._type = mo._type) and (mo2.health > 0) then
      begin
        // other boss not dead
        exit;
      end;
    end;
  until Pointer(th) = Pointer(@thinkercap);}


  // victory!
  if gamemode = commercial then
  begin
    if gamemap = 7 then
    begin
      if mo._type = MT_FATSO then
      begin
        junk.tag := 666;
        EV_DoFloor(@junk, lowerFloorToLowest);
        exit;
      end;
      if mo._type = MT_BABY then
      begin
        junk.tag := 667;
        EV_DoFloor(@junk, raiseToTexture);
        exit;
      end;
    end;
  end
  else
  begin
    case gameepisode of
      1:
        begin
          junk.tag := 666;
          EV_DoFloor(@junk, lowerFloorToLowest);
          exit;
        end;
      4:
        begin
          if gamemap = 6 then
          begin
            junk.tag := 666;
            EV_DoDoor(@junk, blazeOpen);
            exit;
          end
          else if gamemap = 6 then
          begin
            junk.tag := 666;
            EV_DoFloor(@junk, lowerFloorToLowest);
            exit;
          end;
        end;
    end;
  end;
  G_ExitLevel;
end;

procedure A_Hoof(mo: Pmobj_t);
begin
  S_StartSound(mo, Ord(sfx_hoof));
  A_Chase(mo);
end;

procedure A_Metal(mo: Pmobj_t);
begin
  S_StartSound(mo, Ord(sfx_metal));
  A_Chase(mo);
end;

procedure A_BabyMetal(mo: Pmobj_t);
begin
  S_StartSound(mo, Ord(sfx_bspwlk));
  A_Chase(mo);
end;

procedure A_OpenShotgun2(player: Pplayer_t; psp: Pplayer_t);
begin
  S_StartSound(player.mo, Ord(sfx_dbopn));
end;

procedure A_LoadShotgun2(player: Pplayer_t; psp: Ppspdef_t);
begin
  S_StartSound(player.mo, Ord(sfx_dbload));
end;

procedure A_CloseShotgun2(player: Pplayer_t; psp: Ppspdef_t);
begin
  S_StartSound(player.mo, Ord(sfx_dbcls));
  A_ReFire(player, psp);
end;

const
  MAXBRAINTARGETS = 32;

var
  braintargets: array[0..MAXBRAINTARGETS - 1] of Pmobj_t;
  numbraintargets: integer;
  braintargeton: integer;

procedure A_BrainAwake(mo: Pmobj_t);
var
  thinker: Pthinker_t;
  m: Pmobj_t;
begin
  // find all the target spots
  numbraintargets := 0;
  braintargeton := 0;

  thinker := thinkercap.next;
  while Pointer(thinker) <> Pointer(@thinkercap) do
  begin
    if @thinker._function.acp1 = @P_MobjThinker then // is a mobj
    begin
      m := Pmobj_t(thinker);
      if m._type = MT_BOSSTARGET then
      begin
        if numbraintargets >= MAXBRAINTARGETS then
          I_Error('A_BrainAwake(): numbraintargets = %d >= MAXBRAINTARGETS', [numbraintargets]);
        braintargets[numbraintargets] := m;
        inc(numbraintargets);
      end;
    end;
    thinker := thinker.next;
  end;

  S_StartSound(nil, Ord(sfx_bossit));
end;

procedure A_BrainPain(mo: Pmobj_t);
begin
  S_StartSound(nil, Ord(sfx_bospn));
end;

procedure A_BrainScream(mo: Pmobj_t);
var
  x: integer;
  y: integer;
  z: integer;
  th: Pmobj_t;
begin
  x := mo.x - 196 * FRACUNIT;
  while x < mo.x + 320 * FRACUNIT do
  begin
    y := mo.y - 320 * FRACUNIT;
    z := 128 + P_Random * 2 * FRACUNIT;
    th := P_SpawnMobj(x, y, z, MT_ROCKET);
    th.momz := P_Random * 512;

    P_SetMobjState(th, S_BRAINEXPLODE1);

    th.tics := th.tics - (P_Random and 7);
    if th.tics < 1 then
      th.tics := 1;
    x := x + FRACUNIT * 8;
  end;

  S_StartSound(nil, Ord(sfx_bosdth));
end;

procedure A_BrainExplode(mo: Pmobj_t);
var
  x: integer;
  y: integer;
  z: integer;
  th: Pmobj_t;
begin
  x := mo.x + (P_Random - P_Random) * 2048;
  y := mo.y;
  z := 128 + P_Random * 2 * FRACUNIT;
  th := P_SpawnMobj(x, y, z, MT_ROCKET);
  th.momz := P_Random * 512;

  P_SetMobjState(th, S_BRAINEXPLODE1);

  th.tics := th.tics - (P_Random and 7);
  if th.tics < 1 then
    th.tics := 1;
end;

procedure A_BrainDie(mo: Pmobj_t);
begin
  G_ExitLevel;
end;

var
  easy: integer = 0;

procedure A_BrainSpit(mo: Pmobj_t);
var
  targ: Pmobj_t;
  newmobj: Pmobj_t;
begin
  easy := easy xor 1;
  if (gameskill <= sk_easy) and (not boolval(easy)) then
    exit;

  if numbraintargets = 0 then // VJ
  begin
    A_BrainAwake(mo);
    exit;
  end;

  // shoot a cube at current target
  targ := braintargets[braintargeton];
  braintargeton := (braintargeton + 1) mod numbraintargets;

  // spawn brain missile
  newmobj := P_SpawnMissile(mo, targ, MT_SPAWNSHOT);
  newmobj.target := targ;
  newmobj.reactiontime := ((targ.y - mo.y) div newmobj.momy) div newmobj.state.tics;

  S_StartSound(nil, Ord(sfx_bospit));
end;

procedure A_SpawnFly(mo: Pmobj_t);
var
  newmobj: Pmobj_t;
  fog: Pmobj_t;
  targ: Pmobj_t;
  r: integer;
  _type: mobjtype_t;
begin
  mo.reactiontime := mo.reactiontime - 1;
  if mo.reactiontime > 0 then
    exit; // still flying

  targ := mo.target;

  if targ = nil then  // VJ
  begin
    P_RemoveMobj(mo);
    exit;
  end;

  // First spawn teleport fog.
  fog := P_SpawnMobj(targ.x, targ.y, targ.z, MT_SPAWNFIRE);
  S_StartSound(fog, Ord(sfx_telept));

  // Randomly select monster to spawn.
  r := P_Random;

  // Probability distribution (kind of :),
  // decreasing likelihood.
  if r < 50 then
    _type := MT_TROOP
  else if r < 90 then
    _type := MT_SERGEANT
  else if r < 120 then
    _type := MT_SHADOWS
  else if r < 130 then
    _type := MT_PAIN
  else if r < 160 then
    _type := MT_HEAD
  else if r < 162 then
    _type := MT_VILE
  else if r < 172 then
    _type := MT_UNDEAD
  else if r < 192 then
    _type := MT_BABY
  else if r < 222 then
    _type := MT_FATSO
  else if r < 246 then
    _type := MT_KNIGHT
  else
    _type := MT_BRUISER;

  newmobj := P_SpawnMobj(targ.x, targ.y, targ.z, _type);
  if P_LookForPlayers(newmobj, true) then
    P_SetMobjState(newmobj, statenum_t(newmobj.info.seestate));

  // telefrag anything in this spot
  P_TeleportMove(newmobj, newmobj.x, newmobj.y);

  // remove self (i.e., cube).
  P_RemoveMobj(mo);
end;

// travelling cube sound
procedure A_SpawnSound(mo: Pmobj_t);
begin
  S_StartSound(mo, Ord(sfx_boscub));
  A_SpawnFly(mo);
end;

procedure A_PlayerScream(mo: Pmobj_t);
var
  sound: integer;
begin
    // Default death sound.
  sound := Ord(sfx_pldeth);

  if (gamemode = commercial) and (mo.health < -50) then
  begin
    // IF THE PLAYER DIES
    // LESS THAN -50% WITHOUT GIBBING
    sound := Ord(sfx_pdiehi);
  end;

  S_StartSound(mo, sound);
end;



(* Unit am_map implementation *)


                                                           

{ how much zoom-in per tic }
function M_ZOOMIN : integer;
begin
  result := trunc(1.02 * FRACUNIT);
end;

{ how much zoom-out per tic }
function M_ZOOMOUT : integer;
begin
  result := trunc(FRACUNIT / 1.02);
end;

function FTOM(x : integer): integer;
begin
  result := FixedMul(_SHL(x, 16), scale_ftom);
end;

function MTOF(x : integer): integer;
begin
  result := _SHR(FixedMul(x, scale_mtof), 16);
end;

function CXMTOF(x : integer): integer;
begin
  result := f_x + MTOF(x - m_x);
end;

function CYMTOF(y : integer): integer;
begin
  result := f_y + (f_h - MTOF(y - m_y))
end;

//
//
//
procedure AM_getIslope(ml: Pmline_t; _is: Pislope_t);
var dx, dy: integer;
begin
  dx := ml.b.x - ml.a.x;
  dy := ml.a.y - ml.b.y;

  if dy = 0 then
  begin
    if dx < 0 then
      _is.islp := -MAXINT
    else
      _is.islp := MAXINT
  end
  else
    _is.islp := FixedDiv(dx, dy);

  if dx = 0 then
  begin
    if dy < 0 then
      _is.slp := -MAXINT
    else
      _is.slp := MAXINT
  end
  else
    _is.slp := FixedDiv(dy, dx);
end;

//
//
//
procedure AM_activateNewScale;
begin
  m_x := m_x + m_w div 2;
  m_y := m_y + m_h div 2;
  m_w := FTOM(f_w);
  m_h := FTOM(f_h);
  m_x := m_x - m_w div 2;
  m_y := m_y - m_h div 2;
  m_x2 := m_x + m_w;
  m_y2 := m_y + m_h;
end;

//
//
//
procedure AM_saveScaleAndLoc;
begin
  old_m_x := m_x;
  old_m_y := m_y;
  old_m_w := m_w;
  old_m_h := m_h;
end;

//
//
//
procedure AM_restoreScaleAndLoc;
begin
  m_w := old_m_w;
  m_h := old_m_h;
  if not followplayer then
  begin
    m_x := old_m_x;
    m_y := old_m_y;
  end
  else
  begin
    m_x := plr.mo.x - m_w div 2;
    m_y := plr.mo.y - m_h div 2;
  end;

  m_x2 := m_x + m_w;
  m_y2 := m_y + m_h;

  // Change the scaling multipliers
  scale_mtof := FixedDiv(_SHL(f_w, FRACBITS), m_w);
  scale_ftom := FixedDiv(FRACUNIT, scale_mtof);
end;

//
// adds a marker at the current location
//
procedure AM_addMark;
begin
  markpoints[markpointnum].x := m_x + m_w div 2;
  markpoints[markpointnum].y := m_y + m_h div 2;
  markpointnum := (markpointnum + 1) mod AM_NUMMARKPOINTS;
end;

//
// Determines bounding box of all vertices,
// sets global variables controlling zoom range.
//
procedure AM_findMinMaxBoundaries;
var i: integer;
    a, b: fixed_t;
begin
  min_x := MAXINT;
  min_y := MAXINT;
  max_x := -MAXINT;
  max_y := -MAXINT;

  for i := 0 to numvertexes - 1 do
  begin
    if vertexes[i].x < min_x then
      min_x := vertexes[i].x
    else if vertexes[i].x > max_x then
      max_x := vertexes[i].x;

    if vertexes[i].y < min_y then
      min_y := vertexes[i].y
    else if vertexes[i].y > max_y then
      max_y := vertexes[i].y;

  end;

  max_w := max_x - min_x;
  max_h := max_y - min_y;

  min_w := 2 * PLAYERRADIUS; // const? never changed?
  min_h := 2 * PLAYERRADIUS;

  a := FixedDiv(f_w * FRACUNIT, max_w);
  b := FixedDiv(f_h * FRACUNIT, max_h);

  if a < b then
    min_scale_mtof := a
  else
    min_scale_mtof := b;

  max_scale_mtof := FixedDiv(f_h * FRACUNIT, 2 * PLAYERRADIUS);
end;

//
//
//
procedure AM_changeWindowLoc;
begin
  if (m_paninc.x <> 0) or (m_paninc.y <> 0) then
  begin
    followplayer := false;
    f_oldloc.x := MAXINT;
  end;

  m_x := m_x + m_paninc.x;
  m_y := m_y + m_paninc.y;

  if m_x + m_w div 2 > max_x then
    m_x := max_x - m_w div 2
  else if m_x + m_w div 2 < min_x then
    m_x := min_x - m_w div 2;

  if m_y + m_h div 2 > max_y then
    m_y := max_y - m_h div 2
  else if m_y + m_h div 2 < min_y then
    m_y := min_y - m_h div 2;

  m_x2 := m_x + m_w;
  m_y2 := m_y + m_h;
end;

//
//
//
var
  st_notify_AM_initVariables: event_t;

procedure AM_initVariables;
var
  pnum: integer;
  i: integer;
begin
  automapactive := true;
  fb := screens[_FG];

  f_oldloc.x := MAXINT;
  amclock := 0;
  lightlev := 0;

  m_paninc.x := 0;
  m_paninc.y := 0;

  ftom_zoommul := FRACUNIT;
  mtof_zoommul := FRACUNIT;

  m_w := FTOM(f_w);
  m_h := FTOM(f_h);

    // find player to center on initially
  pnum := consoleplayer;
  if not playeringame[pnum] then
    for i := 0 to MAXPLAYERS - 1 do
      if playeringame[i] then
      begin
        pnum := i;
        break;
      end;

  plr := @players[pnum];
  m_x := plr.mo.x - m_w div 2;
  m_y := plr.mo.y - m_h div 2;
  AM_changeWindowLoc;

  // for saving & restoring
  //AM_saveScaleAndLoc;
  old_m_x := m_x;
  old_m_y := m_y;
  old_m_w := m_w;
  old_m_h := m_h;

  // inform the status bar of the change
  ST_Responder(@st_notify_AM_initVariables);
end;

//
//
//
procedure AM_loadPics;
var i: integer;
    namebuf: string;
begin
  for i := 0 to AM_NUMMARKPOINTS - 1 do
  begin
    sprintf(namebuf, 'AMMNUM%d', [i]);
    marknums[i] := W_CacheLumpName(namebuf, PU_STATIC);
  end;
end;

procedure AM_unloadPics;
var i: integer;
begin
  for i := 0 to AM_NUMMARKPOINTS - 1 do
    Z_ChangeTag(marknums[i], PU_CACHE);
end;

procedure AM_clearMarks;
var i: integer;
begin
  for i := 0 to AM_NUMMARKPOINTS - 1 do
    markpoints[i].x := -1; // means empty
  markpointnum := 0;
end;

//
// should be called at the start of every level
// right now, i figure it out myself
//
procedure AM_LevelInit;
begin
  leveljuststarted := 0;

  f_x := 0;
  f_y := 0;
  f_w := SCREENWIDTH;
  f_h := (200 - 32) * SCREENHEIGHT div 200;

  AM_clearMarks;

  AM_findMinMaxBoundaries;
  scale_mtof := FixedDiv(min_scale_mtof, trunc(0.7 * FRACUNIT));
  if scale_mtof > max_scale_mtof then
    scale_mtof := min_scale_mtof;
  scale_ftom := FixedDiv(FRACUNIT, scale_mtof);
end;

//
//
//
var
  st_notify_AM_Stop: event_t;

procedure AM_Stop;
begin
  AM_unloadPics;
  automapactive := false;
  ST_Responder(@st_notify_AM_Stop);
  stopped := true;
end;

//
//
//
var
  lastlevel: integer = -1;
  lastepisode: integer = -1;

procedure AM_Start;
begin
  if not stopped then
    AM_Stop;
  stopped := false;

  if (lastlevel <> gamemap) or (lastepisode <> gameepisode) then
  begin
    AM_LevelInit;
    lastlevel := gamemap;
    lastepisode := gameepisode;
  end;
  AM_initVariables;
  AM_loadPics;
end;

//
// set the window scale to the maximum size
//
procedure AM_minOutWindowScale;
begin
  scale_mtof := min_scale_mtof;
  scale_ftom := FixedDiv(FRACUNIT, scale_mtof);
  AM_activateNewScale;
end;

//
// set the window scale to the minimum size
//
procedure AM_maxOutWindowScale;
begin
  scale_mtof := max_scale_mtof;
  scale_ftom := FixedDiv(FRACUNIT, scale_mtof);
  AM_activateNewScale;
end;

//
// Handle events (user inputs) in automap mode
//
var
  cheatstate: integer = 0;
  bigstate: boolean = false;

function AM_Responder(ev: Pevent_t): boolean;
var
  _message: string;
begin
  result := false;

  if not automapactive then
  begin
    if (ev._type = ev_keydown) and (ev.data1 = AM_STARTKEY) then
    begin
      AM_Start;
      viewactive := false;
      result := true;
    end;
  end
  else if ev._type = ev_keydown then
  begin
    result := true;
    case ev.data1 of
      AM_PANRIGHTKEY: // pan right
        begin
          if not followplayer then
            m_paninc.x := FTOM(F_PANINC)
          else
            result := false;
        end;
      AM_PANLEFTKEY: // pan left
        begin
          if not followplayer then
            m_paninc.x := -FTOM(F_PANINC)
          else
            result := false;
        end;
      AM_PANUPKEY: // pan up
        begin
          if not followplayer then
            m_paninc.y := FTOM(F_PANINC)
          else
            result := false;
        end;
      AM_PANDOWNKEY: // pan down
        begin
          if not followplayer then
            m_paninc.y := -FTOM(F_PANINC)
          else
            result := false;
        end;
      Ord(AM_ZOOMOUTKEY): // zoom out
        begin
          mtof_zoommul := M_ZOOMOUT;
          ftom_zoommul := M_ZOOMIN;
        end;
      Ord(AM_ZOOMINKEY): // zoom in
        begin
          mtof_zoommul := M_ZOOMIN;
          ftom_zoommul := M_ZOOMOUT;
        end;
      AM_ENDKEY:
        begin
          bigstate := false;
          viewactive := true;
          AM_Stop;
        end;
      Ord(AM_GOBIGKEY):
        begin
          bigstate := not bigstate;
          if bigstate then
          begin
            AM_saveScaleAndLoc;
            AM_minOutWindowScale;
          end
          else
            AM_restoreScaleAndLoc;
        end;
      Ord(AM_FOLLOWKEY):
        begin
          followplayer := not followplayer;
          f_oldloc.x := MAXINT;
          if followplayer then
            plr._message := AMSTR_FOLLOWON
          else
            plr._message := AMSTR_FOLLOWOFF;
        end;
      Ord(AM_GRIDKEY):
        begin
          grid := not grid;
          if grid then
            plr._message := AMSTR_GRIDON
          else
            plr._message := AMSTR_GRIDOFF;
        end;
      Ord(AM_MARKKEY):
        begin
          sprintf(_message, '%s %d', [AMSTR_MARKEDSPOT, markpointnum]);
          plr._message := _message;
          AM_addMark;
        end;
      Ord(AM_CLEARMARKKEY):
        begin
          AM_clearMarks;
          plr._message := AMSTR_MARKSCLEARED;
        end
      else
      begin
        cheatstate := 0;
        result := false;
      end
    end;
    if (deathmatch = 0) and cht_CheckCheat(@cheat_amap, Chr(ev.data1)) then
    begin
      result := false;
      cheating := (cheating + 1) mod 3;
    end;
  end
  else if ev._type = ev_keyup then
  begin
    result := false;
    case ev.data1 of
      AM_PANRIGHTKEY:
        begin
          if not followplayer then
            m_paninc.x := 0;
        end;
      AM_PANLEFTKEY:
        begin
          if not followplayer then
            m_paninc.x := 0;
        end;
      AM_PANUPKEY:
        begin
          if not followplayer then
            m_paninc.y := 0;
        end;
      AM_PANDOWNKEY:
        begin
          if not followplayer then
            m_paninc.y := 0;
        end;
      Ord(AM_ZOOMOUTKEY),
      Ord(AM_ZOOMINKEY):
        begin
          mtof_zoommul := FRACUNIT;
          ftom_zoommul := FRACUNIT;
        end;
    end;
  end;

end;

//
// Zooming
//
procedure AM_changeWindowScale;
begin
  // Change the scaling multipliers
  scale_mtof := FixedMul(scale_mtof, mtof_zoommul);
  scale_ftom := FixedDiv(FRACUNIT, scale_mtof);

  if scale_mtof < min_scale_mtof then
    AM_minOutWindowScale
  else if scale_mtof > max_scale_mtof then
    AM_maxOutWindowScale
  else
    AM_activateNewScale;
end;

//
//
//
procedure AM_doFollowPlayer;
begin
  if (f_oldloc.x <> plr.mo.x) or (f_oldloc.y <> plr.mo.y) then
  begin
    m_x := FTOM(MTOF(plr.mo.x)) - m_w div 2;
    m_y := FTOM(MTOF(plr.mo.y)) - m_h div 2;
    m_x2 := m_x + m_w;
    m_y2 := m_y + m_h;
    f_oldloc.x := plr.mo.x;
    f_oldloc.y := plr.mo.y;
  end;
end;

//
//
//
const
  NUMLITELEVELS = 8;

var nexttic: integer = 0;
    litelevels: array[0..NUMLITELEVELS - 1] of integer = (0, 4, 7, 10, 12, 14, 15, 15);
    litelevelscnt: integer = 0;

procedure AM_updateLightLev;
begin
  // Change light level
  if amclock > nexttic then
  begin
    lightlev := litelevels[litelevelscnt];
    litelevelscnt := (litelevelscnt + 1) mod NUMLITELEVELS;
    nexttic := amclock + 6 - (amclock mod 6);
  end;
end;

procedure AM_Ticker;
begin
  if not automapactive then
    exit;

  inc(amclock);

  if followplayer then
    AM_doFollowPlayer;

  // Change the zoom if necessary
  if ftom_zoommul <> FRACUNIT then
    AM_changeWindowScale;

  // Change x,y location
  if (m_paninc.x <> 0) or (m_paninc.y <> 0) then
    AM_changeWindowLoc;

    // Update light level
    // AM_updateLightLev();

end;

//
// Clear automap frame buffer.
//
procedure AM_clearFB(color: integer);
begin
  FillChar(fb^, f_w * f_h, color);
end;

//
// Automap clipping of lines.
//
// Based on Cohen-Sutherland clipping algorithm but with a slightly
// faster reject and precalculated slopes.  If the speed is needed,
// use a hash algorithm to handle  the common cases.
//
function AM_clipMline(ml:  Pmline_t; fl: Pfline_t): boolean;
const
  LEFT = 1;
  RIGHT  = 2;
  BOTTOM = 4;
  TOP = 8;
var
  outcode1, outcode2 ,outside : integer;
  tmp: fpoint_t;
  dx, dy: integer;

  procedure DOOUTCODE(var oc: integer; mx, my: integer);
  begin
    oc := 0;
    if my < 0 then
      oc := oc or TOP
    else if my >= f_h then
      oc := oc or BOTTOM;
    if mx < 0 then
      oc := oc or LEFT
    else if mx >= f_w then
      oc := oc or RIGHT;
  end;

begin
  // do trivial rejects and outcodes
  if ml.a.y > m_y2 then
    outcode1 := TOP
  else if ml.a.y < m_y then
    outcode1 := BOTTOM
  else
    outcode1 := 0;

  if ml.b.y > m_y2 then
    outcode2 := TOP
  else if ml.b.y < m_y then
    outcode2 := BOTTOM
  else
    outcode2 := 0;

  if (outcode1 and outcode2) <> 0 then
  begin
    result := false; // trivially outside
    exit;
  end;

  if ml.a.x < m_x then
    outcode1 := outcode1 or LEFT
  else if ml.a.x > m_x2 then
    outcode1 := outcode1 or RIGHT;

  if ml.b.x < m_x then
    outcode2 := outcode2 or LEFT
  else if ml.b.x > m_x2 then
    outcode2 := outcode2 or RIGHT;

  if (outcode1 and outcode2) <> 0 then
  begin
    result := false; // trivially outside
    exit;
  end;

  // transform to frame-buffer coordinates.
  fl.a.x := CXMTOF(ml.a.x);
  fl.a.y := CYMTOF(ml.a.y);
  fl.b.x := CXMTOF(ml.b.x);
  fl.b.y := CYMTOF(ml.b.y);

  DOOUTCODE(outcode1, fl.a.x, fl.a.y);
  DOOUTCODE(outcode2, fl.b.x, fl.b.y);

  if (outcode1 and outcode2) <> 0 then
  begin
    result := false; // trivially outside
    exit;
  end;

  while (outcode1 or outcode2) <> 0 do
  begin
  // may be partially inside box
  // find an outside point
    if outcode1 <> 0 then
      outside := outcode1
    else
      outside := outcode2;

  // clip to each side
    if (outside and TOP) <> 0 then
    begin
      dy := fl.a.y - fl.b.y;
      dx := fl.b.x - fl.a.x;
      tmp.x := fl.a.x + (dx * (fl.a.y)) div dy;
      tmp.y := 0;
    end
    else if (outside and BOTTOM) <> 0 then
    begin
      dy := fl.a.y - fl.b.y;
      dx := fl.b.x - fl.a.x;
      tmp.x := fl.a.x + (dx * (fl.a.y - f_h)) div dy;
      tmp.y := f_h - 1;
    end
    else if (outside and RIGHT) <> 0 then
    begin
      dy := fl.b.y - fl.a.y;
      dx := fl.b.x - fl.a.x;
      tmp.y := fl.a.y + (dy * (f_w - 1 - fl.a.x)) div dx;
      tmp.x := f_w - 1;
    end
    else if (outside and LEFT) <> 0 then
    begin
      dy := fl.b.y - fl.a.y;
      dx := fl.b.x - fl.a.x;
      tmp.y := fl.a.y + (dy * (-fl.a.x)) div dx;
      tmp.x := 0;
    end;

    if outside = outcode1 then
    begin
      fl.a := tmp;
      DOOUTCODE(outcode1, fl.a.x, fl.a.y);
    end
    else
    begin
      fl.b := tmp;
      DOOUTCODE(outcode2, fl.b.x, fl.b.y);
    end;

    if (outcode1 and outcode2) <> 0 then
    begin
      result := false; // trivially outside
      exit;
    end;
  end;

  result := true;
end;

//
// Classic Bresenham w/ whatever optimizations needed for speed
//
procedure AM_drawFline(fl: Pfline_t; color: integer);
var
  x, y,
  dx, dy,
  sx, sy,
  ax, ay,
  d: integer;

  procedure PUTDOT(xx,yy,cc: integer);
  begin
    fb[yy * f_w + xx] := cc;
  end;

begin
  // For debugging only
  if (fl.a.x < 0) or (fl.a.x >= f_w) or
     (fl.a.y < 0) or (fl.a.y >= f_h) or
     (fl.b.x < 0) or (fl.b.x >= f_w) or
     (fl.b.y < 0) or (fl.b.y >= f_h) then
  begin
    I_Error('AM_drawFline(): fuck!');
    exit;
  end;

  dx := fl.b.x - fl.a.x;
  ax := 2 * abs(dx);
  if dx < 0 then
    sx := -1
  else
    sx := 1;

  dy := fl.b.y - fl.a.y;
  ay := 2 * abs(dy);
  if dy < 0 then
    sy := -1
  else
    sy := 1;

  x := fl.a.x;
  y := fl.a.y;

  if ax > ay then
  begin
    d := ay - ax div 2;
    while true do
    begin
      PUTDOT(x, y, color);
      if x = fl.b.x then exit;
      if d >= 0 then
      begin
        y := y + sy;
        d := d - ax;
      end;
      x := x + sx;
      d := d + ay;
    end;
  end
  else
  begin
    d := ax - ay div 2;
    while true do
    begin
      PUTDOT(x, y, color);
      if (y = fl.b.y) then exit;
      if d >= 0 then
      begin
        x := x + sx;
        d := d - ay;
      end;
      y := y + sy;
      d := d + ax;
    end;
  end;
end;

//
// Clip lines, draw visible part sof lines.
//
var
  fl: fline_t;

procedure AM_drawMline(ml: Pmline_t; color: integer);
begin
  if AM_clipMline(ml, @fl) then
    AM_drawFline(@fl, color); // draws it on frame buffer using fb coords
end;

//
// Draws flat (floor/ceiling tile) aligned grid lines.
//
procedure AM_drawGrid(color: integer);
var
  x, y: fixed_t;
  start, _end: fixed_t;
  ml: mline_t;
begin
  // Figure out start of vertical gridlines
  start := m_x;
  if ((start - bmaporgx) mod (MAPBLOCKUNITS * FRACUNIT)) <> 0 then
    start := start + (MAPBLOCKUNITS * FRACUNIT)
          - ((start - bmaporgx) mod (MAPBLOCKUNITS * FRACUNIT));
  _end := m_x + m_w;

  // draw vertical gridlines
  ml.a.y := m_y;
  ml.b.y := m_y + m_h;
  x := start;
  while x < _end do
  begin
    ml.a.x := x;
    ml.b.x := x;
    AM_drawMline(@ml, color);
    x := x + (MAPBLOCKUNITS * FRACUNIT);
  end;

  // Figure out start of horizontal gridlines
  start := m_y;
  if ((start - bmaporgy) mod (MAPBLOCKUNITS * FRACUNIT)) <> 0 then
    start := start + (MAPBLOCKUNITS * FRACUNIT)
          - ((start - bmaporgy) mod (MAPBLOCKUNITS * FRACUNIT));
  _end := m_y + m_h;

  // draw horizontal gridlines
  ml.a.x := m_x;
  ml.b.x := m_x + m_w;
  y := start;
  while y < _end do
  begin
    ml.a.y := y;
    ml.b.y := y;
    AM_drawMline(@ml, color);
    y := y + (MAPBLOCKUNITS * FRACUNIT);
  end;
end;

//
// Determines visible lines, draws them.
// This is LineDef based, not LineSeg based.
//
procedure AM_drawWalls;
var
  i: integer;
  l: mline_t;
begin
  for i := 0 to numlines - 1 do
  begin
    l.a.x := lines[i].v1.x;
    l.a.y := lines[i].v1.y;
    l.b.x := lines[i].v2.x;
    l.b.y := lines[i].v2.y;
    if (cheating <> 0) or ((lines[i].flags and ML_MAPPED) <> 0) then
    begin
      if ((lines[i].flags and LINE_NEVERSEE) <> 0) and (cheating = 0) then
        continue;
      if not boolval(lines[i].backsector) then
      begin
        AM_drawMline(@l, WALLCOLORS + lightlev);
      end
      else
      begin
        if lines[i].special = 39 then
        begin // teleporters
          AM_drawMline(@l, WALLCOLORS + WALLRANGE div 2);
        end
        else if (lines[i].flags and ML_SECRET) <> 0 then // secret door
        begin
          if cheating <> 0 then
            AM_drawMline(@l, SECRETWALLCOLORS + lightlev)
          else
            AM_drawMline(@l, WALLCOLORS + lightlev);
        end
        else if lines[i].backsector.floorheight <> lines[i].frontsector.floorheight then
        begin
          AM_drawMline(@l, FDWALLCOLORS + lightlev); // floor level change
        end
        else if lines[i].backsector.ceilingheight <> lines[i].frontsector.ceilingheight then
        begin
          AM_drawMline(@l, CDWALLCOLORS + lightlev); // ceiling level change
        end
        else if cheating <> 0 then
        begin
          AM_drawMline(@l, TSWALLCOLORS + lightlev);
        end;
      end;
    end
    else if boolval(plr.powers[Ord(pw_allmap)]) then
    begin
      if (lines[i].flags and LINE_NEVERSEE) = 0 then
        AM_drawMline(@l, GRAYS + 3);
    end;
  end;
end;


//
// Rotation in 2D.
// Used to rotate player arrow line character.
//
procedure AM_rotate(x: Pfixed_t; y: Pfixed_t; a: angle_t);
var
  tmpx: fixed_t;
begin
  tmpx :=  FixedMul(x^, finecosine[_SHRW(a, ANGLETOFINESHIFT)])
        - FixedMul(y^, finesine[_SHRW(a, ANGLETOFINESHIFT)]);

  y^ :=  FixedMul(x^, finesine[_SHRW(a, ANGLETOFINESHIFT)])
      + FixedMul(y^, finecosine[_SHRW(a, ANGLETOFINESHIFT)]);

  x^ := tmpx;
end;

procedure AM_drawLineCharacter(lineguy: Pmline_tArray; lineguylines: integer;
  scale: fixed_t; angle: angle_t; color: integer;
  x: fixed_t; y: fixed_t);
var
  i: integer;
  l: mline_t;
begin
  for i := 0 to lineguylines - 1 do
  begin
    l.a.x := lineguy[i].a.x;
    l.a.y := lineguy[i].a.y;

    if scale <> 0 then
    begin
      l.a.x := FixedMul(scale, l.a.x);
      l.a.y := FixedMul(scale, l.a.y);
    end;

    if angle <> 0 then
      AM_rotate(@l.a.x, @l.a.y, angle);

    l.a.x := l.a.x + x;
    l.a.y := l.a.y + y;

    l.b.x := lineguy[i].b.x;
    l.b.y := lineguy[i].b.y;

    if scale <> 0 then
    begin
      l.b.x := FixedMul(scale, l.b.x);
      l.b.y := FixedMul(scale, l.b.y);
    end;

    if angle <> 0 then
      AM_rotate(@l.b.x, @l.b.y, angle);

    l.b.x := l.b.x + x;
    l.b.y := l.b.y + y;

    AM_drawMline(@l, color);
  end;
end;

procedure AM_drawPlayers;
const
  their_colors: array[0..MAXPLAYERS - 1] of integer = (GREENS, GRAYS, BROWNS, REDS);
var i: integer;
    p: Pplayer_t;
    their_color, color: integer;
begin
  if not netgame then
  begin
    if cheating <> 0 then
      AM_drawLineCharacter
        (@cheat_player_arrow, NUMCHEATPLYRLINES, 0,
        plr.mo.angle, WHITE, plr.mo.x, plr.mo.y)
    else
      AM_drawLineCharacter
        (@player_arrow, NUMPLYRLINES, 0, plr.mo.angle,
        WHITE, plr.mo.x, plr.mo.y);
    exit;
  end;

  their_color := -1;
  for i := 0 to MAXPLAYERS - 1 do
  begin
    inc(their_color);
    p := @players[i];

    if (deathmatch <> 0) and (not singledemo) and (p <> plr) then
      continue;

    if not playeringame[i] then
      continue;

    if boolval(p.powers[Ord(pw_invisibility)]) then
      color := 246 // *close* to black
    else
      color := their_colors[their_color];

    AM_drawLineCharacter
      (@player_arrow, NUMPLYRLINES, 0, p.mo.angle,
       color, p.mo.x, p.mo.y);
  end;
end;

procedure AM_drawThings(colors: integer; colorrange: integer);
var i: integer;
    t: Pmobj_t;
begin
  for i := 0 to numsectors - 1 do
  begin
    t := sectors[i].thinglist;
    while t <> nil do
    begin
      AM_drawLineCharacter
        (@thintriangle_guy, NUMTHINTRIANGLEGUYLINES,
        16 * FRACUNIT, t.angle, colors + lightlev, t.x, t.y);
      t := t.snext;
    end;
  end;
end;

procedure AM_drawMarks;
var i, fx, fy, w, h: integer;
begin
  for i := 0 to AM_NUMMARKPOINTS - 1 do
  begin
    if markpoints[i].x <> -1 then
    begin
      w := 5; // because something's wrong with the wad, i guess
      h := 6; // because something's wrong with the wad, i guess
      fx := CXMTOF(markpoints[i].x);
      fy := CYMTOF(markpoints[i].y);
      if (fx >= f_x) and (fx <= f_w - w) and (fy >= f_y) and (fy <= f_h - h) then
        V_DrawPatch(fx, fy, _FG, marknums[i], false);
    end;
  end;
end;

procedure AM_drawCrosshair(color: integer);
begin
  fb[(f_w * (f_h + 1)) div 2] := color; // single point for now
end;

procedure AM_Drawer;
begin
  if not automapactive then
    exit;

  AM_clearFB(BACKGROUND);
  if grid then
    AM_drawGrid(GRIDCOLORS);
  AM_drawWalls;
  AM_drawPlayers;
  if cheating = 2 then
    AM_drawThings(THINGCOLORS, THINGRANGE);
  AM_drawCrosshair(XHAIRCOLORS);

  AM_drawMarks;

  V_MarkRect(f_x, f_y, f_w, f_h, false);
end;


(* Unit p_inter implementation *)


                                                                                                                       

const
  BONUSADD = 6;

//
// GET STUFF
//

//
// P_GiveAmmo
// Num is the number of clip loads,
// not the individual count (0= 1/2 clip).
// Returns false if the ammo can't be picked up at all
//
function P_GiveAmmo(player: Pplayer_t; ammo: ammotype_t; num: integer): boolean;
var
  oldammo: integer;
begin
  if ammo = am_noammo then
  begin
    result := false;
    exit;
  end;

  if (Ord(ammo) < 0) or (Ord(ammo) > Ord(NUMAMMO)) then
    I_Error('P_GiveAmmo(): bad type %d', [Ord(ammo)]);

  if player.ammo[Ord(ammo)] = player.maxammo[Ord(ammo)] then
  begin
    result := false;
    exit;
  end;

  if num <> 0 then
    num := num * clipammo[Ord(ammo)]
  else
    num := clipammo[Ord(ammo)] div 2;

  if (gameskill = sk_baby) or (gameskill = sk_nightmare) then
  begin
    // give double ammo in trainer mode,
    // you'll need in nightmare
    num := num * 2
  end;


  oldammo := player.ammo[Ord(ammo)];
  player.ammo[Ord(ammo)] := player.ammo[Ord(ammo)] + num;

  if player.ammo[Ord(ammo)] > player.maxammo[Ord(ammo)] then
    player.ammo[Ord(ammo)] := player.maxammo[Ord(ammo)];

  // If non zero ammo,
  // don't change up weapons,
  // player was lower on purpose.
  if boolval(oldammo) then
  begin
    result := true;
    exit;
  end;

  // We were down to zero,
  // so select a new weapon.
  // Preferences are not user selectable.
  case ammo of
    am_clip:
      begin
        if player.readyweapon = wp_fist then
        begin
          if boolval(player.weaponowned[Ord(wp_chaingun)]) then
            player.pendingweapon := wp_chaingun
          else
            player.pendingweapon := wp_pistol;
        end;
      end;
    am_shell:
      begin
        if (player.readyweapon = wp_fist) or
           (player.readyweapon = wp_pistol) then
        begin
          if boolval(player.weaponowned[Ord(wp_shotgun)]) then
            player.pendingweapon := wp_shotgun;
        end;
      end;
    am_cell:
      begin
        if (player.readyweapon = wp_fist) or
           (player.readyweapon = wp_pistol) then
        begin
          if boolval(player.weaponowned[Ord(wp_plasma)]) then
            player.pendingweapon := wp_plasma;
        end;
      end;
    am_misl:
      begin
        if player.readyweapon = wp_fist then
        begin
          if boolval(player.weaponowned[Ord(wp_missile)]) then
            player.pendingweapon := wp_missile;
        end
      end;
  end;

  result := true;
end;

//
// P_GiveWeapon
// The weapon name may have a MF_DROPPED flag ored in.
//
function P_GiveWeapon(player: Pplayer_t; weapon: weapontype_t; dropped: boolean): boolean;
var
  gaveammo: boolean;
  gaveweapon: boolean;
begin
  if netgame and (deathmatch <> 2) and (not dropped) then
  begin
	// leave placed weapons forever on net games
    if boolval(player.weaponowned[Ord(weapon)]) then
    begin
      result := false;
      exit;
    end;

    player.bonuscount := player.bonuscount + BONUSADD;
    player.weaponowned[Ord(weapon)] := 1;

    if boolval(deathmatch) then
      P_GiveAmmo(player, weaponinfo[Ord(weapon)].ammo, 5)
    else
      P_GiveAmmo(player, weaponinfo[Ord(weapon)].ammo, 2);
    player.pendingweapon := weapon;

    if (player = @players[consoleplayer]) then
      S_StartSound(nil, Ord(sfx_wpnup));
    result := false;
    exit;
  end;

  if weaponinfo[Ord(weapon)].ammo <> am_noammo then
  begin
  // give one clip with a dropped weapon,
  // two clips with a found weapon
    if dropped then
      gaveammo := P_GiveAmmo(player, weaponinfo[Ord(weapon)].ammo, 1)
    else
      gaveammo := P_GiveAmmo(player, weaponinfo[Ord(weapon)].ammo, 2);
  end
  else
    gaveammo := false;
	
  if boolval(player.weaponowned[Ord(weapon)]) then
    gaveweapon := false
  else
  begin
    gaveweapon := true;
    player.weaponowned[Ord(weapon)] := 1;
    player.pendingweapon := weapon;
  end;

  result := gaveweapon or gaveammo;
end;

//
// P_GiveBody
// Returns false if the body isn't needed at all
//
function P_GiveBody(player: Pplayer_t; num: integer): boolean;
begin
  if player.health >= MAXHEALTH then
  begin
    result := false;
    exit;
  end;

  player.health := player.health + num;
  if player.health > MAXHEALTH then
    player.health := MAXHEALTH;
  player.mo.health := player.health;

  result := true;
end;

//
// P_GiveArmor
// Returns false if the armor is worse
// than the current armor.
//
function P_GiveArmor(player: Pplayer_t; armortype: integer): boolean;
var
  hits: integer;
begin
  hits := armortype * 100;
  if player.armorpoints >= hits then
  begin
    result := false;  // don't pick up
    exit;
  end;

  player.armortype := armortype;
  player.armorpoints := hits;

  result := true;
end;

//
// P_GiveCard
//
procedure P_GiveCard(player: Pplayer_t; card: card_t);
begin
  if player.cards[Ord(card)] then
    exit;

  player.bonuscount := BONUSADD;
  player.cards[Ord(card)] := true;
end;

//
// P_GivePower
//
function P_GivePower(player: Pplayer_t; power: (*powertype_t*)integer): boolean;
begin
  if power = Ord(pw_invulnerability) then
  begin
    player.powers[power] := INVULNTICS;
    result := true;
    exit;
  end;

  if power = Ord(pw_invisibility) then
  begin
    player.powers[power] := INVISTICS;
    player.mo.flags := player.mo.flags or MF_SHADOW;
    result := true;
    exit;
  end;

  if power = Ord(pw_infrared) then
  begin
    player.powers[power] := INFRATICS;
    result := true;
    exit;
  end;

  if power = Ord(pw_ironfeet) then
  begin
    player.powers[power] := IRONTICS;
    result := true;
    exit;
  end;

  if power = Ord(pw_strength) then
  begin
    P_GiveBody(player, 100);
    player.powers[power] := 1;
    result := true;
    exit;
  end;

  if boolval(player.powers[power]) then
    result := false // already got it
  else
  begin
    player.powers[power] := 1;
    result := true;
  end;
end;

//
// P_TouchSpecialThing
//
procedure P_TouchSpecialThing(special: Pmobj_t; toucher: Pmobj_t);
var
  player: Pplayer_t;
  i: integer;
  delta: fixed_t;
  sound: integer;
begin
  delta := special.z - toucher.z;

  if (delta > toucher.height) or (delta < - 8 * FRACUNIT) then
	// out of reach
    exit;


  sound := Ord(sfx_itemup);
  player := toucher.player;

  // Dead thing touching.
  // Can happen with a sliding player corpse.
  if toucher.health <= 0 then
    exit;

  // Identify by sprite.
  case special.sprite of
	// armor
    SPR_ARM1:
      begin
        if not P_GiveArmor(player, 1) then
          exit;
        player._message := GOTARMOR;
      end;

    SPR_ARM2:
      begin
        if not P_GiveArmor(player, 2) then
          exit;
        player._message := GOTMEGA;
      end;

	// bonus items
    SPR_BON1:
      begin
        player.health := player.health + 1; // can go over 100%
        if player.health > 200 then
          player.health := 200;
        player.mo.health := player.health;
        player._message := GOTHTHBONUS;
      end;

    SPR_BON2:
      begin
        player.armorpoints := player.armorpoints + 1; // can go over 100%
        if player.armorpoints > 200 then
          player.armorpoints := 200;
        if not boolval(player.armortype) then
          player.armortype := 1;
        player._message := GOTARMBONUS;
      end;

    SPR_SOUL:
      begin
        player.health := player.health + 100;
        if player.health > 200 then
          player.health := 200;
        player.mo.health := player.health;
        player._message := GOTSUPER;
        sound := Ord(sfx_getpow);
      end;

    SPR_MEGA:
      begin
        if gamemode <> commercial then
          exit;
        player.health := 200;
        player.mo.health := player.health;
        P_GiveArmor(player, 2);
        player._message := GOTMSPHERE;
        sound := Ord(sfx_getpow);
      end;

	// cards
	// leave cards for everyone
    SPR_BKEY:
      begin
        if not player.cards[Ord(it_bluecard)] then
          player._message := GOTBLUECARD;
        P_GiveCard(player, it_bluecard);
      if netgame then
        exit;
      end;

    SPR_YKEY:
      begin
        if not player.cards[Ord(it_yellowcard)] then
          player._message := GOTYELWCARD;
        P_GiveCard(player, it_yellowcard);
        if netgame then
          exit;
      end;

    SPR_RKEY:
      begin
        if not player.cards[Ord(it_redcard)] then
          player._message := GOTREDCARD;
        P_GiveCard(player, it_redcard);
        if netgame then
          exit;
      end;

    SPR_BSKU:
      begin
        if not player.cards[Ord(it_blueskull)] then
          player._message := GOTBLUESKUL;
        P_GiveCard(player, it_blueskull);
        if netgame then
          exit;
      end;
      	
    SPR_YSKU:
      begin
        if not player.cards[Ord(it_yellowskull)] then
          player._message := GOTYELWSKUL;
        P_GiveCard(player, it_yellowskull);
        if netgame then
          exit;
      end;

    SPR_RSKU:
      begin
        if not player.cards[Ord(it_redskull)] then
          player._message := GOTREDSKULL;
        P_GiveCard(player, it_redskull);
        if netgame then
          exit;
      end;
	
	// medikits, heals
    SPR_STIM:
      begin
        if not P_GiveBody(player, 10) then
          exit;
        player._message := GOTSTIM;
      end;

    SPR_MEDI:
      begin
        if not P_GiveBody(player, 25) then
          exit;

        if player.health < 25 then
          player._message := GOTMEDINEED
        else
          player._message := GOTMEDIKIT;
      end;

	// power ups
    SPR_PINV:
      begin
        if not P_GivePower(player, Ord(pw_invulnerability)) then
          exit;
        player._message := GOTINVUL;
        sound := Ord(sfx_getpow);
      end;

    SPR_PSTR:
      begin
        if not P_GivePower(player, Ord(pw_strength)) then
          exit;
        player._message := GOTBERSERK;
        if player.readyweapon <> wp_fist then
          player.pendingweapon := wp_fist;
        sound := Ord(sfx_getpow);
      end;

    SPR_PINS:
      begin
        if not P_GivePower(player, Ord(pw_invisibility)) then
          exit;
        player._message := GOTINVIS;
        sound := Ord(sfx_getpow);
      end;

    SPR_SUIT:
      begin
        if not P_GivePower(player, Ord(pw_ironfeet)) then
          exit;
        player._message := GOTSUIT;
        sound := Ord(sfx_getpow);
      end;
	
    SPR_PMAP:
      begin
        if not P_GivePower(player, Ord(pw_allmap)) then
          exit;
        player._message := GOTMAP;
        sound := Ord(sfx_getpow);
      end;

    SPR_PVIS:
      begin
        if not P_GivePower(player, Ord(pw_infrared)) then
          exit;
        player._message := GOTVISOR;
        sound := Ord(sfx_getpow);
      end;

	// ammo
    SPR_CLIP:
      begin
        if boolval(special.flags and MF_DROPPED) then
        begin
          if not P_GiveAmmo(player, am_clip, 0) then
            exit;
        end
        else
        begin
          if not P_GiveAmmo(player, am_clip, 1) then
            exit;
        end;
        player._message := GOTCLIP;
      end;

    SPR_AMMO:
      begin
        if not P_GiveAmmo(player, am_clip,5) then
          exit;
        player._message := GOTCLIPBOX;
      end;

    SPR_ROCK:
      begin
        if not P_GiveAmmo(player, am_misl, 1) then
          exit;
        player._message := GOTROCKET;
      end;

    SPR_BROK:
      begin
        if not P_GiveAmmo(player, am_misl, 5) then
          exit;
        player._message := GOTROCKBOX;
      end;

    SPR_CELL:
      begin
        if not P_GiveAmmo(player, am_cell, 1) then
          exit;
        player._message := GOTCELL;
      end;

    SPR_CELP:
      begin
        if not P_GiveAmmo(player, am_cell, 5) then
          exit;
        player._message := GOTCELLBOX;
      end;

    SPR_SHEL:
      begin
        if not P_GiveAmmo(player, am_shell, 1) then
          exit;
        player._message := GOTSHELLS;
      end;

    SPR_SBOX:
      begin
        if not P_GiveAmmo(player, am_shell, 5) then
          exit;
        player._message := GOTSHELLBOX;
      end;

    SPR_BPAK:
      begin
        if not player.backpack then
        begin
          for i := 0 to Ord(NUMAMMO) - 1 do
            player.maxammo[i] := player.maxammo[i] * 2;
          player.backpack := true;
        end;
        for i := 0 to Ord(NUMAMMO) - 1 do
          P_GiveAmmo(player, ammotype_t(i), 1);
        player._message := GOTBACKPACK;
      end;

	// weapons
    SPR_BFUG:
      begin
        if not P_GiveWeapon(player, wp_bfg, false) then
          exit;
        player._message := GOTBFG9000;
        sound := Ord(sfx_wpnup);
      end;

    SPR_MGUN:
      begin
        if not P_GiveWeapon(player, wp_chaingun, boolval(special.flags and MF_DROPPED)) then
          exit;
        player._message := GOTCHAINGUN;
        sound := Ord(sfx_wpnup);
      end;

    SPR_CSAW:
      begin
        if not P_GiveWeapon(player, wp_chainsaw, false) then
          exit;
        player._message := GOTCHAINSAW;
        sound := Ord(sfx_wpnup);
      end;

    SPR_LAUN:
      begin
        if not P_GiveWeapon(player, wp_missile, false) then
          exit;
        player._message := GOTLAUNCHER;
        sound := Ord(sfx_wpnup);
      end;

    SPR_PLAS:
      begin
        if not P_GiveWeapon(player, wp_plasma, false) then
          exit;
        player._message := GOTPLASMA;
        sound := Ord(sfx_wpnup);
      end;

    SPR_SHOT:
      begin
        if not P_GiveWeapon(player, wp_shotgun, boolval(special.flags and MF_DROPPED)) then
          exit;
        player._message := GOTSHOTGUN;
        sound := Ord(sfx_wpnup);
      end;

    SPR_SGN2:
      begin
        if not P_GiveWeapon(player, wp_supershotgun, boolval(special.flags and MF_DROPPED)) then
          exit;
        player._message := GOTSHOTGUN2;
        sound := Ord(sfx_wpnup);
      end;

  else
    I_Error('P_TouchSpecialThing(): Unknown gettable thing');
  end;

  if (special.flags and MF_COUNTITEM) <> 0 then
    player.itemcount := player.itemcount + 1;
  P_RemoveMobj(special);
  player.bonuscount := player.bonuscount + BONUSADD;
  if player = @players[consoleplayer] then
    S_StartSound(nil, sound);
end;

//
// KillMobj
//
procedure P_KillMobj(source: Pmobj_t; target: Pmobj_t);
var
  item: mobjtype_t;
  mo: Pmobj_t;
begin
  target.flags := target.flags and (not (MF_SHOOTABLE or MF_FLOAT or MF_SKULLFLY));

  if target._type <> MT_SKULL then
    target.flags := target.flags and (not MF_NOGRAVITY);

  target.flags := target.flags or (MF_CORPSE or MF_DROPOFF);
  target.height := _SHR(target.height, 2);

  if boolval(source) and boolval(source.player) then
  begin
    // count for intermission
    if (target.flags and MF_COUNTKILL) <> 0 then
      Pplayer_t(source.player).killcount := Pplayer_t(source.player).killcount + 1;

    if target.player <> nil then
      Pplayer_t(source.player).frags[pOperation(target.player, @players[0], '-', SizeOf(players[0]))] :=
        Pplayer_t(source.player).frags[pOperation(target.player, @players[0], '-', SizeOf(players[0]))] + 1;
  end
  else if (not netgame) and boolval(target.flags and MF_COUNTKILL) then
  begin
    // count all monster deaths,
    // even those caused by other monsters
    players[0].killcount := players[0].killcount + 1;
  end;

  if boolval(target.player) then
  begin
    // count environment kills against you
    if not boolval(source) then
      Pplayer_t(target.player).frags[pOperation(target.player, @players[0], '-', SizeOf(players[0]))] :=
        Pplayer_t(target.player).frags[pOperation(target.player, @players[0], '-', SizeOf(players[0]))] + 1;

    target.flags := target.flags and (not MF_SOLID);
    Pplayer_t(target.player).playerstate := PST_DEAD;
    P_DropWeapon(target.player);

    if (target.player = @players[consoleplayer]) and automapactive then
    begin
      // don't die in auto map,
      // switch view prior to dying
      AM_Stop;
    end;

  end;

  if (target.health < -target.info.spawnhealth) and boolval(target.info.xdeathstate) then
    P_SetMobjState(target, statenum_t(target.info.xdeathstate))
  else
    P_SetMobjState(target, statenum_t(target.info.deathstate));
  target.tics := target.tics - (P_Random and 3);

  if target.tics < 1 then
    target.tics := 1;

    //	I_StartSound (&actor->r, actor->info->deathsound);


  // Drop stuff.
  // This determines the kind of object spawned
  // during the death frame of a thing.
  case target._type of
    MT_WOLFSS,
    MT_POSSESSED:
      item := MT_CLIP;
    MT_SHOTGUY:
      item := MT_SHOTGUN;
    MT_CHAINGUY:
      item := MT_CHAINGUN;
  else
    exit;
  end;

  mo := P_SpawnMobj(target.x,target.y, ONFLOORZ, item);
  mo.flags := mo.flags or MF_DROPPED; // special versions of items
end;

//
// P_DamageMobj
// Damages both enemies and players
// "inflictor" is the thing that caused the damage
//  creature or missile, can be NULL (slime, etc)
// "source" is the thing to target after taking damage
//  creature or NULL
// Source and inflictor are the same for melee attacks.
// Source can be NULL for slime, barrel explosions
// and other environmental stuff.
//
procedure P_DamageMobj(target, inflictor, source: Pmobj_t; damage: integer);
var
  ang: angle_t;
  saved: integer;
  player: Pplayer_t;
  thrust: fixed_t;
begin	
  if target.flags and MF_SHOOTABLE = 0 then
    exit; // shouldn't happen...

  if target.health <= 0 then
    exit;

  if target.flags and MF_SKULLFLY <> 0 then
  begin
    target.momx := 0;
    target.momy := 0;
    target.momz := 0;
  end;

  player := target.player;
  if (player <> nil) and (gameskill = sk_baby) then
    damage := _SHR(damage, 1); // take half damage in trainer mode


  // Some close combat weapons should not
  // inflict thrust and push the victim out of reach,
  // thus kick away unless using the chainsaw.
  if boolval(inflictor) and (not boolval(target.flags and MF_NOCLIP)) and
    ((source = nil) or (source.player = nil) or (Pplayer_t(source.player).readyweapon <> wp_chainsaw)) then
  begin
    ang := R_PointToAngle2(inflictor.x, inflictor.y, target.x, target.y);

    thrust := (damage * (_SHR(FRACUNIT, 3)) * 100) div target.info.mass;

    // make fall forwards sometimes
    if (damage < 40) and (damage > target.health) and
       (target.z - inflictor.z > 64 * FRACUNIT) and boolval(P_Random and 1) then
    begin
	    ang := ang + ANG180;
	    thrust := thrust * 4;
    end;

    ang := _SHRW(ang, ANGLETOFINESHIFT);
    target.momx := target.momx + FixedMul(thrust, finecosine[ang]);
    target.momy := target.momy + FixedMul(thrust, finesine[ang]);
  end;

  // player specific
  if player <> nil then
  begin
    // end of game hell hack
    if (Psubsector_t(target.subsector).sector.special = 11) and (damage >= target.health) then
      damage := target.health - 1;

    // Below certain threshold,
    // ignore damage in GOD mode, or with INVUL power.
    if (damage < 1000) and
       (boolval(player.cheats and CF_GODMODE) or boolval(player.powers[Ord(pw_invulnerability)])) then
      exit;

    if player.armortype <> 0 then
    begin
      if player.armortype = 1 then
        saved := damage div 3
      else
        saved := damage div 2;

      if player.armorpoints <= saved then
      begin
        // armor is used up
        saved := player.armorpoints;
        player.armortype := 0;
	    end;
	    player.armorpoints := player.armorpoints - saved;
      damage := damage - saved;
    end;
    player.health := player.health - damage;  // mirror mobj health here for Dave
    if player.health < 0 then
      player.health := 0;

    player.attacker := source;
    player.damagecount := player.damagecount + damage;  // add damage after armor / invuln

    if player.damagecount > 100 then
      player.damagecount := 100;  // teleport stomp does 10k points...

  end;

  // do the damage
  target.health := target.health - damage;
  if target.health <= 0 then
  begin
    P_KillMobj(source, target);
    exit;
  end;

  if (P_Random < target.info.painchance) and
	   (not boolval(target.flags and MF_SKULLFLY)) then
  begin
    target.flags := target.flags or MF_JUSTHIT; // fight back!
    P_SetMobjState(target, statenum_t(target.info.painstate));
  end;

  target.reactiontime := 0; // we're awake now...

  if ((not boolval(target.threshold)) or (target._type = MT_VILE)) and
     (source <> nil) and (source <> target) and (source._type <> MT_VILE) then
  begin
    // if not intent on another player,
    // chase after this one
    target.target := source;
    target.threshold := BASETHRESHOLD;
    if (target.state = @states[target.info.spawnstate]) and
       (target.info.seestate <> Ord(S_NULL)) then
      P_SetMobjState(target, statenum_t(target.info.seestate));
  end;
end;



(* Unit p_user implementation *)


                                                                                                                                                                                                                      

const
// Index of the special effects (INVUL inverse) map.
  INVERSECOLORMAP = 32;

//
// Movement.
//
const
// 16 pixels of bob
  MAXBOB = $100000;

var
  onground: boolean;

//
// P_Thrust
// Moves the given origin along a given angle.
//
procedure P_Thrust(player: Pplayer_t; angle: angle_t; const move: fixed_t);
begin
  angle := _SHRW(angle, ANGLETOFINESHIFT);

  player.mo.momx := player.mo.momx + FixedMul(move, finecosine[angle]);
  player.mo.momy := player.mo.momy + FixedMul(move, finesine[angle]);
end;

//
// P_CalcHeight
// Calculate the walking / running height adjustment
//
procedure P_CalcHeight(player: Pplayer_t);
var
  angle: integer;
  bob: fixed_t;
begin
  // Regular movement bobbing
  // (needs to be calculated for gun swing
  // even if not on ground)
  // OPTIMIZE: tablify angle
  // Note: a LUT allows for effects
  //  like a ramp with low health.

  player.bob :=	FixedMul(player.mo.momx, player.mo.momx) +
                FixedMul(player.mo.momy, player.mo.momy);
  player.bob :=	player.bob div 4;

  if player.bob > MAXBOB then
    player.bob := MAXBOB;

  if boolval(player.cheats and CF_NOMOMENTUM) or (not onground) then
  begin
    player.viewz := player.mo.z + PVIEWHEIGHT;

    if player.viewz > player.mo.ceilingz - 4 * FRACUNIT then
      player.viewz := player.mo.ceilingz - 4 * FRACUNIT;

    player.viewz := player.mo.z + player.viewheight; 
    exit;
  end;

  angle := (FINEANGLES div 20 * leveltime) and FINEMASK;
  bob := FixedMul(player.bob div 2, finesine[angle]);


  // move viewheight
  if player.playerstate = PST_LIVE then
  begin
    player.viewheight := player.viewheight + player.deltaviewheight;

    if player.viewheight > PVIEWHEIGHT then
    begin
      player.viewheight := PVIEWHEIGHT;
      player.deltaviewheight := 0;
    end;

    if player.viewheight < PVIEWHEIGHT div 2 then
    begin
      player.viewheight := PVIEWHEIGHT div 2;
      if player.deltaviewheight <= 0 then
        player.deltaviewheight := 1;
    end;

    if boolval(player.deltaviewheight) then
    begin
      player.deltaviewheight := player.deltaviewheight + FRACUNIT div 4;
      if not boolval(player.deltaviewheight) then
        player.deltaviewheight := 1;
    end;
  end;
  player.viewz := player.mo.z + player.viewheight + bob;

  if player.viewz > player.mo.ceilingz - 4 * FRACUNIT then
    player.viewz := player.mo.ceilingz - 4 * FRACUNIT;
end;

//
// P_MovePlayer
//
procedure P_MovePlayer(player: Pplayer_t);
var
  cmd: Pticcmd_t;
  look: integer;
begin
{$IFDEF DEBUG}
  fprintf(debugfile, '--------------------------' + #13#10);
  fprintf(debugfile, 'P_MovePlayer()' + #13#10);
  fprintf(debugfile, 'leveltime = %d' + #13#10, [leveltime]);
  fprintf(debugfile, 'player.mo.angle = %d' + #13#10, [player.mo.angle]);
  fprintf(debugfile, 'player.mo.momz = %d' + #13#10, [player.mo.momz]);
  fprintf(debugfile, 'player.mo.z = %d' + #13#10, [player.mo.z]);
  fprintf(debugfile, 'player.mo.floorz = %d' + #13#10, [player.mo.floorz]);
  fprintf(debugfile, 'player.viewheight = %d' + #13#10, [player.viewheight]);
  fprintf(debugfile, 'player.viewz = %d' + #13#10, [player.viewz]);
{$ENDIF}

  cmd := @player.cmd;

  player.mo.angle := player.mo.angle + _SHLW(cmd.angleturn, 16);

  // Do not let the player control movement
  //  if not onground.
  onground := player.mo.z <= player.mo.floorz;

  if boolval(cmd.forwardmove) and onground then
    P_Thrust(player, player.mo.angle, cmd.forwardmove * 2048);

  if boolval(cmd.sidemove) and onground then
    P_Thrust(player, player.mo.angle - ANG90, cmd.sidemove * 2048);

  if (boolval(cmd.forwardmove) or boolval(cmd.sidemove)) and
     (player.mo.state = @states[Ord(S_PLAY)]) then
    P_SetMobjState(player.mo, S_PLAY_RUN1);

  look := cmd.look;
	if look > 7 then
    look := look - 16;

	if boolval(look) then
  begin
    if look = TOCENTER then
      player.centering := true
    else
    begin
      player.lookdir := player.lookdir + 5 * look;
      if (player.lookdir > 90) or (player.lookdir < -110) then
        player.lookdir := player.lookdir - 5 * look;
    end;
  end;

  if player.centering then
  begin
    if player.lookdir > 0 then
      player.lookdir := player.lookdir - 8
    else if player.lookdir < 0 then
      player.lookdir := player.lookdir + 8;

    if abs(player.lookdir) < 8 then
    begin
      player.lookdir := 0;
      player.centering := false;
    end;
  end;

end;

//
// P_DeathThink
// Fall on your face when dying.
// Decrease POV height to floor height.
//
const
  ANG5 = ANG90 div 18;
  ANG355 = ANG270 +  ANG5 * 17; // add by VJ 

procedure P_DeathThink(player: Pplayer_t);
var
  angle: angle_t;
  delta: angle_t;
begin
  P_MovePsprites(player);

  // fall to the ground
  if player.viewheight > 6 * FRACUNIT then
    player.viewheight := player.viewheight - FRACUNIT;

  if player.viewheight < 6 * FRACUNIT then
    player.viewheight := 6 * FRACUNIT;

  player.deltaviewheight := 0;
  onground := player.mo.z <= player.mo.floorz;
  P_CalcHeight(player);

  if boolval(player.attacker) and (player.attacker <> player.mo) then
  begin
    angle := R_PointToAngle2(
      player.mo.x, player.mo.y, player.attacker.x, player.attacker.y);

    delta := angle - player.mo.angle;

//    if (delta < ANG5) or (delta > LongWord(-ANG5)) then // VJ changed!, angle_t is integer now, not longword
    if (delta < ANG5) or (delta > ANG355) then
    begin
      // Looking at killer,
      //  so fade damage flash down.
      player.mo.angle := angle;

      if boolval(player.damagecount) then
        player.damagecount := player.damagecount - 1;
    end
    else if delta < ANG180 then
      player.mo.angle := player.mo.angle + ANG5
    else
      player.mo.angle := player.mo.angle - ANG5;
  end
  else if boolval(player.damagecount) then
    player.damagecount := player.damagecount - 1;


  if boolval(player.cmd.buttons and BT_USE) then
    player.playerstate := PST_REBORN;
end;

//
// P_PlayerThink
//
procedure P_PlayerThink(player: Pplayer_t);
var
  cmd: Pticcmd_t;
  newweapon: weapontype_t;
begin
  // fixme: do this in the cheat code
  if boolval(player.cheats and CF_NOCLIP) then
    player.mo.flags := player.mo.flags or MF_NOCLIP
  else
    player.mo.flags := player.mo.flags and (not MF_NOCLIP);

  // chain saw run forward
  cmd := @player.cmd;
  if boolval(player.mo.flags and MF_JUSTATTACKED) then
  begin
    cmd.angleturn := 0;
    cmd.forwardmove := $c800 div 512;
    cmd.sidemove := 0;
    player.mo.flags := player.mo.flags and (not MF_JUSTATTACKED);
  end;


  if player.playerstate = PST_DEAD then
  begin
    P_DeathThink(player);
    exit;
  end;

  // Move around.
  // Reactiontime is used to prevent movement
  //  for a bit after a teleport.
  if boolval(player.mo.reactiontime) then
    player.mo.reactiontime := player.mo.reactiontime - 1
  else
    P_MovePlayer(player);

  P_CalcHeight(player);

  if boolval(Psubsector_t(player.mo.subsector).sector.special) then
    P_PlayerInSpecialSector(player);

  // Check for weapon change.

  // A special event has no other buttons.
  if boolval(cmd.buttons and BT_SPECIAL) then
    cmd.buttons := 0;

  if boolval(cmd.buttons and BT_CHANGE) then
  begin
    // The actual changing of the weapon is done
    //  when the weapon psprite can do it
    //  (read: not in the middle of an attack).
    newweapon := weapontype_t(_SHR(cmd.buttons and BT_WEAPONMASK, BT_WEAPONSHIFT));

  	if (newweapon = wp_fist) and
       boolval(player.weaponowned[Ord(wp_chainsaw)]) and (not (
       (player.readyweapon = wp_chainsaw) and boolval(player.powers[Ord(pw_strength)]))) then
      newweapon := wp_chainsaw;


	  if (gamemode = commercial) and
  	   (newweapon = wp_shotgun) and
  	   boolval(player.weaponowned[Ord(wp_supershotgun)]) and
  	   (player.readyweapon <> wp_supershotgun) then
      newweapon := wp_supershotgun;


	  if boolval(player.weaponowned[Ord(newweapon)]) and
       (newweapon <> player.readyweapon) then
      // Do not go to plasma or BFG in shareware,
      //  even if cheated.
      if ((newweapon <> wp_plasma) and (newweapon <> wp_bfg)) or
         (gamemode <> shareware) then
        player.pendingweapon := newweapon;

  end;

  // check for use
  if boolval(cmd.buttons and BT_USE) then
  begin
    if not player.usedown then
    begin
      P_UseLines(player);
      player.usedown := true;
    end;
  end
  else
    player.usedown := false;

  // cycle psprites
  P_MovePsprites(player);

  // Counters, time dependend power ups.

  // Strength counts up to diminish fade.
  if boolval(player.powers[Ord(pw_strength)]) then
    player.powers[Ord(pw_strength)] := player.powers[Ord(pw_strength)] + 1;

  if boolval(player.powers[Ord(pw_invulnerability)]) then
    player.powers[Ord(pw_invulnerability)] := player.powers[Ord(pw_invulnerability)] - 1;

  if boolval(player.powers[Ord(pw_invisibility)]) then
  begin
    player.powers[Ord(pw_invisibility)] := player.powers[Ord(pw_invisibility)] - 1;
    if not boolval(player.powers[Ord(pw_invisibility)]) then
      player.mo.flags := player.mo.flags and (not MF_SHADOW);
  end;

  if boolval(player.powers[Ord(pw_infrared)]) then
    player.powers[Ord(pw_infrared)] := player.powers[Ord(pw_infrared)] - 1;

  if boolval(player.powers[Ord(pw_ironfeet)]) then
    player.powers[Ord(pw_infrared)] := player.powers[Ord(pw_infrared)] - 1;

  if boolval(player.damagecount) then
    player.damagecount := player.damagecount - 1;

  if boolval(player.bonuscount) then
    player.bonuscount := player.bonuscount - 1;


  // Handling colormaps.
  if boolval(player.powers[Ord(pw_invulnerability)]) then
  begin
    if (player.powers[Ord(pw_invulnerability)] > 4 * 32) or
       boolval(player.powers[Ord(pw_invulnerability)] and 8) then
      player.fixedcolormap := INVERSECOLORMAP
    else
      player.fixedcolormap := 0;
  end
  else if boolval(player.powers[Ord(pw_infrared)]) then
  begin
    if (player.powers[Ord(pw_infrared)] > 4 * 32) or
       boolval(player.powers[Ord(pw_infrared)] and 8) then
      // almost full bright
      player.fixedcolormap := 1
    else
      player.fixedcolormap := 0;
  end
  else
    player.fixedcolormap := 0;
end;



(* Unit p_pspr implementation *)


                                                                                                                                                                                                                                                                                                                                                                                                                                                

//
// VJ
// Adjust weapon bottom and top
//
const
  adjweapons: set of weapontype_t =
    [wp_chainsaw, wp_fist, wp_pistol, wp_shotgun, wp_plasma, wp_supershotgun];

function WEAPONBOTTOM(wp: weapontype_t): fixed_t;
begin
  if (SCREENHEIGHT <> 200) and (wp in adjweapons) then
    result := 148 * FRACUNIT
  else
    result := 128 * FRACUNIT;
end;

function WEAPONTOP(wp: weapontype_t): fixed_t;
begin
  if (SCREENHEIGHT <> 200) and (wp in adjweapons) then
    result := 52 * FRACUNIT
  else
    result := 32 * FRACUNIT;
end;

const
  LOWERSPEED = FRACUNIT * 6;
  RAISESPEED = FRACUNIT * 6;

// plasma cells for a bfg attack
  BFGCELLS = 40;

//
// P_SetPsprite
//
procedure P_SetPsprite(player: Pplayer_t; position: integer; stnum: statenum_t);
var
  psp: Ppspdef_t;
  state: Pstate_t;
begin
  psp := @player.psprites[position];
  repeat
    if not boolval(Ord(stnum)) then
    begin
      // object removed itself
      psp.state := nil;
      break;
    end;

    state := @states[Ord(stnum)];
    psp.state := state;
    psp.tics := state.tics; // could be 0

    if boolval(state.misc1) then
    begin
      // coordinate set
      psp.sx := (state.misc1 * FRACUNIT) div 320 * SCREENWIDTH;
      psp.sy := (state.misc2 * FRACUNIT) div 200 * SCREENHEIGHT;
    end;

    // Call action routine.
    // Modified handling.
    if Assigned(state.action.acp2) then
    begin
      state.action.acp2(player, psp);
      if not boolval(psp.state) then
        break;
    end;

    stnum := psp.state.nextstate;

  until boolval(psp.tics);
  // an initial state of 0 could cycle through
end;

//
// P_CalcSwing
//
var
  swingx: fixed_t;
  swingy: fixed_t;

procedure P_CalcSwing(player: Pplayer_t);
var
  swing: fixed_t;
  angle: integer;
begin
  // OPTIMIZE: tablify this.
  // A LUT would allow for different modes,
  //  and add flexibility.

  swing := player.bob;

  angle := (FINEANGLES div 70 * leveltime) and FINEMASK;
  swingx := FixedMul(swing, finesine[angle]);

  angle := (FINEANGLES div 70 * leveltime + FINEANGLES div 2) and FINEMASK;
  swingy := -FixedMul(swingx, finesine[angle]);
end;

//
// P_BringUpWeapon
// Starts bringing the pending weapon up
// from the bottom of the screen.
// Uses player
//
procedure P_BringUpWeapon(player: Pplayer_t);
var
  newstate: statenum_t;
begin
  if player.pendingweapon = wp_nochange then
    player.pendingweapon := player.readyweapon;

  if player.pendingweapon = wp_chainsaw then
    S_StartSound(player.mo, Ord(sfx_sawup));

  newstate := statenum_t(weaponinfo[Ord(player.pendingweapon)].upstate);

  player.pendingweapon := wp_nochange;
  player.psprites[Ord(ps_weapon)].sy := WEAPONBOTTOM(player.readyweapon);

  P_SetPsprite(player, Ord(ps_weapon), newstate);
end;

//
// P_CheckAmmo
// Returns true if there is enough ammo to shoot.
// If not, selects the next weapon to use.
//
function P_CheckAmmo(player: Pplayer_t): boolean;
var
  ammo: ammotype_t;
  count: integer;
begin
  ammo := weaponinfo[Ord(player.readyweapon)].ammo;

  // Minimal amount for one shot varies.
  if player.readyweapon = wp_bfg then
    count := BFGCELLS
  else if player.readyweapon = wp_supershotgun then
    count := 2 // Double barrel.
  else
    count := 1; // Regular.

  // Some do not need ammunition anyway.
  // Return if current ammunition sufficient.
  if (ammo = am_noammo) or (player.ammo[Ord(ammo)] >= count) then
  begin
    result := true;
    exit;
  end;

  // Out of ammo, pick a weapon to change to.
  // Preferences are set here.
  repeat
    if boolval(player.weaponowned[Ord(wp_plasma)]) and
       boolval(player.ammo[Ord(am_cell)]) and
      (gamemode <> shareware) then
      player.pendingweapon := wp_plasma
    else if boolval(player.weaponowned[Ord(wp_supershotgun)]) and
            (player.ammo[Ord(am_shell)] > 2) and
            (gamemode = commercial) then
      player.pendingweapon := wp_supershotgun
    else if boolval(player.weaponowned[Ord(wp_chaingun)]) and
            boolval(player.ammo[Ord(am_clip)]) then
      player.pendingweapon := wp_chaingun
    else if boolval(player.weaponowned[Ord(wp_shotgun)]) and
            boolval(player.ammo[Ord(am_shell)]) then
      player.pendingweapon := wp_shotgun
    else if boolval(player.ammo[Ord(am_clip)]) then
      player.pendingweapon := wp_pistol
    else if boolval(player.weaponowned[Ord(wp_chainsaw)]) then
      player.pendingweapon := wp_chainsaw
    else if boolval(player.weaponowned[Ord(wp_missile)]) and
            boolval(player.ammo[Ord(am_misl)]) then
      player.pendingweapon := wp_missile
    else if boolval(player.weaponowned[Ord(wp_bfg)]) and
            (player.ammo[Ord(am_cell)] > 40) and
            (gamemode <> shareware) then
      player.pendingweapon := wp_bfg
    else
      // If everything fails.
      player.pendingweapon := wp_fist;

  until not (player.pendingweapon = wp_nochange);

  // Now set appropriate weapon overlay.
  P_SetPsprite(player, Ord(ps_weapon), statenum_t(weaponinfo[Ord(player.readyweapon)].downstate));

  result := false;
end;

//
// P_FireWeapon.
//
procedure P_FireWeapon(player: Pplayer_t);
var
  newstate: statenum_t;
begin
  if P_CheckAmmo(player) then
  begin
    P_SetMobjState(player.mo, S_PLAY_ATK1);
    newstate := statenum_t(weaponinfo[Ord(player.readyweapon)].atkstate);
    P_SetPsprite(player, Ord(ps_weapon), newstate);
    P_NoiseAlert(player.mo, player.mo);
  end;
end;

//
// P_DropWeapon
// Player died, so put the weapon away.
//
procedure P_DropWeapon(player: Pplayer_t);
begin
  P_SetPsprite(player, Ord(ps_weapon), statenum_t(weaponinfo[Ord(player.readyweapon)].downstate));
end;

//
// A_WeaponReady
// The player can fire the weapon
// or change to another weapon at this time.
// Follows after getting weapon up,
// or after previous attack/fire sequence.
//
procedure A_WeaponReady(player: Pplayer_t; psp: Ppspdef_t);
var
  newstate: statenum_t;
  angle: integer;
begin
  // get out of attack state
  if (player.mo.state = @states[Ord(S_PLAY_ATK1)]) or
     (player.mo.state = @states[Ord(S_PLAY_ATK2)]) then
    P_SetMobjState(player.mo, S_PLAY);

  if (player.readyweapon = wp_chainsaw) and
     (psp.state = @states[Ord(S_SAW)]) then
    S_StartSound(player.mo, Ord(sfx_sawidl));

  // check for change
  //  if player is dead, put the weapon away
  if (player.pendingweapon <> wp_nochange) or (not boolval(player.health)) then
  begin
    // change weapon
    //  (pending weapon should allready be validated)
    newstate := statenum_t(weaponinfo[Ord(player.readyweapon)].downstate);
    P_SetPsprite(player, Ord(ps_weapon), newstate);
    exit;
  end;

  // check for fire
  //  the missile launcher and bfg do not auto fire
  if boolval(player.cmd.buttons and BT_ATTACK) then
  begin
    if (not player.attackdown) or
       ((player.readyweapon <> wp_missile) and (player.readyweapon <> wp_bfg)) then
    begin
      player.attackdown := true;
      P_FireWeapon(player);
      exit;
    end;
  end
  else
    player.attackdown := false;

  // bob the weapon based on movement speed
  angle := (128 * leveltime) and FINEMASK;
  psp.sx := FRACUNIT + FixedMul(player.bob, finecosine[angle]);
  angle := angle and (FINEANGLES div 2 - 1);
//  psp.sy := WEAPONTOP + FixedMul(player.bob, finesine[angle]) div 200 * SCREENHEIGHT;
  psp.sy := WEAPONTOP(player.readyweapon) + FixedMul(player.bob, finesine[angle]);
end;

//
// A_ReFire
// The player can re-fire the weapon
// without lowering it entirely.
//
procedure A_ReFire(player: Pplayer_t; psp: Ppspdef_t);
begin
  // check for fire
  //  (if a weaponchange is pending, let it go through instead)
  if boolval(player.cmd.buttons and BT_ATTACK) and
     (player.pendingweapon = wp_nochange) and
     boolval(player.health) then
  begin
    player.refire := player.refire + 1;
    P_FireWeapon(player);
  end
  else
  begin
    player.refire := 0;
    P_CheckAmmo(player);
  end;
end;

procedure A_CheckReload(player: Pplayer_t; psp: Ppspdef_t);
begin
  P_CheckAmmo(player);
{
#if 0
    if (player->ammo[am_shell]<2)
	P_SetPsprite (player, ps_weapon, S_DSNR1);
#endif
}
end;

//
// A_Lower
// Lowers current weapon,
//  and changes weapon at bottom.
//
procedure A_Lower(player: Pplayer_t; psp: Ppspdef_t);
begin
  psp.sy := psp.sy + LOWERSPEED;

  // Is already down.
  if psp.sy < WEAPONBOTTOM(player.readyweapon) then
    exit;

  // Player is dead.
  if player.playerstate = PST_DEAD then
  begin
    psp.sy := WEAPONBOTTOM(player.readyweapon);
    // don't bring weapon back up
    exit;
  end;

  // The old weapon has been lowered off the screen,
  // so change the weapon and start raising it
  if not boolval(player.health) then
  begin
    // Player is dead, so keep the weapon off screen.
    P_SetPsprite(player, Ord(ps_weapon), S_NULL);
    exit;
  end;

  player.readyweapon := player.pendingweapon;

  P_BringUpWeapon(player);
end;

//
// A_Raise
//
procedure A_Raise(player: Pplayer_t; psp: Ppspdef_t);
var
  newstate: statenum_t;
begin
  psp.sy := psp.sy - RAISESPEED;

  if psp.sy > WEAPONTOP(player.readyweapon) then
    exit;

  psp.sy := WEAPONTOP(player.readyweapon);

  // The weapon has been raised all the way,
  //  so change to the ready state.
  newstate := statenum_t(weaponinfo[Ord(player.readyweapon)].readystate);

  P_SetPsprite(player, Ord(ps_weapon), newstate);
end;

//
// A_GunFlash
//
procedure A_GunFlash(player: Pplayer_t; psp: Ppspdef_t);
begin
  P_SetMobjState(player.mo, S_PLAY_ATK2);
  P_SetPsprite(player, Ord(ps_flash), statenum_t(weaponinfo[Ord(player.readyweapon)].flashstate));
end;

//
// WEAPON ATTACKS
//


//
// A_Punch
//
procedure A_Punch(player: Pplayer_t; psp: Ppspdef_t);
var
  angle: angle_t;
  damage: integer;
  slope: integer;
begin
  damage := _SHL(P_Random mod 10 + 1, 1);

  if boolval(player.powers[Ord(pw_strength)]) then
    damage := damage * 10;

  angle := player.mo.angle;
  angle := angle + _SHLW(P_Random - P_Random, 18);
  slope := P_AimLineAttack(player.mo, angle, MELEERANGE);
  P_LineAttack(player.mo, angle, MELEERANGE, slope, damage);

  // turn to face target
  if linetarget <> nil then
  begin
    S_StartSound(player.mo, Ord(sfx_punch));
    player.mo.angle :=
      R_PointToAngle2(player.mo.x, player.mo.y, linetarget.x, linetarget.y);
  end;
end;

//
// A_Saw
//
procedure A_Saw(player: Pplayer_t; psp: Ppspdef_t);
var
  angle: angle_t;
  damage: integer;
  slope: integer;
begin
  damage := 2 * (P_Random mod 10 + 1);
  angle := player.mo.angle;
  angle := angle + _SHLW(P_Random - P_Random, 18);

  // use meleerange + 1 se the puff doesn't skip the flash
  slope := P_AimLineAttack(player.mo, angle, MELEERANGE + 1);
  P_LineAttack(player.mo, angle, MELEERANGE + 1, slope, damage);

  if linetarget = nil then
  begin
    S_StartSound(player.mo, Ord(sfx_sawful));
    exit;
  end;

  S_StartSound(player.mo, Ord(sfx_sawhit));

  // turn to face target
  angle :=
    R_PointToAngle2(player.mo.x, player.mo.y, linetarget.x, linetarget.y);
  if angle - player.mo.angle > ANG180 then
  begin
    if integer(angle - player.mo.angle) < - integer(ANG90 div 20) then
      player.mo.angle := angle + ANG90 div 21
  	else
      player.mo.angle := player.mo.angle - ANG90 div 20;
  end
  else
  begin
    if angle - player.mo.angle > ANG90 div 20 then
      player.mo.angle := angle - ANG90 div 21
    else
      player.mo.angle := player.mo.angle + ANG90 div 20;
  end;

  player.mo.flags := player.mo.flags or MF_JUSTATTACKED;
end;

//
// A_FireMissile
//
procedure A_FireMissile(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 1;
  P_SpawnPlayerMissile(player.mo, MT_ROCKET);
end;

//
// A_FireBFG
//
procedure A_FireBFG(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - BFGCELLS;
  P_SpawnPlayerMissile(player.mo, MT_BFG);
end;

//
// A_FirePlasma
//
procedure A_FirePlasma(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 1;

  P_SetPsprite(player,
    Ord(ps_flash), statenum_t(weaponinfo[Ord(player.readyweapon)].flashstate + (P_Random and 1)));

  P_SpawnPlayerMissile(player.mo, MT_PLASMA);
end;

//
// P_BulletSlope
// Sets a slope so a near miss is at aproximately
// the height of the intended target
//
var
  bulletslope: fixed_t;


procedure P_BulletSlope(mo: Pmobj_t);
var
  an: angle_t;
begin
  // see which target is to be aimed at
  an := mo.angle;
  bulletslope := P_AimLineAttack(mo, an, 16 * 64 * FRACUNIT);

  if linetarget = nil then
  begin
    an := an + _SHLW(1, 26);
    bulletslope := P_AimLineAttack (mo, an, 16 * 64 * FRACUNIT);
    if linetarget = nil then
    begin
      an := an - _SHLW(2, 26);
      bulletslope := P_AimLineAttack(mo, an, 16 * 64 * FRACUNIT);
      if linetarget = nil then
      begin
//        an := an + _SHLW(2, 26);
        bulletslope := (Pplayer_t(mo.player).lookdir * FRACUNIT) div 173;
      end;
    end;
  end;
end;

//
// P_GunShot
//
procedure P_GunShot(mo: Pmobj_t; accurate: boolean);
var
  angle: angle_t;
  damage: integer;
begin
  damage := 5 * ((P_Random mod 3) +1);
  angle := mo.angle;

  if not accurate then
	angle := angle + _SHLW(P_Random - P_Random, 18);

  P_LineAttack(mo, angle, MISSILERANGE, bulletslope, damage);
end;

//
// A_FirePistol
//
procedure A_FirePistol(player: Pplayer_t; psp: Ppspdef_t);
begin
  S_StartSound(player.mo, Ord(sfx_pistol));

  P_SetMobjState(player.mo, S_PLAY_ATK2);
  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 1;

  P_SetPsprite(player, Ord(ps_flash), statenum_t(weaponinfo[Ord(player.readyweapon)].flashstate));

  P_BulletSlope(player.mo);
  P_GunShot(player.mo, not boolval(player.refire));
end;

//
// A_FireShotgun
//
procedure A_FireShotgun(player: Pplayer_t; psp: Ppspdef_t);
var
  i: integer;
begin
  S_StartSound(player.mo, Ord(sfx_shotgn));
  P_SetMobjState(player.mo, S_PLAY_ATK2);

  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 1;

  P_SetPsprite(player, Ord(ps_flash), statenum_t(weaponinfo[Ord(player.readyweapon)].flashstate));

  P_BulletSlope(player.mo);

  for i := 0 to 6 do
    P_GunShot(player.mo, false);
end;

//
// A_FireShotgun2
//
procedure A_FireShotgun2(player: Pplayer_t; psp: Ppspdef_t);
var
  i: integer;
  angle: angle_t;
  damage: integer;
begin
  S_StartSound(player.mo, Ord(sfx_dshtgn));
  P_SetMobjState(player.mo, S_PLAY_ATK2);

  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 2;

  P_SetPsprite(player, Ord(ps_flash), statenum_t(weaponinfo[Ord(player.readyweapon)].flashstate));

  P_BulletSlope(player.mo);

  for i := 0 to 19 do
  begin
    damage := 5 * ((P_Random mod 3) + 1);
    angle := player.mo.angle;
    angle := angle + _SHLW(P_Random - P_Random, 19);  // VJ??? 19 = ANGLETOFINESHIFT
    P_LineAttack(player.mo, angle, MISSILERANGE,
      bulletslope + _SHL(P_Random - P_Random, 5), damage);
  end;
end;

//
// A_FireCGun
//
procedure A_FireCGun(player: Pplayer_t; psp: Ppspdef_t);
begin
  S_StartSound(player.mo, Ord(sfx_pistol));

  if not boolval(player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)]) then
    exit;

  P_SetMobjState(player.mo, S_PLAY_ATK2);
  player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] :=
    player.ammo[Ord(weaponinfo[Ord(player.readyweapon)].ammo)] - 1;

  P_SetPsprite(player, Ord(ps_flash),
    statenum_t(
      weaponinfo[Ord(player.readyweapon)].flashstate +
      pOperation(psp.state, @states[Ord(S_CHAIN1)], '-', SizeOf(states[0]))
    ));

  P_BulletSlope(player.mo);

  P_GunShot(player.mo, not boolval(player.refire));
end;

//
// ?
//
procedure A_Light0(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.extralight := 0;
end;

procedure A_Light1(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.extralight := 1;
end;

procedure A_Light2(player: Pplayer_t; psp: Ppspdef_t);
begin
  player.extralight := 2;
end;

//
// A_BFGSpray
// Spawn a BFG explosion on every monster in view
//
procedure A_BFGSpray(mo: Pmobj_t);
var
  i: integer;
  j: integer;
  damage: integer;
  an: angle_t;
  dan: angle_t;
begin
  // offset angles from its attack angle
{ VJ -> original code was: ---------------
  for i := 0 to 39 do
  begin
    an := mo.angle - ANG90 div 2 + (ANG90 div 40) * i;
-----------------}
  an := mo.angle - ANG90 div 2;
  dan := ANG90 div 40;
  for i := 0 to 39 do
  begin
    an := an + dan;

    // mo->target is the originator (player)
    //  of the missile
    P_AimLineAttack(mo.target, an, 16 * 64 * FRACUNIT);

    if linetarget = nil then
      continue;

    P_SpawnMobj(
      linetarget.x, linetarget.y, linetarget.z + _SHR(linetarget.height, 2), MT_EXTRABFG);

    damage := 0;
    for j := 0 to 14 do
      damage := damage + (P_Random and 7) + 1;

    P_DamageMobj(linetarget, mo.target, mo.target, damage);
  end;
end;

//
// A_BFGsound
//
procedure A_BFGsound(player: Pplayer_t; psp: Ppspdef_t);
begin
  S_StartSound(player.mo, Ord(sfx_bfg));
end;

//
// P_SetupPsprites
// Called at start of level for each player.
//
procedure P_SetupPsprites(player: Pplayer_t);
var
  i: integer;
begin
  // remove all psprites
  for i := 0 to Ord(NUMPSPRITES) - 1 do
    player.psprites[i].state := nil;

  // spawn the gun
  player.pendingweapon := player.readyweapon;
  P_BringUpWeapon(player);
end;

//
// P_MovePsprites
// Called every tic by player thinking routine.
//
procedure P_MovePsprites(player: Pplayer_t);
var
  i: integer;
  psp: Ppspdef_t;
  state: Pstate_t;
begin
  for i := 0 to Ord(NUMPSPRITES) - 1 do
  begin
    psp := @player.psprites[i];
    // a null state means not active
    state := psp.state;
    if boolval(state) then
    begin
      // drop tic count and possibly change state
      // a -1 tic count never changes
      if psp.tics <> -1 then
      begin
        psp.tics := psp.tics - 1;
        if not boolval(psp.tics) then
          P_SetPsprite(player, i, psp.state.nextstate);
      end;
    end;
  end;

  player.psprites[Ord(ps_flash)].sx := player.psprites[Ord(ps_weapon)].sx;
  player.psprites[Ord(ps_flash)].sy := player.psprites[Ord(ps_weapon)].sy;
end;


(* Unit d_main implementation *)


                                                                                                                                                                                                                                                                                                                           

const
  BGCOLOR = 7;
  FGCOLOR = 8;

//
// D-DoomLoop()
// Not a globally visible function,
//  just included for source reference,
//  called by D_DoomMain, never exits.
// Manages timing and IO,
//  calls all ?_Responder, ?_Ticker, and ?_Drawer,
//  calls I_GetTime, I_StartFrame, and I_StartTic
//
procedure D_DoomLoop; forward;


//
// D_PostEvent
// Called by the I/O functions when input is detected
//
procedure D_PostEvent(ev: Pevent_t);
begin
  events[eventhead] := ev^;
  inc(eventhead);
  eventhead := eventhead and (MAXEVENTS - 1);
end;

//
// D_ProcessEvents
// Send all the events of the given timestamp down the responder chain
//
procedure D_ProcessEvents;
var
  ev: Pevent_t;
begin
// IF STORE DEMO, DO NOT ACCEPT INPUT
  if (gamemode = commercial) and (W_CheckNumForName('map01') < 0) then
    exit;

  if I_GameFinished then
    exit;

  while eventtail <> eventhead do
  begin
    ev := @events[eventtail];
    if M_Responder(ev) then
       // menu ate the event
    else
      G_Responder(ev);
    if I_GameFinished then
    begin
      eventtail := eventhead;
      exit;
    end;
    inc(eventtail);
    eventtail := eventtail and (MAXEVENTS - 1);
  end;
end;

//
// D_Display
//  draw current display, possibly wiping it from the previous
//

var
  viewactivestate: boolean = false;
  menuactivestate: boolean = false;
  viewfullscreen: boolean = false;
  inhelpscreensstate: boolean = false;
  oldgamestate: integer = -1;
  borderdrawcount: integer;

procedure D_Display;
var
  nowtime: integer;
  tics: integer;
  wipestart: integer;
  y: integer;
  done: boolean;
  wipe: boolean;
  redrawsbar: boolean;
begin
  if nodrawers then exit; // for comparative timing / profiling

  redrawsbar := false;

  // change the view size if needed
  if setsizeneeded then
  begin
    R_ExecuteSetViewSize;
    oldgamestate := -1; // force background redraw
    borderdrawcount := 3;
  end;

  // save the current screen if about to wipe
  if Ord(gamestate) <> wipegamestate then
  begin
    wipe := true;
    wipe_StartScreen(0, 0, SCREENWIDTH, SCREENHEIGHT);
  end
  else
    wipe := false;

  if (gamestate = GS_LEVEL) and (gametic <> 0) then
    HU_Erase;

  // do buffered drawing
  case gamestate of
    GS_LEVEL:
      begin
        if gametic <> 0 then
        begin
          if automapactive then
            AM_Drawer;
          if wipe or ((viewheight <> SCREENHEIGHT) and viewfullscreen) then
            redrawsbar := true;
          if inhelpscreensstate and (not inhelpscreens) then
            redrawsbar := true; // just put away the help screen
          ST_Drawer(viewheight = SCREENHEIGHT{200}, redrawsbar);
          viewfullscreen := viewheight = SCREENHEIGHT{200};
        end;
      end;
    GS_INTERMISSION:
      WI_Drawer;
    GS_FINALE:
      F_Drawer;
    GS_DEMOSCREEN:
      D_PageDrawer;
  end;

  // draw buffered stuff to screen
  I_UpdateNoBlit;

  // draw the view directly
  if (gamestate = GS_LEVEL) and (not automapactive) and (gametic <> 0) then
    R_RenderPlayerView(@players[displayplayer]);

  if (gamestate = GS_LEVEL) and (gametic <> 0) then
    HU_Drawer;

  // clean up border stuff
  if (Ord(gamestate) <> oldgamestate) and (gamestate <> GS_LEVEL) then
    I_SetPalette(W_CacheLumpName('PLAYPAL', PU_CACHE));

  // see if the border needs to be initially drawn
  if (gamestate = GS_LEVEL) and (oldgamestate <> Ord(GS_LEVEL)) then
  begin
    viewactivestate := false; // view was not active
    R_FillBackScreen;         // draw the pattern into the back screen
  end;

  // see if the border needs to be updated to the screen
  if (gamestate = GS_LEVEL) and (not automapactive) and (scaledviewwidth <> SCREENWIDTH{320}) then
  begin
    if menuactive or menuactivestate or (not viewactivestate) then
      borderdrawcount := 3;
    if borderdrawcount > 0 then
    begin
      R_DrawViewBorder; // erase old menu stuff
      dec(borderdrawcount);
    end;
  end;

  menuactivestate := menuactive;
  viewactivestate := viewactive;
  inhelpscreensstate := inhelpscreens;
  oldgamestate := Ord(gamestate);
  wipegamestate := Ord(gamestate);

  // draw pause pic
  if paused then
  begin
    if automapactive then
      y := 4
    else
      y := viewwindowy + 4;
    V_DrawPatch(viewwindowx + (scaledviewwidth - 68) div 2,
       y, 0, W_CacheLumpName('M_PAUSE', PU_CACHE), false);
  end;

  // menus go directly to the screen
  M_Drawer;  // menu is drawn even on top of everything
  NetUpdate; // send out any new accumulation

  // normal update
  if not wipe then
  begin
    I_FinishUpdate; // page flip or blit buffer
    exit;
  end;

  // wipe update
  wipe_EndScreen(0, 0, SCREENWIDTH, SCREENHEIGHT);

  wipestart := I_GetTime - 1;

  repeat
    repeat
      nowtime := I_GetTime;
      tics := nowtime - wipestart;
    until (tics <> 0);
    wipestart := nowtime;
    done := wipe_ScreenWipe(Ord(wipe_Melt), 0, 0, SCREENWIDTH, SCREENHEIGHT, tics);
    I_UpdateNoBlit;
    M_Drawer;         // menu is drawn even on top of wipes
    I_FinishUpdate;   // page flip or blit buffer
  until done;
end;

//
//  D_DoomLoop
//

procedure D_DoomLoop;
begin
  if demorecording then
    G_BeginRecording;

  I_InitGraphics;

  while true do
  begin
    // frame syncronous IO operations
    I_StartFrame;

    // process one or more tics
    if singletics then
    begin
      I_StartTic;
      D_ProcessEvents;
      G_BuildTiccmd(@netcmds[consoleplayer][maketic mod BACKUPTICS]);
      if advancedemo then
        D_DoAdvanceDemo;
      M_Ticker;
      G_Ticker;
      inc(gametic);
      inc(maketic);
    end
    else
      TryRunTics; // will run at least one tic

    if I_GameFinished then
      break;
    S_UpdateSounds(players[consoleplayer].mo);// move positional sounds

    // Update display, next frame, with current state.
    D_Display;
  end;
end;

//
//  DEMO LOOP
//
var
  demosequence: integer;
  pagetic: integer;
  pagename: string;

//
// D_PageTicker
// Handles timing for warped projection
//
procedure D_PageTicker;
begin
  dec(pagetic);
  if pagetic < 0 then
    D_AdvanceDemo;
end;

//
// D_PageDrawer
//
procedure D_PageDrawer;
begin
  V_DrawPatch (0, 0, 0, W_CacheLumpName(pagename, PU_CACHE), true);
end;

//
// D_AdvanceDemo
// Called after each demo or intro demosequence finishes
//
procedure D_AdvanceDemo;
begin
  advancedemo := true;
end;

//
// This cycles through the demo sequences.
// FIXME - version dependend demo numbers?
//
procedure D_DoAdvanceDemo;
begin
  players[consoleplayer].playerstate := PST_LIVE;  // not reborn
  advancedemo := false;
  usergame := false;               // no save / end game here
  paused := false;
  gameaction := ga_nothing;

  if gamemode = retail then
    demosequence := (demosequence + 1) mod 7
  else
    demosequence := (demosequence + 1) mod 6;

  case demosequence of
    0:
      begin
        if gamemode = commercial then
          pagetic := 35 * 11
        else
          pagetic := 170;
        gamestate := GS_DEMOSCREEN;
        pagename := 'TITLEPIC';
        if gamemode = commercial then
          S_StartMusic(Ord(mus_dm2ttl))
        else
          S_StartMusic(Ord(mus_intro));
      end;
    1:
      begin
        G_DeferedPlayDemo('demo1');
      end;
    2:
      begin
        pagetic := 200;
        gamestate := GS_DEMOSCREEN;
        pagename := 'CREDIT';
      end;
    3:
      begin
        G_DeferedPlayDemo('demo2');
      end;
    4:
      begin
        gamestate := GS_DEMOSCREEN;
        if gamemode = commercial then
        begin
          pagetic := 35 * 11;
          pagename := 'TITLEPIC';
          S_StartMusic(Ord(mus_dm2ttl));
        end
        else
        begin
          pagetic := 200;
          if gamemode = retail then
            pagename := 'CREDIT'
          else
            pagename := 'HELP2';
        end;
      end;
    5:
      begin
        G_DeferedPlayDemo('demo3');
      end;
        // THE DEFINITIVE DOOM Special Edition demo
    6:
      begin
        G_DeferedPlayDemo('demo4');
      end;
  end;
end;

//
// D_StartTitle
//
procedure D_StartTitle;
begin
  gameaction := ga_nothing;
  demosequence := -1;
  D_AdvanceDemo;
end;

//      print title for every printed line
var title: string;

//
// D_AddFile
//
procedure D_AddFile(_file: string);
var
  i: integer;
begin
  for i := 0 to MAXWADFILES - 1 do
    if wadfiles[i] = '' then
    begin
      wadfiles[i] := _file;
      numwadfiles := i + 1;
      exit;
    end;
end;

//
// IdentifyVersion
// Checks availability of IWAD files by name,
// to determine whether registered/commercial features
// should be executed (notably loading PWAD's).
//
procedure IdentifyVersion;
var
  doom1wad: string;
  doomwad: string;
  doomuwad: string;
  doom2wad: string;

  doom2fwad: string;
  plutoniawad: string;
  tntwad: string;
  doomwaddir: string;
begin
  doomwaddir := getenv('DOOMWADDIR');
  if doomwaddir = '' then
    doomwaddir := '.';

  // Commercial.
  sprintf(doom2wad, '%s\doom2.wad', [doomwaddir]);

  // Retail.
  sprintf(doomuwad, '%s\doomu.wad', [doomwaddir]);

  // Registered.
  sprintf(doomwad, '%s\doom.wad', [doomwaddir]);

  // Shareware.
  sprintf(doom1wad, '%s\doom1.wad', [doomwaddir]);

  // plutonia pack
  sprintf(plutoniawad, '%s\plutonia.wad', [doomwaddir]);

  // tnt pack
  sprintf(tntwad, '%s\tnt.wad', [doomwaddir]);


  // French stuff.
  sprintf(doom2fwad, '%s\doom2f.wad', [doomwaddir]);

  if M_CheckParm('-shdev') > 0 then
  begin
    gamemode := shareware;
    devparm := true;
    D_AddFile(DEVDATA + 'doom1.wad');
    D_AddFile(DEVMAPS + 'data_se/texture1.lmp');
    D_AddFile(DEVMAPS + 'data_se/pnames.lmp');
    basedefault := DEVDATA + 'default.cfg';
    exit;
  end;

  if M_CheckParm('-regdev') > 0 then
  begin
    gamemode := registered;
    devparm := true;
    D_AddFile(DEVDATA + 'doom.wad');
    D_AddFile(DEVMAPS + 'data_se/texture1.lmp');
    D_AddFile(DEVMAPS + 'data_se/texture2.lmp');
    D_AddFile(DEVMAPS + 'data_se/pnames.lmp');
    basedefault := DEVDATA + 'default.cfg';
    exit;
  end;

  if M_CheckParm('-comdev') > 0 then
  begin
    gamemode := commercial;
    devparm := true;
  (* I don't bother
  if(plutonia)
      D_AddFile (DEVDATA"plutonia.wad");
  else if(tnt)
      D_AddFile (DEVDATA"tnt.wad");
  else*)
    D_AddFile(DEVDATA + 'doom2.wad');

    D_AddFile(DEVMAPS + 'cdata/texture1.lmp');
    D_AddFile(DEVMAPS + 'cdata/pnames.lmp');
    basedefault := DEVDATA + 'default.cfg';
    exit;
  end;

  basedefault := 'default.cfg';

  if fexists(doom2fwad) then
  begin
    gamemode := commercial;
    // C'est ridicule!
    // Let's handle languages in config files, okay?
    language := french;
    printf('French version' + #13#10);
    D_AddFile(doom2fwad);
    exit;
  end;

  if fexists(doom2wad) then
  begin
    gamemode := commercial;
    D_AddFile(doom2wad);
    exit;
  end;

  if fexists(plutoniawad) then
  begin
    gamemode := commercial;
    D_AddFile(plutoniawad);
    exit;
  end;

  if fexists(tntwad) then
  begin
    gamemode := commercial;
    D_AddFile(tntwad);
    exit;
  end;

  if fexists(doomuwad) then
  begin
    gamemode := retail;
    D_AddFile(doomuwad);
    exit;
  end;

  if fexists(doomwad) then
  begin
    gamemode := registered;
    D_AddFile(doomwad);
    exit;
  end;

  if fexists(doom1wad) then
  begin
    gamemode := shareware;
    D_AddFile(doom1wad);
    exit;
  end;

  printf('Game mode indeterminate.' + #13#10);
  gamemode := indetermined;

  // We don't abort. Let's see what the PWAD contains.
  //exit(1);
  //I_Error ("Game mode indeterminate\n");
end;

//
// Find a Response File
//
// VJ: Changed to handle more that 1 response files
procedure FindResponseFile;
var
  i: integer;
  handle: file;
  size: integer;
  index: integer;
  myargv1: string;
  infile: string;
  _file: string;
  s: TStringList;
begin
  s := TStringList.Create;
  try
    s.Add(myargv[0]);

    for i := 1 to myargc - 1 do
    begin
      if myargv[i][1] = '@' then
      begin
        // READ THE RESPONSE FILE INTO MEMORY
        myargv1 := Copy(myargv[i], 2, length(myargv[i]) - 1);
        {$I-}
        assign(handle, myargv1);
        reset(handle, 1);
        {$I+}
        if IOResult <> 0 then
        begin
          printf(#13#10 + 'No such response file: %s!' + #13#10, [myargv1]);
          halt(1);
        end;
        printf('Found response file %s!' + #13#10, [myargv1]);

        size := FileSize(handle);
        seek(handle, 0);
        SetLength(_file, size);
        BlockRead(handle, (@_file[1])^, size);
        closefile(handle);

        infile := '';
        for index := 1 to Length(_file) do
          if  _file[index] = ' ' then
            infile := infile + #13#10
          else
            infile := infile + _file[i];

        s.Text := s.Text + infile;
      end
      else
        s.Add(myargv[i])
    end;

    index := 0;
    for i := 0 to s.Count - 1 do
      if s.Strings[i] <> '' then
      begin
        myargv[index] := s.Strings[i];
        inc(index);
      end;
    myargc := index;
  finally
    s.Free;
  end;
end;

//
// D_DoomMain
//
procedure D_DoomMain;
var
  p: integer;
  _file: string;
  scale: integer;
  _time: integer;
  s_error: string;
  i: integer;
  j: integer;
  oldoutproc: TOutProc;
  mb_min: integer; // minimum zone size
begin
  outproc := @I_IOprintf;

  printf('M_InitArgv: Initializing command line parameters.' + #13#10);
  M_InitArgv;

  FindResponseFile;

  IdentifyVersion;

  printf('I_InitializeIO: Initializing input/output streams.' + #13#10);
  I_InitializeIO;

  modifiedgame := false;

  nomonsters := M_CheckParm('-nomonsters') > 0;
  respawnparm := M_CheckParm('-respawn') > 0;
  fastparm := M_CheckParm('-fast') > 0;
  devparm := M_CheckParm('-devparm') > 0;

  if M_CheckParm('-altdeath') > 0 then
    deathmatch := 2
  else if M_CheckParm('-deathmatch') > 0 then
    deathmatch := 1;

  case gamemode of
    retail:
      begin
        sprintf(title,
           '                         ' +
           'The Ultimate DOOM Startup v%d.%d' +
           '                           ',
            [VERSION div 100, VERSION mod 100]);
      end;
    shareware:
      begin
        sprintf(title,
           '                            ' +
           'DOOM Shareware Startup v%d.%d' +
           '                           ',
            [VERSION div 100, VERSION mod 100]);
      end;
    registered:
      begin
        sprintf(title,
           '                            ' +
           'DOOM Registered Startup v%d.%d' +
           '                           ',
            [VERSION div 100, VERSION mod 100]);
      end;
    commercial:
      begin
        sprintf(title,
           '                         ' +
           'DOOM 2: Hell on Earth v%d.%d' +
           '                           ',
            [VERSION div 100, VERSION mod 100]);
      end;
  else
    begin
      sprintf(title,
         '                         ' +
         'Public DOOM - v%d.%d' +
         '                           ',
          [VERSION div 100, VERSION mod 100]);
    end;
  end;

  printf('%s' + #13#10, [title]);

  if devparm then
    printf(D_DEVSTR);

  if M_CheckParmCDROM then
  begin
    printf(D_CDROM);
    mkdir('c:\doomdata');
    basedefault := 'c:\doomdata\default.cfg';
  end;

  // turbo option
  p := M_CheckParm('-turbo');
  if p <> 0 then
  begin
    scale := 200;
    if p < myargc - 1 then
      scale := atoi(myargv[p + 1]);
    if scale < 10 then
      scale := 10
    else if scale > 400 then
      scale := 400;
    printf('turbo scale: %d%' + #13#10, [scale]);
    forwardmove[0] := forwardmove[0] * scale div 100;
    forwardmove[1] := forwardmove[1] * scale div 100;
    sidemove[0] := sidemove[0] * scale div 100;
    sidemove[1] := sidemove[1] * scale div 100;
  end;

  // add any files specified on the command line with -file wadfile
  // to the wad list
  //
  // convenience hack to allow -wart e m to add a wad file
  // prepend a tilde to the filename so wadfile will be reloadable
  p := M_CheckParm('-wart');
  if p <> 0 then
  begin
    myargv[p][5] := 'p';     // big hack, change to -warp

  // Map name handling.
    case gamemode of
      shareware,
      retail,
      registered:
        begin
          sprintf(_file, '~' + DEVMAPS + 'E%sM%s.wad',
            [myargv[p + 1][1], myargv[p + 2][1]]);
          printf('Warping to Episode %s, Map %s.' + #13#10,
            [myargv[p + 1], myargv[p + 2]]);
        end;
    else
      begin
        p := atoi(myargv[p + 1]);
        if p < 10 then
          sprintf(_file, '~' + DEVMAPS + 'cdata/map0%i.wad', [p])
        else
          sprintf (_file,'~' + DEVMAPS + 'cdata/map%i.wad', [p]);
      end;
    end;

    D_AddFile(_file);
  end;

  p := M_CheckParm('-file');
  if p <> 0 then
  begin
  // the parms after p are wadfile/lump names,
  // until end of parms or another - preceded parm
    modifiedgame := true;            // homebrew levels
    inc(p);
    while (p <> myargc) and (myargv[p][1] <> '-') do
    begin
      D_AddFile(myargv[p]);
      inc(p);
    end;
  end;

  p := M_CheckParm('-playdemo');

  if p = 0 then
    p := M_CheckParm('-timedemo');

  if (p <> 0) and (p < myargc - 1) then
  begin
    sprintf(_file,'%s.lmp', [myargv[p + 1]]);
    D_AddFile(_file);
    printf('Playing demo %s.lmp.' + #13#10, [myargv[p + 1]]);
  end;

  // get skill / episode / map from parms
  startskill := sk_medium;
  startepisode := 1;
  startmap := 1;
  autostart := false;

  p := M_CheckParm('-skill');
  if (p <> 0) and (p < myargc - 1) then
  begin
    startskill := skill_t(ord(myargv[p + 1][1]) - ord('1'));
    autostart := true;
  end;

  p := M_CheckParm ('-episode');
  if (p <> 0) and (p < myargc - 1) then
  begin
    startepisode := ord(myargv[p + 1][1]) - ord('0');
    startmap := 1;
    autostart := true;
  end;

  p := M_CheckParm ('-timer');
  if (p <> 0) and (p < myargc - 1) and (deathmatch <> 0) then
  begin
    _time := atoi(myargv[p + 1]);
    if _time > 1 then
      printf('Levels will end after %d minutes' + #13#10, [_time])
    else
      printf('Levels will end after %d minute' + #13#10, [_time]);
  end;

  p := M_CheckParm('-avg');
  if (p <> 0) and (p <= myargc - 1) and boolval(deathmatch) then
    printf('Austin Virtual Gaming: Levels will end after 20 minutes' + #13#10);

  p := M_CheckParm('-warp');
  if (p <> 0) and (p < myargc - 1) then
  begin
    if gamemode = commercial then
      startmap := atoi(myargv[p + 1])
    else
    begin
      startepisode := ord(myargv[p + 1][1]) - ord('0');
      startmap := ord(myargv[p + 2][1]) - ord('0');
    end;
    autostart := true;
  end;

  printf('M_LoadDefaults: Load system defaults.' + #13#10);
  M_LoadDefaults;              // load before initing other systems

  p := M_CheckParm('-fullscreen');
  if (p <> 0) and (p <= myargc - 1) then
    fullscreen := true;

  p := M_CheckParm('-nofullscreen');
  if (p <> 0) and (p <= myargc - 1) then
    fullscreen := false;

  p := M_CheckParm('-screenwidth');
  if (p <> 0) and (p < myargc - 1) then
    SCREENWIDTH := atoi(myargv[p + 1]);
  if SCREENWIDTH > MAXWIDTH then
    SCREENWIDTH := MAXWIDTH;

  p := M_CheckParm('-screenheight');
  if (p <> 0) and (p < myargc - 1) then
    SCREENHEIGHT := atoi(myargv[p + 1]);
  if SCREENHEIGHT > MAXHEIGHT then
    SCREENHEIGHT := MAXHEIGHT;

// Try to guess minimum zone memory to allocate    
  mb_min := 6 + (3 * SCREENWIDTH * SCREENHEIGHT) div (1024 * 1024);
  mb_used := mb_min;

  p := M_CheckParm('-zone');
  if (p <> 0) and (p < myargc - 1) then
  begin
    mb_used := atoi(myargv[p + 1]);
    if mb_used < mb_min then
    begin
      printf('Zone memory allocation needs at least %dMB (%d).' + #13#10, [mb_min, mb_used]);
      mb_used := mb_min;
    end;
  end;

  // init subsystems
  printf('Z_Init: Init zone memory allocation daemon, allocation %dMB.' + #13#10, [mb_used]);
  Z_Init;

  printf('I_InitInfo: Initialize information tables.' + #13#10);
  I_InitInfo;

  printf('V_Init: allocate screens.' + #13#10);
  V_Init;

  printf('M_InitMenus: Initializing menus.' + #13#10);
  M_InitMenus;


  printf('W_Init: Init WADfiles.' + #13#10);
  W_InitMultipleFiles(@wadfiles);

  if gamemode = registered then
    if W_CheckNumForName('e4m1') >= 0 then
      gamemode := retail;

  // Check for -file in shareware
  if modifiedgame then
  begin
    if gamemode = shareware then
      I_Error(#13#10 + 'D_DoomMain(): You cannot -file with the shareware version. Register!');
  // Check for fake IWAD with right name,
  // but w/o all the lumps of the registered version.
    if gamemode in [registered, retail] then
    begin
  // These are the lumps that will be checked in IWAD,
  // if any one is not present, execution will be aborted.
      s_error := #13#10 + 'D_DoomMain(): This is not the registered version.';
      for i := 2 to 3 do
        for j := 1 to 9 do
          if W_CheckNumForName('e' + itoa(i) + 'm' + itoa(j)) < 0 then
            I_Error(s_error);
      if W_CheckNumForName('dphoof') < 0 then
        I_Error(s_error);
      if W_CheckNumForName('bfgga0') < 0 then
        I_Error(s_error);
      if W_CheckNumForName('heada1') < 0 then
        I_Error(s_error);
      if W_CheckNumForName('cybra1') < 0 then
        I_Error(s_error);
      if W_CheckNumForName('spida1d1') < 0 then
        I_Error(s_error);
    end;

  // If additonal PWAD files are used, print modified banner
    oldoutproc := outproc;
    outproc := @I_IOMessageBox;
    printf(MSG_MODIFIEDGAME);
    outproc := oldoutproc;
    printf(MSG_MODIFIEDGAME); // Print the message again to console
  end;

  case gamemode of
    shareware,
    indetermined:
      printf(MSG_SHAREWARE);
    registered,
    retail,
    commercial:
      printf(MSG_COMMERCIAL);
  else
    begin
      printf(MSG_UNDETERMINED);
    end;
  end;

  printf('M_Init: Init miscellaneous info.' + #13#10);
  M_Init;

  printf('R_Init: Init DOOM refresh daemon - ');
  R_Init;

  printf(#13#10 + 'P_Init: Init Playloop state.' + #13#10);
  P_Init;

  printf('I_Init: Setting up machine state.' + #13#10);
  I_Init;

  printf('D_CheckNetGame: Checking network game status.' + #13#10);
  D_CheckNetGame;

  printf('S_Init: Setting up sound.' + #13#10);
  S_Init(snd_SfxVolume, snd_MusicVolume);

  printf('HU_Init: Setting up heads up display.' + #13#10);
  HU_Init;

  printf('ST_Init: Init status bar.' + #13#10);
  ST_Init;

////////////////////////////////////////////////////////////////////////////////
//    // check for a driver that wants intermission stats
//    p = M_CheckParm ("-statcopy");
//    if (p && p<myargc-1)
//    {
//  // for statistics driver
//  extern  void*  statcopy;

//  statcopy = (void*)atoi(myargv[p+1]);
//  printf ("External statistics registered.\n");
//    }
////////////////////////////////////////////////////////////////////////////////

  // start the apropriate game based on parms
  p := M_CheckParm('-record');

  if (p <> 0) and (p < myargc - 1) then
  begin
    G_RecordDemo(myargv[p + 1]);
    autostart := true;
  end;

  p := M_CheckParm('-playdemo');
  if (p <> 0) and (p < myargc - 1) then
  begin
    singledemo := true;              // quit after one demo
    G_DeferedPlayDemo(myargv[p + 1]);
    D_DoomLoop;  // never returns
  end;

  p := M_CheckParm('-timedemo');
  if (p <> 0) and (p < myargc - 1) then
  begin
    G_TimeDemo(myargv[p + 1]);
    D_DoomLoop;  // never returns
  end;

  p := M_CheckParm('-loadgame');
  if (p <> 0) and (p <= myargc - 1) then
  begin
    if M_CheckParmCDROM then
    begin
      sprintf(_file, 'c:\doomdata\' + SAVEGAMENAME + '%s.dsg', [myargv[p + 1][1]]);
      G_LoadGame(_file);
    end
    else if p <> myargc - 1 then
    begin
      sprintf(_file, SAVEGAMENAME + '%s.dsg', [myargv[p + 1][1]]);
      G_LoadGame(_file);
    end;
  end;

  if gameaction <> ga_loadgame then
  begin
    if autostart or netgame then
      G_InitNew(startskill, startepisode, startmap)
    else
      D_StartTitle;                // start up intro loop
  end;

  D_DoomLoop;  // never returns
end;

var
  i: integer;


(* Unit g_game implementation *)


                                                                                                                                                                                                                                                                                                                                                                                                                                                                          

const
  SAVEGAMESIZE = $2C000;
  SAVESTRINGSIZE = 24;

procedure G_ReadDemoTiccmd(cmd: Pticcmd_t); forward;
procedure G_WriteDemoTiccmd (cmd: Pticcmd_t); forward;

procedure G_DoReborn(playernum: integer); forward;

procedure G_DoLoadLevel; forward;
procedure G_DoNewGame; forward;
procedure G_DoPlayDemo; forward;
procedure G_DoCompleted; forward;
procedure G_DoWorldDone; forward;
procedure G_DoSaveGame; forward;

var
  sendsave: boolean;         // send a save event next tic

  timingdemo: boolean;       // if true, exit with report on completion
  noblit: boolean;           // for comparative timing purposes
  starttime: integer;        // for comparative timing purposes

  levelstarttic: integer;          // gametic at level start

  demoname: string;
  netdemo: boolean;
  demobuffer: PByteArray;
  demo_p: PByteArray;
  demoend: PByteArray;

  consistancy: array[0..MAXPLAYERS - 1] of array[0..BACKUPTICS - 1] of smallint;

  savebuffer: PByteArray;

const
  TURBOTHRESHOLD = $32;

function MAXPLMOVE: fixed_t;
begin
  result := forwardmove[1];
end;

const
  SLOWTURNTICS = 6;
  NUMKEYS = 256;

var
  gamekeydown: array[0..NUMKEYS - 1] of boolean;
  turnheld: integer;

  lookheld: integer;

  mousearray: array[0..3] of boolean;
  mousebuttons: PBooleanArray;

// mouse values are used once 
  g_game_mousex: integer;
  g_game_mousey: integer;

  dclicktime: integer;
  dclickstate: integer;
  dclicks: integer;
  dclicktime2: integer;
  dclickstate2: integer;
  dclicks2: integer;

// joystick values are repeated
  joyxmove: integer;
  joyymove: integer;
  joyarray: array[0..4] of boolean;
  joybuttons: PBooleanArray;

  savegameslot: integer;
  savedescription: string;

const
  BODYQUESIZE  = 32;

var
  bodyque: array[0..BODYQUESIZE - 1] of Pmobj_t;

  statcopy: pointer;        // for statistics driver

function G_CmdChecksum(cmd: Pticcmd_t): integer;
var
  i: integer;
begin
  result := 0;
  for i := 0 to sizeof(cmd^) div 4 - 2 do
    result := result + PIntegerArray(cmd)[i];
end;

//
// G_BuildTiccmd
// Builds a ticcmd from all of the available inputs
// or reads it from the demo buffer.
// If recording a demo, write it out
//
procedure G_BuildTiccmd(cmd: Pticcmd_t);
var
  i: integer;
  strafe: boolean;
  bstrafe: boolean;
  speed: integer;
  tspeed: integer;
  lspeed: integer;
  _forward: integer;
  side: integer;
  look: integer;
  base: Pticcmd_t;
begin
  base := I_BaseTiccmd;    // empty, or external driver

  memcpy(cmd, base, SizeOf(cmd^));

  cmd.consistancy := consistancy[consoleplayer][maketic mod BACKUPTICS];

  strafe := gamekeydown[key_strafe] or
            mousebuttons[mousebstrafe] or
            joybuttons[joybstrafe];
  speed := intval(gamekeydown[key_speed] or joybuttons[joybspeed]);

  _forward := 0;
  side := 0;
  look := 0;

  // use two stage accelerative turning
  // on the keyboard and joystick
  if (joyxmove < 0) or
     (joyxmove > 0) or
     (gamekeydown[key_right]) or
     (gamekeydown[key_left]) then
    turnheld := turnheld + ticdup
  else
    turnheld := 0;

  if turnheld < SLOWTURNTICS then
    tspeed := 2             // slow turn
  else
    tspeed := speed;

  if gamekeydown[key_lookdown] or gamekeydown[key_lookup] then
    lookheld := lookheld + ticdup
  else
    lookheld := 0;

  if lookheld < SLOWTURNTICS then
    lspeed := 1
  else
    lspeed := 2;

  // let movement keys cancel each other out
  if strafe then
  begin
    if gamekeydown[key_right] then
      side := side + sidemove[speed]; // fprintf(stderr, "strafe right\n");
    if gamekeydown[key_left] then
      side := side - sidemove[speed]; // fprintf(stderr, "strafe left\n");
    if joyxmove > 0 then
      side := side + sidemove[speed];
    if joyxmove < 0 then
      side := side - sidemove[speed];
  end
  else
  begin
    if gamekeydown[key_right] then
      cmd.angleturn := cmd.angleturn - angleturn[tspeed];
    if gamekeydown[key_left] then
      cmd.angleturn := cmd.angleturn + angleturn[tspeed];
    if joyxmove > 0 then
      cmd.angleturn := cmd.angleturn - angleturn[tspeed];
    if joyxmove < 0 then
      cmd.angleturn := cmd.angleturn + angleturn[tspeed];
  end;

  if gamekeydown[key_up] then
    _forward := _forward + forwardmove[speed]; // fprintf(stderr, "up\n");

  if gamekeydown[key_down] then
    _forward := _forward - forwardmove[speed]; // fprintf(stderr, "down\n");

  // Look up/down/center keys
  if gamekeydown[key_lookup] then
    look := lspeed;

  if gamekeydown[key_lookdown] then
    look := -lspeed;

  if gamekeydown[key_lookcenter] then
    look := TOCENTER;

  if joyymove < 0 then
    _forward := _forward + forwardmove[speed];

  if joyymove > 0 then
    _forward := _forward - forwardmove[speed];

  if gamekeydown[key_straferight] then
    side := side + sidemove[speed];

  if gamekeydown[key_strafeleft] then
    side := side - sidemove[speed];

  // buttons
  cmd.chatchar := Ord(HU_dequeueChatChar);

  if gamekeydown[key_fire] or
     (usemouse and mousebuttons[mousebfire]) or
     (usejoystick and joybuttons[joybfire]) then
    cmd.buttons := cmd.buttons or BT_ATTACK;

  if gamekeydown[key_use] or (usejoystick and joybuttons[joybuse]) then
  begin
    cmd.buttons := cmd.buttons or BT_USE;
  // clear double clicks if hit use button
    dclicks := 0;
  end;

  // chainsaw overrides
  for i := 0 to Ord(NUMWEAPONS) - 2 do
    if gamekeydown[ord('1') + i] then
    begin
      cmd.buttons := cmd.buttons or BT_CHANGE;
      cmd.buttons := cmd.buttons or _SHL(i, BT_WEAPONSHIFT);
      break;
    end;

  // mouse
  if (usemouse and mousebuttons[mousebforward]) then
    _forward := _forward + forwardmove[speed];

  // forward double click
  if usemouse and (mousebuttons[mousebforward] <> boolval(dclickstate)) and (dclicktime > 1) then
  begin
    dclickstate := intval(mousebuttons[mousebforward]);
    if dclickstate <> 0 then
      inc(dclicks);
    if dclicks = 2 then
    begin
      cmd.buttons := cmd.buttons or BT_USE;
      dclicks := 0;
    end
    else
      dclicktime := 0;
  end
  else
  begin
    dclicktime := dclicktime + ticdup;
    if dclicktime > 20 then
    begin
      dclicks := 0;
      dclickstate := 0;
    end
  end;

  // strafe double click
  bstrafe := (usemouse and mousebuttons[mousebstrafe]) or
             (usejoystick and joybuttons[joybstrafe]);
  if (bstrafe <> boolval(dclickstate2)) and (dclicktime2 > 1) then
  begin
    dclickstate2 := intval(bstrafe);
    if bstrafe then
      inc(dclicks2);
    if dclicks2 = 2 then
    begin
      cmd.buttons := cmd.buttons or BT_USE;
      dclicks2 := 0;
    end
    else
      dclicktime2 := 0;
  end
  else
  begin
    dclicktime2 := dclicktime2 + ticdup;
    if dclicktime2 > 20 then
    begin
      dclicks2 := 0;
      dclickstate2 := 0;
    end;
  end;

  _forward := _forward + g_game_mousey;
  if strafe then
    side := side + g_game_mousex * 2
  else
    cmd.angleturn := cmd.angleturn - g_game_mousex * $8;

  g_game_mousex := 0;
  g_game_mousey := 0;

  if _forward > MAXPLMOVE then
    _forward := MAXPLMOVE
  else if _forward < -MAXPLMOVE then
    _forward := -MAXPLMOVE;

  if side > MAXPLMOVE then
    side := MAXPLMOVE
  else if side < -MAXPLMOVE then
    side := -MAXPLMOVE;

  cmd.forwardmove := cmd.forwardmove + _forward;
  cmd.sidemove := cmd.sidemove + side;
  if players[consoleplayer].playerstate = PST_LIVE then
  begin
    if look < 0 then
      look := look + 16;
    cmd.look := look;
  end;

  // special buttons
  if sendpause then
  begin
    sendpause := false;
    cmd.buttons := BT_SPECIAL or BTS_PAUSE;
  end;

  if sendsave then
  begin
    sendsave := false;
    cmd.buttons := BT_SPECIAL or BTS_SAVEGAME or _SHL(savegameslot, BTS_SAVESHIFT);
  end;
end;

//
// G_DoLoadLevel 
//
procedure G_DoLoadLevel;
var
  i: integer;
begin
  // Set the sky map.
  // First thing, we have a dummy sky texture name,
  //  a flat. The data is in the WAD only because
  //  we look for an actual index, instead of simply
  //  setting one.
  skyflatnum := R_FlatNumForName(SKYFLATNAME);

  // DOOM determines the sky texture to be used
  // depending on the current episode, and the game version.
  if (gamemode = commercial) or
     (Ord(gamemode) = Ord(pack_tnt)) or
     (Ord(gamemode) = Ord(pack_plut)) then
  begin
    skytexture := R_TextureNumForName('SKY3');
    if gamemap < 12 then
      skytexture := R_TextureNumForName('SKY1')
    else if gamemap < 21 then
      skytexture := R_TextureNumForName('SKY2');
  end;

  levelstarttic := gametic;        // for time calculation

  if wipegamestate = Ord(GS_LEVEL) then
    wipegamestate := -1;             // force a wipe

  gamestate := GS_LEVEL;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] and (players[i].playerstate = PST_DEAD) then
      players[i].playerstate := PST_REBORN;
    memset(@players[i].frags, 0, SizeOf(players[i].frags));
  end;

  P_SetupLevel(gameepisode, gamemap, 0, gameskill);
  displayplayer := consoleplayer;    // view the guy you are playing
  starttime := I_GetTime;
  gameaction := ga_nothing;
  Z_CheckHeap;

  // clear cmd building stuff
  memset(@gamekeydown, 0, SizeOf(gamekeydown));
  joyxmove := 0;
  joyymove := 0;
  g_game_mousex := 0;
  g_game_mousey := 0;
  sendpause := false;
  sendsave := false;
  paused := false;
  memset(mousebuttons, 0, SizeOf(mousebuttons));
  memset(joybuttons, 0, SizeOf(joybuttons));
end;

//
// G_Responder  
// Get info needed to make ticcmd_ts for the players.
//
function G_Responder(ev: Pevent_t): boolean;
begin
  // allow spy mode changes even during the demo
  if (gamestate = GS_LEVEL) and (ev._type = ev_keydown) and
     (ev.data1 = KEY_F12) and (singledemo or (not boolval(deathmatch))) then
  begin
  // spy mode
    repeat
      inc(displayplayer);
      if displayplayer = MAXPLAYERS then
        displayplayer := 0;
    until not ((not playeringame[displayplayer]) and (displayplayer <> consoleplayer));
    result := true;
    exit;
  end;

  // any other key pops up menu if in demos
  if (gameaction = ga_nothing) and (not singledemo) and
     (demoplayback or (gamestate = GS_DEMOSCREEN)) then
  begin
    if (ev._type = ev_keydown) or
       ((ev._type = ev_mouse) and (ev.data1 <> 0)) or
       ((ev._type = ev_joystick) and (ev.data1 <> 0)) then
    begin
      M_StartControlPanel;
      result := true;
      exit;
    end;
    result := false;
    exit;
  end;

  if gamestate = GS_LEVEL then
  begin
    if HU_Responder(ev) then
    begin
      result := true; // chat ate the event
      exit;
    end;
    if ST_Responder(ev) then
    begin
      result := true; // status window ate it
      exit;
    end;
    if AM_Responder(ev) then
    begin
      result := true; // automap ate it
      exit;
    end;
  end;

  if gamestate = GS_FINALE then
  begin
    if F_Responder(ev) then
    begin
      result := true; // finale ate the event
      exit;
    end;
  end;

  case ev._type of
    ev_keydown:
      begin
        if ev.data1 = KEY_PAUSE then
        begin
          sendpause := true;
          result := true;
          exit;
        end;
        if ev.data1 < NUMKEYS then
          gamekeydown[ev.data1] := true;
        result := true; // eat key down events
        exit;
      end;
    ev_keyup:
      begin
        if ev.data1 < NUMKEYS then
          gamekeydown[ev.data1] := false;
        result := false; // always let key up events filter down
        exit;
      end;
    ev_mouse:
      begin
        if usemouse then
        begin
          mousebuttons[0] := (ev.data1 and 1) <> 0;
          mousebuttons[1] := (ev.data1 and 2) <> 0;
          mousebuttons[2] := (ev.data1 and 4) <> 0;
          g_game_mousex := ev.data2 * (mouseSensitivity + 5) div 10;
          g_game_mousey := ev.data3 * (mouseSensitivity + 5) div 10;
        end
        else
        begin
          mousebuttons[0] := false;
          mousebuttons[1] := false;
          mousebuttons[2] := false;
          g_game_mousex := 0;
          g_game_mousey := 0;
        end;
        result := true;    // eat events
        exit;
      end;
    ev_joystick:
      begin
        if usejoystick then
        begin
          joybuttons[0] := (ev.data1 and 1) <> 0;
          joybuttons[1] := (ev.data1 and 2) <> 0;
          joybuttons[2] := (ev.data1 and 4) <> 0;
          joybuttons[3] := (ev.data1 and 8) <> 0;
          joyxmove := ev.data2;
          joyymove := ev.data3;
        end
        else
        begin
          joybuttons[0] := false;
          joybuttons[1] := false;
          joybuttons[2] := false;
          joybuttons[3] := false;
          joyxmove := 0;
          joyymove := 0;
        end;
        result := true;    // eat events
        exit;
      end;
  end;

  result := false;
end;

//
// G_Ticker
// Make ticcmd_ts for the players.
//
procedure G_Ticker;
var
  i: integer;
  buf: integer;
  cmd: Pticcmd_t;
  msg: string;
begin    
  // do player reborns if needed
  for i := 0 to MAXPLAYERS - 1 do
    if playeringame[i] and (players[i].playerstate = PST_REBORN) then
      G_DoReborn(i);

  // do things to change the game state
  while (gameaction <> ga_nothing) do
  begin
    case gameaction of
      ga_loadlevel:
        G_DoLoadLevel;
      ga_newgame:
        G_DoNewGame;
      ga_loadgame:
        G_DoLoadGame;
      ga_savegame:
        G_DoSaveGame;
      ga_playdemo:
        G_DoPlayDemo;
      ga_completed:
        G_DoCompleted;
      ga_victory: 
        F_StartFinale;
      ga_worlddone:
        G_DoWorldDone;
      ga_screenshot:
        begin
          M_ScreenShot;
          gameaction := ga_nothing;
        end;
    end;
  end;

  // get commands, check consistancy,
  // and build new consistancy check
  buf := (gametic div ticdup) mod BACKUPTICS;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] then
    begin
      cmd := @players[i].cmd;

      memcpy (cmd, @netcmds[i][buf], SizeOf(ticcmd_t));

      if demoplayback then
        G_ReadDemoTiccmd(cmd);
      if demorecording then
        G_WriteDemoTiccmd(cmd);

      // check for turbo cheats
      if (cmd.forwardmove > TURBOTHRESHOLD) and
         ((gametic and 31) = 0) and
         (((_SHR(gametic, 5)) and 3) = i) then
      begin
        sprintf(msg, '%s is turbo!', [player_names[i]]);
        players[consoleplayer]._message := msg;
      end;

      if netgame and (not netdemo) and ((gametic mod ticdup) = 0) then
      begin
        if (gametic > BACKUPTICS) and
           (consistancy[i][buf] <> cmd.consistancy) then
          I_Error('G_Ticker(): consistency failure (%d should be %d)',
            [cmd.consistancy, consistancy[i][buf]]);
        if players[i].mo <> nil then
          consistancy[i][buf] := players[i].mo.x
        else
          consistancy[i][buf] := rndindex;
      end;
    end;
  end;

  // check for special buttons
  for i := 0 to MAXPLAYERS - 1 do
  begin
    if playeringame[i] then
    begin
      if (players[i].cmd.buttons and BT_SPECIAL) <> 0 then
      begin
        case players[i].cmd.buttons and BT_SPECIALMASK of
          BTS_PAUSE:
            begin
              paused := not paused; 
              if paused then
                S_PauseSound
              else
                S_ResumeSound;
            end;
          BTS_SAVEGAME:
            begin
              if savedescription = '' then
                savedescription := 'NET GAME';
              savegameslot :=
                _SHR((players[i].cmd.buttons and BTS_SAVEMASK), BTS_SAVESHIFT);
              gameaction := ga_savegame;
            end;
        end;
      end;
    end;
  end;

  // do main actions
  case gamestate of
    GS_LEVEL:
      begin
        P_Ticker;
        ST_Ticker;
        AM_Ticker;
        HU_Ticker;
      end;
    GS_INTERMISSION:
      begin
        WI_Ticker;
      end;
    GS_FINALE:
      begin
        F_Ticker;
      end;
    GS_DEMOSCREEN:
      begin
        D_PageTicker;
      end;
  end;
end;

//
// PLAYER STRUCTURE FUNCTIONS
// also see P_SpawnPlayer in P_Things
//

{
//
// G_InitPlayer
// Called at the start.
// Called by the game initialization functions.
//
procedure G_InitPlayer(player: integer);
var
  p: Pplayer_t;
begin
  // set up the saved info
  p := @players[player];

  // clear everything else to defaults
  G_PlayerReborn(player);
end;
}

//
// G_PlayerFinishLevel
// Can when a player completes a level.
//
procedure G_PlayerFinishLevel(player: integer);
var
  p: Pplayer_t;
begin
  p := @players[player];

  memset(@p.powers, 0, SizeOf(p.powers));
  memset(@p.cards, 0, SizeOf(p.cards));
  p.mo.flags := p.mo.flags and (not MF_SHADOW); // cancel invisibility
  p.lookdir := 0;       // cancel lookdir
  p.centering := false; 
  p.extralight := 0;    // cancel gun flashes
  p.fixedcolormap := 0; // cancel ir gogles
  p.damagecount := 0;   // no palette changes
  p.bonuscount := 0;
end;

//
// G_PlayerReborn
// Called after a player dies 
// almost everything is cleared and initialized 
//
procedure G_PlayerReborn(player: integer);
var
  p: Pplayer_t;
  i: integer;
  frags: array[0..MAXPLAYERS - 1] of integer;
  killcount: integer;
  itemcount: integer;
  secretcount: integer;
begin
  memcpy(@frags, @players[player].frags, SizeOf(frags));
  killcount := players[player].killcount;
  itemcount := players[player].itemcount;
  secretcount := players[player].secretcount;

  p := @players[player];
  memset(p, 0, SizeOf(p^));

  memcpy(@players[player].frags, @frags, SizeOf(players[player].frags));
  players[player].killcount := killcount;
  players[player].itemcount := itemcount;
  players[player].secretcount := secretcount;

  p.usedown := true;
  p.attackdown := true;  // don't do anything immediately
  p.playerstate := PST_LIVE;
  p.health := MAXHEALTH;
  p.readyweapon := wp_pistol;
  p.pendingweapon := wp_pistol;
  p.weaponowned[Ord(wp_fist)] := 1;
  p.weaponowned[Ord(wp_pistol)] := 1;
  p.ammo[Ord(am_clip)] := 50;
  p.lookdir := 0;
  p.centering := false;

  for i := 0 to Ord(NUMAMMO) - 1 do
    p.maxammo[i] := maxammo[i];
end;

//
// G_CheckSpot
// Returns false if the player cannot be respawned
// at the given mapthing_t spot  
// because something is occupying it 
//
function G_CheckSpot(playernum: integer; mthing: Pmapthing_t): boolean;
var
  x: fixed_t;
  y: fixed_t;
  ss: Psubsector_t;
  an: angle_t; // VJ was u long
  mo: Pmobj_t;
  i: integer;
begin
  if not boolval(players[playernum].mo) then
  begin
    // first spawn of level, before corpses
    for i := 0 to playernum - 1 do
      if (players[i].mo.x = (mthing.x * FRACUNIT)) and
         (players[i].mo.y = (mthing.y * FRACUNIT)) then
      begin
        result := false;
        exit;
      end;
    result := true;
    exit;
  end;

  x := mthing.x * FRACUNIT;
  y := mthing.y * FRACUNIT;

  if not P_CheckPosition(players[playernum].mo, x, y) then
  begin
    result := false;
    exit;
  end;

  // flush an old corpse if needed
  if bodyqueslot >= BODYQUESIZE then
    P_RemoveMobj(bodyque[bodyqueslot mod BODYQUESIZE]);
  bodyque[bodyqueslot mod BODYQUESIZE] := players[playernum].mo;
  inc(bodyqueslot);

  // spawn a teleport fog
  ss := R_PointInSubsector(x, y);
  an := _SHR((ANG45 * (mthing.angle div 45)), ANGLETOFINESHIFT);

  mo := P_SpawnMobj(x + 20 * finecosine[an], y + 20 * finesine[an],
          ss.sector.floorheight, MT_TFOG);

  if players[consoleplayer].viewz <> 1 then
    S_StartSound(mo, Ord(sfx_telept));  // don't start sound on first frame

  result := true;
end;

//
// G_DeathMatchSpawnPlayer 
// Spawns a player at one of the random death match spots 
// called at level load and each death 
//
procedure G_DeathMatchSpawnPlayer(playernum: integer);
var
  i, j: integer;
  selections: integer;
begin
  selections := deathmatch_p; // VJ - deathmatchstarts;
  if selections < 4 then
    I_Error('G_DeathMatchSpawnPlayer(): Only %d deathmatch spots, 4 required', [selections]);

  for j := 0 to 19 do
  begin
    i := P_Random mod selections;
    if G_CheckSpot (playernum, @deathmatchstarts[i]) then
    begin
      deathmatchstarts[i]._type := playernum + 1;
      P_SpawnPlayer(@deathmatchstarts[i]);
      exit;
    end;
  end;

  // no good spot, so the player will probably get stuck
  P_SpawnPlayer(@playerstarts[playernum]);
end;

//
// G_DoReborn 
//
procedure G_DoReborn(playernum: integer);
var
  i: integer;
begin
  if not netgame then
    // reload the level from scratch
    gameaction := ga_loadlevel
  else
  begin
    // respawn at the start

    // first dissasociate the corpse
    players[playernum].mo.player := nil;

    // spawn at random spot if in death match
    if deathmatch <> 0 then
    begin
      G_DeathMatchSpawnPlayer(playernum);
      exit;
    end;

    if G_CheckSpot(playernum, @playerstarts[playernum]) then
    begin
      P_SpawnPlayer(@playerstarts[playernum]);
      exit;
    end;

    // try to spawn at one of the other players spots
    for i := 0 to MAXPLAYERS - 1 do
    begin
      if G_CheckSpot (playernum, @playerstarts[i]) then
      begin
        playerstarts[i]._type := playernum + 1; // fake as other player
        P_SpawnPlayer(@playerstarts[i]);
        playerstarts[i]._type := i + 1; // restore 
        exit;
      end;
      // he's going to be inside something.  Too bad.
    end;
    P_SpawnPlayer(@playerstarts[playernum]);
  end;
end; 

procedure G_ScreenShot;
begin
  gameaction := ga_screenshot;
end;

var
// DOOM Par Times
  pars: array[0..3, 0..9] of integer;

// DOOM II Par Times
  cpars: array[0..31] of integer;

//
// G_DoCompleted
//
  secretexit: boolean;

procedure G_ExitLevel;
begin
  secretexit := false;
  gameaction := ga_completed;
end;

// Here's for the german edition.
procedure G_SecretExitLevel;
begin
  // IF NO WOLF3D LEVELS, NO SECRET EXIT!
  if (gamemode = commercial) and (W_CheckNumForName('map31') < 0) then
    secretexit := false
  else
    secretexit := true;
  gameaction := ga_completed;
end;

procedure G_DoCompleted;
var
  i: integer;
begin
  gameaction := ga_nothing;

  for i := 0 to MAXPLAYERS - 1 do
    if playeringame[i] then
      G_PlayerFinishLevel(i); // take away cards and stuff

  if automapactive then
    AM_Stop;

  if gamemode <> commercial then
  begin
    case gamemap of
      8:
        begin
          gameaction := ga_victory;
          exit;
        end;
      9:
        begin
          for i := 0 to MAXPLAYERS - 1 do
            players[i].didsecret := true;
        end;
    end;
  end;


  wminfo.didsecret := players[consoleplayer].didsecret;
  wminfo.epsd := gameepisode - 1;
  wminfo.last := gamemap - 1;

  // wminfo.next is 0 biased, unlike gamemap
  if gamemode = commercial then
  begin
    if secretexit then
    begin
      case gamemap of
        15: wminfo.next := 30;
        31: wminfo.next := 31;
      end
    end
    else
    begin
      case gamemap of
        31,
        32: wminfo.next := 15;
      else
        wminfo.next := gamemap;
      end;
    end
  end
  else
  begin
    if secretexit then
      wminfo.next := 8  // go to secret level
    else if gamemap = 9 then
    begin
      // returning from secret level
      case gameepisode of
        1: wminfo.next := 3;
        2: wminfo.next := 5;
        3: wminfo.next := 6;
        4: wminfo.next := 2;
      end
    end
    else
      wminfo.next := gamemap; // go to next level
  end;

  wminfo.maxkills := totalkills;
  wminfo.maxitems := totalitems;
  wminfo.maxsecret := totalsecret;
  wminfo.maxfrags := 0;
  if gamemode = commercial then
    wminfo.partime := 35 * cpars[gamemap-1]
  else
    wminfo.partime := 35 * pars[gameepisode][gamemap];
  wminfo.pnum := consoleplayer;

  for i := 0 to MAXPLAYERS - 1 do
  begin
    wminfo.plyr[i]._in := playeringame[i];
    wminfo.plyr[i].skills := players[i].killcount;
    wminfo.plyr[i].sitems := players[i].itemcount;
    wminfo.plyr[i].ssecret := players[i].secretcount;
    wminfo.plyr[i].stime := leveltime;
    memcpy(@wminfo.plyr[i].frags, @players[i].frags, SizeOf(wminfo.plyr[i].frags));
  end;

  gamestate := GS_INTERMISSION;
  viewactive := false;
  automapactive := false;

  if statcopy <> nil then
    memcpy(statcopy, @wminfo, SizeOf(wminfo));

  WI_Start(@wminfo);
end;

//
// G_WorldDone 
//
procedure G_WorldDone;
begin
  gameaction := ga_worlddone;

  if secretexit then
    players[consoleplayer].didsecret := true;

  if gamemode = commercial then
  begin
    if secretexit then
    begin
      if gamemap in [15, 31, 6, 11, 20, 30] then
        F_StartFinale
    end
    else if gamemap in [6, 11, 20, 30] then
      F_StartFinale;
  end;
end;

procedure G_DoWorldDone;
begin
  gamestate := GS_LEVEL;
  gamemap := wminfo.next + 1;
  G_DoLoadLevel;
  gameaction := ga_nothing;
  viewactive := true;
end;

//
// G_InitFromSavegame
// Can be called by the startup code or the menu task.
//
var
  savename: string;

procedure G_LoadGame(const name: string);
begin
  savename := name;
  gameaction := ga_loadgame;
end;

const
  VERSIONSIZE = 16;

procedure G_DoLoadGame;
var
  len: integer;
  i: integer;
  a, b, c: integer;
  vcheck: string;
begin
  gameaction := ga_nothing;

  len := M_ReadFile(savename, pointer(savebuffer));
  save_p := PByteArray(integer(savebuffer) + SAVESTRINGSIZE);

  // skip the description field
  vcheck := '';
  sprintf(vcheck, 'version %d', [VERSION]);

  if len < Length(vcheck) then
    exit; // bad version // by VJ extra checking

  for i := 0 to Length(vcheck) - 1 do
    if save_p[i] <> ord(vcheck[i + 1]) then
      exit; // bad version

  save_p := PByteArray(integer(save_p) + VERSIONSIZE);

  gameskill := skill_t(save_p[0]);
  save_p := PByteArray(integer(save_p) + 1);

  gameepisode := save_p[0];
  save_p := PByteArray(integer(save_p) + 1);

  gamemap := save_p[0];
  save_p := PByteArray(integer(save_p) + 1);

  for i := 0 to MAXPLAYERS - 1 do
  begin
    playeringame[i] := save_p[0] <> 0;
    save_p := PByteArray(integer(save_p) + 1);
  end;

  // load a base level
  G_InitNew(gameskill, gameepisode, gamemap);

  // get the times
  a := save_p[0];
  save_p := PByteArray(integer(save_p) + 1);

  b := save_p[0];
  save_p := PByteArray(integer(save_p) + 1);

  c := save_p[0];
  save_p := PByteArray(integer(save_p) + 1);

  leveltime := _SHL(a, 16) + _SHL(b, 8) + c;

  // dearchive all the modifications
  P_UnArchivePlayers;
  P_UnArchiveWorld;
  P_UnArchiveThinkers;
  P_UnArchiveSpecials;

  if save_p[0] <> $1d then
    I_Error('G_DoLoadGame(): Bad savegame');

  // done
  Z_Free(savebuffer);

  if setsizeneeded then
    R_ExecuteSetViewSize;

  // draw the pattern into the back screen
  R_FillBackScreen;
end;

//
// G_SaveGame
// Called by the menu task.
// Description is a 24 byte text string
//
procedure G_SaveGame(slot: integer; const description: string);
begin
  savegameslot := slot;
  savedescription := description;
  sendsave := true;
end;

procedure G_DoSaveGame;
var
  name: string;
  name2: string;
  description: string;
  length: integer;
  i: integer;
begin
  if M_CheckParmCDROM then
    sprintf(name, 'c:\doomdata\' + SAVEGAMENAME + '%d.dsg', [savegameslot])
  else
    sprintf(name, SAVEGAMENAME + '%d.dsg', [savegameslot]);
  description := savedescription;

  save_p := PByteArray(integer(screens[1]) + $4000);
  savebuffer := PByteArray(integer(screens[1]) + $4000);

  memcpy(save_p, @description[1], SAVESTRINGSIZE);

  save_p := PByteArray(integer(save_p) + SAVESTRINGSIZE);
  name2 := '';

  sprintf(name2, 'version %d', [VERSION]);
  memcpy(save_p, @name2[1], VERSIONSIZE);
  save_p := PByteArray(integer(save_p) + VERSIONSIZE);

  save_p[0] := Ord(gameskill);
  save_p := PByteArray(integer(save_p) + 1);

  save_p[0] := gameepisode;
  save_p := PByteArray(integer(save_p) + 1);

  save_p[0] := gamemap;
  save_p := PByteArray(integer(save_p) + 1);

  for i := 0 to MAXPLAYERS - 1 do
  begin
    save_p[0] := intval(playeringame[i]);
    save_p := PByteArray(integer(save_p) + 1);
  end;

  save_p[0] := _SHR(leveltime, 16);
  save_p := PByteArray(integer(save_p) + 1);

  save_p[0] := _SHR(leveltime, 8);
  save_p := PByteArray(integer(save_p) + 1);

  save_p[0] := leveltime;
  save_p := PByteArray(integer(save_p) + 1);

  P_ArchivePlayers;
  P_ArchiveWorld;
  P_ArchiveThinkers;
  P_ArchiveSpecials;

  save_p[0] := $1d; // consistancy marker
  save_p := PByteArray(integer(save_p) + 1);

  length := integer(save_p) - integer(savebuffer);
  if length > SAVEGAMESIZE then
    I_Error('G_DoSaveGame(): Savegame buffer overrun');
  M_WriteFile(name, savebuffer, length);
  gameaction := ga_nothing;
  savedescription := '';

  players[consoleplayer]._message := GGSAVED;

  // draw the pattern into the back screen
  R_FillBackScreen;
end; 

//
// G_InitNew
// Can be called by the startup code or the menu task,
// consoleplayer, displayplayer, playeringame[] should be set. 
//
var
  d_skill: skill_t;
  d_episode: integer;
  d_map: integer;

procedure G_DeferedInitNew(skill: skill_t; episode, map: integer);
begin
  d_skill := skill;
  d_episode := episode;
  d_map := map;
  gameaction := ga_newgame;
end;

procedure G_DoNewGame;
begin
  demoplayback := false;
  netdemo := false;
  netgame := false;
  deathmatch := 0;
  playeringame[1] := false;
  playeringame[2] := false;
  playeringame[3] := false;
  respawnparm := false;
  fastparm := false;
  nomonsters := false;
  consoleplayer := 0;
  G_InitNew (d_skill, d_episode, d_map);
  gameaction := ga_nothing;
end;

procedure G_InitNew(skill: skill_t; episode, map: integer);
var
  i: integer;
begin
  if paused then
  begin
    paused := false;
    S_ResumeSound;
  end;

  if skill > sk_nightmare then
    skill := sk_nightmare;


  // This was quite messy with SPECIAL and commented parts.
  // Supposedly hacks to make the latest edition work.
  // It might not work properly.
  if episode < 1 then
    episode := 1;

  if gamemode = retail then
  begin
    if episode > 4 then
      episode := 4;
  end
  else if gamemode = shareware then
  begin
    if episode > 1 then
     episode := 1;  // only start episode 1 on shareware
  end
  else
  begin
    if episode > 3 then
      episode := 3;
  end;



  if map < 1 then
    map := 1;

  if (map > 9) and (gamemode <> commercial) then
    map := 9;

  M_ClearRandom;

  if (skill = sk_nightmare) or respawnparm then
    respawnmonsters := true
  else
    respawnmonsters := false;

  if fastparm or ((skill = sk_nightmare) and (gameskill <> sk_nightmare)) then
  begin
    for i := Ord(S_SARG_RUN1) to Ord(S_SARG_PAIN2) do
      states[i].tics := _SHR(states[i].tics, 1);
    mobjinfo[Ord(MT_BRUISERSHOT)].speed := 20 * FRACUNIT;
    mobjinfo[Ord(MT_HEADSHOT)].speed := 20 * FRACUNIT;
    mobjinfo[Ord(MT_TROOPSHOT)].speed := 20 * FRACUNIT;
  end
  else if (skill <> sk_nightmare) and (gameskill = sk_nightmare) then
  begin
    for i := Ord(S_SARG_RUN1) to Ord(S_SARG_PAIN2) do
      states[i].tics := _SHL(states[i].tics, 1);
    mobjinfo[Ord(MT_BRUISERSHOT)].speed := 15 * FRACUNIT;
    mobjinfo[Ord(MT_HEADSHOT)].speed := 10 * FRACUNIT;
    mobjinfo[Ord(MT_TROOPSHOT)].speed := 10 * FRACUNIT;
  end;


  // force players to be initialized upon first level load
  for i := 0 to MAXPLAYERS - 1 do
    players[i].playerstate := PST_REBORN;

  usergame := true;  // will be set false if a demo
  paused := false;
  demoplayback := false;
  automapactive := false;
  viewactive := true;
  gameepisode := episode;
  gamemap := map;
  gameskill := skill;

  viewactive := true;

  // set the sky map for the episode
  if gamemode = commercial then
  begin
    skytexture := R_TextureNumForName ('SKY3');
    if gamemap < 12 then
      skytexture := R_TextureNumForName('SKY1')
    else if gamemap < 21 then
      skytexture := R_TextureNumForName ('SKY2');
  end
  else
  begin
    skytexture := R_TextureNumForName('SKY' + Chr(Ord('1') + episode - 1));
{    case episode of
      1: skytexture := R_TextureNumForName('SKY1');
      2: skytexture := R_TextureNumForName('SKY2');
      3: skytexture := R_TextureNumForName('SKY3');
      4: skytexture := R_TextureNumForName('SKY4'); // Special Edition sky
    end;}
  end;

  G_DoLoadLevel;
end;

//
// DEMO RECORDING
//
const
  DEMOMARKER = $80;

procedure G_ReadDemoTiccmd(cmd: Pticcmd_t);
begin
  if demo_p[0] = DEMOMARKER then
  begin
    // end of demo data stream
    G_CheckDemoStatus;
    exit;
  end;
  cmd.forwardmove := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  cmd.sidemove := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  cmd.angleturn := _SHL(demo_p[0], 8);
  demo_p := PByteArray(integer(demo_p) + 1);

  cmd.buttons := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  cmd.look := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);
end;

procedure G_WriteDemoTiccmd(cmd: Pticcmd_t);
begin
  if gamekeydown[Ord('q')] then // press q to end demo recording
    G_CheckDemoStatus;

  demo_p[0] := Ord(cmd.forwardmove);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := Ord(cmd.sidemove);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := _SHR((cmd.angleturn + 128), 8);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := cmd.buttons;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := cmd.look;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p := PByteArray(integer(demo_p) - 5);

  if integer(demo_p) > integer(demoend) - 16 then
  begin
    // no more space
    G_CheckDemoStatus;
    exit;
  end;
  G_ReadDemoTiccmd(cmd);  // make SURE it is exactly the same
end;

//
// G_RecordDemo
//
procedure G_RecordDemo(const name: string);
var
  i: integer;
  maxsize: integer;
begin
  usergame := false;
  demoname := name + '.lmp';

  i := M_CheckParm ('-maxdemo');
  if (i <> 0) and (i < myargc - 1) then
    maxsize := atoi(myargv[i + 1]) * 1024
  else
    maxsize := $20000;

  demobuffer := Z_Malloc(maxsize, PU_STATIC, nil);
  demoend := @demobuffer[maxsize];

  demorecording := true;
end;

procedure G_BeginRecording;
var
  i: integer;
begin
  demo_p := demobuffer;

  demo_p[0] := VERSION;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := Ord(gameskill);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := gameepisode;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := gamemap;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := deathmatch;
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := intval(respawnparm);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := intval(fastparm);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := intval(nomonsters);
  demo_p := PByteArray(integer(demo_p) + 1);

  demo_p[0] := consoleplayer;
  demo_p := PByteArray(integer(demo_p) + 1);

  for i := 0 to MAXPLAYERS - 1 do
  begin
    demo_p[0] := intval(playeringame[i]);
    demo_p := PByteArray(integer(demo_p) + 1);
  end;
end;

//
// G_PlayDemo
//
var
  defdemoname: string;

procedure G_DeferedPlayDemo(const name: string);
begin
  defdemoname := name;
  gameaction := ga_playdemo;
end;

procedure G_DoPlayDemo;
var
  skill: skill_t;
  i, episode, map: integer;
begin
  gameaction := ga_nothing;
  demobuffer := W_CacheLumpName(defdemoname, PU_STATIC);
  demo_p := demobuffer;

  if demo_p[0] <> VERSION then
  begin
    fprintf(stderr, 'G_DoPlayDemo(): Demo is from a different game version = %d.%d!' + #13#10,
      [demo_p[0] div 100, demo_p[0] mod 100]);
    gameaction := ga_nothing;
    exit;
  end;
  demo_p := PByteArray(integer(demo_p) + 1);

  skill := skill_t(demo_p[0]);
  demo_p := PByteArray(integer(demo_p) + 1);

  episode := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  map := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  deathmatch := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  respawnparm := demo_p[0] <> 0;
  demo_p := PByteArray(integer(demo_p) + 1);

  fastparm := demo_p[0] <> 0;
  demo_p := PByteArray(integer(demo_p) + 1);

  nomonsters := demo_p[0] <> 0;
  demo_p := PByteArray(integer(demo_p) + 1);

  consoleplayer := demo_p[0];
  demo_p := PByteArray(integer(demo_p) + 1);

  for i := 0 to MAXPLAYERS - 1 do
  begin
    playeringame[i] := demo_p[0] <> 0;
    demo_p := PByteArray(integer(demo_p) + 1);
  end;

  if playeringame[1] then
  begin
    netgame := true;
    netdemo := true;
  end;

  // don't spend a lot of time in loadlevel
  precache := false;
  G_InitNew(skill, episode, map);
  precache := true;

  usergame := false;
  demoplayback := true;
end;

//
// G_TimeDemo
//
procedure G_TimeDemo(const name: string);
begin
  nodrawers := M_CheckParm('-nodraw') <> 0;
  noblit := M_CheckParm ('-noblit') <> 0;
  timingdemo := true;
  singletics := true;

  defdemoname := name;
  gameaction := ga_playdemo;
end;

(*
===================
=
= G_CheckDemoStatus
=
= Called after a death or level completion to allow demos to be cleaned up
= Returns true if a new demo loop action will take place
===================
*)

function G_CheckDemoStatus: boolean;
var
  endtime: integer;
begin
  if timingdemo then
  begin
    endtime := I_GetTime;
    I_Error('G_CheckDemoStatus(): timed %d gametics in %d realtics', [gametic, endtime - starttime]);
  end;

  if demoplayback then
  begin
    if singledemo then
      I_Quit;

    Z_ChangeTag(demobuffer, PU_CACHE);
    demoplayback := false;
    netdemo := false;
    netgame := false;
    deathmatch := 0;
    playeringame[1] := false;
    playeringame[2] := false;
    playeringame[3] := false;
    respawnparm := false;
    fastparm := false;
    nomonsters := false;
    consoleplayer := 0;
    D_AdvanceDemo;
    result := true;
    exit;
  end;

  if demorecording then
  begin
    demo_p[0] := DEMOMARKER;
    demo_p := PByteArray(integer(demo_p) + 1);

    M_WriteFile(demoname, demobuffer, POperation(demo_p, demobuffer, '-', SizeOf(byte)));
    Z_Free(demobuffer);
    demorecording := false;
    I_Error('G_CheckDemoStatus(): Demo %s recorded', [demoname]);
  end;

  result := false;
end;

function G_IsDemoRecording: boolean;
begin
  result := demorecording;
end;


(* Unit i_video implementation *)


                                                                     

var
  curpal: array[0..255] of LongWord;
  g_pDD: IDirectDraw7 = nil; // DirectDraw object
  g_pDDSPrimary: IDirectDrawSurface7 = nil;// DirectDraw primary surface
  g_pDDScreen: IDirectDrawSurface7 = nil;   // DirectDraw surface

type
  screenitem_t = LongWord;
  screenitem_tArray = array[0..$FFFF] of screenitem_t;
  Pscreenitem_tArray = ^screenitem_tArray;

var
  screen: Pscreenitem_tArray; //array[0..MAXWIDTH * MAXHEIGHT - 1] of screenitem_t;

var
  bpp: integer;

var
  s_alttab_disabled: boolean = false;

procedure I_DisableAltTab;
var
  old: Boolean;
begin
  if s_alttab_disabled then
    Exit;

  if Win32Platform = VER_PLATFORM_WIN32_NT then
  begin
    if isLibrary then
      RegisterHotKey(0, $C000, MOD_ALT, VK_TAB)
    else
      RegisterHotKey(0, 0, MOD_ALT, VK_TAB)
  end
  else
    SystemParametersInfo(SPI_SCREENSAVERRUNNING, 1, @old, 0);

  s_alttab_disabled := True;
end;

procedure I_EnableAltTab;
var
  old: Boolean;
begin
  if s_alttab_disabled then
  begin
    if Win32Platform = VER_PLATFORM_WIN32_NT then
    begin
      if isLibrary then
        UnregisterHotKey(0, $C000)
      else
        UnregisterHotKey(0, 0)
    end
    else
      SystemParametersInfo(SPI_SCREENSAVERRUNNING, 0, @old, 0);

    s_alttab_disabled := False;
  end;
end;

procedure I_ShutdownGraphics1;
begin
  FreeMem(screen);
  if g_pDD <> nil then
  begin
    if g_pDDSPrimary <> nil then
    begin
      g_pDDSPrimary._Release;
    end;
    g_pDD._Release;
  end;
  I_EnableAltTab;
end;

procedure I_ShutdownGraphics;
begin
  FreeMem(screen);
  I_EnableAltTab;
end;

{// Takes full 8 bit values.
procedure I_SetPalette(palette: PByteArray);
var
  i: integer;
begin
  for i := 0 to 255 do
    curpal[i] := (palette[3 * i] shl 16) +
                 (palette[3 * i + 1] shl 8) +
                 (palette[3 * i + 2]);
end;}

procedure I_UpdateNoBlit;
begin
end;

//
// I_FinishUpdate
//
procedure I_FinishUpdate;
var
  i: integer;
  dest: PLongWord;
  destw: PWord;
  pixel: LongWord;
  r, g, b: LongWord;
  src: PByte;
  rect: TRect;
begin
  if hMainWnd = 0 then
    exit;
  if screens[_FG] = nil then
    exit;

  rect.Left := 0;
  rect.Top := 0;
  rect.Right := SCREENWIDTH;
  rect.Bottom := SCREENHEIGHT;

  src := @(screens[0]^);
  if bpp = 32 then
  begin
    dest := @screen[0];
    for i := 0 to SCREENWIDTH * SCREENHEIGHT - 1 do
    begin
      dest^ := curpal[src^];
      inc(dest);
      inc(src);
    end;
  end
  else if bpp = 16 then
  begin
    destw := @screen[0];
    for i := 0 to SCREENWIDTH * SCREENHEIGHT - 1 do
    begin
      pixel := curpal[src^];
      r := (pixel shr 19) and 31;
      g := (pixel shr 11) and 31;
      b := (pixel shr 3) and 31;
      destw^ := (r shl 11) or (g shl 6) or b;
      inc(destw);
      inc(src);
    end;
  end;
  g_pDDSPrimary.BltFast(0, 0, g_pDDScreen, rect, 0);
end;

//
// Palette stuff.
//

//
// I_SetPalette
//
procedure I_SetPalette(palette: PByteArray);
var
  dest: PLongWord;
  src: PByteArray;
begin
  dest := @curpal[0];
  src := palette;
  while integer(src) < integer(@palette[256 * 3]) do
  begin
		dest^ := (LongWord(gammatable[usegamma, src[0]]) shl 16) or
             (LongWord(gammatable[usegamma, src[1]]) shl 8) or
             (LongWord(gammatable[usegamma, src[2]]));
    inc(dest);
    incp(pointer(src), 3);
  end;
end;

// Called by D_DoomMain,
// determines the hardware configuration
// and sets up the video mode
procedure I_InitGraphics;
var
  hres: HRESULT;
  ddsd: DDSURFACEDESC2;

  procedure I_ErrorInitGraphics(const proc: string);
  begin
    I_Error('I_InitGraphics(): %s failed, result = %d', [proc, hres]);
  end;

begin
  if g_pDD <> nil then
    exit;

  screen := malloc(SCREENWIDTH * SCREENHEIGHT * SizeOf(screenitem_t));

///////////////////////////////////////////////////////////////////////////
// Create the main DirectDraw object
///////////////////////////////////////////////////////////////////////////
  hres := DirectDrawCreateEx(nil, g_pDD, IID_IDirectDraw7, nil);
  if hres <> DD_OK then
    I_ErrorInitGraphics('DirectDrawCreateEx');

  if fullscreen then
  begin
    // Get exclusive mode
    hres := g_pDD.SetCooperativeLevel(hMainWnd, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN);
    if hres <> DD_OK then
      I_ErrorInitGraphics('SetCooperativeLevel');

    // Set the video mode to SCREENWIDTH x SCREENHEIGHT x 32
    hres := g_pDD.SetDisplayMode(SCREENWIDTH, SCREENHEIGHT, 32, 0, 0);
    if hres <> DD_OK then
    begin
    // Fullscreen mode failed, trying window mode
      fullscreen := false;

      SetWindowPos(hMainWnd, 0, 0, 0, SCREENWIDTH, SCREENHEIGHT, SWP_SHOWWINDOW);

      printf('SetDisplayMode(): Failed to fullscreen %dx%dx%d, trying window mode',
        [SCREENWIDTH, SCREENHEIGHT, 32]);

      hres := g_pDD.SetCooperativeLevel(hMainWnd, DDSCL_NORMAL);
      if hres <> DD_OK then
        I_ErrorInitGraphics('SetDisplayMode');
    end
    else
      I_DisableAltTab;
  end
  else
  begin
    hres := g_pDD.SetCooperativeLevel(hMainWnd, DDSCL_NORMAL);
    if hres <> DD_OK then
      I_ErrorInitGraphics('SetCooperativeLevel');
  end;

  ZeroMemory(ddsd, SizeOf(ddsd));
  ddsd.dwSize := SizeOf(ddsd);
  ddsd.dwFlags := DDSD_CAPS;
  ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
  hres := g_pDD.CreateSurface(ddsd, g_pDDSPrimary, nil);
  if hres <> DD_OK then
    I_ErrorInitGraphics('CreateSurface');

  ZeroMemory(ddsd, SizeOf(ddsd));
  ZeroMemory(ddsd.ddpfPixelFormat, SizeOf(ddsd.ddpfPixelFormat));

  ddsd.ddpfPixelFormat.dwSize := SizeOf(ddsd.ddpfPixelFormat);
  g_pDDSPrimary.GetPixelFormat(ddsd.ddpfPixelFormat);

  ddsd.dwSize := SizeOf(ddsd);
  ddsd.dwFlags := DDSD_WIDTH or DDSD_HEIGHT or DDSD_LPSURFACE or
                  DDSD_PITCH or DDSD_PIXELFORMAT or DDSD_CAPS;
  ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN or DDSCAPS_SYSTEMMEMORY;

  bpp := ddsd.ddpfPixelFormat.dwRGBBitCount;

  ddsd.dwWidth := SCREENWIDTH;
  ddsd.dwHeight := SCREENHEIGHT;

  if bpp = 32 then
    ddsd.lPitch := 4 * SCREENWIDTH // Display is true color
  else if bpp = 16 then
    ddsd.lPitch := 2 * SCREENWIDTH
  else
    I_Error('I_InitGraphics(): invalid colordepth = %d, only 16 and 32 bit color depth allowed', [bpp]);

  ddsd.lpSurface := screen;

  hres := g_pDD.CreateSurface(ddsd, g_pDDScreen, nil);
  if hres <> DD_OK then
    I_ErrorInitGraphics('CreateSurface');
end;

procedure I_ReadScreen(scr: PByteArray);
begin
  memcpy(scr, screens[_FG], SCREENWIDTH * SCREENHEIGHT);
end;

procedure I_ReadScreen32bit(scr: PByteArray);
begin
  memcpy(scr, screen, SCREENWIDTH * SCREENHEIGHT * SizeOf(screenitem_t));
end;


(* Unit p_map implementation *)


                                                                                                                                                    

var
  tmbbox: array[0..3] of fixed_t;
  tmthing: Pmobj_t;
  tmflags: integer;
  tmx: fixed_t;
  tmy: fixed_t;

//
// TELEPORT MOVE
// 

//
// PIT_StompThing
//
function PIT_StompThing(thing: Pmobj_t): boolean;
var
  blockdist: fixed_t;
begin
  if thing.flags and MF_SHOOTABLE = 0 then
  begin
    result := true;
    exit;
  end;

  blockdist := thing.radius + tmthing.radius;

  if (abs(thing.x - tmx) >= blockdist) or (abs(thing.y - tmy) >= blockdist) then
  begin
    // didn't hit it
    result := true;
    exit;
  end;

  // don't clip against self
  if thing = tmthing then
  begin
    result := true;
    exit;
  end;

  // monsters don't stomp things except on boss level
  if (not boolval(tmthing.player)) and (gamemap <> 30) then
  begin
    result := false;
    exit;
  end;

  P_DamageMobj(thing, tmthing, tmthing, 10000);

  result := true;
end;

//
// P_TeleportMove
//
function P_TeleportMove(thing: Pmobj_t; x, y: fixed_t): boolean;
var
  xl: integer;
  xh: integer;
  yl: integer;
  yh: integer;
  bx: integer;
  by: integer;
  newsubsec: Psubsector_t;
begin
  // kill anything occupying the position
  tmthing := thing;
  tmflags := thing.flags;

  tmx := x;
  tmy := y;

  tmbbox[BOXTOP] := y + tmthing.radius;
  tmbbox[BOXBOTTOM] := y - tmthing.radius;
  tmbbox[BOXRIGHT] := x + tmthing.radius;
  tmbbox[BOXLEFT] := x - tmthing.radius;

  newsubsec := R_PointInSubsector(x, y);
  ceilingline := nil;

  // The base floor/ceiling is from the subsector
  // that contains the point.
  // Any contacted lines the step closer together
  // will adjust them.
  tmdropoffz := newsubsec.sector.floorheight;
  tmfloorz := tmdropoffz;

  tmceilingz := newsubsec.sector.ceilingheight;

  inc(validcount); 
  numspechit := 0;

  // stomp on any things contacted
  xl := _SHR((tmbbox[BOXLEFT] - bmaporgx - MAXRADIUS), MAPBLOCKSHIFT);
  xh := _SHR((tmbbox[BOXRIGHT] - bmaporgx + MAXRADIUS), MAPBLOCKSHIFT);
  yl := _SHR((tmbbox[BOXBOTTOM] - bmaporgy - MAXRADIUS), MAPBLOCKSHIFT);
  yh := _SHR((tmbbox[BOXTOP] - bmaporgy + MAXRADIUS), MAPBLOCKSHIFT);

  for bx := xl to xh do
    for by := yl to yh do
      if not P_BlockThingsIterator(bx, by, PIT_StompThing) then
      begin
        result := false;
        exit;
      end;

  // the move is ok,
  // so link the thing into its new position
  P_UnsetThingPosition(thing);

  thing.floorz := tmfloorz;
  thing.ceilingz := tmceilingz;
  thing.x := x;
  thing.y := y;

  P_SetThingPosition(thing);

  result := true;
end;

//
// MOVEMENT ITERATOR FUNCTIONS
//


//
// PIT_CheckLine
// Adjusts tmfloorz and tmceilingz as lines are contacted
//
function PIT_CheckLine(ld: Pline_t): boolean;
begin
  if (tmbbox[BOXRIGHT] <= ld.bbox[BOXLEFT]) or
     (tmbbox[BOXLEFT] >= ld.bbox[BOXRIGHT]) or
     (tmbbox[BOXTOP] <= ld.bbox[BOXBOTTOM]) or
     (tmbbox[BOXBOTTOM] >= ld.bbox[BOXTOP]) then
  begin
    result := true;
    exit;
  end;

  if P_BoxOnLineSide(@tmbbox, ld) <> -1 then
  begin
    result := true;
    exit;
  end;
		
  // A line has been hit

  // The moving thing's destination position will cross
  // the given line.
  // If this should not be allowed, return false.
  // If the line is special, keep track of it
  // to process later if the move is proven ok.
  // NOTE: specials are NOT sorted by order,
  // so two special lines that are only 8 pixels apart
  // could be crossed in either order.

  if not boolval(ld.backsector) then
  begin
    result := false;  // one sided line
    exit;
  end;

  if not boolval(tmthing.flags and MF_MISSILE) then
  begin
    if (ld.flags and ML_BLOCKING) <> 0 then
    begin
      result := false;  // explicitly blocking everything
      exit;
    end;

    if (not boolval(tmthing.player)) and ((ld.flags and ML_BLOCKMONSTERS) <> 0) then
    begin
      result := false;  // block monsters only
      exit;
    end;
  end;

  // set openrange, opentop, openbottom
  P_LineOpening(ld);

  // adjust floor / ceiling heights
  if opentop < tmceilingz then
  begin
    tmceilingz := opentop;
    ceilingline := ld;
  end;

  if openbottom > tmfloorz then
    tmfloorz := openbottom;

  if lowfloor < tmdropoffz then
    tmdropoffz := lowfloor;

  // if contacted a special line, add it to the list
  if boolval(ld.special) then
  begin
    spechit[numspechit] := ld;

    inc(numspechit);
//    fprintf(stderr, 'numspechit = %d' + #13#10, [numspechit]);
  end;

  result := true;
end;

//
// PIT_CheckThing
//
function PIT_CheckThing(thing: Pmobj_t): boolean;
var
  blockdist: fixed_t;
  solid: boolean;
  damage: integer;
begin
  if not boolval(thing.flags and (MF_SOLID or MF_SPECIAL or MF_SHOOTABLE)) then
  begin
    result := true;
    exit;
  end;

  blockdist := thing.radius + tmthing.radius;

  if (abs(thing.x - tmx) >= blockdist) or (abs(thing.y - tmy) >= blockdist) then
  begin
    // didn't hit it
    result := true;
    exit;
  end;

  // don't clip against self
  if thing = tmthing then
  begin
    result := true;
    exit;
  end;

  // check for skulls slamming into things
  if (tmthing.flags and MF_SKULLFLY) <> 0 then
  begin
    damage := ((P_Random mod 8) + 1) * tmthing.info.damage;
    P_DamageMobj(thing, tmthing, tmthing, damage);

    tmthing.flags := tmthing.flags and (not MF_SKULLFLY);
    tmthing.momx := 0;
    tmthing.momy := 0;
    tmthing.momz := 0;

    P_SetMobjState(tmthing, statenum_t(tmthing.info.spawnstate));

    result := false;  // stop moving
    exit;
  end;

  // missiles can hit other things
  if (tmthing.flags and MF_MISSILE) <> 0 then
  begin
    // see if it went over / under
    if tmthing.z > thing.z + thing.height then
    begin
      result := true; // overhead
      exit;
    end;
    if tmthing.z + tmthing.height < thing.z then
    begin
      result := true; // underneath
      exit;
    end;

    if boolval(tmthing.target) and (
        (tmthing.target._type = thing._type) or
        ((tmthing.target._type = MT_KNIGHT) and (thing._type = MT_BRUISER)) or
        ((tmthing.target._type = MT_BRUISER) and (thing._type = MT_KNIGHT)) ) then
    begin
      // Don't hit same species as originator.
      if thing = tmthing.target then
      begin
        result := true;
        exit;
      end;

      if thing._type <> MT_PLAYER then
      begin
        // Explode, but do no damage.
        // Let players missile other players.
        result := false;
        exit;
      end;
    end;

    if not boolval(thing.flags and MF_SHOOTABLE) then
    begin
      // didn't do any damage
      result := not boolval(thing.flags and MF_SOLID);
      exit;
    end;

    // damage / explode
    damage := ((P_Random mod 8) + 1) * tmthing.info.damage;
    P_DamageMobj(thing, tmthing, tmthing.target, damage);

    // don't traverse any more
    result := false;
    exit;
  end;
    
  // check for special pickup
  if (thing.flags and MF_SPECIAL) <> 0 then
  begin
    solid := (thing.flags and MF_SOLID) <> 0;
    if (tmflags and MF_PICKUP) <> 0 then
    begin
      // can remove thing
      P_TouchSpecialThing(thing, tmthing);
    end;
    result := not solid;
  end
  else
    result := (thing.flags and MF_SOLID) = 0;
end;

//
// MOVEMENT CLIPPING
//

//
// P_CheckPosition
// This is purely informative, nothing is modified
// (except things picked up).
// 
// in:
//  a mobj_t (can be valid or invalid)
//  a position to be checked
//   (doesn't need to be related to the mobj_t->x,y)
//
// during:
//  special things are touched if MF_PICKUP
//  early out on solid lines?
//
// out:
//  newsubsec
//  floorz
//  ceilingz
//  tmdropoffz
//   the lowest point contacted
//   (monsters won't move to a dropoff)
//  speciallines[]
//  numspeciallines
//
function P_CheckPosition(thing: Pmobj_t; x, y: fixed_t): boolean;
var
  xl: integer;
  xh: integer;
  yl: integer;
  yh: integer;
  bx: integer;
  by: integer;
  newsubsec: Psubsector_t;
begin
  tmthing := thing;
  tmflags := thing.flags;

  tmx := x;
  tmy := y;

  tmbbox[BOXTOP] := y + tmthing.radius;
  tmbbox[BOXBOTTOM] := y - tmthing.radius;
  tmbbox[BOXRIGHT] := x + tmthing.radius;
  tmbbox[BOXLEFT] := x - tmthing.radius;

  newsubsec := R_PointInSubsector(x, y);
  ceilingline := nil;

  // The base floor / ceiling is from the subsector
  // that contains the point.
  // Any contacted lines the step closer together
  // will adjust them.
  tmdropoffz := newsubsec.sector.floorheight;
  tmfloorz := tmdropoffz;
  tmceilingz := newsubsec.sector.ceilingheight;

  inc(validcount);
  numspechit := 0;

  if (tmflags and MF_NOCLIP) <> 0 then
  begin
    result := true;
    exit;
  end;

  // Check things first, possibly picking things up.
  // The bounding box is extended by MAXRADIUS
  // because mobj_ts are grouped into mapblocks
  // based on their origin point, and can overlap
  // into adjacent blocks by up to MAXRADIUS units.
  xl := _SHR((tmbbox[BOXLEFT] - bmaporgx - MAXRADIUS), MAPBLOCKSHIFT);
  xh := _SHR((tmbbox[BOXRIGHT] - bmaporgx + MAXRADIUS), MAPBLOCKSHIFT);
  yl := _SHR((tmbbox[BOXBOTTOM] - bmaporgy - MAXRADIUS), MAPBLOCKSHIFT);
  yh := _SHR((tmbbox[BOXTOP] - bmaporgy + MAXRADIUS), MAPBLOCKSHIFT);

  for bx := xl to xh do
    for by := yl to yh do
      if not P_BlockThingsIterator(bx, by, PIT_CheckThing) then
      begin
        result := false;
        exit;
      end;

  // check lines
  xl := _SHR((tmbbox[BOXLEFT] - bmaporgx), MAPBLOCKSHIFT);
  xh := _SHR((tmbbox[BOXRIGHT] - bmaporgx), MAPBLOCKSHIFT);
  yl := _SHR((tmbbox[BOXBOTTOM] - bmaporgy), MAPBLOCKSHIFT);
  yh := _SHR((tmbbox[BOXTOP] - bmaporgy), MAPBLOCKSHIFT);

  for bx := xl to xh do
    for by := yl to yh do
      if not P_BlockLinesIterator(bx, by, PIT_CheckLine) then
      begin
        result := false;
        exit;
      end;

  result := true;
end;

//
// P_TryMove
// Attempt to move to a new position,
// crossing special lines unless MF_TELEPORT is set.
//
function P_TryMove(thing: Pmobj_t; x, y: fixed_t): boolean;
var
  oldx: fixed_t;
  oldy: fixed_t;
  side: integer;
  oldside: integer;
  ld: Pline_t;
begin
  floatok := false;
  if not P_CheckPosition(thing, x, y) then
  begin
    result := false;  // solid wall or thing
    exit;
  end;

  if (thing.flags and MF_NOCLIP) = 0 then
  begin
    if tmceilingz - tmfloorz < thing.height then
    begin
      result := false;  // doesn't fit
      exit;
    end;

    floatok := true;

    if ((thing.flags and MF_TELEPORT) = 0) and
	     (tmceilingz - thing.z < thing.height) then
    begin
      result := false;  // mobj must lower itself to fit
      exit;
    end;

    if ((thing.flags and MF_TELEPORT) = 0) and
       (tmfloorz - thing.z > 24 * FRACUNIT) then
    begin
      result := false;  // too big a step up
      exit;
    end;

    if ((thing.flags and (MF_DROPOFF or MF_FLOAT)) = 0) and
       (tmfloorz - tmdropoffz > 24 * FRACUNIT) then
    begin
      result := false;  // don't stand over a dropoff
      exit;
    end;
  end;

  // the move is ok,
  // so link the thing into its new position
  P_UnsetThingPosition(thing);

  oldx := thing.x;
  oldy := thing.y;
  thing.floorz := tmfloorz;
  thing.ceilingz := tmceilingz;
  thing.x := x;
  thing.y := y;

  P_SetThingPosition(thing);

  // if any special lines were hit, do the effect
  if (thing.flags and (MF_TELEPORT or MF_NOCLIP)) = 0 then
  begin
    while numspechit > 0 do
    begin
      // see if the line was crossed
      dec(numspechit); // VJ what happens to while loops ???
      ld := spechit[numspechit];
      side := P_PointOnLineSide(thing.x, thing.y, ld);
      oldside := P_PointOnLineSide(oldx, oldy, ld);
      if side <> oldside then
      begin
        if boolval(ld.special) then
          P_CrossSpecialLine(pOperation(ld, @lines[0], '-', SizeOf(ld^)), oldside, thing);
      end;
    end;
  end;

  result := true;
end;

//
// P_ThingHeightClip
// Takes a valid thing and adjusts the thing->floorz,
// thing->ceilingz, and possibly thing->z.
// This is called for all nearby monsters
// whenever a sector changes height.
// If the thing doesn't fit,
// the z will be set to the lowest value
// and false will be returned.
//
function P_ThingHeightClip(thing: Pmobj_t): boolean;
var
  onfloor: boolean;
begin
  onfloor := thing.z = thing.floorz;

  P_CheckPosition(thing, thing.x, thing.y);
  // what about stranding a monster partially off an edge?

  thing.floorz := tmfloorz;
  thing.ceilingz := tmceilingz;

  if onfloor then
  begin
    // walking monsters rise and fall with the floor
    thing.z := thing.floorz;
  end
  else
  begin
    // don't adjust a floating monster unless forced to
    if thing.z + thing.height > thing.ceilingz then
      thing.z := thing.ceilingz - thing.height;
  end;

  result := thing.ceilingz - thing.floorz >= thing.height;
end;

//
// SLIDE MOVE
// Allows the player to slide along any angled walls.
//
var
  bestslidefrac: fixed_t;
  secondslidefrac: fixed_t;

  bestslideline: Pline_t;
  secondslideline: Pline_t;

  slidemo: Pmobj_t;

  tmxmove: fixed_t;
  tmymove: fixed_t;

//
// P_HitSlideLine
// Adjusts the xmove / ymove
// so that the next move will slide along the wall.
//
procedure P_HitSlideLine(ld: Pline_t);
var
  side: integer;
  lineangle: angle_t;
  moveangle: angle_t;
  deltaangle: angle_t;
  movelen: fixed_t;
  newlen: fixed_t;
begin
  if ld.slopetype = ST_HORIZONTAL then
  begin
    tmymove := 0;
    exit;
  end;

  if ld.slopetype = ST_VERTICAL then
  begin
    tmxmove := 0;
    exit;
  end;

  side := P_PointOnLineSide(slidemo.x, slidemo.y, ld);

  lineangle := R_PointToAngle2(0, 0, ld.dx, ld.dy);

  if side = 1 then
    lineangle := lineangle + ANG180;

  moveangle := R_PointToAngle2(0, 0, tmxmove, tmymove);
  deltaangle := moveangle - lineangle;

  if deltaangle > ANG180 then
    deltaangle := deltaangle + ANG180;
    //	I_Error ("SlideLine: ang>ANG180");

  lineangle := _SHRW(lineangle, ANGLETOFINESHIFT);
  deltaangle := _SHRW(deltaangle, ANGLETOFINESHIFT);

  movelen := P_AproxDistance(tmxmove, tmymove);
  newlen := FixedMul(movelen, finecosine[deltaangle]);

  tmxmove := FixedMul(newlen, finecosine[lineangle]);
  tmymove := FixedMul(newlen, finesine[lineangle]);
end;

//
// PTR_SlideTraverse
//
function PTR_SlideTraverse(_in: Pintercept_t): boolean;
var
  li: Pline_t;

  procedure isblocking;
  begin
    // the line does block movement,
    // see if it is closer than best so far
    if _in.frac < bestslidefrac then
    begin
      secondslidefrac := bestslidefrac;
      secondslideline := bestslideline;
      bestslidefrac := _in.frac;
      bestslideline := li;
    end;
  end;

begin
  if not _in.isaline then
    I_Error('PTR_SlideTraverse(): not a line?');

  li := _in.d.line;

  if (li.flags and ML_TWOSIDED) = 0 then
  begin
    if boolval(P_PointOnLineSide(slidemo.x, slidemo.y, li)) then
    begin
      // don't hit the back side
      result := true;
      exit;
    end;
    isblocking;
    result := false; // stop
    exit;
  end;

  // set openrange, opentop, openbottom
  P_LineOpening(li);

  if openrange < slidemo.height then
  begin
    isblocking; // doesn't fit
    result := false; // stop
    exit;
  end;

  if opentop - slidemo.z < slidemo.height then
  begin
    isblocking; // mobj is too high
    result := false; // stop
    exit;
  end;

  if openbottom - slidemo.z > 24 * FRACUNIT then
  begin
    isblocking; // too big a step up
    result := false; // stop
    exit;
  end;

  // this line doesn't block movement
  result := true;
end;

//
// P_SlideMove
// The momx / momy move is bad, so try to slide
// along a wall.
// Find the first line hit, move flush to it,
// and slide along it
//
// This is a kludgy mess.
//
procedure P_SlideMove(mo: Pmobj_t);
var
  leadx: fixed_t;
  leady: fixed_t;
  trailx: fixed_t;
  traily: fixed_t;
  newx: fixed_t;
  newy: fixed_t;
  hitcount: integer;

  procedure stairstep;
  begin
    if not P_TryMove(mo, mo.x, mo.y + mo.momy) then
      P_TryMove(mo, mo.x + mo.momx, mo.y);
  end;

begin
  slidemo := mo;
  hitcount := 0;

  repeat
    inc(hitcount);
    if hitcount = 3 then
    begin
      stairstep;
      exit;  // don't loop forever
    end;

    // trace along the three leading corners
    if mo.momx > 0 then
    begin
      leadx := mo.x + mo.radius;
      trailx := mo.x - mo.radius;
    end
    else
    begin
      leadx := mo.x - mo.radius;
      trailx := mo.x + mo.radius;
    end;

    if mo.momy > 0 then
    begin
      leady := mo.y + mo.radius;
      traily := mo.y - mo.radius;
    end
    else
    begin
      leady := mo.y - mo.radius;
      traily := mo.y + mo.radius;
    end;

    bestslidefrac := FRACUNIT + 1;

    P_PathTraverse(leadx, leady, leadx + mo.momx, leady + mo.momy,
      PT_ADDLINES, PTR_SlideTraverse);
    P_PathTraverse(trailx, leady, trailx + mo.momx, leady + mo.momy,
      PT_ADDLINES, PTR_SlideTraverse);
    P_PathTraverse(leadx, traily, leadx + mo.momx, traily + mo.momy,
      PT_ADDLINES, PTR_SlideTraverse);

    // move up to the wall
    if bestslidefrac = FRACUNIT + 1 then
    begin
      // the move most have hit the middle, so stairstep
      stairstep;
      exit;
    end;

    // fudge a bit to make sure it doesn't hit
    bestslidefrac := bestslidefrac - $800;
    if bestslidefrac > 0 then
    begin
      newx := FixedMul(mo.momx, bestslidefrac);
      newy := FixedMul(mo.momy, bestslidefrac);

      if not P_TryMove(mo, mo.x + newx, mo.y + newy) then
      begin
        stairstep;
        exit;
      end;
    end;

    // Now continue along the wall.
    // First calculate remainder.
    bestslidefrac := FRACUNIT - (bestslidefrac + $800);

    if bestslidefrac > FRACUNIT then
      bestslidefrac := FRACUNIT;

    if bestslidefrac <= 0 then
      exit;

    tmxmove := FixedMul(mo.momx, bestslidefrac);
    tmymove := FixedMul(mo.momy, bestslidefrac);

    P_HitSlideLine(bestslideline);	// clip the moves

    mo.momx := tmxmove;
    mo.momy := tmymove;

  until P_TryMove(mo, mo.x + tmxmove, mo.y + tmymove);

end;

//
// P_LineAttack
//
var
  shootthing: Pmobj_t;

// Height if not aiming up or down
// ???: use slope for monsters?
  shootz: fixed_t;

  la_damage: integer;

  aimslope: fixed_t;

//
// PTR_AimTraverse
// Sets linetaget and aimslope when a target is aimed at.
//
function PTR_AimTraverse(_in: Pintercept_t): boolean;
var
  li: Pline_t;
  th: Pmobj_t;
  slope: fixed_t;
  thingtopslope: fixed_t;
  thingbottomslope: fixed_t;
  dist: fixed_t;
begin
  if _in.isaline then
  begin
    li := _in.d.line;

    if (li.flags and ML_TWOSIDED) = 0 then
    begin
      result := false; // stop
      exit;
    end;
	
    // Crosses a two sided line.
    // A two sided line will restrict
    // the possible target ranges.
    P_LineOpening(li);

    if openbottom >= opentop then
    begin
      result := false; // stop
      exit;
    end;

    dist := FixedMul(attackrange, _in.frac);

    if li.frontsector.floorheight <> li.backsector.floorheight then
    begin
      slope := FixedDiv(openbottom - shootz, dist);
      if slope > bottomslope then
        bottomslope := slope;
    end;

    if li.frontsector.ceilingheight <> li.backsector.ceilingheight then
    begin
      slope := FixedDiv(opentop - shootz, dist);
      if slope < topslope then
        topslope := slope;
    end;

    if topslope <= bottomslope then
    begin
      result := false; // stop
      exit;
    end;

	  result := true;	// shot continues
    exit;
  end;

  // shoot a thing
  th := _in.d.thing;
  if th = shootthing then
  begin
    result := true;	// can't shoot self
    exit;
  end;

  if (th.flags and MF_SHOOTABLE) = 0 then
  begin
    result := true; // corpse or something
    exit;
  end;

  // check angles to see if the thing can be aimed at
  dist := FixedMul(attackrange, _in.frac);
  thingtopslope := FixedDiv(th.z + th.height - shootz, dist);

  if thingtopslope < bottomslope then
  begin
    result := true; // shot over the thing
    exit;
  end;

  thingbottomslope := FixedDiv(th.z - shootz, dist);

  if thingbottomslope > topslope then
  begin
    result := true; // shot under the thing
    exit;
  end;

  // this thing can be hit!
  if thingtopslope > topslope then
    thingtopslope := topslope;

  if thingbottomslope < bottomslope then
    thingbottomslope := bottomslope;

  aimslope := (thingtopslope + thingbottomslope) div 2;
  linetarget := th;

  result := false; // don't go any farther
end;

//
// PTR_ShootTraverse
//
function PTR_ShootTraverse(_in: Pintercept_t): boolean;
var
  x: fixed_t;
  y: fixed_t;
  z: fixed_t;
  frac: fixed_t;
  li: Pline_t;
  th: Pmobj_t;
  slope: fixed_t;
  dist: fixed_t;
  thingtopslope: fixed_t;
  thingbottomslope: fixed_t;

  function hitline: boolean;
  begin
    // hit line
    // position a bit closer
    frac := _in.frac - FixedDiv(4 * FRACUNIT, attackrange);
    x := trace.x + FixedMul(trace.dx, frac);
    y := trace.y + FixedMul(trace.dy, frac);
    z := shootz + FixedMul(aimslope, FixedMul(frac, attackrange));

    if li.frontsector.ceilingpic = skyflatnum then
    begin
      // don't shoot the sky!
      if z > li.frontsector.ceilingheight then
      begin
        result := false;
        exit;
      end;

	    // it's a sky hack wall
      if boolval(li.backsector) and (li.backsector.ceilingpic = skyflatnum) then
      begin
        result := false;
        exit;
      end;
    end;
    // Spawn bullet puffs.
    P_SpawnPuff(x, y, z);

    // don't go any farther
    result := false;
  end;

begin
  if _in.isaline then
  begin
    li := _in.d.line;

    if boolval(li.special) then
      P_ShootSpecialLine(shootthing, li);

    if (li.flags and ML_TWOSIDED) = 0 then
    begin
      result := hitline;
      exit;
    end;

    // crosses a two sided line
    P_LineOpening(li);

    dist := FixedMul(attackrange, _in.frac);

    if li.frontsector.floorheight <> li.backsector.floorheight then
    begin
      slope := FixedDiv(openbottom - shootz, dist);
      if slope > aimslope then
      begin
        result := hitline;
        exit;
      end;
    end;

    if li.frontsector.ceilingheight <> li.backsector.ceilingheight then
    begin
      slope := FixedDiv(opentop - shootz, dist);
      if slope < aimslope then
      begin
        result := hitline;
        exit;
      end;
    end;

    // shot continues
    result := true;
    exit;
  end;

  // shoot a thing
  th := _in.d.thing;
  if th = shootthing then
  begin
    result := true; // can't shoot self
    exit;
  end;

  if (th.flags and MF_SHOOTABLE) = 0 then
  begin
    result := true; // corpse or something
    exit;
  end;

  // check angles to see if the thing can be aimed at
  dist := FixedMul(attackrange, _in.frac);
  thingtopslope := FixedDiv(th.z + th.height - shootz, dist);

  if thingtopslope < aimslope then
  begin
    result := true; // shot over the thing
    exit;
  end;

  thingbottomslope := FixedDiv(th.z - shootz, dist);

  if thingbottomslope > aimslope then
  begin
    result := true; // shot under the thing
    exit;
  end;


  // hit thing
  // position a bit closer
  frac := _in.frac - FixedDiv(10 * FRACUNIT, attackrange);

  x := trace.x + FixedMul(trace.dx, frac);
  y := trace.y + FixedMul(trace.dy, frac);
  z := shootz + FixedMul(aimslope, FixedMul(frac, attackrange));

  // Spawn bullet puffs or blod spots,
  // depending on target type.
  if (_in.d.thing.flags and MF_NOBLOOD) <> 0 then
    P_SpawnPuff(x, y, z)
  else
    P_SpawnBlood(x, y, z, la_damage);

  if la_damage <> 0 then
    P_DamageMobj(th, shootthing, shootthing, la_damage);

  // don't go any farther
  result := false;
end;

//
// P_AimLineAttack
//

function P_AimLineAttack(t1: Pmobj_t; angle: angle_t; distance: fixed_t): fixed_t;
var
  x2: fixed_t;
  y2: fixed_t;
begin
  angle := _SHRW(angle, ANGLETOFINESHIFT);
  shootthing := t1;

  x2 := t1.x + (distance div FRACUNIT) * finecosine[angle];
  y2 := t1.y + (distance div FRACUNIT) * finesine[angle];
  shootz := t1.z + _SHR(t1.height, 1) + 8 * FRACUNIT;

  // can't shoot outside view angles
  topslope := (100 * FRACUNIT) div 160; // VJ maybe screenwidth / 2
  bottomslope := -topslope; // VJ

  attackrange := distance;
  linetarget := nil;

  P_PathTraverse(t1.x, t1.y, x2, y2, PT_ADDLINES or PT_ADDTHINGS, PTR_AimTraverse);

  if linetarget <> nil then
    result := aimslope
  else
    result := 0;
end;

//
// P_LineAttack
// If damage == 0, it is just a test trace
// that will leave linetarget set.
//
procedure P_LineAttack(t1: Pmobj_t; angle: angle_t;
  distance: fixed_t; slope: fixed_t; damage: integer);
var
  x2: fixed_t;
  y2: fixed_t;
begin
  angle := _SHRW(angle, ANGLETOFINESHIFT);
  shootthing := t1;
  la_damage := damage;
  x2 := t1.x + (distance div FRACUNIT) * finecosine[angle];
  y2 := t1.y + (distance div FRACUNIT) * finesine[angle];
  shootz := t1.z + (_SHR(t1.height, 1)) + 8 * FRACUNIT;
  attackrange := distance;
  aimslope := slope;

  P_PathTraverse(t1.x, t1.y, x2, y2, PT_ADDLINES or PT_ADDTHINGS, PTR_ShootTraverse);
end;

//
// USE LINES
//
var
  usething: Pmobj_t;

function PTR_UseTraverse(_in: Pintercept_t): boolean;
var
  side: integer;
begin
  if not boolval(_in.d.line.special) then
  begin
    P_LineOpening(_in.d.line);
    if openrange <= 0 then
    begin
      S_StartSound(usething, Ord(sfx_noway));
	    // can't use through a wall
      result := false;
      exit;
    end;
    // not a special line, but keep checking
    result := true;
    exit;
  end;

  side := 0;
  if P_PointOnLineSide(usething.x, usething.y, _in.d.line) = 1 then
    side := 1;

    //	return false;		// don't use back side

  P_UseSpecialLine(usething, _in.d.line, side);

  // can't use for than one special line in a row
  result := false;
end;

//
// P_UseLines
// Looks for special lines in front of the player to activate.
//
procedure P_UseLines(player: Pplayer_t);
var
  angle: angle_t;
  x1: fixed_t;
  y1: fixed_t;
  x2: fixed_t;
  y2: fixed_t;
begin
  usething := player.mo;

  angle := _SHRW(player.mo.angle, ANGLETOFINESHIFT);

  x1 := player.mo.x;
  y1 := player.mo.y;
  x2 := x1 + (USERANGE div FRACUNIT) * finecosine[angle];
  y2 := y1 + (USERANGE div FRACUNIT) * finesine[angle];

  P_PathTraverse(x1, y1, x2, y2, PT_ADDLINES, PTR_UseTraverse);
end;

//
// RADIUS ATTACK
//
var
  bombsource: Pmobj_t;
  bombspot: Pmobj_t;
  bombdamage: integer;


//
// PIT_RadiusAttack
// "bombsource" is the creature
// that caused the explosion at "bombspot".
//
function PIT_RadiusAttack(thing: Pmobj_t): boolean;
var
  dx: fixed_t;
  dy: fixed_t;
  dist: fixed_t;
begin
  if (thing.flags and MF_SHOOTABLE) = 0 then
  begin
    result := true;
    exit;
  end;

  // Boss spider and cyborg
  // take no damage from concussion.
  if (thing._type = MT_CYBORG) or (thing._type = MT_SPIDER) then
  begin
    result := true;
    exit;
  end;
		
  dx := abs(thing.x - bombspot.x);
  dy := abs(thing.y - bombspot.y);

  if dx > dy then
    dist := dx
  else
    dist := dy;
  dist := (dist - thing.radius) div FRACUNIT;

  if dist < 0 then
    dist := 0;

  if dist >= bombdamage then
  begin
    result := true; // out of range
    exit;
  end;

  if P_CheckSight(thing, bombspot) then
  begin
    // must be in direct path
    P_DamageMobj(thing, bombspot, bombsource, bombdamage - dist);
  end;

  result := true;
end;

//
// P_RadiusAttack
// Source is the creature that caused the explosion at spot.
//
procedure P_RadiusAttack(spot: Pmobj_t; source: Pmobj_t; damage: integer);
var
  x: integer;
  y: integer;
  xl: integer;
  xh: integer;
  yl: integer;
  yh: integer;
  dist: fixed_t;
begin
  dist := (damage + MAXRADIUS) * FRACUNIT;
  yh := _SHR((spot.y + dist - bmaporgy), MAPBLOCKSHIFT);
  yl := _SHR((spot.y - dist - bmaporgy), MAPBLOCKSHIFT);
  xh := _SHR((spot.x + dist - bmaporgx), MAPBLOCKSHIFT);
  xl := _SHR((spot.x - dist - bmaporgx), MAPBLOCKSHIFT);
  bombspot := spot;
  bombsource := source;
  bombdamage := damage;

  for y := yl to yh do
    for x := xl to xh do
      P_BlockThingsIterator(x, y, PIT_RadiusAttack);
end;

//
// SECTOR HEIGHT CHANGING
// After modifying a sectors floor or ceiling height,
// call this routine to adjust the positions
// of all things that touch the sector.
//
// If anything doesn't fit anymore, true will be returned.
// If crunch is true, they will take damage
//  as they are being crushed.
// If Crunch is false, you should set the sector height back
//  the way it was and call P_ChangeSector again
//  to undo the changes.
//
var
  crushchange: boolean;
  nofit: boolean;

//
// PIT_ChangeSector
//
function PIT_ChangeSector(thing: Pmobj_t): boolean;
var
  mo: Pmobj_t;
begin
  if P_ThingHeightClip(thing) then
  begin
    // keep checking
    result := true;
    exit;
  end;

  // crunch bodies to giblets
  if thing.health <= 0 then
  begin
    P_SetMobjState(thing, S_GIBS);

    thing.flags := thing.flags and (not MF_SOLID);
    thing.height := 0;
    thing.radius := 0;

    // keep checking
    result := true;
    exit;
  end;

  // crunch dropped items
  if (thing.flags and MF_DROPPED) <> 0 then
  begin
    P_RemoveMobj(thing);

    // keep checking
    result := true;
    exit;
  end;

  if (thing.flags and MF_SHOOTABLE) = 0 then
  begin
    // assume it is bloody gibs or something
    result := true;
    exit;
  end;

  nofit := true;

  if crushchange and (not boolval(leveltime and 3)) then
  begin
    P_DamageMobj(thing, nil, nil, 10);

    // spray blood in a random direction
    mo := P_SpawnMobj(thing.x, thing.y, thing.z + thing.height div 2, MT_BLOOD);

    mo.momx := _SHL(P_Random - P_Random, 12);
    mo.momy := _SHL(P_Random - P_Random, 12);
  end;

  // keep checking (crush other things)
  result := true;
end;

//
// P_ChangeSector
//
function P_ChangeSector(sector: Psector_t; crunch: boolean): boolean;
var
  x: integer;
  y: integer;
begin
  nofit := false;
  crushchange := crunch;

  // re-check heights for all things near the moving sector
  for x := sector.blockbox[BOXLEFT] to sector.blockbox[BOXRIGHT] do
    for y := sector.blockbox[BOXBOTTOM] to sector.blockbox[BOXTOP] do
      P_BlockThingsIterator(x, y, PIT_ChangeSector);

  result := nofit;
end;


(* Unit r_things implementation *)


                                                                                                                                             

const
  MINZ = FRACUNIT * 4;
  BASEYCENTER = 100;

type
  maskdraw_t = record
    x1: integer;
    x2: integer;

    column: integer;
    topclip: integer;
    bottomclip: integer;
  end;
  Pmaskdraw_t = ^maskdraw_t;

//
// Sprite rotation 0 is facing the viewer,
//  rotation 1 is one angle turn CLOCKWISE around the axis.
// This is not the same as the angle,
//  which increases counter clockwise (protractor).
// There was a lot of stuff grabbed wrong, so I changed it...
//
var

  spritelights: Plighttable_tPArray;

//
// INITIALIZATION FUNCTIONS
//
const
  MAXFRAMES = 29; // Maximun number of frames in sprite

var
  sprtemp: array[0..MAXFRAMES - 1] of spriteframe_t;
  maxframe: integer;
  spritename: string;

//
// R_InstallSpriteLump
// Local function for R_InitSprites.
//
procedure R_InstallSpriteLump(lump: integer;
  frame: LongWord; rotation: LongWord; flipped: boolean);
var
  r: integer;
begin
  if (frame >= MAXFRAMES) or (rotation > 8) then
    I_Error('R_InstallSpriteLump(): Bad frame characters in lump %d', [lump]);

  if integer(frame) > maxframe then
    maxframe := frame;

  if rotation = 0 then
  begin
    // the lump should be used for all rotations
    if sprtemp[frame].rotate = 0 then
      I_Error('R_InitSprites(): Sprite %s frame %s has multip rot=0 lump',
        [spritename, Chr(Ord('A') + frame)]);

    if sprtemp[frame].rotate = 1 then
      I_Error('R_InitSprites(): Sprite %s frame %s has rotations and a rot=0 lump',
        [spritename, Chr(Ord('A') + frame)]);

    sprtemp[frame].rotate := 0;
    for r := 0 to 7 do
    begin
      sprtemp[frame].lump[r] := lump - firstspritelump;
	    sprtemp[frame].flip[r] := flipped;
    end;
    exit;
  end;

  // the lump is only used for one rotation
  if sprtemp[frame].rotate = 0 then
    I_Error('R_InitSprites(): Sprite %s frame %s has rotations and a rot=0 lump',
      [spritename, Chr(Ord('A') + frame)]);

  sprtemp[frame].rotate := 1;

  // make 0 based
  dec(rotation);
  if sprtemp[frame].lump[rotation] <> -1 then
    I_Error('R_InitSprites(): Sprite %s : %s : %s has two lumps mapped to it',
      [spritename, Chr(Ord('A') + frame), Chr(Ord('1') + rotation)]);

  sprtemp[frame].lump[rotation] := lump - firstspritelump;
  sprtemp[frame].flip[rotation] := flipped;
end;

//
// R_InitSpriteDefs
// Pass a null terminated list of sprite names
//  (4 chars exactly) to be used.
// Builds the sprite rotation matrixes to account
//  for horizontally flipped sprites.
// Will report an error if the lumps are inconsistant.
// Only called at startup.
//
// Sprite lump names are 4 characters for the actor,
//  a letter for the frame, and a number for the rotation.
// A sprite that is flippable will have an additional
//  letter/number appended.
// The rotation character can be 0 to signify no rotations.
//
procedure R_InitSpriteDefs(namelist: PsprnamesArray_t);

  procedure sprtempreset;
  var
    i: integer;
    j: integer;
  begin
    for i := 0 to MAXFRAMES - 1 do
    begin
      sprtemp[i].rotate := -1;
      for j := 0 to 7 do
      begin
        sprtemp[i].lump[j] := -1;
        sprtemp[i].flip[j] := false;
      end;
    end;
  end;

var
  i: integer;
  l: integer;
  intname: integer;
  frame: integer;
  rotation: integer;
  start: integer;
  _end: integer;
  patched: integer;
begin
  // count the number of sprite names

  r_things_numsprites := 0;
  while namelist[r_things_numsprites] <> '' do
    inc(r_things_numsprites);

  if not boolval(r_things_numsprites) then
    exit;

  sprites := Z_Malloc(r_things_numsprites * SizeOf(spritedef_t), PU_STATIC, nil);

  start := firstspritelump - 1;
  _end := lastspritelump + 1;

  // scan all the lump names for each of the names,
  //  noting the highest frame letter.
  // Just compare 4 characters as ints
  for i := 0 to r_things_numsprites - 1 do
  begin
    spritename := namelist[i];

    sprtempreset;

    maxframe := -1;
    intname := PInteger(@spritename[1])^;

{    intname := integer(Ord(spritename[1]) or
                      (Ord(spritename[2]) shl 8) or
                      (Ord(spritename[3]) shl 16) or
                      (Ord(spritename[4]) shl 24)); }

    // scan the lumps,
    //  filling in the frames for whatever is found
    for l := start + 1 to _end - 1 do
    begin
      if lumpinfo[l].v1 = intname then // VJ
      begin
        frame := Ord(lumpinfo[l].name[4]) - Ord('A');
        rotation := Ord(lumpinfo[l].name[5]) - Ord('0');

        if modifiedgame then
          patched := W_GetNumForName(lumpinfo[l].name)
        else
          patched := l;

        R_InstallSpriteLump(patched, frame, rotation, false);

//        if boolval(Ord(lumpinfo[l].name[6])) then
        if lumpinfo[l].name[6] <> #0 then
        begin
          frame := Ord(lumpinfo[l].name[6]) - Ord('A');
          rotation := Ord(lumpinfo[l].name[7]) - Ord('0');
          R_InstallSpriteLump(l, frame, rotation, true);
        end;
      end;
    end;

    // check the frames that were found for completeness
    if maxframe = -1 then
    begin
      sprites[i].numframes := 0;
	    continue;
    end;

    inc(maxframe);

    for frame := 0 to maxframe - 1 do
    begin
      case sprtemp[frame].rotate of
        -1:
          begin
            // no rotations were found for that frame at all
            I_Error('R_InitSprites(): No patches found for %s frame %s',
              [namelist[i], Chr(frame + Ord('A'))]);
          end;
         0:
          begin
            // only the first rotation is needed
          end;
         1:
          begin
            // must have all 8 frames
            for rotation := 0 to 7 do
              if sprtemp[frame].lump[rotation] = -1 then
                I_Error('R_InitSprites(): Sprite %s frame %s is missing rotations',
                  [namelist[i], Chr(frame + Ord('A'))]);
          end;
      end;
    end;

    // allocate space for the frames present and copy sprtemp to it
    sprites[i].numframes := maxframe;
    sprites[i].spriteframes :=
      Z_Malloc(maxframe * SizeOf(spriteframe_t), PU_STATIC, nil);
    memcpy(sprites[i].spriteframes, @sprtemp, maxframe * SizeOf(spriteframe_t));
  end;
end;

//
// GAME FUNCTIONS
//
var
  vissprites: array[0..MAXVISSPRITES - 1] of vissprite_t;
  vissprite_p: integer;


//
// R_InitSprites
// Called at program start.
//
procedure R_InitSprites(namelist: PsprnamesArray_t);
var
  i: integer;
begin
  for i := 0 to SCREENWIDTH - 1 do
    negonearray[i] := -1;

  R_InitSpriteDefs(namelist);
end;

//
// R_ClearSprites
// Called at frame start.
//
procedure R_ClearSprites;
begin
  vissprite_p := 0;
end;

//
// R_NewVisSprite
//
var
  overflowsprite: vissprite_t;

function R_NewVisSprite: Pvissprite_t;
begin
  if vissprite_p = MAXVISSPRITES then
    result := @overflowsprite
  else
  begin
    result := @vissprites[vissprite_p];
    inc(vissprite_p);
  end;
end;

//
// R_DrawMaskedColumn
// Used for sprites and masked mid textures.
// Masked means: partly transparent, i.e. stored
//  in posts/runs of opaque pixels.
//

procedure R_DrawMaskedColumn(column: Pcolumn_t);
var
  topscreen: integer;
  bottomscreen: integer;
  basetexturemid: fixed_t;
begin
  basetexturemid := dc_texturemid;

  while column.topdelta <> $ff do
  begin
    // calculate unclipped screen coordinates
    // for post
    topscreen := sprtopscreen + spryscale * column.topdelta;
    bottomscreen := topscreen + spryscale * column.length;

    dc_yl := (topscreen + FRACUNIT - 1) div FRACUNIT;
    dc_yh := (bottomscreen - 1) div FRACUNIT;

    if dc_yh >= mfloorclip[dc_x] then
      dc_yh := mfloorclip[dc_x] - 1;
    if dc_yl <= mceilingclip[dc_x] then
      dc_yl := mceilingclip[dc_x] + 1;

    if dc_yl <= dc_yh then
    begin
      dc_source := PByteArray(integer(column) + 3);
      dc_texturemid := basetexturemid - (column.topdelta * FRACUNIT);
	    // dc_source = (byte *)column + 3 - column->topdelta;

	    // Drawn by either R_DrawColumn
	    //  or (SHADOW) R_DrawFuzzColumn.
      colfunc;
    end;
    //column := (column_t *)(  (byte *)column + column->length + 4);
    incp(pointer(column), column.length + 4);
  end;

  dc_texturemid := basetexturemid;
end;

//
// R_DrawVisSprite
//  mfloorclip and mceilingclip should also be set.
//
procedure R_DrawVisSprite(vis: Pvissprite_t; x1: integer; x2: integer);
var
  column: Pcolumn_t;
  texturecolumn: integer;
  frac: fixed_t;
  patch: Ppatch_t;
  i: integer;
begin
  patch := W_CacheLumpNum(vis.patch + firstspritelump, PU_CACHE);

  dc_colormap := vis.colormap;

  if not boolval(dc_colormap) then
  begin
    // NULL colormap = shadow draw
    colfunc := fuzzcolfunc;
  end
  else if boolval(vis.mobjflags and MF_TRANSLATION) then
  begin
    colfunc := R_DrawTranslatedColumn;
    dc_translation := PByteArray(integer(translationtables) - 256 +
	    ( _SHR((vis.mobjflags and MF_TRANSLATION), (MF_TRANSSHIFT - 8)) ));
  end;

//  dc_iscale := _SHR(abs(vis.xiscale), detailshift);
  dc_iscale := abs(vis.xiscale);
  dc_texturemid := vis.texturemid;
  frac := vis.startfrac;
  spryscale := vis.scale;
  sprtopscreen := centeryfrac - FixedMul(dc_texturemid, spryscale);

  for i := vis.x1 to vis.x2 do
  begin
    dc_x := i;
    texturecolumn := frac div FRACUNIT;

    column := Pcolumn_t(integer(patch) + patch.columnofs[texturecolumn]);
    R_DrawMaskedColumn(column);
    frac := frac + vis.xiscale;
  end;

  colfunc := basecolfunc;
end;

procedure R_DrawPVisSprite(vis: Pvissprite_t);
var
  column: Pcolumn_t;
  texturecolumn: integer;
  frac: fixed_t;
  fracstep: fixed_t;
  patch: Ppatch_t;
  i: integer;
{  x1: integer;
  x2: integer;
  y1: integer;
  y2: integer;}
begin
  patch := W_CacheLumpNum(vis.patch + firstspritelump, PU_CACHE);

  dc_colormap := vis.colormap;

  if not boolval(dc_colormap) then
  begin
    // NULL colormap = shadow draw
    colfunc := fuzzcolfunc;
  end
  else if boolval(vis.mobjflags and MF_TRANSLATION) then
  begin
    colfunc := R_DrawTranslatedColumn;
    dc_translation := PByteArray(integer(translationtables) - 256 +
	    ( _SHR((vis.mobjflags and MF_TRANSLATION), (MF_TRANSSHIFT - 8)) ));
  end;

(************************
  dc_iscale := _SHR(abs(vis.xiscale), detailshift) * 200 div SCREENHEIGHT;
  dc_texturemid := vis.texturemid * 200 div SCREENHEIGHT;
  frac := vis.startfrac;
  fracstep := vis.xiscale * 320 div SCREENWIDTH;
  spryscale := vis.scale * SCREENHEIGHT div 200;
  sprtopscreen := centeryfrac - FixedMul(dc_texturemid, spryscale);
//  sprtopscreen := centeryfrac - FixedMul(vis.texturemid, vis.xiscale) * SCREENHEIGHT div 200;

//    topscreen := sprtopscreen + spryscale * column.topdelta;
//    bottomscreen := topscreen + spryscale * column.length;


{  x1 := vis.x1 * SCREENWIDTH div 320;
  x2 := vis.x2 * SCREENWIDTH div 320;}
  x1 := centerx - (centerx - vis.x1) * SCREENWIDTH div 320;
  x2 := centerx - (centerx - vis.x2) * SCREENWIDTH div 320;
  for i := x1 to x2 do
*********************)
//  dc_iscale := _SHR(abs(vis.xiscale), detailshift);
  dc_iscale := vis.xiscale;
  dc_texturemid := vis.texturemid;// * 200 div SCREENHEIGHT;
  frac := vis.startfrac;
  fracstep := vis.xiscale;
  spryscale := vis.scale;
  sprtopscreen := centeryfrac - FixedMul(dc_texturemid, spryscale);
//  sprtopscreen := centeryfrac - FixedMul(vis.texturemid, vis.xiscale) * SCREENHEIGHT div 200;

//    topscreen := sprtopscreen + spryscale * column.topdelta;
//    bottomscreen := topscreen + spryscale * column.length;


  for i := vis.x1 to vis.x2 do
  begin
    dc_x := i;
    texturecolumn := frac div FRACUNIT;

    column := Pcolumn_t(integer(patch) + patch.columnofs[texturecolumn]);
    R_DrawMaskedColumn(column);
    frac := frac + fracstep;
  end;

  colfunc := basecolfunc;
end;

//
// R_ProjectSprite
// Generates a vissprite for a thing
//  if it might be visible.
//
procedure R_ProjectSprite(thing: Pmobj_t);
var
  tr_x: fixed_t;
  tr_y: fixed_t;
  gxt: fixed_t;
  gyt: fixed_t;
  tx: fixed_t;
  tz: fixed_t;
  xscale: fixed_t;
  x1: integer;
  x2: integer;
  sprdef: Pspritedef_t;
  sprframe: Pspriteframe_t;
  lump: integer;
  rot: LongWord;
  flip: boolean;
  index: integer;
  vis: Pvissprite_t;
  ang: angle_t;
  iscale: fixed_t;
begin
  // transform the origin point
  tr_x := thing.x - viewx;
  tr_y := thing.y - viewy;

  gxt := FixedMul(tr_x, viewcos);
  gyt := -FixedMul(tr_y, viewsin);

  tz := gxt - gyt;

  // thing is behind view plane?
  if tz < MINZ then
    exit;

  xscale := FixedDiv(projection, tz);

  gxt := -FixedMul(tr_x, viewsin);
  gyt := FixedMul(tr_y, viewcos);
  tx := -(gyt + gxt);

  // too far off the side?
  if abs(tx) > _SHL(tz, 2) then
    exit;

  // decide which patch to use for sprite relative to player

  sprdef := @sprites[Ord(thing.sprite)];

  sprframe := @sprdef.spriteframes[thing.frame and FF_FRAMEMASK];

  if boolval(sprframe.rotate) then
  begin
    // choose a different rotation based on player view
    ang := R_PointToAngle(thing.x, thing.y);
    rot := _SHRW(ang - thing.angle + LongWord(ANG45 div 2) * 9, 29);
    lump := sprframe.lump[rot];
    flip := sprframe.flip[rot];
  end
  else
  begin
    // use single rotation for all views
    lump := sprframe.lump[0];
    flip := sprframe.flip[0];
  end;

  // calculate edges of the shape
  tx := tx - spriteoffset[lump];
  x1 := (centerxfrac + FixedMul(tx, xscale) ) div FRACUNIT;

  // off the right side?
  if x1 > viewwidth then
    exit;

  tx := tx + spritewidth[lump];
  x2 := ((centerxfrac + FixedMul(tx, xscale)) div FRACUNIT) - 1;

  // off the left side
  if x2 < 0 then
    exit;

  // store information in a vissprite
  vis := R_NewVisSprite;
  vis.mobjflags := thing.flags;
//  vis.scale := _SHL(xscale, detailshift);
  vis.scale := xscale;
  vis.gx := thing.x;
  vis.gy := thing.y;
  vis.gz := thing.z;
  vis.gzt := thing.z + spritetopoffset[lump];
  vis.texturemid := vis.gzt - viewz;
  vis.x1 := decide(x1 < 0, 0, x1);
  vis.x2 := decide(x2 >= viewwidth, viewwidth - 1, x2);
  iscale := FixedDiv(FRACUNIT, xscale);

  if flip then
  begin
    vis.startfrac := spritewidth[lump] - 1;
    vis.xiscale := -iscale;
  end
  else
  begin
    vis.startfrac := 0;
    vis.xiscale := iscale;
  end;

  if vis.x1 > x1 then
    vis.startfrac := vis.startfrac + vis.xiscale * (vis.x1 - x1);
  vis.patch := lump;

  // get light level
  if boolval(thing.flags and MF_SHADOW) then
  begin
    // shadow draw
    vis.colormap := nil;
  end
  else if boolval(fixedcolormap) then
  begin
    // fixed map
    vis.colormap := fixedcolormap;
  end
  else if boolval(thing.frame and FF_FULLBRIGHT) then
  begin
    // full bright
    vis.colormap := Plighttable_tArray(colormaps);
  end
  else
  begin
    // diminished light
    index := _SHR(xscale, LIGHTSCALESHIFT);

    if index >= MAXLIGHTSCALE then
      index := MAXLIGHTSCALE - 1;

    vis.colormap := spritelights[index];
  end;
end;

//
// R_AddSprites
// During BSP traversal, this adds sprites by sector.
//
procedure R_AddSprites(sec: Psector_t);
var
  thing: Pmobj_t;
  lightnum: integer;
begin
  // BSP is traversed by subsector.
  // A sector might have been split into several
  // subsectors during BSP building.
  // Thus we check whether its already added.
  if sec.validcount = validcount then
    exit;

  // Well, now it will be done.
  sec.validcount := validcount;

  lightnum := _SHR(sec.lightlevel, LIGHTSEGSHIFT) + extralight;

  if lightnum < 0 then
    spritelights := @scalelight[0, 0] // VJ ???
  else if lightnum >= LIGHTLEVELS then
    spritelights := @scalelight[LIGHTLEVELS - 1, 0]   // VJ ???
  else
    spritelights := @scalelight[lightnum, 0];   // VJ ???

  // Handle all things in sector.
  thing := sec.thinglist;
  while boolval(thing) do
  begin
    R_ProjectSprite(thing);
    thing := thing.snext;
  end;
end;

//
// R_DrawPSprite
//
procedure R_DrawPSprite(psp: Ppspdef_t);
var
  tx: fixed_t;
  x1: integer;
  x2: integer;
  sprdef: Pspritedef_t;
  sprframe: Pspriteframe_t;
  lump: integer;
  flip: boolean;
  vis: Pvissprite_t;
  avis: vissprite_t;
begin
  // decide which patch to use

  sprdef := @sprites[Ord(psp.state.sprite)];

  sprframe := @sprdef.spriteframes[psp.state.frame and FF_FRAMEMASK];

  lump := sprframe.lump[0];
  flip := sprframe.flip[0];

  // calculate edges of the shape
  tx := psp.sx - 160 * FRACUNIT;

  tx := tx - spriteoffset[lump];
  x1 := (centerxfrac + FixedMul(tx, pspritescale)) div FRACUNIT;

  // off the right side
  if x1 > viewwidth then
    exit;

  tx := tx + spritewidth[lump];
  x2 := ((centerxfrac + FixedMul(tx, pspritescale)) div FRACUNIT) - 1;

  // off the left side
  if x2 < 0 then
    exit;

  // store information in a vissprite
  vis := @avis;
  vis.mobjflags := 0;
  vis.texturemid := (BASEYCENTER * FRACUNIT) + FRACUNIT div 2 - (psp.sy - spritetopoffset[lump]);
  vis.x1 := decide(x1 < 0, 0, x1);
  vis.x2 := decide(x2 >= viewwidth, viewwidth - 1, x2);
  vis.scale := pspritescale;

  if flip then
  begin
    vis.xiscale := -pspriteiscale;
    vis.startfrac := spritewidth[lump] - 1;
  end
  else
  begin
    vis.xiscale := pspriteiscale;
    vis.startfrac := 0;
  end;

  if vis.x1 > x1 then
    vis.startfrac := vis.startfrac + vis.xiscale * (vis.x1 - x1);

  vis.patch := lump;

  if (viewplayer.powers[Ord(pw_invisibility)] > 4 * 32) or
     (boolval(viewplayer.powers[Ord(pw_invisibility)] and 8)) then
  begin
    // shadow draw
    vis.colormap := nil;
  end
  else if boolval(fixedcolormap) then
  begin
    // fixed color
    vis.colormap := fixedcolormap;
  end
  else if boolval(psp.state.frame and FF_FULLBRIGHT) then
  begin
    // full bright
    vis.colormap := colormaps;
  end
  else
  begin
    // local light
    vis.colormap := spritelights[MAXLIGHTSCALE - 1];
  end;

  R_DrawPVisSprite(vis);
end;

//
// R_DrawPlayerSprites
//
procedure R_DrawPlayerSprites;
var
  i: integer;
  lightnum: integer;
begin
  // get light level
  lightnum :=
    _SHR(Psubsector_t(viewplayer.mo.subsector).sector.lightlevel, LIGHTSEGSHIFT) +
      extralight;

  if lightnum < 0 then
    spritelights := @scalelight[0, 0]
  else if lightnum >= LIGHTLEVELS then
    spritelights := @scalelight[LIGHTLEVELS - 1, 0]
  else
    spritelights := @scalelight[lightnum, 0];

  // clip to screen bounds
  mfloorclip := @screenheightarray;
  mceilingclip := @negonearray;

  // add all active psprites
  for i := 0 to Ord(NUMPSPRITES) - 1 do
  begin
    if boolval(viewplayer.psprites[i].state) then
      R_DrawPSprite(@viewplayer.psprites[i]);
  end;
end;

//
// R_SortVisSprites
//
var
  vsprsortedhead: vissprite_t;

procedure R_SortVisSprites;
var
  i: integer;
  count: integer;
  ds: Pvissprite_t;
  best: Pvissprite_t;
  unsorted: vissprite_t;
  bestscale: fixed_t;
begin
  count := vissprite_p;

  if not boolval(count) then
    exit;

  unsorted.next := @unsorted;
  unsorted.prev := @unsorted;

  vissprites[0].next := @vissprites[1];
  vissprites[0].prev := @unsorted;

  for i := 1 to count - 1 do
  begin
    vissprites[i].next := @vissprites[i + 1];
    vissprites[i].prev := @vissprites[i - 1];
  end;

  unsorted.prev := @vissprites[vissprite_p - 1];
  unsorted.next := @vissprites[0];
  vissprites[vissprite_p - 1].next := @unsorted;

  // pull the vissprites out by scale
  //best = 0;		// shut up the compiler warning
  vsprsortedhead.next := @vsprsortedhead;
  vsprsortedhead.prev := @vsprsortedhead;
  for i := 0 to count - 1 do
  begin
    bestscale := MAXINT;
    ds := unsorted.next;
    best := nil; // VJ - > avoid compiler warning
    while ds <> @unsorted do
    begin
      if ds.scale < bestscale then
      begin
        bestscale := ds.scale;
        best := ds;
      end;
      ds := ds.next;
    end;

    if best <> nil then // VJ - > avoid compiler warning
    begin
      best.next.prev := best.prev;
      best.prev.next := best.next;
      best.next := @vsprsortedhead;
      best.prev := vsprsortedhead.prev;
      vsprsortedhead.prev.next := best;
      vsprsortedhead.prev := best;
    end;
  end;
end;

//
// R_DrawSprite
//

var
  clipbot: packed array[0..MAXWIDTH - 1] of smallint;
  cliptop: packed array[0..MAXWIDTH - 1] of smallint;

procedure R_DrawSprite(spr: Pvissprite_t);
var
  ds: Pdrawseg_t;
  x: integer;
  r1: integer;
  r2: integer;
  scale: fixed_t;
  lowscale: fixed_t;
  silhouette: integer;
  i: integer;
begin
  for x := spr.x1 to spr.x2 do
  begin
    clipbot[x] := -2;
    cliptop[x] := -2;
  end;

  // Scan drawsegs from end to start for obscuring segs.
  // The first drawseg that has a greater scale
  //  is the clip seg.
  for i := ds_p - 1 downto 0 do
  begin
//    ds := @drawsegs[i];
    ds := drawsegs[i]; // test
    // determine if the drawseg obscures the sprite
    if (ds.x1 > spr.x2) or
       (ds.x2 < spr.x1) or
       ((not boolval(ds.silhouette)) and (not boolval(ds.maskedtexturecol))) then
    begin
      // does not cover sprite
      continue;
    end;

    r1 := decide(ds.x1 < spr.x1, spr.x1, ds.x1);
    r2 := decide(ds.x2 > spr.x2, spr.x2, ds.x2);

    if ds.scale1 > ds.scale2 then
    begin
      lowscale := ds.scale2;
      scale := ds.scale1;
    end
    else
    begin
      lowscale := ds.scale1;
      scale := ds.scale2;
    end;

    if (scale < spr.scale) or
	     ((lowscale < spr.scale) and (not R_PointOnSegSide(spr.gx, spr.gy, ds.curline))) then
    begin
      // masked mid texture?
      if boolval(ds.maskedtexturecol) then
        R_RenderMaskedSegRange(ds, r1, r2);
      // seg is behind sprite
      continue;
    end;

    // clip this piece of the sprite
    silhouette := ds.silhouette;

    if spr.gz >= ds.bsilheight then
      silhouette := silhouette and (not SIL_BOTTOM);

    if spr.gzt <= ds.tsilheight then
      silhouette := silhouette and (not SIL_TOP);

    if silhouette = 1 then
    begin
      // bottom sil
      for x := r1 to r2 do
        if clipbot[x] = -2 then
          clipbot[x] := ds.sprbottomclip[x];
    end
    else if silhouette = 2 then
    begin
      // top sil
      for x := r1 to r2 do
        if cliptop[x] = -2 then
          cliptop[x] := ds.sprtopclip[x];
    end
    else if silhouette = 3 then
    begin
      // both
      for x := r1 to r2 do
      begin
        if clipbot[x] = -2 then
          clipbot[x] := ds.sprbottomclip[x];
        if cliptop[x] = -2 then
          cliptop[x] := ds.sprtopclip[x];
      end;
    end;
  end;

  // all clipping has been performed, so draw the sprite

  // check for unclipped columns
  for x := spr.x1 to spr.x2 do
  begin
    if clipbot[x] = -2 then
      clipbot[x] := viewheight;

    if cliptop[x] = -2 then
      cliptop[x] := -1;
  end;

  mfloorclip := @clipbot;
  mceilingclip := @cliptop;
  R_DrawVisSprite(spr, spr.x1, spr.x2);
end;

//
// R_DrawMasked
//
procedure R_DrawMasked;
var
  spr: Pvissprite_t;
  ds: Pdrawseg_t;
  i: integer;
begin
  R_SortVisSprites;

  if vissprite_p > 0 then
  begin
    // draw all vissprites back to front
    spr := vsprsortedhead.next;
    while spr <> @vsprsortedhead do
    begin
      R_DrawSprite(spr);
      spr := spr.next;
    end;
  end;

  // render any remaining masked mid textures
  for i := ds_p - 1 downto 0 do
  begin
//    ds := @drawsegs[i];
    ds := drawsegs[i]; // test
    if boolval(ds.maskedtexturecol) then
      R_RenderMaskedSegRange(ds, ds.x1, ds.x2);
  end;

  // draw the psprites on top of everything
  //  but does not draw on side views
  if not boolval(viewangleoffset) then
    R_DrawPlayerSprites;
end;



(* Unit i_net implementation *)


                                                                                                                     

const
  IPPORT_USERRESERVED = 5000;

type
  socklen_t = integer;

//
// NETWORKING
//
var
  DOOMPORT: word;
  mysocket: TSocket;
  sendaddress: array[0..MAXNETNODES - 1] of sockaddr_in;
  sendplayer: array[0..MAXNETNODES] of byte;

var
  netget: PProcedure;
  netsend: PProcedure;

var
  sendsocket: integer;
  insocket: integer;

function neterror: string; forward;

const
  PRE_CONNECT = 0;
  PRE_DISCONNECT = 1;
  PRE_ALLHERE = 2;
  PRE_CONACK = 3;
  PRE_ALLHEREACK = 4;
  PRE_GO = 5;

// Set PreGamePacket.fake to this so that the game rejects any pregame packets
// after it starts. This translates to NCMD_SETUP|NCMD_MULTI.
  PRE_FAKE = $30;

type
  machine_t = record
    address: LongWord;
    port: word;
    player: byte;
    pad: byte;
  end;

  PreGamePacket_t = record
    fake: byte;
    _message: byte;
    numnodes: byte;
    consolenum: byte;
    machines: array[0..MAXNETNODES - 1] of machine_t;
  end;

//
// UDPsocket
//
function UDPsocket: TSocket;
begin
  // allocate a socket
  result := socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if result = INVALID_SOCKET then
    I_Error('UDPsocket(): Can not create socket: %s', [neterror]);
end;

//
// BindToLocalPort
//
procedure BindToLocalPort(s: TSocket; port: word);
var
  v: integer;
  address: sockaddr_in;
begin
  memset(@address, 0, SizeOf(address));
  address.sin_family := AF_INET;
  address.sin_addr.s_addr := INADDR_ANY;
  address.sin_port := htons(port);

  v := bind(s, address, SizeOf(address));
  if v = SOCKET_ERROR then
    I_Error('BindToLocalPort(): %s', [neterror]);
end;

function FindNode(const address: sockaddr_in): integer;
var
  i: integer;
begin
  // find remote node number
  i := 0;
  while i < doomcom.numnodes do
  begin
    if (address.sin_addr.s_addr = sendaddress[i].sin_addr.s_addr) and
       (address.sin_port = sendaddress[i].sin_port) then
      break;
    inc(i);
  end;

  if i = doomcom.numnodes then
    // packet is not from one of the players (new game broadcast?)
    i := -1;

  result := i;
end;

//
// PacketSend
//
procedure PacketSend;
var
  c: integer;
begin
  //printf ("sending %i\n",gametic);
  c := sendto(mysocket, doomcom.data, doomcom.datalength,
        0, sendaddress[doomcom.remotenode], SizeOf(sendaddress[doomcom.remotenode]));

  if c = -1 then
    I_Error('PacketSend(): sendto() returned -1');
end;

const
  I_NET_BACKUPTICS = 36;
  MAX_MSGLEN = I_NET_BACKUPTICS * 10;

//
// PacketGet
//
procedure PacketGet;
var
  c: integer;
  fromlen: socklen_t;
  fromaddress: sockaddr_in;
  node: integer;
  err: integer;
begin
  fromlen := SizeOf(fromaddress);
  c := recvfrom(mysocket, doomcom.data, MAX_MSGLEN, 0, fromaddress, fromlen);
  node := FindNode(fromaddress);

  if c = SOCKET_ERROR then
  begin
    err := WSAGetLastError;

    if (err = WSAECONNRESET) and (node >= 0) then
    begin
      // The remote node aborted unexpectedly, so pretend it sent an exit packet
      printf('The connection from player %d was dropped' + #13#10,
        [sendplayer[node]]);

// VJ      doomcom.data[0] = $80;  // NCMD_EXIT
      c := 1;
    end
    else if err <> WSAEWOULDBLOCK then
    begin
      I_Error('PacketGet(): %s', [neterror]);
    end
    else
    begin
      doomcom.remotenode := -1; // no packet
      exit;
    end;
  end;

  doomcom.remotenode := node;
  doomcom.datalength := c;
end;

var
  fromaddress_pg: sockaddr_in;

function PreGet(buffer: pointer; bufferlen: integer; noabort: boolean): PSOCKADDR;
var
  fromlen: socklen_t;
  c: integer;
  err: integer;
begin
  fromlen := SizeOf(fromaddress_pg);
  c := recvfrom(mysocket, buffer^, bufferlen, 0, fromaddress_pg, fromlen);

  if c = SOCKET_ERROR then
  begin
    err := WSAGetLastError;
    if (err = WSAEWOULDBLOCK) or (noabort and (err = WSAECONNRESET)) then
    begin
      result := nil; // no packet
      exit;
    end;
    I_Error('PreGet(): %s', [neterror]);
  end;
  result := @fromaddress_pg;
end;

procedure PreSend(buffer: pointer; bufferlen: integer; _to: PSOCKADDR);
begin
  sendto(mysocket, buffer^, bufferlen, 0, _to^, SizeOf(_to^));
end;

procedure BuildAddress(address: PSOCKADDR; var name: string);
var
  hostentry: PHostEnt;  // host information entry
  port: word;
  portpart: string;
  isnamed: boolean;
  _pos: integer;
  i: integer;
begin
  isnamed := false;

  address.sin_family := AF_INET;

  _pos := Pos(':', name);
  if _pos > 0 then
  begin
    portpart := '';
    for i := _pos + 1 to Length(name) do
      portpart := portpart + name[i];
    port := atoi(portpart);
    if port = 0 then
    begin
      printf('Weird port: %s (using %d)' + #13#10, [portpart, DOOMPORT]);
      port := DOOMPORT;
    end
  end
  else
    port := DOOMPORT;

  address.sin_port := htons(port);

  for i := 1 to Length(name) do
  begin
    if not (name[i] in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.']) then
    begin
      isnamed := true;
      break;
    end;
  end;

  if not isnamed then
  begin
    address.sin_addr.s_addr := inet_addr(PChar(name));
    printf('Node number %d address %s' + #13#10, [doomcom.numnodes, name]);
  end
  else
  begin
    hostentry := gethostbyname(PChar(name));
    if hostentry = nil then
      I_Error('gethostbyname(): Could not find %s' + #13#10 + '%s',
        [name, neterror]);
    address.sin_addr.s_addr := LongWord(hostentry.h_addr_list);
    printf('Node number %d hostname %s' + #13#10,
      [doomcom.numnodes, StringVal(hostentry.h_name)]);
  end;
end;

procedure CloseNetwork;
begin
  if mysocket <> INVALID_SOCKET then
  begin
    closesocket(mysocket);
    mysocket := INVALID_SOCKET;
  end;
  WSACleanup;
end;

procedure StartNetwork(autoPort: boolean);
var
  trueval: integer;
  wsad: WSADATA;
begin
  trueval := 1;
  if WSAStartup($0101, wsad) <> 0 then
    I_Error('StartNetwork(): Could not initialize Windows Sockets');

//  atterm (CloseNetwork);
  CloseNetwork;

  netsend := PacketSend;
  netget := PacketGet;
  netgame := true;
//  multiplayer := true; // VJ -> removed

  // create communication socket
  mysocket := UDPsocket;
  if autoPort then
    BindToLocalPort(mysocket, 0)
  else
    BindToLocalPort(mysocket, DOOMPORT);
  ioctlsocket(mysocket, FIONBIO, trueval);
end;

procedure WaitForPlayers(i: integer);
begin
  if i = myargc then
    I_Error('WaitForPlayers(): Not enough parameters after -net');

  StartNetwork(false);

  // parse player number and host list
  doomcom.consoleplayer := Ord(myargv[i + 1][1]) - Ord('1');
  printf('Console player number: %d' + #13#10, [doomcom.consoleplayer]);

  doomcom.numnodes := 1;    // this node for sure

  inc(i, 2);
  while (i <= myargc) and (myargv[i][1] <> '-') and (myargv[i][1] <> '+') do
  begin
    BuildAddress(@sendaddress[doomcom.numnodes], myargv[i]);
    inc(doomcom.numnodes);
    inc(i);
  end;

  printf('Total players: %d' + #13#10, [doomcom.numnodes]);

  doomcom.id := DOOMCOM_ID;
  doomcom.numplayers := doomcom.numnodes;
end;

procedure SendAbort;
var
  dis: array[0..1] of byte;
begin
  dis[0] := PRE_FAKE;
  dis[1] := PRE_DISCONNECT;

  dec(doomcom.numnodes);
  while doomcom.numnodes > 0 do
  begin
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    PreSend(@dis, 2, @sendaddress[doomcom.numnodes]);
    dec(doomcom.numnodes);
  end;
end;


procedure I_InitNetwork;
var
  trueval: integer;
  i: integer;
  j: integer;
  p: integer;
  hostentry: PHostEnt; // host information entry
  addr: string;
begin
  trueval := 1;
  doomcom := malloc(SizeOf(doomcom^));

  // set up for network
  i := M_CheckParm('-dup');
  if (i <> 0) and (i < myargc - 1) then
  begin
    doomcom.ticdup := Ord(myargv[i + 1][1]) - Ord('0');
    if doomcom.ticdup < 1 then
      doomcom.ticdup := 1
    else if doomcom.ticdup > 9 then
      doomcom.ticdup := 9;
  end
  else
    doomcom.ticdup := 1;

  if M_CheckParm('-extratic') <> 0 then
    doomcom.extratics := 1
  else
    doomcom.extratics := 0;

  p := M_CheckParm('-port');
  if (p <> 0) and (p < myargc - 1) then
  begin
    DOOMPORT := atoi(myargv[p + 1]);
    printf('using alternate port %d' + #13#10, [DOOMPORT]);
  end;

  // parse network game options,
  //  -net <consoleplayer> <host> <host> ...
  i := M_CheckParm('-net');
  if (i = 0) or (i = myargc) then
  begin
    // single player game
    netgame := false;
    doomcom.id := DOOMCOM_ID;
    doomcom.numplayers := 1;
    doomcom.numnodes := 1;
    doomcom.deathmatch := 0;
    doomcom.consoleplayer := 0;
    exit;
  end;

  netsend := PacketSend;
  netget := PacketGet;
  netgame := true;

  // parse player number and host list
  if i < myargc - 1 then
    doomcom.consoleplayer := Ord(myargv[i + 1][1]) - Ord('1')
  else
    I_Error('I_InitNetwork(): You must specify player number after -net');

  doomcom.numnodes := 1; // this node for sure

  inc(i);
  while (i < myargc - 1) and (myargv[i + 1][1] <> '-') do
  begin
    inc(i);
    sendaddress[doomcom.numnodes].sin_family := AF_INET;
    sendaddress[doomcom.numnodes].sin_port := htons(DOOMPORT);
    if myargv[i][1] = '.' then
    begin
      addr := '';
      for j := 2 to Length(myargv[i]) do
        addr := addr + myargv[i][j];
      sendaddress[doomcom.numnodes].sin_addr.s_addr := inet_addr(PChar(addr));
    end
    else
    begin
      hostentry := gethostbyname(PChar(myargv[i]));
      if hostentry = nil then
        I_Error('I_InitNetwork(): gethostbyname: couldn''t find %s', [myargv[i]]);
      sendaddress[doomcom.numnodes].sin_addr.s_addr := PLongWord(hostentry.h_addr_list)^;
    end;
    doomcom.numnodes := doomcom.numnodes + 1;
  end;

  doomcom.id := DOOMCOM_ID;
  doomcom.numplayers := doomcom.numnodes;

  // build message to receive
  insocket := UDPsocket;
  BindToLocalPort(insocket, htons(DOOMPORT));
  ioctlsocket(insocket, FIONBIO, trueval);

  sendsocket := UDPsocket;
end;

procedure I_NetCmd;
begin
  if doomcom.command = CMD_SEND then
    netsend
  else if doomcom.command = CMD_GET then
    netget
  else
    I_Error('I_NetCmd(): Bad net cmd: %d', [doomcom.command]);
end;

function neterror: string;
begin
  result := '';
end;




initialization

(* Unit m_rnd initialization *)


  M_ClearRandom;


(* Unit i_midi initialization *)

  Window := 0;
  MidiFileNames := TStringList.Create;
  MidiID := 0;
  fIsPlaying := false;


(* Unit tables initialization *)

  finecosine := Pfixed_tArray(@finesine[FINEANGLES div 4]);


(* Unit m_argv initialization *)

{$IFDEF DEBUG}
  cmdparams := TStringList.Create;
{$ENDIF}


(* Unit s_sound initialization *)

  snd_SfxVolume := 15;
  snd_MusicVolume := 15;

  mus_playing := nil;



(* Unit i_sound initialization *)

  pDS := nil;
  pDSBPrimary := nil;
  HandleCount := 1;


(* Unit dstrings initialization *)


  // DOOM1
  endmsg[0] := QUITMSG;
  endmsg[1] := 'please don''t leave, there''s more' + #13#10 + 'demons to toast!';
  endmsg[2] := 'let''s beat it -- this is turning' + #13#10 + 'into a bloodbath!';
  endmsg[3] := 'i wouldn''t leave if i were you.' + #13#10 + 'dos is much worse.';
  endmsg[4] := 'you''re trying to say you like dos' + #13#10 + 'better than me, right?';
  endmsg[5] := 'don''t leave yet -- there''s a' + #13#10 + 'demon around that corner!';
  endmsg[6] := 'ya know, next time you come in here' + #13#10 + 'i''m gonna toast ya.';
  endmsg[7] := 'go ahead and leave. see if i care.';

  // QuitDOOM II messages
  endmsg[8] := 'you want to quit?' + #13#10 + 'then, thou hast lost an eighth!';
  endmsg[9] := 'don''t go now, there''s a ' + #13#10 + 'dimensional shambler waiting' + #13#10 + 'at the dos prompt!';
  endmsg[10] := 'get outta here and go back' + #13#10 + 'to your boring programs.';
  endmsg[11] := 'if i were your boss, i''d ' + #13#10 + ' deathmatch ya in a minute!';
  endmsg[12] := 'look, bud. you leave now' + #13#10 + 'and you forfeit your body count!';
  endmsg[13] := 'just leave. when you come' + #13#10 + 'back, i''ll be waiting with a bat.';
  endmsg[14] := 'you''re lucky i don''t smack' + #13#10 + 'you for thinking about leaving.';

  // FinalDOOM?
  endmsg[15] := 'fuck you, pussy!' + #13#10 + 'get the fuck out!';
  endmsg[16] := 'you quit and i''ll jizz' + #13#10 + 'in your cystholes!';
  endmsg[17] := 'if you leave, i''ll make' + #13#10 + 'the lord drink my jizz.';
  endmsg[18] := 'hey, ron! can we say' + #13#10 + '''fuck'' in the game?';
  endmsg[19] := 'i''d leave: this is just' + #13#10 + 'more monsters and levels.' + #13#10 + 'what a load.';
  endmsg[20] := 'suck it down, asshole!' + #13#10 + 'you''re a fucking wimp!';
  endmsg[21] := 'don''t quit now! we''re ' + #13#10 + 'still spending your money!';

  // Internal debug. Different style, too.
  endmsg[22] := 'THIS IS NO MESSAGE!' + #13#10 + 'Page intentionally left blank.';


(* Unit i_system initialization *)

  basetime := 0;

  if not QueryPerformanceFrequency(Freq) then
    Freq := 1000;


(* Unit f_finale initialization *)

  castorder[0].name := CC_ZOMBIE;
  castorder[0]._type := MT_POSSESSED;

  castorder[1].name := CC_SHOTGUN;
  castorder[1]._type := MT_SHOTGUY;

  castorder[2].name := CC_HEAVY;
  castorder[2]._type := MT_CHAINGUY;

  castorder[3].name := CC_IMP;
  castorder[3]._type := MT_TROOP;

  castorder[4].name := CC_DEMON;
  castorder[4]._type := MT_SERGEANT;

  castorder[5].name := CC_LOST;
  castorder[5]._type := MT_SKULL;

  castorder[6].name := CC_CACO;
  castorder[6]._type := MT_HEAD;

  castorder[7].name := CC_HELL;
  castorder[7]._type := MT_KNIGHT;

  castorder[8].name := CC_BARON;
  castorder[8]._type := MT_BRUISER;

  castorder[9].name := CC_ARACH;
  castorder[9]._type := MT_BABY;

  castorder[10].name := CC_PAIN;
  castorder[10]._type := MT_PAIN;

  castorder[11].name := CC_REVEN;
  castorder[11]._type := MT_UNDEAD;

  castorder[12].name := CC_MANCU;
  castorder[12]._type := MT_FATSO;

  castorder[13].name := CC_ARCH;
  castorder[13]._type := MT_VILE;

  castorder[14].name := CC_SPIDER;
  castorder[14]._type := MT_SPIDER;

  castorder[15].name := CC_CYBER;
  castorder[15]._type := MT_CYBORG;

  castorder[16].name := CC_HERO;
  castorder[16]._type := MT_PLAYER;

  castorder[17].name := '';
  castorder[17]._type := mobjtype_t(0);


  laststage := 0;


(* Unit doomdef initialization *)



(* Unit doomstat initialization *)

  gamemode := indetermined;
  gamemission := doom;
  language := english;
  modifiedgame := false;


(* Unit st_stuff initialization *)

////////////////////////////////////////////////////////////////////////////////
  veryfirsttime := 1;
  st_msgcounter := 0;
  st_oldhealth := -1;
  st_facecount := 0;
  st_faceindex := 0;

////////////////////////////////////////////////////////////////////////////////
// Now what?
  cheat_mus.sequence := get_cheatseq_string(cheat_mus_seq);
  cheat_mus.p := get_cheatseq_string(0);
  cheat_god.sequence := get_cheatseq_string(cheat_god_seq);
  cheat_god.p := get_cheatseq_string(0);
  cheat_ammo.sequence := get_cheatseq_string(cheat_ammo_seq);
  cheat_ammo.p := get_cheatseq_string(0);
  cheat_ammonokey.sequence := get_cheatseq_string(cheat_ammonokey_seq);
  cheat_ammonokey.p := get_cheatseq_string(0);
  cheat_noclip.sequence := get_cheatseq_string(cheat_noclip_seq);
  cheat_noclip.p := get_cheatseq_string(0);
  cheat_commercial_noclip.sequence := get_cheatseq_string(cheat_commercial_noclip_seq);
  cheat_commercial_noclip.p := get_cheatseq_string(0);

  cheat_powerup[0].sequence := get_cheatseq_string(cheat_powerup_seq0);
  cheat_powerup[0].p := get_cheatseq_string(0);
  cheat_powerup[1].sequence := get_cheatseq_string(cheat_powerup_seq1);
  cheat_powerup[1].p := get_cheatseq_string(0);
  cheat_powerup[2].sequence := get_cheatseq_string(cheat_powerup_seq2);
  cheat_powerup[2].p := get_cheatseq_string(0);
  cheat_powerup[3].sequence := get_cheatseq_string(cheat_powerup_seq3);
  cheat_powerup[3].p := get_cheatseq_string(0);
  cheat_powerup[4].sequence := get_cheatseq_string(cheat_powerup_seq4);
  cheat_powerup[4].p := get_cheatseq_string(0);
  cheat_powerup[5].sequence := get_cheatseq_string(cheat_powerup_seq5);
  cheat_powerup[5].p := get_cheatseq_string(0);
  cheat_powerup[6].sequence := get_cheatseq_string(cheat_powerup_seq6);
  cheat_powerup[6].p := get_cheatseq_string(0);

  cheat_choppers.sequence := get_cheatseq_string(cheat_choppers_seq);
  cheat_choppers.p := get_cheatseq_string(0);
  cheat_clev.sequence := get_cheatseq_string(cheat_clev_seq);
  cheat_clev.p := get_cheatseq_string(0);
  cheat_mypos.sequence := get_cheatseq_string(cheat_mypos_seq);
  cheat_mypos.p := get_cheatseq_string(0);

////////////////////////////////////////////////////////////////////////////////
  lastcalc := 0;
  oldhealth := -1;

////////////////////////////////////////////////////////////////////////////////
  lastattackdown := -1;
  priority := 0;

////////////////////////////////////////////////////////////////////////////////
  largeammo := 1994; // means "n/a"

////////////////////////////////////////////////////////////////////////////////
  st_palette := 0;

////////////////////////////////////////////////////////////////////////////////
  st_stopped := true;


(* Unit d_items initialization *)

//
// PSPRITE ACTIONS for waepons.
// This struct controls the weapon animations.
//
// Each entry is:
//   ammo/amunition type
//  upstate
//  downstate
// readystate
// atkstate, i.e. attack/fire/hit frame
// flashstate, muzzle flash
//

  // fist
  weaponinfo[0].ammo := am_noammo;
  weaponinfo[0].upstate := Ord(S_PUNCHUP);
  weaponinfo[0].downstate := Ord(S_PUNCHDOWN);
  weaponinfo[0].readystate := Ord(S_PUNCH);
  weaponinfo[0].atkstate := Ord(S_PUNCH1);
  weaponinfo[0].flashstate := Ord(S_NULL);

  // pistol
  weaponinfo[1].ammo := am_clip;
  weaponinfo[1].upstate := Ord(S_PISTOLUP);
  weaponinfo[1].downstate := Ord(S_PISTOLDOWN);
  weaponinfo[1].readystate := Ord(S_PISTOL);
  weaponinfo[1].atkstate := Ord(S_PISTOL1);
  weaponinfo[1].flashstate := Ord(S_PISTOLFLASH);

  // shotgun
  weaponinfo[2].ammo := am_shell;
  weaponinfo[2].upstate := Ord(S_SGUNUP);
  weaponinfo[2].downstate := Ord(S_SGUNDOWN);
  weaponinfo[2].readystate := Ord(S_SGUN);
  weaponinfo[2].atkstate := Ord(S_SGUN1);
  weaponinfo[2].flashstate := Ord(S_SGUNFLASH1);

  // chaingun
  weaponinfo[3].ammo := am_clip;
  weaponinfo[3].upstate := Ord(S_CHAINUP);
  weaponinfo[3].downstate := Ord(S_CHAINDOWN);
  weaponinfo[3].readystate := Ord(S_CHAIN);
  weaponinfo[3].atkstate := Ord(S_CHAIN1);
  weaponinfo[3].flashstate := Ord(S_CHAINFLASH1);

  // missile launcher
  weaponinfo[4].ammo := am_misl;
  weaponinfo[4].upstate := Ord(S_MISSILEUP);
  weaponinfo[4].downstate := Ord(S_MISSILEDOWN);
  weaponinfo[4].readystate := Ord(S_MISSILE);
  weaponinfo[4].atkstate := Ord(S_MISSILE1);
  weaponinfo[4].flashstate := Ord(S_MISSILEFLASH1);

  // plasma rifle
  weaponinfo[5].ammo := am_cell;
  weaponinfo[5].upstate := Ord(S_PLASMAUP);
  weaponinfo[5].downstate := Ord(S_PLASMADOWN);
  weaponinfo[5].readystate := Ord(S_PLASMA);
  weaponinfo[5].atkstate := Ord(S_PLASMA1);
  weaponinfo[5].flashstate := Ord(S_PLASMAFLASH1);

  // bfg 9000
  weaponinfo[6].ammo := am_cell;
  weaponinfo[6].upstate := Ord(S_BFGUP);
  weaponinfo[6].downstate := Ord(S_BFGDOWN);
  weaponinfo[6].readystate := Ord(S_BFG);
  weaponinfo[6].atkstate := Ord(S_BFG1);
  weaponinfo[6].flashstate := Ord(S_BFGFLASH1);

  // chainsaw
  weaponinfo[7].ammo := am_noammo;
  weaponinfo[7].upstate := Ord(S_SAWUP);
  weaponinfo[7].downstate := Ord(S_SAWDOWN);
  weaponinfo[7].readystate := Ord(S_SAW);
  weaponinfo[7].atkstate := Ord(S_SAW1);
  weaponinfo[7].flashstate := Ord(S_NULL);

  // super shotgun
  weaponinfo[8].ammo := am_shell;
  weaponinfo[8].upstate := Ord(S_DSGUNUP);
  weaponinfo[8].downstate := Ord(S_DSGUNDOWN);
  weaponinfo[8].readystate := Ord(S_DSGUN);
  weaponinfo[8].atkstate := Ord(S_DSGUN1);
  weaponinfo[8].flashstate := Ord(S_DSGUNFLASH1);


(* Unit hu_lib initialization *)

  lastautomapactive := true;


(* Unit hu_stuff initialization *)

  chat_macros[0] := HUSTR_CHATMACRO0;
  chat_macros[1] := HUSTR_CHATMACRO1;
  chat_macros[2] := HUSTR_CHATMACRO2;
  chat_macros[3] := HUSTR_CHATMACRO3;
  chat_macros[4] := HUSTR_CHATMACRO4;
  chat_macros[5] := HUSTR_CHATMACRO5;
  chat_macros[6] := HUSTR_CHATMACRO6;
  chat_macros[7] := HUSTR_CHATMACRO7;
  chat_macros[8] := HUSTR_CHATMACRO8;
  chat_macros[9] := HUSTR_CHATMACRO9;

  player_names[0] := HUSTR_PLRGREEN;
  player_names[1] := HUSTR_PLRINDIGO;
  player_names[2] := HUSTR_PLRBROWN;
  player_names[3] := HUSTR_PLRRED;

  always_off := false;

////////////////////////////////////////////////////////////////////////////////
//
// Builtin map names.
// The actual names can be found in DStrings.h.
//
////////////////////////////////////////////////////////////////////////////////

// DOOM shareware/registered/retail (Ultimate) names.

  mapnames[0] := HUSTR_E1M1;
  mapnames[1] := HUSTR_E1M2;
  mapnames[2] := HUSTR_E1M3;
  mapnames[3] := HUSTR_E1M4;
  mapnames[4] := HUSTR_E1M5;
  mapnames[5] := HUSTR_E1M6;
  mapnames[6] := HUSTR_E1M7;
  mapnames[7] := HUSTR_E1M8;
  mapnames[8] := HUSTR_E1M9;

  mapnames[9] := HUSTR_E2M1;
  mapnames[10] := HUSTR_E2M2;
  mapnames[11] := HUSTR_E2M3;
  mapnames[12] := HUSTR_E2M4;
  mapnames[13] := HUSTR_E2M5;
  mapnames[14] := HUSTR_E2M6;
  mapnames[15] := HUSTR_E2M7;
  mapnames[16] := HUSTR_E2M8;
  mapnames[17] := HUSTR_E2M9;

  mapnames[18] := HUSTR_E3M1;
  mapnames[19] := HUSTR_E3M2;
  mapnames[20] := HUSTR_E3M3;
  mapnames[21] := HUSTR_E3M4;
  mapnames[22] := HUSTR_E3M5;
  mapnames[23] := HUSTR_E3M6;
  mapnames[24] := HUSTR_E3M7;
  mapnames[25] := HUSTR_E3M8;
  mapnames[26] := HUSTR_E3M9;

  mapnames[27] := HUSTR_E4M1;
  mapnames[28] := HUSTR_E4M2;
  mapnames[29] := HUSTR_E4M3;
  mapnames[30] := HUSTR_E4M4;
  mapnames[31] := HUSTR_E4M5;
  mapnames[32] := HUSTR_E4M6;
  mapnames[33] := HUSTR_E4M7;
  mapnames[34] := HUSTR_E4M8;
  mapnames[35] := HUSTR_E4M9;

  mapnames[36] := 'NEWLEVEL';
  mapnames[37] := mapnames[36];
  mapnames[38] := mapnames[36];
  mapnames[39] := mapnames[36];
  mapnames[40] := mapnames[36];
  mapnames[41] := mapnames[36];
  mapnames[42] := mapnames[36];
  mapnames[43] := mapnames[36];
  mapnames[44] := mapnames[36];

////////////////////////////////////////////////////////////////////////////////

// DOOM 2 map names.

  mapnames2[0] := HUSTR_1;
  mapnames2[1] := HUSTR_2;
  mapnames2[2] := HUSTR_3;
  mapnames2[3] := HUSTR_4;
  mapnames2[4] := HUSTR_5;
  mapnames2[5] := HUSTR_6;
  mapnames2[6] := HUSTR_7;
  mapnames2[7] := HUSTR_8;
  mapnames2[8] := HUSTR_9;
  mapnames2[9] := HUSTR_10;
  mapnames2[10] := HUSTR_11;

  mapnames2[11] := HUSTR_12;
  mapnames2[12] := HUSTR_13;
  mapnames2[13] := HUSTR_14;
  mapnames2[14] := HUSTR_15;
  mapnames2[15] := HUSTR_16;
  mapnames2[16] := HUSTR_17;
  mapnames2[17] := HUSTR_18;
  mapnames2[18] := HUSTR_19;
  mapnames2[19] := HUSTR_20;

  mapnames2[20] := HUSTR_21;
  mapnames2[21] := HUSTR_22;
  mapnames2[22] := HUSTR_23;
  mapnames2[23] := HUSTR_24;
  mapnames2[24] := HUSTR_25;
  mapnames2[25] := HUSTR_26;
  mapnames2[26] := HUSTR_27;
  mapnames2[27] := HUSTR_28;
  mapnames2[28] := HUSTR_29;
  mapnames2[29] := HUSTR_30;
  mapnames2[30] := HUSTR_31;
  mapnames2[31] := HUSTR_32;

////////////////////////////////////////////////////////////////////////////////

// Plutonia WAD map names.

  mapnamesp[0] := PHUSTR_1;
  mapnamesp[1] := PHUSTR_2;
  mapnamesp[2] := PHUSTR_3;
  mapnamesp[3] := PHUSTR_4;
  mapnamesp[4] := PHUSTR_5;
  mapnamesp[5] := PHUSTR_6;
  mapnamesp[6] := PHUSTR_7;
  mapnamesp[7] := PHUSTR_8;
  mapnamesp[8] := PHUSTR_9;
  mapnamesp[9] := PHUSTR_10;
  mapnamesp[10] := PHUSTR_11;

  mapnamesp[11] := PHUSTR_12;
  mapnamesp[12] := PHUSTR_13;
  mapnamesp[13] := PHUSTR_14;
  mapnamesp[14] := PHUSTR_15;
  mapnamesp[15] := PHUSTR_16;
  mapnamesp[16] := PHUSTR_17;
  mapnamesp[17] := PHUSTR_18;
  mapnamesp[18] := PHUSTR_19;
  mapnamesp[19] := PHUSTR_20;

  mapnamesp[20] := PHUSTR_21;
  mapnamesp[21] := PHUSTR_22;
  mapnamesp[22] := PHUSTR_23;
  mapnamesp[23] := PHUSTR_24;
  mapnamesp[24] := PHUSTR_25;
  mapnamesp[25] := PHUSTR_26;
  mapnamesp[26] := PHUSTR_27;
  mapnamesp[27] := PHUSTR_28;
  mapnamesp[28] := PHUSTR_29;
  mapnamesp[29] := PHUSTR_30;
  mapnamesp[30] := PHUSTR_31;
  mapnamesp[31] := PHUSTR_32;

////////////////////////////////////////////////////////////////////////////////

// TNT WAD map names.

  mapnamest[0] := THUSTR_1;
  mapnamest[1] := THUSTR_2;
  mapnamest[2] := THUSTR_3;
  mapnamest[3] := THUSTR_4;
  mapnamest[4] := THUSTR_5;
  mapnamest[5] := THUSTR_6;
  mapnamest[6] := THUSTR_7;
  mapnamest[7] := THUSTR_8;
  mapnamest[8] := THUSTR_9;
  mapnamest[9] := THUSTR_10;
  mapnamest[10] := THUSTR_11;

  mapnamest[11] := THUSTR_12;
  mapnamest[12] := THUSTR_13;
  mapnamest[13] := THUSTR_14;
  mapnamest[14] := THUSTR_15;
  mapnamest[15] := THUSTR_16;
  mapnamest[16] := THUSTR_17;
  mapnamest[17] := THUSTR_18;
  mapnamest[18] := THUSTR_19;
  mapnamest[19] := THUSTR_20;

  mapnamest[20] := THUSTR_21;
  mapnamest[21] := THUSTR_22;
  mapnamest[22] := THUSTR_23;
  mapnamest[23] := THUSTR_24;
  mapnamest[24] := THUSTR_25;
  mapnamest[25] := THUSTR_26;
  mapnamest[26] := THUSTR_27;
  mapnamest[27] := THUSTR_28;
  mapnamest[28] := THUSTR_29;
  mapnamest[29] := THUSTR_30;
  mapnamest[30] := THUSTR_31;
  mapnamest[31] := THUSTR_32;

////////////////////////////////////////////////////////////////////////////////

  headsupactive := false;

  head := 0;
  tail := 0;

  shiftdown := false;
  altdown := false;

  destination_keys[0] := HUSTR_KEYGREEN;
  destination_keys[1] := HUSTR_KEYINDIGO;
  destination_keys[2] := HUSTR_KEYBROWN;
  destination_keys[3] := HUSTR_KEYRED;

  num_nobrainers := 0;


(* Unit r_draw initialization *)

  fuzzpos := 0;


(* Unit d_net initialization *)

  oldentertics := 0;
  frameon := 0;


(* Unit wi_stuff initialization *)

  wi_stuff_anims[0] := @epsd0animinfo;
  wi_stuff_anims[1] := @epsd1animinfo;
  wi_stuff_anims[2] := @epsd2animinfo;
  wi_stuff_anims[3] := nil;



(* Unit am_map initialization *)


////////////////////////////////////////////////////////////////////////////////

  player_arrow[0].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[0].a.y := 0;
  player_arrow[0].b.x := ((8 * PLAYERRADIUS) div 7);
  player_arrow[0].b.y := 0;

  player_arrow[1].a.x := ((8 * PLAYERRADIUS) div 7);
  player_arrow[1].a.y := 0;
  player_arrow[1].b.x := ((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 2;
  player_arrow[1].b.y := ((8 * PLAYERRADIUS) div 7) div 4;

  player_arrow[2].a.x := ((8 * PLAYERRADIUS) div 7);
  player_arrow[2].a.y := 0;
  player_arrow[2].b.x := ((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 2;
  player_arrow[2].b.y := -((8 * PLAYERRADIUS) div 7) div 4;

  player_arrow[3].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[3].a.y := 0;
  player_arrow[3].b.x := -((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[3].b.y := ((8 * PLAYERRADIUS) div 7) div 4;

  player_arrow[4].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[4].a.y := 0;
  player_arrow[4].b.x := -((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[4].b.y := -((8 * PLAYERRADIUS) div 7) div 4;

  player_arrow[5].a.x := -((8 * PLAYERRADIUS) div 7)+3*((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[5].a.y := 0;
  player_arrow[5].b.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[5].b.y := ((8 * PLAYERRADIUS) div 7) div 4;

  player_arrow[6].a.x := -((8 * PLAYERRADIUS) div 7)+3*((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[6].a.y := 0;
  player_arrow[6].b.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  player_arrow[6].b.y := -((8 * PLAYERRADIUS) div 7) div 4;

////////////////////////////////////////////////////////////////////////////////

  cheat_player_arrow[0].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[0].a.y := 0;
  cheat_player_arrow[0].b.x := ((8 * PLAYERRADIUS) div 7);
  cheat_player_arrow[0].b.y := 0;

  cheat_player_arrow[1].a.x := ((8 * PLAYERRADIUS) div 7);
  cheat_player_arrow[1].a.y := 0;
  cheat_player_arrow[1].b.x := ((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 2;
  cheat_player_arrow[1].b.y := ((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[2].a.x := ((8 * PLAYERRADIUS) div 7);
  cheat_player_arrow[2].a.y := 0;
  cheat_player_arrow[2].b.x := ((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 2;
  cheat_player_arrow[2].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[3].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[3].a.y := 0;
  cheat_player_arrow[3].b.x := -((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[3].b.y := ((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[4].a.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[4].a.y := 0;
  cheat_player_arrow[4].b.x := -((8 * PLAYERRADIUS) div 7)-((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[4].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[5].a.x := -((8 * PLAYERRADIUS) div 7)+3*((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[5].a.y := 0;
  cheat_player_arrow[5].b.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[5].b.y := ((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[6].a.x := -((8 * PLAYERRADIUS) div 7)+3*((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[6].a.y := 0;
  cheat_player_arrow[6].b.x := -((8 * PLAYERRADIUS) div 7)+((8 * PLAYERRADIUS) div 7) div 8;
  cheat_player_arrow[6].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[7].a.x := -((8 * PLAYERRADIUS) div 7) div 2;
  cheat_player_arrow[7].a.y := 0;
  cheat_player_arrow[7].b.x := -((8 * PLAYERRADIUS) div 7) div 2;
  cheat_player_arrow[7].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[8].a.x := -((8 * PLAYERRADIUS) div 7) div 2;
  cheat_player_arrow[8].a.y := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[8].b.x := -((8 * PLAYERRADIUS) div 7) div 2+((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[8].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[9].a.x := -((8 * PLAYERRADIUS) div 7) div 2+((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[9].a.y := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[9].b.x := -((8 * PLAYERRADIUS) div 7) div 2+((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[9].b.y := ((8 * PLAYERRADIUS) div 7) div 4;

  cheat_player_arrow[10].a.x := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[10].a.y := 0;
  cheat_player_arrow[10].b.x := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[10].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[11].a.x := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[11].a.y := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[11].b.x := 0;
  cheat_player_arrow[11].b.y := -((8 * PLAYERRADIUS) div 7) div 6;

  cheat_player_arrow[12].a.x := 0;
  cheat_player_arrow[12].a.y := -((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[12].b.x := 0;
  cheat_player_arrow[12].b.y := ((8 * PLAYERRADIUS) div 7) div 4;

  cheat_player_arrow[13].a.x := ((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[13].a.y := ((8 * PLAYERRADIUS) div 7) div 4;
  cheat_player_arrow[13].b.x := ((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[13].b.y := -((8 * PLAYERRADIUS) div 7) div 7;

  cheat_player_arrow[14].a.x := ((8 * PLAYERRADIUS) div 7) div 6;
  cheat_player_arrow[14].a.y := -((8 * PLAYERRADIUS) div 7) div 7;
  cheat_player_arrow[14].b.x := ((8 * PLAYERRADIUS) div 7) div 6+((8 * PLAYERRADIUS) div 7) div 32;
  cheat_player_arrow[14].b.y := -((8 * PLAYERRADIUS) div 7) div 7-((8 * PLAYERRADIUS) div 7) div 32;

  cheat_player_arrow[15].a.x := ((8 * PLAYERRADIUS) div 7) div 6+((8 * PLAYERRADIUS) div 7) div 32;
  cheat_player_arrow[15].a.y := -((8 * PLAYERRADIUS) div 7) div 7-((8 * PLAYERRADIUS) div 7) div 32;
  cheat_player_arrow[15].b.x := ((8 * PLAYERRADIUS) div 7) div 6+((8 * PLAYERRADIUS) div 7) div 10;
  cheat_player_arrow[15].b.y := -((8 * PLAYERRADIUS) div 7) div 7;

////////////////////////////////////////////////////////////////////////////////

  triangle_guy[0].a.x := round(-0.867 * FRACUNIT);
  triangle_guy[0].a.y := round(-0.5 * FRACUNIT);
  triangle_guy[0].b.x := round(0.867 * FRACUNIT);
  triangle_guy[0].b.y := round(-0.5 * FRACUNIT);

  triangle_guy[1].a.x := round(0.867 * FRACUNIT);
  triangle_guy[1].a.y := round(-0.5 * FRACUNIT);
  triangle_guy[1].b.x := 0;
  triangle_guy[1].b.y := FRACUNIT;

  triangle_guy[2].a.x := 0;
  triangle_guy[2].a.y := FRACUNIT;
  triangle_guy[2].b.x := round(-0.867 * FRACUNIT);
  triangle_guy[2].b.y := round(-0.5 * FRACUNIT);

////////////////////////////////////////////////////////////////////////////////
  thintriangle_guy[0].a.x := round(-0.5 * FRACUNIT);
  thintriangle_guy[0].a.y := round(-0.7 * FRACUNIT);
  thintriangle_guy[0].b.x := FRACUNIT;
  thintriangle_guy[0].b.y := 0;

  thintriangle_guy[1].a.x := FRACUNIT;
  thintriangle_guy[1].a.y := 0;
  thintriangle_guy[1].b.x := round(- 0.5 * FRACUNIT);
  thintriangle_guy[1].b.y := round(- 0.7 * FRACUNIT);

  thintriangle_guy[2].a.x := round(- 0.5 * FRACUNIT);
  thintriangle_guy[2].a.y := round(- 0.7 * FRACUNIT);
  thintriangle_guy[2].b.x := round(- 0.5 * FRACUNIT);
  thintriangle_guy[2].b.y := round(- 0.7 * FRACUNIT);

////////////////////////////////////////////////////////////////////////////////
  cheat_amap.sequence := get_cheatseq_string(cheat_amap_seq);
  cheat_amap.p := get_cheatseq_string(0);

////////////////////////////////////////////////////////////////////////////////
  ZeroMemory(st_notify_AM_initVariables, SizeOf(st_notify_AM_initVariables));
  st_notify_AM_initVariables._type := ev_keyup;
  st_notify_AM_initVariables.data1 := AM_MSGENTERED;

////////////////////////////////////////////////////////////////////////////////
  ZeroMemory(st_notify_AM_Stop, SizeOf(st_notify_AM_Stop));
  st_notify_AM_Stop._type := ev_keyup;
  st_notify_AM_Stop.data1 := AM_MSGEXITED;


(* Unit d_main initialization *)


  for i := 0 to MAXWADFILES - 1 do
    wadfiles[i] := '';

  wipegamestate := Ord(GS_DEMOSCREEN);


(* Unit g_game initialization *)


  forwardmove[0] := $19;
  forwardmove[1] := $32;
  sidemove[0] := $18;
  sidemove[1] := $28;
  angleturn[0] := 640;
  angleturn[1] := 1280;
  angleturn[2] := 320;

  mousebuttons := PBooleanArray(@mousearray[1]);
  joybuttons := PBooleanArray(@joyarray[1]);


  ZeroMemory(pars, SizeOf(pars));

  pars[1, 1] := 30;
  pars[1, 2] := 75;
  pars[1, 3] := 120;
  pars[1, 4] := 90;
  pars[1, 5] := 165;
  pars[1, 6] := 180;
  pars[1, 7] := 180;
  pars[1, 8] := 30;
  pars[1, 9] := 165;

  pars[2, 1] := 90;
  pars[2, 2] := 90;
  pars[2, 3] := 90;
  pars[2, 4] := 120;
  pars[2, 5] := 90;
  pars[2, 6] := 360;
  pars[2, 7] := 240;
  pars[2, 8] := 30;
  pars[2, 9] := 170;

  pars[3, 1] := 90;
  pars[3, 2] := 45;
  pars[3, 3] := 90;
  pars[3, 4] := 150;
  pars[3, 5] := 90;
  pars[3, 6] := 90;
  pars[3, 7] := 165;
  pars[3, 8] := 30;
  pars[3, 9] := 135;


  ZeroMemory(cpars, SizeOf(cpars));

  cpars[0] := 30;
  cpars[1] := 90;
  cpars[2] := 120;
  cpars[3] := 120;
  cpars[4] := 90;
  cpars[5] := 150;
  cpars[6] := 120;
  cpars[7] := 120;
  cpars[8] := 270;
  cpars[9] := 90;
  cpars[10] := 210;
  cpars[11] := 150;
  cpars[12] := 150;
  cpars[13] := 150;
  cpars[14] := 210;
  cpars[15] := 150;
  cpars[16] := 420;
  cpars[17] := 150;
  cpars[18] := 210;
  cpars[19] := 150;
  cpars[20] := 240;
  cpars[21] := 150;
  cpars[22] := 180;
  cpars[23] := 150;
  cpars[24] := 150;
  cpars[25] := 300;
  cpars[26] := 330;
  cpars[27] := 420;
  cpars[28] := 300;
  cpars[29] := 180;
  cpars[30] := 120;
  cpars[31] := 30;

  precache := true;


(* Unit i_net initialization *)

  DOOMPORT := (IPPORT_USERRESERVED + 29);
  mysocket := INVALID_SOCKET;




finalization

(* Unit i_midi finalization *)

  StopPlaying;
  MidiFileNames.Free;


(* Unit m_argv finalization *)

{$IFDEF DEBUG}
  cmdparams.SaveToFile('cmdparams.txt');
  cmdparams.Free;
{$ENDIF}




end.
